/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1999 Leverage 
 * Information Systems. All Rights reserved.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#include <stdio.h>
#include <sys/socket.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/types.h>

#define LOCO_SCAN_LINEAR 1
#define LOCO_SCAN_BALANCE 2
#define LOCO_MAXLINE 255
#define LOCO_MAX_DESTINATIONS 32 
#define LOCO_FILEBUFSIZE 4096
#define LOCO_NETBUFSIZE 32768  /* used for readback too */
#define LOCO_PROTOCOL_VERSION 2  /* We can use the same version as CGI here */
#define LOCO_MAX_COOKIE_NAME_LEN 255

/* change this to one to get debugging statemtents */
#define LOCO_DEBUG -1

struct loco_location
{
    unsigned long address;     /* ip address in network order */
    int port;       /* tcp port number */
    unsigned int weight;       /* weight used in random pick calculation */
    unsigned int inst_id;      /* distinguish this dest. from others in pool */
};

struct loco_location_config 
{
    int style;                      /* SCAN_LINEAR or SCAN_BALANCE */
    int location_count;             /* up to LOCO_MAX_DESTINATIONS */
    char vs_cookie_name[LOCO_MAX_COOKIE_NAME_LEN];
                                    /* e.g., "LOCO_INST_ID_LOCO_ORG_SYSTAG=" */
    struct loco_location *loc_list; /* location_count structures */
};

typedef struct loco_module_config
{
    char *this_request_config_file_path;
    struct loco_location_config *llc;
    time_t previous_update_time;
    time_t config_file_previous_mod_time;
    pool *configpool;
    char *locolog_filename;

} loco_module_config;

/* declare our module, then define all functions below */
module MODULE_VAR_EXPORT locolink_module;

/* declare main handler function.  This is what actually handles the request */
static int locolink_response_handler (request_rec *);

static void senderrormsg (request_rec *r, char *msg);
static struct loco_location_config *get_loco_location_config (request_rec *r);
static int load_locations (request_rec *r, loco_module_config *dcfg);
static int parse_cfgfile (request_rec *r, loco_module_config *dcfg, FILE *fp);
static int get_connection (request_rec *r, struct loco_location_config *llc);
static void set_locolog(cmd_parms *cmd, void *mconfig, char *arg);
static void locolog(request_rec *r, char *logstr);
static void locolog_inner(char *logstr);
static int read_scan_type (request_rec *r, struct loco_location_config *llc, 
			   FILE *fp);
static int read_cookie_name(request_rec *r, struct loco_location_config *llc, 
			    FILE *fp); 
static int read_locations (request_rec *r, struct loco_location_config *llc, 
			   FILE *fp);
static int read_line_cfg (request_rec *r, char linebuf[], FILE *fp);
static char *first_token (char linebuf[]);
static char *next_token ();
static int get_sock (request_rec *r, struct loco_location_config *llc, 
		     int cursor);
static int try_virtual_session_connect(request_rec *r, 
				       struct loco_location_config *llc);
static int get_loco_inst_id (request_rec *r, struct loco_location_config *llc);

/*
 * This method log errors that occured in the module in the standard 
 * error log file for apache. 
 */
static void logerror(request_rec *r, char *logstr, ...)
{
  va_list args;
  char *buf;

  va_start(args, logstr);

  buf =  ap_pvsprintf (r->pool, logstr, args);
  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "mod_locolink: %s", 
		buf);

  va_end(args);
}




static void locolog (request_rec *r, char *logstr)
{
    if (LOCO_DEBUG > 0)
    {
	FILE *fp;
	pid_t pid;
	loco_module_config *conf;
	
	conf = (loco_module_config *) 
	    ap_get_module_config (r->per_dir_config, &locolink_module);
	
	if (conf->locolog_filename && 
	    ( fp = ap_pfopen( r->pool, 
			      ap_server_root_relative(r->pool, 
						      conf->locolog_filename), 
			      "a")) != NULL )
	{
	    pid = getpid();
	    fprintf (fp, "%d - %s\n", pid, logstr);
	    ap_pfclose(r->pool, fp);
	}
    } 	
}

/* this is only for internal debugging- remove it as soon as we can... */
static void locolog_inner(char *logstr)
{
    if (LOCO_DEBUG > 0)
    {
	FILE *fp;
	pid_t pid;
	
	if ( (fp = fopen("/usr/local/apache/logs/locoinner.log", "a")) != NULL )
	{
	    pid = getpid();
	    fprintf (fp, "%d - %s\n", pid, logstr);
	    fclose(fp);
	}
    }
}    


static void *locolink_create_dir_config (pool *p, char *dirspec)
{
    loco_module_config *loco_mconf;

    loco_mconf = (loco_module_config *) 
	ap_pcalloc (p, sizeof (loco_module_config));

    loco_mconf->this_request_config_file_path = NULL;
    
    loco_mconf->llc = (struct loco_location_config *)
	ap_pcalloc (p, sizeof (struct loco_location_config));
    
    loco_mconf->llc->loc_list = (struct loco_location *) 
	ap_pcalloc (p, LOCO_MAX_DESTINATIONS * sizeof (struct loco_location));

    loco_mconf->previous_update_time = 0;
    loco_mconf->config_file_previous_mod_time = 0;

    locolog_inner ("dir config 1 created");

    return (loco_mconf);
}

/* called to pull the pathname out of the config file and 
 * store it where it can easily be accessed by the module 
 */
static const char *set_loco_locations_config_file (cmd_parms *cmd, 
						   void *mconfig, 
						   char *arg)
{
    loco_module_config *loco_config = (loco_module_config *) mconfig;

    loco_config->this_request_config_file_path = arg;

    locolog_inner ("path set in config struct");
    return (NULL);
}


/* sets the locolog location to whatever the LocoLogFile config is set to */
static const char *set_locolog_file(cmd_parms *cmd, void *mconfig, char *arg)
{
    server_rec *s = cmd->server;

    loco_module_config *conf = (loco_module_config *) mconfig;

    conf->locolog_filename = arg;
    return (NULL);
}  


static const command_rec locolink_cmds[] =
{
    {"LocoLocationsFile", set_loco_locations_config_file, NULL, ACCESS_CONF|RSRC_CONF, 
     TAKE1, "Absolute path to the Locomotive Tunnel configuration file"}, 
    {"LocoLogFile", set_locolog_file, NULL, ACCESS_CONF|RSRC_CONF, TAKE1, 
     "The log to use for debugging statements"},
    {NULL}
};

/* the handler.  This is the function called every hit by the apache core.
 * The function pointer is listed in the module table (below).  */
static int locolink_response_handler (request_rec *r)
{
    loco_module_config *pcfg;
    int lfd;
    BUFF *locobuf;
    struct loco_location_config *llc;
    int i;

    locolog (r, "Made it to the response_handler 1");

    if ((llc = get_loco_location_config (r)) == NULL)
    {
	senderrormsg (r, "Config load failed");
	return OK;
    }

    locolog (r, "Got loco_location_config");

    if ((lfd = get_connection (r, llc)) < 0)
    {
	/* get_connection handles its own errors */
	return OK;
    }

    locolog (r, "Got a connection");

    locobuf = ap_bcreate (r->pool, B_RDWR |B_SOCKET); 
    ap_bpushfd (locobuf, lfd, lfd);

    if (locobuf == NULL)
    {
	ap_pclosesocket (r->pool, lfd);
	senderrormsg (r, "send envs failed");
	return OK;
    }

    locolog (r, "Sending envs");

    if (send_envs (locobuf, r) != 0)
    {
	ap_bclose(locobuf);
	ap_pclosesocket (r->pool, lfd);
	senderrormsg (r, "send envs failed");
	return OK;
    }

    locolog (r, "Writing stdin");

    if (write_stdin (locobuf, r) != 0)
    {
	ap_bclose(locobuf);
	ap_pclosesocket (r->pool, lfd);
	senderrormsg (r, "write stdin failed");
	return OK;
    }

    locolog (r, "Receiving response");

    if (read_and_write_response (locobuf, r) != 0)
    {
	senderrormsg (r, "write loco response failed");
    }
    
    locolog (r, "Finishing request");

    ap_bclose(locobuf);
    ap_pclosesocket(r->pool, lfd);
    locolog (r, "through close");
    return OK;
}

/* Sends the protocol version to the locomotive, along with the
   appropriate name value pairs from the server internals and
   http headers */
int send_envs (BUFF *locobuf, request_rec *r)
{
    int written;
    unsigned int version;
    char **names;
    char **values;
    char port_str[10];
    char *tmp_value;
    int nv_array_length = 32;  /* this number is manually allocated */
    int nv_array_cursor = 0;
    int i = 0;
    char **env; 
    array_header *hdrs_arr;
    table_entry *hdrs;
    
    ap_add_common_vars(r);
    ap_add_cgi_vars(r);
    env = ap_create_environment (r->pool, r->subprocess_env);
    
    /* write version to buffer as first int */
    version = htonl ((unsigned int) LOCO_PROTOCOL_VERSION);
    locolog (r, "writing version");
    ap_bwrite (locobuf, &version, sizeof (unsigned int));


    locolog (r, "Environment follows");

    /* Ok, let's write the CGI vars to the buff  */
    for (i = 0; env[i]; i++)
    {
	unsigned int length = 0;
	unsigned int net_len = 0;
	char *foo;
	length = strlen(env[i]);
	net_len = htonl(length);

	foo = ap_psprintf (r->pool, "env: %s len: %d", env[i], length);
	locolog(r, foo);

	ap_bwrite (locobuf, &net_len, sizeof(unsigned int));
	ap_bwrite (locobuf, env[i], length);
    }
    
    locolog (r, "before final flush");

    /* write the 4 byte zero terminator */
    written = 0;
    if ((written = ap_bwrite (locobuf, &written, sizeof(int))) == -1)
    {
        return (-1);
    }
    
    ap_bflush(locobuf);

    locolog (r, "after flush");

    return (0); /* everything went through */
}

/* check for any stdin.  If there is any, forward through to
   the locomotive.  returns -1 on error, otherwise the length of the
   stdin that was written, or 0 if none */
int write_stdin (BUFF *locobuf , request_rec *r)
{
    int ret;
    int numread = 0;
    char postbuf[HUGE_STRING_LEN];

    if ((ret = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR)))
    {
	return (ret);
    }

    /* If we have no POST, then return */
    if (ap_should_client_block(r))
    {
	ap_soft_timeout ("transfer stdin to loco", r); 

	while ((numread = ap_get_client_block(r, postbuf, HUGE_STRING_LEN)) >0)
	{
	    ap_reset_timeout(r);

	    locolog(r, "writing stdin");
	    if (ap_bwrite(locobuf, postbuf, numread) < numread)
	    {
		/* troubles- dump the rest */
		logerror(r, "dumping stdin");
		while ( ap_get_client_block(r, postbuf, HUGE_STRING_LEN) > 0);
		return (-1);
	    }
	}
	
	locolog(r, "flushing stdin");
	ap_bflush(locobuf);
    }
	
    ap_kill_timeout(r);

    return (0);
}


/* Here we get the response from the tunnel and write it out.  We'll
   redirect internally if we find a 'Location: ' header */
int read_and_write_response (BUFF *locobuf, request_rec *r)
{
    char respbuf[HUGE_STRING_LEN];
    int ret;
    
    if ((ret = ap_scan_script_header_err_buff(r, locobuf, respbuf)))
    {
	ap_kill_timeout(r);
	return (ret);
    }

    /* handle the 'Location' fun */
    if (ap_table_get(r->headers_out, "Location") && r->status == 200)
    {
	ap_kill_timeout(r);
	return (REDIRECT);
    }

    locolog(r, "sending headers");
    ap_send_http_header(r);
    if (! r->header_only)
    {
	locolog(r, "sending content");
	ap_send_fb(locobuf, r);
	locolog(r, "through content");
    }

    ap_kill_timeout(r);
    return (0);
}
    


/* opens up a connection to a locomotive.  The location and scan pattern
   are stored in llc */
static int get_connection (request_rec *r, struct loco_location_config *llc)
{
    int lsock = 0; 
    unsigned int cursor;
    int count = 0;
    int lid = 0;

    if (llc->location_count < 1)
    {
        /* no locations! */
        logerror (r, "get_connection(): no locations");
        return (-1);
    }

    /* If the browser already knows what server it wants, then return a 
     * connection to that server */
    lsock = try_virtual_session_connect(r, llc);
    if (lsock >= 0)
	return lsock; 

    locolog (r, "No virtual session connection; proceeding with normal "
	     "connection");

    /* pick our starting position in the list */
    if (llc->style == LOCO_SCAN_LINEAR)
    {
        cursor = 0;
    }
    else  /* start at a weighted random position in the list */
    {
        int mod = 0;
        int sum = 0;
        struct timeval tp;

        for (cursor = 0; cursor < llc->location_count; cursor++)
        {
            sum += llc->loc_list[cursor].weight;
        }

        /* get current time */
        gettimeofday (&tp, NULL);

        /* initialize quasi-random number generator.  Won't be very random
           at certain slow hit rates.  Perhaps use a millisec timer too? */
        srand ((int) (tp.tv_sec + tp.tv_usec));

        mod = rand () % sum;

        sum = 0;
        for (cursor = 0; cursor < llc->location_count; cursor++)
        {
            sum += llc->loc_list[cursor].weight;
            if (mod < sum) { break; } /* will get called on last loop, but will
                                         always be true.  This is an easy way
                                         to get the cursor incremented */
            /* now cursor is at the correct starting point in the list */
        }
    }

    /* snarfed from original "locolink_weight.c", then NSAPI */
    /* enter loop, starting at destination pointed to by cursor */
    for (count = 0; count < llc->location_count; count++)
    {
	lsock = get_sock(r, llc, cursor);

	if (lsock >= 0)      /* found an available loco */
	    return lsock;
	
	if (lsock == -1)     /* can't open a socket! */
	    return -1;

        /* check if we need to reset the counter to the bottom or increment */
        if (cursor >= (llc->location_count - 1))
            cursor = 0;
        else
            cursor++;
    }

    /* none of the destinations worked!  */
    logerror (r, "get_connection(): no working destinations");
    senderrormsg (r, "connection failure");
    return -1;
}

/* returns a connected socket on success, -1 on fatal error (could not open
 * a socket), -2 on non-fatal error (could not connect to a locomotive) */
static int get_sock (request_rec *r, struct loco_location_config *llc, 
		     int cursor)
{
    struct sockaddr_in sa;
    char *tmp;
    int lsock;
    
    sa.sin_family = AF_INET;
    sa.sin_port = htons(llc->loc_list[cursor].port);
    sa.sin_addr.s_addr = llc->loc_list[cursor].address;
    
    /* get a socket */
    if ((lsock = ap_psocket(r->pool, AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
    {
	logerror (r, "get_connection(): net_socket failed");
	senderrormsg (r, "create socket failed");
	return -1;
    }

    /* connect to Application Server.  maybe we should do a select instead ? */
    if (connect (lsock, (struct sockaddr *) &sa,
		 sizeof (struct sockaddr_in)) == 0)
    {
	return lsock;
    }

    /* log the error */
    tmp = ap_psprintf (r->pool,
		       "get_connection(): connect to %s failed: %s", 
		       inet_ntoa (sa.sin_addr), strerror(errno));
    logerror (r, tmp);
    
    ap_pclosesocket (r->pool, lsock);
    return -2;
}

/* Looks at the cookie and finds the instance ID.  If it's a valid ID, we try 
 * to connect to the indicated server.  Returns a nonnegative port no. on 
 * success; -1 on failure 
 */
static int try_virtual_session_connect(request_rec *r, 
				       struct loco_location_config *llc)
{
    int inst_id;
    int cursor;
    int lsock;
    char tmp[255];

    inst_id = get_loco_inst_id (r, llc);

    sprintf (tmp, "Examining loco_inst_id=%d", inst_id);
    locolog (r, tmp);

    if (inst_id <= 0)
	return -1;

    for (cursor = 0; cursor < llc->location_count; cursor++) {
	if (inst_id == llc->loc_list[cursor].inst_id) {

	    /* try to get a connection to the (first) matching instance */
	    lsock = get_sock (r, llc, cursor);
	    if (lsock >= 0)
		return lsock;
	    else
		return -1;
	}
    }

    return -1;
}

/* Return an integer representing the instance ID in the cookie. 
 * Return value is 0 if no valid instance ID was found. 
 * Code taken from src/c/locolink/cgi/locolink.c. 
 */
static int get_loco_inst_id (request_rec *r, struct loco_location_config *llc)
{
    char *cookie;

    cookie = ap_pstrdup (r->pool, ap_table_get (r->headers_in, "Cookie"));

    if (cookie == NULL) {
	locolog (r, "Null cookie");
	return 0;
    }

    locolog (r, "Fetched cookie:");
    locolog (r, cookie);
    
    /* At this point, the value of cookie should look something like this:
     * "COOKIE_VAL_A=foo ; {llc->vs_cookie_name}2; COOKIE_VAL_C=bar"
     *
     * We'll advance the pointer to end of "{llc->vs_cookie_name}" and call 
     * atoi().  atoi() will stop as soon as it sees a non-whitespace character
     * that doesn't look like a decimal integer.  Leading whitespace is 
     * ignored.
     */

    if((cookie = strstr (cookie, llc->vs_cookie_name)) == NULL)
	return 0;

    /* If we got this far, we have a pointer to the beginning of vs_cookie_name
     * in the cookie string.  Now we advance the pointer to the first 
     * character after vs_cookie_name, which could possibly be the the string-
     * terminating NULL byte or some other piece of non-integer cruft.  atoi() 
     * returns 0 for such garbage.
     */

    cookie += strlen(llc->vs_cookie_name);

    return (atoi(cookie));
}

/* returns the reference to the in-memory copy of the location config.  
 * If enough time has elapsed (one second), it will re-fstat the config file
 * and if it has changed, it will reparse it and update the in-memory copy. 
 */
static struct loco_location_config *get_loco_location_config (request_rec *r)
{
    loco_module_config *dcfg;
    time_t current_second;

    dcfg = (loco_module_config *) ap_get_module_config (r->per_dir_config,
							&locolink_module);
    current_second = time (NULL);

    /* boy, this is a hell of alot easier in here than the nsapi version!
     * We don't have to mutex, nor make thread specific copies.  yeah!
     * on the other hand, we stat the file a maximum of once per second 
     * per child httpd process, rather than once per sec period.  
     * lets assume that you have 32 child processes, and you were receiving
     * requests for 100 dynamic page view per second.  You'd top out at 
     * a maximum of 32 fstats / sec.  How bad is this?  Hardly noticable.
     * A 233MHz pentium II running FreeBSD 2.2.7 can do about 35000 per 
     * second on a local file.
     */
    if (current_second > dcfg->previous_update_time)
    {
	int num_locations = 0;

	locolog (r, "made it to get_loco_location_config, calling "
		 "load_locations");
	
	num_locations = load_locations (r, dcfg);
	if (num_locations < 1) /* must'a been a parse error or fnf */ 
	    return NULL;
	
	dcfg->previous_update_time = current_second;
	locolog (r, "get_loco_location_config: finished load_locations");

	return (dcfg->llc);
    }
    else
    {
	return (dcfg->llc);
    }
}

static int load_locations (request_rec *r, loco_module_config *dcfg)
{
    struct stat file_stat; /* automatic garbage collection on the stack*/
    int parse_return_val;
    FILE *fp;

    if (dcfg->this_request_config_file_path == NULL)
    {
	/* serious error - no config file specified */
	logerror (r, "load_locations: missing config file path!");
	return (-1);
    }

    if ((stat (dcfg->this_request_config_file_path, &file_stat)) != 0)
    {
	logerror (r, "load_locations: fstat failed for file = %s", 
		  dcfg->this_request_config_file_path);
	/* fstat failed - file missing  ? */
	return (-1);
    }
    
    /* if the file on disk hasn't been changed, skip parsing and simply 
       return the count in the resident copy */
    if (file_stat.st_mtime <= dcfg->config_file_previous_mod_time)
    {
	return (dcfg->llc->location_count);
    }

    /* ok, must parse the file.  open it. */
    if ((fp = ap_pfopen (r->pool, dcfg->this_request_config_file_path, "r"))
	== NULL)
    {
	logerror (r, "load_locations: fopen config file failed for file = %s", 
		  dcfg->this_request_config_file_path);
	return (-1);
    }
    
    if ((parse_return_val = parse_cfgfile (r, dcfg, fp)) < 1)
    {
	ap_pfclose (r->pool, fp);
	logerror (r, "load_locations: parse cfg file failed");
	return (-1);
    }
    
    ap_pfclose (r->pool, fp);
    /* update our cached notion of the mod time */
    dcfg->config_file_previous_mod_time = file_stat.st_mtime;
    return (parse_return_val);
}

/* 
 * takes the input FILE and scans it, loading locations and scan-style
 * into dcfg->llc.  returns the number of locations found in the file.
 */
static int parse_cfgfile (request_rec *r, loco_module_config *dcfg, FILE *fp)
{
    struct loco_location_config *llc;
    llc = dcfg->llc;

    /* Read the scan type line */
    if (read_scan_type (r, llc, fp) < 0)
      return -1;

    /* Read the cookie name line */
    if (read_cookie_name (r, llc, fp) < 0)
      return -1;

    /* Read the locations */
    return read_locations (r, llc, fp);
}

/* Read the scan type from the config file and store in llc->style.
 * Returns negative if error; otherwise zero.
 */
static int read_scan_type (request_rec *r, struct loco_location_config *llc, 
			   FILE *fp)
{
    char linebuf[LOCO_MAXLINE];

    /* There better be a first line... */
    if (read_line_cfg (r, linebuf, fp) < 0)
      return -1;

    /* parse the scan type */
    /* note we use strncmp, which may not be valid on platforms other than
     * unix.  For whatever reason at the time of writing, the
     * util_strncmp provided by netscape didn't want to be found.
     * Maybe we need to have an extra define somewhere?   
     */
    if (strlen (linebuf) >= 6  &&  !strncmp ("linear", linebuf, 6)) 
    {
	llc->style = LOCO_SCAN_LINEAR;
	locolog (r, "locolink: config: linear scantype");
    }
    else if (strlen (linebuf) >= 7 &&  !strncmp ("balance", linebuf, 7)) 
    {
        llc->style = LOCO_SCAN_BALANCE;
        locolog (r, "locolink: config: balance scantype");
    }
    else 
    {
        logerror (r, "config file: invalid scantype = %s", linebuf);
        /* style listed is bogus */
        return -1;
    }
    return 0;
}

/* Reads the virtual sessions cookie name and returns 0 if successful, negative
 * if error.
 */
static int read_cookie_name (request_rec *r, struct loco_location_config *llc, 
			     FILE *fp) 
{
    char linebuf[LOCO_MAXLINE];
    int i;
    int c = 0;
    int len = 0;
    char tmp[255];

    if (read_line_cfg (r, linebuf, fp) < 0)
      return -1;

    /* parse the cookie name */
    /* We will just check that the string consists of some non-whitespace 
     * characters and that the last character is an equals sign ("=").  
     * Furthermore, we'll barf if there are any commas.  This will cause a 
     * meaningful error to be thrown if the cookie line is omitted.
     *
     * If everything is okay, return 0, else -1.  
     */

    /* Here strtok() has the effect of truncating linebuf at the first newline.
     */
    strtok (linebuf, "\n");
    len = strlen (linebuf);

    /* The string must have at least one character in addition to the "=" */
    if (len < 2) 
    {
        logerror (r, "config file: empty cookie name");
        return -1;
    }

    /* The string must not be too long */
    if (len > LOCO_MAX_COOKIE_NAME_LEN) 
    {
        logerror (r, "config file: cookie name line too long");
        return -1;
    }

    /* The string must not contain whitespace or commas */
    for (i = 0; (c = (int)linebuf[i]) != '\0'; ++i) 
    {
	if (c == ' ' || c == '\t') 
	{
	    logerror (r, "config file: illegal whitespace in "
		     "cookie name");
	    return -1;
	}
	if (c == ',') 
	{
	    logerror (r, "config file: illegal comma in cookie name. " 
		      "second line must contain cookie name for virtual "
		      "sessions");
	    return -1;
	}
	if (c == '=') 
	  break;
    }

    /* The last character examined must be an equal sign */
    if (c != '=') 
    {
	logerror (r, "config file: last char of cookie name "
		  "must be '='");
	return -1;
    }

    strncpy(llc->vs_cookie_name, linebuf, LOCO_MAX_COOKIE_NAME_LEN);
    sprintf(tmp, "locolink: config: got cookie name: %s", llc->vs_cookie_name);
    locolog (r, tmp);	 

    return 0;
}

/* Read locations from config file fp.  Returns the number of locations found,
 * or -1 on error.
 */ 
static int read_locations (request_rec *r, struct loco_location_config *llc, 
			   FILE *fp) 
{
    char linebuf[LOCO_MAXLINE];
    unsigned long address = 0;
    int port = 0;
    unsigned int weight = 0;
    unsigned int inst_id = 0;
    struct in_addr s_inaddr;
    char tmp[255];
    char *token;

    /* just in case it wasn't already zero */
    llc->location_count = 0;

    while (read_line_cfg (r, linebuf, fp) >= 0 ) 
    {
	address = 0;
	port = 0;
	weight = 0;
	inst_id = 0;

	/*** Get address ***/
	token = first_token (linebuf);
	if (token == NULL) 
	{
	    logerror (r, "config: no address given");
	    return -1;
	} 
	address = inet_addr (token);
	if (address == -1) 
	{
	    logerror (r, "config file: bogus net address = %s",
		    token);
	    return -1;
	} 

	/*** Get port ***/
	token = next_token ();
	if (token == NULL) 
	{
	    logerror (r, "config: no port given for line = %s", linebuf);
	    return -1;
	}
	port = atoi (token);
	if (port < 1 || port > 65535) 
	{
	    /* port number is bogus */
	    logerror (r, "config: bogus port number = %s", token);
	    return -1;
	}

	/*** Get weight ***/
	token = next_token ();
	if (token == NULL) 
	{
	    logerror (r, "config: no weight given");
	    return -1;
	}
	weight = atoi (token);
	if (weight < 1) 
	{
	    /* weight val is bogus */
	    logerror (r, "config: invalid weight value = %s", token);
	    return -1;
	}

	/*** Get inst ID ***/
	token = next_token ();
	if (token == NULL) 
	{
	    logerror (r, "config: no instance ID given");
	    return -1;
	}
	inst_id = atoi (token);
	if (inst_id < 1) 
	{
	    logerror (r, "config: invalid inst_id value = %s", token);
	    return -1;
	}
	
	/* if we got this far, set the variables and write log */
	llc->loc_list[llc->location_count].address = address;
	llc->loc_list[llc->location_count].port = port;
	llc->loc_list[llc->location_count].weight = weight;
	llc->loc_list[llc->location_count].inst_id = inst_id;
	llc->location_count++;
	
	s_inaddr.s_addr = address;
	sprintf(tmp, "locolink: config: scanned location %s,%d,%d,%d", 
		inet_ntoa (s_inaddr), port, weight, inst_id);
	locolog (r, tmp);	    
    }

    /* There should never be fewer than one location in the config file */
    if (llc->location_count < 1) {
        logerror (r, "config: read error or zero location count");
        return -1;
    }

    /* return the number of new locations found */    
    return llc->location_count;
}

/* Reads a line from file fp and writes it into linebuf.  If there is
 * a failure or EOF, return -1.  Otherwise return 0.
 */
static int read_line_cfg (request_rec *r, char linebuf[], FILE *fp)
{
    if ((fgets (linebuf, LOCO_MAXLINE - 1, fp)) == NULL)
      return -1;
    else
      return 0;
}

/* Initialize the string tokenizer and get the first token */
static char *first_token (char linebuf[]) 
{
    return strtok (linebuf, ",");
}

/* Get next token from the string we provided as an argument to first_token() 
 */
static char *next_token () 
{
    return strtok (NULL, ",");
}


static void senderrormsg (request_rec *r, char *msg)
{
    r->content_type = "text/html";
    
    ap_send_http_header(r);
    
    ap_rputs ("<HTML><HEAD>Locolink - Error</HEAD>\n<BODY>\n", r);
    ap_rputs ("<H2>Locolink Failure: </H2><BR>\n", r);
    ap_rputs ("Error: ", r);
    ap_rputs (msg, r); 
    ap_rputs ("<BR>\n", r);
    ap_rputs ("</BODY></HTML>", r);
    
    ap_kill_timeout (r);
}

static const handler_rec locolink_handlers[] =
{
    {"locolink-tunnel", locolink_response_handler},
    {"application/locomotive", locolink_response_handler},
    {NULL}
};


module locolink_module = 
{
    STANDARD_MODULE_STUFF,
    NULL,                       /* initializer */
    locolink_create_dir_config, /* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    NULL,                       /* server config */
    NULL,                       /* merge server config */
    locolink_cmds,              /* command table */
    locolink_handlers,          /* handlers */
    NULL,                       /* filename translation */
    NULL,                       /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    NULL,                       /* child_init */
    NULL,                       /* child_exit */
    NULL                        /* post read-request */
};
