/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1999 Leverage 
 * Information Systems. All Rights reserved.
 *
 *
 * locolink.c
 * Passes CGI environmental variables and 
 * stdin to the Leverage Locomotive via a simple 
 * protocol.  Reads the response and writes it back to STDOUT, which is  
 * tied into the webserver.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>

#define BUF_LENGTH 32768

int get_connection ();
void exit_with_error(char *msg);
int send_envs (int s, char *buf);
int write_stdin (int s, char *buf);
int read_and_write_response (int s, char *buf);
int write_fully (int s, char *buf, int buflen);

/* To implement virtual sessions, we're going to add a fourth column to the
 * destinations matrix: instance ID.  The loco.conf entry LOCO_INST_ID must 
 * match the value entered here.  This value will be stored in the browser 
 * cookie to identify the instance to which the user has been assigned.
 *
 * Since the destinations array does not change, I changed the 
 * destination_count variable to a preprocessor variable.  The compiler 
 * should issue a warning if the destination count is less than the number 
 * of elements in the destinations array initializer.
 *
 * With the following definitions, we can access the columns of the 
 * destinations matrix by name.  For example, the the weight of the first 
 * destination is destinations[0][L_WEIGHT].  -- DAS
 */

#define L_PORT 0
#define L_ADDR 1
#define L_WEIGHT 2
#define L_INST_ID 3

/* IMPORTANT: DEST_COLS is the number of columns in the destinations array */
#define DEST_COLS 4
 
/**************************/
/* CONFIGURATION SECTION  */
/**************************/

/* In order for virtual sessions to work, provide the name of the virtual 
 * sessions cookie.  The name consists of the string "LOCO_INST_ID_", 
 * concatenated with the SYSTEM_TAG name as configured in loco.conf, 
 * concatenated with the string "=".  For example, if SYSTEM_TAG is 
 * "LOCO_ORG_SYSTAG", then the cookie name is "LOCO_INST_ID_LOCO_ORG_SYSTAG=".
 *
 * This is not an elegant solution, but it'll do for the first VS implementation.
 */
#define VS_COOKIE_NAME "LOCO_INST_ID_LOCO_ORG_SYSTAG="

/* Declare the number of server destinations and enter them:
 * portnum, expanded address, weight, instance ID. 
 * Note: 2130706433 is 127.0.0.1 expanded 
 */
#define DESTINATION_COUNT 1
static unsigned int destinations[DESTINATION_COUNT][DEST_COLS] = 
{ 
  {5421, 2130706433, 2, 1}
};
/* PORT, ADDRESS,   WT,ID */  
/**************************/

extern char **environ;

int main (int argc, char **argv)
{
    int s = 0;
    char *buf;

    /* get a connection */
    if ((s = get_connection()) < 0)
    {
	exit_with_error("connect_failed");
    }

    /* allocate a buffer */
    if ((buf = malloc (BUF_LENGTH * sizeof (char))) == NULL)
    {
	close (s);
	exit_with_error("transfer_buffer_malloc_failed");
    }

    /* set timer for death - don't want to fill up the system with a 
     * million of these things if something goes wrong */
    alarm (60);

    /* send protocol */
    /* amass envs in databuffer, send them */
    if (send_envs (s, buf) != 0)
    {
	close (s);
	exit_with_error ("send_env_failed");
    }

    /* write the remaining stdin (if any) */
    if (write_stdin (s, buf) != 0)
    {
	close (s);
	exit_with_error ("write_stdin_failed");
    }

    /* read response, write to stdout */
    if (read_and_write_response (s, buf) != 0)
    {
	close (s);
	exit_with_error ("read_loco_resp_failed");
    }
    /* die happily */

    exit(0);
}    


int get_connection ()
{
    struct sockaddr_in sa;
    int s = 0;
    unsigned int i = 0;
    int count = 0;
    int mod = 0;
    int sum = 0;
    int lid = 0;
    extern unsigned int destinations[DESTINATION_COUNT][DEST_COLS];

    /* set up socket address for connect */
    sa.sin_family = AF_INET;
    
    
    /* Get the LOCO_INST_ID value out of the cookie. */
    lid = get_loco_inst_id();
    
    if (lid > 0) {
	/* Try to get a connection with the (first) server whose instance ID
	   matches lid.  There should be at most one such server. */

	for(i=0; i<DESTINATION_COUNT; i++) {
	    if(lid == destinations[i][L_INST_ID]) {

       		sa.sin_port = htons((unsigned short) destinations[i][L_PORT]);
		sa.sin_addr.s_addr = htonl(destinations[i][L_ADDR]);
				   					
		/* get a socket */
		if ((s = socket (AF_INET, SOCK_STREAM, 6)) < 0) 
		    break;
		
		/* connect to Application Server */
		if (connect (s, (struct sockaddr *) &sa, sizeof (struct sockaddr)) == 0)
		    return (s);
		
		close (s);
		
		break;
	    }
	}
    }

    /* If our attempt to connect to the server with the right LOCO_INST_ID failed,
       we now find a server the old-fashioned way.  That is, we find a server 
       based on availability and load-balancing considerations. */

    /* prepare for randomly picking the weighted starting point */
    /* sum the weights */
    for (i = 0; i < DESTINATION_COUNT; i++)
	sum += destinations[i][L_WEIGHT];

    /* initialize the quasi-random number generator */
    srand (time (NULL) + getpid());

    /* pick the modulus */
    mod = rand() % sum;
    
    sum = 0;
    for (i = 0; i < DESTINATION_COUNT; i++)
    {
	sum += destinations[i][L_WEIGHT];
	if (mod < sum) { break; } /* will get called on last loop, but will
				     always be true.  This is an easy way 
				     to get the cursor incremented */

	/* NOTE: The value of i is preserved past this loop. */
    }

    /* enter loop, starting at destination pointed to by cursor */
    while (count < DESTINATION_COUNT)
    {
	sa.sin_port = htons((unsigned short) destinations[i][L_PORT]);
	sa.sin_addr.s_addr = htonl(destinations[i][L_ADDR]);

	/* get a socket */
	if ((s = socket (AF_INET, SOCK_STREAM, 6)) < 0) 
	    return (-1);

	/* connect to Application Server */
	if (connect (s, (struct sockaddr *) &sa, sizeof (struct sockaddr))== 0)
	    return (s);

	close (s);

	/* check if we need to reset the counter to the bottom or increment */
	if (i >= (DESTINATION_COUNT - 1))
	    i = 0;
	else
	    i++;
	    
	count++;
    }

    return (-1);
}


/* 
 * Pull the instance ID out of the HTTP_COOKIE envar.  If the cookie was not found,
 * return 0. 
 */ 

int get_loco_inst_id () 
{
    char *cookie;

    if ((cookie = getenv ("HTTP_COOKIE")) == NULL)
	return 0;

    /* At this point, the value of cookie should look something like this:
     * "COOKIE_VAL_A=foo ; <VS_COOKIE_NAME>2; COOKIE_VAL_C=bar"
     *
     * We'll advance the pointer to end of "<VS_COOKIE_NAME>" and call atoi().
     * atoi() will stop as soon as it sees a non-whitespace character that doesn't 
     * look like a decimal integer.  Leading whitespace is ignored.
     */

    if((cookie = strstr (cookie, VS_COOKIE_NAME)) == NULL)
	return 0;

    /* If we got this far, we have a pointer to the beginning of VS_COOKIE_NAME 
     * in the cookie string.  Now we advance the pointer to the first character 
     * after VS_COOKIE_NAME, which could possibly be the the string-terminating 
     * NULL byte or some other piece of non-integer cruft.  atoi() returns 0
     * for such garbage.
     */

    cookie += strlen(VS_COOKIE_NAME);

    return (atoi(cookie));
}



/* 
 * package up the envs in the simple protocol, try to be efficient,
 * and send them to the application server
 * s - an open socket 
 * buf - a char buf, to fill and send 
 */
int send_envs (int s, char *buf)
{
    extern char **environ;
    
    char *name_value_pair;
    int version = 2;
    int buf_cursor = 0;
    int env_count = 0;
    int native_length = 0;
    int netord_length = 0;

    /* prepend the version */
    netord_length = htonl (version);
    memcpy (buf, &netord_length, 4);
    buf_cursor = 4;

    /* for each env name-value pair */
    while ((name_value_pair = environ[env_count++]) != NULL)
    {
	native_length = strlen (name_value_pair);

	if ((buf_cursor + native_length + 4) <= BUF_LENGTH)
	{
	    netord_length = htonl (native_length);

	    memcpy (buf + buf_cursor, &netord_length, 4);
	    buf_cursor = buf_cursor + 4;
	    memcpy (buf + buf_cursor, name_value_pair, native_length);
	    buf_cursor += native_length;
	}
	else if ((4 + native_length) <= BUF_LENGTH)
	{
	    if (write_fully (s, buf, buf_cursor) != 0)
	    {
		return (-1);
	    }
	    buf_cursor = 0;

	    netord_length = htonl (native_length);

	    memcpy (buf + buf_cursor, &netord_length, 4);
	    buf_cursor += 4;
	    memcpy (buf + buf_cursor, name_value_pair, native_length);
	    buf_cursor += native_length;
	}
	/* silently skip envs that are greater than buffer size */
    }

    /* add zero terminator + flush */
    if ((buf_cursor + 4) <= BUF_LENGTH)
    {
	netord_length = htonl (0);
	memcpy (buf + buf_cursor, &netord_length, 4);
	buf_cursor += 4;

	if (write_fully (s, buf, buf_cursor) != 0)
	{
	    return (-1);
	}
    }
    else
    {
	if (write_fully (s, buf, buf_cursor) != 0)
	{
	    return (-1);
	}

	netord_length = htonl (0);
	memcpy (buf + buf_cursor, &netord_length, 4);
	buf_cursor += 4;

	if (write_fully (s, buf, buf_cursor) != 0)
	{
	    return (-1);
	}
    }

    return (0);
}

int write_stdin (int s, char *buf)
{
    char *cl;
    int numread = 0;
    int numwritten = 0;
    int content_length = 0;

    if ((cl = getenv ("CONTENT_LENGTH")) == NULL)
    {
	return (0);
    }
    
    content_length = atoi(cl);
    if (content_length == 0)
    {
	return (0);
    }
    else if (content_length < 0)
    {
	return (-1);
    }

    while (numwritten < content_length)
    {
	numread = read (STDIN_FILENO, buf, BUF_LENGTH);
	
	/* unexpected eof or read error */
	if (numread < 1)
	{
	    return (-1);
	}
	
	if (write_fully (s, buf, numread) != 0)
	{
	    return (-1);
	}

	numwritten += numread;
    }

    return (0);
}


int write_fully (int s, char *buf, int buflen)
{
    int total_written = 0;
    int numwritten = 0;

    while (total_written < buflen)
    {
	numwritten = write (s, buf + total_written, buflen - total_written);

	if (numwritten < 1)
	{
	    return (-1);
	}
	total_written += numwritten;
    }
    return (0);
}

int read_and_write_response (int s, char *buf)
{
    int numread = 0;
    int totalread = 0;

    /* read until eof */
    while (1)
    {
	numread = read (s, buf + totalread, BUF_LENGTH - totalread);

/*	printf ("read %d bytes\n", numread); */

	if (numread < 0)
	{
	    return (-1);
	}
	else if (numread > 0)
	{
	    totalread = totalread + numread;
	}
	else if (numread == 0)
	{
	    if (totalread > 0)
	    {
/*		printf ("%s\n", "now writing"); */
		if (write_fully (STDOUT_FILENO, buf, totalread) != 0)
		{
		    return (-1);
		}
	    }

	    if (totalread != BUF_LENGTH)
	    {
		return (0);
	    }

	    totalread = 0;
	}
    }

    return (0);
}

void exit_with_error(char *msg)
{
    printf ("Content-Type: text/html\r\n\r\n");
    
    printf ("<HTML>\n<HEAD>\n<TITLE>LocoLink CGI</TITLE>\n</HEAD>\n<BODY>\n");
    printf ("<H1>Request Failed at LocoLink: </H1>\n");
    printf ("%s: %s\n", msg, strerror(errno));
    printf ("</BODY>\n</HTML>\n");

    exit (-1);
}
