/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1999 Leverage 
 * Information Systems. All Rights reserved.
 */

/*
 * nsapi_locolink.c 
 * purpose: provide high performance connectivity between NSAPI 
 * enabled web servers (Enterprise, FastTrack) and the Leverage Locomotive
 */

#include <nsapi.h>
#include <base/file.h>
#include <base/buffer.h>
#include <frame/log.h>
#include <base/util.h>
#include <frame/protocol.h>

#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/time.h>

#define LOCO_SCAN_LINEAR 1
#define LOCO_SCAN_BALANCE 2
#define LOCO_MAXLINE 255
#define LOCO_MAX_DESTINATIONS 32
#define LOCO_MAX_DESTINATION_LISTS 32
#define LOCO_FILEBUFSIZE 4096
#define LOCO_NETBUFSIZE 32768  /* used for readback too */
#define LOCO_PROTOCOL_VERSION 4
#define LOCO_SOCK_SUCCESS 0
#define LOCO_SOCK_FATAL_ERROR -1
#define LOCO_SOCK_NON_FATAL_ERROR -2
#define LOCO_MAX_COOKIE_NAME_LEN 255

struct loco_location
{
    unsigned long address;   /* ip address in network order */
    int port;                /* tcp port number */
    unsigned int weight;     /* weight used in random pick calculation */
    unsigned int inst_id;    /* distinguish this dest. from others in pool */
};

struct loco_location_config 
{
    int style;                      /* SCAN_LINEAR or SCAN_BALANCE */
    int location_count;             /* up to LOCO_MAX_DESTINATIONS */
    char vs_cookie_name[LOCO_MAX_COOKIE_NAME_LEN];
                                    /* e.g., "LOCO_INST_ID_LOCO_ORG_SYSTAG=" */
    struct loco_location *loc_list; /* location_count structures */
};

typedef struct loco_header
{
    char *name;
    char *nsapi_equiv;
} loco_header;
	
static const loco_header headers [] =
{
    /* use rq->header for these */
    { "CONTENT_LENGTH", "content-length"},
    { "CONTENT_TYPE", "content-type" },
    { "HTTP_ACCEPT", "accept" },
    { "HTTP_ACCEPT_ENCODING", "accept-encoding" },
    { "HTTP_ACCEPT_LANGUAGE", "accept-language" },
    { "HTTP_AUTHORIZATION", "authorization" },
    { "HTTP_COOKIE", "cookie" },
    { "HTTP_IF_MODIFIED_SINCE", "if-modified-since" },
    { "HTTP_REFERER", "referer" },
    { "HTTP_USER_AGENT", "user-agent" },
    { "HTTP_USER_DEFINED", "user-defined" },
    {NULL, NULL},
    /* use rq->reqpb for these */
    { "QUERY_STRING", "query" },
    { "REQUEST_LINE", "clf-request" },
    { "REQUEST_METHOD", "method" },
    { "SCRIPT_NAME", "uri" },
    { "SERVER_PROTOCOL", "protocol" },
    {NULL, NULL},
    /* use rq->vars for these */
    { "AUTH_TYPE", "auth-type" },
    { "PATH_INFO", "path-info" },
    { "REMOTE_USER", "auth-user" },
    {NULL, NULL},
    /* use sn_client for these */
    { "HTTPS_KEYSIZE", "keysize" },
    { "HTTPS_SECRETSIZE", "secret-keysize" },
    { "REMOTE_ADDR", "ip" },
    {NULL, NULL}
};

/* the number of our normal headers, plus a couple extra.  Last I checked,
   it was about 31 */
#define NUM_ENVARS (( sizeof(headers) / sizeof(loco_header) ) + 5)

/* function declarations */
/* primary entry point, per hit */
int locolink (pblock *pb, Session *sn, Request *rq);

/* others */
struct loco_location_config *get_loco_location_config(pblock *pb, Request *rq);
int load_locations (Request *rq, int llc_index);
int get_llcindex (char *path); 
int parse_cfgfile (filebuffer *fb, struct loco_location_config *llc);
SYS_NETFD get_connection (struct loco_location_config *, Request *rq);
void write_error_msg (Session *sn, Request *rq, char *msg);
int send_envs (SYS_NETFD loco_nfd, Session *sn, Request *rq);
int write_header(char *name, char*value, SYS_NETFD l_nfd, char *buf, 
		 int *bufpos);
int buffered_writestring (SYS_NETFD l_nfd, char *buf, int *bufpos, char *val);
int write_stdin (SYS_NETFD loco_nfd, Session *sn, Request *rq);
int read_and_write_response (SYS_NETFD loco_nfd, Session *sn, Request *rq);
int get_pstring (char *buf, int *pos, int buflen, char **value);
int locolink_init (pblock *pb, Session *sn, Request *rq);
void loco_free (void *data);
int get_sock (struct loco_location_config *llc, int cursor, SYS_NETFD *sn);
SYS_NETFD try_virtual_session_connect(struct loco_location_config *llc, 
				      Request *rq);
int get_loco_inst_id (struct loco_location_config *llc, Request *rq);
int read_scan_type (filebuffer *fb, struct loco_location_config *llc);
int read_cookie_name (filebuffer *fb, struct loco_location_config *llc);
int read_locations (filebuffer *fb, struct loco_location_config *llc);
char *first_token (char linebuf[]); 
char *next_token ();

/* global variables.  My programming professor for Pascal said when I was 
 * a freshman not to use global variables.  Does Pascal even have them? 
 */
char **config_pathname_list = NULL;
struct loco_location_config **llc_exts = NULL;
CRITICAL *sync_objs;
time_t *previous_update_times = NULL;
time_t *config_file_previous_mod_times = NULL;


/* primary entry point, called per hit */
int locolink (pblock *pb, Session *sn, Request *rq)
{
    struct loco_location_config *llc;
    SYS_NETFD loco_nfd;

    /* get the fd of the client out where it is handy */

    /* get the list of locomotive locations, exit on error */
    if ((llc = get_loco_location_config (pb, rq)) == NULL)
    {
	write_error_msg (sn, rq, "get locations failure");
	return (REQ_PROCEED);
    }
    
    /* connect to a locomotive */
    if ((loco_nfd = get_connection (llc, rq)) == SYS_NET_ERRORFD)
    {
	/* connect failed, leave now */
	FREE (llc->loc_list);
	FREE (llc);
	write_error_msg (sn, rq, "connect failure");
	return (REQ_PROCEED);
    }

    /* free the stack-local copy of the configs.  Now that we have an 
       open connection, we don't need them any more */
    FREE (llc->loc_list);
    FREE (llc);

    /* send necessary envs.  Non zero if error */
    if (send_envs (loco_nfd, sn, rq) != 0)
    {
	net_close (loco_nfd);
	write_error_msg (sn, rq, "send envs failed");
	return (REQ_PROCEED);
    }

    /* write any stdin */
    if (write_stdin (loco_nfd, sn, rq) != 0)
    {
	net_close (loco_nfd);
	write_error_msg (sn, rq, "write stdin failed");
	return (REQ_PROCEED);
    }

    /* read back results from locomotive */
    if (read_and_write_response (loco_nfd, sn, rq) != 0)
    {
	write_error_msg (sn, rq, "write loco response");
    }

    net_close (loco_nfd);

    return (REQ_PROCEED);
}

/* Sends the protocol version to the locomotive, along with the 
   appropriate name value pairs from the server internals and 
   http headers.
   Right now we use a protocol like: [len]name[len]value, as opposed
   to the generic [len]name=value.  I implemented both, just in case
   we want to switch later...
   */
int send_envs (SYS_NETFD l_nfd, Session *sn, Request *rq)
{
    char *buf;
    int bufpos = 0;
    unsigned int version; 
    pblock **pblocks;
    char **name;
    char **value;
    char *tmp_value = NULL;
    /*char **env; */
    char port_str[10];
    int i = 0;
    int j = 0;
    int written = 0;
    int nv_pos = 0;

    if ((buf = (char *) MALLOC (LOCO_FILEBUFSIZE * sizeof (char))) == NULL)
    {
	/* malloc of writebuf failed */
	return (-1); 
    }

    /*
    if ((env = (char **) MALLOC ( NUM_ENVARS * sizeof (char *))) == NULL)
    {
	FREE (buf);
	return (-1);
    }
    */

    if ((name = (char **) MALLOC ( NUM_ENVARS * sizeof (char *))) == NULL)
    {
	FREE (buf);
	return (-1);
    }

    if ((value = (char **) MALLOC ( NUM_ENVARS * sizeof (char *))) == NULL)
    {
	FREE (buf);
	FREE (name);
	return (-1);
    }

    if ((pblocks = (pblock **) MALLOC (4 * sizeof(pblock *))) == NULL)
    {
	FREE(buf);
	/*FREE(env);*/
	FREE(name);
	FREE(value);
	return (-1);
    }

    /* the four different tables we need to get stuff out of */
    pblocks[0] = rq->headers;
    pblocks[1] = rq->reqpb;
    pblocks[2] = rq->vars;
    pblocks[3] = sn->client;

    for (i = 0; i < 4; i++)
    {
	while (headers[j].name)
	{
	    tmp_value = pblock_findval(headers[j].nsapi_equiv, pblocks[i]);
	    if (tmp_value)
	    {
		name[nv_pos] = headers[j].name;
		value[nv_pos++] = tmp_value;
/*	       env[nv_pos++] = util_env_str(headers[j].name, tmp_value); */
	    }
	    j++;
	}
	j++; /* skip other the null that marks a new table */
    }

    FREE(pblocks);

    /* write out the miscellaneous other headers name value pairs */
/*    env[nv_pos++] = util_env_str("NSAPI", "ON"); 
    
    env[nv_pos++] = util_env_str("HTTPS", security_active ? "ON" : "OFF"); */

    name[nv_pos] = "NSAPI"; value[nv_pos++] = "ON";
    name[nv_pos] = "HTTPS"; value[nv_pos++] = security_active ? "ON" : "OFF";
    
    if (server_hostname) 
    {
	name[nv_pos] = "SERVER_NAME";
	value[nv_pos++] = server_hostname;
	/*env[nv_pos++] = util_env_str("SERVER_NAME", server_hostname); */
    }

    tmp_value = http_uri2url("", "");
    if (tmp_value) 
    {
	name[nv_pos] = "SERVER_URL";
	value[nv_pos++] = tmp_value;
	/*env[nv_pos++] = util_env_str("SERVER_URL", http_uri2url("", ""));*/
    }

    tmp_value = session_dns(sn);
    if (tmp_value) 
    {
	name[nv_pos] = "REMOTE_HOST";
	value[nv_pos++] = tmp_value;
	/*env[nv_pos++] = util_env_str("REMOTE_HOST", tmp_value);*/
    }

    util_sprintf(port_str, "%d", conf_getglobals()->Vport);
    if (port_str) 
    {
	name[nv_pos] = "SERVER_PORT";
	value[nv_pos++] = port_str;
	/*env[nv_pos++] = util_env_str("SERVER_PORT", port_str);*/
    }

    /* Phew!  finally finished with the wretched envars */

    /* write version to buffer as first int */
    version = htonl ((unsigned int) LOCO_PROTOCOL_VERSION);

    memcpy (buf, &version, sizeof (unsigned int));
    bufpos = sizeof(unsigned int);

    for (i = 0; i < nv_pos; i++)
    {
	if ((written = buffered_writestring (l_nfd, buf, &bufpos, name[i])) 
	    == -1)
	{
	    FREE(buf);
	    /*FREE(env);*/
	    FREE(name); 
	    FREE(value);
	    return (-1);
	}

	if ((written = buffered_writestring (l_nfd, buf, &bufpos, value[i])) 
	    == -1)
	{
	    FREE(buf);
	    /*FREE(env);*/
	    FREE(name); 
	    FREE(value);
	    return (-1);
	}
    }

    /* write the 4 byte zero terminator */
    if ((written = buffered_writestring (l_nfd, buf, &bufpos, "")) == -1)
    {
	FREE (buf);
	/*FREE(env);*/
	FREE(name); 
	FREE(value);
	return (-1);	
    }

    /* finish flushing the buffer */
    if ((bufpos > 0) && (net_write (l_nfd, buf, bufpos) != bufpos))
    {
	FREE (buf);
	/*FREE(env);*/
	FREE(name); 
	FREE(value);
	return (-1);	    
    }
    
    FREE (buf);
    /*FREE(env);*/
    FREE(name); 
    FREE(value);

    return (0); /* everything went through */
}

/* Copy a pstring to the output buffer.  Flush buffer if necessary.  
   If the value of the string is longer than the buffer, write the 
   string itself directly. returns -1 on error or the number of bytes 
   written to the Locomotive or buffer */
int buffered_writestring (SYS_NETFD l_nfd, char *buf, int *bufpos, char *val)
{
    int length;
    int netorder_length; 

    length = strlen (val);
    netorder_length = htonl (length);

    /* ereport(LOG_VERBOSE, val);  */

    /* if buffer is near full, flush it first.  If it still too big
       write the length, then the value directly, rather than copying into 
       the output buffer */
    if ((*bufpos + length + 4) > LOCO_FILEBUFSIZE)
    {
	if ((*bufpos > 0) && net_write (l_nfd, buf, *bufpos) != *bufpos)
	{
	    /* write error */
	    ereport (LOG_FAILURE, 
		     "locolink: buf_wr_str buf flush to loco failed");
	    return (-1);
	}
	
	*bufpos = 0;
	
	if ((*bufpos + length + 4) > LOCO_FILEBUFSIZE)
	{
	    if (net_write (l_nfd, (char *) &netorder_length, 4) != 4)
	    {
		/* write error on length */
		ereport (LOG_FAILURE, 
			 "locolink: buf_wr_str len flush to loco failed");
		return (-1);
	    }
	    
	    if (net_write (l_nfd, val, length) != length)
	    {
		/* write error for value */
		ereport (LOG_FAILURE, 
			 "locolink: buf_wr_str val flush to loco failed");
		return (-1);
	    }
	    
	    return (length + 4);
	}
    }

    memcpy (buf + *bufpos, &netorder_length, 4);
    *bufpos += 4;
    memcpy (buf + *bufpos, val, length);
    *bufpos += length;

    return (length + 4);
}

/* check for any stdin.  If there is any, forward through to 
   the locomotive.  returns -1 on error, otherwise the length of the
   stdin that was written, or 0 if none */
int write_stdin (SYS_NETFD loco_nfd, Session *sn, Request *rq)
{
    char *length_string;
    int content_length = 0;
    int total_written = 0;

    length_string = pblock_findval ("content-length", rq->headers);

    if (length_string == NULL)
    {
	return (0);
    }
    else if (strlen (length_string) == 0 || 
	     ((content_length = atoi (length_string)) <= 0))
    {
	/* notice that above we silently ignore a parse error */
	ereport (LOG_INFORM, "locolink: content-length parse error");
	return (-1);
    }
    
/*    ereport (LOG_INFORM, 
      "content-length from browser is %d", content_length); */

    total_written = netbuf_buf2sd (sn->inbuf, loco_nfd, content_length);

    if (total_written != content_length) 
    {
	return (-1);
    }
    return (0);
}

/* purpose: 
 * using either a linear scan or a random technique, get an 
 * open connection to a locomotive.  If any are not answering, 
 * move onto the next one.  Might be nice to do a 'select'ed failover
 * if we can't connect after one second.  Usually this means that the
 * machine is off the air or its accept queue is full for that process. 
 */
SYS_NETFD get_connection (struct loco_location_config *llc, Request *rq)
{
    int sock_return_code;
    SYS_NETFD lsock;
    
    unsigned int cursor;
    int count = 0;
    
    if (llc->location_count < 1)
    {
	/* no locations! */
	ereport (LOG_FAILURE, "locolink get_conn: no locations");
	return SYS_NET_ERRORFD;
    }

    lsock = try_virtual_session_connect (llc, rq);
    if (lsock != SYS_NET_ERRORFD)
	return lsock; 

    ereport (LOG_INFORM, "No virtual session connection; proceeding with" 
	     " normal connection");

    
    /* pick our starting position in the list */
    if (llc->style == LOCO_SCAN_LINEAR)
    {
	cursor = 0;
    }
    else  /* start at a weighted random position in the list */
    {
	int mod = 0;
	int sum = 0;
	struct timeval tp;
	
	for (cursor = 0; cursor < llc->location_count; cursor++)
	{
	    sum += llc->loc_list[cursor].weight;
	}

	/* get current time */
	gettimeofday (&tp, NULL);

	/* initialize quasi-random number generator.  Won't be very random
	   at certain slow hit rates.  Perhaps use a millisec timer too? */  
	srand ((int) (tp.tv_sec + tp.tv_usec));

	mod = rand () % sum;

	sum = 0;
	for (cursor = 0; cursor < llc->location_count; cursor++)
	{
	    sum += llc->loc_list[cursor].weight;
	    if (mod < sum) { break; } /* will get called on last loop, but will
					 always be true.  This is an easy way 
					 to get the cursor incremented */
	    /* now cursor is at the correct starting point in the list */
	}
    }

    /* below code snarfed from "locolink_weight.c" */
    /* enter loop, starting at destination pointed to by cursor */
    for (count=0; count < llc->location_count; count++)
    {
	sock_return_code = get_sock(llc, cursor, &lsock);
	
	if (sock_return_code == LOCO_SOCK_SUCCESS)
	{
	    /* found an available loco */
	    return lsock;
	}
	
	if (sock_return_code == LOCO_SOCK_FATAL_ERROR)
	{
	    /* can't open a socket! */
	    return SYS_NET_ERRORFD;
	}

        /* check if we need to reset the counter to the bottom or increment */
        if (cursor >= (llc->location_count - 1))
            cursor = 0;
        else
            cursor++;
    }

    /* none of the destinations worked!  */
    ereport (LOG_FAILURE, "locolink get_conn: no working destinations");
    return SYS_NET_ERRORFD;
}

int get_sock (struct loco_location_config *llc, int cursor, SYS_NETFD *sn)
{
    struct sockaddr_in sa;

    sa.sin_family = AF_INET;
    sa.sin_port = llc->loc_list[cursor].port;
    sa.sin_addr.s_addr = llc->loc_list[cursor].address;

    /* get a socket */
    if ((*sn = net_socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) == 
	SYS_NET_ERRORFD)
    {
	ereport (LOG_FAILURE, "locolink get_conn: net_socket failed");
	return LOCO_SOCK_FATAL_ERROR;
    }
    
    /* connect to Application Server */
    if (net_connect (*sn, (struct sockaddr *) &sa, 
		     sizeof (struct sockaddr)) == 0)
    {
	return LOCO_SOCK_SUCCESS;
    }
    
    ereport (LOG_FAILURE, "locolink get_conn: connect to %s failed", 
	     inet_ntoa (sa.sin_addr));
    
    net_close (*sn);

    *sn = SYS_NET_ERRORFD;
    return LOCO_SOCK_NON_FATAL_ERROR;
}


/* Looks at the cookie and finds the instance ID.  If it's a valid ID, we try 
 * to connect to the indicated server.  Returns a nonnegative port no. on 
 * success; -1 on failure 
 */
SYS_NETFD try_virtual_session_connect(struct loco_location_config *llc, 
				      Request *rq)
{
    int inst_id;
    int cursor;
    int sock_return_code;
    SYS_NETFD lsock;

    inst_id = get_loco_inst_id (llc, rq);

    ereport (LOG_INFORM, "Examining loco_inst_id=%d", inst_id);

    if (inst_id <= 0) 
	return SYS_NET_ERRORFD;

    for (cursor=0; cursor < llc->location_count; cursor++) 
    {
	if (inst_id == llc->loc_list[cursor].inst_id) 
	{
	    /* try to get a connection to the (first) matching instance */
	    sock_return_code = get_sock (llc, cursor, &lsock);
	    if (sock_return_code == LOCO_SOCK_SUCCESS) 
	    {
		return lsock;
	    }
	    else 
	    {
		return SYS_NET_ERRORFD;
	    }
	}
    }
    
    return SYS_NET_ERRORFD;
}


/* Return an integer representing the instance ID in the cookie. 
 * Return value is 0 if no valid instance ID was found. 
 * Code taken from src/c/locolink/cgi/locolink.c. 
 */
int get_loco_inst_id (struct loco_location_config *llc, Request *rq)
{
    char *cookie;

    cookie = pblock_findval ("cookie", rq->headers);

    if (cookie == NULL) {
	ereport (LOG_INFORM, "Null cookie");
	return 0;
    }

    ereport (LOG_INFORM, "Fetched cookie: %s", cookie);
    
    /* At this point, the value of cookie should look something like this:
     * "COOKIE_VAL_A=foo ; {llc->vs_cookie_name}2; COOKIE_VAL_C=bar"
     *
     * We'll advance the pointer to end of "{llc->vs_cookie_name}" and call 
     * atoi().  atoi() will stop as soon as it sees a non-whitespace character
     * that doesn't  look like a decimal integer.  Leading whitespace is 
     * ignored.
     */

    if((cookie = strstr (cookie, llc->vs_cookie_name)) == NULL) {
	ereport (LOG_INFORM, "VS cookie not located in cookie string");
	return 0;
    }

    /* If we got this far, we have a pointer to the beginning of vs_cookie_name
     * in the cookie string.  Now we advance the pointer to the first character
     * after vs_cookie_name, which could possibly be the the string-
     * terminating NULL byte or some other piece of non-integer cruft.  atoi() 
     * returns 0 for such garbage.
     */

    /* ereport (LOG_INFORM, "cookie name: %s", llc->vs_cookie_name); */
    /* ereport (LOG_INFORM, "cookie name length: %d", 
                strlen(llc->vs_cookie_name)); */
    
    cookie += strlen(llc->vs_cookie_name);
    /* ereport (LOG_INFORM, "Value sent to atoi: %s", cookie); */

    return (atoi(cookie));
}


/* reads all data back from the locomotive and tells the webserver
 * to send it back to the client after being properly formatted.  
 * The data from the locomotive comes in as a list of name value pairs, 
 * followed by the actual "data body", typically HTML.
 */
int read_and_write_response (SYS_NETFD loco_nfd, Session *sn, Request *rq)
{
    int numread = 0;
    int total_read = 0;
    int done = 0;
    int eof_reached = 0;
    int status_was_set = 0;
    int response_status = 0;
    char *name = NULL;
    char *value = NULL;
    char *buf;
    int pos = 0;

    /* strategy:  Attempt to read first LOCO_NETBUFSIZE back from 
     * locomotive before processing any of the data.  It is likely, 
     * with current web page sizes that the entire document can 
     * be read and the connection to the locomotive can be closed.  
     * This allows locomotive threads to get onto another task, rather 
     * than becoming tied up pushing data.  If not all can be read, 
     * start sending the headers back (and do other NSAPI setup), then
     * write the data out while continuuing to read from the loco. */

    buf = MALLOC (LOCO_NETBUFSIZE);
    if (buf == NULL)
    {
	ereport (LOG_CATASTROPHE, 
		 "locolink: rw_response: buf malloc failed");
	return (-1);
    }

    /* try to read everything back from locomotive now */
    while (!done && (LOCO_NETBUFSIZE - total_read) > 0)
    {
	numread = net_read (loco_nfd, buf + total_read, 
			    LOCO_NETBUFSIZE - total_read, 180);

	if (numread == 0)
	{
	    eof_reached = 1;
	    done = 1;
	}
	else if (numread < 0)
	{
	    eof_reached = 1;
	    done = 1;
	    ereport (LOG_FAILURE, "locolink: read error on loco_nfd");
	}
	else 
	{
	    total_read += numread;
	}
    }

    done = 0;
    while (!done)
    {
	int retval = 0;

	retval = get_pstring (buf, &pos, total_read, &name);
	if (retval == 0)
	{
	    done = 1;
	    continue;
	}
	else if (retval < 0)
	{
	    ereport (LOG_FAILURE, "locolink: get_pstring failed (name)");
	    FREE (buf);
	    return (-1);
	}

	retval = get_pstring (buf, &pos, total_read, &value);
	if (retval == 0)
	{
	    value = "";
	}
	else if (retval < 0)
	{
	    ereport (LOG_FAILURE, "locolink: get_pstring failed (value)");
	    FREE (buf);
	    return (-1);	    
	}

	if (! strcasecmp (name, "content-type"))
	{
	    param_free(pblock_remove("content-type", rq->srvhdrs));
	    pblock_nvinsert("content-type", value, rq->srvhdrs);
	}
	else if (! strcasecmp (name, "location"))
	{
	    protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);
	    param_free(pblock_remove("location", rq->srvhdrs));
	    pblock_nvinsert("location", value, rq->srvhdrs);
	    status_was_set = 1;
	}
	else
	{
	    pblock_nvinsert(name, value, rq->srvhdrs);
	}
    }

    if (! status_was_set)
    {
	protocol_status(sn, rq, PROTOCOL_OK, NULL);
    }

    /* start the proper response */
    response_status = protocol_start_response (sn, rq);
    if (response_status != REQ_PROCEED)
    {
	FREE (buf);
	return (0);
    }

    /* write remainder of buffer */
    if (pos < total_read)
    {
	int written = net_write (sn->csd, buf + pos, total_read - pos);
	if (written != total_read - pos)
	{
	    FREE (buf);
	    return (0);
	}
    }

    /* read anything else from the loco */
    if (! eof_reached)
    {
	while ((numread = net_read (loco_nfd, buf, LOCO_NETBUFSIZE, 180)) > 0)
	{
	    if (net_write (sn->csd, buf, numread) != numread)
	    {
		FREE (buf);
		return (-1);
	    }
	}
    }

    /* write remainder of socket, if any */
    FREE (buf);
    return (0);
}

/* used by read_and_write_response (above) to read back name value pairs from 
   the loco.  */
int get_pstring (char *buf, int *pos, int buflen, char **value)
{
    int len = 0;

    /* check for data availability */
    if (buflen - *pos < 4)
    {
	return (-1);
    }

    memcpy (&len, buf + *pos, 4);
    *pos += 4;

    if (len == 0)
    {
	return (0);
    }

    /* check for data availability again */
    if (buflen - *pos < len)
    {
	return (-1);
    }

    /* From observation of memory behaviour, it seems possible 
     * that fasttrack FREE()s anything 
     * allocated with MALLOC at the end of a hit automatically. 
     * Testing confirms, but is not def
     */
    *value = (char *) MALLOC (len + 1);  /* leave room for null terminator */
    memcpy (*value, buf + *pos, len);

    *pos += len;
    (*value)[len] = 0;

    return (len);
}

/* attempts to send a 'fatal' looking error message back to the 
 * browser */
void write_error_msg (Session *sn, Request *rq, char *msg)
{
    int response_status = 0;
    char outbuf[1024]; 
    int msglen = 0;
    /* what we do here is send back the basic server headers, 
     * then add a tidbit of HTML and a simple error message */

    /* We need to get rid of the internal content type. */
    param_free(pblock_remove("content-type", rq->srvhdrs));

    pblock_nvinsert("content-type", "text/html", rq->srvhdrs);

    protocol_status(sn, rq, PROTOCOL_OK, NULL);
    response_status = protocol_start_response (sn, rq);

    if (response_status != REQ_PROCEED)
    {
	return;
    }

    sprintf (outbuf, "<HTML>\n<H1>Locolink ERROR: </H1>%s\n</HTML>", msg);
    msglen = strlen (outbuf);
    
    net_write (sn->csd, outbuf, msglen);
}

/* purpose: return a structure that contains the search list for 
   locomotives.  The list can then be traversed linearly or randomly, 
   using weights.  */
struct loco_location_config *get_loco_location_config(pblock *pb, Request *rq)
{
    extern struct loco_location_config **llc_exts;
    extern CRITICAL *sync_objs;
    extern time_t *previous_update_times;

    struct loco_location_config *llc_copy;
    time_t time_now; 
    int llc_index = 0;  /* offset into llc_exts */
    char *path; 

    path = pblock_findval ("cfgpath", pb);
    if (path == NULL)
    {
	ereport (LOG_FAILURE, "locolink: must define cfgpath in obj.conf");
	return (NULL);
    }

    llc_index = get_llcindex (path);

    /* one second resolution, comparison below locolink will update
     itself at a maximum of once per second.  At the time of this 
     writing one second intervals seems appropriate and will not 
     impose much load on the system */
    time_now = time (NULL);

    /* enter sensitive area - don't want more than one thread updating
       the config at once. */
    crit_enter (sync_objs[llc_index]);

    /* load new if time has changed */
    if (time_now > previous_update_times[llc_index])
    {
	int num_locations = 0;
	
/*	ereport (LOG_VERBOSE, "locolink: getloc: now: %d, previous: %d", 
	         time_now, previous_update_times[llc_index]); */
	
	num_locations = load_locations (rq, llc_index);
	if (num_locations < 1)
	{
	    crit_exit (sync_objs[llc_index]);
	    return (NULL);
	}
	
	previous_update_times[llc_index] = time_now;
    }

    /* make a copy to return for this thread */
    llc_copy = (struct loco_location_config *)
	MALLOC (sizeof (struct loco_location_config));

    llc_copy->style = llc_exts[llc_index]->style;
    llc_copy->location_count = llc_exts[llc_index]->location_count;

    strncpy (llc_copy->vs_cookie_name, llc_exts[llc_index]->vs_cookie_name, 
	     LOCO_MAX_COOKIE_NAME_LEN);

    llc_copy->loc_list = (struct loco_location *)
	MALLOC (llc_copy->location_count * sizeof (struct loco_location));
    
    memcpy (llc_copy->loc_list, llc_exts[llc_index]->loc_list, 
	    llc_copy->location_count * sizeof (struct loco_location));

    /* leave sensitive area */
    crit_exit (sync_objs[llc_index]);
    
    return (llc_copy);
}


/* purpose: called by get_loco_location_config to get the offset integer
   that points into llc_exts.  The config path is compared to all known
   config paths until a match is found.  If not found, the function will
   return a suitable index for the location of a new config.  If there 
   is no more remaining space for new config entries, the index of the 
   last valid entry is returned. */
int get_llcindex (char *path)
{
    extern char **config_pathname_list;
    int index = 0; 
    
    while (config_pathname_list[index] != NULL && 
	   index < LOCO_MAX_DESTINATION_LISTS)
    {
	if (! strcmp (config_pathname_list[index], path))
	{
	    return (index);
	}
	
	index++;
    }

    /* no match thus far.  If room, add the entry so that it can be found
     next time. */
    if (index < LOCO_MAX_DESTINATION_LISTS)
    {
	int length = 0;

	config_pathname_list[index] = 
	    (char *) PERM_MALLOC (LOCO_MAXLINE * sizeof (char));
	/* I feel like a virtual machine - allocating and deallocating 
	 * memory left and right.  Computers are much better 
	 * at keeping track of all these things 
	 */
	
	length = strlen (path);
	if (length < (LOCO_MAXLINE -1)) /* leave space for null terminator */
	{
	    strcpy (config_pathname_list[index], path);
	}
	else
	{
	    memcpy (config_pathname_list[index], path, (LOCO_MAXLINE - 1));
	    *(config_pathname_list[index] + (LOCO_MAXLINE - 1)) = 0;
	}

	return (index);
    }

    /* to many config names ignore the new entry */
    return (LOCO_MAX_DESTINATION_LISTS - 1);
}

/* gets called a max of once per second.  Attempts to read in 
   destination information from a config file whose location is passed
   in from above.
   - check file status - if not changed, skip reload
   - open the config file at the location specified in the pblock
   - read in the resultant binary data structure.
   - check it as best as possible 
   - if there is any failure, return a genericly configured structure 
   */
int load_locations (Request *rq, int llc_index)
{
    extern char **config_pathname_list; 
    extern struct loco_location_config **llc_exts;
    extern time_t *config_file_previous_mod_times;

    struct loco_location_config *newllc;
    struct loco_location_config *oldllc;
    SYS_FILE cfg_fd;
    filebuffer *fb;
    struct stat *cfgstat;
    int parse_return_val = 0;
    
    /* get the stat.  Note that NSAPI docs discourage freeing the stat pointer 
     */
    if ((cfgstat = 
	 request_stat_path (config_pathname_list[llc_index], rq)) == NULL)
    {
	ereport (LOG_FAILURE, "stat of config file %s failed.  Is it there?",
		 config_pathname_list[llc_index]);
	return (-1);
    }

    /* if the file has already been loaded, skip it, go home */
    if (cfgstat->st_mtime <= config_file_previous_mod_times[llc_index])
    {
/*	ereport (LOG_VERBOSE, "configs cached already: %s", path);  */
	return (llc_exts[llc_index]->location_count);
    }

    ereport (LOG_INFORM, "locolink: starting config reload: %s", 
	     config_pathname_list[llc_index]);

    config_file_previous_mod_times[llc_index] = cfgstat->st_mtime;

    /* open the file */
    if ((cfg_fd = system_fopenRO (config_pathname_list[llc_index])) == 0)
    {
	ereport (LOG_FAILURE, "locolink: open_config failed for: ", 
		 config_pathname_list[llc_index]);
	return (-1);
    }

    if ((fb = filebuf_open_nostat (cfg_fd, LOCO_FILEBUFSIZE, cfgstat)) == NULL)
    {
	ereport (LOG_FAILURE, 
		 "locolink: filebuf create failed for config file");
	/* handle error */
	if (system_fclose (cfg_fd))
	{
	    ereport (LOG_FAILURE, "locolink: onerror close cfg_fd failed");
	    /* handle the close error */
	    return (-1);
	}

	return (-1);
    }

    /* parse the file buf */
    newllc = (struct loco_location_config *) 
	PERM_MALLOC (sizeof (struct loco_location_config));

    if ((parse_return_val = parse_cfgfile (fb, newllc)) < 1)
    {
	ereport (LOG_FAILURE, "locolink: config_file_parse_failed");
	PERM_FREE (newllc);
    }
    else
    {
	/* switch a roo - structure space is pre allocated in init below */
	oldllc = llc_exts[llc_index];
	llc_exts[llc_index] = newllc; 
	PERM_FREE (oldllc->loc_list);
	PERM_FREE (oldllc);
    }
    
    /* cleanup */
    filebuf_close (fb);
    system_fclose (cfg_fd);

    return (parse_return_val);
}

/* go through the filebuffer, scan the lines for contents of the 
 * struct loco_location_config in ascii form.  Why can't we just live
 * in a binary world?  Must we punish ourselves? 
 * returns: either a -1 on error, or the number of successfully scanned 
 * lines, which will be gt than zero, and lt or eq to LOCO_MAX_DESTINATIONS 
 */
int parse_cfgfile (filebuffer *fb, struct loco_location_config *llc)
{
    /* Read the scan type line */
    if (read_scan_type (fb, llc) < 0)
      return -1;

    /* Read the cookie name line */
    if (read_cookie_name (fb, llc) < 0)
      return -1;

    /* ereport (LOG_INFORM, "pcfg: cookie name: %s", llc->vs_cookie_name); */

    /* Read the locations */
    return read_locations (fb, llc);
}


/* Read the scan type from the config file and store in llc->style.
 * Returns negative if error; otherwise zero.
 */
int read_scan_type (filebuffer *fb, struct loco_location_config *llc)
{
    int callstatus = 0;
    char linebuf[LOCO_MAXLINE];

    if ((callstatus = util_getline (fb, 1, LOCO_MAXLINE - 1, linebuf)) != 0)
    {
	ereport (LOG_FAILURE, 
		 "locolink: scantype readtoken failed, status %d", callstatus);
	/* EOF or read failed */
	return -1;
    }

    if (strlen (linebuf) >= 6  &&  !strncmp ("linear", linebuf, 6))
    {
	llc->style = LOCO_SCAN_LINEAR;
	ereport (LOG_INFORM, "locolink: config: linear scantype");
    }
    else if (strlen (linebuf) >= 7 &&  !strncmp ("balance", linebuf, 7))
    {
	llc->style = LOCO_SCAN_BALANCE;	
	ereport (LOG_INFORM, "locolink: config: balance scantype");
    }
    else
    {
	ereport (LOG_FAILURE, "locolink: config file: scantype invalid");
	/* style listed is bogus */
	return -1;
    }
    return 0;
}

int read_cookie_name (filebuffer *fb, struct loco_location_config *llc)
{
    char linebuf[LOCO_MAXLINE];
    int i;
    int c = 0;
    int len = 0;
    int callstatus = 0;

    if ((callstatus = util_getline (fb, 2, LOCO_MAXLINE - 1, linebuf)) != 0)
    {
	ereport (LOG_FAILURE, 
		 "locolink: cookie name readtoken failed, status %d", 
		 callstatus);
	/* EOF or read failed */
	return -1;
    }

    /* parse the cookie name */
    /* We will just check that the string consists of some non-whitespace 
     * characters and that the last character is an equals sign ("=").  
     * Furthermore, we'll barf if there are any commas.  This will cause a 
     * meaningful error to be thrown if the cookie line is omitted.
     *
     * If everything is okay, return 0, else -1.  
     */

    /* Here strtok() has the effect of truncating linebuf at the first newline.
     */
    strtok (linebuf, "\n");
    len = strlen (linebuf);

    /* The string must have at least one character in addition to the "=" */
    if (len < 2) 
    {
	ereport (LOG_FAILURE, 
		 "ERROR: locolink: config file: empty cookie name");
        return -1;
    }

    /* The string must not be too long */
    if (len > LOCO_MAX_COOKIE_NAME_LEN) 
    {
	ereport (LOG_FAILURE, 
		 "ERROR: locolink: config file: cookie name line too long");
        return -1;
    }

    /* The string must not contain whitespace or commas */
    for (i = 0; (c = (int)linebuf[i]) != '\0'; ++i) 
    {
	if (c == ' ' || c == '\t') 
	{
	    ereport (LOG_FAILURE, "ERROR: locolink: config file: illegal "
		     "whitespace in cookie name");
	    return -1;
	}
	if (c == ',') 
	{
	    ereport (LOG_FAILURE, "ERROR: locolink: config file: illegal comma"
		     " in cookie name; second line must contain cookie name "
		     "for virtual sessions");
	    return -1;
	}
	if (c == '=') 
	  break;
    }

    /* The last character examined must be an equal sign */
    if (c != '=') 
    {
	ereport (LOG_FAILURE, "ERROR: locolink: config file: last char of "
		 "cookie name must be '='");
	return -1;
    }
    
    strncpy(llc->vs_cookie_name, linebuf, LOCO_MAX_COOKIE_NAME_LEN);
    
    ereport (LOG_INFORM, "locolink: config: got cookie name: %s", 
	     llc->vs_cookie_name);
    return 0;
}
  
int read_locations (filebuffer *fb, struct loco_location_config *llc) 
{
    char linebuf[LOCO_MAXLINE];
    unsigned long address = 0;
    int port = 0;
    unsigned int weight = 0;
    unsigned int inst_id = 0;
    struct in_addr s_inaddr;
    char *token;
    int linenum;
    int callstatus = 0;

    /* ereport (LOG_INFORM, "rloc: cookie name: %s", llc->vs_cookie_name); */

    /* just in case it wasn't already zero */
    llc->location_count = 0;

    /* make room for the new list - must free if we leave early */
    llc->loc_list = (struct loco_location *) 
	PERM_MALLOC (LOCO_MAX_DESTINATIONS * sizeof (struct loco_location));

    /* ereport (LOG_INFORM, "Starting read_locations"); */

    for (linenum = 3; 
	 callstatus == 0 && llc->location_count < LOCO_MAX_DESTINATIONS; 
	 linenum++)
    {
	/* util_getline returns a 0 as long as it successfully retrieves a line
	 * and EOF has not been reached.
	 */
	callstatus = util_getline (fb, linenum, LOCO_MAXLINE - 1, linebuf);

	/* ereport (LOG_INFORM, "util_getline returned %d: %s", callstatus, 
	            linebuf); */
	
	if (callstatus > 0)   /* EOF */
	    break;
	
	if (callstatus < 0)   /* error */
	{
	    ereport (LOG_FAILURE, "ERROR: locolink: config: getline fail: %s", 
		     linebuf);
	    PERM_FREE(llc->loc_list);
	    return -1;	   
	}
	
	address = 0;
	port = 0;
	weight = 0;
	inst_id = 0;

	/* ereport (LOG_INFORM, "Preparing to tokenize line"); */

	/*** Get address ***/
	token = first_token (linebuf);
	/* ereport (LOG_INFORM, "Got first token %s", token); */
	if (token == NULL) 
	{
	    ereport (LOG_FAILURE, "ERROR: locolink: config: no address given");
	    PERM_FREE(llc->loc_list);
	    return -1;
	} 
	address = inet_addr (token);
	/* ereport (LOG_INFORM, "Got address: %lu", address); */
	if (address == -1) 
	{
	    ereport (LOG_FAILURE, 
		     "ERROR: locolink: config: bogus net address");
	    PERM_FREE(llc->loc_list);
	    return -1;
	} 
	/* ereport (LOG_INFORM, 
	            "Finished getting address token: %lu", address); */

	/*** Get port ***/
	token = next_token ();
	if (token == NULL) 
	{
	    ereport (LOG_FAILURE, "ERROR: locolink: config: no port given");
	    PERM_FREE(llc->loc_list);
	    return -1;
	}
	port = atoi (token);
	if (port < 1 || port > 65535) 
	{
	    /* port number is bogus */
	    ereport (LOG_FAILURE, 
		     "ERROR: locolink: config: bogus port number");
	    PERM_FREE(llc->loc_list);
	    return -1;
	}

	/*** Get weight ***/
        token = next_token ();
	if (token == NULL) 
	{	    
	    ereport (LOG_FAILURE, "ERROR: locolink: config: no weight given");
	    PERM_FREE(llc->loc_list);
	    return -1;
	}
	weight = atoi (token);
	if (weight < 1) 
	{
	    /* weight val is bogus */
	    ereport (LOG_FAILURE, 
		     "ERROR: locolink: config: invalid weight value");
	    PERM_FREE(llc->loc_list);
	    return -1;
	}

	/*** Get inst ID ***/
	token = next_token ();
	if (token == NULL) 
	{
	    ereport (LOG_FAILURE, 
		     "ERROR: locolink: config: no instance ID given");
	    PERM_FREE(llc->loc_list);
	    return -1;
	}
	inst_id = atoi (token);
	if (inst_id < 1) 
	{
	    ereport (LOG_FAILURE, 
		     "ERROR: locolink: config: invalid inst_id value");
	    PERM_FREE(llc->loc_list);
	    return -1;
	}

	/* ereport (LOG_INFORM, "Finished getting tokens %lu %d %u %u", 
	            address, port, weight, inst_id); */

	/* if we got this far, set the variables and write log */

	llc->loc_list[llc->location_count].address = address;
	llc->loc_list[llc->location_count].port = port;
	llc->loc_list[llc->location_count].weight = weight;
	llc->loc_list[llc->location_count].inst_id = inst_id;
	llc->location_count++;

	s_inaddr.s_addr = address;
	ereport (LOG_INFORM, "locolink: config: scanned location %s,%d,%u,%u", 
		 inet_ntoa (s_inaddr), port, weight, inst_id);

    }

    /* There should never be fewer than one location in the config file */
    if (llc->location_count < 1) {
	ereport (LOG_FAILURE, 
		 "locolink: config: read error or zero location count");
	PERM_FREE(llc->loc_list);
        return -1;
    }

    /* return the number of new locations found */    
    return llc->location_count;    
}

/* Initialize the string tokenizer and get the first token.
 * Should set the delimiter to null and return a pointer 
 */
char *first_token (char linebuf[]) 
{
    return strtok (linebuf, ",");
}

/* Get next token from the string we provided as an argument to first_token () 
 */
char *next_token () 
{
    return strtok (NULL, ",");
}


/* called before the server forks.  Question: if the server forks to
   create a bunch of child processes, does it mean this is executed
   every time?  
   Note: there is essentially a one to one mapping between the
   code sections contained within and the external variable defined 
   before the locolink fuction */
int locolink_init (pblock *pb, Session *sn, Request *rp)
{
    extern char **config_pathname_list;
    extern struct loco_location_config **llc_exts;
    extern CRITICAL *sync_objs;
    extern time_t *previous_update_times;
    extern time_t *config_file_previous_mod_times;
    int i = 0;

    /* allocate space to keep path names */
    if (config_pathname_list == NULL)
    {
	config_pathname_list = 
	    (char **) PERM_MALLOC (
		LOCO_MAX_DESTINATION_LISTS * sizeof (char *));

	for (i = 0; i < LOCO_MAX_DESTINATION_LISTS; i++)
	{
	    config_pathname_list[i] = NULL;
	}
    }
    
    /* allocate memory to contain our config data */
    if (llc_exts == NULL)
    {
	/* allocate enought room for all of the pointers */
	llc_exts = (struct loco_location_config **)
	    PERM_MALLOC (LOCO_MAX_DESTINATION_LISTS * 
			 sizeof (struct loco_location_config *));

	/* initialize to enough space for each one */  
	for (i = 0; i < LOCO_MAX_DESTINATION_LISTS; i++)
	{
	    llc_exts[i] = (struct loco_location_config *)
		PERM_MALLOC (sizeof (struct loco_location_config));
	}
    }

    /* create sync objects used above in get_locations() */
    if (sync_objs == NULL)
    {
	sync_objs = (CRITICAL *) PERM_MALLOC (LOCO_MAX_DESTINATION_LISTS *
					      sizeof (CRITICAL));
	for (i = 0; i < LOCO_MAX_DESTINATION_LISTS; i++)
	{
	    sync_objs[i] = crit_init ();
	}
    }

    /* update times */
    if (previous_update_times == NULL)
    {
	previous_update_times = (time_t *)
	    PERM_MALLOC (LOCO_MAX_DESTINATION_LISTS * sizeof (time_t));

	for (i = 0; i < LOCO_MAX_DESTINATION_LISTS; i++)
	{
	    previous_update_times[i] = 0;
	}
    }

    /* file mod times */
    if (config_file_previous_mod_times == NULL)
    {
	config_file_previous_mod_times = (time_t *)
	    PERM_MALLOC (LOCO_MAX_DESTINATION_LISTS * sizeof (time_t));

	for (i = 0; i < LOCO_MAX_DESTINATION_LISTS; i++)
	{
	    config_file_previous_mod_times[i] = 0;
	}
    }

    daemon_atrestart (loco_free, NULL);
    return (REQ_PROCEED);
}

/* called via callback set up in locolink_init 
 * purpose is to free up any resources associated with locks
 * or with the locations config object */
void loco_free (void *unused)
{
    extern char **config_pathname_list;
    extern struct loco_location_config **llc_exts; 
    extern CRITICAL *sync_objs;
    extern time_t *previous_update_times;
    extern time_t *config_file_previous_mod_times;
    int i = 0;

    /* free up path name allocations */ 
    while (config_pathname_list[i] != NULL && i < LOCO_MAX_DESTINATION_LISTS)
    {
	PERM_FREE ( config_pathname_list[i] );
	i++;
    }
    PERM_FREE (config_pathname_list);
    config_pathname_list = NULL;


    /* free things allocated in init or in load_locations */
    i = 0; /* reset i */
    while (llc_exts[i] != NULL && i < LOCO_MAX_DESTINATION_LISTS)
    {
	PERM_FREE ( llc_exts[i]->loc_list );
	PERM_FREE (llc_exts[i]);
	i++;
    }
    PERM_FREE (llc_exts);
    llc_exts = NULL;


    /* free mutexes */
    i = 0; 
    while (sync_objs[i] != NULL && i < LOCO_MAX_DESTINATION_LISTS)
    {
	crit_terminate (sync_objs[i]);
	i++;
    }
    PERM_FREE (sync_objs);
    sync_objs = NULL;


    /* free update times list */
    PERM_FREE (previous_update_times);
    previous_update_times = NULL;

    /* free mod times list */
    PERM_FREE (config_file_previous_mod_times);
    config_file_previous_mod_times = NULL;
}
