/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

/* 
 * stoploco.c  
 * Passes a magic number to the locomotive, which, if from a valid 
 * host, will cause the locomotive to shut down.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <netdb.h>

#define SHUTDOWN_MAGIC_NUMBER 0x09080706

int get_connection (unsigned int destination_ip, 
		    unsigned short destination_port);
void exit_with_error(char *msg);
int parse_command_line (int argc, char **argv);

/* enter server destinations - portnum, expanded address 
   up the count to the number of port/address pairs 
   Note: 2130706433 is 127.0.0.1 expanded */
unsigned int destination_ip = 2130706433;
unsigned short destination_port = 5421;

/* standard env, not currently used */
extern char **environ;

int main (int argc, char **argv)
{
    char outbuf[4];
    unsigned int magic_int; 
    struct in_addr loco_in_addr;
    int deliver_count = 0;

    /* parse command line options returns -1 if we should exit */
    /* it will set the external variables dest_ip and dest_port */
    if (parse_command_line (argc, argv) != 0)
    {
	exit (-1);
    }

    magic_int = htonl(SHUTDOWN_MAGIC_NUMBER);

    /* copy the magic number into the data buffer to be written */
    memcpy (outbuf, &magic_int, sizeof (unsigned int));

    /* set the ip address */
    loco_in_addr.s_addr = htonl (destination_ip);

    /* do some prep for logging */
    printf ("Attempting shutdown of Locomotive at: %s, port %d...\n", 
	    inet_ntoa (loco_in_addr), destination_port);

    /* connect, deliver data, close.  Do twice. */
    while (deliver_count < 2)
    {
	int socket_fd = 0;
	int bytes_written = 0;

	/* get a connection */
	if ((socket_fd = 
	     get_connection (destination_ip, destination_port)) < 0)
	{
	    exit_with_error(
		"Locomotive is dead or not at location specified");
	}
	
	bytes_written = write (socket_fd, outbuf, sizeof (unsigned int));

	if (bytes_written == -1)
	{
	    close (socket_fd);
	    exit_with_error ("Message Delivery failed");
	}
	else if (bytes_written < sizeof (unsigned int))
	{
	    printf ("that's odd, got short write count of %d", bytes_written);
	}
	
	close (socket_fd);
	deliver_count++;
    }
    /* die happily */
    exit(0);
}    

int get_connection (unsigned int destination_ip, 
		    unsigned short destination_port)
{
    struct sockaddr_in sa;
    int s = 0;

    /* set up socket address for connect */
    sa.sin_family = AF_INET;

    /* set the cursor to a random location in the list */
    sa.sin_addr.s_addr = htonl(destination_ip);
    sa.sin_port = htons (destination_port);

    /* get a socket */
    if ((s = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) 
    {
	exit_with_error ("socket allocation failed");
    }

    /* connect to Application Server */
    if (connect (s, (struct sockaddr *) &sa, sizeof (struct sockaddr)) == 0)
    {
	return (s);
    }

    return (-1);
}

/* parse command line.  Looks like this
   stoploco [-h] [-p port] [-a ipaddress or hostname] 
   */
int parse_command_line (int argc, char **argv)
{
    int cursor = 0;
    static char *synopsis = "Usage: %s [-h | --help] [-p port] [-a hostname|ipaddress]\n";

    /* loop though, skip over name in argv[0] */
    for (cursor = 1; cursor < argc; cursor++)
    {
	if (! strcmp ("-h", argv[cursor]) || !strcmp ("-help", argv[cursor]))
	{
	    printf (synopsis, argv[0]);
	    return (-1);
	}
	else if (! strcmp ("-p", argv[cursor]))
	{
	    int tmp_val = 0;
	    /* must have port afterwards */
	    if (cursor + 1 >= argc)
	    {
		printf ("Must specify port after -p\n");
		printf (synopsis, argv[0]);
		return (-1);
	    }
	    
	    cursor++;
	    tmp_val = atoi (argv[cursor]);

	    if (tmp_val > 65535 || tmp_val <= 0)
	    {
		printf ("Port number invalid\n");
		printf (synopsis, argv[0]);
		return (-1);
	    }
	    
	    destination_port = (unsigned short) tmp_val;
	}
	else if (! strcmp ("-a", argv[cursor]))
	{
	    struct hostent *hent; 

	    if (cursor + 1 >= argc)
	    {
		printf ("Must specify host after -a\n");
		printf (synopsis, argv[0]);
		return (-1);
	    }
	    
	    cursor++;
	    hent = gethostbyname (argv[cursor]);
	    
	    if (hent == NULL)
	    {
		if (h_errno == HOST_NOT_FOUND)
		{
		    printf ("Error on host lookup: host not found\n");
		}
		else if (h_errno == TRY_AGAIN)
		{
		    printf ("Error on host lookup: resource unavailable\n");
		}
		else if (h_errno == NO_RECOVERY)
		{
		    printf ("Error on host lookup: name server failure\n");
		}
		else if (h_errno == NO_DATA)
		{
		    printf ("Error on host lookup: host has no address\n");
		}
		printf (synopsis, argv[0]);
		return (-1);
	    }
	    else if (hent->h_addrtype == AF_INET)
	    {
		int tmp = 0;

		/* debug 
		printf ("found %s\n", hent->h_addr_list[0]); */

		memcpy (&tmp, hent->h_addr_list[0], 4);
		destination_ip = htonl(tmp); /* undone by htonl() above */
	    }
	    else
	    {
		printf ("gethostbyname(): namespace type not understood\n");
		printf (synopsis, argv[0]);
		return (-1);
	    }
	}
	else
	{
	    printf (synopsis, argv[0]);
	    return (-1);
	}

	/* debug */
/*	printf ("position: %d, %s\n", cursor, argv[cursor]);  */
    }

    return (0);
}

void exit_with_error(char *msg)
{
    printf ("%s: %s\n", msg, strerror(errno));
    exit (-1);
}
