/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.commerce.ics;

import java.util.*;

/**
 * Represents a response returned from Cybersource.  The fields
 * that are contained in the reply are specific to the requested services. 
 * Please see the 
 * <a href="http://msc.ic3.com/html/docs/flex/intgd/intintro.html">Cybersource 
 * CommerceFLEX API documentation</a> for more information.  
 * This object is not designed for multiple thread access. 
 */
public class ICSReply
{
    private SCMPMessage msg;
    private int offer_count;

    //* called by ICSRequest.send ()
    ICSReply (SCMPMessage reply_msg, int offer_cnt) 
    {
	msg = reply_msg;
	offer_count = offer_cnt;
    }

    /**
     * Gets the value for the given return code name.  For each ICS service 
     * requested there should be specific codes and response values 
     * that give the details of the processing that occured along with 
     * any errors.
     * If the given return code does not exist in the underlying SCMPMessage, 
     * this method returns null.  
     */
    public String getValueByName (String name)
    {
	return (msg.getValueByName (name));
    }

    /**
     * Useful for debugging.
     * @returns a reference to the underlying SCMPMessage that was returned
     * from Cybersource.
     */
    public SCMPMessage getSCMPMessage ()
    {
	return (msg);
    }

    /**
     * Returns an array of URLS generated by Cybersource if the 
     * ICSRequest contained a request for the &quot;ics_download&quot;
     * service.  The urls will be ordered in the order of the Offers
     * submitted in the ICSRequest.  If a download URL was not relavant 
     * to a particular offer, the value will be a null string.
     */
    public String[] getDownloadURLs ()
    {
	return (getValueArray ("download_url"));
    }

    /**
     * Gets the reference numbers (elc_ref_nums) for offers when 
     * the ICS service &quot;ics_elc&quot; was requested in the Request.
     * If the offer was not electronic and therefore does not have 
     * corresponding reference number, value will be a null string.
     */
    public String[] getElcRefNumbers ()
    {
	return (getValueArray ("elc_ref_nums"));
    }

    /**
     * Gets the accept URLs (preapp_accept_url) for offers when 
     * the ICS service &quot;ics_preapp&quot; was requested in the Request.
     * For offers were a url is nonsense, the value will be a null string.
     */
    public String[] getPreAppAcceptURLs ()
    {
	return (getValueArray ("preapp_accept_url"));
    }

    /**
     * Gets the tax amounts associated (tax_taxamount0..n) with 
     * the offers submitted in the ICSRequest.  Words with the 
     * &quot;ics_tax&quot; service
     */
    public String[] getTaxAmounts ()
    {
	return (getValueArray ("tax_taxamount"));
    }

    /**
     * utility method for looping through multi-entried name-value pairs.
     */
    private String[] getValueArray(String value_basename)
    {
	String values[] = new String[offer_count];
	String tmp;

	for (int offer_number = 0; offer_number < offer_count; offer_number++)
	{
	    tmp = msg.getValueByName (value_basename + offer_number);
	    values[offer_number] = (tmp == null) ? "" : tmp;
	}

	return (values);
    }
}
