/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.commerce.ics;

import java.util.*;

/**
 * A wrapper class for generating proper SCMPMessages.  This implementation
 * is not written to support access by multiple concurrent threads, 
 * so it is suggested that you avoid doing it.  Also, Cybersource 
 * does not recommend adding more than 32 offers to SCMPMessage via their
 * C library, and therefore we don't recommend it here either.  :-)
 * The characters ':', '^', ',' (colon, caret, comma) must not be in 
 * any of the names or values, as they are used by the underlying transport
 * mechanism as field separators.  There is no known way to escape these
 * characters - don't use them.
 * The addField () calls used internally are not always 
 * checked to see if they fail.  With the exception of low memory 
 * circumstances, this should not be a problem.
 * Please see the 
 * <a href="http://msc.ic3.com/html/docs/flex/intgd/intintro.html">Cybersource 
 * CommerceFLEX API documentation</a> for more information.  
 */
public class ICSRequest 
{
    private SCMPMessage msg;
    private Vector service_name_vector;
    private Vector offer_vector;
    private boolean request_sent = false;

    public ICSRequest (boolean debug_on, String hostname, String keyname) 
	throws SCMPException
    {
	msg = new SCMPMessage (debug_on);
	service_name_vector = new Vector ();
	offer_vector = new Vector ();

	request_sent = false;

	if (msg.addField ("ics_svrhost", hostname) < 0)
	{
	    throw new SCMPException ("add hostname failed for scmpmessage");
	}
	
	if (msg.addField ("ics_merchant", keyname) < 0)
	{
	    throw new SCMPException ("add merchant name failed (scmpmessage)");
	}
    }

    /**
     * request the ICS service &quot;ics_bill&quot;
     */ 
    public void requestBill (String ordernumber)
    {
	service_name_vector.addElement ("ics_bill");
	msg.addField ("order_number", ordernumber);
    }

    /**
     * request the ICS service &quot;ics_bill&quot; for the specified amount
     */ 
    public void requestBill (String ordernumber, String totalamount)
    {
	service_name_vector.addElement ("ics_bill");
	msg.addField ("order_number", ordernumber);
	msg.addField("total_amount", totalamount);
    }

    /**
     * request the ICS service &quot;ics_credit&quot;
     */
    public void requestCredit (String ordernumber, String totalamount)
    {
	service_name_vector.addElement ("ics_credit");
	msg.addField ("order_number", ordernumber);
	msg.addField ("total_amount", totalamount);
    }

    /**
     * request the ICS service &quot;ics_export&quot;
     */
    public void requestExportCheck ()
    {
	service_name_vector.addElement ("ics_export");
    }

    /**
     * request the ICS service &quot;ics_preapp&quot; Requesting this 
     * service is necessary before any funds can be transferred from 
     * customer credit cards to the account of the merchant.
     */
    public void requestPreApp ()
    {
	service_name_vector.addElement ("ics_preapp");
    }

    /**
     * Necessary for any order. 
     * There are limits on value sizes: <BR>
     * <UL>
     * <LI> firstname: 60 characters
     * <LI> lastname: 60 characters
     * <LI> email: 255 characters
     * <LI> phone: 15 characters (used for scoring, along with cc number)
     * <LI> address1: 60 characters (cc billing address)
     * <LI> city: 50 characters (cc billing address)
     * <LI> state: two characters, matching postal service abbreviations
     * <LI> zip: minimum 5, max 10 (with hyphen for zip+4)
     * <LI> country: 2 chars, 'US' for United States, or ISO3166 country code
     * <LI> ccnum: up to 20 characters (non-numerics ignored)
     * <LI> expmo: numeric - 01 through 12
     * <LI> expyr: numeric - 'yyyy' (example 1999, 2000)
     * </UL>
     */
    public void setBuyerInfo (String firstname, String lastname, String email,
			      String phone, String address1, String city, 
			      String state, String zip, String country, 
			      String ccnum, String expmo, String expyr)
    {
	msg.addField ("firstname", firstname);
	msg.addField ("lastname",  lastname);
	msg.addField ("email",     email);
	msg.addField ("phone",     phone);
	msg.addField ("address1",  address1);
	msg.addField ("city",      city);
	msg.addField ("state",     state);
	msg.addField ("zip",       zip);
	msg.addField ("country",   country);
	msg.addField ("ccnum",     ccnum);
	msg.addField ("expmo",     expmo);
	msg.addField ("expyr",     expyr);	
    }

    /**
     * Optional information to accompany an order.
     */
    public void setOptionalBuyerInfo (String company, String address2)
    {
	msg.addField ("company", company);
	msg.addField ("address2", address2);
    }

    /**
     * The following method allows you to add your own fields to the
     * SCMP message.  Use at your own discretion!  Make sure that you're
     * not redefining already existing ICS fields.  An example of using
     * this method might be adding the credit card type for the fulfillment
     * house, for their accounting purposes.  In that case, you would call
     * this method by saying:
     * <pre>    ics_req.addField ("cctype", "visa");</pre>
     * In this example, "cctype" is not a field that is defined in ICS.
     */
    public void addField (String name, String value)
    {
	msg.addField (name, value);
    }
    
    /**
     * Set optional Shipping address information.
     */
    public void setShipAddress (String shipaddress1, String shipaddress2, 
				String shipcity, String shipstate, 
				String shipzip, String shipcountry) 
    {
	msg.addField ("shipaddress1", shipaddress1);
	if (shipaddress2 != null && shipaddress2.length() > 0)
	    msg.addField ("shipaddress2", shipaddress2);
	msg.addField ("shipcity", shipcity);
	msg.addField ("shipstate", shipstate);
	msg.addField ("shipzip", shipzip);
	msg.addField ("shipcountry", shipcountry);
    }

    /**
     * set the comment field.  
     */
    public void setComments (String comments)
    {
	msg.addField ("comments", comments);
    }
    

    /**
     * set one of the 8 comment fields.  
     * @param comment the comment to enter
     * @param field_no the comment field to enter the comment into.  Can
     *        be from 0 to 7.
     */
    public void setComment(String comment, int field_no)
    {
	if (field_no < 0 || field_no > 7)
	{
	    throw new IllegalArgumentException (
		"Field numbers must be between 0 and 7");
	}
	else
	{
	    msg.addField("comment" + field_no, comment);
	}
    }

    /**
     * the SubRefNumber will be passed along to the fulfillment house.  
     * The specific contents of this field should be determined by the 
     * merchant and the fulfillment house.
     */
    public void setSubRefNumber (String sub_ref_no)
    {
	msg.addField ("sub_ref_no", sub_ref_no);
    }

    /**
     * adds the specified offer to the list (Vector) of offers associated
     * with this ICSRequest.
     */
    public void addOffer (Offer next_offer)
    {
	offer_vector.addElement (next_offer);
    }

    /**
     * deletes or removes any offers that were added to this request.
     */
    public void clearOffers ()
    {
	offer_vector.removeAllElements();
    }

    /**
     * request that the fulfillment house be notified.  Typically this 
     * is used when doing an 'ics_bill'.  
     */
    public void requestFulfillmentNotify (String who)
    {
	service_name_vector.addElement ("ics_notify");
	msg.addField ("notifywho", who);
    }

    public void requestDownload ()
    {
	service_name_vector.addElement ("ics_download");
    }
    
    /**
     * request that the fraud score be generated.  Valid score metrics 
     * will be generated for each offer.
     */
    public void requestFraudScore (String ipaddress)
    {
	service_name_vector.addElement ("ics_score");
    }

    /**
     * request the tax calculation service.  Each offer may require 
     * additional tax fields to be set.  
     */
    public void requestTaxCalculation ()
    {
	service_name_vector.addElement ("ics_tax");
    }

    /**
     * sends the request to the 
     */
    public ICSReply send () throws SCMPException 
    {
	if (request_sent)
	{
	    throw new SCMPException ("Request can only be sent once");
	}
	
	request_sent = true;

	// append the service names together in comma separated list
	// then add it to the message.
	StringBuffer sb = new StringBuffer ();

	Enumeration e = service_name_vector.elements ();
	String s;
	while (e.hasMoreElements())
	{
	    s = (String) e.nextElement ();
	    sb.append (s);
	    if (e.hasMoreElements ())
	    {
		sb.append (",");
	    }
	}
	
	// add the concatenated service request string
	msg.addField ("ics_svcs", sb.toString());

	// now add all of the offers...
	int offer_count = offer_vector.size ();
	Offer o;
	for (int i = 0; i < offer_count; i++)
	{
	    o = (Offer) offer_vector.elementAt (i);
	    msg.addField ("offer" + i, o.getEncodedString ());
	}

	// throws SCMPException
	SCMPMessage reply_msg = msg.send();
	return (new ICSReply (reply_msg, offer_count));
    }
}
