/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.commerce.ics;

import java.util.*;

/**
 * Offer is a wrapper class for the ICS Offer field of an ICS Message.
 * No fields should contain the characters '^', ',' or ':', as these
 * are used as field separators in the SCMPMessages sent to the Cybersource 
 * gateway.  
 * Please see the 
 * <a href="http://msc.ic3.com/html/docs/flex/intgd/intintro.html">Cybersource 
 * CommerceFLEX API documentation</a> for more information.  
 */
public class Offer
{
    private Hashtable ht;
    /**
     * Creates a new Offer.  ICS calls this a &quot;Product Description&quot;
     * (digital offer).  The constructor has required fields, but depending
     * upon the other services requested in the message sent to Cybersource, 
     * setting some additional fields may be required.
     */
    public Offer (String offerid, String price, String productname, 
		  String productsku, String producttype, int quantity)
    {
	ht = new Hashtable (15);
	ht.put ("offerid", offerid);
	ht.put ("price", price);
	ht.put ("productname", productname);
	ht.put ("productsku", productsku);
	ht.put ("producttype", producttype);
	ht.put ("quantity", String.valueOf (quantity));
    }

    /**
     * nexus is optional and used by the &quot;ics_tax&quot; service
     */
    public void setNexus (String s)
    {
	ht.put ("nexus", s);
    }
    
    /**
     * taxclass is optional and is used by the &quot;ics_tax&quot; service
     */
    public void setTaxClass (String s)
    {
	ht.put ("taxclass", s);
    }

    /**
     * shipamount is optional and is used by the &quot;ics_tax&quot; 
     * and &quot;ics_preapp&quot; services
     */
    public void setShipAmount (String s)
    {
	ht.put ("shipamount",s);
    }

    /**
     * taxamount is optional and is used by the &quot;ics_tax&quot; and
     * &quot;ics_preapp&quot; services
     */
    public void setTaxAmount (String s)
    {
	ht.put ("taxamount", s);
    }

    /**
     * Overrides the default score metric for this offer.  Ignored unless
     * &quot;ics_score&quot; service is requested.
     */
    public void setScoreMetric (String s)
    {
	ht.put ("scoremetric", s);
    }

    /**
     * These fields are used when requesting the &quot;ics_elc&quot; 
     * (Electronic License) service.  Fields may be left null or blank.
     */
    public void setELicenseFields (String description, String export,
				   String serial_number, String distributor,
				   String distributor_acct)
    {
	if (description != null && ! description.equals (""))
	{
	    ht.put ("description", description);
	}

	if (export != null && ! export.equals (""))
	{
	    ht.put ("export", export);
	}

	if (serial_number != null && ! serial_number.equals (""))
	{
	    ht.put ("serial_number", serial_number);
	}

	if (distributor != null && ! distributor.equals (""))
	{
	    ht.put ("distributor", distributor);
	}

	if (distributor_acct != null && ! distributor_acct.equals (""))
	{
	    ht.put ("distributor_acct", distributor_acct);
	}
    }

    /**
     * returns the string suitable for inclusion in an SCMPMessage.  
     * This method is called by ICSRequest.send() and 
     */
    public String getEncodedString ()
    {
	StringBuffer sb = new StringBuffer (128);
	Enumeration e = ht.keys();
	while (e.hasMoreElements())
	{
	    String key = (String) e.nextElement();
	    sb.append (key);
	    sb.append (":");
	    sb.append ((String) ht.get (key));
	    if (e.hasMoreElements ())
	    {
		sb.append ("^");
	    }
	}
	
	return (sb.toString());
    }
}
