/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.commerce.ics;
import java.util.*;

/**
 * A wrapper for the CyberSource C-Library ics_msg API
 * funtions essentially mirror their C counterparts.  Used by ICSRequest
 * and ICSReply.  Most often an implementor will not need to use this 
 * lower level interface, but it is provided for additional flexibility.
 * Please see the 
 * <a href="http://msc.ic3.com/html/docs/flex/intgd/intintro.html">Cybersource 
 * CommerceFLEX API documentation</a> for more information.  
 */
public class SCMPMessage
{
    private static int list_growth_chunk_size = 100;
    private static int list_length = 0; // initial number of scmp pointers
    private static Stack free_msg_index_stack;
    
    // since we don't know whether the ics C libraries are thread-safe
    // (they probably aren't, since there's no mention of
    // thread-safeness in their docs anywhere), let's assume that
    // they're not thread-safe and serialize them.  We're
    // synchronizing on free_msg_index_stack not for any particular
    // reason, but only because it's a static object variable.  The
    // following lock is used to serialize calls to the native method,
    // nSend().
    private static Integer serialization_lock = null;

    private int msg_index;  // used as index into array of *ics_msgs in C-side
    private boolean is_reply;

    static
    {
	System.loadLibrary ("lisscmp");
	free_msg_index_stack = new Stack ();
	serialization_lock = new Integer(1);
    }

    /**
     * Creates a new SCMPMessage.  
     * @param debug_on  if true, turns on debugging to the stdout of the 
     *                  C based ics_msg library.
     */
    public SCMPMessage (boolean debug_on) throws SCMPException
    {
	int debug_flag = 0;
	debug_flag = (debug_on) ? 1 : 0;

	synchronized (serialization_lock)
	{
	    msg_index = getNextMessageIndex ();
	    if (msg_index < 0)
	    {
		throw new SCMPException ("lib memory alloc failed");
	    }

	    if (initialize (debug_flag, msg_index) != 0)
	    {
		free_msg_index_stack.push (new Integer (msg_index));
		throw new SCMPException ("allocation failure");
	    }
	}
    }

    /**
     * private constructor used by send()
     */ 
    private SCMPMessage (int msg_index_in)
    {
	is_reply = true;
	msg_index = msg_index_in;
    }
    
    /** 
     * Adds a name value pair to the message.  If the name match a
     * name currently specified in the message, the name value pair
     * will be overwritten.  As of version2.4 of the Cybersource libraries, 
     * there is a limit of 64 name value pairs.
     * @param name - the name of the field. Should not be null or &quot;&quot;
     * @param value - the field value.  Can be nullstring. (&quot;&quot;) 
     * @returns the current field count, -1 if the field couldn't be added.  
     */
    public int addField (String name, String value)
    {
	if (name == null || value == null || name.equals(""))
	{
	    return (-1);
	}

	synchronized (serialization_lock)
	{
	    return (nAddField (msg_index, name, value));
	}
    }
    
    /**
     * Gets the field count.  The entire list of name-value pairs can be 
     * iterated using a 'for' loop and this number.
     * @returns the number of name value pairs in the message.
     */
    public int getFieldCount ()
    {
	synchronized (serialization_lock)
	{
	    return nGetFieldCount (msg_index);
	}
    }
    
    /**
     * gets the field name at index postion <B>field_index</B>.  Returns null
     * if the index is out of range. Field postions start at index 0.
     * @returns the field name or null
     */
    public String getName (int field_index)
    {
	synchronized (serialization_lock)
	{
	    return (nGetName (msg_index, field_index));
	}
    }

    /**
     * gets the field value at index postion <B>field_index</B>.  Returns null
     * if the index is out of range. Field postions start at index 0.
     * @returns the field value or null
     */
    public String getValue (int field_index)
    {
	synchronized (serialization_lock)
	{
	    return (nGetValue (msg_index, field_index));
	}
    }

    /**
     * Gets the field for a specified name.  If the name does not exist
     * in the message, null is returned
     * @param name - the name of the name value pair
     * @returns - the field contents, or null if field is missing
     */
    public String getValueByName (String name)
    {
	synchronized (serialization_lock)
	{
	    return (nGetValueByName (msg_index, name));
	}
    }

    /**
     * removes a field of the specified name
     */
    public int removeField (String name)
    {
	synchronized (serialization_lock)
	{
	    return (nRemoveField (msg_index, name));
	}
    }
    
    /**
     * sends a message to the ICS gateway.  The message returned is 
     * the reply.
     * @returns the reply message from the ICS gateway
     * @throws SCMPException if the message cannot be sent, if the reply 
     *                        cannot be created if this message is a response
     */
    public SCMPMessage send () throws SCMPException
    {
	if (is_reply == true)
	{
	    throw new SCMPException ("resending a received reply not valid");
	}

	synchronized (serialization_lock)
	{
	    int temp_reply_msg_index = getNextMessageIndex ();
	    if (temp_reply_msg_index < 0)
	    {
		throw new SCMPException ("required realloc failed");
	    }

	    // now we send properly.  we must pass in the reference to the
	    // current message (via index) using the clib call ics_send, a
	    // new ics_msg will hopefully be created, and stored in the
	    // array of pointers indexed by 'msg_index'
	    if (nSend (msg_index, temp_reply_msg_index) == 0)
	    {
		// successful send, use private constructor.
		return new SCMPMessage (temp_reply_msg_index);
	    }
	    else
	    {
		// doh!  
		free_msg_index_stack.push (new Integer (temp_reply_msg_index));
		throw new SCMPException ("delivery failure");
	    }
	}
    }

    /**
     * Provides debugging information and complete message contents, 
     * as a String.
     */ 
    public String toString ()
    {
	StringBuffer sb = new StringBuffer (256);
	sb.append ("----SCMPMessage Contents----\n");
	sb.append ("C Pointer value (as int32): ");

	synchronized (serialization_lock)
	{
	    sb.append (String.valueOf (getCPointer (msg_index)));

	    int pair_count = getFieldCount ();
	    sb.append ("\nTotal Pairs: ");
	    sb.append (pair_count);
	    sb.append ("\n\n");
	    for (int i = 0; i < pair_count; i++)
	    {
		sb.append ("Name: ");
		sb.append (getName (i));
		sb.append ("\nValue: ");
		sb.append (getValue (i));
		sb.append ("\n\n");
	    }
	}

	sb.append ("----END SCMPMessage Contents----\n");
	    
	return (sb.toString());
    }

    private int getNextMessageIndex ()
    {
	/* don't want to take out the inner locks either, since 
	   they were there for a reason.  If cybersources libraries are
	   found to be thread safe, simply remove the outer lock to 
	   "push down" synchronization. -thede */
	   
	synchronized (serialization_lock)
	{
	synchronized (free_msg_index_stack)
	{
	    // free_msg_index_stack is empty the first time through.
	    if (free_msg_index_stack.empty ())
	    {
		int old_length = list_length;
		list_length += list_growth_chunk_size;

		if (reallocateStorage (list_length) != 0)
		{
		    return (-1);
		}

		// insert new msg_indexs into stack
		for (int i = old_length; i < list_length; i++)
		{
		    free_msg_index_stack.push (new Integer (i));
		}
	    }

	    return(((Integer) free_msg_index_stack.pop()).intValue());
	}
	}
    }

    /**
     * When the SCMPMessages are unreferenced, the garbage collector 
     * eventually comes along to free the memory.  We need to make 
     * sure that the memory allocated in the native libraries is 
     * freed too.  Once this occurs, we can make the 'indirect handle'
     * or index available again via the stack.
     */
    protected void finalize ()
    {
	// just to be safe...
	synchronized (serialization_lock)
	{
	    // free up memory used in 'C' land, adjust stack.

	    destroy (msg_index);
	    synchronized (free_msg_index_stack)
	    {
		free_msg_index_stack.push (new Integer (msg_index));
	    }
	}
    }

    /**
     * Initialize a ics_msg in C land of the given msg_index.
     */
    private native int initialize (int debug_on, int msg_index);
    private native int getCPointer (int msg_index);

    private native int nAddField (int msg_index, String name, String value);
    private native int nGetFieldCount (int msg_index);
    private native String nGetName (int msg_index, int field_index);
    private native String nGetValue (int msg_index, int field_index);
    private native String nGetValueByName (int msg_index, String name);
    private native int nRemoveField (int msg_index, String name);
    private native int nSend (int msg_index, int reply_msg_index);

    private native int reallocateStorage (int newlength);
    private native void destroy (int msg_index);
}
