/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

#include <jni.h>
#include "leverage_commerce_ics_SCMPMessage.h"
#include <stdio.h>
#include <ics.h>
#include <string.h>

/* for outbound message memory management */
static ics_msg **msgs;
static int storage_initialized = 0;
static jint msgs_storage_current_length = 0;

/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    initialize
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_leverage_commerce_ics_SCMPMessage_initialize
(JNIEnv *env, jobject jobj, jint debug_on, jint msg_id)
{
    /* create the actual new ics_msg object, store pointer */    
    msgs[msg_id] = ics_init (debug_on);

    if (msgs[msg_id] == NULL)
    {
	return (-1); /* there was a memory error */
    }

    return (0);
}


/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    getCPointer
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_leverage_commerce_ics_SCMPMessage_getCPointer
(JNIEnv *jenv, jobject jobj, jint msg_id)
{
    return ((jint) msgs[msg_id]);
}


/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    nAddField
 * Signature: (ILjava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_leverage_commerce_ics_SCMPMessage_nAddField
  (JNIEnv *env, jobject jobj, jint msg_id, jstring jname, jstring jvalue)
{
    int retval = 0;
    const char *name; 
    const char *value; 

    /* get the strings into c-compatible form */
    name = (*env)->GetStringUTFChars (env, jname, 0);
    value = (*env)->GetStringUTFChars (env, jvalue, 0);
    
    retval = ics_fadd(msgs[msg_id], (char *) name, (char *) value);

    (*env)->ReleaseStringUTFChars (env, jname, name);
    (*env)->ReleaseStringUTFChars (env, jvalue, value);
    return ((jint) retval);
}


/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    nGetName
 * Signature: (II)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_leverage_commerce_ics_SCMPMessage_nGetName
  (JNIEnv *env, jobject jobj, jint msg_id, jint field_index)
{
    char *name;

    /* check index */
    int cnt = ics_fcount (msgs[msg_id]);
    if (field_index < 0 || field_index >= cnt)
    {
	return (NULL);
    }

    name = msgs[msg_id]->names[field_index];
    if (name == NULL || (strcmp (name, "") == 0))
    {
	return (NULL);
    }
    else
    {
	/* return a java.lang.String object somehow. */
	jstring retstr = (*env)->NewStringUTF (env, name);
	return (retstr);
    }
}

/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    nGetValue
 * Signature: (II)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_leverage_commerce_ics_SCMPMessage_nGetValue
  (JNIEnv *env, jobject jobj, jint msg_id, jint field_index)
{
    char *value;

    /* check index */
    int cnt = ics_fcount (msgs[msg_id]);
    if (field_index < 0 || field_index >= cnt)
    {
	return (NULL);
    }

    value = msgs[msg_id]->values[field_index];
    if (value == NULL || (strcmp (value, "") == 0))
    {
	return (NULL);
    }
    else
    {
	/* return a java.lang.String object somehow. */
	jstring retstr = (*env)->NewStringUTF (env, value);
	return (retstr);
    }
}


/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    nGetValueByName
 * Signature: (ILjava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_leverage_commerce_ics_SCMPMessage_nGetValueByName
  (JNIEnv *env, jobject jobj, jint msg_id, jstring jname)
{
    const char *field;
    const char *name;

    name = (*env)->GetStringUTFChars (env, jname, 0);
    field = ics_fgetbyname (msgs[msg_id], (char *) name);
    (*env)->ReleaseStringUTFChars (env, jname, name);

    if (field == NULL)
    {
	return (NULL);
    }
    else
    {
	/* return a java.lang.String object somehow. */
	jstring retstr = (*env)->NewStringUTF (env, field);
	return (retstr);
    }
}

/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    nGetFieldCount
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_leverage_commerce_ics_SCMPMessage_nGetFieldCount
  (JNIEnv *env, jobject jobj, jint msg_id)
{
    return ((jint) ics_fcount (msgs[msg_id]));
}


/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    nRemoveField
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_leverage_commerce_ics_SCMPMessage_nRemoveField
  (JNIEnv *env, jobject jobj, jint msg_id, jstring jname)
{
    int result = 0;
    const char *name = (*env)->GetStringUTFChars (env, jname, 0);

    result = ics_fremove (msgs[msg_id], (char *) name);

    /* release char array created above */
    (*env)->ReleaseStringUTFChars (env, jname, name);

    return (result);
}


/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    nSend 
 * Signature: (I,I)I
 */
JNIEXPORT jint JNICALL Java_leverage_commerce_ics_SCMPMessage_nSend
(JNIEnv *jenv, jobject jobj, jint msg_id, jint reply_msg_id)
{
    ics_msg *new_msg; 

    new_msg = ics_send (msgs[msg_id]);
    if (new_msg == NULL)
    {
	return (-1);
    }
    
    msgs[reply_msg_id] = new_msg;
    return (0);
}



/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    reallocateStorage
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_leverage_commerce_ics_SCMPMessage_reallocateStorage
(JNIEnv *env, jobject jobj, jint new_length)
{
    /* here is where we manage the pointer array allocation... */
    if (storage_initialized == 0)
    {
	msgs = (ics_msg **) malloc (new_length * sizeof(ics_msg *));
	if (msgs == NULL)
	{
	    return (-1);
	}

	msgs_storage_current_length = new_length;
	storage_initialized = 1;
	return (0);
    }
    else if (new_length > msgs_storage_current_length)
    {
	int i = 0;
	ics_msg **old_msgs;
	ics_msg **new_msgs = 
	    (ics_msg **) malloc (new_length * sizeof(ics_msg *));

	if (new_msgs == NULL)
	{
	    return (-1);
	}
	
	/* copy the contents of the previous list into the new one */
	for (i = 0; i < msgs_storage_current_length; i++)
	{
	    new_msgs[i] = msgs[i];
	}
	
	old_msgs = msgs;
	msgs = new_msgs;
	
	msgs_storage_current_length = new_length;
	free (old_msgs);
	return (0);
    }
}



/*
 * Class:     leverage_commerce_ics_SCMPMessage
 * Method:    destroy
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_leverage_commerce_ics_SCMPMessage_destroy
(JNIEnv *env, jobject jobj, jint msg_id)
{
    ics_destroy (msgs[msg_id]);
    msgs[msg_id] = NULL;
}


