/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Candleweb AS 
 * Portions created by Candleweb AS are Copyright (C) 1999 Candleweb AS. 
 * All Rights reserved.
 */
package org.locomotive.directory;

import java.util.*;

/**
 * A class representing a bean configuration.
 *
 * @author Gunnar Rnning (gunnar@candleweb.no).
 * @version $Id: BeanConfiguration.java,v 1.2 1999/05/05 10:19:31 gunnar Exp $
 */
public class BeanConfiguration {
    String scopeType = "default";
    Class clazz;
    String className;
    Hashtable properties = new Hashtable();

    // contains the same BeanProperty objects as the properties Hashtable, 
    // but the BeanProperty values occur in the order they were declared.
    Vector orderedProps = new Vector();

    DirectoryContext directoryContext;
    String fullName;
    // When the this bean configuration last was read from the file system
    // using System.currentTimeMillis().
    protected long lastReadTime;

    /**
     * Constructor for BeanConfiguration. 
     * @param context The DirectoryContext for this bean configuration.
     */
    public BeanConfiguration(DirectoryContext context){
	this.directoryContext = context;
    }

    /**
     * Get the scope type for this bean. The method returns "default"
     * if no scope was defined in the configuration of this bean.
     */
    public String getScopeType(){
	return scopeType;
    }

    /**
     * Set the scope type for this bean.
     * @param scopeType A string containg the scopeType for this component.
     */
    public void setScopeType(String scopeType){
	this.scopeType = scopeType;
    }

    /**
     * Get class for this bean configuration.
     */
    public Class getClazz(){
	return clazz;
    }

    /**
     * Set the class for this bean configuration.
     */
    public void setClazz(Class clazz){
	this.clazz = clazz;
    }

    /**
     * Set the fully qualified classname for this bean configuration.
     * @param className The class name for this bean.
     */
    public void setClassName(String className){
	this.className = className;
    }

    /**
     * Get the fully qualified classname for this bean configuration.
     * @return The the class for this bean configuration.
     */
    public String getClassName(){
	return className;
    }

    /**
     * Get a bean property by name.
     * @return The specified property if defined, or null otherwise.
     */
    public BeanProperty getBeanProperty(String name){
	return (BeanProperty) properties.get(name);
    }
  
    /**
     * Add a bean property to this bean configuration.
     * @param beanProperty The bean property to add, the name of 
     *        beanProperty will be the key for retrieving the property.
     */
    public void addBeanProperty(BeanProperty beanProperty){
	properties.put(beanProperty.getName(), beanProperty);
	orderedProps.addElement(beanProperty);
    }

    /**
     * Get all the properties defined for this bean.
     * @return an Enumeration with all the BeanProperty objects of this
     *         configuration.
     */
    public Enumeration getBeanProperties(){
	return orderedProps.elements();
    }

    /**
     * Get the directory context for this configuration.
     */
    public DirectoryContext getDirectoryContext(){
	return directoryContext;
    }

    /**
     * Set the directory context for this configuration.
     * @param directoryContext The directory context for this 
     *        bean configuration. 
     */
    public void setDirectoryContext(DirectoryContext directoryContext){
	this.directoryContext = directoryContext;
    }

    /**
     * Get the time this configuration last was read from the file system. 
     * The time stamp is generated using System.currentTimeMillis().
     * @return Return The time this configuration last was read from the 
     *	     file system. 
     */
    public long getLastReadTime(){
	return lastReadTime;
    }

    /**
     * Set the time this configuration last was read from the file system.
     * @param lastReadTime The time we last read this bean configuration 
     *        from the file system, expressed using System.currentTimeMillis().
     */
    public void setLastReadTime(long lastReadTime){
	this.lastReadTime = lastReadTime;
    }

    /**
     * Set the full name of this configuration.
     * @param fullName The full name of this configuration.
     */
    public void setFullName(String fullName){
	this.fullName = fullName;
    }

    /**
     * Get the full name of this configuration.
     * @return Return the full name of this configuration.
     */
    public String getFullName(){
	return fullName;
    }
}
