/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Candleweb AS 
 * Portions created by Candleweb AS are Copyright (C) 1999 Candleweb AS. 
 * All Rights reserved.
 */
package org.locomotive.directory;

import java.lang.reflect.*;
import java.util.*;

/**
 * A NameService object delegates the responsibility for creating new bean 
 * instances from BeanConfiguration objects to a object of this class. 
 * The BeanInstantiator uses the Java reflection mechanism to
 * instatiate beans with the the configured values.
 *
 * @author Gunnar Rnning (gunnar@candleweb.no).
 * @version $Id: BeanInstantiator.java,v 1.4 1999/05/05 10:19:31 gunnar Exp $
 */
public class BeanInstantiator {
  private NameService nameService;

  /**
   * Constructor for the BeanInstantiator. 
   * @param nameService The NameService that will collaborate with 
   *    this BeanInstantiator.
   */
  BeanInstantiator(NameService nameService) {
    this.nameService = nameService;
  }

  /**
   * Created an instance of a component.
   * @param beanConfig A bean configuration object.
   * @return The instantiated component.
   */
  protected Object createBeanInstance(BeanConfiguration beanConfig)
  {
    Class clazz = getClassObject(beanConfig);
    if(clazz == null)
      return null;
    Object o;
    try {
      o = clazz.newInstance();
    } catch(IllegalAccessException iae){
      nameService.dispatchErrorEvent(this, "Instantiation of class = " + 
				     clazz.getName() + " failed.", 
				     beanConfig, null, iae);
      return null;
    } catch(InstantiationException iae){
      nameService.dispatchErrorEvent(this, "Instantiation of class = " + 
				     clazz.getName() + " failed.", 
				     beanConfig, null, iae);
      return null;
    } catch(Throwable t){
      nameService.dispatchErrorEvent(this, "Instantiation of class = " + 
				     clazz.getName() + " failed.", 
				     beanConfig, null, t);
      return null;
    }
    return o;
  }


  /**
   * Set the properties of a component according to properties
   * set in a BeanConfiguration object.
   * @param scopeLookup The scope lookup component. This must be provided
   *                    in the bean configuration refers to other components. 
   * @param component The component to set the properties on.
   * @param beanConfig A bean configuration object.
   */
  protected void setProperties(ScopeLookup scopeLookup, Object component,
			       BeanConfiguration beanConfig)
  {
    Enumeration e = beanConfig.getBeanProperties();
    while(e.hasMoreElements()){
      BeanProperty p = (BeanProperty) e.nextElement();
      setBeanProperty(component, p, scopeLookup, beanConfig);
    }    

    if(component instanceof LocomotiveService){
      LocomotiveService ls = (LocomotiveService) component;
      try {
	ls.startService();
      } catch(ServiceException se){
	nameService.dispatchErrorEvent(this, "startService() failed.",
				       beanConfig, null, se);
      }
    }
  }



  /**
   * Set a given bean property in an object. If an error occurs, the 
   * property is not set.
   */
  protected void setBeanProperty(Object o, BeanProperty p, 
				 ScopeLookup scopeLookup, 
				 BeanConfiguration beanConfig)
				   
  {
    String propName = p.getName();
    String methodName = "set" + propName.substring(0, 1).toUpperCase() + 
      propName.substring(1);

    Class c = o.getClass();

    // Assumes that there is only one setMethod with this name.
    // Can we do this according to the spec. ?
    Method methods[] = c.getMethods();
    Method method = null;
    for(int i = 0; i < methods.length; i++){
      if(methodName.equals(methods[i].getName())){
	method = methods[i];
	break;
      }
    }
    if(method == null){
	nameService.dispatchErrorEvent(this, "Method not found for property.", 
				       beanConfig, p, null);
	return;
    }
    Vector values = p.getValues();
    
    Class pars[] = method.getParameterTypes();
    if(pars.length == 1){
      Class convClass = null;
      Vector v = null;
      Object arr = null;
      Object res = null;
      if(pars[0] == Vector.class){
	res = v = new Vector();
      } else if(pars[0].isArray()){
	convClass = pars[0].getComponentType();
	res = arr = Array.newInstance(convClass, values.size());
      } else {
	convClass = pars[0];
      }

      PropertyConverter converter = null;
      if(convClass != null) 
	converter = nameService.getPropertyConverter(convClass);

      for(int i = 0; i < values.size(); i++){
	BeanPropertyValue value = (BeanPropertyValue) values.elementAt(i);
	String propval = value.getValue();
	Object convval = null; 
	if(value.getType() == BeanPropertyValue.CONSTANT_TYPE){
	  if(converter != null) {
	      try {
		  convval = converter.convertProperty(propval);
	      } catch (Exception e) {
		  nameService.dispatchErrorEvent(this, 
						 "Setting of property failed.",
						 beanConfig, p, e);
	  
		  return;
	      }
	  } else
	    convval = propval;
	} else if(value.getType() == BeanPropertyValue.COMPONENT_TYPE){

	  nameService.logDebug(beanConfig, 
			       "Looking up component = " + propval);

	  convval = nameService.resolve(scopeLookup, 
					beanConfig.getDirectoryContext(),
					propval);
	  if(converter != null){
	      try {
		  convval = converter.convertProperty(convval);
	      } catch (Exception e) {
		  nameService.dispatchErrorEvent(this, 
						 "Setting of property failed.",
						 beanConfig, p, e);
	  
		  return;
	      }
	  }
	  nameService.logDebug(beanConfig, "Result was = " + convval);
	}
	if(v != null)
	  v.addElement(convval);
	else if(arr != null){
	  try {
	    Array.set(arr, i, convval);
	  } catch(Exception e){
	    nameService.dispatchErrorEvent(this, "Setting of property failed.",
					   beanConfig, p, e);
	    return;
	  }
	} else {
	  res = convval;
	}
      }
      Object objs[] = { res };
      
      try {
	method.invoke(o, objs);
      } catch(Exception e){
	nameService.dispatchErrorEvent(this, "Setting of property failed.",
				       beanConfig, p, e);
	return;
      }

    } else {
      if(pars.length > 1){

      }
	
    }
  }

  

  /**
   * Get the Java Class object for a bean configuration.
   */
  private Class getClassObject(BeanConfiguration beanConfig){
    Class clazz = beanConfig.getClazz();
    if(clazz == null){
      try {
	clazz = Class.forName(beanConfig.getClassName());
	beanConfig.setClazz(clazz);
      } catch(ClassNotFoundException ce){
	nameService.dispatchErrorEvent(this, "Loading of class=" + 
				       beanConfig.getClassName() + " failed.",
				       beanConfig, null, ce);
      }
    }
    return clazz;
  }

  public String capitalize(String str){
    String firstPart = str.substring(0, 1);
    String lastPart = str.substring(1);
    firstPart = firstPart.toUpperCase();
    return firstPart + lastPart;
  }

  /**
   * Get a property for an object.
   * @param beanConfig The BeanConfiguration context for this lookup.
   * @param object The object to lookup the property in.
   * @param property The property to lookup.
   * @return The property if it exists, or null else. Properties of 
   *         primitive types are wrapped.
   */
  public Object getProperty(BeanConfiguration beanConfig, 
			    Object object, String property)
  {
      if(property == null){
	  nameService.dispatchErrorEvent(this, "Cannot resolve null property.",
					 beanConfig, null, null);
	  return null;
      }
      if(object == null) {
	  nameService.dispatchErrorEvent(this, "Property lookup failed " + 
					 "because component was null.",
					 beanConfig, null, null);
	  
	  return null;
      }


    int nextDot = property.indexOf('.');
    String nextPart;
    if(nextDot == -1)
      nextPart = property;
    else
      nextPart = property.substring(0, nextDot);
    int openBracket = nextPart.indexOf('[');

    String methodName = "get" + capitalize(nextPart);
    Class clazz = object.getClass();
    Method method = null;
    
    try {
      method = clazz.getMethod(methodName, null);
    } catch(NoSuchMethodException nme){
      nameService.dispatchErrorEvent(this, "Method = " + methodName + 
				     "() not found in object of class = " + 
				     clazz.getName() + ".", 
				     beanConfig, null, nme);
      return null;
    }
    Object nextComp = null;

    try {
      nextComp = method.invoke(object, null);
    } catch(Exception e){
      nameService.dispatchErrorEvent(this, "Invocation of method = " + 
				     methodName + 
				     "() failed in object of class = " + 
				     clazz.getName() + ".", 
				     beanConfig, null, e);
      return null;
    }
    // TODO handle array indexing for Vector, array and indexed properties.

    if(nextDot != -1){
      nextComp = getProperty(beanConfig, nextComp, 
			     property.substring(nextDot + 1));
    }
    return nextComp;
  }

}
