/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Candleweb AS 
 * Portions created by Candleweb AS are Copyright (C) 1999 Candleweb AS. 
 * All Rights reserved.
 */
package org.locomotive.directory;

import java.util.EventObject;

/**
 * All events that occurs when using the bean directory will generate
 * this event.
 *
 * @author Gunnar Rnning (gunnar@candleweb.no).
 * @version $Id: DirectoryEvent.java,v 1.1 1999/05/05 10:19:32 gunnar Exp $
 */
public class DirectoryEvent extends EventObject {
  protected String message;
  protected BeanConfiguration beanConfig;
  protected BeanProperty beanProperty;
  protected Throwable throwable;
  int type;

  /**
   * Constructor for the DirectoryEvent.
   * @param type The type of DirectoryEvent, see DirectoryEventTypes for
   *		 valid types.
   * @param source The object that generated the event.
   * @param message Descriptive message of the event.
   */
  public DirectoryEvent(int type, Object source, String message){
   this(type, source, message, null, null);
  }

  /**
   * Constructor for the DirectoryEvent.
   * @param type The type of DirectoryEvent, see DirectoryEventTypes for
   *		 valid types.
   * @param source The object that generated the event.
   * @param message Descriptive message of the event.
   * @param beanConfig The BeanConfiguration processed when the event occured, 
   *		       or null.
   */
  public DirectoryEvent(int type, Object source, String message, 
			BeanConfiguration beanConfig){    
    this(type, source, message, beanConfig, null);
  }


  /**
   * Constructor for the DirectoryEvent.
   * @param type The type of DirectoryEvent, see DirectoryEventTypes for
   *		 valid types.
   * @param source The object that generated the event.
   * @param message Descriptive message of the event.
   * @param beanConfig The BeanConfiguration processed when the event occured, 
   *		       or null.
   * @param beanProperty The BeanProperty processed when the event occured, 
   *		         or null.
   */
  public DirectoryEvent(int type, Object source, String message, 
			     BeanConfiguration beanConfig, 
			     BeanProperty beanProperty){
    this(type, source, message, beanConfig, beanProperty, null);
  }
  
  /**
   * Constructor for the DirectoryEvent.
   * @param type The type of DirectoryEvent, see DirectoryEventTypes for
   *		 valid types.
   * @param source The object that generated the event.
   * @param message Descriptive message of the event.
   * @param beanConfig The BeanConfiguration processed when the event occured, 
   *		       or null.
   * @param beanProperty The BeanProperty processed when the event occured, 
   *		         or null.
   * @param throwable The Throwable that triggered the event, or null.
   */
  public DirectoryEvent(int type, Object source, String message, 
			BeanConfiguration beanConfig, 
			BeanProperty beanProperty, Throwable throwable){
    super(source);
    this.message = message;
    this.beanConfig = beanConfig;
    this.beanProperty = beanProperty;
    this.throwable = throwable;
    this.type = type;
  }
  
  /**
   * Get the type of this directory event. See DirectoryEventTypes for valid
   * types.
   * @return The type of this directory event.
   */
  public int getType(){
    return type;
  }

  /**
   * Get a descriptive message of the event.
   * @return Returns a String containing a descriptive message of the event.
   */
  public String getMessage(){
    return message;
  }

  /**
   * If this event occured in the context of a particular bean 
   * configuration, this method will return it.
   * @return Return a BeanConfiguration or null if the event didn't occur
   *         in the context of a particular BeanConfiguration.
   */
  public BeanConfiguration getBeanConfig(){
    return beanConfig;
  }

  /**
   * If this event occured in the context of a particular bean 
   * property, this method will return it.
   * @return Return a BeanProperty or null if the event didn't occur
   *         in the context of a particular BeanProperty.
   */
  public BeanProperty getBeanProperty(){
    return beanProperty;
  }

  /**
   * If this event was triggered by an throwable, this method will 
   * return that throwable.
   * @return An throwable or null if the event wasn't triggered 
   *         by an throwable. 
   */
  public Throwable getThrowable(){
    return throwable;
  }
}

