/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Candleweb AS 
 * Portions created by Candleweb AS are Copyright (C) 1999 Candleweb AS. 
 * All Rights reserved.
 */
package org.locomotive.directory;

/**
 * NameParse provides a static method for parsing component references. 
 * A component name can be on the form 
 * "/locomotive/server/httpd.serverName". 
 * This is then partioned into : 
 * <ul>
 * <li> The directory context : /locomotive/server/
 * <li> The component name : httpd
 * <li> The property part : serverName.
 * </ul>
 *
 * @author Gunnar Rnning (gunnar@candleweb.no).
 * @version $Id: NameParse.java,v 1.2 1999/05/05 10:19:32 gunnar Exp $
 */
public class NameParse {
  DirectoryContext context;
  String name;
  String property;

  /**
   * Parse a component name relative to a directory context, and return 
   * a NameParse object containing the different segments of the parse.
   */
  public static NameParse parseName(DirectoryContext context, 
				    String componentName){
    if(componentName == null)
      return null;

    NameParse nameParse = new NameParse();

    int lastSlash = componentName.lastIndexOf('/');
    
    String contextPart = null;
    if(lastSlash != -1)
      contextPart = componentName.substring(0, lastSlash);

    String componentPart = componentName.substring(lastSlash + 1);
    int firstDot = componentPart.indexOf('.');
    
    DirectoryContext absoluteContext = 
      DirectoryContext.getDirectoryContext(context, contextPart);
    nameParse.context = absoluteContext;
    
    if(firstDot == -1){
      nameParse.name = componentPart;
    } else {
      nameParse.name = componentPart.substring(0, firstDot);
      nameParse.property = componentPart.substring(firstDot + 1);
    }
    return nameParse;
  }

  /**
   * Get the directory context part of this NameParse.
   */ 
  public DirectoryContext getContext(){
    return context;
  }
  
  /**
   * Get the component name part of this NameParse.
   */
  public String getName(){
    return name;
  }

  /**
   * Get the property part of this NameParse, or null if
   * there is no property part.
   */
  public String getProperty(){
    return property;
  }

  public static void main(String args[]){
    DirectoryContext context;
    String contexts[] = { "/server", 
			  "/server/httpd/../snmp/",
			  "../server/httpd/../snmp", 
			  "/server/httpd/child/../snmp/", 
			  "/server/httpd/../snmp/.."
    };
    String components[] = {
      "/locomotive/server/httpd.serverName", 
      "../httpd", 
      "httpd.serverName"
    };

    for(int i = 0; i < contexts.length; i++){
      context = DirectoryContext.getDirectoryContext(contexts[i]);
      for(int j = 0; j < components.length; j++){
	NameParse nameParse = NameParse.parseName(context, components[j]);
	System.out.println(nameParse.getContext() + ":" + 
			   nameParse.getName() + ":" + 
			   nameParse.getProperty());
      }
    }
    
  }


 
}
