/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Candleweb AS 
 * Portions created by Candleweb AS are Copyright (C) 1999 Candleweb AS. 
 * All Rights reserved.
 */
package org.locomotive.directory;

import java.util.*;
import java.io.*;

/**
 * A naming service for Java components. This is the core
 * functionality provided by NameService : 
 * <ul>
 * <li> A hierarchical naming convention for Java components.
 * <li> Automatic instantation and configuration of Java components backed by 
 *      configuration files. 
 * <li> Scoping of components, provides the functionality for implementing 
 *      session and request scoped components.
 * </ul>
 *
 * <p> 
 * The configurations is organized in a tree structure of DirectoryContexts.
 * 
 * <h3> Synopsis </h3>
 * 
 * <h3> Open Issues </h3>
 * <ul>
 * <li> The default storage mechanism for configurations is on the local 
 *      filesystem. It is however conceivable that some might wish to read 
 *      configurations from other sources like databases, web servers etc.
 *      To accomplish this I hope it is enough to subclass this NameService
 *      class and override the readBeanConfiguration method.
 * <li> Should allow for components to be looked up without being 
 *      associated with any scope.
 * <li> 
 * </ul>
 *
 * @author Gunnar Rnning (gunnar@candleweb.no).
 * @version $Id: NameService.java,v 1.4 1999/05/05 10:19:32 gunnar Exp $
 */
public class NameService implements DirectoryEventTypes {
  Hashtable scopes = new Hashtable();
  String configPath[];
  int checkIntervalSeconds = -1;
  BeanParser beanParser;
  String configFileExtension = ".cfg";
  Hashtable propertyConverters = new Hashtable();
  Vector eventListeners = new Vector();
  boolean logDebug = false;

  BeanInstantiator beanInstantiator;
  
  // Hashtable containing BeanConfiuration objects representing 
  // parsed configuration files.
  Hashtable beanConfigurations = new Hashtable();

  /**
   * Create a new NameService.  Relative entries in the config path are
   * computed relative to the current directory of the Java application.
   * @param configPath A String array containing the configuration path.
   */
  public NameService(String configPath[]){
    this(null, configPath);
  }
  
  /**
   * Create a new NameService.  Relative entries in the config path are
   * computed relative to the current directory of the Java application.
   *
   * @param beanParser The parser to use when parsing configuration files.
   * @param configPath A String array containing the configuration path.
   */
  public NameService(BeanParser beanParser, String configPath[]){
    if(beanParser == null)
      this.beanParser = new XmlBeanParser(this);
    else
      this.beanParser = beanParser;
    this.configPath = configPath;
    beanInstantiator = new BeanInstantiator(this);

    addPropertyConverter(String.class, new StringConverter());
    addPropertyConverter(Integer.TYPE, new IntegerConverter());
    addPropertyConverter(Integer.class, new IntegerConverter());
  }

  public NameService(){
    this(null, null);
  }

  public void setConfigPath(String configPath[]){
    this.configPath = configPath;
  }
  
  /**
   * Resolve a component relative to a given directory context.
   * @param scopeLookup The object responsible for looking up the scope
   *        for this component.
   * @param context The directory context to resolve this name relative to.
   * @param name The name of the component.
   * @return Returns the resolved component or null if the component 
   *         not could be resolved. 
   **/
  public Object resolve(ScopeLookup scopeLookup, DirectoryContext context, 
			String name)
  {
    NameParse nameParse = NameParse.parseName(context, name);

    BeanConfiguration beanConfig = 
      getBeanConfiguration(nameParse.getContext(), nameParse.getName());

    if(beanConfig == null){
      return null;
    }
    Object scope = scopeLookup.getScope(beanConfig.getScopeType());
    if(scope == null){
      return null;
    }
    ScopeStore scopeStore = (ScopeStore) scopes.get(scope);

    Object resolvedComponent = null;

    if(scopeStore != null){
      resolvedComponent = scopeStore.getComponent(nameParse.getContext(), 
						  nameParse.getName());
    } else {
      synchronized(scopes){
	scopeStore = (ScopeStore) scopes.get(scope);
	if(scopeStore == null){
	  scopeStore = new ScopeStore();
	  scopes.put(scope, scopeStore);
	}
      }
    }
    
    if(resolvedComponent == null){
      synchronized(scopeStore) {
	resolvedComponent = scopeStore.getComponent(nameParse.getContext(), 
						    nameParse.getName());

	if(resolvedComponent == null) {
	  resolvedComponent = 
	    beanInstantiator.createBeanInstance(beanConfig);

	  if(resolvedComponent != null){
	    scopeStore.putComponent(nameParse.getContext(), 
				    nameParse.getName(), 
				    resolvedComponent);
	    beanInstantiator.setProperties(scopeLookup, resolvedComponent, 
					 beanConfig);
	  }
	}
      }
    }
    
    if(nameParse.getProperty() != null)
      return beanInstantiator.getProperty(beanConfig, resolvedComponent, 
					  nameParse.getProperty());
    else 
      return resolvedComponent;
  }

  /**
   * Resolve a component with a given scope. 
   * @param scopeLookup The object responsible for looking up the scope
   *        for this component.
   * @param name The name of the component.
   * @return Returns the resolved component or null if the component 
   *         not could be resolved. 
   */
  public Object resolve(ScopeLookup scopeLookup, String name){
    return resolve(scopeLookup, null, name);
  }

  /**
   * Register a component with a given name and scope. This method is 
   * useful in two instances :
   * <ol>
   * <li> When bootstrapping the NameService, you want it to know about
   *      itself.
   * <li> Generally when you want to register components that are created
   *      by other means than configuration files.
   * </ol>
   * @param scopeLookup The object used to lookup the scope for this 
   *        component.
   * @param scopeType The scope type for this component. This is only
   *                  set if no configuration file is found for the
   *                  component. 
   * @param name      The full name of the component.
   * @param component The component to register.
   */
  public void registerComponent(ScopeLookup scopeLookup, String scopeType,
				String name, Object component)
  {
    NameParse nameParse = NameParse.parseName(null, name);

    BeanConfiguration beanConfig = 
      getBeanConfiguration(nameParse.getContext(), nameParse.getName());

    if(beanConfig == null){
      beanConfig = new BeanConfiguration(nameParse.getContext());
      beanConfig.setScopeType(scopeType);
      beanConfigurations.put(nameParse.getContext().getContext(), 
			     nameParse.getName());
    }
    Object scope = scopeLookup.getScope(beanConfig.getScopeType());
    if(scope == null){
      return;
    }

    ScopeStore scopeStore = (ScopeStore) scopes.get(scope);

    if(scopeStore == null){
      scopeStore = new ScopeStore();
      scopes.put(scope, scopeStore);
    }
    
    scopeStore.putComponent(nameParse.getContext(), nameParse.getName(), 
			    component);
  }

  /**
   * Remove a scope from the  NameService. This removes references to 
   * objects associated with this scope.
   * @param scope The scope to remove.
   */
  public void removeScope(Object scope){
    scopes.remove(scope);
  }

  /**
   * Get the BeanConfiguration for a given component.
   * @param context The DirectoryContext to get this component is in.
   * @param name The name of the component.
   * @return The BeanConfiguration object for this component.
   */
  protected BeanConfiguration getBeanConfiguration(DirectoryContext context, 
						   String name){
    String configName = context.getContext() + name;
    BeanConfiguration beanConfig = 
      (BeanConfiguration) beanConfigurations.get(configName);

    if(beanConfig != null){
      if(checkIntervalSeconds < 0){
	return beanConfig;
      }
      
      // The bean configuration is valid if the time it last was read  
      // is checkIntervalSeconds 
      if((beanConfig.getLastReadTime() + (checkIntervalSeconds * 1000)) < 
	 System.currentTimeMillis())
      {
	return beanConfig;
      }
    } 

    // We can synchronize on context since all context are created through
    // factory pattern and there can only be one DirectoryContext object 
    // for a directory.
    synchronized(context){
      if(beanConfig == null){
	beanConfig = (BeanConfiguration) beanConfigurations.get(configName);
	if(beanConfig != null)
	  return beanConfig;
      } else {
	beanConfigurations.remove(configName);
      }
      
      beanConfig = readBeanConfiguration(context, name);


      if(beanConfig != null){
	beanConfig.setLastReadTime(System.currentTimeMillis());	
	beanConfigurations.put(configName, beanConfig);
      }

    }
    
    return beanConfig;
  }

  /**
   * Search one directory in the filesystem for a bean configuration file, 
   * and update the associated with the values in this file. If a 
   * configuration file for the bean was not found, the beanConfig object
   * will not be modified.
   * @param beanConfig The BeanConfiguration object to update.
   * @param configDir The filesystem directory used as the root when 
   *                  looking up this component.
   * @param name      The name of the bean configuration. 
   */
  protected void updateBeanConfigFromDir(BeanConfiguration beanConfig, 
					 String configDir, 
					 String name)
  {
    StringBuffer fileName = new StringBuffer(configDir);
    
    // Append file system dependent separator if it is not there already.
    if(!configDir.endsWith(File.separator))
      fileName.append(File.separator);
    
    int currentIndex = 0;

    while(true) {
      /**
       * INVARIANT: fileName always ends with a fileseparator character
       * when at the top of the loop.
       */

      int nextIndex = name.indexOf('/', currentIndex);
      
      if(nextIndex == -1) {
	fileName.append(name.substring(currentIndex));
	break;
      }
      /**
       * Note: If nextIndex == currentIndex means that we do not
       * have any new directory name in this iteration, so
       * we should not add another fileseparator character either.
       */ 
      if(nextIndex != currentIndex){
	fileName.append(name.substring(currentIndex, nextIndex));
	fileName.append(File.separator);
      }
      currentIndex = nextIndex + 1;
    }     
    fileName.append(configFileExtension);

    beanParser.parseBeanConfiguration(fileName.toString(), beanConfig);
  }
  



  /**
   * Search the config path for a bean configuration file matching a name, 
   * and return a BeanConfiguration object describing the configuration.
   * <h4> Subclassing </h4>
   * This method could be overridden by subclasses to read bean configurations
   * from sources other than the local filesystem.
   *
   * @param context The directory context of the bean configuration.
   * @param name The name of the bean configuration.
   * @return Returns a BeanConfiguration object, or null if the configuration
   *         was not found.
   */ 
  protected BeanConfiguration readBeanConfiguration(DirectoryContext context, 
						    String name)
  {
    String absName = context.getContext() + name;
    BeanConfiguration beanConfig = new BeanConfiguration(context);
    beanConfig.setFullName(absName);
    for(int i = 0; i < configPath.length; i++){
      updateBeanConfigFromDir(beanConfig, configPath[i], absName);
    }

    // If no class name is found return null. 
    if(beanConfig.getClassName() == null)
      return null;

    return beanConfig;
  }



  /**
   * Check seconds defines how many seconds a BeanConfiguration read from 
   * file is valid before it will be reread if changed. 
   * When deploying an application it is recommended that
   * you set this parameter to -1(the default) to improve performance
   * on component lookups. In a development situtation you might want to 
   * set this property to a low so changes in your
   * configuration files are reflected without 
   * restarting the application. 
   * @param checkSeconds How often a BeanConfiguration should be checked
   *                     for changes on file. 
   *                     -1 is the default value means that configuration
   *                     files not will be reparsed. If the value is 0 the 
   *                     configuration file will always be reread if changed.
   *                     
   */
  public void setCheckIntervalSeconds(int checkIntervalSeconds){
    this.checkIntervalSeconds = checkIntervalSeconds;
  }

  /**
   * Get the check interval seconds property. See setCheckIntervalSeconds
   * for a detailed description.
   */
  public int getCheckIntervalSeconds(){
    return checkIntervalSeconds;
  }

  /**
   * The parser to use when parsing bean configuration files.
   */
  public BeanParser getBeanParser(){
    return beanParser;
  }
  
  /**
   * Set the parser to use when parsing bean configuration files.
   * @param beanParser Object implementing the BeanParser interface.
   */
  public void setBeanParser(BeanParser beanParser){
    this.beanParser = beanParser;
  }

  /**
   * Add property converter.
   * Registers a new PropertyConverter object for a specific class.
   * @param clazz The class of objects this PropertyConverter creates.
   * @param converter An object implementing the PropertyConverter interface.
   */
  public void addPropertyConverter(Class clazz, PropertyConverter converter){
    propertyConverters.put(clazz, converter);
  }
  
  /**
   * Get the property converter for a specific class.
   * @param clazz The class the class to get the property converter for.
   * @return The PropertyConverter object for this clazz, or null if
   *         there no property converter is registered.
   */
  public PropertyConverter getPropertyConverter(Class clazz){
    return (PropertyConverter) propertyConverters.get(clazz);
  }
  
  /**
   * Add a directory event listener. A DirectoryEventListener will 
   * receive event notifications on all events in this directory.
   * @param listener The DirectoryEventListener to add.
   */
  public void addDirectoryEventListener(DirectoryEventListener listener){
    eventListeners.addElement(listener);
  }

  /**
   * Remove a directory event listener.
   * @param listener The DirectoryEventListener to remove.
   */
  public void removeDirectoryEventListener(DirectoryEventListener listener){
    eventListeners.removeElement(listener);
  }
  
  /**
   * Dispatch a DirectoryEvent to event listeners.
   * <p>
   * TODO: Implement an event queue, so we can have asynchronous 
   * event logging.
   * @param event The event to dispatch.
   */
  public void dispatchEvent(DirectoryEvent event){
    for(int i = 0; i < eventListeners.size(); i++){
      DirectoryEventListener listener = (DirectoryEventListener) 
	eventListeners.elementAt(i);
      listener.directoryEvent(event);
    }
  }

  /**
   * Convenience method for creating and dispatching an event event.
   * @param type   The type of this event.
   * @param source The object that generated the event.
   * @param message Descriptive message of the event.
   * @param beanConfig The BeanConfiguration processed when the event occured, 
   *		       or null.
   * @param beanProperty The BeanProperty processed when the event occured, 
   *		         or null.
   * @param throwable The Throwable that triggered the event, or null.
   */
  public void dispatchEvent(int type, Object source, String message, 
			    BeanConfiguration beanConfig, 
			    BeanProperty beanProperty, Throwable throwable){
    dispatchEvent(new DirectoryEvent(type, source, message, beanConfig, 
					  beanProperty, throwable));
  }

  /**
   * Convenience method for creating and dispatching an event event.
   * @param source The object that generated the event.
   * @param message Descriptive message of the event.
   * @param beanConfig The BeanConfiguration processed when the event occured, 
   *		       or null.
   * @param beanProperty The BeanProperty processed when the event occured, 
   *		         or null.
   * @param throwable The Throwable that triggered the event, or null.
   */
  public void dispatchErrorEvent(Object source, String message, 
			    BeanConfiguration beanConfig, 
			    BeanProperty beanProperty, Throwable throwable){
    dispatchEvent(new DirectoryEvent(ERROR, source, message, beanConfig, 
					  beanProperty, throwable));
  }

  /**
   * Control logging of debug messages for the bean directory.
   * @param logDebug True if debug messages should be logged, 
   *                 false otherwise. 
   */
  public void setLogDebug(boolean logDebug){
    this.logDebug = logDebug;
  }
  
  /**
   * Control logging of debug messages for the bean directory.
   * @return Returns True if debug messages should be logged, 
   *                 false otherwise. 
   */
  public boolean getLogDebug(){
    return logDebug;
  }

  /**
   * Log a debug message for the bean directory.
   */
  public void logDebug(BeanConfiguration beanConfig, String message){
    if(logDebug){
      dispatchEvent(DirectoryEventTypes.DEBUG, this, message, 
		    beanConfig, null, null);    
    }
  }

  
}
