/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Candleweb AS 
 * Portions created by Candleweb AS are Copyright (C) 1999 Candleweb AS. 
 * All Rights reserved.
 */
package org.locomotive.directory;

import org.openxml.parser.*;
import org.w3c.dom.*;
import java.io.*;
import java.util.*;

/**
 * This class is responsible for parsing a single bean configuration file and 
 * building the appropriate BeanConfiguration structure.
 *
 * @author Gunnar Rnning (gunnar@candleweb.no).
 * @version $Id: XmlBeanParser.java,v 1.4 1999/05/05 10:19:32 gunnar Exp $
 */
public class XmlBeanParser implements BeanParser {
  NameService nameService;

  public XmlBeanParser(NameService nameService){
    this.nameService = nameService;
  }
  
  /**
   * Parse a bean property configuration file and build
   * the approriate BeanConfiguration structure. The method can be called
   * several times with the same bean configuration structure, and will
   * the update the structure to reflect the changes introduced by the
   * new file.
   * @param fileName The fileName of the configuration file.
   * @param beanConfig The beanConfig structure to build/modify.
   */
  public void parseBeanConfiguration(String fileName, 
				     BeanConfiguration beanConfig)
  {
    if (fileName != null) {
      Reader reader;
      File file = new File(fileName);
      if(!file.exists() || !file.canRead()){
	  return;
      }

      try {
	  reader = new FileReader(file);
      } catch (FileNotFoundException notFound) {
	  return;
      }

      XMLParser parser = new XMLParser(reader, fileName);
      Document doc = null;
      try {
	  doc = parser.parseDocument();
      } catch(ParseException pe){
	nameService.dispatchErrorEvent(this, "Error parsing file = " + 
				       fileName + " at line " + 
				       pe.getLineNumber() + ".", 
				       beanConfig, null, pe);
	return;
      }

      Element root = (Element)doc.getDocumentElement();
      
      if(!root.getTagName().equals("bean")){
	nameService.dispatchErrorEvent(this, "Root element in file was not " +
				       "<bean>, but "+ root.getTagName() + ".",
				       beanConfig, null, null);
	return;
      }
      beanConfig.setClassName(root.getAttribute("name"));
      String scope = root.getAttribute("scope");
      if(scope != null && !scope.equals("")){
	beanConfig.setScopeType(scope);
      }

      NodeList nl = root.getElementsByTagName("property");
      int size = nl.getLength();

      for (int i = 0; i < size; i++) {
	Element propElem = (Element) nl.item(i);
	String name = propElem.getAttribute("name");
	if(name == null || name.equals("")){
	  nameService.dispatchEvent(DirectoryEventTypes.WARNING, this,
				    "Property declaration without " + 
				    "name tag is ignored.",
				    beanConfig, null, null);
	  continue;
	}
	BeanProperty prop = beanConfig.getBeanProperty(name);
	if(prop == null){
	  prop = new BeanProperty();
	  prop.setName(name);
	  beanConfig.addBeanProperty(prop);
	}

	String constvalue = propElem.getAttribute("value");
	String component = propElem.getAttribute("component");
	
	BeanPropertyValue propvalue = new BeanPropertyValue();
	if(constvalue != null &&
	   component == null){
	  propvalue.setValue(constvalue);
	} else if(component != null && constvalue == null){
	  propvalue.setValue(component);
	  propvalue.setType(BeanPropertyValue.COMPONENT_TYPE);
	} else {
	  if(constvalue == null)
	    nameService.dispatchErrorEvent(this,
					   "No value or component attribute" +
					   " specified for property = " + name,
					   beanConfig, null, null);
	  else
	    nameService.dispatchErrorEvent(this,
					   "Both value and component " + 
					   "attribute specified for" + 
					   " property = " + name, 
					   beanConfig, null, null);
	  return;
	}
	String append = propElem.getAttribute("append");
	
	
	if(append != null && append.equals("true"))
	  prop.addValue(propvalue);
	else 
	  prop.setValue(propvalue);
      }
    }
  }
}
