/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import java.sql.*;


/** This class provides integration of the Locomotive with Cloudscape.  Each
 * of the methods here are to be called via Cloudscape's database-side method
 * invocation.  For example, to call insertNewSession(), use the following:
 * <PRE>
 *      "select (CLASS org.locomotive.loco.Cloudscape).createSession(...)"
 * <PRE>
 * <P>Note that the methods in this class have been abstracted from their more
 * logical location in the User, Session, and RequestManager classes because
 * there is some hope that someday cloudscape will be able to run in a seperate
 * server JVM.  If this is the case, then this will be the only loco class which
 * have to be in the cloudscape server's classpath.
 */
public class Cloudscape 
{

    /** gets a unique bstamp_id.  All this does is get the last number
     * out of the loco_bstamp_sq table, increment by 1, update the table.
     * and return the value. We shouldn't need to synchronize this method
     * becauase everything is handled within a transaction, which will lock
     * the table we're worried about.  To invoke with SQL, use the following:
     * <PRE>
     *   SELECT (CLASS org.locomotive.loco.Cloudscape).getNextBStampID(
     *          GETCURRENTCONNECTION()) from LOCO_BSTAMP_SQ;
     * </PRE>
     * @Throws java.sql.SQLException if there problems accessing or updating 
     * the data
     * @return a unique bstamp id
     */
    public static long getNextBStampID(Connection conn) throws SQLException
    {
	PreparedStatement pstmt = null;
	ResultSet rs = null;
	long bstamp_id = 0;
	String statement1 = "select max(id) from LOCO_BSTAMP_SQ";
	String statement2 = "update LOCO_BSTAMP_SQ set id = ?";

	try
	{
	    pstmt = conn.prepareStatement(statement1);
	    rs = pstmt.executeQuery();
	    rs.next();
	    bstamp_id = rs.getLong(1) + 1;
	    rs.close();
	    pstmt.close();

	    pstmt = conn.prepareStatement(statement2);
	    pstmt.setLong(1, bstamp_id);
	    pstmt.executeUpdate();
	    pstmt.close();
	    
	    rs = null;  // that way we don't close them again in the finally
	    pstmt = null;
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (pstmt != null) pstmt.close();
	}
	
	return bstamp_id;
    }
		


    /** use this to create a new session.  This shouldn't need to be 
     * synchronized because it does everything within a transaction, 
     * which will lock the LOCO_SESSIONS table anyway
     * handles three JDBC calls: <PRE>
     * 1. get the last session_id from session_id_sq  
     * 2. increment the session_id_sq
     * 3. insert the new session
     * </PRE> 
     * Here's a sample invocation:
     * <PRE>
     *   SELECT (CLASS org.locomotive.loco.Cloudscape).createSession(
     *          GETCURRENTCONNECTION(), 12121, 1010, CURRENT_TIMESTAMP) 
     *          from LOCO_SESSION_SQ
     * </PRE>
     * @Return the session_id of the created session
     * @throws java.sql.SQLException if there problems accessing or updating 
     * the data
     */
    public static long createSession(Connection conn, int srid, int userid, 
				    Timestamp expire)
	throws SQLException
    {
	boolean auto_commit = false;
	PreparedStatement pstmt = null;
	ResultSet rs = null;
	long sid = 0;
//	String statement1 = "select max(sid) from LOCO_SESSIONS";
        String statement1 = "select max(id) from LOCO_SESSION_SQ";
	String statement2 = "update LOCO_SESSION_SQ set id = ?";
	String statement3 = "insert into LOCO_SESSIONS  " +
	    " (sid, srid, userid, added, active, expire) " +
	    " VALUES (?, ?, ?, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP, ?)";

	try
	{
	    // get the last sid
	    pstmt = conn.prepareStatement(statement1);
	    rs = pstmt.executeQuery();
	    rs.next();
	    sid = rs.getLong(1) + 1;
	    rs.close();
	    pstmt.close();
	    
	    // replace it with an incremented one
	    pstmt = conn.prepareStatement(statement2);
	    pstmt.setLong(1, sid);
	    pstmt.executeUpdate();
	    pstmt.close();
	    
	    // insert the new session
	    pstmt = conn.prepareStatement(statement3);
	    pstmt.setLong(1, sid);
	    pstmt.setInt(2, srid);
	    pstmt.setInt(3, userid);
	    pstmt.setTimestamp(4, expire);
	    pstmt.executeUpdate();
	    pstmt.close();

	    rs = null;
	    pstmt = null;
	}
	finally
	{
	    if (rs != null) { rs.close(); }
	    if (pstmt != null) { pstmt.close(); }
	}

	return sid;
    }
	    
		
    /** use this to create a new user.  This shouldn't need to be 
     * synchronized because it does everything within a transaction, 
     * which will lock the LOCO_USERS table anyway
     * handles three JDBC calls: <PRE>
     * 1. get the last user_id from user_id_sq  
     * 2. increment the user_id_sq
     * 3. insert the new user
     * </PRE>
     * Here's an example SQL invocation:
     * <PRE>
     *   SELECT (CLASS org.locomotive.loco.Cloudscape).createUser(
     *           GETCURRENTCONNECTION(), 'username', 'password', 0)
     *           from LOCO_USER_SQ
     * </PRE>
     * @return the user_id of the created user
     * @throws java.sql.SQLException if there problems accessing or updating 
     * the data
     */
    public static int createUser(Connection conn, String username, 
				  String password, int loco_flags) 
	throws SQLException
    {
	PreparedStatement pstmt = null;
	ResultSet rs = null;
	int userid = 0;
//	String statement1 = "select max(userid) from LOCO_USERS";
	String statement1 = "select max(id) from LOCO_USER_SQ";
	String statement2 = "update LOCO_USER_SQ set id = ?";
	String statement3 = "insert into LOCO_USERS " +
	    " (userid, username, password, loco_flags, added, modified, " +
	    " last_login)  VALUES (?, ?, ?, ?, CURRENT_TIMESTAMP, " +
	    " CURRENT_TIMESTAMP, CURRENT_TIMESTAMP)";

	try
	{
	    // get the last userid
	    pstmt = conn.prepareStatement(statement1);
	    rs = pstmt.executeQuery();
	    rs.next();
	    userid = rs.getInt(1) + 1;
	    rs.close();
	    pstmt.close();
	    
	    // replace it with an incremented one
	    pstmt = conn.prepareStatement(statement2);
 	    pstmt.setLong(1, userid);
 	    pstmt.executeUpdate();
 	    pstmt.close();
	    
	    // insert the new user
	    pstmt = conn.prepareStatement(statement3);
	    pstmt.setInt(1, userid);
	    pstmt.setString(2, username);
	    pstmt.setString(3, password);
	    pstmt.setInt(4, loco_flags);
	    pstmt.executeUpdate();
	    pstmt.close();
	    
	    rs = null;
	    pstmt = null;
	}
	finally
	{
	    if (rs != null) { rs.close(); }
	    if (pstmt != null) { pstmt.close(); }
	}

	return userid;

    }


    /** gets a unique graffiti tag id.  All this does is get the last number
     * out of the loco_graffiti_tag_sq table, increment by 1, update the table.
     * and return the value. We shouldn't need to synchronize this method
     * becauase everything is handled within a transaction, which will lock
     * the table we're worried about.  To invoke with SQL, use the following:
     * <PRE>
     *   SELECT (CLASS org.locomotive.loco.Cloudscape).getNextGraffitiTagID(
     *          GETCURRENTCONNECTION()) from LOCO_GRAFFITI_TAG_SQ;
     * </PRE>
     * @Throws java.sql.SQLException if there problems accessing or updating 
     * the data
     * @return a unique bstamp id
     */
    public static long getNextGraffitiTagID(Connection conn) throws SQLException
    {
	PreparedStatement pstmt = null;
	ResultSet rs = null;
	long tag_id = 0;
	String statement1 = "select max(id) from LOCO_GRAFFITI_TAG_SQ";
	String statement2 = "update LOCO_GRAFFITI_TAG_SQ set id = ?";

	try
	{
	    pstmt = conn.prepareStatement(statement1);
	    rs = pstmt.executeQuery();
	    rs.next();
	    tag_id = rs.getLong(1) + 1;
	    rs.close();
	    pstmt.close();

	    pstmt = conn.prepareStatement(statement2);
	    pstmt.setLong(1, tag_id);
	    pstmt.executeUpdate();
	    pstmt.close();
	    
	    rs = null;  // that way we don't close them again in the finally
	    pstmt = null;
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (pstmt != null) pstmt.close();
	}
	
	return tag_id;
    }
		



}  // end of Cloudscape.java
