/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.server.*;
import org.locomotive.util.*;
import java.io.*;
import java.util.*;

/**
 * FormData combines the features of a Hashtable with a
 * string-checker and a StringBuffer into one powerful class.
 * FormData allows you to get the values of the fields in a form,
 * check whether those values fit a specific format, and generate
 * automatic error messages when those values do not fit the format.
 *
 * You do not need to create an instance of FormData, if you use
 * the GenericRequest class.  The constructor for GenericRequest
 * automatically instantiates a local variable called form_data, which
 * is an instance of FormData that has been filled out with the form
 * values from the Web browser.
 *
 * There is extensive documentation about how to use the FormData
 * class, in Chapter III of the Locomotive Developer's Guide.
 * However, we will present a few examples here.  In general, you can
 * use the methods whose names start with get...() to retrieve the
 * values of fields.  For example:
 *
 * <pre>
 *     String username = form_data.getAlphanum ("LOGIN_USERNAME", "username", 2, 12, true);
 * </pre>
 *
 * In general, all the get...() methods have the same five arguments:
 * <ol>
 * <li> The first argument is the actual name of the field in the form.
 * <li> the second argument is a "friendly" name, to be displayed to the user in an error message.
 * <li> the third argument is the minimum length that the field must be.
 * <li> the fourth argument is the maximum length that the field must be.
 * <li> you can use the fifth argument to tell this FormData object whether to generate an error code and an error message when there is an error.
 * </ol>
 *
 * If the field does not fit your formatting criteria, or it could not
 * be found in the form, or it was empty, the FormData object will
 * generate an error code and an error message.  Then, unless you ignore or
 * override this error, the error message will be added to an accumulation
 * of all error messages.  So, this object "keeps track" of the errors for you.
 * 
 * These are the different error codes that are thrown (which are very
 * similar to the error codes of org.locomotive.util.ParseStringException):
 *
 * <ol>
 * <li> If the field is not found in the form, the FormData object generates an error code of FormData.IS_NULL.
 * <li> if the field exists in the form, but it's just an empty string "", and the minimum length is not 0, then the FormData generates FormData.IS_EMPTY.
 * <li> if the field is shorter than the minimum length, the FormData generates FormData.TOO_SHORT.
 * <li> if it is longer than the maximum length, the FormData generates FormData.TOO_LONG.
 * <li> if the field does not fit the specific type you specified, for example, if you expected digits, and there were alphabetical letters in the field, then the FormData will generate FormData.DOESNT_FIT_TYPE.
 * <li> if the field is numeric, and the field is smaller than the lower bound which you specified, then the FormData will generate FormData.TOO_SMALL.
 * <li> however, if a numeric field is larger than the upper bound, then the FormData will generate FormData.TOO_LARGE.
 * </ol>
 *
 * In order to find out whether the last get...() generated one or
 * more error codes, you call the method hasError(int your_error_code).
 * In order to see the last error message that was generated for a particular
 * error code, call the method lastErrorMsg(int your_error_code).
 * Here's an example:
 *
 * <pre>
 *     boolean too_short = form_data.hasError(FormData.TOO_SHORT);
 *     String too_short_message = form_data.lastErrorMsg(FormData.TOO_SHORT);
 * </pre>
 * You can override the error message that was generated for a
 * particular error code during the last get...() method call, using
 * setErrorMsg(...).  Or, if you want to ignore an error code, use
 * ignoreErrorCode(...).
 * If you want to see the sum total of all error messages generated for all 
 * fields, use allErrorMsgs().
 */
public class FormData
{
    // For an explanation of these error codes, see the comments at
    // the top of the FormData class.
    public static final int NO_ERROR = 0;
    public static final int IS_NULL = 1;
    public static final int IS_EMPTY = 2;
    public static final int TOO_SHORT = 3;
    public static final int TOO_LONG = 4;
    public static final int TOO_SMALL = 5;
    public static final int TOO_LARGE = 6;
    public static final int DOESNT_FIT_TYPE = 7;

    String current_error_msgs[] = new String [8];
    StringBuffer all_error_messages = null;

    Hashtable form_fields;

    /**
     * create a new FormData object, and initialize it with the values
     * of the fields from the Hashtable, form_hash.
     */
    public FormData (Hashtable form_hash)
    {
	form_fields = form_hash;
	    
	if (form_fields == null)
	    form_fields = new Hashtable();
	    
	for (int i=0; i < current_error_msgs.length; i++)
	{
	    current_error_msgs[i] = null;
	}
	
	all_error_messages = new StringBuffer();
    }
    
    
    /**
     * returns the value of the field named field_name.  If the form does not
     * contain this field, will return null.
     * NOTE: this is the only method which returns the value of a field
     * without generating any error codes or error messages.
     */
    public String getField (String field_name)
    {
	Object o = form_fields.get(field_name);
	if (o == null || o instanceof String)
	{
	    return (  (String) o);
	}
	else
	{
	    return (  ((String []) o) [0] );
	}
    }
    
    
    
    /**
     * returns the value of a Hashtable containing all fields in this form.
     * if the form is empty, will return a new Hashtable().
     */
    public Hashtable getHashtable()
    {
	return form_fields;
    }
    
    
    /**
     *  tells whether the form is empty (contains no fields).
     */
    public boolean isEmpty()
    {
	return form_fields.isEmpty();
    }
    
    
    
    /**
     * copies the values of all fields in this form to the Hashtable
     * containing all the Steam variables, so that you can use
     * the Steam variables to put in default values for the form fields 
     * in the HTML template.
     */
    public void copyAllIntoSteam (Hashtable steam_variables)
    {
	if (steam_variables == null)
	    return;
	    
	Enumeration values = form_fields.keys();
	while (values.hasMoreElements())
	{
	    String key = (String) values.nextElement();
	    steam_variables.put (key, getField (key));
	}
    }
    
    /**
     * tells whether the form contains values for all of the fields mentioned.
     * Example:
     * <pre>
     *     String[] required_fields = { "USERNAME", "USERID", ... };
     *     boolean is_good = my_form.containsAll(required_fields);
     * </pre>
     */
    public boolean containsAll (String [] names_of_all_fields)
    {
	for (int i=0; i < names_of_all_fields.length; i++)
	    if (!form_fields.containsKey (names_of_all_fields [i]))
		return false;
	    
	return true;
    }
    
    
    /**
     *  tells whether the form contains non-empty values for all of the
     *  fields mentioned.
     */
    public boolean containsAllNonEmpty (String [] names_of_all_fields)
    {
	for (int i=0; i < names_of_all_fields.length; i++)
	    if (!form_fields.containsKey (names_of_all_fields [i]) ||
		getField(names_of_all_fields[i]).toString().equals(""))
		return false;
	    
	return true;
    }
    
    
    
    /**
     * returns the value of the field named field_name.  if the form does not
     * contain this field, will return an empty String, "", rather than null,
     * and generate the error code IS_NULL.
     */
    public String getNonNullField (String field_name, String name_to_display)
    {
	saveLastError();
	
	String value = (String) getField (field_name);
	if (value == null)
	{
	    current_error_msgs[IS_NULL] = "Error: your form did not contain the " +
		name_to_display + " field.  Please try submitting it again.<BR>";
	    return "";
	}
	else
	{
	    return value.trim();
	}
    }
    
    
    /**
     * returns the value of the field named field_name.  if the form does not
     * contain this field, or if the field is an empty string (""), will set
     * the error code, IS_EMPTY, and store an error message, 
     * then return an empty string,
     * "".
     */
    public String getNonEmptyField (String field_name, String name_to_display)
    {
	saveLastError();
	
	String value = (String) getField (field_name);
	if (value == null)
	{
	    current_error_msgs[IS_NULL] = "Error: your form did not contain the " +
		name_to_display + " field.  Please try submitting it again.<BR>";
	    return "";
	}
	else if (value.equals(""))
	{
	    current_error_msgs[IS_EMPTY] = "You did not enter a value for the " +
		name_to_display + " field.<BR>";
	    return value;
	}
	else
	{
	    return value.trim();
	}
    }
    
    
    void saveLastError()
    {
	for (int i=0; i < current_error_msgs.length; i++)
	{
	    if (current_error_msgs[i] != null)
	    {
		all_error_messages.append (current_error_msgs[i]);
	    }
	    current_error_msgs[i] = null;
	}
    }
    

    /**
     * tells whether the last get...() method generated an error whose code 
     * matches error_code.  Remember, it may have generated multiple error
     * codes, so you can check each individual error code.
     */
    public boolean hasError (int error_code)
    {
	return (current_error_msgs[error_code] != null);
    }
    
    
    /**
     * tells whether the last get...() method generated any errors of any kind.
     */
    public boolean hasAnyErrors ()
    {
	for (int i=0; i < current_error_msgs.length; i++)
	    if (current_error_msgs[i] != null)
		return true;

	return false;
    }
    
    
    /**
     * if the last get...() method generated an error matching error_code, 
     * this method will return the error message.  If there was no error,
     * it will return "".
     */
    public String lastErrorMsg (int error_code)
    {
	if (error_code < 0 || error_code > DOESNT_FIT_TYPE)
	    return "BAD ERROR CODE!!!";
	
	if (current_error_msgs[error_code] != null)
	    return current_error_msgs[error_code];
	else
	    return "";
    }
    
    
    /**
     * allows you to set the error message for a particular error code.
     * However, if the last get...() method did not generate that error code,
     * your error message will not be added to the accumulation of all error
     * messages.
     */
    public void setMsgForError (int error_code, String error_msg)
    {
	if (error_code < 0 || error_code > DOESNT_FIT_TYPE ||
	    error_msg == null)
	    return;

	current_error_msgs[error_code] = error_msg;
    }
    
    
    /**
     * returns a String containing all the error messages generated while
     * accessing all the fields in this form.
     */
    public String allErrorMsgs ()
    {
	saveLastError();
	
	return all_error_messages.toString();
    }
    
    
    /**
     * allows you to ignore the error message associated with error_code
     * in the last get...() method call.
     */
    public void ignoreErrorCode (int error_code)
    {
	if (error_code < 0 || error_code > DOESNT_FIT_TYPE)
	    return;

	current_error_msgs[error_code] = null;
    }

    
    /**
     * allows you to ignore all the error messages in the last get...()
     * method call.
     */
    public void ignoreErrors ()
    {
	for (int i=0; i < current_error_msgs.length; i++)
	{
	    current_error_msgs[i] = null;
	}
    }

    
    /**
     * tries to read a field as an int.
     * @param field_name the name of the int field in the form.
     * @param name_to_display the name to display in an error message to the user.
     * May generate the errors IS_NULL, IS_EMPTY, and DOESNT_FIT_TYPE.
     */
    public int getInt (String field_name, String name_to_display)
    {
	String field_value = getNonEmptyField (field_name, name_to_display);
	    
	if (hasAnyErrors())
	    return 0;
	    
	try {
	    return Integer.parseInt (field_value);
	}
	catch (NumberFormatException exception) {
	    current_error_msgs[DOESNT_FIT_TYPE] =
		"Please type in a number for the " + field_name + " field.<BR>";
	    return 0;
	}
    }
    
    /**
     * tries to read a field as an int, bounded by a minimum and a maximum.
     * @param field_name the name of the int field in the form.
     * @param name_to_display the name to display in an error message to the user.
     * @param minimum the minimum this number can be.
     * @param maximum the maximum this number can be.
     * May generate the errors IS_NULL, IS_EMPTY, DOESNT_FIT_TYPE, 
     * TOO_SMALL, and TOO_LARGE.
     */
    public int getIntWithinBounds (String field_name, String name_to_display,
				   int minimum, int maximum) 
    {
	int result = getInt (field_name, name_to_display);
	if (hasAnyErrors())
	    return 0;
	    
	if (result < minimum)
	{
	    current_error_msgs[TOO_SMALL] =
		"You must enter a number greater than or equal to " + 
		minimum + " for the " + name_to_display + " field.<BR>";
	    return 0;
	}
	else if (result > maximum)
	{
	    current_error_msgs[TOO_LARGE] = 
		"You must enter a number less than or equal to " +
		maximum + " for the " + name_to_display + " field.<BR>";
	    return 0;
	}
	else
	{
	    return result;
	}
    }
    
    
    /**
     *  if this field contains non-digits, then the non-digits will be removed.
     *  Also, if it does contain non-digits, and warn_non_digits is true,
     *  then will generate an error code of DOESNT_FIT_TYPE.
     */
    public String getDigits (String field_name, String name_to_display, 
			     int minimum_length, int maximum_length, 
			     boolean warn_non_digits)
    {
	return getSpecialChars ("digits", field_name, name_to_display,
				minimum_length, maximum_length,
				warn_non_digits); 
    }
    
    
    /**
     *  if this field contains non_alphanumerics, then they will be removed.
     *  also, if it does contain non_alphanumerics, and warn_non_alphanum is 
     *  true, then will generate an error code of DOESNT_FIT_TYPE.
     */
    public String getAlphanum (String field_name, String name_to_display,
			       int minimum_length, int maximum_length, 
			       boolean warn_non_alphanum)
    {
	return getSpecialChars ("letters or digits", field_name, name_to_display,
				minimum_length, maximum_length,
				warn_non_alphanum); 
    }
    
    
    
    /**
     * Returns the string field named field_name.  Does NOT trim any leading
     * or trailing whitespace in the field, unlike all the other methods.
     * If it's null, too short, or too long, then this method will set
     * the error code.
     */
    public String getString (String field_name, String name_to_display,
			     int minimum_length, int maximum_length) 
    {
	saveLastError();

	String value = (String) getField (field_name);
	if (value == null)
	{
	    current_error_msgs[IS_NULL] = 
		"Error: your form did not contain the " + name_to_display +
		" field.  Please try submitting it again.<BR>";
	    return "";
	}
	else if (value.length() == 0 && minimum_length > 0)
	{
	    current_error_msgs[IS_EMPTY] = 
		"Please enter a value for the " + name_to_display + " field.<BR>";
	}
	else if (value.length() < minimum_length)
	{
	    current_error_msgs[TOO_SHORT] = 
		"Please type at least " + minimum_length +
		" letters in the " + name_to_display + " field.<BR>";
	}
	else if (value.length() > maximum_length)
	{
	    current_error_msgs[TOO_LONG] =
		"Please type no more than " + maximum_length +
		" letters in the " + name_to_display + " field.<BR>";
	}

	return value;
    }
    
    
    
    /**
     * This method trims the leading and trailing whitespace in
     * a field, and returns its value. 
     * It will NOT give an error if there is whitespace in the field.
     */
    public String getTrimmedString (String field_name, String name_to_display,
				    int minimum_length, int maximum_length) 
    {
	return getSpecialChars ("trimmed_string", field_name, name_to_display,
				minimum_length, maximum_length, false); 
    }
    
    
    
    String getSpecialChars (String char_type, String field_name, 
			    String name_to_display, int minimum_length,
			    int maximum_length, boolean warn_bad_chars) 
    {
	String field_value = getNonNullField (field_name, name_to_display);
	    
	if (current_error_msgs[IS_NULL] != null)
	{
	    return "";
	}
	    
	char[] chars = field_value.toCharArray();
	char[] new_value = new char[chars.length];
	int counter = 0;
	    
	boolean found_bad_char = false;
	    
	if (char_type.equals("digits"))
	{
	    for (int i=0; i < chars.length; i++)
		if (Character.isDigit (chars[i]))
		    new_value[counter++] = chars[i];
		else
		    found_bad_char = true;
	}
	else if (char_type.equals("letters or digits"))
	{
	    for (int i=0; i < chars.length; i++)
		if (Character.isLetterOrDigit (chars[i]))
		    new_value[counter++] = chars[i];
		else
		    found_bad_char = true;
	}
	else if (char_type.equals("trimmed_string"))
	{
	    new_value = chars;
	    counter = chars.length;
	}
	    
	    
        // if the user should not enter any bad characters in this field
        // but then we encounter one, then we should give an error message:
	if (warn_bad_chars && found_bad_char)
	{
	    current_error_msgs[DOESNT_FIT_TYPE] = 
		"Please type only " + 
		char_type + " in the " + name_to_display + " field.<BR>";
	}
	
	if (counter == 0 && minimum_length > 0)
	{
	    current_error_msgs[IS_EMPTY] = 
		"Please enter a value for the " + name_to_display + " field.<BR>";
	}
	else if (counter < minimum_length)
	{
	    current_error_msgs[TOO_SHORT] =
		"Please type at least " + minimum_length + " " + char_type +
		" in the " + name_to_display + " field.<BR>";
	}
	else if (counter > maximum_length)
	{
	    current_error_msgs[TOO_LONG] = 
		"Please type no more than " + maximum_length + " " + 
		char_type + " in the " + name_to_display + " field.<BR>";
	}
	    
	return new String(new_value, 0, counter);
    }
    

    /**
     * tries to parse a field as a Date in one of the following formats:
     * "MMM d, yyyy", "MM-d-yy", "MM-d-yyyy", "MM/d/yy",
     * "MM/d/yyyy", "d.MM.yy", "d.MM.yyyy".
     * Will generate DOESNT_FIT_TYPE if it could not parse the field as a Date,
     * and then return null.
     * This method uses the StringUtil.getDate() method.
     */  
    public java.util.Date getDate (String field_name, String name_to_display)
    {
	String field_value = getNonEmptyField (field_name, name_to_display);
	    
	if (hasAnyErrors())
	    return null;

	java.util.Date result = StringUtil.getDate (field_value);

	if (result == null)
	{
	    current_error_msgs[DOESNT_FIT_TYPE] = 
		"Please enter a date in the format: 'Jan 1, 1998' in the " + 
		name_to_display + " field.<BR>";
	}
	    
	return result;
    }
    
    
    /**
     * tries to parse the three fields, month_field, day_field, and
     * year_field, and create a Date out of them.  The month_field
     * argument can be either a number from 1-12 or at least
     * the first three letters of the name of a month, upper or lower case.
     * This method uses the StringUtil.getDate() method.
     * It will generate an error code of DOESNT_FIT_TYPE if
     * it could not parse the field as a Date, and then, it will return null.
     * This method uses the StringUtil.getDate() method.
     */
    public java.util.Date getDate (String month_field, String day_field, 
				   String year_field, String name_to_display)
    {
	saveLastError();
	    
	String month_value = getField (month_field);
	String day_value = getField (day_field);
	String year_value = getField (year_field);
	    
	try {
	    return StringUtil.getDate (month_value, day_value, year_value);
	}
	catch (ParseStringException exception) {
	    int last_error_code = exception.errorCode();
	    switch (last_error_code)
	    {
	    case ParseStringException.IS_NULL:
		current_error_msgs[IS_NULL] =
		    "Error: your form did not contain the " + name_to_display +
		    " field.  Please try submitting it again.<BR>";
		return null;
		    
	    case ParseStringException.IS_EMPTY:
		current_error_msgs[IS_EMPTY] = 
		    "Please enter the month, day, and year" +
		    " values for the " + name_to_display + " field.<BR>";
		return null;
		    
	    case ParseStringException.DOESNT_FIT_TYPE:
		current_error_msgs[DOESNT_FIT_TYPE] =
		    "Please enter a date in the format:" +
		    " 'Jan 1, 1998' in the " + name_to_display + " field.<BR>";
		return null;
	    }
	}
	
	return null;
    }
    

    /**
     * Will return the email address no matter what, even if it doesn't fit,
     * but if there are errors, will give IS_NULL, IS_EMPTY, TOO_SHORT,
     * TOO_LONG, and DOESNT_FIT_TYPE.  
     * The email address should have at least one '@', at least one '.' after
     * the '@', and contain only alphanumeric characters or characters
     * in: [+@.'%-_!] (not including the brackets [ ]).
     * (This method uses StringUtil.isEmailAddress().)
     */
    public String getEmailAddress (String field_name, String name_to_display,
				   int minimum_length, int maximum_length)
    {
	String field_value = getNonEmptyField (field_name, name_to_display);
	    
	if (hasAnyErrors())
	{
	    return "";
	}
	
	if (!StringUtil.isEmailAddress (field_value))
	{
	    current_error_msgs[DOESNT_FIT_TYPE] =
		"I'm sorry, I could not recognize the" +
		" format of the email address you entered for the " +
		name_to_display + " field.<BR>";
	}
	
	if (field_value.length() < minimum_length)
	{
	    current_error_msgs[TOO_SHORT] =
		"Your email address in the " + name_to_display +
		" field must be at least " +
		minimum_length + " characters.<BR>";
	}
	else if (field_value.length() > maximum_length)
	{
	    current_error_msgs[TOO_LONG] =
		"Please limit your email address in the " + name_to_display +
		" field to " + maximum_length + " characters.<BR>";
	}		    
	
	return field_value;
    }	    

}
