/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.server.*;
import org.locomotive.util.*;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.SimpleDateFormat;

/**
 * HTTPRequestManagers do several things:<BR>
 * 1) wait for a GET or POST request from the client socket<BR>
 * 2) send the request on to the Locomotive if necessary (not yet
 *     implemented), or just serve a file, depending on the request<BR>
 * 3) post serving cleanup <BR>
 * 4) management of server data <BR>
 * <P>
 * HTTPRequestManagers are invoked as separate threads (one thread to a 
 * request) after being initialized by the HTTPServer at run-time
 */
class HTTPRequestManager implements Runnable
{
    private HTTPServer http_server;
    private HTTPServerConfig config;
    private static LocoConfig loco_config;
    private Log server_log;
    private Log event_log;
    
    // these are arrays of ints holding the ips  and subnets which are
    // allowed to access the HTTPServer
    private int ip_accept[] = null;
    private int ip_subnet[] = null;
    private int ip_calc[]   = null;  // ip_accept & ip_subnet
    
    private static final String DOTDOT = "..";
    private static final String SLASHSLASH = "//";
    private static final String SLASHDOTDOT = "/..";
    private static final String HTTP_NEWLINE = "\r\n";
    private static final String LOCALHOST = "localhost";
    private static final String LOCALHOST_IP = "127.0.0.1";
    private static final String LOCALHOSTLONG = "127.0.0.1/localhost";

    // This one should be made into a configurable static variable, so
    // the webmaster can decide what String JSP files should end with.
    private static final String JSP_EXTENSION = ".jsp";
    
    // these are all that come in on a new HTTPRequestManager (instantiated by 
    // the HTTPServer at hit receive time)
    private int request_id;
    private Socket sock;
    private int reply_protocol_type; 
    
    private BufferedReader is;
    private BufferedOutputStream os;
    
    private Hashtable cgi_envars_hash;
    private Hashtable form_hash;
    private Hashtable subs;
    private String url_tokens[];
    private String http_request_line;  // Stores the GET/POST request line
    private String loco_request_token;
    private boolean http_response_sent = false;

    private Response resp;

    /** 
     * constructor used by HTTPServer
     */
    HTTPRequestManager(HTTPServer server, Socket s, int rq_id)
    {
	http_server = server;
	sock = s;
	request_id = rq_id;
    }

    /**
     * Invoked by the HTTPServer for each request.  Do not use publicly.
     */
    public void run ()
    {
	// if setup is successful, hand it off
	if (doSetup())
	{
	    // This is where the HTTP request is actually looked at, and
	    // where it decides what to do about it.
	    handoff();
	}

	doCleanup();
	
	// indicates completion of this thread to the Loco for 
	// active thread limit management.  Important!
    }

    /*--------------- SETUP ----------------------*/


    /**
     * do all pre-handoff setup:  
     * cookies, sessions, user objects, get a connection, open streams, 
     * parse encoded form data, whatever.  Return false if we are to skip
     * the handoff.
     */ 
    private boolean doSetup ()
    {
	if (! openStreams())
	{
	    return (false);
	}

	// Check to see if this client is coming from an acceptable IP address
	if (! authorize())
	{
	    // Send the user a "Forbidden" error page, and hang up!
	    String errorString = "<H1>403 FORBIDDEN</H1> - Your client is "
		                 + "disallowed access to the server.";
	    sendErrorMsg(errorString);
	    return(false);
	}

	// Now, let's create the request object we'll use for the setup.
	// we'll create an HTTPReqeust object to handle this as any web
	// server would..
	OutputStream sock_output_stream;
	try
	{
	    sock_output_stream = sock.getOutputStream();
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: HTTPRM(" + request_id + ") doSetup(): " + 
		 StringUtil.exceptionStackTrace(ioe));
	    return (false);
	}

	// We may need to do this later..
  	//resp = new HTTPResponse(sock_output_stream, request_id);

  	return (true);
    }


    // create a BufferedReader and a BufferedWriter from the client Socket
    private boolean openStreams()
    {
	try
	{
            InputStream input = sock.getInputStream();
            is = new BufferedReader(new InputStreamReader(input));

    	    OutputStream output_stream = sock.getOutputStream();
  	    os = new BufferedOutputStream(output_stream);
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: HTTPRM(" + request_id + ") openStreams(): " + 
		  StringUtil.exceptionStackTrace(ioe));
	    return (false);
	}
	
	return (true);
    }


    /** 
     * This will check the IP address of the incoming connection, and
     * if it passes the test (is an authorized client address), this
     * method allows the request.
     */
    private boolean authorize()
    {
	// Bugfixes by Ludovic Lange.

	// make sure it's coming from somplace we like
	byte [] sock_ar = sock.getInetAddress().getAddress();
	int sock_addr = 0;
	int temp2 = 0;
	for (int i = 0; i < sock_ar.length; i++)
	{
	    temp2 = sock_ar[i];
	    if (temp2 < 0)
	    {
		temp2 = 256 + sock_ar[i];
	    }
	    int temp = temp2 << (24 - (i*8));
	    sock_addr += temp;
	}

	boolean ok = false;
	int ip;
	for (ip = 0; ip < ip_accept.length; ip++)
	{
	    int so_su = sock_addr & ip_subnet[ip];
	    if ( so_su == ip_calc[ip])
	    {
		ok = true;
		break;
	    }
	}
	    

	if (ip >= ip_accept.length) // just use the last ip in the config
	    ip -= 1;

	// let's log all this for posterity
	StringBuffer sb_ip = new StringBuffer();
	StringBuffer sb_su = new StringBuffer();
	StringBuffer sb_so = new StringBuffer();
	for (int i = 0; i < 4; i++)
	{
	    int temp = (ip_accept[ip] >> (24 - i*8)) & 0x000000FF;
	    sb_ip.append(temp); sb_ip.append('.');

	    temp = (ip_subnet[ip] >> (24 - i*8)) & 0x000000FF;
	    sb_su.append(temp); sb_su.append('.');

	    temp = (sock_addr >> (24 - i*8)) & 0x000000FF;
	    sb_so.append(temp); sb_so.append('.');
	}

	sb_ip.setLength(sb_ip.length() - 1);
	sb_su.setLength(sb_su.length() - 1);
	sb_so.setLength(sb_so.length() - 1);
	
	log(0, "Client request from: " + sb_so.toString() +
	    "\nIPs accepted from " + sb_ip.toString() +
	    " with subnet " + sb_su.toString() +
	    "\t Accept: " + ok);
	
	return(ok);
    }



/*----------------- HANDOFF  --------------------*/


    /**
     * This method waits for an HTTP request from the client, and decides what
     * to do with it.  Usually, it will just instantiate a Response object
     * to serve a file from the filesystem (somewhere under the
     * HTTPD_DOCUMENT_ROOT_PATH).  It could also forward requests to
     * the Locomotive if it sees that the request URI begins with the
     * LOCO_REQUEST_TOKEN.
     */
    void handoff ()
    {
	Vector environment_lines; // Stores all other data lines sent to us

	try
	{
	    // Wait here for an HTTP GET/POST request line
	    http_request_line = is.readLine();
	    //System.out.println("Request Line: " + http_request_line);

	    // Now parse the request line and decide what to do
	    StringTokenizer st = new StringTokenizer(http_request_line);
	    String method = st.nextToken();
	    String req_path = st.nextToken();
	    String version = "[none]";

	    // Get the HTTP version string that the client sent in, if any
	    if (st.hasMoreTokens())
	    {
		version = st.nextToken();
	    }

	    // Read lines until we get a blank one (a single blank
	    // line marks the end of request lines)
	    environment_lines = new Vector();
	    String environment_line;

	    while (!(environment_line = is.readLine()).equals(""))
	    {
		// Add the new environment data line to our Vector of them
		environment_lines.addElement(environment_line);
		//System.out.println("HTTPRM: Header line: " +
		//                     environment_line);
	    }

	    //System.out.println("HTTPRM: Done reading environment lines.");

	    //
	    // Now determine if we should send this request to the
	    // Locomotive, or if the Embedded Web Server should simply
	    // serve a file..
	    //

	    // If the request path begins with the Locomotive's request token
	    // (like "/locolink"), then it's a Locomotive request.
	    if (req_path.startsWith(loco_request_token))
	    {
		// Request URI starts with the loco_request_token
		// (like "/locolink"), send this request into
		// the Locomotive!  Yeeehaw!
		handleLocomotiveRequest(method, req_path, version,
					environment_lines);

		return;
	    }

	    // If the URI ends with a filename having the extension
	    // ".jsp" (Java Server Page!), then that's also a Locomotive
	    // request.
            //
            // Later, this section of code should be replaced with a method
            // call to translate any/all URI aliases that are in the EWS's
            // config file.  When that happens, move this section of code
            // above the check for Locomotive requests.
	    String uri = req_path;
	    int index = uri.indexOf('?');
	    if (index != -1) uri = uri.substring(0, index);
	    if (uri.endsWith(JSP_EXTENSION))
	    {
		// It's a Locomotive request for a Java Server Page
		handleLocomotiveRequest(method, loco_request_token + req_path,
					version, environment_lines);
		return;
	    }

	    // The Embedded Web Server will just serve a file..
		
	    if (method.equalsIgnoreCase("GET"))
	    {
		// Serve the document that the client requested
		serveDocument(req_path, version, environment_lines);
	    }
	    else
	    {
		// The method wasn't a GET, and right now that's
		// all we handle..

		// Later, we should allow special web server configurations
		// that map other URI path prefixes to Locomotive servlets,
		// so that if you don't want something like "/locolink" on
		// the front of your paths, you don't *have* to have it..
		sendNotImplementedError(version);
	    }
	}
	catch (IOException ioe)
	{
	    log(0, "ERROR: HTTPRM(" + request_id + ") handoff(): " +
	        StringUtil.exceptionStackTrace(ioe));
	}
    }

    /**
     * This method hands the current request to the Locomotive Servlet
     * engine.  It sets all of the appropriate environment variables
     * that the Locomotive needs, and then instantiates a RequestManager
     * that will take over serving the request.
     *                                                             <BR><BR>
     * @param method the type of request, either "GET" or "POST".
     * @param req_path the request URI, looks like "/locolink/disp?blah".
     * @param environment_lines a Vector of HTTP header lines, just as
     *                          they came to us from the web browser.
     */
    protected void handleLocomotiveRequest(String method, String req_path,
					   String version,
					   Vector environment_lines)
    {
	//System.out.println("HTTPRM: It's a Locomotive request!");

	//
	// Set some more header/environment lines
	//
	environment_lines.addElement("SERVER_PROTOCOL: HTTP/1.0");

	// Get the name of the remote client machine
	InetAddress remote_inet_address = sock.getInetAddress();
	String remote_ip_address = remote_inet_address.getHostAddress();
	String remote_addr = LOCALHOSTLONG;
	String remote_host = LOCALHOST;
	// If the request came from localhost, don't do a DNS lookup!
	if (null != remote_ip_address &&
	    !remote_ip_address.equals(LOCALHOST_IP))
	{
	    // It's not a local client, so do a DNS lookup for the hostname.
            remote_addr = sock.getInetAddress().toString();
	    remote_host = sock.getLocalAddress().getHostName();
	}

	// Remove the slash and IP address, if they're there.
        int slash_index = remote_addr.indexOf('/');
        if (slash_index != -1)
            remote_addr = remote_addr.substring(slash_index + 1,
                                       remote_addr.length());
	environment_lines.addElement("REMOTE_ADDR: " + remote_addr);
	environment_lines.addElement("REMOTE_PORT: " + sock.getPort());
	environment_lines.addElement("HTTP_HOST: " + remote_host);
	environment_lines.addElement("REMOTE_HOST: " + remote_host);

	String server_name = null;
	try
	{
	    // Get the name of the local machine via files, not via DNS
	    server_name = InetAddress.getLocalHost().toString();

	    // Remove the slash and IP address, if they're there.
	    slash_index = server_name.indexOf('/');
	    if (slash_index != -1)
		server_name = server_name.substring(0, slash_index);
	}
	catch (Exception e)
	{
	    // If something went wrong, try using DNS instead.
	    server_name = sock.getLocalAddress().getHostName();
	}
	environment_lines.addElement("SERVER_NAME: " + server_name);
	environment_lines.addElement("SERVER_PORT: " + sock.getLocalPort());
	environment_lines.addElement("DOCUMENT_ROOT: " + config.getString(
				     "HTTPD_DOCUMENT_ROOT_PATH"));
	environment_lines.addElement("SCRIPT_NAME: " + loco_request_token);
	environment_lines.addElement("SERVER_SOFTWARE: Locomotive/" + 
				     Loco.getVersionString());
	environment_lines.addElement("GATEWAY_INTERFACE: CGI/1.1");
	// We need to also set SERVER_ADMIN and TZ here!

	String path_info = null;
	String query_string = null;

        // Andrew Mercer took this code out of get so that it also works for
        // post

        // Separate the path from the query, and chop of the
        // LOCO_REQUEST_TOKEN from the path..
        int index = req_path.indexOf("?");

        if (index != -1)
        {
            // We have a path and a query string
            path_info = req_path.substring(loco_request_token.length(), index);
            query_string = req_path.substring(index + 1, req_path.length());
        }
        else
        {
            // We only have a path
            path_info = req_path.substring(loco_request_token.length(),
                                           req_path.length());
        }

        if (method.equalsIgnoreCase("GET"))
        {
            //System.out.println("HTTPRM: It's a GET!");
            environment_lines.addElement("REQUEST_METHOD: GET");
        }
        else if (method.equalsIgnoreCase("POST"))
        {
            //System.out.println("HTTPRM: It's a POST!");
            //System.out.println("req_path: " + req_path);

            environment_lines.addElement("REQUEST_METHOD: POST");
        }

	//
	// In the future, right here is where we'd make the decision
	// whether to hand this request to another Locomotive (via a
	// TCP Socket), or to serve this request in this VM's Locomotive.
	// Eventually, we should offer the same load-balancing options
	// (or even more?) as the Locolink tunnels do.
	//

	//System.out.println("HTTPRM: Starting RM with " + 
	//		   path_info + ", " + query_string);

	// Make a new RequestManager to serve this request
	RequestManager rm = new RequestManager(sock, request_id, 
					       path_info, query_string,
					       environment_lines, is);

	// Run the request manager
	rm.run();

	return;
    }

    /**
     * This method serves a file right out into the socket.  If it cannot
     * open and read the file, it will send a "Not Found" error page to the
     * client.
     *                                                              <BR><BR>
     * @param document_path     the request URI to what the browser wants,
     *                          it probably looks something like
     *                          "/docs/index.html".
     * @param http_version      the version of HTTP that we got from the
     *                          browser, if any.  We need this in case we're
     *                          sending an error page so that we know if we
     *                          need to send HTTP headers along with the
     *                          error or not.
     * @param environment_lines a Vector of Strings, each one containing an
     *                          HTTP request header line received from the
     *                          client.
     */
    protected void serveDocument(String document_path, String http_version,
				 Vector environment_lines)
    {
	String original_requested_path = document_path;

	// SECURITY!!  Make sure that clients can't back out of the
	// document root path using "../" dirnames!
	document_path = enforceDocumentRoot(document_path, http_version);

	// If enforceDocumentRoot() already sent an error page, then
	// we're done with this request.
	if (http_response_sent) return;

	if (document_path.endsWith("/"))
	    document_path += config.getString("HTTPD_DOCUMENT_INDEX_FILE");

	//System.out.println("PATH: " + document_path);

	// Guess at the MIME Content-type
	String content_type = guessContentTypeFromName(document_path);
	
	// Try to open the file and serve it
	try
	{
	    File the_file = new File(
			     config.getString("HTTPD_DOCUMENT_ROOT_PATH") +
			     document_path);

	    // If it's a directory, and the URI doesn't have a slash on the
	    // end, then send a redirect to the same URI ending with a slash.
	    if (the_file.isDirectory() && (document_path.length() == 0 ||
		document_path.charAt(document_path.length() - 1) != '/'))
	    {
		//
		// Redirect the client to the same URI, just with a slash
		// on the end.
		//

		// First, see if we have a "Host:" header line that we can
		// get the host:port from..
		String host_colon_port = null;
		for (Enumeration e = environment_lines.elements();
		     e.hasMoreElements(); )
		{
		    host_colon_port = (String) e.nextElement();
		    if (host_colon_port.startsWith("Host:"))
		    {
			host_colon_port = host_colon_port.substring(6,
                                          host_colon_port.length());
			break;
		    }
		    else host_colon_port = null;
		}

		String src_url;

		if (null == host_colon_port)
		{
		    src_url = "http://" + sock.getLocalAddress().getHostName();
		    int src_port = sock.getLocalPort();
		    if (src_port != 80)
		    {
			src_url += ":" + src_port;
		    }
		    src_url += original_requested_path;
		}
		else
		    src_url = "http://" + host_colon_port +
			      original_requested_path;

		String dst_url = src_url + "/";

		// Send the redirect
		sendRedirect(src_url, dst_url, http_version, null);
		return;
	    }

	    FileInputStream fis = new FileInputStream(the_file);

	    byte[] the_content = new byte[(int) the_file.length()];
	    int read_count = fis.read(the_content);
	    //System.out.println("Read " + read_count + " bytes from file, " +
	    //	       "should be " + the_file.length());
	    fis.close();

            // Serve the content
	    Vector extra_headers = new Vector();
	    extra_headers.addElement("Content-Length: " + the_file.length());
	    sendHTTPResponse(200, "OK", extra_headers, http_version,
			     content_type, the_content);
	}
	catch (IOException ioe)
	{
	    // We could get here either if the file was not found, or
	    // if something pretty bad went wrong.. for now we'll assume
	    // that the file wasn't found.
	    sendNotFoundError(http_version);

	    // Log the exception/error
	    log(0, "ERROR: HTTPRM(" + request_id + ") serveDocument(): "
		+ StringUtil.exceptionStackTrace(ioe));
	}
    }

    /**
     * This method makes sure that the client is only served files that
     * reside within the document root directory, regardless of the OS
     * that this web server is running on.  We convert all slashes to
     * backslashes (Is this the right thing to do with those?  Anyone
     * know?), we remove all double slashes (making them single), and
     * then interpret ".." dirnames in the path.  After rewriting the
     * path and making sure we're not going outside of the document
     * root, we return the rewritten path to be served.
     *                                                            <BR><BR>
     * @param requested_path the path that the browser is requesting
     *                       this web server to serve, before we make
     *                       any modifications to it.
     * @param http_version the HTTP version (if any) that the browser
     *                     sent in along with this request.  We need
     *                     this so that we know if we need to send HTTP
     *                     headers along with any error page we may
     *                     send.
     */
    public String enforceDocumentRoot(String requested_path,
				      String http_version)
    {
	//System.out.println("HTTPRM: Requested path: " + requested_path);

	// For added performance in the future, use a StringBuffer in
	// this method instead of lots of Strings..

	String enforced_path = requested_path;
	int index;

	// Make sure all backslashes are replaced with slashes, so that
	// on Windows nobody can back out of the document root using a
	// path like "\windows\win.ini"
	enforced_path = enforced_path.replace('\\', '/');

	// Now, make sure the path doesn't say "//" anywhere
	while ((index = enforced_path.indexOf(SLASHSLASH)) != -1)
	{
	    enforced_path = enforced_path.substring(0, index) +
		            enforced_path.substring(index + 1,
			    enforced_path.length());
	}

	// Next, starting from the beginning of the path and going
	// towards the end of the path, interpret any ".." to mean
	// that we should rip out the last directory from the path,
	// unless we go back beyond the document root dir (in which
	// case we'll send a Bad Request error page.
	int last_index = 0;
	while ((index = enforced_path.indexOf('/', last_index + 1)) != -1)
	{
	    // If the next token is 2 chars long (will short-circuit most
	    // cases!), and if the token is "..", then we should try to
	    // drop the last token from the path
	    if ((index - last_index) == 3 &&
		enforced_path.substring(last_index + 1, index).equals(DOTDOT))
	    {
		// Make sure we don't back up too far..
		if (last_index == 0)
		{
		    sendBadRequestError(http_version);
		    return enforced_path;
		}

		// Otherwise, clip out the last token and the ".." token
		int before_last_index = enforced_path.lastIndexOf('/',
					last_index - 1);
		if (before_last_index == -1)
		{
		    // We can't find a previous token to cut out, so this
		    // is a bad request.
		    sendBadRequestError(http_version);
		    return enforced_path;
		}

		String s = enforced_path.substring(0, before_last_index + 1);

		enforced_path = s + enforced_path.substring(index + 1,
				enforced_path.length());

		index = before_last_index;
	    }
	    last_index = index;
	}

	// The above while loop does not catch the case where you have
	// just ".." on the end (without a slash on the very end), so
	// we'll check for that here..
	if (enforced_path.endsWith(DOTDOT))
	{
	    // If the whole enforced_path now equals "/..", then this
	    // is a bad request.
	    if (enforced_path.equals(SLASHDOTDOT))
	    {
		sendBadRequestError(http_version);
		return enforced_path;
	    }

	    // Otherwise, we'll remove the last token and the "/.."
	    index = enforced_path.lastIndexOf('/');
	    last_index = enforced_path.lastIndexOf('/', index - 1);

	    // Make sure we know where to start the chop
	    if (last_index == -1)
	    {
		// We couldn't find the spot to cut.. it may not have
		// even been there, so regardless we'll send a Bad Request
		// Error page.
		sendBadRequestError(http_version);
		return enforced_path;
	    }

	    enforced_path = enforced_path.substring(0, last_index);
	}

	//if (!enforced_path.equals(requested_path))
	//    System.out.println("HTTPRM: Secured path to be: " +
	//			       enforced_path);

	return enforced_path;
    }

    /**
     * This is a generic method for sending HTTP coded responses back
     * to the browser in response to a request.  This method does not
     * automatically use a STEAM template, but you could always pass
     * in the already-filled-in content from any kind of template you
     * like.
     *                                                               <BR><BR>
     * Here are some useful error codes and error strings to use:
     *                                                           <BLOCKQUOTE>
     *     200 "OK"                                                      <BR>
     *     201 "Created"                                                 <BR>
     *     202 "Accepted"                                                <BR>
     *     204 "No Content"                                              <BR>
     *     300 "Multiple Choices"                                        <BR>
     *     301 "Moved Permanently"                                       <BR>
     *     302 "Moved Temporarily"                                       <BR>
     *     304 "Not Modified"                                            <BR>
     *     400 "Bad Request"                                             <BR>
     *     401 "Unauthorized"                                            <BR>
     *     403 "Forbidden"                                               <BR>
     *     404 "Not Found"                                               <BR>
     *     500 "Internal Server Error"                                   <BR>
     *     501 "Not Implemented"                                         <BR>
     *     502 "Bad Gateway"                                             <BR>
     *     503 "Service Unavailable"
     *                                                      </BLOCKQUOTE><BR>
     *
     * @param http_code          the HTTP result code for the response you
     *                           wish to send to the browser.
     * @param http_code_string   the String explanation of the HTTP result
     *                           code, like "Not Found".
     * @param extra_headers      a Vector of Strings, each one an HTTP header
     *                           line to be sent with the response (if any
     *                           headers get sent).  This method appends the
     *                           HTTP line termination String to each header
     *                           line, so the Strings in this Vector must not
     *                           already have them on the end!  Also, this
     *                           set of headers should not include Date,
     *                           Server, or Content-Type header lines.
     * @param http_version       the HTTP version String that the browser
     *                           sent to us, if any.
     * @param content_type       the content type String (like "text/html")
     *                           to use for this response.
     * @param http_response_body a byte array containing the body of the
     *                           response to send, instead of what this
     *                           method would send by default.
     */
    public void sendHTTPResponse(int http_code, String http_code_string,
				 Vector extra_headers, String http_version,
				 String content_type,
				 byte[] http_response_body)
    {
	// Don't send more than one HTTP response to the client..
	if (http_response_sent) return;
	
	try
	{
	    // Check to see if we should send a header at all
	    if (http_version.startsWith("HTTP/"))
	    {
		// Yes, we should send a header..
		os.write(("HTTP/1.0 " + http_code + " " + http_code_string +
			 HTTP_NEWLINE).getBytes());
		Date now = new Date();
		os.write(("Date: " + now + HTTP_NEWLINE).getBytes());
		os.write(("Server: Locomotive/" + Loco.getVersionString() +
			 HTTP_NEWLINE).getBytes());

		// Send out any extra headers, if we were given any
		if (null != extra_headers)
		{
		    String header_line;

		    // Loop through the header line Strings
		    for (Enumeration e = extra_headers.elements();
			 e.hasMoreElements(); )
		    {
			// Send the header line out to the client
			header_line = (String) e.nextElement();
			header_line = header_line + HTTP_NEWLINE;
			os.write(header_line.getBytes());
		    }
		}

		// Send the Content-Type
		if (null == content_type)
		{
		    // We'll default to text/html
		    os.write(("Content-type: text/html" + HTTP_NEWLINE +
			     HTTP_NEWLINE).getBytes());
                }
                else
                {
                    os.write(("Content-type: " + content_type + HTTP_NEWLINE +
                             HTTP_NEWLINE).getBytes());
                }
	    }

            if (null == http_response_body)
            {
                // Send the default response body, it's a simple "error" page
		String response_content = "<HTML>\n<HEAD>\n<TITLE>" +
		    http_code_string + "</TITLE>\n</HEAD>\n<BODY>\n" +
		    "<H1>HTTP " + http_code + ": " + http_code_string +
		    "</H1>\n<BR>\n<BR>\n<HR>\n<I><A HREF=\"" +
		    "http://www.locomotive.org/\">Locomotive " +
		    http_server.getVersionString() + "</A></I>\n<BR>\n" +
                    "</BODY>\n</HTML>\n";

		http_response_body = response_content.getBytes();
            }

	    // Send the response body
	    os.write(http_response_body);

	    // Send anything that may be waiting to go!
	    os.flush();
	}
	catch (IOException ioe)
	{
	    log(0, "ERROR: HTTPRM(" + request_id + ") sendHTTPResponse(): " +
		StringUtil.exceptionStackTrace(ioe));
	}

	// We just sent a response, so don't send another one to this same
	// client socket later!
	http_response_sent = true;

	//
	// Log the request to the access log
	//

	// Put the client's hostname/ip address into a String
	InetAddress address = sock.getInetAddress();
	String client_host = "unknown";
	if (null != address) client_host = address.toString();
	int index = client_host.indexOf('/');
	if (index != -1)
	    client_host = client_host.substring(0, index);

	// Log this request to the access log
	http_server.logAccess(client_host, "-", "-", http_request_line,
			      http_code, http_response_body.length);
    }

    public void sendNotImplementedError(String http_version)
    {
	sendHTTPResponse(501, "Not Implemented", null, http_version, null,
                         null);
    }

    public void sendNotFoundError(String http_version)    {
	sendHTTPResponse(404, "Not Found", null, http_version, null, null);
    }

    public void sendBadRequestError(String http_version)
    {
	sendHTTPResponse(400, "Bad Request", null, http_version, null, null);
    }

    public void sendRedirect(String source_url, String destination_url,
			     String http_version, byte[] response_body)
    {
	// Add the "Location: " header
	Vector extra_headers = new Vector();
	extra_headers.addElement("Location: " + destination_url);
	
	// If the caller didn't send us a response body, we'll just create
	// and use our own simple default one..
	if (null == response_body)
	{
	    // Create our own default redirect response body
	    StringBuffer b = new StringBuffer();

	    String response_page =
                "<HTML>\n<HEAD>\n<TITLE>301: Moved Permenantly</TITLE>\n" +
                "</HEAD>\n<BODY>\n<H1>HTTP 301: Moved Permenantly</H1>\n" +
                "<BR>\n<BR>\nThis resource has been permenantly moved to:\n" +
                "<A HREF=\"" + destination_url + "\">" + destination_url +
                "</A>\n<BR>\n<BR>\nPlease update your " +
                "bookmarks.\n<BR>\n<BR>\n<HR>\n<A HREF=\"http://" +
                "www.locomotive.org/\"><I>Locomotive " +
		Loco.getVersionString() + "</I></A>\n</BODY>\n</HTML>\n";

	    response_body = response_page.getBytes();
	}

	sendHTTPResponse(301, "Moved Permenantly", extra_headers,
			 http_version, null, response_body);
    }

    /**
     * This method guesses the MIME Content-type by looking at the file
     * extension.  Right now it doesn't know many MIME types, so this will
     * need to be expanded!     */
    public String guessContentTypeFromName(String name)
    {
	if (name.endsWith(".html") || name.endsWith(".htm"))
	{
	    return "text/html";
	}
	if (name.endsWith(".gif"))
	{
	    return "image/gif";
	}
	if (name.endsWith(".class"))
	{
	    return "application/octet-stream";
	}
	if (name.endsWith(".jpg") || name.endsWith(".jpeg"))
	{
	    return "image/jpeg";
	}

	// And, finally, if we have no idea, it's just text/plain.
	return "text/plain";
    }


/*---------------- CLEANUP ---------------------*/


    private void doCleanup()
    {
	closeStreams();
    }

    private void closeStreams()
    {
	try
	{
	    is.close();
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: HTTPRM(" + request_id + ") closeStreams(): at is: "
		 + StringUtil.exceptionStackTrace(ioe));
	}

	try
	{
	    os.close();
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: HTTPRM(" + request_id + ") closeStreams(): at os: "
		 + StringUtil.exceptionStackTrace(ioe));
	}
    }



/*------------------------- UTILITY ---------------------*/



    /** logs to the server_log */
    protected void log(int lev, String s)
    {
	server_log.log(lev, "HTTPRM(" + request_id + ") " + s);
    }
    

    /** 
     * Use this to send an error message via the generic error STEAM template.
     */
    protected void sendErrorMsg(String message)
    {
	java.util.Date dt = new java.util.Date();

	// Now, let's create the request object we'll use for the setup.
	// we'll create an HTTPReqeust object to handle this as any web
	// server would..
	OutputStream sock_output_stream;
	try
	{
	    sock_output_stream = sock.getOutputStream();

	    resp = new HTTPResponse(sock_output_stream, request_id);
	    
	    // note that if the Response has been flushed at all, the
	    // http headers have already been sent.  There is no way to avoid
	    // strange behaviours in the output at this point.  

	    // clean up whatever else has been placed in the response thus far.
	    resp.setContentType("text/html");
	    resp.clearHeaders();
	    resp.clearBody();
	    
	    // do the error message. now uses template, new for 3.3
	    Hashtable esubs = new Hashtable ();
	    esubs.put ("ERROR_MSG", message);
	    esubs.put ("REQUEST_ID", String.valueOf (request_id));
	    esubs.put ("SYSTEM_TAG", config.getString ("SYSTEM_TAG"));
	    esubs.put ("SYSTEM_INSTANCE_ID", 
		       config.getString ("SYSTEM_INSTANCE_ID"));
	    esubs.put ("ERROR_DATE", dt.toString());

	    String s = Loco.evaluateTemplate ("error/generic_error.tmpl", 
					      esubs);
	    if (s.equals(""))
	    {
		s = "<HTML>LOCOMOTIVE HTTPD ERROR: " +
		    "<BR>ERROR_MSG: " + message + 
		    "<BR> (The generic error template is missing. " +
		    " Administrator, please configure) </HTML>";
	    }

	    resp.addString (s);
	    resp.flush (server_log);

	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: HTTPRM(" + request_id + ") doSetup(): " + 
		 StringUtil.exceptionStackTrace(ioe));
	}
    }


/*-------------------- INIT AND SHUTDOWN METHODS --------------*/

    /**
     * Set up all persistant data and initialize the various services
     */
    boolean initialize(HTTPServerConfig hc, Log s_log, Log e_log)
    {
	config = hc;
	server_log = s_log;
	event_log = e_log;

	// Save a reference to our LOCO_REQUEST_TOKEN
	loco_request_token = hc.getString("LOCO_REQUEST_TOKEN");

	// store the accept ip and subnet in byte arrays
	// see the explanations at the beginning of the class
	StringTokenizer ips = 
	    new StringTokenizer(hc.getString("HTTPD_ACCEPT_IP"), ",");
	StringTokenizer sus = 
	    new StringTokenizer(hc.getString("HTTPD_ACCEPT_SUBNET"), ",");

	int ip_length = ips.countTokens();
	if (ip_length != sus.countTokens())
	{
	    s_log.log(0, "ERROR in config: HTTPD_ACCEPT_IP does not have the" +
		"same number of addresses as HTTPD_ACCEPT_SUBNET");
	    return false;
	}

	ip_accept = new int [ip_length];
	ip_subnet = new int [ip_length];
	ip_calc =   new int [ip_length];

	for (int j = 0; j < ip_length; j++)
	{
	    String ip_str = ips.nextToken();
	    StringTokenizer ip = new StringTokenizer(ip_str, ".");

	    if (ip.countTokens() != 4)
	    {
		s_log.log(0, "ERROR in HTTPD_ACCEPT_IP config: " +
			  " IP addresses must have 4 octets: " + ip_str);
		return false;
	    }
		
	    for (int i = 0; i < 4; i++)
	    {
		try
		{
		    int temp = Integer.parseInt(ip.nextToken());
		    temp = temp << (24 - (i*8));
//	            Loco.log(9, "ip[" + i + "]: " + temp);
		    ip_accept[j] += temp;
		}
		catch (NumberFormatException e) 
		{
		    s_log.log(0, "ERROR in HTTPD_ACCEPT_IP config: " +
			  "could not convert IP to number: " + ip_str);
		    return false;
		}
	    }

	    ip_str = sus.nextToken();
	    StringTokenizer su = new StringTokenizer(ip_str, ".");
	    if (su.countTokens() != 4)
	    {
		s_log.log(0, "ERROR in HTTPD_ACCEPT_IP config: " +
			  " Subnet masks must have 4 octets: " + ip_str);
		return false;
	    }
	    
	    for (int i = 0; i < 4; i++)
	    {
		try
		{
		    int temp = Integer.parseInt(su.nextToken());
		    temp = temp << (24 - (i*8));
//	                Loco.log(9, "su[" + i + "]: " + temp);
		    ip_subnet[j] += temp;
		}
		catch (NumberFormatException e) 
		{
		    s_log.log(0, "ERROR in HTTPD_ACCEPT_SUBNET: " +
			  "could not convert subnet to number: " + ip_str);
		    return false;
		}
	    }
	    
	    ip_calc[j] = ip_subnet[j] & ip_accept[j];

	    s_log.log(9, "ip accept pair: " + ip_accept[j] +  
		    " subnet: " + ip_subnet[j]);
	    
	}

	return (true);
    }


    /**
     * shuts down the various services currently running, and saves
     * the cached session objects, if session object caching is turned on
     * called by Loco.shutdownAll () during server initiated bail 
     * which, we hope, is how it always goes down
     */
    static boolean shutdown()
    {
	// What should we do here???  Do we even need this anymore?

	return (true);
    }
}
