/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.server.*;
import org.locomotive.util.*;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;


/**
 * The response class for regular HTTP requests.
 * It uses the Response outputStream for its HTTPOutputStream, 
 * and its own internal printwriter for the printWriter
 */
public class HTTPResponse extends LocoServletResponse 
{

    // vars
    protected boolean used_writer = false;
    protected boolean used_os = false;
    private static int MAX_HEADERS = 32;
    protected PrintWriter resw = null;
    private String header_names[];
    private String header_values[];
    private int next_header_index = 0;

    public HTTPResponse(OutputStream os, int request_id)
    {
	super(os, request_id, Response.REPLY_TYPE_CGI);
	content_type = "text/plain";
	header_names = new String[MAX_HEADERS];
	header_values = new String [MAX_HEADERS];
    }
    /**
     * adds an HTTP protocol header to the response.  Adding any headers
     * after the first flush() will be ineffective, since the HTTP 
     * header has already been sent off to the webserver/client. 
     * The values will be formatted in a manner appropriate for 
     * HTTP - similar to RFC 822 -  &quot; Name: Value &quot; <BR>
     * An example might be "Pragma: no-cache"
     * @param name - the name 
     * @param value - the value associated with name.
     */
    public void addHTTPHeader (String name, String value)
    {
	/* developers note: with the NSAPI and ISAPI, the name value pairs
	   are formatted at the web server.  Only in the case of CGI are
	   we supposed to do it.  Often in the case of CGI the headers must 
	   be reinterpreted by the CGI handler of the web server, which 
	   adds to CPU overhead on the webserver machine...
	   */
	if (next_header_index < MAX_HEADERS)
	{
	    header_names[next_header_index] = name;
	    header_values[next_header_index] = value;
	    next_header_index++;
	}
    }

    /**
     * returns the number of items currently scheduled to be sent in the 
     * HTTP header
     */
    public int getHeaderCount()
    {
	return (next_header_index);
    }

    /**
     * Returns true if a header with the specified name has been set.
     */
    public boolean containsHeader(String name)
    {
	if (name == null)
	{
	    return (false); 
	}

	for (int i = 0; i < header_names.length; i++)
	{
	   if (name.equals(header_names [i]))
	   {
	       return (true);
	   }
	}
	
	return (false);
    }

    /** This writes the headers to the OutputStream in correct protocol. 
     * This get's called automatically when either of the two
     * inner stream flush, so we don't have to worry about calling it
     * explicitly.
     */
    public void writeHeader (Log server_log) throws IOException
    {
	StringBuffer sb = new StringBuffer ();
        String status_value = null;

	addStatusToHeaders();

        // First, we need to send the HTTP response line, so we need to
        // find the status header in order to send it..
        for (int i = 0; i < next_header_index; i++)
        {
            // Look for a line starting with "status"
            if (header_names[i].startsWith("status"))
            {
                status_value = header_values[i];

                // If we have a status value, use it..
                if (null != status_value)
                {
                    // Output the HTTP response line with the status
                    sb.append("HTTP/1.0 " + status_value + "\r\n");
                }
                else
                {
                    // Otherwise, just send HTTP/1.0 200 OK
                    sb.append("HTTP/1.0 200 OK\r\n");
                }

                // Replace this header line with the Server: header
                header_names[i] = "Server";
                header_values[i] = "Locomotive/" + Loco.getVersionString();
            }
        }

	// always do content-type - CGI requires it.
	sb.append ("Content-Type: ");
	sb.append (content_type);
	sb.append ("\r\n");

	for (int i = 0; i < next_header_index; i++)
	{
	    sb.append (header_names[i]);
	    sb.append (": ");
	    sb.append (header_values[i]);
	    sb.append ("\r\n");
	}

	// new line after header seperates header from data to follow
	sb.append("\r\n");

	// java IO is a bit inefficient - alot of data gets copied about in 
	// memory, plus many looping function calls - have a look in 
	// the source for OutputStream.write() 
	// some time.  Pointers still have their uses. See 
	// "Why Object Orientation is Inappropriate for Building Operating 
	// Systems"
	String output = sb.toString();

	byte b[] = output.getBytes();

	try
	{
	    os.write (b, 0, b.length);
	    os.flush();
	}
	catch (IOException ioe)
	{
	    if (server_log != null)
	    {
		server_log.log (0, "ERROR: RESP(" + request_id + 
				") write header failed: " + ioe.toString());
	    }
	    throw ioe;
	}
	    
	header_was_flushed = true;
    }
}
