/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.server.*;
import org.locomotive.util.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.text.SimpleDateFormat;

/**
 * The org.locomotive.loco.HTTPServer is the Locomotive's embedded web server.
 * The server instances are launched by the org.locomotive.loco.Loco object on
 * startup.
 */
public class HTTPServer extends Thread
{
    private ServerSocket ss;
    private Socket cs;
    private HTTPServerConfig config;
    private Log server_log;
    private Log event_log;
    private Log access_log;

    private java.util.Date startup_date;
    private boolean stop_accepting = false;
    private int request_id = 0;
    private int max_pooled_threads = 0;
    private int min_pooled_threads = 0;
    private ThreadPoolManager tpm; 

    private int initial_cpl_bufsize = 16384;
    String document_root;                  // set below
    String index_file;                     // set below
    private String instance_name;          // set below
    private Thread current;                // set below
    private LocoConfig locoConfig;         // set below

    private SimpleDateFormat access_time_formatter = new SimpleDateFormat(
        "dd/MMM/yyyy:HH:mm:ss z");

    /** a basic wrapper for the constructor to package protect it */
    HTTPServer(String name)
    {
	super(name + "-http-server");
	instance_name = name;
	current = this;
    }

    /**
     * This method is called by the Loco Object
     */
    public void run() {
    {
	// get the configs

	// As long as we got a reference to the Locomotive's LocoConfig
	// object, we can continue.
	locoConfig = Loco.getConfig();
	if (null == locoConfig)
	{
	    server_log.log(0, "HTTPD" + instance_name 
			   + " - ERROR: Did not get a valid Locomotive " 
			   + "Config object reference!");
	    System.exit(-1);
	}	

	// Formulate the path to this HTTPServer instance's config file
	String server_root = locoConfig.getString("SYSTEM_SERVER_ROOT_PATH");

	// Remove any trailing file separator (either slash or backslash)
	if (server_root.endsWith(File.separator))
	    server_root = server_root.substring(0, server_root.length() - 1);

	try 
	{
	    config = new HTTPServerConfig (new String[0], server_root +
		     File.separator + "config" + File.separator + "httpd" +
		     instance_name + ".conf");
	}
	catch (IllegalArgumentException iae)
        {
            System.exit(-1);
        }

        try
        {
            // Set the name of this HTTPServer instance into the Config object
            config.changeConfig("SYSTEM_INSTANCE_ID", instance_name);
        }
        catch (Exception e)
        {
            // Probably an org.locomotive.server.ConfigInitializationException
            e.printStackTrace();
        }

	startup_date = new java.util.Date();

	// setup the primary logs
	try
	{
	    server_log = new Log(config.getInt("LOG_LEVEL"), 
				 locoConfig.getString(
				 "SYSTEM_SERVER_ROOT_PATH"),
				 config.getString("LOG_PATH"));

	    event_log = new Log(config.getInt("EVENT_LOG_LEVEL"), 
			        locoConfig.getString(
				"SYSTEM_SERVER_ROOT_PATH"),
				config.getString("EVENT_LOG_PATH"));

	    access_log = new Log(Log.INFO, locoConfig.getString(
				 "SYSTEM_SERVER_ROOT_PATH"),
				 config.getString("ACCESS_LOG_PATH"));
	    // Don't prepend the date to the access log output lines
	    access_log.setPrependDate(false);
	}
	catch (IOException ioe)
	{
	    System.out.println ("Open logs failed: " + ioe.toString());
	    System.exit (-1);
	}

	server_log.log (0, "HTTP" + instance_name + " started: config file: " 
			+ config.getConfigFilePath());


	// startup the 
	max_pooled_threads = config.getInt ("HTTPD_MAX_POOLED_THREADS");
	min_pooled_threads = config.getInt ("HTTPD_MIN_POOLED_THREADS");

	tpm = new ThreadPoolManager (server_log, min_pooled_threads, 
				     max_pooled_threads);

	// initialize the CachedPageLoader used in evaluateTemplate 
	// below.
	document_root = config.getString ("HTTPD_DOCUMENT_ROOT_PATH");

	// prepare server socket for accept
	int loco_port = config.getInt("HTTPD_PORT");
	int loco_port_backlog = config.getInt ("HTTPD_PORT_BACKLOG");

	try
	{
	    ss = new ServerSocket (loco_port, loco_port_backlog);
	}
	catch (IOException ioe)
	{
	    server_log.log(0, "HTTPD - ERROR: ServerSocket on port " 
			   + loco_port + ": " + ioe.toString());
	    shutdownAll();
	    System.exit(-1);
	}

	server_log.log(0, "HTTPD" + instance_name 
		       + " - Now accepting connections...");

	// main loop
	while (! stop_accepting)
	{
	    request_id++;
	    try
	    {
		//if (request_id % 50 == 7)
		//{
		//    Runtime r = Runtime.getRuntime();
		//    server_log.log (0, "before gc (free, total): " + 
		//		    r.freeMemory() + ", " + r.totalMemory());
		//    System.gc();
		//    server_log.log (0, "after gc (free, total): " + 
		//		    r.freeMemory() + ", " + r.totalMemory());
		//}

		cs = ss.accept();
		dispatch ();
	    }
	    catch (InterruptedIOException ie)
	    {
		server_log.log(0, "HTTPD - INTERRUPT! shutdown initiated: " + 
			       ie.getMessage());
	    }
	    catch (IOException ioe)
	    {
		server_log.log (1, "HTTPD - ERROR: accept failed: " + 
				ioe.toString());
	    }
	}

	server_log.log (0, "HTTPD" + instance_name 
			+ " - Shutdown issued, closing accept socket");
	try 
	{
	    ss.close();
	}
	catch (IOException ioe)
	{
	    server_log.log(0, "HTTPD" + instance_name 
			   + " - ERROR: close Server Socket failure: " + 
			   ioe.toString()); 
	}
	server_log.log (0, "HTTPD" + instance_name 
			+ " - accept socket closed");

	shutdownAll();
    }
    }
	

    /** 
     * used by main accept loop to serve HTTP requests (either serve
     * a file or launch new RequestManagers).  
     * This will block and wait if too many RMs are active.
     */ 
    private void dispatch()
    {
	// getPooledThread blocks until a thread becomes available
	PooledThread pt = tpm.getPooledThread();

	server_log.log(2, "HTTPD" + instance_name 
		       + " - dispatch REQ(" + request_id + 
		       ") active: " + tpm.getActiveCount() + " out of " + 
		       tpm.getPoolSize());

	HTTPRequestManager rm = new HTTPRequestManager(this, cs, request_id);

	// initialized globals in HTTPRequestManager - ready it for handling
	if (! rm.initialize(config, server_log, event_log))
	{
	    System.out.println("HTTPRequestManager initialization failure" + 
			       " - see logs");
	    System.exit(-1);
	}

	pt.loadRunnable(rm);
        pt.go();
    }



    /**
     * log to the HTTPServer log
     * @param lev the log level
     * @param s a string to be logged
     */
    public void log(int lev, String s)
    {
	server_log.log(lev, s);
    }

    /**
     * log to the HTTPServer event log - simplified version
     * @param level - the log level
     * @param hd - a HandlerData object as passed into a handler
     * @param major - the major event region or code section
     * @param minor - the minor event code
     * @param data  - event specific data
     */
//      public void logEvent (int level, HandlerData hd, String major, 
//  				 String minor, String data)
//      {
//  	StringBuffer sb = new StringBuffer (128);

//  	sb.append(hd.getRequestID());
//  	sb.append("\t");
//  	sb.append(hd.session.getId());
//  	sb.append("\t");
//  	sb.append(hd.user.getUserid());
//  	sb.append("\t");
//  	sb.append(hd.user.getUsername());
//  	sb.append("\t");
//  	sb.append(major);
//  	sb.append("\t");
//  	sb.append(minor);
//  	sb.append("\t");
//  	sb.append(data);
//  	event_log.log(level, sb.toString());
//      }

    /**
     * log to the HTTPServer event log
     * @param level - the log level
     * @param rid - the response id (used mostly for debugging)
     * @param sess - the current session
     * @param u    - the User
     * @param major - the major event region or code section
     * @param minor - the minor event code
     * @param data  - event specific data
     */
    public void logEvent(int level, int rid, Session sess, User u, 
				  String major, String minor, String data)
    {
	StringBuffer sb = new StringBuffer (128);

	sb.append(rid);
	sb.append("\t");
	sb.append(sess.getId());
	sb.append("\t");
	sb.append(u.getUserid());
	sb.append("\t");
	sb.append(u.getUsername());
	sb.append("\t");
	sb.append(major);
	sb.append("\t");
	sb.append(minor);
	sb.append("\t");
	sb.append(data);
	event_log.log(level, sb.toString());
    }

    /**
     * Log a request to the HTTPServer access log
     *                                                             <BR><BR>
     * @param client_host The client's IP address (for now)
     * @param remote_username The rfc931 username of the user at the client
     * @param authenticated_username The authenticated username as specified
     *                               by rfc931
     * @param request_line The request line exactly as it came from the client
     * @param response_status The numeric HTTP status code returned to the
     *                        client
     * @param response_length The number of bytes of the response body that
     *                        was sent to the client
     */
    public void logAccess(String client_host, String remote_username,
			  String authenticated_username, String request_line,
                          int response_status, int response_length)
    {
	//System.out.println("HTTPServer.logAccess: called.");

	StringBuffer sb = new StringBuffer(128);

	// Format the current time for the common web log format
	Date access_date = new Date();
	String request_date = access_time_formatter.format(access_date);

	sb.append(client_host);
	sb.append(" ");
	sb.append(remote_username);
	sb.append(" ");
	sb.append(authenticated_username);
	sb.append(" [");
	sb.append(request_date);
	sb.append("] \"");
	sb.append(request_line);
	sb.append("\" ");
	sb.append(response_status);
	sb.append(" ");
	sb.append(response_length);
	access_log.log(1, sb.toString());
    }

    /**
     * gets the reference to the main server log of this HTTPServer instance
     */
    public Log getServerLog()
    {
	return (server_log);
    }

    /**
     * gets the reference to the event log of this HTTPServer instance
     */
    public Log getEventLog()
    {
	return (event_log);
    }

    /**
     * gets the reference to the access log of this HTTPServer instance
     */
    public Log getAccessLog()
    {
	return (access_log);
    }

    /**
     * Returns the number of active RequestManager Threads.  Allows you to 
     * get and idea of current load/concurrency.  
     */
    public int getActiveThreadCount()
    {
	return (tpm.getActiveCount());
    }

    /**
     * Returns the number of active RequestManager Threads.  Allows you to 
     * get and idea of current load/concurrency.  
     */
    public int getThreadPoolSize()
    {
	return (tpm.getPoolSize ());
    }

    /**
     * Returns the number of active RequestManager Threads.  Allows you to 
     * get and idea of current load/concurrency.  
     */
    public int getMaxThreadPoolSize()
    {
	return (tpm.getMaxCount ());
    }

    /**
     * returns the startup date of the HTTPServer
     */
    public java.util.Date getStartupDate ()
    {
	long mills = startup_date.getTime ();
	java.util.Date d = new java.util.Date ();
	d.setTime (mills);
	return (d);
    }

    /**
     * gets the current value of the request id.  The value is 
     * incremented just before pulling a new client off the queue. 
     */
    public int getCurrentRequestID ()
    {
	return (request_id);
    }


/*------- WRAPPED METHODS FROM REQUEST MANAGER - IN BOTH PLACES FOR NOW ----*/

    
    /**
     * returns the current value of the auto_session_handling flag.
     * @see setAutoSessionHandling
     */ 
    public static boolean getAutoSessionHandling ()
    {
	return (RequestManager.auto_session_handling);
    }

    /**
     * for shutting down the HTTPServer.  After handling the next inbound 
     * connection, the Loco will stop accepting new connections, do various
     * cleanups, then exit
     */
    public void setStopAccepting()
    {
	stop_accepting = true;
	current.interrupt();
	server_log.log(0, "setStopAccepting: Interrupt issued");
    }

    /**
     * called internally for cleanup 
     */
    private void shutdownAll()
    {
	// wait for all active rms to end (thus returning their 
	// db connections to the internal dbcp
	int loopcount = 0;
	int active_count = 0;

	while (((active_count = tpm.getActiveCount ()) > 0) && loopcount < 15)
	{
	    server_log.log (0, "Loco - shutdown waiting: still active: " + 
			    active_count);
	    loopcount++;
	    try
	    {
		Thread.sleep (1000);
	    }
	    catch (InterruptedException ie) {}
	}

	// oh dear, their stuck.  
	if (active_count > 0)
	{
	    server_log.log (0, "Loco - WARNING: wedged rm - forging onward");
	}

	server_log.log (0, "Loco - calling RM.shutdown()");

	if (! RequestManager.shutdown ())
	{
	    // no reason for this to happen 
	    server_log.log(0, "ERROR: RequestManager: could not shut down " +
			   "successfully");
	}

	server_log.log (0, "HTTPServer - Shutdown Complete. Exiting..a.a...argh!");
	//System.exit (0);  // System.exit() is a bad call to make!
    }

    /**
     * returns the HTTPServerConfig of the server
     */
    public HTTPServerConfig getConfig ()
    {
	return (config);
    }
    
    /**
     * returns the Version Number of the server (for now, returns the same
     * String as Loco.getVersionString();
     */
    public String getVersionString ()
    {
	return Loco.getVersionString();
    }
}
