/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.server.*;
import org.locomotive.util.*;


/** 
 * The config object for the HTTPServer.  Each of the HTTPServer configs
 * that can be set are listed and documented in the sample httptest.conf file
 * included with each distribution.  To see what the current config
 * settings are, log into the Locomotive has an admin user and
 * request the srv URL (i.e. 'http://www.yourdomain.com/cgi-bin/locolink/srv').
 * To change any configs during runtime, use the changeConfig() method.
 * For more information, see the 
 * <a href="http://www.leverage.com/docs/install/install_noframe.html#_Toc424012341">Locomotive Installation Guide</a> or the 
 * <a href="http://www.leverage.com/docs/admin/admin_noframe.html#_Toc425673336">Locomotive Administrator's Guide</a>.
 */  
public class HTTPServerConfig extends org.locomotive.server.Config
{

    /** This creates an HTTPServerConfig object. It takes the same comman 
     *	arguments as its super class. 
     */
    public HTTPServerConfig (String args[], String defaultpath) 
	throws IllegalArgumentException
    {
	super (args, defaultpath);
    }
    

    /** This establishes the default setting for all the Locomotive configs.
     */
    public void setupDefaults()
    {
	// system tag for email and cookies
	addConfig ("SYSTEM_TAG", "DEF_SYS_TAG", 
		   forstring, "System Identity Tag");
	addConfig ("SYSTEM_INSTANCE_ID", "test",
		   forstring, "HTTPServer Instance Identifier");
	addConfig ("HTTPD_DOCUMENT_ROOT_PATH", "/usr/local/loco/www", 
		   forstring, "Basedir for all file paths - absolute");
	addConfig ("HTTPD_DOCUMENT_INDEX_FILE", "index.html", 
		   forstring, "Path to home page - absolute");
	addConfig ("LOCO_REQUEST_TOKEN", "/locolink", 
		   forstring, "Token to route requests to Locomotive");

	addConfig ("HTTPD_PORT", "5421",
		   forint, "TCP port for Request servicing");
	addConfig ("HTTPD_PORT_BACKLOG", "5",
		   forint, "TCP Request servicing port backlog (listen)");
		   
	// threads
	addConfig ("HTTPD_MAX_POOLED_THREADS", "16", 
		   forint, "Maximum number Pooled Threads for RequestManagers");
	addConfig ("HTTPD_MIN_POOLED_THREADS", "2", 
		   forint, "Minimum number Pooled Threads for RequestManagers");
	// Logs
	addConfig ("LOG_LEVEL", "9", 
		   forint, "HTTPServer instance Log log verbosity value");
	addConfig ("LOG_PATH", "logs/httpdtest-%.log", 
		   forstring, "Relative path for HTTPServer instance Logs");

	addConfig ("EVENT_LOG_LEVEL", "10", 
		   forint, "HTTPServer instance events logging verbosity");
	addConfig ("EVENT_LOG_PATH", "logs/httpdtestevents-%.log", 
		   forstring, "path to HTTPServer instance event log");

	addConfig ("ACCESS_LOG_PATH", "logs/httpdtestaccess-%.log", forstring,
		   "path to this HTTPServer instance's access log.");

	addConfig("HTTPD_ACCEPT_IP", "127.0.0.0", forstring,
		  "The network ip to allow connections from");
	addConfig("HTTPD_ACCEPT_SUBNET", "255.255.255.0", forstring,
		  "the size of the subnet to allow");

    }

    /**
     * a package friendly method that outputs an html table with rows
     * containing the configurations names and current values. 
     * Used by the ServerHandler status display
     */
    String toHTMLRows ()
    {
/*	String names[] = {"SYSTEM_TAG", "SYSTEM_INSTANCE_ID", 
			  "SYSTEM_SERVER_ROOT_PATH", 
			  "HANDLER_ROUTING_TABLE_PATH", "LOCO_PORT", 
			  "LOCO_PORT_BACKLOG", "LOCO_MAX_POOLED_THREADS", 
			  "LOCO_MIN_POOLED_THREADS", "LOCO_SESSION_EXPIRE_TIME",
			  "LOCO_SESSION_AUTOHANDLE", 
			  "LOCO_CHECK_NEW_TEMPLATE_FREQ", 
			  "DB_AUTOCONNECT", "DB_TYPE",
			  "DB_DRIVER_CLASSNAME", "DB_CONNPOOL_URL", 
			  "DB_CONNPOOL_MIN_CONNECTIONS", 
			  "DB_CONNPOOL_MAX_CONNECTIONS", 
			  "LOCO_SESSION_AUTOHANDLE",
			  "LOG_LEVEL", "EVENT_LOG_LEVEL", 
			  "LOCO_URL_MAIN", "HTTPD_HOME_PAGE", 
			  "LOCO_COOKIE_DOMAIN_MAIN"};
			  */

	StringBuffer sb = new StringBuffer ();
	// sb.append("<TABLE>\n");
	java.util.Enumeration e = getConfigNamesEnumeration();
	while (e.hasMoreElements())
	{
	    String name = (String) e.nextElement();
	    sb.append ("\t<TR><TD>");
	    sb.append (name);
	    sb.append ("</TD><TD align=right>");
	    sb.append (StringUtil.htmlQuoteString (getString (name)));
	    sb.append ("</TD></TR>\n");
	}
//	sb.append("</TABLE>");
	return (sb.toString());
    }


    /** changes the value of a config at runtime.  Checks for change in 
     * DB_TYPE, and, if changed, resets the DB_TYPE variable. Also checks to
     * make sure SERVICE_DEFAULT is either HANDLER or SERVLET.
     * @param name: the name of the config
     * @param value the new value for the config.
     * @returns the value of changed config, which may have been altered 
     *          to allow it to match the config's constraints.
     * @exception ConfigNotFoundException if the config does not exist.
     * @exception ConfigInitializationException if the config value does not
     *            match the required constraints.  For DB_TYPE, value must be
     *            either 'Oracle', 'Informix', 'Cloudscape', MySQL, or 'Solid'.
     *            For SERVICE_DEFAULT, this must be either 'Handler' or 
     *            'Servlet'.
     */
    public String changeConfig (String name, String value)
	throws ConfigNotFoundException, ConfigInitializationException

    {
	return super.changeConfig(name, value);
    }
}
