/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.loco.handler.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.server.*;
import org.locomotive.util.*;
import org.locomotive.steam.CachedPageLoader;
import org.locomotive.util.cache.DBCache;
import org.locomotive.directory.loco.BeanDirectory;
import java.io.*;
import java.net.*;
import java.util.*;
import java.sql.*;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * The org.locomotive.loco.Loco is the primary entry point for the Org.Locomotive.
 * Application Server.  The server is launched by invoking the java 
 * interpreter on this class, which contains the main (String args[]) method.
 * <P>A series of srguments may be specified as command line options.
 * These include:<UL>
 * <LI>-h  help- prints usage.
 * <LI>-f  sets config file path.   It defaults to ./loco.conf
 * <LI>-d  dumps configs
 * <LI>-dv dumps configs- verbose
 * <LI>-0  enters any number of properties into Config file. 
 *         Usage: NAME=VALUE
 * </UL>
 * Only config properties which already exist in the default config tables
 * can be added.  If property is NAME="", then the value is entered as "", 
 * erasing any previous values.  
 */
public class Loco {
   private static ServerSocket ss;
   private static Socket cs;
   private static LocoConfig config;
   private static Log server_log;
   private static Log event_log;
   private static DBConnectionPool dbcp;
   private static DBCache dbCache;

   private static java.util.Date startup_date;
   private static boolean stop_accepting = false;
   private static int request_id = 0;
   private static int max_pooled_threads = 0;
   private static int min_pooled_threads = 0;
   private static ThreadPoolManager tpm; 

   // new for 3.3
   private static int initial_cpl_bufsize = 16384;
   private static String template_root;          // set below
   private static int check_templates_freq = 1;  // overridden below
   private static Thread current;

   private String serverRoot;

   /** a basic wrapper for the constructor */
   public Loco(){
      setServerRoot(System.getProperty("loco.home"));
   }

   /**
    * This method is called by the JVM at server startup.   
    */
   public static void main (String argv[])
   {
       // THIS TEMPORARY VARIABLE ALLOWS US TO IMPLEMENT BEAN DIRECTORY 
       // CONFIGURATION FOR THE LOCOCONFIG CLASS WITHOUT BREAKING THE BUILD.  
       // WHILE DEVELOPING THIS FEATURE, FLIP THE VARIABLE TO TRUE.  WHEN 
       // COMMITTING CODE, FLIP IT TO FALSE.  WHEN DEVELOPMENT IS COMPLETE, 
       // WE'LL REMOVE THE VARIABLE ENTIRELY.
      boolean LOCO_CONFIG_BEAN_DIRECTORY_ACTIVATED = false;

      // get the configs
      try {
	  if (LOCO_CONFIG_BEAN_DIRECTORY_ACTIVATED && 
	      BeanDirectory.isRunning())
	  {
	      config = (LocoConfig) BeanDirectory.resolve(
		  "/services/LocoConfig");
	      System.out.print("Calling config.init with args = ");
	      int argvlen = argv.length;
	      for (int i=0; i<argvlen; i++)
		  System.out.print(argv[i] + " ");
	      System.out.println();

	      //TODO: need to disable -f argument, but pass others.
	      // perhaps this is best done in the Config class
	      //config.init (argv, org.locomotive.server.Config.BEAN);
	      config.init (new String[0], org.locomotive.server.Config.BEAN);

	      System.out.println("Configs: ");
	      config.doConfigDump();
	  }
	  else
	      config = new LocoConfig (argv, "./loco.conf");
      } catch (IllegalArgumentException iae) {
	  System.exit (-1);
      }
      
      startup_date = new java.util.Date();

      if (!BeanDirectory.isRunning()) {
         // setup the two primary logs
         try {
            server_log = new Log (config.getInt ("LOG_LEVEL"), 
                                  config.getString ("SYSTEM_SERVER_ROOT_PATH"),
                                  config.getString ("LOG_PATH"));

            event_log = new Log (config.getInt ("EVENT_LOG_LEVEL"), 
                                 config.getString ("SYSTEM_SERVER_ROOT_PATH"),
                                 config.getString ("EVENT_LOG_PATH"));
         } catch (IOException ioe) {
            System.out.println ("Open logs failed: " + ioe.toString());
            System.exit (-1);
         }
      }

      server_log.log (0, "Loco started: config file: " + 
                      config.getConfigFilePath());

      if (!BeanDirectory.isRunning()) {
         // Create the DBConnectionPool
         boolean db_autoconnect = 
         config.getString("DB_AUTOCONNECT").equalsIgnoreCase("YES");
         if (db_autoconnect) {
            if (! createDBConnectionPool(config)) {
               server_log.log(0,
                              "DBConnectionPool initialization failed.");
               System.out.println("Warning: DBConnectionPool creation " +
                                  "failure - see logs.");
               System.exit(-1);
            }
         }
      }

      // initialized static globals in Request - ready it for handling
      if (! RequestManager.initialize (config, server_log, event_log)) {
         System.out.println ("RequestManager initialization failure" + 
                             " - see logs");

         shutdownDBConnectionPool();
         System.exit(-1);
      }

      if (BeanDirectory.isRunning()) {
         tpm = (ThreadPoolManager) 
               BeanDirectory.resolve("/services/ThreadPoolManager");
      } else {
         // startup the 
         max_pooled_threads = config.getInt ("LOCO_MAX_POOLED_THREADS");
         min_pooled_threads = config.getInt ("LOCO_MIN_POOLED_THREADS");

         tpm = new ThreadPoolManager (server_log, min_pooled_threads, 
                                      max_pooled_threads);
      }

      // initialize the CachedPageLoader used in evaluateTemplate 
      // below.
      template_root = 
      config.getString ("SYSTEM_SERVER_ROOT_PATH") +
      "/" + config.getString ("LOCO_TEMPLATE_ROOT_PATH");
      check_templates_freq = config.getInt ("LOCO_CHECK_NEW_TEMPLATE_FREQ");

      CachedPageLoader.initialize (template_root, check_templates_freq, 
                                   initial_cpl_bufsize);

      // Start up any HTTP server instances that we're configured to start
      String http_servers = config.getString("WEB_SERVER_INSTANCE_NAMES");

      if (!http_servers.equalsIgnoreCase("NONE")) {
         // Start up any/all HTTP servers
         startWebServers(http_servers);
      }

      // prepare server socket for accept
      int loco_port = config.getInt("LOCO_PORT");
      int loco_port_backlog = config.getInt ("LOCO_PORT_BACKLOG");

      try {
         ss = new ServerSocket (loco_port, loco_port_backlog);
      } catch (IOException ioe) {
         server_log.log(0, "Loco - ERROR: ServerSocket on port " + loco_port
                        + ": " + ioe.toString());
         shutdownAll();
         System.exit(-1);
      }

      server_log.log(0, "Loco - Now accepting connections...");

      current = Thread.currentThread();

      // main loop
      while (! stop_accepting) {
         request_id++;
         try {
            if (request_id % 50 == 7) {
               Runtime r = Runtime.getRuntime();
               server_log.log (0, "before gc (free, total): " + 
                               r.freeMemory() + ", " + r.totalMemory());
               System.gc();
               server_log.log (0, "after gc (free, total): " + 
                               r.freeMemory() + ", " + r.totalMemory());
            }

            cs = ss.accept();
            dispatch ();
         } catch (InterruptedIOException ie) {
            server_log.log(0, "Loco - INTERRUPT! shutdown initiated: " + 
                           ie.getMessage());
         } catch (IOException ioe) {
            server_log.log (1, "Loco - ERROR: accept failed: " + 
                            ioe.toString());
         }
      }

      server_log.log (0, "Loco - Shutdown issued, closing accept socket");
      try {
         ss.close();
      } catch (IOException ioe) {
         server_log.log(0, "Loco - ERROR: close Server Socket failure: " + 
                        ioe.toString()); 
      }
      server_log.log (0, "Loco - accept socket closed");

      shutdownAll();
   }


   /** 
    * used by main accept loop to launch new RequestManagers.  
    * will block and wait if too many RMs are active
    */ 
   private static void dispatch ()
   {
      // getPooledThread blocks until a thread becomes available
      PooledThread pt = tpm.getPooledThread ();

      server_log.log (2, "Loco - dispatch REQ(" + request_id + 
                      ") active: " + tpm.getActiveCount() + " out of " + 
                      tpm.getPoolSize());

      RequestManager rm = new RequestManager (cs, request_id);
      pt.loadRunnable (rm);
      pt.go ();
   }


   /**
    * log to the Loco server log
    * @param lev the log level
    * @param s a string to be logged
    */
   public static void log (int lev, String s)
   {
      server_log.log (lev, s);
   }

   /**
    * log to the Loco event log - simplified version
    * @param level - the log level
    * @param hd - a HandlerData object as passed into a handler
    * @param major - the major event region or code section
    * @param minor - the minor event code
    * @param data  - event specific data
    */
   public static void logEvent (int level, HandlerData hd, String major, 
                                String minor, String data)
   {
      StringBuffer sb = new StringBuffer (128);

      sb.append (hd.getRequestID());
      sb.append ("\t");
      sb.append (hd.session.getId());
      sb.append ("\t");
      sb.append (hd.user.getUserid());
      sb.append ("\t");
      sb.append (hd.user.getUsername());
      sb.append ("\t");
      sb.append (major);
      sb.append ("\t");
      sb.append (minor);
      sb.append ("\t");
      sb.append (data);
      event_log.log (level, sb.toString());
   }

   /**
    * log to the Loco event log
    * @param level - the log level
    * @param rid - the response id (used mostly for debugging)
    * @param sess - the current session
    * @param u    - the User
    * @param major - the major event region or code section
    * @param minor - the minor event code
    * @param data  - event specific data
    */
   public static void logEvent (int level, int rid, Session sess, User u, 
                                String major, String minor, String data)
   {
      StringBuffer sb = new StringBuffer (128);

      sb.append (rid);
      sb.append ("\t");
      sb.append (sess.getId());
      sb.append ("\t");
      sb.append (u.getUserid());
      sb.append ("\t");
      sb.append (u.getUsername());
      sb.append ("\t");
      sb.append (major);
      sb.append ("\t");
      sb.append (minor);
      sb.append ("\t");
      sb.append (data);
      event_log.log (level, sb.toString());
   }

   /**
    * Start up any/all web servers from the space-separated list of
    * instance names passed to this method.  They each start in their
    * own thread, load their own configuration files, etc.
    *
    * @param server_list a space-separated list of web server instance names
    *                    to start.  The instance names can be anything, but
    *                    there must be an aptly-named config file for each.
    */
   public static void startWebServers(String server_list)
   {
      // Grab each web server instance name using a StringTokenizer
      StringTokenizer tokenizer = new StringTokenizer(server_list);

      // loop through all of the names, starting up each web server
      while (tokenizer.hasMoreTokens()) {
         // Get a single instance name
         String instance_name = tokenizer.nextToken();

         // start this web server instance in its own thread
         HTTPServer http_server = new HTTPServer(instance_name);
         http_server.setDaemon(true);  // These die when Locomotive dies
         http_server.start();
      }
   }

   /**
    * Evalute a template.  Uses the default template root path as a base. 
    * The default template root is set via the primary Loco configuration file
    * &quot;loco.conf&quot;.  If the default setup is not sufficient, it is 
    * possible for a handler implementor to use the CachedPageLoader directly.
    * @param rel_path  the relative path to the template file 
    * @param env       the evaluation environment hash
    * @see org.locomotive.server.CachedPageLoader
    */
   public static String evaluateTemplate (String rel_path, Hashtable env)
   {
      // null, 1 are for logging
      CachedPageLoader cpl = new CachedPageLoader (server_log, 1);
      return (cpl.evalTemplate (rel_path, env));
   } 

   /**
    * gets the reference to the main server log of the Loco
    */
   public static Log getServerLog ()
   {
      return (server_log);
   }

   /**
    * Set the log for this loco.
    */
   public static void setServerLog(Log serverLog){
      Loco.server_log = serverLog;
   }

   /**
    * gets the reference to the main server log of the Loco
    */
   public static Log getEventLog ()
   {
      return (event_log);
   }

   /**
    * set the event log for this loco.
    */
   public static void setEventLog(Log eventLog){
      Loco.event_log = eventLog;
   }


   /** 
    * Redirects a given request to a new handler for completion.
    * This routine might be useful when the originally called handler
    * (first handler) determines that a second handler should be 
    * called to complete a request.  <P>
    * The handler_uri is parsed to create a new hd.url_tokens, 
    * which is then used to determine how the new request will be handled.  
    * It should appear as if the first handler was never called, 
    * with a few exceptions: <BR> 
    * Any changes to variables in hd made by the first handler 
    * will remain, and the value of 'hd.url' (called hd.handler_uri 
    * in versions beyond 3.6) is left unmodified by the redirect.
    * In future versions, it too may
    * be reset to the value passed in via the redirect call.  <P> If it
    * is important for the second handler to know something about the first, 
    * we recommend passing a flag or data via the cgi_envars or subs hashes 
    * in hd. <P>
    * 
    * @param handler_uri the hander_uri specifying any parameters and 
    *        the handler name to redirect to.  A null or empty string 
    *        will result in the DefaultHandler.  Typically of the forms: <P>
    *        &quot;/handler/pathtoken/other?querytoken1+querytoken2&quot; or <P>
    *        &quot;handler?querytoken1+querytoken2&quot;  or <P>
    *        &quot;handler&quot;  <P>
    *        with zero or more path tokens and or zero or more query tokens, 
    *        depending on how the url structure has been designed.
    * 
    * @param hd the HandlerData instance passed into the first Handler, or 
    *           a newly constructed one if necessary.
    */
   public static void redirect (String request_uri, HandlerData hd)
   {
      RequestManager rm = new RequestManager(request_uri, hd);

      Loco.log(4, "Redirect: " + request_uri);
      rm.handoff();
   }


   /**
    * Returns the number of active RequestManager Threads.  Allows you to 
    * get and idea of current load/concurrency.  
    */
   public static int getActiveThreadCount()
   {
      return (tpm.getActiveCount());
   }

   /**
    * Returns the number of active RequestManager Threads.  Allows you to 
    * get and idea of current load/concurrency.  
    */
   public static int getThreadPoolSize()
   {
      return (tpm.getPoolSize ());
   }

   /**
    * Returns the number of active RequestManager Threads.  Allows you to 
    * get and idea of current load/concurrency.  
    */
   public static int getMaxThreadPoolSize()
   {
      return (tpm.getMaxCount ());
   }

   /**
    * Set the thread pool manager for this loco.
    */
   public static void setThreadPoolManager(ThreadPoolManager tpm){
      Loco.tpm = tpm;
   }

   /**
    * returns the startup date of the Loco
    */
   public static java.util.Date getStartupDate ()
   {
      long mills = startup_date.getTime ();
      java.util.Date d = new java.util.Date ();
      d.setTime (mills);
      return (d);
   }

   /**
    * gets the current value of the request id.  The value is 
    * incremented just before pulling a new client off the queue. 
    */
   public static int getCurrentRequestID ()
   {
      return (request_id);
   }


   /*------------ DB Connection Pool access methods --------------*/


   /**
    * Gets a Connection from the Request Manager's DBConnection Pool.
    * This method might be called by various modules.
    * The connection should be returned after use using putBackDBConnection().
    * @return a connection from the Locomotive's internal database connection
    *         pool, or null if the connection pool has not been created yet.
    */
   public static Connection getDBConnection ()
   {
      if (dbcp != null)
         return (dbcp.getConnection());
      else
         return null;
   }

   /** 
    * Returns a Connection object back to the RequestManager
    * DBConnection Pool.  (used by Module cleanup threads) Putting
    * back connections that were not fetched using getDBConnection()
    * could damage the integrity of the DBConnectionPool and is not advised.  
    **/
   public static void putBackDBConnection (Connection c)
   {
      if (dbcp != null && c != null)
         dbcp.returnToPool (c);
   }


   /**
    * throws away a Connection that has been corrupted, for example,
    * by a SQLException, so that it will be deleted from the
    * DBConnection Pool.
    */
   public static void disposeDBConnection (Connection c)
   {
      if (dbcp != null && c != null)
         dbcp.wipeFromPool (c);
   }


   /**
    * used by the Server Handler for status
    */
   public static int getOpenDBConnectionCount ()
   {
      if (dbcp != null)
         return (dbcp.getPoolSize());
      else
         return (-1);
   }


   /** 
    * creates the main Locomotive Database connection pool.
    * if the DB_AUTOCONNECT is set to yes, then this occurs when the
    * Locomotive is started.  Otherwise it must be called by a handler.
    * @returns  true if the connpool was created.
    */
   public static boolean createDBConnectionPool(LocoConfig lc)
   {
      if (dbcp != null)
         return false;



      String db_conn_pool_classname = config.getString ("DB_CONN_POOL_CLASSNAME");

      String db_driver_classname = config.getString ("DB_DRIVER_CLASSNAME");
      String db_url = config.getString ("DB_CONNPOOL_URL");
      String db_username = config.getString ("DB_CONNPOOL_USERNAME");
      String db_password = config.getString ("DB_CONNPOOL_PASSWORD");
      int min_open = config.getInt ("DB_CONNPOOL_MIN_CONNECTIONS");
      int max_open = config.getInt ("DB_CONNPOOL_MAX_CONNECTIONS");


      // David Warnock change to support alternative DBConnectionPools
      // by default the test will be passed as the class name is set by default in
      // LocoConfig. In that case behaviour will be unchanged from earlier versions
      // Otherwise the class will be constructed from the class name (should be a fully 
      // qualified name in the config file including it's package). All the parameters will
      // then be set one by one as the default constructor is used.
      if (db_conn_pool_classname.equals("org.locomotive.server.DBConnectionPool")) {

         try {
            dbcp = new DBConnectionPool (db_driver_classname, db_url, 
                                         db_username, db_password,
                                         min_open, max_open, server_log);
            dbcp.initialize();
         } catch (InitializationException ie) {
            server_log.log (0, "ERROR: create DBConnectionPool(): " + 
                            StringUtil.exceptionStackTrace(ie));
            return (false);
         }
      } else {
          Class db_conn_pool_class = null;
          Object dbcp_o = null;
         try {
            db_conn_pool_class = Class.forName(db_conn_pool_classname);
         } catch (ClassNotFoundException e) {
            server_log.log (0, "ERROR: create DBConnectionPool(): " + 
                            StringUtil.exceptionStackTrace(e));
            return (false);
         }

         try {
            dbcp_o = db_conn_pool_class.newInstance();
         } catch (InstantiationException e) {
            server_log.log (0, "ERROR: create DBConnectionPool(): " + 
                            StringUtil.exceptionStackTrace(e));
            return (false);
         } catch (IllegalAccessException e) {
            server_log.log (0, "ERROR: create DBConnectionPool(): " + 
                            StringUtil.exceptionStackTrace(e));
            return (false);
         }
         dbcp = (DBConnectionPool)dbcp_o;

         dbcp.setDriverClassName(db_driver_classname);
         dbcp.setUrl(db_url);
         dbcp.setUserName(db_username);
         dbcp.setPassword(db_password);
         dbcp.setMinSize(min_open);
         dbcp.setMaxSize(max_open);
         dbcp.setLog(server_log);
         dbcp.setLogLevel( dbcp.getDefaultLogLevel() );
         try {
            dbcp.initialize();
         } catch (InitializationException ie) {
            server_log.log (0, "ERROR: create DBConnectionPool(): " + 
                            StringUtil.exceptionStackTrace(ie));
            return (false);
         }
      }
      // END change by David Warnock change to support alternative DBConnectionPools

      Loco.log(1, "DBConnection Pool Initialized");
      return true;
   } 



   /**
    * Wipes all the current database connections from the
    * DBConnectionPool and reloads it with new ones.  It uses the
    * JDBC driver class, connection URL, username, and password
    * specified in the loco.conf file.  You may want to use this
    * periodically if connections are getting locked thanks to bad
    * code, drivers, or database bugs.  
    * <P>This isn't currently as safe as it should be, as it wipes
    * the current connections without checking to see if it creates
    * new ones.  So, calling this could potentially result in a loco
    * hang, if new connections can't be created.  Watch out!  Note,
    * however, that this isn't as serious as it seems, because
    * inability to create a new connection is often a symptom of
    * problems that affect general database communication, and so the
    * inability to reload would probably only happen when the loco
    * isn't able to talk to the database at all.
    */
   protected static boolean reloadDBConnectionPool()
   {
      if (dbcp != null) {
         try {
            // I don't think we'll need this, but just to be safe... 
            synchronized (dbcp)
            {
               dbcp.closeCurrentPoolMembers();
               dbcp.initialize();
            }
            return (true);
         } catch (InitializationException ie) {
            server_log.log(0, "ERROR: could not re-initialize database " +
                           "connection pool: " + ie.toString());
         }
      }

      return (false);
   } 

   /**
    * Set the connection pool to use for this loco.
    * @param connectionPool The connection pool.
    */
   public static void setConnectionPool(DBConnectionPool connectionPool){
      dbcp = connectionPool;
   }

/*------- WRAPPED METHODS FROM REQUEST MANAGER - IN BOTH PLACES FOR NOW ----*/


   /**
    * toggle auto session handling <BR>
    * If true, which is the default, the RequestManager will automatically 
    * sense expired sessions.  The default timeout is 3600 seconds (one hour)
    * and is configured via the primary Loco configuration file.  Sessions 
    * that have not been active for greater than the timeout will be reset
    * to a new session, stored in the database, and the a new browser 
    * session cookie will set in the current response object.  If the 
    * previous session belonged to an authenticated user (aka, one who 
    * was logged in) they will be sent the default 'session expired' template,
    * the request will then end and a handler will not be invoked.   If the 
    * old session contained an anonymous user (or an &quot; 
    * unauthenticated user&quot;) no message is sent, and the request 
    * is passed into the apprpriate handler with the new anonymous session.  
    * processing.  If a handler needs to handle expired sessions on its own, 
    * setAutoSessionHandling should be passed &quot;false&quot;.  Note 
    * that this cannot be set on a per-handler basis.  
    * @param on new value of auto session handling flag
    */
   public static void setAutoSessionHandling (boolean on)
   {
      RequestManager.auto_session_handling = on;
   }


   /**
    * returns the current value of the auto_session_handling flag.
    * @see setAutoSessionHandling
    */ 
   public static boolean getAutoSessionHandling ()
   {
      return (RequestManager.auto_session_handling);
   }


   /** 
    * Marks the Handler class specified by the string for reloading, which
    * means that upon the next request for that handler, the handler's
    * properties file will be reloaded and the handler's init() method will
    * be called. <BR> 
    * If the string is left blank ("") or is null, then all Handler 
    * classes are reloaded- see markAllHandlersForReloading(). <BR>
    * This only reinstantiates the class and re-reads the properties files.
    * It doesn't go back and reload the hrt.conf file, nor does it actually
    * reload the class into the VM.
    * @param token the token that refers to the Handler class in the
    *        hrt.conf file
    */
   protected static void reloadHandler(String handler)
   {
      RequestManager.reloadHandler(handler);

   }

   /** 
    * Reloads the hrt.conf file, then reinitializes all the handlers and
    * reloads each of their properties files.
    * It doesn't actually reload Classes from their class files, though.
    * The reload will occur on the next request to the Locomotive.
    * @return true if we reloaded successfully, false if otherwise
    */
   protected static boolean reloadHandlerRoutingTable()
   {
      return (RequestManager.reloadHandlerRoutingTable());
   }


   /**
    * This will return the Properties object associated with a Handler,
    * if there is any.  The Properties object is created from a file
    * specified in the hrt.conf file, and is loaded at startup and
    * whenever the Handler is reloaded.  If no Properties file was
    * specified in the hrt.conf file, then this returns null.
    * <P>Each Properties object associated with a Handler will have
    * at least one variable in it- <PRE>
    *   PROPERTIES_PATH </PRE>
    * which is the absolute path to the file the Properties were loaded
    * from.  This will allow Handlers to manipulate and re-store their
    * Properties.
    * @see org.locomotive.loco.handler.Handler
    * @see org.locomotive.loco.handler.HandlerData#handler_config
    */
   public static Properties getHandlerProperties(Handler h)
   {
      Class c = h.getClass();
      return (RequestManager.getHandlerProperties(c));
   }


   /** Marks the Servlet class specified by the string for reloading, which
    * means that upon the next request for that Servlet, the Servlet's
    * properties file will be reloaded and the Servlet's init() method will
    * be called. <BR> 
    * If the string is left blank ("") or is null, then all Servlet 
    * classes are reloaded- see markAllServletsForReloading(). <BR>
    * This only reinstantiates the class and re-reads the properties files.
    * It doesn't go back and reload the hrt.conf file, nor does it actually
    * reload the class into the VM.
    * @param token the token that refers to the Servlet class in the
    *        hrt.conf file
    */
   public static void reloadServlet(String Servlet)
   {
      RequestManager.reloadServlet(Servlet);

   }

   /** 
    * Reloads the hrt.conf file, then reinitializes all the Servlets and
    * reloads each of their properties files.
    * It doesn't actually reload Classes from their class files, though.
    * The reload will occur on the next request to the Locomotive.
    * @return true if we reloaded successfully, false if otherwise
    */
   public static boolean reloadServletRoutingTable()
   {
      return (RequestManager.reloadServletRoutingTable());
   }


   /**
    * for shutting down the Loco.  After handling the next inbound connection, 
    * the Loco will stop accepting new connections, do various cleanups, 
    * then exit
    */
   public static void setStopAccepting()
   {
      stop_accepting = true;
      current.interrupt();
      server_log.log(0, "setStopAccepting: Interrupt issued");
   }

   /**
    * called internally for cleanup 
    */
   private static void shutdownAll()
   {
      // wait for all active rms to end (thus returning their 
      // db connections to the internal dbcp
      int loopcount = 0;
      int active_count = 0;

      while (((active_count = tpm.getActiveCount ()) > 0) && loopcount < 15) {
         server_log.log (0, "Loco - shutdown waiting: still active: " + 
                         active_count);
         loopcount++;
         try {
            Thread.sleep (1000);
         } catch (InterruptedException ie) {
            
         }
      }

      // oh dear, their stuck.  
      if (active_count > 0) {
         server_log.log (0, "Loco - WARNING: wedged rm - forging onward");
      }

      server_log.log (0, "Loco - calling RM.shutdown()");

      if (! RequestManager.shutdown ()) {
         // no reason for this to happen 
         server_log.log(0, "ERROR: RequestManager: could not shut down " +
                        "successfully");
      }

      server_log.log (0, "Loco - closing Database Connection Pool");

      shutdownDBConnectionPool();

      server_log.log (0, "Loco - RM Shutdown Complete. Exiting..a.a...argh!");
      System.exit (0);
   }


   private static void shutdownDBConnectionPool()
   {
      // get rid of the database connection pool
      if (dbcp != null) {
         dbcp.closeCurrentPoolMembers();

         // let's make sure Cloudscape shuts down too...
         switch (config.DB_TYPE) {
         case SQL.DB_TYPE_CLOUDSCAPE:
            try {
               DriverManager.getConnection(
                                          "jdbc:cloudscape:;shutdown=true;");
            } catch (SQLException e) {
               server_log.log(0, "Shutting down cloudscape: " + 
                              e.getMessage());
            }
         }
      }
   }


   /**
    * returns the LocoConfig of the server
    */
   public static LocoConfig getConfig ()
   {
      return (config);
   }

   /**
    * Get the global DBCache for Loco.
    */
   public static DBCache getGlobalDBCache()
   {
      if (dbCache == null) {
         // Ensure connection pool is created.
         if (dbcp == null)
            createDBConnectionPool(getConfig());

         dbCache = new DBCache(getConfig(), dbcp, server_log);
         dbCache.startService();
      }
      return dbCache;
   }

   /**
    * returns the Version Number of the server
    */
   public static String getVersionString ()
   {
      return ("4.2.0");
   }


   /**
    * Displays and error page with Standard insertions, plus a 
    * the error string for the class.  The error string is set in
    * the constructor.  Evaluates the template
    * LOCO_TEMPLATE_ROOT_PATH/error/generic_error.tmpl.
    * LOCO_TEMPLATE_ROOT_PATH is the template directory, as specified in
    * the Loco configuration file.
    * If this template is missing, then it will send a generic html file
    * including the message.
    * @param error_msg The mesage to display.  If left blank or null,
    *        then the following will message will be displayed: <BR>
    *        <CODE>Your request could not be processed due to an
    *		     internal handler error.</CODE>
    * @param hd the handlerData object whose request we should send the
    *           response through.
    */
   public static void sendErrorMsg(String error_msg, HandlerData hd)
   {
      if (error_msg == null || error_msg.length() < 1) {
         error_msg = "Your request could not be processed " +
                     "due to an internal handler error. ";
      }

      // clean up whatever else has been placed in the response thus far.
      hd.resp.setContentType("text/html");
      hd.resp.clearHeaders();
      hd.resp.clearBody();

      if (hd.subs == null) hd.subs = new java.util.Hashtable();

      hd.subs.put("ERROR_MSG", error_msg);
      StandardInsertions.insertDateTime(hd);
      StandardInsertions.insertRequestID(hd);

      // identify the loco returning to error
      StandardInsertions.insertSystemTags(hd);  

      String response = 
      Loco.evaluateTemplate("error/generic_error.tmpl", hd.subs);

      // the template is missing
      if (response == null || response.length() < 1) {
         response = 
         "<HTML>\n<HEAD><TITLE>Error</TITLE></HEAD>\n" + 
         "<BODY BGCOLOR=\"#FFFFFF\">\n\n" +
         "<H1>There has been an error</H1>\n" + 
         error_msg + "<BR>\n" +
         "Please notify the server administrator " +
         "and send the details of this message.<BR>\n" + 
         "</BODY>\n</HTML>";
      }

      hd.resp.addString (response);
      hd.resp.flush (hd.server_log);
   } 

   /**
    * Get the server root.
    * @return Returns the server root.
    */
   public String getServerRoot()
   {
      return serverRoot;
   }

   /**
    * Set the server root.
    * @param serverRoot The server root for this locomotive.
    */
   public void setServerRoot(String serverRoot)
   {
      this.serverRoot = serverRoot;
   }

}
