/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1999 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import javax.servlet.ServletInputStream;
import javax.servlet.http.*;

import org.locomotive.server.*;
import org.locomotive.loco.handler.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.util.*;

import org.locomotive.loco.store.*;
import org.locomotive.loco.store.bstamp.*;

import java.io.*;
import java.net.*;
import java.util.*;
import java.sql.*;


/**
 * RequestManagers do several things:<BR>
 * 1) standard environmental setup for modules<BR>
 * 2) invocation of a specific service<BR>
 * 3) post module execution cleanup <BR>
 * 4) management of server data <BR>
 * <P>
 * RequestManagers are invoked as separate threads (one thread to a request)
 * after being initialized (statically) by the Loco at run-time
 */
class RequestManager implements Runnable
{
    private static LocoConfig config;
    private static Log server_log;
    private static Log event_log;
    protected static boolean auto_session_handling = true;

    protected static ServletRoutingTable srt;
    protected static HandlerRoutingTable hrt;
    protected static Vector class_path;
    protected static boolean auto_reload_classes = false;

    // these are arrays of ints holding the ips  and subnets which are
    // allowed to shut down the loco
    private static int ip_accept[] = null;
    private static int ip_subnet[] = null;
    private static int ip_calc[]   = null;  // ip_accept & ip_subnet

    private static final int SHUTDOWN_MAGIC_COOKIE = 0x09080706;
    private static final int PROTOCOL_VERSION_CGI = 2;
    private static final int PROTOCOL_VERSION_ISAPI = 3;
    private static final int PROTOCOL_VERSION_NSAPI = 4;
    private static final int PROTOCOL_VERSION_EWS = 5;

    private static int SERVICE_DEFAULT;

    // gets set in parseURLTokens
    private static final int HANDLER_REQUEST = 11;
    private static final int SERVLET_REQUEST = 12;

    private int request_type;  // holds one of the above

    // these are all that come in on a new RequestManager (instantiated by
    // the Loco at hit receive time)
    private int request_id;
    private Socket sock;
    private int reply_protocol_type;

    private InputStream is;
    private OutputStream os;
    private Connection conn = null;

    private Hashtable cgi_envars_hash;
    private Hashtable form_hash;
    private Hashtable subs;
    private String url_tokens[];
    private String request_uri;

    private Session session = null;
    private Response resp;
    private long bstamp_id;
    private Hashtable attributes;

    //dw
    // This is the UserStore that provides all the persistance
    // call method createUserStore to fill this
    private BStampStore bStampStore = null;

    // This is sent in to us by the Embedded Web Server
    private String http_path_info = null;
    private String http_query_string = null;
    private Vector http_environment_lines = null;
    private BufferedReader http_buffered_reader = null;
    public static Hashtable envar_name_mappings = null;

    static
    {
	// The envar_name_mappings static Hashtable is used by the
	// Embedded Web Server to translate the HTTP environment
	// variable names (HTTP header line names, or whatever you
	// want to call them) into names that web servers commonly
	// translate them into..  names that the Locomotive uses.
	envar_name_mappings = new Hashtable();

	// Populate the Hashtable with all of the name mappings
	envar_name_mappings.put("ACCEPT", "HTTP_ACCEPT");
	envar_name_mappings.put("ACCEPT-CHARSET", "HTTP_ACCEPT_CHARSET");
        envar_name_mappings.put("ACCEPT-ENCODING", "HTTP_ACCEPT_ENCODING");
	envar_name_mappings.put("ACCEPT-LANGUAGE", "HTTP_ACCEPT_LANGUAGE");
	envar_name_mappings.put("AUTHORIZATION", "HTTP_AUTHORIZATION");
	envar_name_mappings.put("AUTH-USER", "REMOTE_USER");
	envar_name_mappings.put("CONNECTION", "HTTP_CONNECTION");
	envar_name_mappings.put("COOKIE", "HTTP_COOKIE");
	envar_name_mappings.put("IF-MODIFIED-SINCE", "HTTP_IF_MODIFIED_SINCE");
	envar_name_mappings.put("METHOD", "REQUEST_METHOD");
	envar_name_mappings.put("REFERER", "HTTP_REFERER");
	envar_name_mappings.put("USER-AGENT", "HTTP_USER_AGENT");
	envar_name_mappings.put("CONTENT-TYPE", "CONTENT_TYPE");
	envar_name_mappings.put("CONTENT-LENGTH", "CONTENT_LENGTH");
    }

    /**
     * constructor used by Loco
     */
    RequestManager (Socket s, int rq_id)
    {
	sock = s;
	request_id = rq_id;
	request_type = SERVICE_DEFAULT;

	createBStampStore();
    }


    /**
     * Constructor used by the redirect.  Will reparse the request and
     * get everything ready for handoff() to be called successfully.
     */
    RequestManager(String request_uri, HandlerData hd)
    {
	this(hd.sock, hd.getRequestID());

	this.request_uri = request_uri;
	this.resp = hd.resp;
	this.is = hd.is;
	this.os = hd.os;
	this.conn = hd.conn;
	this.cgi_envars_hash = hd.cgi_envars_hash;
	this.form_hash = hd.form_hash;
	this.subs = hd.subs;
	this.session = hd.session;
	this.bstamp_id = hd.bstamp_id;
        this.attributes = hd.getAttributes();

	// now, a couple of hacks to get this to work right
	cgi_envars_hash.remove("NSAPI");

	// Set the PATH_INFO and QUERY_STRING vars from the request_uri
	int index = request_uri.indexOf("?");
	if (index == -1)
	{
	    cgi_envars_hash.put("PATH_INFO", request_uri);
	    cgi_envars_hash.remove("QUERY_STRING");
	}
	else
	{
	    cgi_envars_hash.put("PATH_INFO", request_uri.substring(0, index));
	    cgi_envars_hash.put("QUERY_STRING",request_uri.substring(index+1));
	}

	parse_url_tokens ();
    }

    /**
     * Constructor used by the Embedded Web Server
     */
    RequestManager(Socket s, int rq_id, String path_info, String query_string,
		   Vector lines, BufferedReader reader)
    {
	this(s, rq_id);

	// Save the path_info and query_string that the Embedded Web Server
	// gave us.
	http_path_info = path_info;
	http_query_string = query_string;

	// Save the vector of http environment lines that were sent to us
	// by the Embedded Web Server
	http_environment_lines = lines;

        // Save a reference to the web server's buffered reader, in case
        // we need it (only for HTTP POSTs right now).
        http_buffered_reader = reader;

	// In our case, we need to instantiate a Hashtable for
	// the cgi_envars_hash right here
	cgi_envars_hash = new Hashtable(30);

	// Set the PATH_INFO and QUERY_STRING vars
	if (null == query_string)
	{
	    cgi_envars_hash.put("PATH_INFO", path_info);
	    cgi_envars_hash.remove("QUERY_STRING");
	}
	else
	{
	    cgi_envars_hash.put("PATH_INFO", path_info);
	    cgi_envars_hash.put("QUERY_STRING", query_string);
	}

	createBStampStore();
    }


    //dw
    /**
     * get a new BStampStore from the factory if one has not been found already
     * this needs to be called before any db stuff (obviously)
     */
    private void createBStampStore()
    {
	if ( bStampStore == null )
	{
	    bStampStore = StoreFactory.createBStampStore();
	}
    }

    /**
     * Invoked by the Loco for each request.  Do not use publicly.
     */
    public void run ()
    {
	// if setup is successful, hand it off
	if (doSetup())
	{
	    handoff();
	}

	doCleanup();

	// indicates completion of this thread to the Loco for
	// active thread limit management.  Important!
    }

    /*--------------- SETUP ----------------------*/


    /**
     * do all pre-handoff setup:
     * cookies, sessions, user objects, get a connection, open streams,
     * parse encoded form data, whatever.  Return false if we are to skip
     * the handoff.
     */
    private boolean doSetup ()
    {
	if (! openStreams())
        {
	    return(false);
	}

	log (7, "reading envars...");
	if (readCGIEnvars() == false)
        {
	    log (0, "ERROR: Failed to get cgi envars properly!");
	    return(false);
	}

	// process path info and querystring into an array
	parse_url_tokens();


	// get connection, if we've opened the DBConnPool.
	// this will be null if we haven't....
	conn = Loco.getDBConnection();

	// now, let's create the response object we'll use for the setup.
	// we'll create a LocoServletResponse, which we can later
	// cast to a normal request if we need to
        if (null == http_path_info)
            // It's a regular Locomotive request, not from Embedded Web Server
            resp = new LocoServletResponse(os, request_id,
                                           reply_protocol_type);
        else
            // It's an Embedded Web Server request
            resp = new HTTPResponse(os, request_id);

	// throw out a the bstamp cookie, if we're supposed to
	String do_bstamp = config.getString("BSTAMP_COOKIE");
	if (do_bstamp.equalsIgnoreCase("YES"))
        {
	    doBstampCookie();
	}

	// Take care of the LOCO_INST_ID cookie to facilitate virtual sessions
	doLocoInstIDCookie();

	// handle sessions via cookies
	String session_cookie_string =
	    getCookieValue("LOCOSID_" + config.getString("SYSTEM_TAG"));
	try
        {
	    if (conn != null)
            {
		// if no cookie, then add one, and create a session
		if (session_cookie_string == null)
                {
		    session = new Session(
		        config.getInt("LOCO_SESSION_EXPIRE_TIME"), conn);
		    session.updateBrowserCookie(resp);
		}
                else
                {
		    // here we re-create a session from the db.  Note that
		    // handling of expired sessions should occur in Handlers,
		    // allowing each client to deal in their own way
		    session = new Session(conn, session_cookie_string);
		}
	    }
            else
            {
		// We have no Database connection- just create an anon session
		session = new Session ();

		// set a test cookie.  This is added back in
		// after testing reveals that MSIE4.0 can't swallow
		// the bstamp cookie...
		Cookie c = new Cookie ("TC", "return_me");
		c.setDomain (config.getString("LOCO_COOKIE_DOMAIN_MAIN"));
		c.setPath("/");
		resp.addCookie(c);
	    }
	}
        catch (FailedSessionException fse)
        {
	    // We've got a problem- let's try creating a new session
	    // and updating the session cookie, then we'll let them
	    // know there's been a problem
	    log (0, "ERROR: " + StringUtil.exceptionStackTrace(fse) +
		 "\n wiping cookie...");
	    String err = "session failure";
	    Session s = new Session();
	    try
	    {
		s.expire(null);
	    }
            catch (FailedSessionException e)
	    {
		log(0, "ERROR: could not expire session:" + fse);
	    }

	    // we flush the cookie and the content type before sending the
	    // error message, because the error message clears all headers
	    s.updateBrowserCookie(resp);
	    resp.setContentType("text/html");
	    resp.flush();
	    sendErrorMsg (err);
	    return(false);
	}


	// here we deal with expired sessions.  read below on
	// setAutoSessionHandling() for details of behaviour.
	if (auto_session_handling && conn != null)
        {
	    if (! handleSessionExpiration())
	    {
		return(false);
	    }
	}

	subs = new Hashtable();

	return(true);
    }


    // create bufferedInput and BufferedOutput streams from socket
    private boolean openStreams()
    {
	try
        {
	    is = new BufferedInputStream(sock.getInputStream());
	    os = new BufferedOutputStream(sock.getOutputStream());
	}
        catch (IOException ioe)
        {
	    log(0, "ERROR: RM(" + request_id + ") getStreams(): " +
		 StringUtil.exceptionStackTrace(ioe));
	    return(false);
	}

	return(true);
    }




    /** read from the webserver or cgi-like inflow socket
     * here's the protocol:
     * <pstring length><envar name>=<envar value>
     * so - first we read in the length, then we read in that length of bytes
     * to get the envar name and value, then we can split it into to
     * individual strings at the '='
     * For the NSAPI, we bring in the names and values individually.
     */
    private boolean readCGIEnvars()
    {
	DataInputStream dis = new DataInputStream (is);

	if (null == cgi_envars_hash)
	{
	    cgi_envars_hash = new Hashtable(30);
	}

	int dis_vers = 0;
	int inbound_data_format = 0; /* 0: for name=value as single pstring,
					used by CGI, Apache, and ISAPI
					1: name and value comes in as
					seperate pstrings - used by NSAPI,
				        2: straight from browser - used
				        by the Embedded Web Server. */

	// get the version number, unless the Embedded Web Server created us
	if (null == http_path_info)
	{
	    try
	    {
		dis_vers = dis.readInt();
	    }
	    catch (IOException ioe)
	    {
		log(0, "ERROR: reading version: " +
		    StringUtil.exceptionStackTrace(ioe));
		return(false);
	    }
	}
	else
	{
	    // Set the protocol version to EWS for Embedded Web Server requests
	    dis_vers = PROTOCOL_VERSION_EWS;
	}

	/* what should really happen here is that we should split
	   both input and output processing into another class
	   called "Tunnel.java" which handles the various protocols
	   to and from the web server tunnel lib or cgi */
	switch (dis_vers)
        {
	    case (PROTOCOL_VERSION_CGI):
	        reply_protocol_type = Response.REPLY_TYPE_CGI;
	        break;
	    case (PROTOCOL_VERSION_NSAPI):
	        inbound_data_format = 1;
	        reply_protocol_type = Response.REPLY_TYPE_NSAPI;
	        break;
	    case (PROTOCOL_VERSION_ISAPI):
	        /* for now, NSAPI and ISAPI use same reply protocol.
	           See Response.java */
	        reply_protocol_type = Response.REPLY_TYPE_NSAPI;
		break;
            case (PROTOCOL_VERSION_EWS):
		/* it's an Embedded Web Server Request */
		inbound_data_format = 2;
		reply_protocol_type = Response.REPLY_TYPE_CGI;
		break;
	    case (SHUTDOWN_MAGIC_COOKIE):
		attemptShutdown();
		return(false);
	    default:
		log (0, "ERROR: bogus protocol version: " + dis_vers);
	        return(false);
	} /* I love case statements */

	/* We now support three data input formats from the web server.
	   One is the original type, used by the CGI and the ISAPI.
	   For these, the name values pairs come in as a single p-string
	   seprated by the equals sign.  This was done to make the cgi
	   more efficient, since the standard format for environmental
	   variables, where the name values pairs originate, is a single
	   string concatenated with the = sign.  In the second type, the
	   NSAPI, it's a bit easier to pass them in one after the other
	   - both the java code in the locomotive and the C of the NSAPI
	   are simpler.  The third type is coming from the Embedded Web
	   Server, and basically we're reading lines just like they come
           from the web browser to the web server.  */
	try
        {
	    if (inbound_data_format == 0) /* cgi, isapi */
            {
		byte env_buf[];
		int env_length = 0;
		byte value_of_equals = (byte) '=';
		int index = 0;

		/* how to kill a locomotive:  send it an environmental
		   var that is 4 gigs in size.  */
		while ((env_length = dis.readInt ()) > 0)
                {
		    env_buf = new byte[env_length];

		    dis.readFully (env_buf, 0, env_buf.length);

		    // use first '=' to separate name from value (conventional)
		    // scan byte buffer, find the index
		    index = 0;
		    for (int i = 0; i < env_buf.length; i++)
                    {
			if (env_buf[i] == value_of_equals)
                        {
			    index = i;
			    break;
			}
		    }

		    // we ignore name/values without a separator
		    if (index > 0)
                    {
			if (index < (env_buf.length -1))
                        {
			    cgi_envars_hash.put(new String (env_buf, 0, index),
						new String (env_buf, index + 1,
						env_buf.length - 1 - index));
			}
                        else
                        {
			    cgi_envars_hash.put(
                                new String(env_buf, 0, index), "");
			}
		    }
		}
	    }
            else if (inbound_data_format == 1) /* NSAPI
						* names and values seperate */
            {
		byte buf[];
		int pstring_length = 0;
		String name;

		/* get length of name.  When a name length = 0,
		   the env list is finished.  A value can have a zero
		   length, as defined by the protocol, and an empty
		   string will be placed into the hashtable. */
		while ((pstring_length = dis.readInt ()) > 0)
                {
		    buf = new byte[pstring_length];
		    dis.readFully (buf, 0, pstring_length);
		    name = new String (buf);

		    // now read value
		    pstring_length = dis.readInt();
		    buf = new byte[pstring_length];
		    dis.readFully (buf, 0, pstring_length);
		    cgi_envars_hash.put(name,
					(pstring_length > 0) ?
					new String (buf) : "");
		}
	    }
	    else  /* Embedded Web Server */
	    {
		String env_line = null;
		String var_name = null;
		String var_value = null;

		int index = 0;

		// Loop through all of the http_environment_lines
		for (Enumeration e = http_environment_lines.elements();
		     e.hasMoreElements();)
		{
		    env_line = (String) e.nextElement();

		    // Use the first ':' to separate the name from the value
		    index = env_line.indexOf((int) ':');

		    // Ignore lines with no colon (for now)
		    if (-1 != index)
                    {
			// Separate the names and values into their own Strings
			var_name = env_line.substring(0,index);
			var_value = env_line.substring(index + 2,
						       env_line.length());

			// Translate the environment variable name
			var_name = translateEnvarName(var_name);

			//System.out.println("RM: adding env line: " +
			//		   var_name + "," + var_value);

			cgi_envars_hash.put(var_name, var_value);
		    }
		    else
                    {
			System.out.println("RM: Got bogus line: " + env_line);
		    }
		}

                // If we're doing a POST, parse our form variables here..
                if (cgi_envars_hash.get("REQUEST_METHOD").equals("POST"))
                {
                    // Our form variables are waiting on the next line of
                    // input from the client socket..  It's the first line
                    // of the HTTP request body (for an HTTP POST only)..
                    String content_type = (String) cgi_envars_hash.get(
                                          "CONTENT_TYPE");

                    // should we just use ours?
                    if (content_type.indexOf("multipart") != -1)
                    {
                        try
                        {
                            FormParser f = new FormParser(
                                               new EWSRequestInputStream(
                                                   http_buffered_reader), 
                                               cgi_envars_hash, 
                                               Loco.getConfig(),
                                               Loco.getServerLog());
                            form_hash = f.parseForm();
                        }
                        catch (Exception e)
                        {
                            form_hash = new Hashtable();
                        }
                    }
                    else
                    {
                        String content_length = (String) cgi_envars_hash.get(
                                                "CONTENT_LENGTH");
                        int len = Integer.parseInt(content_length);

                        form_hash = 
                            HttpUtils.parsePostData(len, (ServletInputStream)
                                                    new EWSRequestInputStream(
                                                    http_buffered_reader));
                    }
                }
	    }

	    // We need to make sure that "USER-AGENT" is set to the same
	    // thing that "HTTP_USER_AGENT" is set to (for better Servlet
	    // support, aparently).
	    String user_agent =
		(String) cgi_envars_hash.get("HTTP_USER_AGENT");
	    if (null != user_agent)
		cgi_envars_hash.put("USER-AGENT", user_agent);
	}
	catch (EOFException eofe)
        {
	    log(0, "ERROR: reading envHash: " +
		StringUtil.exceptionStackTrace(eofe));
	    return(false);
	}
        catch (IOException ioe)
        {
	    log(0, "ERROR: reading envHash: " +
		StringUtil.exceptionStackTrace(ioe));
	    return(false);
	}

	return(true);
    }

    /**
     * This method is called only for requests that come from the
     * Embedded Web Server.  This method uses the static Hashtable
     * called envar_name_mappings to translate an HTTP header variable
     * name into environment variable names that are commonly used
     * by web servers (and by the Locomotive).  If no translation
     * can be found in the hashtable, then the passed-in String is
     * returned unchanged.
     *                                                              <BR><BR>
     * @param envar_name The HTTP environment variable name that
     *        this method should translate.
     * @return the translated environment variable name, or the
     *         unchanged String if no translation could be found.
     */
    public static String translateEnvarName(String envar_name)
    {
	// Get a translated name from the static Hashtable mapping
	String translated_name =
	    (String) RequestManager.envar_name_mappings.get(
	                            envar_name.toUpperCase());

	// If we got a null, then we can't translate the name, return the
	// name unchanged.
	if (null == translated_name) return envar_name;

	// Otherwise, we got a translated name -- return it.
	return translated_name;
    }

    // parse the url tokens into an array, url_tokens.  Also
    // fill out url_string, which represents which url the browser requested
    private void parse_url_tokens ()
    {
	Vector url_tokens_vector = new Vector (10);

	/* we do things differently with the NSAPI than we do with
	   CGI.  CGI makes it easy, NSAPI is a bit more manual. */
	String nsapi = (String) cgi_envars_hash.get("NSAPI");
	if ((nsapi != null) && (nsapi.equals("ON")))
        {
	    removeNSAPIToken();
	}
	else  // what about the embedded web server?
	{
//	    removeCGITunnelToken();
	}

	if ((nsapi != null) && (nsapi.equals("ON")))
        {
	    removeNSAPIToken();
	}

	String path_info = null;
	String query_string = null;

	if ((path_info = (String) cgi_envars_hash.get("PATH_INFO")) != null)
        {
	    StringTokenizer st = new StringTokenizer(path_info, "/");

	    // first, let's look for the service token, strip it out,
	    // and set the service type
	    if (st.hasMoreTokens())
            {
		String tok = st.nextToken();
		if (tok.equals( config.getString("SERVLET_URI_TOKEN_PREFIX")))
                {
		    request_type = SERVLET_REQUEST;
		}
		else if (tok.equals(
			 config.getString("HANDLER_URI_TOKEN_PREFIX")))
		{
		    request_type = HANDLER_REQUEST;
		}
		else
		{
		    // a default request
		    url_tokens_vector.addElement(tok);
		}
	    }

	    while (st.hasMoreTokens ())
	    {
		url_tokens_vector.addElement(st.nextToken());
	    }
	}

	if ((query_string =
	     (String) cgi_envars_hash.get("QUERY_STRING")) != null)
	{
	    StringTokenizer st = new StringTokenizer(query_string, "+");
	    while (st.hasMoreTokens())
	    {
		url_tokens_vector.addElement(st.nextToken());
	    }
	}

	url_tokens = new String[url_tokens_vector.size()];
	url_tokens_vector.copyInto(url_tokens);

	/* do the handler uri.  Cover the case where everything is
	   sent to locomotive - there might not be a path info or
	   query string */
	request_uri = ((path_info != null) ? path_info : "") +
	    ((query_string != null && !query_string.equals("")) ?
	     "?" + query_string : "");

	cgi_envars_hash.put("REQUEST_URI", request_uri);

	/* log for now.  */

	// log the now captured CGI envars
	Enumeration e = cgi_envars_hash.keys();
	String key;

	while (e.hasMoreElements())
        {
	    key = (String) e.nextElement();
	    if (key.equals ("HTTP_USER_AGENT"))
	    {
		log(4, key + ": " + (String) cgi_envars_hash.get(key));
	    }
	    else
	    {
		log(8, key + ": " + (String) cgi_envars_hash.get(key));
	    }
	}

	StringBuffer sb = new StringBuffer();
	sb.append("url tokens: ");
	for (int i = 0; i < url_tokens.length; i++)
        {
	    sb.append(" ");
	    sb.append(url_tokens[i]);
	}
	log(5, sb.toString());

	/* returns void */
    }


    private void removeNSAPIToken()
    {
	String uri = (String) cgi_envars_hash.get("SCRIPT_NAME");
	if (uri == null)
        {
	    log(0, "ERROR: got NSAPI, but no SCRIPT_NAME!");
	    return;
	}

	/* The NSAPI provides a URI value, but not a path info,
	   since a token to find the tunnel rather than traversing
	   down a path until a executable cgi is found.  If
	   no token is used, we should treat the entire URI as
	   the PATH_INFO, so it will be parsed below into
	   url_tokens */
	String locolink_tag =
	    Loco.getConfig().getString("NSAPI_TUNNEL_URI_TOKEN_PREFIX");
	int index = uri.lastIndexOf(locolink_tag);

	if (index == -1)
        {
	    /* this means that the NSAPI at the webserver has
	       been configured to send essentially everything to
	       the Locomotive.  This can be done with a
	       global 'force-type' in the obj.conf file */
	    cgi_envars_hash.put("PATH_INFO", uri);
	}
	else
	{
	    index = index + locolink_tag.length();

	    while ((index < uri.length()) && uri.charAt(index) != '/')
            {
		index++;
	    }

	    if (index < uri.length())
            {
		cgi_envars_hash.put("PATH_INFO", uri.substring (index));
	    }
	}
    }


    /**
     * sets a persistent cookie in the response if there isn't one,
     * otherwise it plucks the value out
     */
    private void doBstampCookie ()
    {
	String bstamp_cookie_string =
	    getCookieValue("BSTAMP_" + config.getString("SYSTEM_TAG"));

	// set new cookie and exit
	if (bstamp_cookie_string == null && conn != null)
        {
	    try
            {
		bstamp_id = getNextBrowserStampID(conn);
		log(9, "new bstamp id: " + bstamp_id);
	    }
	    catch (SQLException se)
	    {
		log(0, "ERROR: getNextId(): " +
		     StringUtil.exceptionStackTrace(se));
		bstamp_id = 0;
	    }

	    // stamp their brower - also supplies cookie functionality
	    // test for Cookie.supported
	    Cookie c = new Cookie("BSTAMP_" + config.getString("SYSTEM_TAG"),
				  String.valueOf (bstamp_id));
	    c.setDomain(config.getString("LOCO_COOKIE_DOMAIN_MAIN"));
	    c.setPath("/");
	    // four years from now, in seconds
	    c.setMaxAge(4 * 365 * 24 * 3600);
	    resp.addCookie(c);
	}
        else
        {
	    // try and get a value out of the current cookie
	    try
            {
		if (bstamp_cookie_string != null)
		    bstamp_id = Long.parseLong(bstamp_cookie_string);
	    }
	    catch (NumberFormatException nfe)
	    {
		bstamp_id = 0;
		log(0, "ERROR: bstamp parse failed.  Setting to zero");
	    }
	}
    }


    // the database calls to get the bstamp
    private long getNextBrowserStampID(Connection conn) throws SQLException
    {
	return bStampStore.getNextBrowserStampID(conn);
    }


    // Send a LOCO_INST_ID cookie with the value configured in loco.conf
    private void doLocoInstIDCookie ()
    {
	// I'm assuming config has been properly initialized.
	String lid_out = String.valueOf(config.getInt("LOCO_INST_ID"));
	String sys_tag = config.getString("SYSTEM_TAG");

	// Log the incoming and outgoing values of the cookie.
	String lid_in = getCookieValue("LOCO_INST_ID_" + sys_tag);
	log(8, "RequestManager: My LOCO_INST_ID is: " + lid_out +
	    "; Instance requested was: " + lid_in);

	// Issue a new cookie if there is a change
	if ((lid_in == null) || (! lid_in.equals(lid_out)))
	{
	    Cookie c = new Cookie("LOCO_INST_ID_" + sys_tag, lid_out);
	    c.setDomain(config.getString("LOCO_COOKIE_DOMAIN_MAIN"));
	    c.setPath("/");
	    resp.addCookie(c);
	}
    }


    // expire the session and send back a warning page
    private boolean handleSessionExpiration()
    {
	if (! session.isExpired())
	{
	    // advance the expiration timestamp on the session
	    try
            {
		session.resetExpiration(conn,
		    Loco.getConfig().getInt ("LOCO_SESSION_EXPIRE_TIME"));
	    }
	    catch (FailedSessionException fse)
	    {
		log(0, "ERROR: RequestManager: resetExp failure: " +
		    StringUtil.exceptionStackTrace(fse));
		sendErrorMsg("rm_resetsession_fse");
		return(false);
	    }
	}
	else
	{
	    // we keep the old session so we can log with it.
	    Session old_session = session;
	    User old_user = (User) old_session.getValue("locomotive.user");

	    try
	    {
		// create a new session
		session = new Session(Loco.getConfig().getInt(
				      "LOCO_SESSION_EXPIRE_TIME"), conn);

		// this stores a new session cookie in the Response.
		session.updateBrowserCookie(resp);

		// remove the session objects from the old session
		old_session.removeObjects(conn);

	    }
	    catch (FailedSessionException fse)
	    {
		log(0, "ERROR: RequestManager: session reset: create new"+
		     " failed: " +  StringUtil.exceptionStackTrace(fse));
		sendErrorMsg("rm_sessionreset_cn_fse");
		return(false);
	    }

	    // send them the default template message upon expiration
	    // only if they were anonymous
	    if (! old_session.isAnonymous())
	    {
		resp.setContentType("text/html");
		resp.addString(Loco.evaluateTemplate(
		    "/error/session_expired.tmpl", new Hashtable()));
		resp.flush();

		Loco.logEvent(5, request_id, old_session, old_user,
			      "session", "expired",
			      "newsid: " + session.getId());

		return(false);
	    }

	}  // end if session expired
	return(true);
    }


    /** This will check the ip of the incoming connection, and
     * if it passes the test, will shut down the Locomotive.
     */
    private void attemptShutdown()
    {
	// make sure it's coming from somplace we like
	byte [] sock_ar = sock.getInetAddress().getAddress();
	int sock_addr = 0;
	int temp2 = 0;
	for (int i = 0; i < sock_ar.length; i++)
	{
	    temp2 = sock_ar[i];
	    if (temp2 < 0)
	    {
		temp2 = 256 + sock_ar[i];
	    }
	    int temp = temp2 << (24 - (i*8));
	    sock_addr += temp;
	}

	boolean ok = false;
	int ip;
	for (ip = 0; ip < ip_accept.length; ip++)
	{
	    int so_su = sock_addr & ip_subnet[ip];
	    if (so_su == ip_calc[ip])
	    {
		ok = true;
		break;
	    }
	}


	if (ip >= ip_accept.length) // just use the last ip in the config
	    ip -= 1;

	// let's log all this for posterity
	StringBuffer sb_ip = new StringBuffer();
	StringBuffer sb_su = new StringBuffer();
	StringBuffer sb_so = new StringBuffer();
	for (int i = 0; i < 4; i++)
	{
	    int temp = (ip_accept[ip] >> (24 - i*8)) & 0x000000FF;
	    sb_ip.append(temp); sb_ip.append('.');

	    temp = (ip_subnet[ip] >> (24 - i*8)) & 0x000000FF;
	    sb_su.append(temp); sb_su.append('.');

	    temp = (sock_addr >> (24 - i*8)) & 0x000000FF;
	    sb_so.append(temp); sb_so.append('.');
	}

	sb_ip.setLength(sb_ip.length() - 1);
	sb_su.setLength(sb_su.length() - 1);
	sb_so.setLength(sb_so.length() - 1);

	log(0, "Attempt to shutdown from: " + sb_so.toString() +
	    "\nIPs accepted from " + sb_ip.toString() +
	    " with subnet " + sb_su.toString() +
	    "\t Accept shutdown: " + ok);

	if (ok)
	{
	    Loco.setStopAccepting ();
	}
    }


    /*----------------- HANDOFF  --------------------*/


    /**
     * This will pick a particular request type to send our request off
     * to, instantiate the request, send it, then clean it up.
     * This gets called in the Loco.redirect method, so we'll package protect
     * it for now.
     * This should be of a factory design, really. we'll see...
     */
    void handoff ()
    {
	switch (request_type)
        {
	    case SERVLET_REQUEST:
		srt.routeRequest(
		    request_id, sock, is, os, cgi_envars_hash,
		    form_hash, subs, url_tokens, request_uri, bstamp_id,
		    resp, conn, session, attributes );
		break;
	    case HANDLER_REQUEST:
		hrt.routeRequest(
                    request_id, sock, is, os, cgi_envars_hash,
		    form_hash, subs, url_tokens, request_uri, bstamp_id,
		    resp, conn, session, attributes );
		break;
	    default:  // this shouldn't ever happen
	        log(0, "ERROR: could not find appropriate request for uri " +
		    request_uri);
		sendErrorMsg(
                    "Internal Error: could not find service for request");
		break;
	}
    }

    /*---------------- CLEANUP ---------------------*/


    private void doCleanup()
    {
	// return session objects to db
	if (session != null && conn != null)
        {
	    try
	    {
		session.storeObjects(conn);
	    }
	    catch (SQLException se)
	    {
		log(0, "ERROR: session objects save failed: " + se.toString());
	    }
	}

	// return db connection
	if (conn != null)
	{
	    Loco.putBackDBConnection(conn);
	}

	closeStreams();
    }

    private void closeStreams()
    {
	try
        {
	    is.close();
	}
        catch (IOException ioe)
        {
	    log(0, "ERROR: RM(" + request_id + ") closeStreams(): at is: " +
	        StringUtil.exceptionStackTrace(ioe));
	}

	try
        {
	    os.close();
	}
        catch (IOException ioe)
        {
	    log(0, "ERROR: RM(" + request_id + ") closeStreams(): at os: " +
		 StringUtil.exceptionStackTrace(ioe));
	}
    }

    /*------------------------- UTILITY ---------------------*/


    /** logs to the server_log */
    protected void log(int lev, String s)
    {
	server_log.log(lev, "REQ(" + request_id + ") " + s);
    }


    /**
     * Gets the cookie value out of the environment passed from the web
     * server.
     * @param name  -  the name of the cookie.
     */
    public String getCookieValue(String name)
    {
	String http_header_value =
	    (String) cgi_envars_hash.get("HTTP_COOKIE");

	if (http_header_value == null)
	{
	    return null;
	}

	name = name + "=";
	String temp;

	StringTokenizer st = new StringTokenizer(http_header_value, " ;");

	// below construct seems a bit silly, but it allows us an early exit
	// either by finding what we wanted, or by running out of cookie names

	while (st.hasMoreTokens())
	{
	    temp = st.nextToken();
	    if (temp.startsWith(name))
	    {
		return(temp.substring(name.length()));
	    }
	}
	return(null);
    }


    /** Returns true if the client returned any cookies to us; false if
     * otherwise.  This can return false both if the user doesn't have
     * any cookies set for our domain, or if the user's browser doesn't
     * support cookies
     */
    private boolean cookiesReturned()
    {
	return(((String) cgi_envars_hash.get("HTTP_COOKIE")) != null);
    }


    /** use this to send an error response before the handlerdata object
     * is created */
    protected void sendErrorMsg (String message)
    {
	java.util.Date dt = new java.util.Date();

	// note that if the Response has been flushed at all, the
	// http headers have already been sent.  There is no way to avoid
	// strange behaviours in the output at this point.

	// clean up whatever else has been placed in the response thus far.
	resp.setContentType("text/html");
	resp.clearHeaders();
	resp.clearBody();

	// do the error message. now uses template, new for 3.3
	Hashtable esubs = new Hashtable();
	esubs.put("ERROR_MSG", message);
	esubs.put("REQUEST_ID", String.valueOf(request_id));
	esubs.put("SYSTEM_TAG", config.getString("SYSTEM_TAG"));
	esubs.put("SYSTEM_INSTANCE_ID",
		  config.getString("SYSTEM_INSTANCE_ID"));
	esubs.put("ERROR_DATE", dt.toString());

	String s = Loco.evaluateTemplate("error/generic_error.tmpl",
				         esubs);
	if (s.equals(""))
	{
	    s = "<HTML>LOCO ERROR: " +
		"<BR>ERROR_MSG: " + message +
		"<BR> (The generic error template is missing. " +
		" Administrator, please configure) </HTML>";
	}

	resp.addString(s);
	resp.flush(server_log);
    }

    /*-----------------  STATIC ACCESS METHODS ----------------*/
    /* these should probably be in the Loco class..... */


    /**
     *toggle auto session handling <BR> If true, which is the
     * default, the RequestManager will automatically sense and handle
     * expired sessions.  The default timeout is 3600 seconds (one
     * hour) and is configured via the primary Locomotive
     * configuration file.  Sessions that have not been active for
     * greater than the timeout will be reset to a new session, stored
     * in the database, and the a new browser session cookie will set
     * in the current response object.  If the previous session
     * belonged to an authenticated user (aka, one who was logged in)
     * they will be sent the default 'session expired' template,
     * located at templates/error/session_expired.tmpl off the
     * template root.  The request will then end and a handler will
     * not be invoked.  If the old session contained an anonymous user
     * (or an &quot;unauthenticated user&quot;) no message is sent
     * and the request is passed into the apprpriate handler with the
     * new anonymous session.  If a handler needs to
     * process expired sessions on its own, setAutoSessionHandling
     * should be passed a &quot;false&quot;.  Note that this cannot be
     * set on a per-handler basis, and will effect all installed Handlers.
     * @param on new value of auto session handling flag */
    public static void setAutoSessionHandling (boolean auto_handle)
    {
	auto_session_handling = auto_handle;
    }


    /**
     * returns the current value of the auto_session_handling flag.
     * @see setAutoSessionHandling
     */
    public static boolean getAutoSessionHandling()
    {
	return(auto_session_handling);
    }


    /**
     * This will return the Properties object associated with a Handler,
     * if there is any.  The Properties object is created from a file
     * specified in the hrt.conf file, and is loaded at startup and
     * whenever the Handler is reloaded.  If no Properties file was
     * specified in the hrt.conf file, then this returns null.
     * <P>Each Properties object associated with a Handler will have
     * at least one variable in it- <PRE>
     *   PROPERTIES_PATH </PRE>
     * which is the absolute path to the file the Properties were loaded
     * from.  This will allow Handlers to manipulate and re-store their
     * Properties.
     * @see org.locomotive.loco.handler.Handler
     * @see org.locomotive.loco.handler.HandlerData#handler_config
     */
    public static Properties getHandlerProperties(Class theclass)
    {
	return(hrt.getProperties(theclass));
    }


    /** Marks the Handler class specified by the string for reloading, which
     * means that upon the next request for that handler, the handler's
     * properties file will be reloaded and the handler's init() method will
     * be called. <BR>
     * If the string is left blank ("") or is null, then all Handler
     * classes are reloaded- see markAllHandlersForReloading(). <BR>
     * This only reinstantiates the class and re-reads the properties files.
     * It doesn't go back and reload the hrt.conf file, nor does it actually
     * reload the class into the VM.
     * @param token the token that refers to the Handler class in the
     *        hrt.conf file
     */
    public static void reloadHandler(String token)
    {
	try
	{
	    hrt.reloadService(token);
	}
        catch (InitializationException ie)
        {
	    server_log.log(0, "ERROR: HandlerRoutingTable: " +
			   " could not reload service for token: " +
			   token + ": " + ie.toString());
	}
    }


    /**
     * Reloads the hrt.conf file, then reinitializes all the handlers and
     * reloads each of their properties files.
     * It doesn't actually reload Classes from their class files, though.
     * The reload will occur on the next request to the Locomotive.
     * @return true if we reloaded successfully, false if otherwise
     */
    public static boolean reloadHandlerRoutingTable()
    {
	server_log.log(0, "RM: Attempting reload of " +
		       "Handler Routing table...");

	String hrt_path = config.getString("HANDLER_ROUTING_TABLE_PATH");

	if (! hrt_path.equals("none") && !hrt_path.startsWith(File.separator))
        {
	    hrt_path = config.getString("SYSTEM_SERVER_ROOT_PATH") +
		File.separator + hrt_path;
	}

	try
        {
	    synchronized (srt)
	    {
		hrt.shutdown();  // at least try...
		System.gc();
		hrt = new HandlerRoutingTable(hrt_path, Loco.getServerLog(),
					      class_path, auto_reload_classes);
		hrt.initialize();
	    }

	    return(true);
	}
        catch (InitializationException ie)
	{
	    server_log.log(0, "ERROR: could not reload ServletRoutingTable: " +
			   ie.toString());
	    return(false);
	}
    }


    /** Marks the Servlet class specified by the string for reloading, which
     * means that upon the next request for that Servlet, the Servlet's
     * properties file will be reloaded and the Servlet's init() method will
     * be called. <BR>
     * If the string is left blank ("") or is null, then all Servlet
     * classes are reloaded- see markAllServletsForReloading(). <BR>
     * This only reinstantiates the class and re-reads the properties files.
     * It doesn't go back and reload the hrt.conf file, nor does it actually
     * reload the class into the VM.
     * @param token the token that refers to the Servlet class in the
     *        hrt.conf file
     */
    public static void reloadServlet(String token)
    {
	try
	{
	    hrt.reloadService(token);
	}
        catch (InitializationException ie)
	{
	    server_log.log(0, "ERROR: ServletRoutingTable: " +
			   " could not reload service for token: " +
			   token + ": " + ie.toString());
	}
    }


    /**
     * Reloads the servlet.conf file, then reinitializes all the Servlets and
     * reloads each of their properties files.
     * It doesn't actually reload Classes from their class files, though.
     * The reload will occur on the next request to the Locomotive.
     * @return true if we reloaded successfully, false if otherwise
     */
    public static boolean reloadServletRoutingTable()
    {
	server_log.log(0, "RM: Attempting reload of " +
		       "Servlet Routing table...");

	String srt_path = config.getString("SERVLET_ROUTING_TABLE_PATH");

	if (! srt_path.equals("none") && !srt_path.startsWith(File.separator))
        {
	    srt_path = config.getString("SYSTEM_SERVER_ROOT_PATH") +
		File.separator + srt_path;
	}

	try
        {
	    synchronized (srt)
	    {
		srt.shutdown();  // dump from memory...
		System.gc();
		srt = new ServletRoutingTable(srt_path, Loco.getServerLog(),
					      class_path, auto_reload_classes);
		srt.initialize();
	    }

	    return(true);
	}
        catch (InitializationException ie)
	{
	    server_log.log(0, "ERROR: could not reload ServletRoutingTable: " +
			   ie.toString());
	    return(false);
	}
    }

    /*-------------------- INIT AND SHUTDOWN METHODS --------------*/


    /**
     * setup all persistant data and initialize the various services
     */
    static boolean initialize(LocoConfig lc, Log s_log, Log e_log)
    {
	config = lc;
	server_log = s_log;
	event_log = e_log;

	boolean autoHandling = config.getString("LOCO_SESSION_AUTOHANDLE").
	    equalsIgnoreCase("YES");

	setAutoSessionHandling(autoHandling);

	// store the accept ip and subnet in byte arrays
	// see the explanations at the beginning of the class
	StringTokenizer ips =
	    new StringTokenizer(lc.getString("LOCO_ACCEPT_IP"), ",");
	StringTokenizer sus =
	    new StringTokenizer(lc.getString("LOCO_ACCEPT_SUBNET"), ",");

	int ip_length = ips.countTokens();
	if (ip_length != sus.countTokens())
	{
	    s_log.log(0, "ERROR in config: LOCO_ACCEPT_IP does not have the" +
		      "same number of addresses as LOCO_ACCEPT_SUBNET");
	    return false;
	}

	ip_accept = new int[ip_length];
	ip_subnet = new int[ip_length];
	ip_calc =   new int[ip_length];

	for (int j = 0; j < ip_length; j++)
	{
	    String ip_str = ips.nextToken();
	    StringTokenizer ip = new StringTokenizer(ip_str, ".");

	    if (ip.countTokens() != 4)
            {
		s_log.log(0, "ERROR in LOCO_ACCEPT_IP config: " +
			  " IP addresses must have 4 octets: " + ip_str);
		return false;
	    }

	    for (int i = 0; i < 4; i++)
            {
		try
                {
		    int temp = Integer.parseInt(ip.nextToken());
		    temp = temp << (24 - (i*8));
		    //	            Loco.log(9, "ip[" + i + "]: " + temp);
		    ip_accept[j] += temp;
		}
                catch (NumberFormatException e)
                {
		    s_log.log(0, "ERROR in LOCO_ACCEPT_IP config: " +
			      "could not convert IP to number: " + ip_str);
		    return false;
		}
	    }

	    ip_str = sus.nextToken();
	    StringTokenizer su = new StringTokenizer(ip_str, ".");
	    if (su.countTokens() != 4)
            {
		s_log.log(0, "ERROR in LOCO_ACCEPT_IP config: " +
			  " Subnet masks must have 4 octets: " + ip_str);
		return false;
	    }

	    for (int i = 0; i < 4; i++)
            {
		try
                {
		    int temp = Integer.parseInt(su.nextToken());
		    temp = temp << (24 - (i*8));
		    //	                Loco.log(9, "su[" + i + "]: " + temp);
		    ip_subnet[j] += temp;
		}
                catch (NumberFormatException e)
                {
		    s_log.log(0, "ERROR in LOCO_ACCEPT_SUBNET: " +
			      "could not convert subnet to number: " + ip_str);
		    return false;
		}
	    }

	    ip_calc[j] = ip_subnet[j] & ip_accept[j];

	    s_log.log(9, "ip accept pair: " + ip_accept[j] +
		      " subnet: " + ip_subnet[j]);

	}

	// set up the user and session objects
	User.initialize();
	Session.initialize();


	// get the service default
	String service = lc.getString("SERVICE_DEFAULT");

	Loco.log(9, "Setting service default: " + service);

	if (service.equalsIgnoreCase("handler"))
        {
	    SERVICE_DEFAULT = HANDLER_REQUEST;
	}
        else   // we made sure SERVICE_DEFAULT was valid in LocoConfig
        {
	    SERVICE_DEFAULT = SERVLET_REQUEST;
	}

	// finally, let's start up our various service managers
	String class_str = config.getString("SERVICE_CLASSPATH");
	String ar_str = config.getString("AUTO_RELOAD_SERVICES");

	if (ar_str.equalsIgnoreCase("YES") || ar_str.equalsIgnoreCase("TRUE"))
        {
	    auto_reload_classes = true;
	}

	StringTokenizer st = new StringTokenizer(class_str, ":");
	class_path = new Vector(st.countTokens());
	while (st.hasMoreTokens())
        {
	    String next = st.nextToken();
	    File f = null;

	    if (! next.startsWith(File.separator))
	    {
		f = new File(config.getString("SYSTEM_SERVER_ROOT_PATH") +
			     File.separator + next);
	    }
            else
	    {
		f = new File(next);
	    }

	    if (f.exists())
            {
		class_path.addElement(f);
	    }
	    else
            {
		Loco.log(0, "ERROR: SERVICE_CLASSPATH element does not " +
			 "exist: " + f.toString());
		return(false);
	    }
	}

	try
        {
	    // handlers
	    String hrt_path = lc.getString("HANDLER_ROUTING_TABLE_PATH");
	    if (! hrt_path.equals("none") &&
		! hrt_path.startsWith(File.separator))
	    {
		hrt_path = lc.getString("SYSTEM_SERVER_ROOT_PATH") +
		    File.separator + hrt_path;
	    }

	    hrt = new HandlerRoutingTable(hrt_path, server_log, class_path,
					  auto_reload_classes);
	    hrt.initialize();
	    Loco.log(1, "All Handlers loaded successfully.");

	    // servlets
	    String srt_path = lc.getString ("SERVLET_ROUTING_TABLE_PATH");
	    if (! srt_path.equals("none") && !srt_path.startsWith(
		File.separator))
	    {
		srt_path = lc.getString("SYSTEM_SERVER_ROOT_PATH") +
		    File.separator + srt_path;
	    }
	    srt = new ServletRoutingTable(srt_path, server_log, class_path,
					  auto_reload_classes);
	    srt.initialize();
	}
	catch (InitializationException ie)
	{
	    s_log.log(0, "ERROR: " + StringUtil.exceptionStackTrace(ie));
	    System.out.println("RequestManager initialization failure" +
			       " - see logs");
	    return(false);
	}

	return(true);
    }


    /**
     * shuts down the various services currently running, and saves
     * the cached session objects, if session object caching is turned on
     * called by Loco.shutdownAll () during server initiated bail
     * which, we hope, is how it always goes down
     */
    static boolean shutdown()
    {
	Connection conn = Loco.getDBConnection();
	Session.storeSessionObjects(conn);
	Loco.putBackDBConnection(conn);

	if (! hrt.shutdown())
	{
	    server_log.log(0, "ERROR: HandlerRoutingTable: " +
			   "could not shut down all handlers");
	    return(false);
	}

	if (! srt.shutdown())
	{
	    server_log.log(0, "ERROR: ServletRoutingTable: " +
			   "could not shut down all handlers");
	    return(false);
	}

	return(true);
    }

    /**
     * included to make javac find and compile the externally referenced
     * Handlers
     */
    private void includeDefaultHandlers()
    {
	org.locomotive.loco.ServerHandler sh;
	org.locomotive.loco.handler.DefaultHandler dh;
	org.locomotive.loco.handler.UnknownHandler uh;
	org.locomotive.module.base.BaseHandler bh;

	org.locomotive.loco.ServerServlet ss;
	org.locomotive.loco.servlet.DefaultServlet ds;
	org.locomotive.loco.servlet.UnknownServlet us;
	org.locomotive.module.base.BaseServlet bs;
    }

    /**
     * Simple InputStream class that implements ServletInputStream.
     * The Embedded Web Server needs this to support POSTs.
     */
    protected class EWSRequestInputStream extends ServletInputStream 
    {
	private BufferedReader r;

	protected EWSRequestInputStream(BufferedReader r)
	{
	    this.r = r;
	}

	/**
	 * our required method
	 */
	public int read() throws IOException
	{
	    return (r.read());
	}

	public void close() throws IOException
	{
	    r.close();
	}

	public void mark(int limit)
	{
            try
            {
	        r.mark(limit);
            }
            catch (Exception e)
            {
                // not yet sure what to do here
            }
	}
	
	public boolean markSupported()
	{
	    return (r.markSupported());
	}

	public void reset() throws IOException
	{
	    r.reset();
	}

	public long skip(long n) throws IOException
	{
	    return (r.skip(n));
	}
    }
}
