/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.server.Log;
import java.io.*;
import java.util.*;
import java.text.*;

import javax.servlet.http.Cookie;
import javax.servlet.ServletOutputStream;

/** This object encapsulates a response from the Locomotive back to the
 * browser.  Typically one would call addString() to add something to
 * the response, then call flush() to send it off to the browser.
 * Alternatively, one can cast this to a ServletResponse and use
 * getOutputStream to write to the response.
 */
public class Response
{
    public static int REPLY_TYPE_CGI = 1;
    public static int REPLY_TYPE_NSAPI = 2;
    public static int REPLY_TYPE_ISAPI = 3;
    
    private static int MAX_HEADERS = 32;
    protected String content_type;
    protected String char_encoding;
    private String header_names[];
    private String header_values[];
    private int next_header_index = 0;
    private boolean autoflush = false;
    protected boolean header_was_flushed = false;
    protected ResponseOutputStream resos;
    protected OutputStream os;
    protected int request_id = 0;
    protected int reply_type = 0;
    private static SimpleDateFormat sdf = null;

    // this is for the cookies
    static 
    {
	sdf = new SimpleDateFormat ("EEE, dd-MMM-yyyy HH:mm:ss 'GMT'", 
				    Locale.US);
//	sdf.setTimeZone (TimeZone.getTimeZone ("GMT"));
    }
	



    /**
     * constructs a new response.  Package friendly but not public.  
     * Responses are instantiated by the Request Manager for each response.
     * Request Manger is responsible for picking the correct protocol, 
     * which is determined by the protocol version sent to it via the 
     * tunnel.  
     */
    public Response (OutputStream theos, int rid, int reply_type_in)
    {
	os = theos;
	resos = new ResponseOutputStream(os);
	header_names = new String[MAX_HEADERS];
	header_values = new String [MAX_HEADERS];
	content_type = "text/html";
	char_encoding = "ISO-8859-1";
	request_id = rid;

	reply_type = reply_type_in;

	/* default to CGI protocol */
	if (reply_type != REPLY_TYPE_ISAPI && reply_type != REPLY_TYPE_NSAPI)
	{
	    reply_type = REPLY_TYPE_CGI;
	}
    }
    
    /**
     * adds an HTTP protocol header to the response.  Adding any headers
     * after the first flush() will be ineffective, since the HTTP 
     * header has already been sent off to the webserver/client. 
     * The values will be formatted in a manner appropriate for 
     * HTTP - similar to RFC 822 -  &quot; Name: Value &quot; <BR>
     * An example might be "Pragma: no-cache"
     * @param name - the name 
     * @param value - the value associated with name.
     */
    public void addHTTPHeader (String name, String value)
    {
	/* developers note: with the NSAPI and ISAPI, the name value pairs
	   are formatted at the web server.  Only in the case of CGI are
	   we supposed to do it.  Often in the case of CGI the headers must 
	   be reinterpreted by the CGI handler of the web server, which 
	   adds to CPU overhead on the webserver machine...
	   */
	if (next_header_index < MAX_HEADERS)
	{
	    header_names[next_header_index] = name;
	    header_values[next_header_index] = value;
	    next_header_index++;
	}
    }

    /**
     * returns the number of items currently scheduled to be sent in the 
     * HTTP header
     */
    public int getHeaderCount()
    {
	return (next_header_index);
    }


    /**
     * Returns true if a header with the specified name has been set.
     */
    public boolean containsHeader(String name)
    {
	if (name == null)
	{
	    return (false); 
	}

	for (int i = 0; i < header_names.length; i++)
	{
	   if (name.equals(header_names [i]))
	   {
	       return (true);
	   }
	}
	
	return (false);
    }


	


    /**
     * sets the content type HTTP header to the the input string.  The
     * input should be of the form &quot;major/minor&quot;  If left unset
     * the Response will default to type &quot;text/html&quot;
     * @param t  the MIME type of the response data
     */
    public void setContentType (String t)
    {
	content_type = t;
	if (content_type != null)
	{
	    int index = content_type.indexOf("charset=");
	    if (index > 0)
	    {
		int end = content_type.indexOf(";", index);
		char_encoding = content_type.substring(index + 8, end);
	    }
	}
    }


    /**
     * returns the character encoding, if it was included in the 
     * content type.  If no character encoding was included, this
     * will return 'ISO-8859-1'.
     */
    public String getCharacterEncoding()
    {
	return (char_encoding);
    }
 

    /**
     * wipes all headers for this response.  Content-Type is unaffected, 
     * as it is required by the CGI spec and by Loco native webserver 
     * connectivity modules.
     */
    public void clearHeaders()
    {
	next_header_index = 0;
    }


    /**
     * setting this to true will cause the Response to flush automatically
     * during every addString() or addBytes ()
     * @param bol - true if Response should autoflush
     */
    public void setAutoflush (boolean bol)
    {
	autoflush = bol;
    }


    /**
     * clears whatever data is currently resident in the Response body.
     * If set to autoflush, calling this will be next to useless.
     */
    public void clearBody()
    {
	resos.reset();
    }


    /**
     * Flushes the contents of the Response with no logging.
     */
    public void flush ()
    {
	flush (null);
    }

    /**
     * flushes the contents of the response to the OutputStream.  If the 
     * header has already been written, it is not written again, only the 
     * body buffers are emptied.  Flush can be called multiple times 
     * during the life of the response, each time flushing whatever has been
     * added since the last flush.  
     */
    public void flush (Log server_log)
    {
	try
	{
	    resos.flush(server_log);
	}
	catch (IOException ioe)
	{
	    if (server_log != null)
	    {
		server_log.log (0, "ERROR: RESP(" + request_id + 
				") write failed: " + ioe.toString());
	    }
	}
    }

    /**
     * gets the current length of the data in the response buffer.  The
     * length returned does not include the length of the header variables, 
     * only the response body.  
     */
    public int getCurrentResponseLength ()
    {
	return (resos.size());
    }

    /**
     * appends to the output of the Response.  If autoflush is true, the 
     * call will block until the data has been written or fails.  Given that
     * the typically write is destined for a locolink or webserver module, 
     * this will 
     * @param s - the data to be written to the client 
     */
    public void addString (String s)
    {
	byte b[] = s.getBytes();
	resos.write (b, 0, b.length);

	if (autoflush)
	{
	    flush (null);
	}
    }

    /**
     * adds the input byte array to the Response.  If autoflush is true, 
     * the Response will immediately flush the data.
     */
    public void addBytes (byte addme[])
    {
	resos.write (addme, 0, addme.length);

	if (autoflush)
	{
	    flush (null);
	}
    }


    /** Adds a cookie to the headers for this response */
    public void addCookie(Cookie cookie)
    {
	String temp = cookie.getName();
	if (temp == null || temp.length() < 1)
	{
	    return;
	}

	StringBuffer output = new StringBuffer();
	output.append(cookie.getName());
	output.append('=');
	output.append(cookie.getValue());
	output.append(';');

	temp = cookie.getDomain();
	if (temp != null)
	{
	    if (! temp.equals(""))
	    {
		output.append(" domain=");
		output.append(temp);
		output.append(';');
	    }
	}


	temp = cookie.getPath();
	if (temp != null)
	{
	    if (! temp.equals(""))
	    {
		output.append(" path=");
		output.append(temp);
		output.append(';');
	    }
	}

	int maxAge = cookie.getMaxAge();
	if (maxAge != -1)  // -1 is not a persistent cookie
	{
	    Date d = null;
	    if (maxAge >= 0)
	    {
		d = new Date(System.currentTimeMillis() + 
                             (1000 * (long) maxAge) );
	    }
	    else
	    {
		d = new Date(365 * 86400 * 1000); // a long time ago
	    }
	    
	    output.append(" expires=");
	    output.append( sdf.format (d) );
	    output.append(';');
	}

	if (cookie.getSecure())
	{
	    output.append(" secure;");
	}
	output.setLength(output.length() - 1); // get rid o' trailing semicolon

	addHTTPHeader ("Set-Cookie", output.toString());
    }


    /** This writes the headers to the OutputStream in correct protocol. 
     * This get's called automatically when either of the two
     * inner stream flush, so we don't have to worry about calling it
     * explicitly.
     */
    protected void writeHeader (Log server_log) throws IOException
    {
	/* this is basically an inlined subroutine */
	/* currently NSAPI and ISAPI use same protocol */
	if (reply_type == REPLY_TYPE_NSAPI || reply_type == REPLY_TYPE_ISAPI)
	{
	    DataOutputStream dos = new DataOutputStream (os);

	    String ct = "Content-Type";

	    try
	    {
		dos.writeInt (ct.length());
		dos.writeBytes (ct);

		dos.writeInt (content_type.length());
		dos.writeBytes (content_type);
	    
		for (int i = 0; i < next_header_index; i++)
		{
		    // name
		    dos.writeInt (header_names[i].length());
		    dos.writeBytes (header_names[i]);
		
		    //value 
		    dos.writeInt (header_values[i].length());
		    dos.writeBytes (header_values[i]);
		}
	    
		// marks end of header.  Everything else that follows 
		// is document body
		dos.writeInt (0);
		dos.flush ();
	    }
	    catch (IOException ioe)
	    {
		if (server_log != null)
		{
		    server_log.log (0, "ERROR: RESP(" + 
				    request_id + ") write header failed: " + 
				    ioe.toString());
		}
	    }
	    
	    header_was_flushed = true;
	    return;
	}
	
	/* reply_type == REPLY_TYPE_CGI   original style */
	StringBuffer sb = new StringBuffer ();

	// always do content-type - CGI requires it.
	sb.append ("Content-Type: ");
	sb.append (content_type);
	sb.append ("\r\n");

	for (int i = 0; i < next_header_index; i++)
	{
	    sb.append (header_names[i]);
	    sb.append (": ");
	    sb.append (header_values[i]);
	    sb.append ("\r\n");
	}

	// new line after header seperates header from data to follow
	sb.append("\r\n");

	// java IO is a bit inefficient - alot of data gets copied about in 
	// memory, plus many looping function calls - have a look in 
	// the source for OutputStream.write() 
	// some time.  Pointers still have their uses. See 
	// "Why Object Orientation is Inappropriate for Building Operating 
	// Systems"
	String output = sb.toString();
	byte b[] = output.getBytes();

	try
	{
	    os.write (b, 0, b.length);
	    os.flush();
	}
	catch (IOException ioe)
	{
	    if (server_log != null)
	    {
		server_log.log (0, "ERROR: RESP(" + request_id + 
				") write header failed: " + ioe.toString());
	    }
	    throw ioe;
	}
	    
	header_was_flushed = true;
    }




    /** 
     * our little ServletOutputStream.      
     * It stores everything in a ByteArrayOutputStream, and sends it all
     * out when we flush.  Note that flush will automatically send the
     * headers out, so we don't need to worry about that.
     */
    protected class ResponseOutputStream extends ServletOutputStream
    {
	ByteArrayOutputStream baos;
	Log server_log;
	OutputStream os;

	protected ResponseOutputStream(OutputStream os)
	{
	    baos = new ByteArrayOutputStream (8192);
	    this.server_log = server_log;
	    this.os = os;
	}

	/** required */
	public void write(int i)
	{
	    baos.write(i);
	}

	public void write(byte [] b)
	{
	    baos.write(b, 0, b.length);
	}

	public void write(byte [] b, int start, int len)
	{
	    baos.write(b, start, len);
	}
	
	public void flush() throws IOException
	{
	    flush(null);
	}

	public void flush(Log server_log) throws IOException
	{
	    if (! header_was_flushed)
	    {
		writeHeader(server_log);
	    }
	
	    baos.writeTo (os);
	    
	    os.flush();
	    
	    if (server_log != null)
	    {
		server_log.log (8, "RESP(" + request_id + 
				") flushed body size : " + baos.size());
	    }
	    baos.reset();
	}

	public void reset()
	{
	    baos.reset();
	}
	
	public int size()
	{
	    return (baos.size());
	}

	public void close() throws IOException
	{
	    flush(null);

	    os.close();
	}
    }

}
