/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.loco.handler.*;
import org.locomotive.server.Log;
import org.locomotive.steam.CachedPageLoader;


/**
 * The Server Handler.  The server provides a window into the server itself, 
 * rather than to any particular module handler.  It is accessed by the 
 * handler url-name &quot;srv&quot;. Valid commands include: <BR>
 * <UL>
 * <LI> srv - display status
 * <LI> srv?shutdown - mark the server for shutdown.  
 *      calls Loco.setStopAccepting
 * <LI> srv?status - display status
 * <LI> srv?reload_templates - tells the CachePageManager to flush cached 
 *      Templates
 * <LI> srv?reload_hrt - attempts to reload the hrt.conf file and 
 *      re-initialize all the Handlers.
 * <LI> srv?reload_dbconnpool - wipes all the present connections from the
 *      Locomotive database connection pool and fills it with new ones
 * </UL>
 * @see org.locomotive.loco.Loco#setStopAccepting
 */
public class ServerHandler implements Handler
{
    private HandlerData hd;  
    
    /**
     * this is the entry point for the server  
     */
    public void handleRequest (HandlerData hd_in)
    {
	hd = hd_in;

	// exit immediately if they aren't supposed to be here
	if (! hd.user.canViewServerStatus())
	{
	    showUnAcknowledgedMessage ();
	    return;
	}

	if (hd.session.isExpired())
	{
	    sendMessage(getExpiredMessage ());
	    return;
	}

	// default to showing status
	if (hd.url_tokens.length < 2)
	{
	    showStatus();
	    return;
	}
	
	// url_tokens has at least two entries at this point
	if (hd.url_tokens[1].equals("shutdown"))
	{
	    sendMessage(handleShutdownRequest (hd.user));
	}
	else if (hd.url_tokens[1].equals("reload_templates"))
	{
	    sendMessage(handleTemplateReloadRequest (hd.user));
	}
	else if (hd.url_tokens[1].equals("status"))
	{
	    showStatus();
	}
	else if (hd.url_tokens[1].equals("reload_hrt"))
	{
	    sendMessage(handleHRTReloadRequest(hd.user));
	}
	else if (hd.url_tokens[1].equals("reload_dbconnpool"))
	{
	    sendMessage(handleDBConnPoolReloadRequest(hd.user));
	}
	else 
	{
	    sendMessage(getBogusCommandMessage ());
	}
    }

    protected static String handleShutdownRequest (User user)
    {
	if (user.canShutDownLoco())
	{
	    Loco.setStopAccepting ();
	    return ("System shutdown initiated - marked for shutdown");
	}
	else 
	{
	    return ("Shutdown denied: no priviledges");
	}
    }

    protected static String handleTemplateReloadRequest (User user)
    {
	if (user.canChangeLocoVars())
	{
	    CachedPageLoader.clearTemplateCache();
	    return ("Template Cache cleared.");
	}
	else
	{
	    return ("Template Cache Flush failed: no priviledges");
	}
    }


    protected static String handleHRTReloadRequest(User user)
    {
	if (user.canChangeLocoVars())
	{
	    if (Loco.reloadHandlerRoutingTable())
	    {
		return ("Reload of Handler Routing Table " +
			     "successful! ");
	    }
	    else
	    {
		return ("Handler Routing Table reload failed.<BR> " +
			"Please see logs for details");
	    }
	}
	else
	{
	    return ("Handler Routing Table Reload failed: no privileges");
	}
    }


    protected static String handleDBConnPoolReloadRequest(User user)
    {
	if (user.canChangeLocoVars())
	{
	    if (Loco.reloadDBConnectionPool())
	    {
		return ("Reload of Database Connection Pool " +
			     "successful! ");
	    }
	    else
	    {
		return ("Database Connection Pool reload failed.<BR> " +
			"Please see logs for details");
	    }
	}
	else
	{
	    return ("Database Connection Pool reload failed: no privileges");
	}
    }

    private void showStatus()
    {
	StringBuffer output = new StringBuffer();

	// headers
	output.append ("<HTML>\n<HEAD><TITLE>Server Handler" + 
			   "</TITLE></HEAD>\n<BODY BGCOLOR=\"FFFFFF\">\n");
	
	output.append ("<H1 ALIGN=CENTER>Loco - Server Internal Status</H1>\n");
	output.append ("<TABLE BORDER=0 SIZE=100%>\n");

	output.append ("<TR><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?status\">" +
		       "Reload Status</A></B></TD></TR>\n");
	output.append ("<TR><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?shutdown\">" +
		       "Shutdown Locomotive</A></B></TD></TR>\n");
	output.append ("<TR><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?reload_templates\">" +
		       "Reload Templates</A></B></TD></TR>\n");
	output.append ("<TR><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?reload_hrt\">" +
		       "Reload Handlers</A></B></TD></TR>\n");
	output.append ("<TR><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?reload_dbconnpool\">" +
		       "Reload Database Connection Pool</A></B></TD></TR>\n");

	output.append ("<TR><TD COLSPAN=2>&nbsp;</TD></TR>\n");
	output.append ("<TR><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"base\">" +
		       "Locomotive Base handler</A></B></TD></TR>\n");

	output.append ("<TR><TD COLSPAN=2>&nbsp;</TD></TR>\n");

	output.append (getStatusVars());
	
	output.append ("<TR><TD COLSPAN=2>&nbsp;</TD></TR>\n");

	// config info
	output.append ("<TR><TD ALIGN=CENTER COLSPAN=2>" +
		       "<H2>Configuration Info</H2>\n");
	output.append (Loco.getConfig ().toHTMLRows());
	output.append ("</TABLE>");

	output.append ("</BODY>\n</HTML>");

	hd.resp.addString(output.toString());
	hd.resp.flush();
    }
		

    /**
     * send back a nice status page
     */
    protected static String getStatusVars ()
    {
	// get the values
	int thread_count = Loco.getActiveThreadCount ();
	int pool_size =    Loco.getThreadPoolSize ();
	int pool_max =     Loco.getMaxThreadPoolSize ();
	
	int open_db_connections = Loco.getOpenDBConnectionCount ();

	String vers = Loco.getVersionString ();
	java.util.Date since_dt = Loco.getStartupDate ();
	int last_request_id = Loco.getCurrentRequestID () - 1;

	Runtime r =           Runtime.getRuntime();
	long free_mem =        r.freeMemory ();
	long total_mem =       r.totalMemory ();

	// format the output
	StringBuffer output = new StringBuffer();

	// version
	output.append ("<TR><TD>Loco Version</TD><TD align=right>");
	output.append (vers);
	output.append ("</TD></TR>\n");

	// active threads
	output.append ("<TR><TD>Currently Active Threads</TD><TD align=right>");
	output.append (thread_count);
	output.append ("</TD></TR>\n");

	// thread pool size
	output.append ("<TR><TD>Current Thread Pool Size</TD><TD align=right>");
	output.append (pool_size);
	output.append ("</TD></TR>\n");

	// thread pool max size
	output.append ("<TR><TD>Maximum Thread Pool Size</TD><TD align=right>");
	output.append (pool_max);
	output.append ("</TD></TR>\n");

	// open db connections
	output.append ("<TR><TD>Open DB Connections</TD><TD align=right>");
	output.append (open_db_connections);
	output.append ("</TD></TR>\n");

	// uptime
	output.append ("<TR><TD>Up since</TD><TD align=right>");
	output.append (since_dt);
	output.append ("</TD></TR>\n");

	// last request id
	output.append ("<TR><TD>Last Request ID</TD><TD align=right>");
	output.append (last_request_id);
	output.append ("</TD></TR>\n");

	// memory total
	output.append ("<TR><TD>Total Java Heap Memory</TD><TD align=right>");
	output.append (total_mem);
	output.append ("</TD></TR>\n");

	// memory free
	output.append ("<TR><TD>Free Java Heap Memory</TD><TD align=right>");
	output.append (free_mem);
	output.append ("</TD></TR>\n");


	return (output.toString());
    }

    protected static String getBogusCommandMessage ()
    {
	return ("unknown or unimplemented command");
    }
    
    /**
     * used as a simple wrapper to include the surrounding page
     */
    private void sendMessage (String s)
    {
	hd.resp.addString (
	    "<HTML>\n<HEAD><TITLE>Server Handler</TITLE></HEAD>\n" + 
	    "<BODY BGCOLOR=\"#FFFFFF\"><H1>Loco - Server Handler</H1>\n" + 
	    "Command Results: <B>" + s + "</B>\n" + 
	    "<P><A HREF=\"srv?status\"><B>Server Status</B><BR>" +
	    "</BODY>\n</HTML>"
			      );
	    
	hd.resp.flush (hd.server_log);
    }

    private void showUnAcknowledgedMessage ()
    {
	// If we got here then we aren't supposed to know about the srv
	// handler... This is copied from unknown handler right now but should
	// be something called from here so it is guaranteed to be the same.

	hd.resp.addString("<HTML>\n<HEAD><TITLE>Unknown Handler" + 
			  "</TITLE></HEAD>\n<BODY><H1>Unknown Handler</H1>\n"+
			  "The handler specified could not be found<BR>\n" +
			  "</BODY>\n</HTML>");
	hd.resp.flush();
    }

    protected static String getExpiredMessage ()
    {
	// If we got here then we aren't supposed to know about the srv
	// handler... This is copied from unknown handler right now but should
	// be something called from here so it is guaranteed to be the same.

	return ("<HTML>\n<HEAD><TITLE>Server Handler" + 
		"</TITLE></HEAD>\n<BODY><H1>Server Handler</H1>\n"+
		"Your session has expired, go login again<BR>\n" +
		"</BODY>\n</HTML>");
    }

    /**
     * Place holder.  Currently the ServerHandler does not need any 
     * special initializations.  
     */
    public boolean init (Log initlog)
    {
	return (true);
    }

    /**
     * Place holder.  Currently the ServerHandler does not need any 
     * special initializations.  
     */
    public boolean shutdown (Log initlog)
    {
	return (true);
    }

}
