/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.locomotive.loco.servlet.*;


/**
 * The Server Servlet provides bare bones administration facilities for
 * the Locomotive.  It is accessed by the url token 'srv'.  Valid commands 
 * include: 
 * <UL>
 * <LI> srv - display status
 * <LI> srv?shutdown - mark the server for shutdown.  
 *      calls Loco.setStopAccepting
 * <LI> srv?status - display status
 * <LI> srv?reload_templates - tells the CachePageManager to flush cached 
 *      Templates
 * <LI> srv?reload_srt - attempts to reload the hrt.conf file and 
 *      re-initialize all the Handlers.
 * </UL>
 * <P>This class implements the same functionality as the ServerHandler,
 *    except in Servlet form.
 */
public class ServerServlet extends HttpServlet
{

    public void doGet(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	// here's how we get the HandlerData goodies
	Session session = (Session) req.getSession(false);
	User user = (User) session.getValue("locomotive.user");
	String [] url_tokens = 
	    (String []) req.getAttribute("locomotive.url_tokens");

	if (! user.canViewServerStatus())
	{
	    resp.sendError(resp.SC_NOT_FOUND,
			   "The requested service cannot be found.");
	    return;
	}

	resp.setContentType("text/html");
	PrintWriter pw = resp.getWriter();

	if (session.isExpired())
	{
	    pw.print(ServerHandler.getExpiredMessage());
	    pw.flush();
	    return;
	}

	// just show the status
	if (url_tokens.length < 2 || url_tokens[1].equals("status"))
	{
	    pw.print(getStatus());
	    pw.flush();
	}
	else if (url_tokens[1].equals("shutdown"))
	{
	    sendMessage(pw, ServerHandler.handleShutdownRequest(user));
	}
	else if (url_tokens[1].equals("reload_templates"))
	{
	    sendMessage(pw, ServerHandler.handleTemplateReloadRequest (user));
	}
	else if (url_tokens[1].equals("reload_srt"))
	{
	    sendMessage(pw, handleSRTReloadRequest(user));
	}
	else if (url_tokens[1].equals("reload_dbconnpool"))
	{
	    sendMessage(pw, ServerHandler.handleDBConnPoolReloadRequest(user));
	}
	else 
	{
	    sendMessage(pw, ServerHandler.getBogusCommandMessage ());
	}

	pw.flush();
    }


    protected static String handleSRTReloadRequest(User user)
    {
	if (user.canChangeLocoVars())
	{
	    if (Loco.reloadServletRoutingTable())
	    {
		return ("Reload of Servlet Routing Table " +
			     "successful! ");
	    }
	    else
	    {
		return ("Servlet Routing Table reload failed.<BR> " +
			"Please see logs for details");
	    }
	}
	else
	{
	    return ("Servlet Routing Table Reload failed: no privileges");
	}
    }

    private String getStatus()
    {
	StringBuffer output = new StringBuffer();
	
	output.append("<HTML>\n<HEAD><TITLE>Server Administration Servlet" + 
			   "</TITLE></HEAD>\n<BODY BGCOLOR=\"FFFFFF\">\n");
	
	output.append ("<H1 ALIGN=CENTER>Locomotive - Server Internal Status</H1>\n");
	output.append ("<TABLE BORDER=0 SIZE=100%>\n");

	output.append ("<TR bgcolor=\"#ffffff\"><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?status\">" +
		       "Reload Status</A></B></TD></TR>\n");
	output.append ("<TR BGCOLOR=\"#99CC99\"><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?shutdown\">" +
		       "Shutdown Locomotive</A></B></TD></TR>\n");
	output.append ("<TR bgcolor=\"#ffffff\"><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?reload_templates\">" +
		       "Reload Templates</A></B></TD></TR>\n");
	output.append ("<TR BGCOLOR=\"#99CC99\"><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?reload_srt\">" +
		       "Reload Servlets</A></B></TD></TR>\n");
	output.append ("<TR bgcolor=\"#ffffff\"><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"srv?reload_dbconnpool\">" +
		       "Reload Database Connection Pool</A></B></TD></TR>\n");

	output.append ("<TR><TD COLSPAN=2>&nbsp;</TD></TR>\n");
	output.append ("<TR BGCOLOR=\"#99CC99\"><TD COLSPAN=2 ALIGN=CENTER><B>" +
		       "<A HREF=\"base\">" +
		       "Locomotive Base Servlet</A></B></TD></TR>\n");

	output.append ("<TR><TD COLSPAN=2>&nbsp;</TD></TR>\n");

	output.append (ServerHandler.getStatusVars());
	
	output.append ("<TR><TD COLSPAN=2>&nbsp;</TD></TR>\n");

	// config info
	ServletContext sc = getServletContext();
	Enumeration e = sc.getAttributeNames();
	output.append ("<TR><TD ALIGN=CENTER COLSPAN=2>" +
		       "<H2>Configuration Info</H2>\n");

	while (e.hasMoreElements())
	{
	    String name = (String) e.nextElement();
	    output.append("<TR BGCOLOR=\"#99CC99\"><TD>");
	    output.append(name);
	    output.append("</TD><TD align=\"right\">");
	    output.append(sc.getAttribute(name));
	    output.append("</TD></TR>\n");

	    if (e.hasMoreElements())
	    {
		name = (String) e.nextElement();
		output.append("<TR BGCOLOR=\"#ffffff\"><TD>");
		output.append(name);
		output.append("</TD><TD align=\"right\">");
		output.append(sc.getAttribute(name));
		output.append("</TD></TR>\n");
	    }
	}
	output.append ("</TABLE>");

	output.append ("</BODY>\n</HTML>");

	return (output.toString());
    }

    /**
     * used as a simple wrapper to include the surrounding page
     */
    private void sendMessage (PrintWriter pw, String s)
    {
	pw.print(
	    "<HTML>\n<HEAD><TITLE>Server Handler</TITLE></HEAD>\n" + 
	    "<BODY BGCOLOR=\"#FFFFFF\"><H1>Loco - Server Handler</H1>\n" + 
	    "Command Results: <B>" + s + "</B>\n" + 
	    "<P><A HREF=\"srv?status\"><B>Server Status</B><BR>" +
	    "</BODY>\n</HTML>"
	    );
	    
	pw.flush ();
    }
}
