/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.loco.handler.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.util.*;

import org.locomotive.loco.store.*;
import org.locomotive.loco.store.session.*;

import java.io.*;
import java.util.*;
import java.sql.*;

import javax.servlet.http.*;

/**
 * The Locomotive is able to keep track of a user from one page-view to the
 * next, by sending a cookie to that user's browser and then using that cookie
 * to identify that user.
 *
 * A session is an object which represents a continuous series of
 * page-views by the same user.  Since it needs to be persistent, it
 * is stored in the database (in the loco_sessions table).  A unique
 * session is associated with a user's unique cookie.  The session's
 * ID is logged to the Server Log and the Event Log, so that the
 * session can be used to track the user.
 *
 * The Session for this user is passed into all Handlers via
 * the Handler.handleRequest (HandlerData hd) method.  It is set up by
 * the Locomotive, which finds the session in the database that is
 * associated with the unique cookie from a user's browser.

 * This version has changes by David Warnock. Changes are marked by
 * //dw comments.
 * Queries are marked by //dwq

 * The changes are to move out all persistance logic. This class gets
 * an instance of SessionStore from StoreFactory for the currently
 * configured persistance store
 */

public class Session implements HttpSession {
  private boolean istemp = true;
  private boolean from_cookie = false;

  private long sid = 0;
  private int srid = 0;
  private int userid = 0;
  private Timestamp expire = null;
  private Timestamp added = null;
  private Timestamp active = null;
  private Hashtable objects = null;

  private static Random rand;
  private static boolean rand_set = false;

  private static LocoConfig config = null;

  // if this is null, then caching will not be used.  initialize() sets
  // this to whatever the USE_CACHED_SESSION_OBJECTS config var is set to
  private static Hashtable session_objects = null;

  //dw
  // This is the SessionStore that provides all the persistance
  // call method createSessionStore to fill this
  private SessionStore sessionStore = null;

  /**
   * create a temporary anonymous session (no database persistance)
   * with a default expiration time of one hour.
   */
  public Session ()
  {
    this(3600);
  }

  /**
   * create a temporary anonymous session (no database persistance)
   * @param exp_seconds - the time from instantiation this session
   *                      will expire
   */
  public Session (int exp_seconds)
  {
    expire = new Timestamp(System.currentTimeMillis() +
                           (long) 1000 * (long) exp_seconds);
    objects = new Hashtable();
    objects.put("locomotive.user", new User());
    // dw
    createSessionStore();
  }

  /**
   * creates a persistant session, with corresponding db entry
   * @param exp_seconds   the time, in seconds from "now" that the
   *                       session will be expired
   * @param conn          a JDBC database connection with
   *                       access to the Loco tables
   * @exception FailedSessionException if there is a database error
   */
  public Session (int exp_seconds, Connection conn)
  throws FailedSessionException
  {
    this(exp_seconds);
    if (conn != null) {
      createNewSession (conn);
    }
  }

  /**
   * instantiates a session using session data stored persistantly in the
   * session database.
   * @param conn           a JDBC connection, opened to the session db
   * @param cookie_string  the cookie session string, as generated by
   *                       generateCookieString()
   * @exception FailedSessionException if there is a cookie parse error
   *                       or database load problem
   */
  public Session (Connection conn, String cookie_string)
  throws FailedSessionException
  {
    parseCookieString(cookie_string); // throws the exception

    //dw
		createSessionStore();

    if (conn != null) {
      from_cookie = true;
      istemp = false;
      load (conn);
    } else {
      // we expire an hour from now
      expire = new Timestamp(System.currentTimeMillis() +
                             (long) 3600000);
    }
  }

  /**
   * Loads a Session from the session database using the session id.
   * @param conn           a JDBC connection, opened
   * @param session_id  the numerical session id.
   * @exception FailedSessionException if the session can't be found
   *                       or database load problem occurs
   */
  public Session (Connection conn, long session_id)
  throws FailedSessionException
  {
    if (conn != null) {
      // dw
      createSessionStore();

      sid = session_id;
      istemp = false;
      load (conn);
    } else {
      throw new FailedSessionException ("db connection cannot be null");
    }
  }


  //dw
  /**
   * get a new SessionStore from the factory if one has not been found already
   * this needs to be called before any db stuff (obviously)
   */
  private void createSessionStore() {
    if ( sessionStore == null ) {
      sessionStore = StoreFactory.createSessionStore();
    }
  }

  /**
   * Parses a valid cookie string and store the cookie values in the session
   * Returns false if cookie String submitted is bogus, true if successfully
   * parsed
   *
   * @exception FailedSessionException if there is a cookie parse error
   *                       or database load problem
   */
  private void parseCookieString (String cs) throws FailedSessionException
  {
    StringTokenizer st = new StringTokenizer (cs, "_");
    if (st.countTokens() != 2) {
      throw new FailedSessionException("cookie parse failure");
    }

    try {
      sid = Long.parseLong (st.nextToken());
      srid = Integer.parseInt (st.nextToken());
    } catch (NumberFormatException nfe) {
      throw new FailedSessionException("cookie parse failure");
    }

  }

  /**
   * returns a string suitable for usage in a HTTP cookie
   * naming of the cookie is handled externally
   */
  public String generateCookieString()
  {
    return (String.valueOf (sid) + "_" + String.valueOf(srid));
  }

  /**
   * takes the input response object and places in it an outbound cookie
   * to update the browser's knowledge of the current session
   */
  public void updateBrowserCookie (Response resp)
  {
    Cookie c = new Cookie ("LOCOSID_" + config.getString("SYSTEM_TAG"),
                           generateCookieString());

    c.setDomain(config.getString("LOCO_COOKIE_DOMAIN_MAIN"));
    c.setPath("/");

    if (isExpired()) {
      c.setMaxAge(-100);  // this'll expire it
    } else {
      // if our session cookie is persisten, reset it's expiration date
      String persist = config.getString("PERSISTENT_SESSION_COOKIE");
      if (persist != null && persist.equalsIgnoreCase("YES")) {
	  //This seems to be a bug in the spec. The setMaxAge
	  //is supposed to take a value in seconds, but it only
	  //works correctly with milliseconds
	  // For example to set the MaxAge to be 1 day from now
	  // we needed to set it to 86400 * 1000 
	  c.setMaxAge((int)(expire.getTime() -  System.currentTimeMillis()));
      }
    }

    resp.addCookie(c);
  }

  /**
   * returns the userid of this session
   */
  public int getUserId()
  {
    return (userid);
  }

  /**
   * returns a true if this session does not have a specific user
   * associated with it.  @see org.locomotive.loco.User.isAnonymous().
   */
  public boolean isAnonymous()
  {
    return (userid <= 0);
  }


  /**
   * returns the session id (sid) of the session
   */
  public long getID()
  {
    return (sid);
  }


  /**
   * true if this session is temporary, not stored persistantly in the db
   * Temporary sessions cannot have users associated with them.
   * @deprecated in favor of isNew()
   */
  public boolean isTemporary()
  {
    return (istemp);
  }


  /**
   * True if session has expired
   */
  public boolean isExpired ()
  {
    Timestamp ts = new Timestamp(System.currentTimeMillis());

    if (ts.after (expire) || ts.equals (expire)) {
      return (true);
    }

    return (false);
  }


  /** returns the Session cookie string */
  public String toString()
  {
    return (generateCookieString());
  }


  /*----------------- HTTPSession implementation ------------------------*/

  /**
   * returns the session id (sid) of the session
   */
  public String getId()
  {
    return ( String.valueOf(sid) );
  }


  /**
   * true if the browser has not been sent by some notification about the
   * session. New sessions cannot have users associated with them.
   */
  public boolean isNew()
  {
    return (istemp);
  }


  /**
   * Returns the time in milliseconds, since 1/1/1970, when this session
   * was created/
   */
  public long getCreationTime()
  {
    return ( added.getTime() );
  }


  /**
   * Returns the time in milliseconds since the last request assigned to
   * this session
   */
  public long getLastAccessedTime()
  {
    return ( active.getTime() );
  }


  /**
   * Returns the maximum amount of time the session can be inactive before
   * it is expired, in seconds.
   */
  public int getMaxInactiveInterval()
  {
    return (Loco.getConfig().getInt("LOCO_SESSION_EXPIRE_TIME"));
  }

  /**
   * Returns the context in which this session is bound.
   * @deprecated as the SessionContext object is thought to be a
   *             security risk.
   */
  public HttpSessionContext getSessionContext()
  {
    return ( RequestManager.srt.getSessionContext() );
  }


  /**
   * Expires this session
   */
  public void invalidate()
  {
    Connection conn = Loco.getDBConnection();

    try {
      expire(conn);
    } catch (FailedSessionException fse) {
      Loco.log(0, "ERROR: could not expire session: " + fse.toString());
    }

    Loco.putBackDBConnection(conn);
  }


  /**
   * Stores the object persistently, keyed by the name.
   * Notifies the object, if it is an HttpSessionBindingListener.
   */
  public void putValue(String name, Object value)
  {
    if (objects.containsKey(name)) {
      removeValue(name);  // so the HttpSessionBindingEvent gets created
    }

    objects.put(name, value);
    if (value instanceof HttpSessionBindingListener) {
      HttpSessionBindingEvent hsbe =
      new HttpSessionBindingEvent(this, name);
      ( (HttpSessionBindingListener) value).valueBound(hsbe);
    }
  }


  /**
   * returns a stored object keyed by the name.  Returns null if no
   * object is stored by the key.
   */
  public Object getValue(String name)
  {
    return ( objects.get(name) );
  }


  /**
   * Returns an array of Strings representing the names to all the
   * Objects stored with this session
   */
  public String [] getValueNames()
  {
    String [] names = new String [objects.size()];
    Enumeration e = objects.keys();
    for (int i = names.length - 1; i >= 0; i--) {
      names[i] = (String) e.nextElement();
    }

    return ( names );
  }


  /**
   * Removes the object keyed by the name from the persistent store.
   * Notifies the object, if it is an HttpSessionBindingListener.
   */
  public void removeValue(String name)
  {
    Object o =  objects.remove(name);
    if (o != null && o instanceof HttpSessionBindingListener) {
      HttpSessionBindingEvent hsbe =
      new HttpSessionBindingEvent(this, name);
      ( (HttpSessionBindingListener) o).valueUnbound(hsbe);
    }
  }


  /**
   * Sets the expire time for this session to the number of seconds
   * into the future specified by the interval.
   * @return the new exipre time
   */
  public void setMaxInactiveInterval(int interval)
  {
    expire = new Timestamp( System.currentTimeMillis() +
                            (interval * 1000) );
  }



  /*------------------- Database access methods ------------------*/


  /**
   * associate a user with this session, persistantly.  Authentication
   * should be handled externally
   * @param conn    a JDBC connection with access to the Loco tables
   * @param user    the user to be associated
   * @exception FailedSessionException if there is a database storage problem
   */
  public void associateUser (Connection conn, User user)
  throws FailedSessionException
  {
    if (user.getUserid() <= 0) {
      throw new FailedSessionException ("attempt to associate " +
                                        "anonymous user with session");
    }

//	if (istemp)
//	{
//	    throw new FailedSessionException ("attempt to associate " +
//					      "user with temporary session");
//	}

    userid = user.getUserid();
    objects.put("locomotive.user", user);  // just to be sure...

    sessionStore.associateUser(conn, user, sid);
  }


  /**
   * expires this session, in both the object and the database
   * returns false if the operation failed
   * @param conn    a JDBC connection with access to the Loco session tables
   * @exception FailedSessionException if the session cannot be marked in
   *            the database
   */

  public void expire (Connection conn) throws FailedSessionException
  {
    // set the expire date in the db to yesterday
    expire = new Timestamp(System.currentTimeMillis() - (long) 86400000);

    if (istemp || conn == null) {
      return;
    }

    sessionStore.expire(conn, sid, expire);
  }


  /**
   * Stores the session_objects table in the database.
   * Does nothing if caching is being used
   * Should be called to store the session's resources after the session
   * has been used for a request, if caching is not being used.
   * @throw SQLException if there was a database error
   */

  //dwq
  // for consistancy should this not throw FailedSessionException not SQLException
  public void storeObjects(Connection conn) throws SQLException
  {
    if (session_objects != null || isExpired()) {
      return;
    }

    Loco.log(9, "storing objects in DB for sid " + sid + ": " + objects);

    if (objects == null) {
      objects = new Hashtable();
    }

    sessionStore.storeObjects(conn, sid, objects);
  }


  /**
   * Removes the objects for this session, either from the cache, if
   * it's being used, or from the database
   */
  public void removeObjects(Connection conn)
  throws FailedSessionException
  {
    // remove each of the objects, so they get a chance to respond
    // before disappearing, via HttpSessionBindingListeners
    Enumeration e = objects.keys();
    for (int i = objects.size(); i > 0; i--) {
      String name = (String) e.nextElement();
      removeValue(name);
    }

    // remove the session objects from our various storage means
    if (session_objects != null) {
      Object o = session_objects.remove(new Long(sid));
      Loco.log(9, "removed session objects for sid " + sid +
               ": " + o);
    } else {
      if (conn == null) {
        return;
      }

      //dwq
      // Wot no log here?
      sessionStore.removeObjects(conn, sid);
    }
  }

  /**
   * Advances the expiration time of the session by exp_seconds
   * if the session is persistant, the new time will be committed to the
   * session database
   * @param exp_seconds  the timespan in seconds to advance the expire time
   * @param conn          a JDBC connection with access to
   *	                    the Loco session tables
   * @exception FailedSessionException if there is a database error
   */
  public void resetExpiration (Connection conn, int exp_seconds)
  throws FailedSessionException
  {
    if (istemp || conn == null) return;

    // update the active and the expire times in db
    expire = new Timestamp(System.currentTimeMillis() +
                           (long) 1000 * (long) exp_seconds);

    // write it to the db too
    sessionStore.resetExpiration(conn, sid, expire);
  }


  /**
    * loads remainder of session info and validates it given the sid and srid
    * called internally from creator
    * May need to modify later for session loading based purley on sid
    * input, rather than cookie string input
    */
  private void load (Connection conn) throws FailedSessionException
  {

    LocoSessionRow r = sessionStore.getLocoSessionRow(conn, sid);

		int storedrand = r.getStoredrand();
		userid = r.getUserid();
		expire = r.getExpire();
		added  = r.getAdded();
		active = r.getActive();

		// check the srid against the parsed id
		if (from_cookie && storedrand != srid) {
			throw new FailedSessionException(
																			"security alert: cookie randnum != stored value");
		} else {
			srid = storedrand;
		}


		if (session_objects != null) {
			objects = (Hashtable) session_objects.get(new Long(sid));
		}

		// either we're not using the cache, or the cache doesn't
		// have the objects for this session.  Try to get them from
		// the DB
		if (objects == null) {
			Loco.log(9, "loading objects from DB for sid " + sid);

			objects = sessionStore.getSessionObjects(conn, sid);

			// If we're still null, go ahead and create a new set of objects
			if (objects == null) {
				Loco.log(9, "Loading session: " + sid +
								 ": session's objects are null");
				objects = new Hashtable();
			}

			if (session_objects != null) {
				session_objects.put(new Long(sid), objects);
			}
		}

		// if we don't have a user, go get one
		if (! objects.containsKey("locomotive.user") ) {
			User user = null;
			// go get the user for this session, if there is one
			if (userid > 0) {
				try {
					user = new User(conn, userid);
				} catch (UserNotFoundException unfe) {
					throw new FailedSessionException(
																					"User not found: " + unfe.toString());
				} catch (LoadUserException lue) {
					throw new FailedSessionException(
																					"User load failed: " + lue.toString());
				}
			} else {
				user = new User();
			}

			objects.put("locomotive.user", user);
		}
  }


  /** generates a new session, and session objects */
  private void createNewSession (Connection conn)
  throws FailedSessionException
  {

    //create srid (and cache rand)
    if (! rand_set) {
      rand = new Random (Runtime.getRuntime().freeMemory() +
                         expire.getTime()); // set in constructor
      rand_set = true;
    }

    srid = rand.nextInt();
    if (srid < 0) {
      srid = -srid;
    }


    sid = sessionStore.createNewSession( conn, srid, userid, expire );

    // objects hash should have been created in the constructor
    if (session_objects != null) {
      session_objects.put(new Long(sid), objects);
    }
  }



  /*--------------------- Static  methods -----------------*/


  /** Initializes static variables, like the config object and the
   * session_objects hash.  This needs to be run to decide how to cache
   * the session objects.
   * If the config var PERSISTENT_SESSION_OBJECTS is set to TRUE, then
   * the session objects are retrieved from the database at the beginning
   * of each request, and saved to the database at the end of each request.
   * If set to FALSE, they are kept in a static hashtable and will
   * disappear as soon as the VM is shut down.
   */
  protected static void initialize()
  {
    config = Loco.getConfig();

    if (config.getString("USE_CACHED_SESSION_OBJECTS")
        .equalsIgnoreCase("true")) {
      session_objects = new Hashtable();
    } else {
      session_objects = null;
    }
  }

  /**
   * Clean up the session objects cache- save all the session info
   * that is currently cached to the DB.  If we're not using the
   * cache, return.
   */
  protected static void storeSessionObjects(Connection conn)
  {
    if (session_objects == null || conn == null) {
      return;
    }

    //dw Ok I need my own sessionStore as it's a static method

    SessionStore store = StoreFactory.createSessionStore();
    store.storeSessionObjects(conn, session_objects);
  }

}
