/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import org.locomotive.loco.handler.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.loco.*;
import org.locomotive.server.*;
import org.locomotive.util.*;
import java.util.*;
import java.text.*;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * This class has methods which insert some standard variables into the steam
 * substitutions for use in HTML templates.  For example, the requester's 
 * username, userid, and session id are inserted as USERNAME, USERID, 
 * SESSIONID, respectively. The current date and time are also inserted, 
 * as DATE, and TIME.
 */
public class StandardInsertions
{
    protected static SimpleDateFormat dateFormat =
        new SimpleDateFormat ("MMM d, yyyy");
    protected static SimpleDateFormat timeFormat = 
        new SimpleDateFormat ("h:mm a");

    /** to hide from javadoc! */
    StandardInsertions() {}

    /** This method adds all the insertions implemented in the methods in this
     * class. See the other methods for the names and types of these insertions.
     */
    public static void insertAll (HandlerData hd)
    {
	hd.subs.put ("LOCOMOTIVE_SERVER_VERSION", Loco.getVersionString());
	hd.subs.put ("SESSIONID", String.valueOf (hd.session.getID()));
	hd.subs.put ("USERID", String.valueOf (hd.user.getUserid()));
	if (! hd.session.isAnonymous())
	{
	    hd.subs.put ("USERNAME", hd.user.getUsername());
	}
	hd.subs.put ("REQUEST_ID", String.valueOf(hd.getRequestID()));

	Date rightNow = new Date();
	hd.subs.put ("DATE", dateFormat.format(rightNow));
	hd.subs.put ("TIME", timeFormat.format(rightNow));

	if (hd.url_tokens.length > 0)
	{
	    hd.subs.put ("HANDLER_URLTOKEN", hd.url_tokens[0]);
	}

	LocoConfig config =  Loco.getConfig();
	hd.subs.put ("SYSTEM_TAG", config.getString ("SYSTEM_TAG"));
	hd.subs.put ("SYSTEM_INSTANCE_ID", 
		     config.getString ("SYSTEM_INSTANCE_ID"));
    }

    /** Inserts the current Locomotive Version.
     * the version tag is [LOCOMOTIVE_SERVER_VERSION]
     */
    public static void insertVersion(HandlerData hd)
    {
	hd.subs.put ("LOCOMOTIVE_SERVER_VERSION", Loco.getVersionString());
    }

    /** Inserts the current session id.
     * the session tag is [SESSIONID]
     */
    public static void insertSessionID(HandlerData hd)
    {
	hd.subs.put ("SESSIONID", String.valueOf (hd.session.getID()));
    }

    /** Inserts the current User's ID.
     * The User ID tag is [USERID]
     */
    public static void insertUserID(HandlerData hd)
    {
	if (! hd.session.isAnonymous())
	{
	    hd.subs.put ("USERID", String.valueOf (hd.user.getUserid()));
	}
    }

    /** Insters the current user's name, if the session is not anonymous.
     *  The user name tag is [USERNAME]
     */
    public static void insertUserName(HandlerData hd)
    {
	if (! hd.session.isAnonymous())
	{
	    hd.subs.put ("USERNAME", hd.user.getUsername());
	}
    }

    /** Inserts the handler url token, if there is one.
     *  The handler URL tag is [HANDLER_URLTOKEN]
     */
    public static void insertHandlerURL(HandlerData hd)
    {
	if (hd.url_tokens.length > 0)
	{
	    hd.subs.put ("HANDLER_URLTOKEN", hd.url_tokens[0]);
	}
    }

    /** Inserts the current date and time.
     *  the date tag is [DATE]; the time tag is [TIME]
     */
    public static void insertDateTime(HandlerData hd)
    {
	Date rightNow = new Date();
	hd.subs.put ("DATE", dateFormat.format(rightNow));
	hd.subs.put ("TIME", timeFormat.format(rightNow));
    }


    /** 
     * Inserts the request id for this handler.  The tage is
     * [REQUEST_ID]
     */
    public static void insertRequestID(HandlerData hd)
    {
	hd.subs.put("REQUEST_ID", String.valueOf(hd.getRequestID()));
    }


    /** 
     * Inserts two tags that unquely identify the Locomotive the
     * that is handling the request.  The tags are 
     * [SYSTEM_TAG] which represents the main Locomotive tag,
     * and [SYSTEM_INSTANCE_ID], which is each Locomotive's unique ID.
     * These tags are set in the Locomotive configuration files.
     */ 
    public static void insertSystemTags(HandlerData hd)
    {
	LocoConfig config =  Loco.getConfig();
	hd.subs.put ("SYSTEM_TAG", config.getString ("SYSTEM_TAG"));
	hd.subs.put ("SYSTEM_INSTANCE_ID", 
		     config.getString ("SYSTEM_INSTANCE_ID"));
    }	


    /** 
     * This method adds all the insertions implemented in the methods in this
     * class, for servlets. See the other methods for the names and types 
     * of these insertions.  It inserts the following: 
     * <BLOCKQUOTE><CODE>
     * SESSIONID, USERID, USERNAME, REQUEST_ID, LOCOMOTIVE_SERVER_VERSION,
     * SERVLET_API_VERSION, SYSTEM_TAG, SYSTEM_INSTANCE_ID, DATE, TIME, 
     * SERVLET_URL_TOKEN, REQUEST_URI, SERVER_INFO.
     * </CODE></BLOCKQUOTE>
     * @param req the ServletRequest to pull the info from.
     * @param servlet the Servlet to pull info from.
     * @param subs the subs to insert into.
     * @return the subs hashtable with the inserted values.
     */
    public static Hashtable insertAll (ServletRequest req, Hashtable subs)
    {
	LocoConfig config =  Loco.getConfig();
	Integer request_id = 
	    (Integer) req.getAttribute("locomotive.request_id");
	String [] url_tokens = 
	    (String []) req.getAttribute("locomotive.url_tokens");
	String request_uri = 
	    (String) req.getAttribute("locomotive.request_uri");
//  	ServletContext context = null;
//  	if (request_uri != null)
//  	{
//  	    subs.put("REQUEST_URI", request_uri);
//  	    context = Loco.getServletContext(request_uri);
//  	}
//  	else
//  	{
//  	    context = Loco.getServletContext(null);
//  	}

	if (req instanceof HttpServletRequest)
	{
	    HttpSession session = ((HttpServletRequest) req).getSession(true);
	    User user = null;
	    if (session != null)
	    {
		subs.put ("SESSIONID", String.valueOf (session.getId()));
		user = (User) session.getValue("locomotive.user");
		
		if (user != null)
		{
		    subs.put ("USERID", String.valueOf (user.getUserid()));
		    if (! ((Session) session).isAnonymous())
		    {
			subs.put ("USERNAME", user.getUsername());
		    }
		}
	    }
	}

	if (request_id != null) subs.put ("REQUEST_ID", request_id);

//  	if (context != null)
//  	{
//  //	    subs.put("SERVLET_API_VERSION", context.getMajorVersion() + 
//  //		     "." + context.getMinorVersion());

//  //	    subs.put("SERVER_INFO", context.getServerInfo());
//  	}

	subs.put ("LOCOMOTIVE_SERVER_VERSION", Loco.getVersionString());
	subs.put ("SYSTEM_TAG", config.getString ("SYSTEM_TAG"));
	subs.put ("SYSTEM_INSTANCE_ID", 
		     config.getString ("SYSTEM_INSTANCE_ID"));

	Date rightNow = new Date();
	subs.put ("DATE", dateFormat.format(rightNow));
	subs.put ("TIME", timeFormat.format(rightNow));

	if (url_tokens != null && url_tokens.length > 0)
	{
	    subs.put ("SERVLET_URL_TOKEN", url_tokens[0]);
	}

	return (subs);
    }

}
