/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import java.io.*;
import java.net.*;

/**
 * Java-based stoploco client.  Acts more or less exactly as the old c-based
 * client does.  For usage information, see the main method below.
 */
public class StopLoco
{
    /**
     * This attempts to stop a locomotive remotely.  It takes the following
     * arguments: <PRE>
     *   -h      print usage
     *   -p      port the locomotive is listening on.  Defaults to 5421.
     *   -a      host that the locomotive is operating on.  This can be either
     *           an ip address in quad-octet notation, or a host name.
     *           Defaults to localhost.
     * </PRE>
     */
    public static void main(String [] args)
    {
        int port = 5421;
        InetAddress addr = null;
        
        final int MAGIC_NUMBER = 0x09080706;
        
        try
        {
            for (int i = 0; i < args.length; i++)
            {
                if (args[i].equals("-h") || args[i].equals("--help")) 
                { 
                    usage(); System.exit(0); 
                }
                else if (args[i].equals("-p"))
                {
                    try
                    {
                        port = Integer.parseInt(args[++i]);
                    }
                    catch (Exception e)
                    {
                        System.out.println("malformed port number");
                        throw e;
                    }
                }
                else if (args[i].equals("-a"))
                {
                    try
                    {
                        addr = InetAddress.getByName(args[++i]);
                    }
                    catch (Exception e)
                    {
                        System.out.println("malformed host name or ip " +
                                           "address");
                        throw e;
                    }                        
                }
            }
            
            if (addr == null)
            {
                addr = InetAddress.getLocalHost();
            }

            System.out.println("Attempting shutdown of Locomotive at: " + 
                               addr.toString() + ", port " + port + "...");

            for (int i = 0; i < 2; i++)
            {
                Socket s = null;
                try
                {
                    s = new Socket(addr, port);
                }
                catch (IOException ioe)
                {
                    System.out.println("Locomotive is dead or not at " + 
                                       "location specified");
                    throw ioe;
                }
                
                s.setSoTimeout(10000);  // time out after ten seconds
                DataOutputStream out = 
                    new DataOutputStream(s.getOutputStream());
                try
                {
                    out.writeInt(MAGIC_NUMBER);
                }
                catch (IOException ioe)
                {
                    System.out.println("Message delivery failed");
                    throw ioe;
                }
                finally 
                { 
                    try
                    {
                        out.close();
                        s.close();
                    }
                    catch (Exception e)
                    {
                    }
                }
            }
        }
        catch (Exception e)
        {
            System.out.println(e);
            usage();
            System.exit(-1);
        }
    }

    public static void usage()
    {
        System.out.println("Usage: java org.locomotive.loco.StopLoco " +
                           "[-h | --help] [-p port] [-a hostname|ip address]");
    }
}
