/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco;

import java.io.*;
import java.util.*;
import java.sql.*;
import javax.servlet.http.*;

import org.locomotive.util.*;

import org.locomotive.loco.store.*;
import org.locomotive.loco.store.user.*;


/**
 * The User object encapsulates information stored for a user in the
 * Locomotive User Management system.  It is passed to modules as a
 * member of the the HandlerData Object.  The user can be "anonymous",
 * or can be a specific user whose content is persistantly stored in
 * the User Database.  Modules wishing to store additional
 * module-specific data based on users should use the userid field
 * as a key.  If the current session is non-anonymous (aka
 * "authenticated"), the User object will contain pre-loaded user
 * specific data.  Given that it will be loaded every hit, the User object
 * only contains the most important data, for performance reasons.
 */
public class User implements Serializable, HttpSessionBindingListener {
  // these are accessed through method interfaces
  private int userid = 0;
  private String username;
  private String password;
  private int loco_flags;

  // LOCO_FLAGS_SIZE reflects number of flags present
  public static final int LOCO_FLAGS_DEFAULT = 0;
  public static final int LOCO_FLAGS_CAN_SHUT_DOWN_LOCO = 1 << 0;
  public static final int LOCO_FLAGS_CAN_VIEW_SERVER_STATUS = 1 << 1;
  public static final int LOCO_FLAGS_CAN_CHANGE_LOCO_VARS = 1 << 2;

  public static final int LOCO_FLAGS_CAN_VIEW_USER_PREFS = 1 << 3;
  public static final int LOCO_FLAGS_CAN_EDIT_USER_PREFS = 1 << 4;

  public static final int NULL_USERID = -1;
  public static final int ANONYMOUS_ID = 0;
  public static final int MIN_ID = 1;

  public static final String ANONYMOUS_NAME = "anonymous";

  // used for caching and "quicksave"
  private boolean dirty = false;

  // Heads up here- for dynamic updating of Loco's configs, make sure Loco's
  // config object doesn't get reinitialized!
  private static LocoConfig config = null;


  //dw
  // This is the UserStore that provides all the persistance
  // call method createUserStore to fill this
  private UserStore userStore = null;

  private static UserStore staticUserStore = null;
  /**
   * initializes the config object.  Run this if the config object has been
   * reloaded.
   */
  protected static void initialize()
  {
    config = Loco.getConfig();
    staticUserStore = StoreFactory.createUserStore();
  }


  /**
   * create an anonymous User
   */
  public User ()
  {
    userid = 0;
    username = ANONYMOUS_NAME;
    password = "password";
    loco_flags = LOCO_FLAGS_DEFAULT;
    //dw
    createUserStore();
  }

  /**
   * loads persistant User data from the user database.
   * @param conn a JDBC connection with access to the Loco user database
   * @param uid  the userid of the user
   * @exception LoadUserException if there is a database access error
   * @exception UserNotFoundException if a user of the specified userid
   *                   cannot be found
   */
  public User (Connection conn, int uid) throws LoadUserException,
  UserNotFoundException
  {
    //dw
    createUserStore();
    if (! load (conn, uid)) {
      throw new UserNotFoundException ("Userid: " + uid +
                                       " does not exist");
    }
  }

  /**
   * loads persistant User data from the user database.
   * @param conn a JDBC connection with access to the Loco user database
   * @param un   the username of the user
   * @exception LoadUserException if there is a database access error
   * @exception UserNotFoundException if a user of the specified userid
   *                   cannot be found
   */
  public User (Connection conn, String un) throws LoadUserException,
  UserNotFoundException
  {
    //dw
    createUserStore();
    if (! load (conn, un, true)) {
      throw new UserNotFoundException ("Username: " + un
                                       + " does not exist");
    }
  }

  /**
   * loads persistant User data from the user database.
   * @param conn a JDBC connection with access to the Loco user database
   * @param un   the username of the user
   * @param sensitiveToCase whether or not to pay attention to case
   * @exception LoadUserException if there is a database access error
   * @exception UserNotFoundException if a user of the specified userid
   *                   cannot be found
   */
  public User (Connection conn, String un, boolean sensitiveToCase)
  throws LoadUserException, UserNotFoundException
  {
    //dw
    createUserStore();
    if (sensitiveToCase) {
      if (! load (conn, un, true)) {
        throw new UserNotFoundException ("Username: " + un
                                         + " does not exist");
      }
    } else {
      if (! load (conn, un, false)) {
        throw new UserNotFoundException ("Username: " + un
                                         + " does not exist in any"
                                         + " permutation of case.");
      }
    }
  }

  /**
   * creates User and persistantly stores User's data.  Will fail
   * if uid is not unique (uid <= 0 will cause system to pick a userid).
   * Will also fail if un (intended username) is already used.
   *
   * @param conn   a JDBC connection with access to the Loco user database
   * @param uid    the userid, system will pick if <= 0
   * @param un     username of the new user
   * @param pword  password of the new user
   * @param fn     firstname of the new user
   * @param ln     lastname of the new user
   * @param flg    flags of the new user
   * @param locoflg flags of the new user for Loco-specific functions
   * @exception LoadUserException if there is a database storage error
   * @deprecated thanks to schema changes for the loco_user table.
   *              fn, ln, flg are all now ignored.
   */
  public User (Connection conn, int uid, String un, String pword,
               String fn, String ln, int flg, int locoflg)
  throws LoadUserException
  {
    this(conn, uid, un, pword, locoflg);
  }


  /**
   * creates User and persistantly stores User's data.  Will fail
   * if uid is not unique (uid <= 0 will cause system to pick a userid).
   * Will also fail if un (intended username) is already used.
   *
   * @param conn   a JDBC connection with access to the Loco user database
   * @param uid    the userid, system will pick if <= 0
   * @param un     username of the new user
   * @param pword  password of the new user
   * @param locoflg flags of the new user for Loco-specific functions
   * @exception LoadUserException if there is a database storage error
   */
  public User(Connection conn, int uid, String un, String pword,
              int locoflg) throws LoadUserException
  {
    //dw
    createUserStore();

    if ((un == null) || (un.length() > 128) || un.equals(ANONYMOUS_NAME)) {
      throw new LoadUserException ("unacceptable username");
    }

    userid = userStore.createUser(conn, uid, un, pword, locoflg);
    username = un;
    password = pword;
    loco_flags = locoflg;
  }

  /**
   * used internally by loadUsers (Connection, int[]) to
   * create the heap of individual User objects.
   * @deprecated thanks to schema changed to the loco_users table
   *     fn, ln, flg are all now ignored.
   */
  private User (int uid, String un, String pword,
                String fn, String ln, int flg, int locoflg)
  {
    this(uid, un, pword, locoflg);
  }



  /**
   * used internally by loadUsers (Connection, int[]) to
   * create the heap of individual User objects.
	 * DW change visibility from private to public so can be used by UserStore
   */
  public User (int uid, String un, String pword, int locoflg)
  {
    userid = uid;
    username = un;
    password = pword;
    loco_flags = locoflg;
    //dw
    createUserStore();
  }

  //dw
  /**
   * get a new UserStore from the factory if one has not been found already
   * this needs to be called before any db stuff (obviously)
   */
  private void createUserStore() {
    if ( userStore == null ) {
      userStore = StoreFactory.createUserStore();
    }
  }



  /**
   * gets the userid of this User
   */
  public int getUserid()
  {
    return (userid);
  }

  /**
   * gets the userid of the Anonymous User
   */
  public int getAnonymousUserid()
  {
    return (ANONYMOUS_ID);
  }

  /**
   * gets username of this User
   */
  public String getUsername()
  {
    return (username);
  }

  /**
   * gets the username of the Anonymous User
   */
  public String getAnonymousUsername()
  {
    return (ANONYMOUS_NAME);
  }

  /**
   * returns the value of the User's password
   * Note that later this might be made package accessable only
   */
  public String getPassword()
  {
    return (password);
  }


  /**
   * gets the value for this User's Loco-specific flags
   */
  public int getLocoFlags()
  {
//System.out.println("############### getting flags : " + loco_flags);
    return (loco_flags);
  }


  /**
   * sets the loco_flags value for this User
   * @param newlocoflags  the new value for loco_flags
   */
  public void setLocoFlags (int newlocoflags)
  {
    dirty = true;
    loco_flags = newlocoflags;
  }

  /**
   * sets the username value for this User
   * Note: this may be made package friendly only in the future
   * @param un  the new username value
   */
  public void setUsername (String un)
  {
    dirty = true;
    username = un;
  }

  /**
   * sets the password value for this User
   * @param pw  the new value for the User's password
   */
  public void setPassword (String pw)
  {
    dirty = true;
    password = pw;
  }

  /**
   * tests to see if the stored password is equivalent to the input
   * @param pwtry   the password to compare for a match
   */
  public boolean hasPassword (String pwtry)
  {
    if ((password == null) || (pwtry == null)) {
      return (false);
    }

    return (pwtry.equals(password));
  }

  /**
   * tests to see if the stored password is equivalent to the input
   * in any case (i.e. lower(password) == lower(try))
   * @param pwtry   the password to compare for a match
   */
  public boolean hasPasswordAnyCase (String pwtry)
  {
    if ((password == null) || (pwtry == null)) {
      return (false);
    }

    return (pwtry.equalsIgnoreCase(password));
  }


  /**
   * checks whether this user is an anonymous user.
   */
  public boolean isAnonymous ()
  {
    return (userid == ANONYMOUS_ID);
  }


  public boolean canShutDownLoco()
  {
    return ((LOCO_FLAGS_CAN_SHUT_DOWN_LOCO & loco_flags) ==
            LOCO_FLAGS_CAN_SHUT_DOWN_LOCO);
  }

  public boolean canViewServerStatus()
  {
    return ((LOCO_FLAGS_CAN_VIEW_SERVER_STATUS & loco_flags) ==
            LOCO_FLAGS_CAN_VIEW_SERVER_STATUS);
  }

  public boolean canChangeLocoVars()
  {
    return ((LOCO_FLAGS_CAN_CHANGE_LOCO_VARS & loco_flags) ==
            LOCO_FLAGS_CAN_CHANGE_LOCO_VARS);
  }

  public boolean canViewUserPrefs()
  {
    return ((LOCO_FLAGS_CAN_VIEW_USER_PREFS & loco_flags) ==
            LOCO_FLAGS_CAN_VIEW_USER_PREFS);
  }

  public boolean canEditUserPrefs()
  {
    return ((LOCO_FLAGS_CAN_EDIT_USER_PREFS & loco_flags) ==
            LOCO_FLAGS_CAN_EDIT_USER_PREFS);
  }


  /** Returns the userid and username */
  public String toString()
  {
    return ("User: " + userid + " " + username);
  }


  /** overrides Object.equals for User specific checking */
  public boolean equals(Object o)
  {
    if (! (o instanceof User)) return (false);
    User u = (User) o;
    return ( u.getUserid() == userid &&
             u.getUsername().equals(username) &&
             u.getPassword().equals(password) &&
             u.getLocoFlags() == loco_flags );
  }


  /*-------------- HttpSessionBindingListener methods -------------------*/


  /**
   * If the name of the event is 'locomotive.user', this will go ahead
   * and associate this user with this session in the database.
   * We'll try to cast the HttpSession to a org.locomotive.loco.Session,
   * and if that fails, we'll do nothing.
   */
  public void valueBound(HttpSessionBindingEvent hsbe)
  {
    if (! hsbe.getName().equals("locomotive.user")) {
      return;
    }

    Connection conn = Loco.getDBConnection();
    try {
      Session sess = (Session) hsbe.getSession();
      Loco.log(9, "User: inserting user " + userid +
               " into session " + sess.getId());
      sess.associateUser(conn, this);
    } catch (ClassCastException cce) {
      Loco.log(0, "ERROR: couldn't cast HttpSession to Session: " +
               StringUtil.exceptionStackTrace(cce));
    } catch (FailedSessionException se) {
      Loco.log(0, "ERROR: couldn't bind user to session: " +
               StringUtil.exceptionStackTrace(se));
    } finally {
      Loco.putBackDBConnection(conn);
    }
  }


  /**
   * We don't really need to do anything here
   */
  public void valueUnbound(HttpSessionBindingEvent hsbe) {}


  /**
   * returns true if a user with this userid exists in the user database
   * @param conn   a JDBC connection with access to the Loco user database
   * @param id     the userid of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public static boolean exists (Connection conn, int id)
  throws LoadUserException
  {
    return staticUserStore.exists(conn, id);
  }

  /**
   * returns true if a user in the user database has the same name
   * @param conn   a JDBC connection with access to the Loco user database
   * @param un     the username of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public static boolean exists (Connection conn, String un)
  throws LoadUserException
  {
    return staticUserStore.exists(conn, un);
  }

  /**
   * returns true if a user in the user database has the same name,
   * regardless of case.  e.g. user "bob" exists, even though the
   * userbase only contains a user "Bob"
   * @param conn   a JDBC connection with access to the Loco user database
   * @param un     the username of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public static boolean existsAnyCase (Connection conn, String un)
  throws LoadUserException
  {
    return staticUserStore.existsAnyCase(conn, un);
  }

  /**
   * Updates the user.last_login field in the database.
   * silently ignored on anonymous users
   * @param conn an active JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public void saveLoginTime (Connection conn) throws LoadUserException
  {
    if (userid <= 0)
      return;

    userStore.saveLoginTime(conn, userid);
  }


  /**
   * Updates the User data in the database if changes have been made,
   * silently ignored on anonymous users
   * @param conn an active JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public void save (Connection conn) throws LoadUserException
  {
    if ((userid <= 0) || !dirty) {
      return;
    }

    if ((username == null) || (username.length() > 128)) {
      throw new LoadUserException ("unacceptable username value");
    }

    userStore.save(conn, userid, username, password, loco_flags);
  }

  /**
   * used by constructor above - loads User based on un from db
   */
  private boolean load (Connection conn, String un, boolean sensitiveToCase)
  throws LoadUserException
  {
    if ((un == null) || (un.length() > 128) || un.equals(ANONYMOUS_NAME)) {
      return (false);
    }


    LocoUserRow ur = userStore.load(conn, un, sensitiveToCase);
    if (ur != null) {
      userid = ur.getUserid();
      password = ur.getPassword();
      loco_flags = ur.getLoco_Flags();
      username = ur.getUsername();
    }
    return (ur != null);
  }

  /**
   * used by constructor above - loads User based on uid from db
   */
  private boolean load (Connection conn, int uid) throws LoadUserException
  {
    if (uid <= 0) {
      return (false);
    }

    LocoUserRow ur = userStore.load(conn, uid);
    if (ur != null) {
      userid = ur.getUserid();
      password = ur.getPassword();
      loco_flags = ur.getLoco_Flags();
      username = ur.getUsername();
    }
    return (ur != null);
  }

  /**
   * Searches the user database and returns a list of Users, based upon
   * userid.
   * @param conn  a JDBC connection
   * @param userids  a list of userids to retrieve
   * @exception LoadUserException if there is a database storage error
   */
  public static User[] loadUsers (Connection conn, int userids[])
  throws LoadUserException
  {
    if ((userids == null) || (userids.length < 1)) {
      return (new User[0]);
    }

    return staticUserStore.loadUsers(conn, userids);
  }

  /**
   * Returns an array of all the user names in the database.
   * @param conn  a JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public static String[] loadAllUserNames (Connection conn)
  throws LoadUserException
  {
    return staticUserStore.loadAllUserNames(conn);
  }

  /**
   * Given a vector of Integer-ized userids, look up each username for
   * them.
   */
  public static Vector getUsernames(Connection conn, Vector userids)
  throws SQLException
  {
    return staticUserStore.getUsernames(conn, userids);
  }

  /**
   * Given a userid, return the username
   */
  public static String getUsername(Connection conn, int userid)
  throws SQLException
  {
    String username = null;

    if (userid == ANONYMOUS_ID) return (ANONYMOUS_NAME);
    if (userid < MIN_ID) return (null);


    return staticUserStore.getUsername(conn, userid);
  }

  /**
   * Gets an integer appropriate for use as the next userid. <BR>
   * WARNING: Not supported for Informix databases.
   * @param conn  a JDBC connection connected to the user database
   * @exception LoadUserException if there is a database storage error
   *            or if database used is Informix.
   * @deprecated by ck- this call is now inlined in the constructor
   *             which used to call it.
   * NOW NOT MOVED INTO UserStore!!!!
   */
  static int getNextId (Connection conn) throws LoadUserException
  {
    throw new LoadUserException ("User.getNextID: Not supported by UserStore for any database");
  }

  /**
   * Generates a character sequence of length 8 suitable for a
   * password (though probably not easy to remember)
   *
   * @param seedmodifier seed used to assist random character generation
   */
  public static String generateNewPassword (int seedmodifier)
  {
    java.util.Date dt = new java.util.Date();
    Random rand = new Random(dt.getTime() + (long) seedmodifier);

    char pwc[] = new char[8];

    // use characters ascii 40 - 126
    for (int i = 0; i < pwc.length; i++) {
      pwc[i] = (char) ((char) 48 + (char) (rand.nextFloat() * 74.0));
      if ( ((pwc[i] > 57) && (pwc[i] < 65)) ||
           ((pwc[i] > 90) && (pwc[i] < 97))) {
        i--;
      }
    }
    return (new String (pwc));
  }

}
