/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.handler;

import org.locomotive.loco.*;
import org.locomotive.server.*;
import org.locomotive.util.*;
import java.util.*;
import java.sql.*;


/**
 * The GenericHandler class is a generic implementation of the Handler
 * interface.
 * There are two alternative ways to use this class:
 * <ol>
 * <li>You can subclass this class to handle all your user requests.
 * <li>You can create several different subclasses of this class in order
 * to handle different features.  Then, you can create a "master" subclass
 * to route to each of these different subclasses, based on the different URLs.
 * </ol>
 *
 * This class comes with many convenience methods, some of which serve only
 * to make the code more readable; feel free to customize them.
 */
public class GenericHandler implements Handler
{
    /** The handlerData object containing information for this request. */
    protected HandlerData hd;
    /** the url tokens for this request. */
    protected String [] url_tokens;
    /** The last part of the url which specifies which Handler to use */
    protected String handler_uri;
    /** the FormData object that contains any form information from a POST. */
    protected FormData form_data;
    /** the substitutions hashtable to be used when evaluating templates. */
    protected Hashtable steam_vars;
    /** the user associated with this request. */
    protected User user;
    /** the session associated with this request. */
    protected Session session;
    /** The database connection this request can use, if desired. */
    protected Connection conn;
    /** whether or not a page has been flushed back to the browser. */
    protected boolean already_displayed_page;
    
    
    /** Log level to use for serious errors; e.g. log(ERROR, "Wait! There's an 
     * error!"); */
    public static final int ERROR = 0;
    /** Log level to use for warnings */
    public static final int WARNING = 1;
    /** log level to use for normal activities */
    public static final int NORMAL = 6;
    /** log level to use for debugging */
    public static final int DEBUG = 9;
    
    
    
    /**
     * initializes all the local variables.  Whenever you subclass a
     * GenericHandler, be sure to call super.handleRequest(hd) in the first
     * line of your handleRequest() method.
     */
    public void handleRequest (HandlerData hd)
    {
	this.hd = hd;
	url_tokens = hd.url_tokens;
	handler_uri = hd.handler_uri;
	form_data = hd.form_data;
	steam_vars = hd.subs;
	user = hd.user;
	session = hd.session;
	conn = hd.conn;
	already_displayed_page = false;
    }
    

    /**
     * GenericHandler doesn't do anything in the init() method.
     */
    public boolean init (Log server_log)
    {
	return true;
    }
    
    
    /**
     * GenericHandler doesn't do anything in the shutdown() method, either.
     */
    public boolean shutdown (Log server_log)
    {
	return true;
    }
    
    
    /**
     * writes a message to the locomotive's global server log.
     * @@param log_level -- 0 is most severe.  The higher, the less severe.
     * Messages with log_levels higher than the variable LOG_LEVEL (from
     * the file loco.conf) will be suppressed from being written to the log.
     * @@param log_message -- the message to write to the log.
     */
    public void log (int log_level, String log_message)
    {
	hd.server_log.log (log_level, log_message);
    }
    
    
    /**
     * this is a convenience method you can use when you catch
     * serious Exceptions for which you want a stack trace and 
     * you need to send the user an error code, which can help you debug.
     */
    public void handleException (Exception exception, String short_code)
    {
	hd.server_log.log (0, "ERROR: " + short_code + "." +
			   StringUtil.exceptionStackTrace (exception));
	Loco.sendErrorMsg (short_code, hd);
	already_displayed_page = true;
    }
    
    
    /**
     * write one event to the global locomotive event log.
     * (this method is just a wrapper around Loco.logEvent().)
     * @@param major - the major event region or code section
     * @@param minor - the minor event code
     * @@param data  - event specific data
     * example:
     * <BLOCKQUOTE>
     * logEvent (5, "login", "submit_form", user.getUsername());
     * </BLOCKQUOTE>
     */
    public void logEvent (String major, String minor, String data)
    {
	Loco.logEvent (5, hd, major, minor, data);
    }
    

    /**
     * if you want to send back a single string as the response to the user,
     * you can use this method.  (You don't need to put your message within
     * &lt;HTML&gt;&lt;BODY&gt;...&lt;/BODY&gt;&lt;/HTML&gt;)
     */
    public void displayString (String message)
    {
	if (message == null || message.equals(""))
	{
	    Loco.sendErrorMsg ("no message!", hd);
	    already_displayed_page = true;
	    return;
	}
	
	if (message.indexOf ("<HTML>") != -1)
	{
	    hd.resp.addString (message);
	    hd.resp.flush();
	    already_displayed_page = true;
	}
	else
	{
	    hd.resp.addString ("<HTML><BODY>" + message + "</BODY></HTML>");
	    hd.resp.flush();
	    already_displayed_page = true;
	}
	
    }
    
    
    
    /**
     * loads a Steam document, inserts the values of Steam variables into that
     * document, and then writes the result back to the Web server.
     * @@param document_path -- the file path of that document, relative to the
     * template root.
     * @@returns a String containing the fully evaluated Steam
     * document, after inserting all the variables.  If the document
     * with that file path does not exist, then returns "", an empty String.
      * If another page has already been displayed, will return "".
     */
    public String displayPage (String document_path, Hashtable steam_vars)
    {
	if (already_displayed_page)
	{
	    Loco.log (0, "ERROR: can't call GenericHandler.displayPage() twice!");
	    return "";
	}
	
	String result = Loco.evaluateTemplate (document_path, steam_vars);
	if (result.equals(""))
	{
	    Loco.log (0, "ERROR: GenericHandler.displayPage(" + document_path +
		     ") was empty!");
	}
	else
	{
	    already_displayed_page = true;
	    hd.resp.addString (result);
	    hd.resp.flush();
	}
	
	return result;
    }
    
    /**
     * simply tells the user that they typed a bad URL, by sending the message:
     * "I'm sorry, I don't understand that URL."
     */
    public void sayBadURL()
    {
	hd.resp.addString ("<HTML><BODY>\n" +
			   "I'm sorry, I don't understand that URL.\n" +
			   "</BODY></HTML>");
	hd.resp.flush();
	already_displayed_page = true;
    }
    
}

