/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.handler;

import java.util.*;
import java.io.*;
import java.net.*;
import java.sql.*;

import org.locomotive.loco.*;
import org.locomotive.server.*;

/**
 * A HandlerData object is passed to all module Handlers when they are invoked.
 * This object is the primary interface to the server facilities.  
 * New instances are created by the Locomotive, on a per request basis.  
 */
public class HandlerData 
{
    private int request_id = 0;
    private StringBuffer debug_strbuf;
    private String debug_str;
    
    /**
     * The socket connected to the web server/locolink CGI.  
     * @see java.net.Socket
     */
    public Socket sock;
    
    /**
     * The input stream associated with sock.  Currently implemented with 
     * java.io.BufferedInputStream.  
     * @see java.io.BufferedInputStream
     */
    public InputStream is;

    /** 
     * A java.io.OutputStream, associated with Socket sock.  
     */
    public OutputStream os;

    /**
     * A Loco response.  Used to send the text of a response back to the client.
     * The RequestManager will have already set an association with 
     * java.net.Socket sock using java.io.BufferedInputStream is
     */
    public Response resp;
    
    /**
     * A JDBC Connection, cleared, opened, connected, and 
     * associated with the Loco's database namespace.  Used to store and 
     * retrieve persistant data.  
     */
    public Connection conn;
    
    /**
     * A hashtable containing the name/value pairs passed into the Loco 
     * by the web server.  Some possible names might be HTTP_REFERRER, 
     * CONTENT_TYPE, CONTENT_LENGTH, HTTP_COOKIE.
     */
    public Hashtable cgi_envars_hash;
    
    /**
     * A hash of HTML form input variables.  Normally null.
     * Will be non-null only when the current
     * request was generated by an HTTP "POST" from the remote 
     * HTTP client (browser).  Subsequently, in any HTML form, the 
     * "REQUEST_METHOD" should be explicitly set to "POST". <BR>
     * HTTP 'GET' for forms may eventually be supported
     */
    public Hashtable form_hash;
    
    /**
     * A convenient class which allows you to access and error-check 
     * the data in an HTML form that a user submits.
     */
    public FormData form_data;

    /**
     * This Hashtable contains all the names and values of variables to be
     * inserted into a Steam document.  (This variable is called "steam_vars"
     * in the GenericRequest class)
     */
    public Hashtable subs;
    
    /**
     * an ordered list of the url tokens that triggered the current request.
     */
    public String[] url_tokens;

    /**
     * The part of the URI that is after the Locomotive indicator token.
     * This string is effectivly the contents of url_tokens, but as it 
     * actually appears in the browser.  <BR>
     * Examples: <BR>
     * If the uri is /cgi-bin/locolink/foo?bar <BR>
     * handler_uri is /foo?bar <BR>
     * if the uri is /cgi-bin/locolink/foo/bar?baz <BR>
     * handler_uri is /foo/bar?baz <BR>
     * or, in the case of using the NSAPI tunnel, the uri might look like <BR>
     * /locolink/handerfoo?bar <BR>
     * handler_uri is then /handlerfoo?bar <P>
     * Previously called &quot;url&quot;, which was somewhat misleading.  
     */
    public String handler_uri;

    /**
     * The current session
     * @see org.locomotive.loco.Session
     */
    public Session session;

    /**
     * The current simple user
     * @see org.locomotive.loco.User
     */
    public User user;
    
    /** 
     * a numerical id used to track browsers.  Will be set to zero 
     * if a stamp is not available.  Perhaps could be used for logging
     * purposes.
     */
    public long bstamp_id;
    
    /**
     * a org.locomotive.server.Log referencing the Loco server log
     * @see org.locomotive.server.Log
     */
    public Log server_log;
    
    /**
     * a org.locomotive.server.Log referencing the Loco event log.  
     * Typically one would make one call to the Event log per request.  
     * The logfile format is tab-delimited, and can easily be sucked back 
     * into a database for processing, or processed as is with a Perl script. 
     * @see org.locomotive.server.Log
     */
    public Log event_log;

    /**
     * a org.locomotive.loco.LocoConfig referencing the default Loco config object
     * @see org.locomotive.loco.LocoConfig
     */
    public Config cfg;

    /**
     * the properties file associated with the handler this is being sent
     * to, if there is one.  The properties file path is set in the 
     * hrt.conf file.  If there is no properties file set, then this
     * will be null.  When this is not null, it will always have at least
     * one property: <PRE>
     * HANDLER_PROPERTIES_PATH </PRE> 
     * which is the absolute path to the properties file.  This
     * allows handlers to manipulate and re-store their properties.
     */
    public Properties handler_config;


    /**
     * HandlerDataconstructor takes the request id whose fields it contains
     */
    public HandlerData (int r_id)
    {
	request_id = r_id;
    }

    /**
     * gets the id of this request (for logging purposes)
     */
    public int getRequestID()
    {
	return (request_id);
    }

    /**
     * returns a string which contains debugging information about the data
     * contained in this instance of HandlerData.
     */
    public String toString()
    {
	debug_strbuf = new StringBuffer();
    
	debug_strbuf.append("HandlerData ID: " + request_id +
			    ", contains the following information:\n");
	debug_strbuf.append ("---------------------------------------\n\n");

	debug_strbuf.append ("Session ID: " + session.getID() + "\n");

	if (sock != null)
	{
		debug_strbuf.append ("Socket Info:    " +
				     sock.toString() + "\n");
	}

	if (conn != null)
	{
		debug_strbuf.append ("Connection Info:\n    ");
		try
		{
		    debug_strbuf.append ("Current catalog name: " + 
					 conn.getCatalog() + "\n");
		} catch (java.sql.SQLException e) {}

//		debug_strbuf.append ("SQL MetaData: " + conn.getMetaData() + "\n");
//		debug_strbuf.append ("Current Transaction Isolation Mode: "
//			+ conn.getTransactionIsolation() + "\n");
	}

	if (cgi_envars_hash != null)
	{
		debug_strbuf.append ("cgi_envars_hash Info:\n    " +
				     cgi_envars_hash.toString() + "\n");
	}

	if (form_hash != null)
	{
		debug_strbuf.append ("form_hash Info:\n    " +
				     form_hash.toString() + "\n");
	}

	if (subs != null)
	{
		debug_strbuf.append ("subs Info:\n    " +
				     subs.toString() + "\n");
	}

/*	if (url_tokens != null)
	{
		debug_strbuf.append ("url_tokens Info:\n    ");

		int i=0;
		while (i<= url_tokens.length)
		{
			debug_strbuf.append (url_tokens[i] + "\n");
			i = i + 1;
		}
	}
*/

	debug_strbuf.append ("bstamp_id Info:\n    ");
	debug_strbuf.append (String.valueOf(bstamp_id) + "\n");

	if (user != null)
	{
		debug_strbuf.append ("USER INFO:\n--------------\n");
		debug_strbuf.append ("Username: " + user.getUsername() + "\n");
		try
		{
		    debug_strbuf.append ("Exists? : " + 
					 user.exists(conn, user.getUserid()) + 
					 "\n");
		}
		catch (org.locomotive.loco.LoadUserException lue)
		{
			debug_strbuf.append ("Exists? : " + 
					     lue.toString() + "\n");
		}

		// this could be a security hole if we're not careful,
		// so Ben has decided not to put it in.

		//  debug_strbuf.append ("Password: " + user.getPassword() + "\n");
	}

/*
		if (socket.getInetAddress() != null)
		{
			debug_strbuf.append ("INetAddress: " +
				socket.getInetAddress().toString() + "\n");
			debug_strbuf.append ("Local Address: " +
				socket.getLocalAddress().toString +
				" at port: " + socket.getLocalPort() + "\n");
		}
*/
	debug_str = new String(debug_strbuf);
	return debug_str;
    }

    /**
     * returns a string in HTML format which contains debugging information
     * about the data contained in this instance of HandlerData, so we can
     * print it out on a web browse.
     */
    public String toHTMLText()
    {
	debug_strbuf = new StringBuffer();
    
	debug_strbuf.append("HandlerData ID <B>" + request_id +
		"</B> item contains the following information:</H1><HR>");

	debug_strbuf.append ("Session ID: " + session.getID());

	if (sock != null)
	{
		debug_strbuf.append ("<BR><H2>Socket Info:</H2><HR>" +
		sock.toString());
	}

	if (conn != null)
	{
		debug_strbuf.append ("<BR><H2>Connection Info:</H2><HR>");
		try
		{
		debug_strbuf.append ("Current catalog name: " + conn.getCatalog());
		} catch (java.sql.SQLException e) {}

//		debug_strbuf.append ("<BR>SQL MetaData: " + conn.getMetaData());
//		debug_strbuf.append ("<BR>Current Transaction Isolation Mode: "
//			+ conn.getTransactionIsolation());
	}

	if (cgi_envars_hash != null)
	{
		debug_strbuf.append ("<BR><H2>cgi_envars_hash Info:</H2><HR>");
		debug_strbuf.append (cgi_envars_hash.toString());
	}

	if (form_hash != null)
	{
		debug_strbuf.append ("<BR><H2>form_hash Info:</H2><HR>");
		debug_strbuf.append (form_hash.toString());
	}

	if (subs != null)
	{
		debug_strbuf.append ("<BR><H2>subs Info:</H2><HR>");
		debug_strbuf.append (subs.toString());
	}

/*	if (url_tokens != null)
	{
		debug_strbuf.append ("<BR><H2>url_tokens Info:</H2><HR>");

		int i=0;
		while (i<= url_tokens.length)
		{
			debug_strbuf.append (url_tokens[i]);
			i = i + 1;
		}
	}
*/

	debug_strbuf.append ("<BR><H2>bstamp_id Info:</H2><HR>");
	debug_strbuf.append (String.valueOf(bstamp_id));

	if (user != null)
	{
	    debug_strbuf.append ("<BR><H2>user Info:</H2><HR>");
	    debug_strbuf.append ("Username: " + user.getUsername());
//		debug_strbuf.append ("<BR>Loco Flags: " + user.getLocoFlags());
	    try
	    {
		debug_strbuf.append ("<BR>Exists? : " + 
				     user.exists(conn, user.getUserid()));
	    }
	    catch (org.locomotive.loco.LoadUserException lue)
	    {
		debug_strbuf.append ("<BR>Exists? : " + lue.toString());
	    }
	    
	    // this could be a security hole if we're not careful,
	    // so Ben has decided not to put it in.
	    
	    // debug_strbuf.append ("<BR>Password: " + user.getPassword());
	}
	
/*
	if (socket.getInetAddress() != null)
	{
	    debug_strbuf.append ("INetAddress: " +
				 socket.getInetAddress().toString());
	    debug_strbuf.append ("<br>Local Address: " +
				 socket.getLocalAddress().toString +
				 " at port: " + socket.getLocalPort());
	}
*/
	debug_str = new String(debug_strbuf);
	return debug_str;
    }

    Hashtable attributes;
    
    /**
     * Get the value of attributes.
     *
     * @return Value of attributes.
     */
    public Hashtable getAttributes() {return attributes;}
    
    /**
     * Set the value of attributes.
     *
     * @param v  Value to assign to attributes.
     */
    public void setAttributes(Hashtable  v) {this.attributes = v;}
    
}
