/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.handler;

import org.locomotive.loco.*;
import org.locomotive.server.*;
import org.locomotive.util.*;

import java.util.*;
import java.io.*;
import java.net.*;
import java.sql.*;

/**
 * Used to choose which handler classes get invoked.  Used by the 
 * RequestManager.  Has some preset defaults- see putInDefaults
 */
public class HandlerRoutingTable extends ServiceRoutingTable
{

    protected Log server_log;

    /**
     * creates a new (but uninitialized) HanderManager
     * @param path the path to the table file
     * @param log  a server log
     */
    public HandlerRoutingTable (String path, Log log, Vector classpath, 
				boolean auto_reload)
    {
	super(path, log, classpath, auto_reload);
	this.server_log = log;
    }
 

    /**
     * Adds the a HandlerEntry to the service_hash.
     */
    protected ServiceEntry createServiceEntry(String pattern, String classname, 
					      String config_path)
    {
	return ( new HandlerEntry(pattern, classname, config_path) );
    }
	


    /** Overrides the superclass to put in the default handlers.
     * here's what we put in: <PRE>
     * default      org.locomotive.loco.handler.DefaultHandler
     * unknown      org.locomotive.loco.handler.UnknownHandler
     * srv          org.locomotive.loco.ServerHandler
     * </PRE>
     */
    protected void putInDefaults() throws InitializationException
    {
	// store the ever-present entries 

	ServiceEntry se;
	boolean status_ok = true;

	// allow external file to override default handler, 
	// but install the default if one is missing.
	if (! containsService ("default"))
	{
	    addService( "default", "org.locomotive.loco.handler.DefaultHandler", 
			null);
	}

	// The handler to be called if no other one can be found
	// may also be overridden in the config file
	if (! containsService ("unknown"))
	{
	    addService("unknown", "org.locomotive.loco.handler.UnknownHandler", null);
	}

	// always install the server handler.
	addService("srv", "org.locomotive.loco.ServerHandler", null);
    }



    /** 
     * This gets called on every request by the loco, once it
     * determines it has a handler in its hands.  It sets up everything
     * the handler needs, then sends it on its way.
     * 
     * @param request_id the unique id for this request
     * @param sock the socket this request connected on
     * @param is the inputstream from that socket
     * @param os the outputstream from that socket
     * @param reply_protocol the protocol type we received the request from
     * @param cgi_envars the CGI variables passed in with the request
     * @param form_hash the form data, if there was any
     * @param url_tokens a tokenized form of the url
     * @param request_uri the unparsed url tokens
     * @param bstamp_id the browser stamp id that came in from the
     *        bstamp cookie for this request
     * @param conn the Database connection for this request
     * @param session the Session object for this request
     * @param user the user for this request
     */
    public void routeRequest(int request_id,
			     Socket sock, 
			     InputStream is, 
			     OutputStream os,
			     Hashtable cgi_envars,
			     Hashtable form_hash,
			     Hashtable steam_vars,
			     String [] url_tokens,
			     String request_uri, 
			     long bstamp_id,
			     Response resp,
			     Connection conn, 
			     Session session,
                             Hashtable attributes)
    {
	HandlerData hd = new HandlerData (request_id);

	hd.resp = resp;
	resp.setContentType("text/html");  // handler default
	hd.sock = sock;
	hd.is = is;
	hd.os = os;
	hd.conn = conn;
	hd.form_hash = form_hash;
	hd.cgi_envars_hash = cgi_envars;
        hd.setAttributes(attributes);
	    
	// we expect this to be either a GET or POST request
	// if a POST, then process the form waiting on stdin
	String http_method = (String) cgi_envars.get("REQUEST_METHOD");
	if (form_hash == null && http_method != null && 
	    http_method.equals ("POST") )
	{
	    Loco.log (2, "found POST: generating form_hash...");
		
	    try 
	    {
		FormParser f = new FormParser(is, cgi_envars, 
					      Loco.getConfig(), 
					      Loco.getServerLog());
		hd.form_hash = f.parseForm();
	    }
	    catch (ParseFormException pfe)
	    {
		Loco.log (0, "ERROR: form parsing error: " + pfe);
		Loco.sendErrorMsg("Could not parse form: " + 
				  pfe.getMessage(), hd);
		return;
	    }		
	}
	    
	hd.form_data = new FormData(hd.form_hash);
	hd.url_tokens = url_tokens;
	hd.handler_uri = request_uri;
	hd.subs = steam_vars;
	    
	hd.session = session;
	hd.user = (User) session.getValue("locomotive.user");
	hd.cfg = Loco.getConfig();
	hd.server_log = Loco.getServerLog();
	hd.event_log = Loco.getEventLog();
	hd.bstamp_id = bstamp_id;
	    
	String s = "";
	Class c = null;
	if (hd.url_tokens != null && hd.url_tokens.length > 0 && 
	    hd.url_tokens[0] != null)
	{
	    s = hd.url_tokens[0];
	}
	    
	c = routeToClass(Loco.getConfig(), s);
	    
	if (c == null)
	{
	    Loco.log(0, "ERROR: HandlerRoutingTable.routeRequest: " +
		     " Unknown handler not found; could " +
		     "not route to class " + s);
		
	    Loco.sendErrorMsg("Unknown Server Function: " + s, hd);
	}
	else
	{
	    // get the properties and put them in hd
	    // this will occur for both original routes and redirects
	    // this will be null if there wasn't hany props for this handler
	    hd.handler_config = getProperties(c);
		
	    try
	    {
		Handler h = (Handler) c.newInstance();
		h.handleRequest (hd);
	    }
	    catch (Throwable e)
	    {
		Loco.log(0, "ERROR: handoff(): " + c.toString() + ": " +
			 StringUtil.exceptionStackTrace(e));
		Loco.sendErrorMsg("", hd);
	    }
	}
	    
	// now clean up... 
	
	// if the handler connection isn't the one gave them, let's put
	// that back in the pool too.  The normal conn object is put
	// back in the RequestManager
	if (hd != null && hd.conn != null && hd.conn != conn)
	{
	    Loco.putBackDBConnection(hd.conn);
	}
    }




    // Boy is this even uglier 
    protected class HandlerEntry extends ServiceRoutingTable.ServiceEntry
    {

	HandlerEntry (String pat, String cname, String c_path)
	{
	    super(pat, cname, c_path);
	}

	protected void initialize() throws InitializationException
	{
	    try
	    {
		Loco.log(5, theclass.getName() + ": init");
		Handler h = (Handler) theclass.newInstance();
		if (h.init (server_log))
		{
		    cached = true;
		}
		else
		{
		    throw new InitializationException ( 
			"ERROR: HandlerRoutingTable: init failed for " +
			theclass.getName() + ". ");
		}
	    }
	    catch (NoClassDefFoundError ncdfe) { throwInitException(ncdfe); }
	    catch (ClassCastException cce) { throwInitException(cce); }
	    catch (InstantiationException ie) { throwInitException(ie); }
	    catch (IllegalAccessException iae) { throwInitException(iae); }
	}

	private void throwInitException(Throwable t) 
	    throws InitializationException
	{
	    props = null;
	    throw new InitializationException (
		"ERROR: HandlerRoutingTable: classload failed for " + 
		theclass.toString() + ": " + t.toString());
	}

	/**
	 * instantiates the handler for this hrte and calls its 
	 * shutdown method
	 */
	public boolean shutdown(Log server_log)
	{
	    try
	    {
		Handler h = (Handler) theclass.newInstance();
		if (h.shutdown (server_log))
		{
		    return (true);
		}
		else
		{
		    server_log.log(0, "ERROR: HandlerRoutingTable: " +
				   "shutdown for " + pattern + " failed!");
		    return (false);
		}
	    }
	    catch (InstantiationException ie)
	    {
		server_log.log(0, "ERROR: HandlerRoutingTable: shutdown for " + 
			       pattern + " failed: " + ie.toString());
		return (false);
	    }
	    catch (IllegalAccessException iae)
	    {
		server_log.log(0, "ERROR: HandlerRoutingTable: shutdown for " + 
			       pattern + " failed: " + iae.toString());
		return (false);
	    }

	}


    }

}
