/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import org.locomotive.util.*;
import org.locomotive.loco.*;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Maintains all database transactions dealing with acls.
 */
public class ACLManager implements Serializable
{
    // Constant used to represent when an ACL has no users associated with
    // a permission in the database.
    public static final int NO_USERS = -1;

    // Public methods

    /**
     * Adds an ACL to the database.
     *
     * @param acl  ACL to be added to the database
     * @param conn database connection
     *
     * @exception SQLException             when a database error occurs
     * @exception LoadUserException        when a user cannot be loaded from 
     *                                     the database
     * @exception UserNotFoundException    when a user cannot be found in the
     *                                     database
     */
    public static void addACL(AccessControlList acl, Connection conn)
	throws SQLException, LoadUserException, UserNotFoundException
    {
	Enumeration e = (acl.getPermissions()).elements();
	while(e.hasMoreElements())
	{
	    String perm = (String)e.nextElement();
	    Vector userIds = ACLManager.convertToUserIds(
		acl.getUsers(perm), conn);
	    Vector groupIds = GroupManager.
		convertToGroupIds(acl.getGroups(perm), conn);
	    boolean pol = acl.getPolarity(perm);
	    
	    ACLManager.addEntry(acl.getId(),
				perm,
				userIds,
				groupIds,
				pol,
				conn);
	}
    }

    /**
     * Removes an ACL from the database.
     *
     * @param id   id of ACL to be removed
     * @param conn database connection
     *
     * @exception SQLException when a database error occurs
     */
    public static void removeACL(int id, Connection conn)
	throws SQLException
    {	
	if(!ACLManager.exists(id, conn))
	{
	    return;
	}

	Statement stmt = null;

	try
	{ 
	    String sql_statement = "delete from LOCO_ROLES " +
		"where Objectid = " + id;
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Updates an ACL in the database.
     *
     * @param updatedACL updated ACL
     * @param conn       database connection
     *
     * @exception SQLException             when a database error occurs
     * @exception LoadUserException        when a user cannot be loaded from
     *                                     the database
     * @exception UserNotFoundException    when a user cannot be found in the
     *                                     database
     */
    public static void updateACL(AccessControlList updatedACL, Connection conn)
	throws SQLException, LoadUserException, UserNotFoundException
    {
	AccessControlList acl = ACLManager.getACL(updatedACL.getId(), conn);

	if (acl == null)
	{
	    addACL(updatedACL, conn);
	    return;
	}

	Vector updatedPerms = updatedACL.getPermissions();

	Enumeration oldPerms = (acl.getPermissions()).elements();
	while(oldPerms.hasMoreElements())
	{
	    String perm = (String)oldPerms.nextElement();
	    if (!updatedPerms.contains(perm))
	    {
		// Delete no longer existing permissions from DB
		ACLManager.removeEntry(acl.getId(), perm, conn);
	    }
	    else
	    {
		// Update polarity
		if( acl.getPolarity(perm) != updatedACL.getPolarity(perm) )
		{
		    ACLManager.updatePolarity(acl.getId(), updatedACL.
					      getPolarity(perm), perm,
					      conn);
		}
		 
		// Remove no longer existing users
		Enumeration oldUsers = (acl.getUsers(perm)).elements();
		Vector updatedUsers = updatedACL.getUsers(perm);
		while (oldUsers.hasMoreElements())
		{
		    String oldUser = (String)oldUsers.nextElement();
		    if (!updatedUsers.contains(oldUser))
		    {
			ACLManager.removeUser(acl.getId(), oldUser,
					      perm, conn);
		    }
		}

		// Add new users
		Enumeration newUsers = updatedUsers.elements();
		Vector toAddUsers = new Vector();
		while (newUsers.hasMoreElements())
		{
		    String newUser = (String)newUsers.nextElement();
		    if (!(acl.getUsers(perm)).contains(newUser))
		    {
			toAddUsers.addElement(newUser);
		    }
		}
		if (!toAddUsers.isEmpty())
		{
		    ACLManager.addEntry(acl.getId(),
					perm,
					ACLManager.convertToUserIds(toAddUsers,
								    conn),
					null,
					updatedACL.getPolarity(perm),
					conn);
		}

		// Remove no longer existing groups
		Enumeration oldGroups = (acl.getGroups(perm)).elements();
		Vector updatedGroups = updatedACL.getGroups(perm);
		while (oldGroups.hasMoreElements())
		{
		    String oldGroup = (String)oldGroups.nextElement();
		    if (!updatedGroups.contains(oldGroup))
		    {
			ACLManager.removeGroup(acl.getId(), oldGroup,
					       perm, conn);
		    }
		}

		// Add new groups
		Enumeration newGroups = updatedGroups.elements();
		Vector toAddGroups = new Vector();
		while (newGroups.hasMoreElements())
		{
		    String newGroup = (String)newGroups.nextElement();
		    if (!(acl.getGroups(perm)).contains(newGroup))
		    {
			toAddGroups.addElement(newGroup);
		    }
		}
		if (!toAddGroups.isEmpty())
		{
		    ACLManager.addEntry(acl.getId(),
					perm,
					null,
					GroupManager.convertToGroupIds(
					    toAddGroups,
					    conn),
					updatedACL.getPolarity(perm),
					conn);
		}
	    }
	}

	// Add new permissions
	Enumeration newPerms = updatedPerms.elements();
	while(newPerms.hasMoreElements())
	{
	    String perm = (String)newPerms.nextElement();
	    if (!(acl.getPermissions()).contains(perm))
	    {
		ACLManager.addEntry(acl.getId(),
				    perm,
				    ACLManager.convertToUserIds(
					updatedACL.getUsers(perm), conn),
				    GroupManager.convertToGroupIds(
					updatedACL.getGroups(perm), conn),
				    updatedACL.getPolarity(perm),
				    conn);
	    }
	}
	
    }

    /**
     * Retrieves an ACL from the database.
     *
     * @param id   id of ACL
     * @param conn database connection
     *
     * @return AccessControlList object (null if ACL does not exist)
     *
     * @exception SQLException           when a database error occurs
     * @exception LoadUserException      when a user cannot be loaded from the 
     *                                   database
     * @exception UserNotFoundException  when a user cannot be found in the
     *                                   database
     */
    public static AccessControlList getACL(int id, Connection conn)
	throws SQLException, UserNotFoundException, LoadUserException
    {
	AccessControlList acl = new AccessControlList(id);

	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    // Get the permissions
	    String sql_statement = "select Permission from " +
		"loco_access_control_list " +
		"where Objectid = " + id;
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);

	    if (!rs.next())
	    {
		return (null);
	    }
	    else
	    {
		String perm = rs.getString("Permission");
		if (PermissionSet.exists(perm)) {
		    acl.addPermission(perm);
		}
	    }
	    
	    while(rs.next())
	    {
		String perm = rs.getString("Permission");
		if (PermissionSet.exists(perm)) {
		    acl.addPermission(perm);
		}
	    }
	    
	    Enumeration e = (acl.getPermissions()).elements();
	    while(e.hasMoreElements())
	    {
		String perm = (String)e.nextElement();
		acl = ACLManager.loadEntry(acl, perm, conn);
	    }

	    return(acl);
	}
	catch (SQLException se)
	{ 
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}

    }

    /** 
     * Checks the existence of an ACL in the database.
     *
     * @param id   id of the ACL in question
     * @param conn database connection
     *
     * @return <code>true</code> if an ACL associated with the id exists
     *
     * @exception SQLException when a database error occurs
     */
    public static boolean exists(int id, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    String sql_statement = "select * from " +
		"loco_access_control_list " +
		"where Objectid = " + id;

	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);

	    if (!rs.next())
	    {
		return(false);
	    }

	    return (true);
	}
	catch (SQLException se)
	{ 
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}

    }

    /**
     * Returns a boolean indicating whether a particular user, identified
     * by the userid, is allowed the permission in question (indicated by
     * its name) in the ACL specified by the id.
     *
     * @param id     id of ACL
     * @param perm   name of permission
     * @param userid id of user
     * @param conn   database connection
     *
     * @return <code>true</code> if the user has the permission in the ACL
     *
     * @exception SQLException when a database error occurs
     */
    public static boolean permit(int id, String perm, int userid, 
				 Connection conn)
	throws SQLException
    {
	if(!ACLManager.exists(id, conn))
	{
	    return(true);
	}
	
	Statement stmt1 = null;
	Statement stmt2 = null;
	Statement stmt3 = null;
	Statement stmt4 = null;

	ResultSet rs1 = null;;
	ResultSet rs2 = null;
	ResultSet rs3 = null;
	ResultSet rs4 = null;

	try
	{
	    // Check to see if the permission is at all listed in the db.
	    // If not, then the user has free reign.
	    String sql_statement = "select Allow_Deny from " +
		"loco_access_control_list " +
		"where Objectid = " + id +
		" and Permission = " + SQL.quoteString(perm);

	    stmt1 = conn.createStatement();
	    rs1 = stmt1.executeQuery(sql_statement);

	    if (!rs1.next())
	    {
		// permission not in acl; user has free reign
		return(true);
	    }
	    // permission exists; check flag to determine if
	    // the acl is an allow or deny list
	    boolean allow_deny = rs1.getBoolean("Allow_Deny");

	    // See if the individual is in the acl
	    // for this particular permission 
	    sql_statement = "select * from " +
		"loco_access_control_list " +
		"where Objectid = " + id +
		" and Permission = " + SQL.quoteString(perm) +
		" and Userid = " + userid;

	    stmt2 = conn.createStatement();
	    rs2 = stmt2.executeQuery(sql_statement);

	    if (rs2.next())
	    {
		// user is contained in the list
		return (allow_deny);
	    }

	    // Since the permission doesn't have the user listed with it,
	    // check the user's groups.	    
	    
	    sql_statement = "select Groupid from LOCO_USER_GROUPS " +
		"where Userid = " + userid;

	    stmt3 = conn.createStatement();
	    rs3 = stmt3.executeQuery(sql_statement);

            while (rs3.next())
	    {
		// See if the permission exists for any group
		int groupid = rs3.getInt("groupid");
		
		sql_statement = "select * from " +
		    "loco_access_control_list " +
		    "where Objectid = " + id  +
		    " and Permission = " + SQL.quoteString(perm) +
		    " and Groupid = " + groupid;

		stmt4 = conn.createStatement();
		rs4 = stmt4.executeQuery(sql_statement);

		if (rs4.next())
		{
		    // Group exists for the permission
		    return(allow_deny);
		}

		rs4.close();
		stmt4.close();
	    }

	    // Since the user or any of his groups were not
	    // associated with the permission in the entire
	    // acl, then the user has free reign if the list
	    // was a deny list, or no access if the list was
	    // an allow list.
	    return(!allow_deny);
	       
	}
	catch (SQLException se)
	{ 
	    throw(se);
	}
	finally
	{
	    if(rs1 != null) { rs1.close(); }
	    if(stmt1 != null) { stmt1.close(); }
	    if(rs2 != null) { rs2.close(); }
	    if(stmt2 != null) { stmt2.close(); }
	    if(rs3 != null) { rs3.close(); }
	    if(stmt3 != null) { stmt3.close(); }
	    if(rs4 != null) { rs4.close(); }
	    if(stmt4 != null) { stmt4.close(); }
	}
    }

    // Private helper methods

    /**
     * Removes a user from a permission in an ACL in the database.
     *
     * @param id       id of ACL
     * @param userName name of the user to be removed
     * @param permName name of the permission in the ACL
     * @param conn     database connection
     *
     * @exception SQLException             when a database error occurs
     * @exception LoadUserException        when a user cannot be loaded from 
     *                                     the database
     * @exception UserNotFoundException    when a user cannot be found in the
     *                                     database
     */
    private static void removeUser(int id, String userName, String permName,
				   Connection conn)
	throws SQLException, LoadUserException, UserNotFoundException
    {
	Statement stmt = null;

	try
	{
	    int userid = -1;

	    if (userName.equals(User.ANONYMOUS_NAME))
	    {
		userid = User.ANONYMOUS_ID;
	    }
	    else
	    {
		userid = (new User(conn, userName)).getUserid();
	    }
	    
	    String sql_statement = "delete from loco_access_control_list " +
		"where Userid = " + userid +
		" and Permission = " + SQL.quoteString(permName) +
		" and Objectid = " + id;
	    
	    stmt = conn.createStatement();
	    
	    stmt.executeUpdate(sql_statement);
	}
	catch (SQLException se)
	{ 
	    throw(se);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Removes a group from a permission in an ACL in the database.
     *
     * @param id        id of ACL
     * @param groupName name of the group to be removed
     * @param permName  name of the permission in the ACL
     * @param conn      database connection
     *
     * @exception SQLException     when a database error occurs
     */
    private static void removeGroup(int id, String groupName, String permName,
				    Connection conn)
	throws SQLException
    {
	Statement stmt = null;

	try
	{
	    int groupid = GroupManager.convertToGroupId(groupName, conn);

	    String sql_statement = "delete from loco_access_control_list " +
		"where Groupid = " + groupid +
		" and Permission = " + SQL.quoteString(permName) +
		" and Objectid = " + id;
	    
	    stmt = conn.createStatement();
	    
	    stmt.executeUpdate(sql_statement);
	}
	catch (SQLException se)
	{ 
	    throw(se);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Changes the polarity (allow or deny) of a permission in an ACL.
     *
     * @param id        id of ACL
     * @param pol       <code>true</code> for allow, <code>false</code> for
     *                  deny
     * @param permName  name of permission
     * @param conn      database connection
     *
     * @exception SQLException   when a database error occurs
     */
    private static void updatePolarity(int id, boolean pol, String permName,
				       Connection conn)
	throws SQLException
    {
	PreparedStatement pstmt = null;

	try
	{
	    String sql_statement = "update loco_access_control_list" +
		" set Allow_Deny = ?" +
		" where Permission = ?" +
		" and Objectid = ?";

	    pstmt = conn.prepareStatement(sql_statement);
	    
	    pstmt.setBoolean(1, pol);
	    pstmt.setString(2, permName);
	    pstmt.setInt(3, id);

	    pstmt.executeUpdate();
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(pstmt != null) { pstmt.close(); }
	}
    }

    /**
     * Adds an entry for an ACL (which is solely identified by its id number).
     * The entry consists of the permission, an array of users, an array
     * of groups, and a boolean indicating whether the users and groups
     * are allowed or denied that permission.
     *
     * @param id         id of ACL
     * @param perm       name of permission
     * @param userIds    Vector of user ids to be added to the acl
     * @param groupIds   Vector of group ids to be added to the acl
     * @param allow_deny boolean polarity of the permission
     * @param conn       database connection
     *
     * @exception SQLException             when a database error occurs
     */
    private static void addEntry (int id, String perm, Vector userIds,
				  Vector groupIds, boolean allow_deny,
				  Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;
	PreparedStatement pstmt = null;

	// First see if we have a valid permission:
	if(!PermissionSet.exists(perm))
	{
	    return;
	}

	try
	{
	    String sql_statement = "select Allow_Deny " +
		"from loco_access_control_list " +
		"where Permission = " + SQL.quoteString(perm) +
		" and ObjectId = " + id;
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);
	    
	    if(rs.next())
	    {
		boolean a_d = rs.getBoolean("Allow_Deny");
		if (a_d != allow_deny)
		{
		    // Can't have conflicting allow_deny flags for the same
		    // permission
		    return;
		}
	    }
	    	    
	    sql_statement = "insert into " +
		"loco_access_control_list " +
		"(Objectid, Permission, Userid, Groupid, Allow_Deny) " +
		"values (?, ?, ?, ?, ?)";

	    pstmt = conn.prepareStatement(sql_statement); 

	    pstmt.setInt(1, id);
	    pstmt.setString(2, perm);
	    pstmt.setInt(4, User.NULL_USERID);  // -1
	    pstmt.setBoolean(5, allow_deny);

	    if (userIds != null)
	    {
		Enumeration u = userIds.elements();
		while(u.hasMoreElements())
		{
		    Integer uid = (Integer)u.nextElement();
		    pstmt.setInt(3, uid.intValue());
		    pstmt.executeUpdate();
		} 
	    }		

	    pstmt.setInt(3, User.NULL_USERID);  // -1

	    if (groupIds != null)
	    {
		Enumeration g = groupIds.elements();
		while(g.hasMoreElements())
		{   
		    Integer gid = (Integer)g.nextElement();
		    pstmt.setInt(4, gid.intValue());
		    pstmt.executeUpdate();
		}
	    }

	    if ( (userIds == null || userIds.isEmpty()) &&
		 (groupIds == null || groupIds.isEmpty()) )
	    {
		pstmt.setInt(3, NO_USERS);
		pstmt.executeUpdate();
	    } 
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	    if(pstmt != null) { pstmt.close(); }
	}

    }

    /**
     * Loads an individual permission into an ACL from the database.
     *
     * @param acl      AccessControlList to which to add the permission entry
     * @param permName name of permission
     * @param conn     database connection
     *
     * @return resulting AccessControlList with the loaded entry
     *
     * @exception SQLException             when a database error occurs
     * @exception LoadUserException        when a user cannot be loaded from 
     *                                     the database
     * @exception UserNotFoundException    when a user cannot be found in the
     *                                     database
     */
    private static AccessControlList loadEntry(AccessControlList acl,
					       String permName,
					       Connection conn)
	throws SQLException, LoadUserException, UserNotFoundException
       
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    // First get the polarity of the acl entry
	    String sql_statement = "select Allow_Deny from " +
		"loco_access_control_list " +
		"where Objectid = " + acl.getId() +
		" and Permission = " + SQL.quoteString(permName);

	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);

	    rs = stmt.executeQuery(sql_statement);
	    if (!rs.next())
	    {
		// No entries for this permission; therefore return
		return (null);
	    }
	    boolean allow_deny = rs.getBoolean("Allow_Deny");
	    acl.setPolarity(allow_deny, permName);

	    // Get the userids
	    sql_statement = "select Userid from " +
		"loco_access_control_list " +
		"where Objectid = " + acl.getId() +
		" and Permission = " + SQL.quoteString(permName);

	    rs.close();
	    rs = stmt.executeQuery(sql_statement);
	    while(rs.next())
	    {
		int userid = rs.getInt("Userid");
		
		if (userid == User.ANONYMOUS_ID)
		{
		    acl.addUser(User.ANONYMOUS_NAME, permName);
		}
		else if (userid > 0 && User.exists(conn, userid))
		{
		    String username = User.getUsername(conn, userid);
		    if (username != null) {
			acl.addUser(username, permName);
		    }
		}
	    }

	    // Get the groupids
	    sql_statement = "select Groupid from " +
		"loco_access_control_list " +
		"where Objectid = " + acl.getId() +
		" and Permission = " + SQL.quoteString(permName);

	    rs.close();
	    rs = stmt.executeQuery(sql_statement);
	    while(rs.next())
	    {
		int groupid = rs.getInt("Groupid");
		if (groupid > 0)
		{
		    String groupname = GroupManager.convertToGroupName(groupid,
								       conn);
		    if (groupname != null) {
			acl.addGroup(groupname, permName);
		    }
		}
	    }

	}
	catch (SQLException se)
	{ 
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}

	return(acl);
    }

    /**
     * Removes a permission entry from an acl in the database.
     *
     * @param id       int id of ACL
     * @param permName name of permission
     * @param conn     database connection
     *
     * @exception SQLException when a database error occurs
     */
    private static void removeEntry(int id, String permName, Connection conn)
	throws SQLException
    {
	Statement stmt = null;

	try
	{
	    String sql_statement = "delete from " +
		"loco_access_control_list " +
		"where Objectid = " + id +
		" and Permission = " + SQL.quoteString(permName);

	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	}
	catch (SQLException se)
	{ 
	    throw(se);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }
    /**
     * Converts a Vector of user names into a Vector of user ids.
     *
     * @param usernames Vector of user names
     * @param conn      database connection object
     *
     * @return Vector of user ids
     *
     * @exception LoadUserException        when a user cannot be loaded from 
     *                                     the database
     * @exception UserNotFoundException    when a user cannot be found in the
     *                                     database
     */
    private static Vector convertToUserIds(Vector usernames, Connection conn)
	throws LoadUserException, UserNotFoundException
    {
	Vector userids = new Vector();
	
	Enumeration e = usernames.elements();
	while(e.hasMoreElements())
	{
	    String username = (String)e.nextElement();
	    if (username.equals(User.ANONYMOUS_NAME))
	    {
		userids.addElement(new Integer(User.ANONYMOUS_ID));
	    }
	    else
	    {
		User u = new User(conn, username);
		userids.addElement(new Integer(u.getUserid()));
	    }
	}

	return(userids);
    }
}








