/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import java.io.*;
import java.lang.*;
import java.util.*;

/**
 * Defines permission ACLs.  An ACL consists of a list of permissions.
 * Each permission in the ACL has a polarity, a list of users, and a
 * list of groups.  Polarity determines whether that permission has an
 * allow or deny list.  Therefore the list of users and the list of
 * groups associated with the permission are allowed access to that
 * permission if it has positive polarity, or are denied that permission
 * if that permission has negative polarity.
 */

public class AccessControlList implements Serializable
{
    // Fields

    private int       id = -1;
    private Vector    permissions;
    private Hashtable users;
    private Hashtable groups;
    private Hashtable polarity;

    // Constructor

    /**
     * Creates an ACL for an object with owner id.
     *
     * @param ownerId id of the object to which the ACL is attached
     */
    public AccessControlList(int ownerId)
    {
	id = ownerId;
	permissions = new Vector();
	users = new Hashtable();
	groups = new Hashtable();
	polarity = new Hashtable();
    }

    // Methods
    
    /**
     * Returns the owner id of the ACL.
     *
     * @return int owner id of the ACL
     */
    public int getId()
    {
	return(id);
    }

    /**
     * Adds a permission to the ACL.
     *
     * @param permName name of the permission being added to the ACL
     */
    public void addPermission(String permName)
    {
	if (!permissions.contains(permName))
	{
	    permissions.addElement(permName);
	    users.put(permName, new Vector());
	    groups.put(permName, new Vector());
	    polarity.put(permName, new Boolean(true));
	}
    }

    /**
     * Removes a permission from the ACL.
     *
     * @param permName name of the permission to be removed from the ACL.
     */  
    public void removePermission(String permName)
    {
	permissions.removeElement(permName);
	users.remove(permName);
	groups.remove(permName);
	polarity.remove(permName);
    }

    /**
     * Returns a Vector containing all of the permission names contained in
     * the ACL.
     *
     * @return Vector of all permission names contained in the ACL
     */
    public Vector getPermissions()
    {
	return(permissions);
    }

    /**
     * Adds a user corresponding to a permission in the ACL.
     *
     * @param userName name of the user to be added
     * @param permName name of the permission to which the user is being added
     */
    public void addUser(String userName, String permName)
    {
	if (!permissions.contains(permName)) 
	{
	    return;
	}

	Vector currentUsers = (Vector) users.get(permName);
	if (!currentUsers.contains(userName))
	{
	    currentUsers.addElement(userName);
	    users.put(permName, currentUsers);
	}
    }
 
    /**
     * Removes a user corresponding to a permission in the ACL.
     *
     * @param userName name of the user to be removed
     * @param permName name of the permission from which the user is being
     * removed
     */

    public void removeUser(String userName, String permName)
    {
	if (!permissions.contains(permName))
	{
	    return;
	}

	Vector currentUsers = (Vector) users.get(permName);
	if (currentUsers.removeElement(userName))
	{
	    users.put(permName, currentUsers);
	}
    }

    /**
     * Returns a Vector containing the user names associated with a particular
     * permission.
     *
     * @param permName name of the permission
     *
     * @return Vector of String user names
     */
    public Vector getUsers(String permName)
    {
	if (!permissions.contains(permName))
	{
	    return(new Vector());
	}

	return((Vector)users.get(permName));
    }

    /**
     * Adds a group to a particular permission in the ACL.
     *
     * @param groupName name of group to be added
     * @param permName name of the permission to which the group is being added
     */
    public void addGroup(String groupName, String permName)
    {
	if (!permissions.contains(permName))
	{
	    return;
	}

	Vector currentGroups = (Vector) groups.get(permName);
	if (!currentGroups.contains(groupName))
	{
	    currentGroups.addElement(groupName);
	    groups.put(permName, currentGroups);
	}
    }

    /** 
     * Removes a group from a particular permission in the ACL.
     *
     * @param groupName name of the group to be removed
     * @param permName  name of the permission from which the group is being
     * removed
     */
    public void removeGroup(String groupName, String permName)
    {
	if (!permissions.contains(permName))
	{
	    return;
	}

	Vector currentGroups = (Vector) groups.get(permName);
	currentGroups.removeElement(groupName);
	groups.put(permName, currentGroups);
    }

    /**
     * Returns a Vector of group names associated with a particular permission
     *
     * @param permName name of the permission
     *
     * @return Vector of String group names
     */
    public Vector getGroups(String permName)
    {
	if (!permissions.contains(permName))
	{
	    return(new Vector());
	}
	
	return((Vector)groups.get(permName));
    }

    /**
     * Sets the polarity (allow or deny list) of a permission in the ACL.
     *
     * @param pol      boolean polarity value
     * @param permName permission of the ACL whose polarity is being set
     */
    public void setPolarity(boolean pol, String permName)
    {
	if (!permissions.contains(permName))
	{
	    return;
	}

	polarity.put(permName, new Boolean(pol));
    }

    /** 
     * Gets the polarity value of a permission in the ACL.
     * @param permName name of the permission whose polarity is in question
     * @return boolean polarity value
     */
    public boolean getPolarity(String permName)
    {
	if (!permissions.contains(permName))
	{
	    return(false);
	}

	Boolean bool = (Boolean)polarity.get(permName);
	return(bool.booleanValue());
    }
}







