/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import org.locomotive.util.*;

import java.lang.*;
import java.util.*;
import java.io.*;

/**
 * Contains all the information for one group, which represents a set 
 * of users.
 */

public class Group implements Serializable
{
    // Fields (accessed through the interface)

    private String name;
    private int id = -1;
    private Vector users;

    // Constructors

    /**
     * Creates a group with associated with a name.  This constructor is used
     * before inserting a group into the database, which will assign it
     * a unique id.
     *
     * @param groupName name of group to be created
     */
    public Group(String groupName)
    {
	name = groupName;
	users = new Vector();
    }

    /**
     * Creates a group with complete group information.  This constructor
     * is used when the database is loading group information into the
     * Group data structure.
     *
     * @param groupName  name of group
     * @param groupId    id of group
     * @param groupUsers Vector of user names contained in the group
     */
    public Group(String groupName, int groupId, Vector groupUsers)
    {
	name = groupName;
	id = groupId;
	users = groupUsers;
    }

    // Methods

    /**
     * Returns the name of the group.
     *
     * @return String name of the group
     */
    public String getName()
    {
	return(name);
    }
    
    /** 
     * Returns the id of the group.
     *
     * @return int id of the group
     */
    public int getId()
    {
	return(id);
    }

    /**
     * Sets the id of the group.
     *
     * @param id of the group
     */
    public void setId(int groupId)
    {
	id = groupId;
    }

    /** 
     * Adds a user to the group.
     *
     * @param userName name of user to be added to the group
     */
    public void addUser(String userName)
    {
	users.addElement(userName);
    }

    /** 
     * Removes an existing user from the group.
     *
     * @param userName name of user to be removed from the group
     */
    public void removeUser(String userName)
    {
	users.removeElement(userName);
    }

    /**
     * Lists all the user names contained in the group.
     *
     * @return Vector uf user names contained in the group
     */
    public Vector getUsers()
    {
	return users;
    }

    /**
     * Checks for the existence of a user within this group.
     *
     * @param userName name of user in question
     *
     * @return <code>true</code> if the user is contained in this group
     */
    public boolean containsUser(String userName)
    {
	return(users.contains(userName));
    }
}
