/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import org.locomotive.util.*;
import org.locomotive.loco.*;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Maintains all database transactions dealing with groups. 
 */

public class GroupManager implements Serializable
{
    /** 
     * Returns a unique id for a new group.
     *
     * @return a unique int id for a new group
     *
     * @exception SQLException if some database error occurs
     */
    public static int getNextId(Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
        {
            stmt = conn.createStatement();
            rs = stmt.executeQuery ("select LOCO_GROUPS_SQ.nextval " +
				    "from DUAL");

            rs.next();
            int newId = rs.getInt(1);
	    
            return (newId);
        }
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Adds a new group to the database.  Only adds the existence of the
     * group, not any of the users.  Use addUserToGroup or updateGroup to
     * add users to the group (after calling this function to initially create
     * the group).
     *
     * @param group group to be added to the database
     * @param conn  database connection
     *
     * @exception SQLException if some database error occurs
     */
    public static void addGroup(Group group, Connection conn)
	throws SQLException
    {
	if (group.getId() < 0)
	{
	    group.setId(GroupManager.getNextId(conn));
	}
	
	PreparedStatement pstmt = null;

	try
	{
	    String sql_statement = "insert into LOCO_GROUPS " +
		"(Name, Id) values (?, ?)";

	    pstmt = conn.prepareStatement(sql_statement);
	    
	    pstmt.setString(1, group.getName());
	    pstmt.setInt(2, group.getId());

	    pstmt.executeUpdate();
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(pstmt != null) { pstmt.close(); }
	}
    }

    /** 
     * Removes an existing group from the database.
     *
     * @param groupName name of group to be removed from the database
     * @param conn      database connection
     *
     * @exception SQLException if some database error occurs
     */
    public static void removeGroup(String groupName, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{ 
	    String sql_statement = "select Id " +
		"from LOCO_GROUPS where Name = " +
		SQL.quoteString(groupName); 
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery (sql_statement);
	    
	    int groupId = -1;
	    if (rs.next())
	    {
		groupId = rs.getInt("Id");
		rs.close();
		stmt.close();
	    }
	    else
	    {
		return;
	    }	

	    // Delete the group from the LOCO_GROUPS table in the database
	    sql_statement = "delete from LOCO_GROUPS " +
		"where Name = " + SQL.quoteString(groupName);
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	    stmt.close();

	    // Delete the group from the LOCO_USER_GROUPS table in the database
	    sql_statement = "delete from LOCO_USER_GROUPS  " +
		"where Groupid = " + groupId;
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	    stmt.close();	    

	    // Delete the group from any ACLs that may contain the group
	    sql_statement = "delete from loco_access_control_list  " +
		"where Groupid = " + groupId;
	    
	    stmt = conn.createStatement();
	    
	    stmt.executeUpdate(sql_statement);
	    stmt.close();	    
	    
	    stmt = conn.createStatement();
	    
	    stmt.executeUpdate(sql_statement);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }
 
    /** 
     * Updates an existing group in the database with new information.
     *
     * @param group updated group
     * @param conn  database connection
     *
     * @exception SQLException          if some database error occurs
     * @exception LoadUserException     when a user cannot be loaded
     * @exception UserNotFoundException when a user does not exist in the db
     */
    public static void updateGroup(Group group, Connection conn)
	throws SQLException, LoadUserException, UserNotFoundException
    {
	Group oldGroup = getGroup(group.getName(), conn);
	if(oldGroup == null) {
	    return;
	}
	
	Vector users = new Vector();

	Enumeration e = (group.getUsers()).elements();
	while(e.hasMoreElements())
	{
	    User u = new User(conn, (String)e.nextElement());
	    users.addElement(new Integer(u.getUserid()));
	}

	Vector oldUsers = GroupManager.getUserIdsFromGroup(oldGroup.getId(),
							   conn);
	
	Enumeration o = oldUsers.elements();
	while(o.hasMoreElements())
	{
	    Integer userId = (Integer)o.nextElement();
	    if (!users.contains(userId))
	    {
		GroupManager.removeUserFromGroup(group.getId(),
						 userId.intValue(), conn);
	    }
	}
	
	Enumeration u = users.elements();
	while(u.hasMoreElements())
	{
	    Integer userId = (Integer)u.nextElement();
	    if (!oldUsers.contains(userId))
		{
		    GroupManager.addUserToGroup(group.getId(),
						userId.intValue(), conn);
		}
	}
    }

    /** 
     * Returns a group object from the database via its name.
     *
     * @param groupName name of group being retrieved from the database
     * @param conn      database connection
     *
     * @return Group object (<code>null</code> if group does not exist)
     *
     * @exception SQLException if some database error occurs
     */
    public static Group getGroup(String groupName, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    String sql_statement = "select Name, " +
		"Id  from LOCO_GROUPS where Name = " + 
                 SQL.quoteString(groupName);
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);

	    if (!rs.next())
	    {
		return (null);
	    }

	    String name = rs.getString("Name");
	    int id = rs.getInt("Id");
	    
	    Vector userIds = GroupManager.getUserIdsFromGroup(id, conn);
	    Vector userNames = User.getUsernames(conn, userIds);

	    Group group = new Group(name, id, userNames);

	    return(group);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Adds a user to a group.
     *
     * @param groupId id of the group
     * @param userId  id of the user being added to the group
     * @param conn    database connection
     *
     * @exception SQLException if some database error occurs
     */
    public static void addUserToGroup(int groupId, int userId, Connection conn)
	throws SQLException
    {
	PreparedStatement pstmt = null;

	try
	{
	    String sql_statement = "insert into LOCO_USER_GROUPS " +
	    "(Userid, Groupid) values (?, ?)";

	    pstmt = conn.prepareStatement(sql_statement);
	    
	    pstmt.setInt(1, userId);
	    pstmt.setInt(2, groupId);

	    pstmt.executeUpdate();
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(pstmt != null) { pstmt.close(); }
	}

    }

    /** 
     * Removes an existing user from a group.
     *
     * @param groupId id of the group
     * @param userId  id of the user being removed from the group
     * @param conn    database connection
     *
     * @exception SQLException if some database error occurs   
     */
    public static void removeUserFromGroup(int groupId, int userId,
					   Connection conn)
	throws SQLException
    {
	Statement stmt = null;

	try
	{ 
	    String sql_statement = "delete from LOCO_USER_GROUPS " +
		"where Userid = " + userId + 
		"and Groupid = " + groupId;
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Lists all the user ids contained in a group.
     *
     * @param groupId id of the group
     * @param conn    database connection
     *
     * @return Vector of user ids contained in the group
     *
     * @exception SQLException if some database error occurs   
     */
    public static Vector getUserIdsFromGroup(int groupId, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    Vector users = new Vector();

	    String sql_statement = "select Userid " +
		"from LOCO_USER_GROUPS " +
		"where Groupid = " + groupId;

	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);
	    
	    while (rs.next())
	    {
		int userId = rs.getInt("Userid");
		users.addElement(new Integer(userId));
	    }

	    return(users);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}

    }
    
    /**
     * Returns an enumeration of all the group names in the database.
     *
     * @param conn database connection
     *
     * @return Vector containing all group names in the database
     *
     * @exception SQLException if some database error occurs
     */
    public static Vector getGroupNames(Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    Vector groups = new Vector();	
	    String sql_statement = "select Name from LOCO_GROUPS";
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);
	    
	    while (rs.next())
	    {
		String name = rs.getString("Name");
		groups.addElement(name);
	    }
	    
	    return(groups);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /** 
     * Verifies that a group exists in the DB.
     *
     * @param groupName name of group in question
     * @param conn      database connection
     *
     * @return <code>true</code> if the group exists in the database
     *
     * @exception SQLException if some database error occurs
     */
    public static boolean exists(String groupName,
				    Connection conn)
	throws SQLException 
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    boolean exists = true;

	    String sql_statement = 
                 "select Id from LOCO_GROUPS where Name = " +
		 SQL.quoteString(groupName); 
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery (sql_statement);

	    if (rs.next())
	    {
		exists = true;
	    }
	    else
	    {
		exists = false;
	    }

	    return(exists);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Converts a group name into a group id.
     *
     * @param groupName name of group
     * @param conn      database connection
     *
     * @return int id of group (-1 if group does not exist)
     *
     * @exception SQLException if some database error occurs
     */
    public static int convertToGroupId(String groupName, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    String sql_statement = "select Id from LOCO_GROUPS " +
		"where Name = " + SQL.quoteString(groupName);
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);

	    if (!rs.next())
	    {
		return (-1);
	    }

	    int id = rs.getInt("Id");
	    
	    return(id);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }
    
    /**
     * Converts a Vector of group names into a Vector of group ids.
     *
     * @param groupNames Vector containing the names of groups
     * @param conn       database connection
     *
     * @return Vector of group ids
     *
     * @exception SQLException if some database error occurs
     */
    public static Vector convertToGroupIds(Vector groupNames, Connection conn)
	throws SQLException
    {
	Vector groupIds = new Vector();
	Enumeration e = groupNames.elements();
	
	while (e.hasMoreElements())
	{
	    String groupname = (String)e.nextElement();
	    
	    int groupId = GroupManager.convertToGroupId(groupname, conn);
	    if (groupId > -1)
	    {
		groupIds.addElement(new Integer(groupId));
	    }
	}

	return (groupIds);
    }

     /**
     * Converts a group id into a group name.
     *
     * @param groupId id of group
     * @param conn    database connection
     *
     * @return String name of group (<code>null</code> if group does not exist)
     *
     * @exception SQLException if some database error occurs
     */   
    public static String convertToGroupName(int groupId, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    String sql_statement = "select Name from LOCO_GROUPS " +
		"where Id = " + groupId;
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);

	    if (!rs.next())
	    {
		return (null);
	    }

	    String name = rs.getString("Name");

	    return(name);
	}
        catch (SQLException se)
        {
	    throw(se);
        }
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Converts a Vector of group ids into a Vector of group names.
     *
     * @param groupIds Vector containing the ids of groups
     * @param conn     database connection
     *
     * @return Vector of group names
     *
     * @exception SQLException if some database error occurs
     */
    public static Vector convertToGroupNames(Vector groupIds, Connection conn)
	throws SQLException
    {
	Vector groupNames = new Vector();
	Enumeration e = groupIds.elements();
	
	while (e.hasMoreElements())
	{
	    Integer groupId = (Integer)e.nextElement();

	    String groupName = GroupManager.convertToGroupName
		(groupId.intValue(), conn);
	    if (groupName != null)
	    {
		groupNames.addElement(groupName);
	    }
	}

	return (groupNames);
    }
}



