/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import org.locomotive.util.SQL;

import java.io.*;
import java.lang.String;
import java.sql.*;
import java.util.*;

/**
 * Creates and maintains a single atomic permission.  Also contains a cache for
 * all the permissions presently in the database, and methods to access it,
 * as well as methods to create and remove new permissions.
 */
public class Permission implements Serializable
{
    public static int DB_TYPE = SQL.DB_TYPE_ORACLE;    // THIS IS THE DEFAULT

    // This hashtable acts as a cache of all the permissions stored in the db.
    // Permissions are mapped to positions in the internal bitfield
    private static Hashtable perm_hash;
    private static boolean initialized = false;

    // Private Class Members (accessed through the interface)
    
    private String name;
    private int id;

    // Constructors
   
    /**
     * Creates a permission associated by a name and id.  This item is not
     * put in the cache.
     *
     * @param name String name of the permission
     * @param id   int id of the permission
     */
    public Permission(String name, int id)
    {
	this.name = name;
	this.id = id;
    }

    /** 
     * Loads the permission object with the given name from the database, and
     * puts it in the cache.  To be used for atomic updates of the cache, 
     * or to ensure that the permission to be instantiated is the one
     * in the database.  To get a permission from the cache, use 
     * Permission.get();
     * @exception SQLException if we had database problems
     * @see #get
     */
    public Permission (Connection conn, String name)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;
	String query = "select name, id from loco_permissions where name = " +
	    SQL.quoteString(name);
	
	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);
	    rs.next();
	    this.name = rs.getString(1);
	    this.id = rs.getInt(2);

	    perm_hash.put(name, this);
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}
    }


    // Methods
    
    /**
     * Returns the name of the permission.
     *
     * @return String name of the permission
     */
    public String getName() 
    {
	return name;
    }

    /** 
     * Returns the id of the permission.
     *
     * @return int id of the permission
     */
    public int getId() 
    {
	return id;
    }


    /*-------------- static database access methods --------------*/

    /**
     * Reads permissions from the database and puts them into the permission
     * hash.  This function must be executed before any instances of this class
     * are created.
     *
     * @param conn Connection object to the database
     * 
     * @exception SQLException if a database error occurs
     */
    public static void initializeCache(Connection conn)
	throws SQLException
    {
	if (initialized)
	    return;

	perm_hash = new Hashtable();

	// Fill in perm_hash from the DB where the permission name
	// is the key and the Permission object is the value
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    DB_TYPE = SQL.getDBType(conn);

	    String sql_statement = "select Name, Id from loco_permissions";
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);
	    
	    while (rs.next())
	    {
		String name = rs.getString(1);
		int id = rs.getInt(2);
		
		Permission perm = new Permission(name, id);	
		perm_hash.put(name, perm);
	    }

	    initialized = true;
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null ) { stmt.close(); }
	}
    }


    /**
     * Reloads the permissions from the database.  Use this to prevent
     * inconsistency of data across multiple servers.
     * 
     * @param conn the Database connection object
     *
     * @exception SQLException if a database error occurs
     */
    public static void reloadCache(Connection conn) throws SQLException
    {
	initialized = false;
	initializeCache(conn);
    }


    /**
     * Retrieves a permission from the permissions cache.  If the permission
     * is not in the cache, then this returns null.  To instantiate a 
     * permission directly from the database, use the constructor.
     *
     * @param permName String name of the permission to be retrieved
     */
    public static Permission get(String permName)
    {
	Permission perm = (Permission) perm_hash.get(permName);
	return(perm);
    }

    /**
     * Checks the existence of a permission in the cache. Does not check the
     * database.
     *
     * @param permName String name of the permission
     */
    public static boolean exists(String permName)
    {
	return (perm_hash.containsKey(permName));
    }


    /** 
     * Returns the number of permissions currently in the cache
     */
    public static int getSize()
    {
	return (perm_hash.size());
    }


    /** 
     * Returns an enumeration of all the permissions currently in the cache
     */
    public static Enumeration permissions()
    {
	return (perm_hash.elements());
    }


    /**
     * Returns an Vector of all permission names presently in the cache.  
     * Don't confuse this with getPermissions(), which actually goes back
     * to the database and returns the permissions ordered by name.
     *
     * @return Vector of String permission names
     */
    public static Vector names()
    {
	Vector names = new Vector();
	Enumeration e = perm_hash.keys();

	while(e.hasMoreElements())
	{
	    names.addElement((String)e.nextElement());
	}

	return(names);
    }

    /** 
     * Returns all the permissions currently available.  The vector contains
     * a list of all the permissions sorted by their name, in ascending order.
     * This goes directly to the database, bypassing the cache.  The names() 
     * method retrieves from the cache and is more efficient.
     *
     * @exception SQLException if we had database problems
     */
    public static Vector getPermissions(Connection conn) throws SQLException
    {
	Vector perms = new Vector();

	Statement stmt = null;
	ResultSet rs = null;
	String query = "select id, name from loco_permissions " +
	    " order by name asc";
	
	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);
	    Permission p = null;
	    int id = 0;
	    String name = null;

	    while (rs.next())
	    {
		id = rs.getInt(1);
		name = rs.getString(2);
		p = new Permission(name, id);
		perms.addElement(p);
	    }
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}

	return (perms);
    }


    /**
     * Creates a new permission and puts it in the cache.
     * @return the permission object, with the unique id it is associated with
     * in the database.
     * @exception SQLException if we had database problems, or if a
     *            permission with that name already exists
     */
    public static Permission createPermission(Connection conn, String name)
	throws SQLException
    {
	String query = null;
	Statement stmt = null;
	ResultSet rs = null;
	int id = 0;
	
	try
	{
	    stmt = conn.createStatement();

	    switch (DB_TYPE)
	    {
	    case SQL.DB_TYPE_ORACLE:
		query = "select loco_permissions_sq.nextval from dual";
		rs = stmt.executeQuery(query);
		rs.next();
		id = rs.getInt(1);
		rs.close();
		rs = null;
		break;
	    }

	    query = "insert into loco_permissions (id, name) values (" + id +
		", " + SQL.quoteString(name) + ")";
	    stmt.executeUpdate(query);
	    
	    switch (DB_TYPE)
	    {
	    case SQL.DB_TYPE_MYSQL:
		query = "select LAST_INSERT_ID() from loco_permissions";
		rs = stmt.executeQuery(query);
		rs.next();
		id = rs.getInt(1);
		break;
	    }
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}
	
	Permission perm = new Permission(name, id);
	perm_hash.put(name, perm);
	return (perm);
    }

    /**
     * deletes the permission with the specified name from the database
     */
    public static void deletePermission(Connection conn, String perm_name)
	throws SQLException
    {
	Statement stmt = null;
	String query = "delete from loco_permissions where name = " + 
	    SQL.quoteString(perm_name);
	
	try
	{
	    stmt = conn.createStatement();
	    stmt.executeUpdate(query);
	    
	    perm_hash.remove(perm_name);
	}
	finally
	    {
		if (stmt != null) stmt.close();
	    }
	
    }

    
    /** Method for displaying permissions- prints only the name */
    public String toString()
    {
	return (name);
    }
    

}


