/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import org.locomotive.util.*;

import java.util.*;
import java.sql.*;
import java.lang.*;
import java.io.Serializable;

/**
 * Provides a data structure to store a set of permissions.  A PermissionSet
 * contains a bitfield in which each bit is associated with a permission.
 * The PermissionSet is said to contain a permission if the corresponding
 * bit is set in its bitfield.
 */

public class PermissionSet implements Serializable, Cloneable
{
    // Constants
    private static int NUM_BITS_IN_INT = 32; // Number of bits in an Integer

    // Instance Member Variables

    // This BitSet stores all the set bits of an instance of a PermissionSet.
    private BitSet bits;
    // the userid allows us to associate this with a user
    private int userid;

    // Static Functions

    
    /**
     * Reads permissions from the database and puts them into the permission
     * hash.  This function must be executed before any instances of this class
     * are created.
     *
     * @param conn Connection object to the database
     * @deprecated in favor of Permission.initializeCache();
     * @exception SQLException if a database error occurs
     */
    public static void initialize(Connection conn)
	throws SQLException
    {
	Permission.initializeCache(conn);
    }

    /**
     * Inserts a new permission into the database.  Watch out- no checking
     * is done to make sure this permission exists or not.
     *
     * @param perm Permission to be inserted into the database
     * @param conn Connection object to the database
     * @deprecated Permission.createPermission() should be used instead
     * @exception SQLException if some database error occurs
     */
    private static void insert(Permission perm, Connection conn)
	throws SQLException
    {
	PreparedStatement pstmt = null;

	try
	{
	    String sql_statement = "insert into loco_permissions " +
		"(Name, Id) values (?, ?)";

	    pstmt = conn.prepareStatement(sql_statement);
	    
	    pstmt.setString(1, perm.getName());
	    pstmt.setInt(2, perm.getId());

	    pstmt.executeUpdate();

	    Permission.reloadCache(conn);
	}
	finally
	{
	    if(pstmt != null) { pstmt.close(); }
	}
    }

    /*
     * Removes a permission from the database. 
     *
     * @param permName name of permission to be removed from the database
     * @param conn     Connection object to the database
     * @deprecated Permission.deletePermission should be used instead
     * @exception SQLException if some database error occurs     
     */
    private static void remove(String permName, Connection conn)
	throws SQLException
    {	
	Permission.deletePermission(conn, permName);
    }


    /**
     * Retrieves a permission from the database.
     *
     * @param permName String name of the permission to be retrieved
     * @deprecated in favor of Permission.get()
     * @return Permission object of the specified name
     */
    public static Permission get(String permName)
    {
	Permission perm = Permission.get(permName);
	return(perm);
    }

    /**
     * Returns an enumeration of all permission names in the database.
     *
     * @return Vector of String permission names
     * @deprecated in favor of Permission.getPermissions()
     */
    public static Vector names()
    {
	return (Permission.names());
    }

    /**
     * Checks the existence of a permission in the database.
     *
     * @param permName String name of the permission
     * @deprecated in favor of Permission.exists()
     * @return <code>true</code> if the permission exists in the database
     */
    public static boolean exists(String permName)
    {
	return (Permission.exists(permName));
    }


    // Constructors

    /**
     * Creates an empty permission set.
     */
    public PermissionSet()
    {
	bits = new BitSet();
	userid = 0;
    }

    /**
     * Creates an empty permission set with the specified userid
     */
    public PermissionSet(int userid)
    {
	bits = new BitSet();
	this.userid = userid;
    }
    
    /**
     * Creates a permission set initialized to the bitfield contained in the
     * Vector of Integers.  Used for loading a permission set from the
     * database.
     *
     * @param ints Vector of Integers representing a permissions bitfield
     */
    public PermissionSet(Vector ints)
    {
	setBitfield(ints);
	userid = 0;
    }

    /**
     * Creates a permission set initialized to the bitfield contained in the
     * array of booleans.  An alternative means of loading a permission set 
     * from the database.
     *
     * @param ints Vector of Integers representing a permissions bitfield
     */
    public PermissionSet(boolean [] values)
    {
	setBitfield(values);
	userid = 0;
    }


    /**
     *  sets the userid for this permission set
     */
    public void setUserid(int userid)
    {
	this.userid = userid;
    }

    /**
     * gets the userid for this permission set
     */
    public int getUserid()
    {
	return (userid);
    }

    /**
     * Sets a permission in the permission set.
     *
     * @param permName name of the permission to set
     * @param on_or_off whether to set the permission to on or off.
     */
    public PermissionSet setOnOrOff(String permName, boolean on_or_off)
    {
	Permission perm = Permission.get(permName);
	if(perm != null)
	{
	    int pos = perm.getId();
	    if (on_or_off)
		bits.set(pos);
	    else
		bits.clear(pos);
	}

	return (this);
    }

    /**
     * Sets a permission in the permission set.
     *
     * @param perm the permission to set
     * @param on_or_off whether to set the permission to on or off.
     */
    public PermissionSet setOnOrOff(Permission perm, boolean on_or_off)
    {
	if(perm != null)
	{
	    int pos = perm.getId();
	    if (on_or_off)
		bits.set(pos);
	    else
		bits.clear(pos);
	}

	return (this);
    }

    /**
     * Sets a permission in the permission set.
     * @param permName name of the permission to set
     */
    public PermissionSet set(String permName)
    {
	Permission perm = Permission.get(permName);
	if(perm != null)
	{
	    int pos = perm.getId();
	    bits.set(pos);
	}

	return(this);
    }

    /**
     * Sets a permission in the permission set.
     * @param perm the permission to set
     */
    public PermissionSet set(Permission perm)
    {
	if(perm != null)
	{
	    int pos = perm.getId();
	    bits.set(pos);
	}

	return (this);
    }

    /**
     * Sets all permissions in the permission set.
     */
    public PermissionSet setAll()
    {
	for(int i = 0; i < bits.size(); i++)
	{
	    bits.set(i);
	}

	return(this);
    }

    /**
     * Unsets a permission in the permission set.
     * @param permName name of the permission to clear
     */
    public PermissionSet unset(String permName)
    {
	Permission perm = Permission.get(permName);
	if(perm != null)
	{
	    int pos = perm.getId();
	    bits.clear(pos);
	}

	return(this);
    }

    /**
     * Unsets a permission in the permission set.
     * @param perm the permission to clear
     */
    public PermissionSet unset(Permission perm)
    {
	if(perm != null)
	{
	    int pos = perm.getId();
	    bits.clear(pos);
	}
	return (this);
    }

    /** 
     * Clears all permissions in the permission set.
     */
    public PermissionSet clear()
    {
	bits = new BitSet();

	return(this);
    }

    /**
     * Checks if a permission is set in the permission set.
     *
     * @param permName name of the permission in question
     *
     * @return <code>true</code> if the permission is set
     */
    public boolean isSet(String permName)
    {
	Permission perm = Permission.get(permName);
	if(perm == null)
	{
	    return(false);
	}

	int pos = perm.getId();
	return(bits.get(pos));
    }

    /**
     * Checks if a permission is set in the permission set.
     *
     * @param perm the permission in question
     *
     * @return <code>true</code> if the permission is set
     */
    public boolean isSet(Permission perm)
    {
	if(perm == null)
	{
	    return(false);
	}

	int pos = perm.getId();
	return(bits.get(pos));
    }


    /** a handy little method for package internal use in UserPerm */
    protected boolean isSet(int position)
    {
	return (bits.get(position));
    }

    /** 
     * Returns an enumeration of all set permission names.
     *
     * @return Vector of String names
     */
    public Vector getSetPerms()
    {
	Vector set_names = new Vector();

	Enumeration e = Permission.permissions();

	while (e.hasMoreElements())
	{
	    Permission perm = (Permission) e.nextElement();
	    int pos = perm.getId();

	    if(bits.get(pos))
	    {
		set_names.addElement(perm.getName());
	    }
	}

	return(set_names);
    }

    /** 
     * returns a SortableTable of all the permissions and whether or not
     * this permission set has them set.  Has the columns: <PRE>
     *   0 perm_id (int)       1 perm_name (String)
     *   2 is_set  (boolean)
     * </PRE>
     * This table is sorted by Permission name.
     */
    public SortableTable getPermTable()
    {
	SortableTable st = new SortableTable(Permission.getSize(), 3);
	st.setColumnName(0, "perm_id");
	st.setColumnName(1, "perm_name");
	st.setColumnName(2, "is_set");
	
	Enumeration e = Permission.permissions();
	for (int i = 0; e.hasMoreElements(); i++)
	{
	    Permission perm = (Permission) e.nextElement();
	    int pos = perm.getId();
	    st.insert(new Integer(pos), i, 0);
	    st.insert(perm.getName(), i, 1);
	    st.insert(new Boolean(bits.get(pos)), i, 2);
	}
	
	st.sort(1, SortableTable.STRING, SortableTable.ASCENDING);

	return (st);
    }



    /** 
     * 'Ands' this PermissionSet with another PermissionSet. 
     *
     * @param pSet PermissionSet with which to be 'And'ed
     */
    public PermissionSet and(PermissionSet pSet)
    {
	bits.and(pSet.bits);

	return(this);
    }

    /** 
     * 'Ors' this PermissionSet with another PermissionSet.
     *
     * @param pSet PermissionSet with which to be 'Or'ed
     */
    public PermissionSet or(PermissionSet pSet)
    {
	bits.or(pSet.bits);

	return(this);
    }

    /** 
     * 'Xors' the PermissionSet with another PermissionSet.
     *
     * @param pSet PermissionSet with which to be 'Xor'ed
     */
    public PermissionSet xor(PermissionSet pSet)
    {
	bits.xor(pSet.bits);

	return(this);
    }

    /**
     * returns the size of the permission set
     *
     */
    public int size()
    {
	return (bits.size());
    }


    /**  
     * Sees if this PermissionSet is a subset of another PermissionSet.
     *
     * @param pSet PermissionSet being tested with
     *
     * @return <code>true</code> if this PermissionSet is a subset of the
     *         parameter PermissionSet
     */
    public boolean isSubsetOf(PermissionSet pSet)
    {
	BitSet other_bits = pSet.bits;

	for(int i = 0; i < bits.size(); i++)
	{
	    if(!other_bits.get(i))
	    {
		if(bits.get(i))
		{
		    return(false);
		}
	    }
	}
	
	return(true);
    }

    /**
     * Transforms the BitSet's bitfield into a Vector of Integers.  The Vector
     * can have any number of Integers.  Integers are 32 bits, so the BitSet
     * will be a size of X * 32, where X is the number of Integers.  Used
     * primarily for storage in the database.
     *
     * @return Vector of Integers representing the PermissionSet's bitfield.
     */
    public Vector toInts()
    {
	if(bits.size() == 0)
	{
	    return(null);
	}

	Vector collapsed = new Vector();

	// First we have to find how many Integers we will need to contain
	// the bitfield in the BitSet object.
	int num_ints = bits.size()/NUM_BITS_IN_INT;
	if(bits.size()%NUM_BITS_IN_INT != 0)
	{
	    num_ints++;
	}

	// We start at the highest 32 bits of the bitfield, put it into an
	// Integer, and put it as the first in the queue of the Vector.  The
	// next lowest 32 bits will then be added to the Vector, and so on
	// until the lowest 32 bits will be the last added to the Vector, which
	// will then be at index 0.
	for(int j = num_ints-1; j >= 0; j--)
	{
	    int bitfield = 0;
	    
	    // Starting from the 31st bit (remember the range will be 0-31)
	    // we keep left shifting in the 1's or 0's, so that at the end
	    // of the iteration, the first bit that we put in will be in the
	    // 31st index.
	    for (int i = NUM_BITS_IN_INT-1; i >= 0; i--)
	    {
		if (bits.get(i + NUM_BITS_IN_INT*j))
		{
		    bitfield += 1;
		}
		
		if (i > 0)
		{
		    bitfield = bitfield << 1;
		}
	    }
	    
	    // Remember, the highest 32 bits go in first, and will be pushed
	    // to the next index when the next lower 32 bits get added.
	    collapsed.insertElementAt(new Integer(bitfield), 0);
	}

	// Now remove any extra 0 integers from the end of the list
	int index = collapsed.size() - 1;

	while(index > 0)
	{
	    if(((Integer)collapsed.elementAt(index)).intValue() == 0)
	    {
		collapsed.removeElementAt(index);
		index--;
	    }
	    else
	    {
		break;
	    }
	}

	return(collapsed);
    }

    /**
     * Takes a Vector of Integers (in the format created by the
     * <code>toInts<\code> method), and sets it as our new bitfield.
     *
     * @param ints Vector of Integers representing the PermissionsSet bitfield
     */
    public PermissionSet setBitfield(Vector ints)
    {
	// Reset the BitSet
	bits = new BitSet();

	if (ints == null)
	{
	    return(this);
	}

	// Take each Integer and set the appropriate bit in the BitSet object
	for(int j = 0; j < ints.size(); j++)
	{
	    int bitfield = ((Integer)ints.elementAt(j)).intValue();
	    int currbit = 0;

	    for (int i = 0; i < NUM_BITS_IN_INT; i++)
	    {
		currbit = (int) (bitfield&1);
		if (currbit == 1)
		{
		    bits.set(i + j*NUM_BITS_IN_INT);
		}
		bitfield = bitfield >> 1;
	    }	
	}

	return(this);
    }


    /**
     * Transforms the Bitfield into an array of booleans.  The booleans 
     * represent the each of the bits.  Should be used as an alternate 
     * means to toInts() of storage in the database 
     * @see #toInts
     */
    public boolean [] toArray()
    {
	int size = bits.size();
	boolean [] collapsed = new boolean [size];

	for (int i = 0; i < size; i++)
	{
	    collapsed [i] = bits.get(i);
	}
	
	return (collapsed);
    }

    
    /** takes an array of booleans and sets them as our new bitfield.  An 
     * alternative to using the Vector method.
     * @param values the array of boolean values to use as the bitfield
     */
    public void setBitfield(boolean [] values)
    {
	bits = new BitSet(values.length);
	for (int i = 0; i < values.length; i++)
	{
	    if (values[i])
	    {
		bits.set(i);
	    }
	}
    }
	    


    /**
     * Returns the the userid + BitSet's toString().
     */
    public String toString()
    {
	if (userid > 0)
	    return("PermissionSet for userid " + userid + " bits: " + 
		   bits.toString());
	else
	    return ("PermissionSet: " + bits.toString());
    }

    /**
     * Makes a clone of this PermissionSet.
     */
    public Object clone()
    {
	PermissionSet cloned = new PermissionSet();
	cloned.bits = (BitSet)this.bits.clone();
	cloned.userid = this.userid;

	return(cloned);
    }
}
