/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import org.locomotive.util.*;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Maintains all database transactions dealing with roles.
 */

public class RoleManager implements Serializable
{

    /** 
     * Returns a unique id for a new role.
     *
     * @return int id for a new role
     *
     * @exception SQLException if some database error occurs
     */
    public static int getNextId(Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery ("select LOCO_ROLES_SQ.nextval from DUAL");
	    
	    rs.next();
	    int newId = rs.getInt(1);
	   
	    return (newId);
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
	
    }

    /**
     * Adds a new role to the database.
     *
     * @param role role to be added to the database
     * @param conn database connection
     *
     * @exception SQLException if some database error occurs
     */
    public static void addRole(Role role, Connection conn)
	throws SQLException
    {
	if (role.getId() < 0)
	{
	    role.setId(RoleManager.getNextId(conn));
	}

	PreparedStatement pstmt = null;

	try
	{
	    Vector ints = role.toInts();

	    if(ints.isEmpty())
	    {
		String sql_statement = "insert into LOCO_ROLES " +
		    "(Name, Id, Seq, Perms) values " +
		    "(?, ?, ?, ?)";
		
		pstmt = conn.prepareStatement(sql_statement);
		
		pstmt.setString(1, role.getName());
		pstmt.setInt(2, role.getId());
		pstmt.setInt(3, 0);
		pstmt.setInt(4, 0); 

		pstmt.executeUpdate();
	    }
	    else
	    {
		for(int i = 0; i < ints.size(); i++)
		{
		    int bitfield = ((Integer)ints.elementAt(i)).intValue();
		    
		    String sql_statement = "insert into LOCO_ROLES " +
			"(Name, Id, Seq, Perms) values " +
			"(?, ?, ?, ?)";
		    
		    pstmt = conn.prepareStatement(sql_statement);
		    
		    pstmt.setString(1, role.getName());
		    pstmt.setInt(2, role.getId());
		    pstmt.setInt(3, i);
		    pstmt.setInt(4, bitfield); 

		    pstmt.executeUpdate();
		    pstmt.close();
		}
	    }
	}
	catch (SQLException se)
	{	   
	    throw(se);
	}
	finally
	{
	    if(pstmt != null) { pstmt.close(); }
	}
    }

    /** 
     * Removes an existing role from the database.
     *
     * @param roleName name of role to be removed from the database
     * @param conn     database connection
     *
     * @exception SQLException if some database error occurs
     */
    public static void removeRole(String roleName, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{ 
	    String sql_statement = "select Id " +
		"from LOCO_ROLES " +
		"where Name = " +
		SQL.quoteString(roleName); 
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery (sql_statement);
	    
	    int roleId = -1;
	    if (rs.next())
	    {
		roleId = rs.getInt("Id");
		rs.close();
		stmt.close();
	    }
	    else
	    {
		return;
	    }	
	
	    // Delete the role from the LOCO_ROLES table in the database
	    sql_statement = "delete from LOCO_ROLES " +
		"where Name = " + SQL.quoteString(roleName);
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	    stmt.close();

	    // Also delete the role from any users that may contain the role
	    sql_statement = "delete from LOCO_ROLE_USERS " +
		"where Roleid = " + roleId;
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /** 
     * Updates an existing role in the database with new information
     * (for instance, new permissions that get removed or added to a role).
     *
     * @param role updated role
     * @param conn database connection
     *
     * @exception SQLException if some database error occurs
     */
    public static void updateRole(Role role, Connection conn)
	throws SQLException
    {
	if (!exists(role.getName(), conn))
	{
	    return;
	}

	Statement stmt = null;

	try
	{
	    // Delete the current entry(ies) of the role
	    String sql_statement = "delete from LOCO_ROLES " +
		"where Name = " + SQL.quoteString(role.getName());
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);

	    // Now just re-enter the role
	    addRole(role, conn);
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }

    /** 
     * Returns a role object from the database via its name.
     *
     * @param roleName name of role being retrieved from the database
     * @param conn database connection
     *
     * @return <code>Role</code> object (<code>null</code> if role does not
     *         exist)
     *
     * @exception SQLException if some database error occurs
     */
    public static Role getRole(String roleName,
			       Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    String sql_statement = "select Name, " +
		"Id, " +
		"Seq, " +
		"Perms " + 
		"from LOCO_ROLES " +
		"where Name = " +
		SQL.quoteString(roleName); 
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery (sql_statement);

	    String name = null;
	    int id = -1;

	    Hashtable h = new Hashtable();
	    
	    if(!rs.next())
	    {
		return(null);
	    }
	    else
	    {
		name = rs.getString("Name");
		id =  rs.getInt("Id");
		int seq = rs.getInt("Seq");
		int perms = rs.getInt("Perms");	
	
		h.put(new Integer(seq), new Integer(perms));
	    }

	    while(rs.next())
	    {
		int seq = rs.getInt("Seq");
		int perms = rs.getInt("Perms");

		h.put(new Integer(seq), new Integer(perms));
	    }

	    Vector ints = new Vector();
	    ints.setSize(h.size());

	    for(int j = 0; j < h.size(); j++)
	    {
		ints.setElementAt((Integer)h.get(new Integer(j)), j);
	    }	    

	    Role role = new Role(name, id, ints);

	    return(role);
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Returns an enumeration of all the role names in the database.
     *
     * @param conn database connection
     *
     * @return Vector containing all role names in the database
     *
     * @exception SQLException if some database error occurs
     */
    public static Vector getRoleNames(Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    Vector roles = new Vector();
	    String sql_statement = "select Name " +
		"from LOCO_ROLES";
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);
	    
	    while (rs.next())
	    {
		String name = rs.getString("name");
		if(!roles.contains(name))
		{
		    roles.addElement(name);
		}
	    }

	    return(roles);	    
	}
	catch (SQLException se)
	{  
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Verifies that a role exists in the DB.
     *
     * @param roleName name of role in question
     * @param conn     database connection
     *
     * @return <code>true</code> if a role with this name exists
     *
     * @exception SQLException if some database error occurs
     */
    public static boolean exists(String roleName, Connection conn)
	throws SQLException 
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    boolean exists = true;

	    String sql_statement = "select Id " +
		"from LOCO_ROLES " +
		"where Name = " +
		SQL.quoteString(roleName); 
	    
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery (sql_statement);

	    if (rs.next())
	    {
		exists = true;
	    }
	    else
	    {
		exists = false;
	    }

	    return(exists);
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Adds a role to a user.
     *
     * @param userId   id of the user
     * @param roleName name of the role being added to the user
     * @param conn     database connection
     *
     * @exception SQLException if some database error occurs
     */
    public static void addRoleToUser(int userId, String roleName, 
				     Connection conn)
	throws SQLException
    {
	PreparedStatement pstmt = null;

	try
	{
	    String sql_statement = "insert into LOCO_ROLE_USERS " +
	    "(Roleid, Userid) values " +
	    "(?, ?)";

	    pstmt = conn.prepareStatement(sql_statement);
	    
	    Role role = getRole(roleName, conn);
	    if (role == null) {
		pstmt.close();
		return;
	    }
	    pstmt.setInt(1, role.getId());
	    pstmt.setInt(2, userId);

	    pstmt.executeUpdate();
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(pstmt != null) { pstmt.close(); }
	}
    }

    /** 
     * Removes an existing role from a user.
     *
     * @param userId   id of the user
     * @param roleName name of the role being removed from the user
     * @param conn     database connection
     *
     * @exception SQLException if some database error occurs   
     */
    public static void removeRoleFromUser(int userId, String roleName,
					  Connection conn)
	throws SQLException
    {
	Statement stmt = null;

	try
	{ 
	    Role role = getRole(roleName, conn);
	    if (role == null) {
		return;
	    }
	    String sql_statement = "delete from LOCO_ROLE_USERS " +
		"where Roleid = " + role.getId() + 
		"and Userid = " + userId;
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	}
	catch (SQLException se)
	{    
	    throw(se);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Lists all the role names associated with a user.
     *
     * @param userId id of the user
     * @param conn   database connection
     *
     * @return Vector of all role names for a user
     *
     * @exception SQLException if some database error occurs   
     */
    public static Vector getRoleNamesForUser(int userId, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    Vector roleIds = new Vector();

	    String sql_statement = "select Roleid " +
		"from LOCO_ROLE_USERS where Userid = " + userId;

	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(sql_statement);

	    while (rs.next())
	    {
		int roleId = rs.getInt("Roleid");
		roleIds.addElement(new Integer(roleId));
	    }

	    rs.close();
	    stmt.close();

	    Vector roleNames = new Vector();
	    Enumeration enum = roleIds.elements();

	    while(enum.hasMoreElements())
	    {
		sql_statement = "select Name from LOCO_ROLES where Id = " +
		    ((Integer)enum.nextElement()).intValue();

		stmt = conn.createStatement();
		rs = stmt.executeQuery (sql_statement);

		rs.next();
		roleNames.addElement(rs.getString("Name"));
	    }

	    return (roleNames);
	}
	catch (SQLException se)
	{
	    throw(se);
	}
	finally
	{
	    if(rs != null) { rs.close(); }
	    if(stmt != null) { stmt.close(); }
	}
    }

    /**
     * Returns a role of all the user's roles OR'ed together.
     *
     * @param userId id of the user
     * @param conn   database connection
     *
     * @return one role of all the roles 'Or'ed together
     *
     * @exception SQLException if some database error occurs   
     */
    public static Role getRolesForUser(int userid, Connection conn)
	throws SQLException
    {
	Vector userRoles = getRoleNamesForUser(userid, conn);

	Role role = new Role("All Roles For Userid " + userid);

	Enumeration e = userRoles.elements();
	while (e.hasMoreElements())
	{
	    role.or(getRole((String)e.nextElement(), conn));
	}

	return (role);
    }

}
    




