/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.perm;

import org.locomotive.util.*;
import org.locomotive.loco.*;

import java.lang.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Maintains users' permissions in the database.
 */
public class UserPerm
{

    /**
     * Sets the users's permissions in the database
     * @param perms the user's permission set
     * @param conn the database connection to use
     * @exception SQLException if some database error occurs
     */
    public static void setPermissions(PermissionSet perms, Connection conn)
	throws SQLException
    {
	int userid = perms.getUserid();

	if (userid > 0)
	{
	    setPermissions(userid, perms, conn);
	}
    }


    /**
     * Sets the a user's permissions in the database.
     *
     * @param userid userid of user
     * @param perms  PermissionSet containing the user's permissions
     * @param conn   database connection object
     * @deprecated - the userid should be in the perms, so we don't need it here
     * @exception SQLException if some database error occurs
     */
    public static void setPermissions(int userid, PermissionSet perms,
				      Connection conn)
	throws SQLException
    {
	deletePermissions(userid, conn);
	
	PreparedStatement pstmt = null;
	String query = 
	    "insert into loco_user_perms (userid, perm_id) values (?, ?)";
	
	try
	{
	    pstmt = conn.prepareStatement(query);
	    
	    boolean [] values = perms.toArray();
	    
	    for (int i = 0; i < values.length; i++)
	    {
		if (values [i])
		{
		    pstmt.setInt(1, userid);
		    pstmt.setInt(2, i);
		    pstmt.executeUpdate();
		}
	    }
	}
	finally 
	{
	    if(pstmt != null) { pstmt.close(); }
	}

    }


    /**
     * Gets a user's permissions from the database.
     *
     * @param userid userid of user
     * @param conn   database connection object
     *
     * @return PermissionSet containing the user's permissions
     *
     * @exception SQLException if some database error occurs
     */
    public static PermissionSet getPermissions(int userid, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	ResultSet rs = null;
	String query = "select perm_id from loco_user_perms where userid = " + 
	    userid + " order by perm_id desc";
	
	PermissionSet perm = null;

	// since we selected the highest perm_id first, we know how big our
	// array will need to be
	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);

	    int perm_id = 0;

	    if (rs.next())
	    {
		perm_id = rs.getInt(1);
	    }
	    else
	    {
		perm = new PermissionSet(userid);
		return (perm);
	    }

	    boolean [] values = new boolean [perm_id + 1];
	    values[perm_id] = true;

	    while (rs.next())
	    {
		perm_id = rs.getInt(1);
		values [perm_id] = true;
	    }
	    
	    perm = new PermissionSet(values);
	    perm.setUserid(userid);
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}
	
	return (perm);
    }		


    /**
     * Deletes a user's permissions in the database.
     *
     * @param userid userid of user
     * @param conn   database connection object
     *
     * @exception SQLException if some database error occurs
     */
    public static void deletePermissions(int userid, Connection conn)
	throws SQLException
    {
	Statement stmt = null;

	try
	{ 
	    String sql_statement = "delete from loco_user_perms " +
		"where Userid = " + userid;
	    
	    stmt = conn.createStatement();

	    stmt.executeUpdate(sql_statement);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }


    /**
     * Deletes many users' permissions from the database.
     *
     * @param userids userid of users to delete the permissions from
     * @param conn   database connection object
     *
     * @exception SQLException if some database error occurs
     */
    public static void deletePermissions(int [] userids, Connection conn)
	throws SQLException
    {
	Statement stmt = null;
	String sql_statement = "delete from loco_user_perms " +
	    "where " + org.locomotive.util.SQL.buildINClause ("userid", userids);

	try
	{ 
	    stmt = conn.createStatement();
	    stmt.executeUpdate(sql_statement);
	}
	finally
	{
	    if(stmt != null) { stmt.close(); }
	}
    }

    /** 
     * Lets you insert multiple permissions for of the users
     * associated with each of the perms' userid.  Besides that its
     * about the same as the one above.  
     * @param perms the permissions that correspond to the userids 
     * @param conn the db connection to use 
     * @exception java.sql.SQLException if we didn't get along well with our 
     *   database 
     */
    public static void setPermissions(PermissionSet [] perms,
				      Connection conn) throws SQLException
    {
	if (perms.length == 0)
	    return;
	
	int [] userids = new int [perms.length];
	for (int i = 0; i < perms.length; i++)
	{
	    userids [i] = perms [i].getUserid();
	}

	deletePermissions(userids, conn);
	
	PreparedStatement pstmt = null;
	String query = 
	    "insert into loco_user_perms (userid, perm_id) values (?, ?)";
	
	try
	{
	    pstmt = conn.prepareStatement(query);
	    for (int j = 0; j < perms.length; j++)
	    {
		if (userids [j] > 0)
		{
		    boolean [] values = perms [j].toArray();
		    
		    for (int i = 0; i < values.length; i++)
		    {
			if (values [i])
			{
			    pstmt.setInt(1, userids [j]);
			    pstmt.setInt(2, i);
			    pstmt.executeUpdate();
			    pstmt.close();
			}
		    }
		}
	    }
	}
	finally 
	{
	    if(pstmt != null) { pstmt.close(); }
	}

    }

    /** 
     * Gets a bunch of users' permissions from the database. 
     * @param userids the userids for the users whose permissions we'll get
     * @param conn the database connection we'll use
     * @exception SQLException if our db doesn't like what we're doing to it
     * @return an array of PermissionSet objects corresponding to each user.
     *         This array will be ordered by the userid.  If any users
     *         in the userids parameter didn't have any permissions, then
     *         the array will just skip over them.
     */
    public static PermissionSet [] getPermissions(Connection conn, 
						  int [] userids)
						  
	throws SQLException
    {
	// hey lets make this a little more efficient in JDK 1.2 by
	// getting rid of the SortableTable and using the Collections API
	PermissionSet [] perms = new PermissionSet[userids.length];

	if (userids.length < 1) // we didn't get any users
	    return (perms);

	SortableTable st = new SortableTable(userids.length, 1);
	for (int i = 0; i < userids.length; i++)
	{
	    st.insert(new Integer(userids[i]), i, 0);
	}

	st.sort(0, SortableTable.NUMBER, SortableTable.ASCENDING);

	Statement stmt = null;
	ResultSet rs = null;
	String query = "select userid, perm_id from loco_user_perms where " +
	    SQL.buildINClause("userid", userids) +
	    " order by userid asc, perm_id desc";
	
	// alrighty- so we have two sorted lists of userids- let's just merge
	// them and stick the bits into our bitsets
	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);

	    int last_userid = ((Number) st.get(0, 0)).intValue();
	    int userid = last_userid;
	    int perm_id = 0;
	    int user_ix = 0;   // what user in the perms array we're on
	    int st_ix   = 0;   // what user in the st we're on
	    boolean [] values = null;
	    boolean finished = false;

	    for (int t = 0; st_ix < userids.length; t++)
	    {
		if (rs.next())
		{
		    userid = rs.getInt(1);
		    perm_id = rs.getInt(2);
		}
		else
		{
		    finished = true;
		}
		
		// we have the same user- add the bit
		if (userid == last_userid)
		{

//		System.out.println("userid: " + userid + " last_userid " + 
//				   last_userid + " perm: " + perm_id + 
//				   " st_ix " + st_ix);


		    switch (t)  
		    {
		    case 0:  // our weird first case
			values = new boolean [perm_id +1];
		    default:
			values [perm_id] = true;
		    }
		}
		else  
		{
		    switch (t)
		    {
		    case 0: break;  // our odd first case
		    default: 
			// first, lets create the permission from the  
			// values array, increment user_ix, create new value
			perms[user_ix] = new PermissionSet(values);
			perms[user_ix++].setUserid(last_userid);

//			System.out.println("new perm[" + user_ix +"]: " +
//					   perms[user_ix - 1].toString());
			
			values = new boolean [perm_id +1];
			values [perm_id] = true;
		    }

		    last_userid = userid;

		    // now, lets run through the st till we find the new user
		    // we've gotten from the result set
		    for (st_ix++; st_ix < userids.length; st_ix++)
		    {
			// the next userid from the st
			int st_userid = ((Number) st.get(st_ix, 0)).intValue();
			
//		System.out.println("userid: " + userid + " last_userid " + 
//				   last_userid + " perm: " + perm_id + 
//				   " st_ix " + st_ix + " st_userid " + 
//				   st_userid);

			// we found the next user to load perms for
			if (userid == st_userid)
			{
			    break;
			}
			else
			{
			    perms[user_ix++] = new PermissionSet(st_userid);

//			System.out.println("new perm[" + user_ix +"]: " +
//					   perms[user_ix - 1].toString());

			}
		    }
		}
		if (finished) userid = -1;

	    }
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}

	return (perms);
    }		
	

    /** An alternative to the above that uses an outer join.  Slightly slower.
     * Gets a bunch of users' permissions from the database. 
     * @param userids the userids for the users whose permissions we'll get
     * @param conn the database connection we'll use
     * @exception SQLException if our db doesn't like what we're doing to it
     * @return an array of PermissionSet objects corresponding to each user.
     *         This array will be ordered by the userid.  If any users
     *         in the userids parameter didn't have any permissions, then
     *         the array will just skip over them.
     */
    static PermissionSet [] getOuterPermissions(Connection conn, 
						       int [] userids)
	throws SQLException
    {
	PermissionSet [] perms = new PermissionSet[userids.length];

	Statement stmt = null;
	ResultSet rs = null;
	String query = "select lu.userid, lup.perm_id from loco_users lu, " +
	    "loco_user_perms lup where " + 
	    SQL.buildINClause("lu.userid", userids) +
	    " and lu.userid = lup.userid (+) " +
	    "order by lu.userid asc, lup.perm_id desc";

	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);

	    int last_userid = 0;
	    int userid = last_userid;
	    int perm_id = 0;
	    int user_ix = 0;   // what user in the perms array we're on
	    boolean [] values = null;

	    for (int t = 0; rs.next(); t++)
	    {
		userid = rs.getInt(1);
		perm_id = rs.getInt(2);

//		System.out.println("userid: " + userid + " last_userid " + 
//				   last_userid + " perm: " + perm_id);

		if (perm_id == 0)  // we have a blank permission for this user
		{
		    perms [user_ix++] = new PermissionSet(userid);
		}
		else if (userid != last_userid)  // we have a new user
		{
		    switch (t)
		    {
		    case 0: break;
		    default:
			perms[user_ix] = new PermissionSet(values);
			perms[user_ix++].setUserid(last_userid);
			
//			System.out.println("new perm[" + user_ix +"]: " +
//					   perms[user_ix - 1].toString());
		    }
		    values = new boolean [perm_id +1];
		    values [perm_id] = true;
		    
		    last_userid = userid;
		}
		else // add this perm to the present user
		{
		    switch (t)
		    {
		    case 0:  // if its the first time through the loop
			values = new boolean [perm_id +1];
		    default:
			values [perm_id] = true;
		    }
		}
	    }
	    perms[user_ix] = new PermissionSet(values);
	    perms[user_ix++].setUserid(last_userid);
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}

	return (perms);
    }		

    /** 
     * Returns a Vector of userids, which represent the users who 
     * have the permission with the specified name.  If the permission doesn't
     * exist, then an empty vector is returned
     * @exception SQLException is there was a problem with the database
     */
    public static Vector usersWhoHave (Connection conn, String permName)
	throws SQLException
    {
	PermissionSet pset = new PermissionSet();
	pset.set(permName);
	return (usersWhoHaveAll(conn, pset));
    }


    /** 
     * Returns a Vector of userids, which represent the users who 
     * have any of the permissions in the specified permission set.
     * Note that using a blank permission set will return all
     * the users who have any permissions at all.
     * @exception SQLException is there was a problem with the database
     */
    public static Vector usersWhoHaveAny (Connection conn, PermissionSet pset)
	throws SQLException
    {
	int size = pset.size();
	int [] temp = new int [size];
	int temp_ix = 0;

	for (int i = 0; i < size; i++)
	{
	    if (pset.isSet(i))
	    {
		temp[temp_ix++] = i;
	    }
	}
	int [] perm_ids = new int [temp_ix];
	System.arraycopy(temp, 0, perm_ids, 0, temp_ix);
	
	String query = "select distinct userid from loco_user_perms where " +
	    SQL.buildINClause("perm_id", perm_ids);

	Statement stmt = null;
	ResultSet rs = null;

	Vector userids = new Vector();
	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);
	    while (rs.next())
	    {
		userids.addElement(new Integer(rs.getInt(1)));
	    }
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}
	
	return (userids);
    }

    /** 
     * Returns a Vector of userids, which represent the users who 
     * have ALL of the permisssions in the specified permission set.  Basically
     * does an INTERSECT of the set of users that have each of the permissions
     * in the permissions set.  Using a blank permission set will return no
     * users.
     * @exception SQLException is there was a problem with the database
     */
    public static Vector usersWhoHaveAll (Connection conn, PermissionSet pset)
	throws SQLException
    {
	Vector userids = new Vector();
	StringBuffer query = new StringBuffer();

	int size = pset.size();
	for (int i = 0; i < size; i++)
	{
	    if (pset.isSet(i))
	    {
		query.append("select userid from loco_user_perms where ");
		query.append("perm_id = ");
		query.append(i);
		query.append(" intersect ");
	    }
	}

	int len = query.length();

	// if we have an empty permission set, return an empty Vector
	if (len == 0)
	{
	    return (userids);
	}

	// trim that last intersect
	query.setLength(len - 11);

	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query.toString());
	    while (rs.next())
	    {
		userids.addElement(new Integer(rs.getInt(1)));
	    }
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}
	
	return (userids);
    }


}

