/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.server.Log;
import org.locomotive.util.*;

import java.io.*;
import java.sql.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * The LocoServlet class is a Servlet class you can use as
 * a superclass for Servlets you're writing to use with the Locomotive.
 * It provides most of the same functionality as the GenericHandler
 * class.  Notice that it implements SingleThreadModel, which means
 * you can safely store stuff in instance variables of your subclasses,
 * making it easier to pass information between methods during a request.
 *
 * <P>There are two alternative ways to use this class: 
 * <ol>
 * <li>You can subclass this class to handle all your user requests.
 * <li>You can create several different subclasses of this class in order
 * to handle different features.  Then, you can create a "master" subclass
 * to route to each of these different subclasses, based on the different URLs.
 * </ol>
 *
 * This class comes with many convenience methods, some of which serve only
 * to make the code more readable; feel free to customize them.
 * @see org.locomotive.loco.handler.GenericHandler
 */
public class LocoServlet extends HttpServlet implements SingleThreadModel
{
    /** The request object we received for this request */
    protected LocoServletRequest req;
    /** The response object to use for this request */
    protected HttpServletResponse resp;
    /** the url tokens for this request. */
    protected String [] url_tokens;
    /** the FormData object that contains any form information from a POST. */
    protected FormData form_data;
    /** the substitutions hashtable to be used when evaluating templates. */
    protected Hashtable steam_vars;
    /** the user associated with this request. */
    protected User user;
    /** the session associated with this request. */
    protected Session session;
    /** The database connection this request can use, if desired. */
    protected Connection conn;
    
    
    /** Log level to use for serious errors; e.g. log(ERROR, "Wait! There's an 
     * error!"); */
    public static final int ERROR = 0;
    /** Log level to use for warnings */
    public static final int WARNING = 1;
    /** log level to use for normal activities */
    public static final int NORMAL = 6;
    /** log level to use for debugging */
    public static final int DEBUG = 9;
    
    
    /** 
     * This will set up the various fields the class requires.  This
     * sh ould be called at the very beginning of each request, and is
     * called in this class's version of service(), so it should
     * be set up for you as long as you don't override service, and only
     * use goGet(), doPost(), etc., as your request handling methods.
     * @see #service
     */
    public void setup(HttpServletRequest req, HttpServletResponse resp)
	throws IOException, ServletException
    {
	this.req = (LocoServletRequest) req;
	this.resp = resp;

	url_tokens = (String []) req.getAttribute("locomotive.url_tokens");
	steam_vars = (Hashtable) req.getAttribute("locomotive.steam_vars");
	conn = (Connection) req.getAttribute("locomotive.connection");
	form_data = (FormData) req.getAttribute("locomotive.form_data");
	session = (Session) req.getSession(true);
	user = (User) session.getValue("locomotive.user");
	
	resp.setContentType("text/html");
	StandardInsertions.insertAll(req, steam_vars);
    }


    /**
     * The service method is the main entry point for Servlets.  All this
     * method does is call the setup() method to get everything ready for
     * the request, and then call super.service() to send the request
     * off to the appropriate do() method (i.e. doGet(), doPost(), etc).
     * <P>If you choose to override this method in a sub class, don't forget
     * to call the setup() method at the beginning of your version.
     * @see #setup
     * @see javax.servlet.http.HttpServlet.html#service
     */ 
    public void service(HttpServletRequest req, HttpServletResponse resp)
	throws IOException, ServletException
    {
	setup(req, resp);
	super.service(req, resp);
    }


    /**
     * writes a message to the locomotive's global server log.
     * @param log_level 0 is most severe.  The higher, the less severe.
     * Messages with log_levels higher than the variable LOG_LEVEL (from
     * the file loco.conf) will be suppressed from being written to the log.
     * @param log_message the message to write to the log.
     */
    public void log (int log_level, String log_message)
    {
	Loco.log (log_level, getClass().getName() + ": " + log_message);
    }
    
    
    /**
     * this is a convenience method you can use when you catch
     * serious Exceptions for which you want a stack trace and 
     * you need to send the user an error code, which can help you debug.
     * @param exception the exception to print to the server log.
     * @param short_code the error message to send to the user
     */
    public void handleException (Exception exception, String short_code)
    {
	Loco.log (0, getClass().getName() + ": ERROR: " + short_code + "." +
		  StringUtil.exceptionStackTrace (exception));
	ServletRoutingTable.sendErrorMsg (resp.SC_INTERNAL_SERVER_ERROR,
					  short_code, req, resp);
    }
    
    
    /**
     * write one event to the global locomotive event log.
     * (this method is just a wrapper around Loco.logEvent().)
     * <P>example:
     * <BLOCKQUOTE>
     * logEvent (5, "login", "submit_form", user.getUsername());
     * </BLOCKQUOTE>
     * @param major the major event region or code section
     * @param minor the minor event code
     * @param data  event specific data
     */
    public void logEvent (String major, String minor, String data)
    {
	Loco.logEvent (5, req.getRequestID(), session, 
		       user, 
		       major, minor, data);
    }
    

    /**
     * if you want to send back a single string as the response to the user,
     * you can use this method.  (You don't need to put your message within
     * &lt;HTML&gt;&lt;BODY&gt;...&lt;/BODY&gt;&lt;/HTML&gt;)
     */
    public void displayString (String message) throws IOException
    {
	if (message == null || message.equals(""))
	{
	    ServletRoutingTable.sendErrorMsg (resp.SC_INTERNAL_SERVER_ERROR,
					      "no message!", req, resp);
	    return;
	}
	
	PrintWriter pw = resp.getWriter();
	if (message.indexOf ("<HTML>") != -1)
	{
	    pw.print(message);
	    pw.flush();
	}
	else
	{
	    pw.print ("<HTML><BODY>" + message + "</BODY></HTML>");
	    pw.flush();
	}
    }
    
    
    
    /**
     * loads a Steam document, inserts the values of Steam variables into that
     * document, and then writes the result back to the Web server.
     * @param document_path -- the file path of that document, relative to the
     * template root.
     * @return a String containing the fully evaluated Steam
     * document, after inserting all the variables.  If the document
     * with that file path does not exist, then returns "", an empty String.
     * If another page has already been displayed, will return "".
     */
    public String displayPage (String document_path, Hashtable steam_vars)
	throws IOException
    {	
	String result = Loco.evaluateTemplate (document_path, steam_vars);
	if (result.equals(""))
	{
	    Loco.log (0, "ERROR: GenericHandler.displayPage(" + document_path +
		     ") was empty!");
	}
	else
	{
	    PrintWriter pw = resp.getWriter();
	    pw.print(result);
	    pw.flush();
	}
	
	return result;
    }
    
    /**
     * simply tells the user that they typed a bad URL, by sending the message:
     * "I'm sorry, I don't understand that URL."
     */
    public void sayBadURL() throws IOException
    {
	resp.sendError(resp.SC_NOT_FOUND, 
		       "I'm sorry, I don't understand that URL.\n");
    }
    
}
