/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.loco.*;
import org.locomotive.server.*;
import org.locomotive.util.*;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.servlet.*;
//import javax.servlet.http.*;


/** 
 * This class is instantiated for each servlet that the Locomotive initializes.
 * An instance should be created when a servlet is initialized, and it is
 * maintained throughout the life of the servlet.
 */
public class LocoServletContext implements ServletContext
{
    private static Hashtable mimes;

    private ServletRoutingTable my_servlet_routing_table;
    private Hashtable attributes;

    static
    {
	/** loads the very basic mimes.  Should we be getting these from
	 * somewhere else?  Loading them from a file, at least?
	 */
	mimes = new Hashtable();
	mimes.put("html",   "text/html");
	mimes.put("htm",    "text/html");
	mimes.put("shtml",  "text/html");
	mimes.put("tmpl",   "text/html");
	mimes.put("dsht",   "text/html");
	mimes.put("gif",    "image/gif");
	mimes.put("jpg",    "image/jpeg");
	mimes.put("jpe",    "image/jpeg");
	mimes.put("jpeg",   "image/jpeg");
	mimes.put("txt",    "text/plain");
	mimes.put("jar",    "application/java-archive");
	mimes.put("zip",    "application/z-zip-compressed");
    }

    public LocoServletContext(ServletRoutingTable servlet_routing_table)
    {
	my_servlet_routing_table = servlet_routing_table;

	attributes = new Hashtable();
    }

    /**
     * Returns the value of the named attribute of the network service,
     * or null if the attribute does not exist.  This looks first for any
     * servlet specific attributes, then looks in the main server config 
     * object.
     */
    public Object getAttribute(String name)
    {
	String config = (String) attributes.get(name);

	if (config == null) try
	{
	    config = Loco.getConfig().getString(name);
	}
	catch (ConfigNotFoundException cnfe) {}

	return (config);
    }

    /**
     * returns an enumeration of all the attribute names in the
     * server.  Presenly returns an enumeration of all the
     * config names in the main server config, as well as any attributes
     * added by any servlets since the server started up.
     */
    public Enumeration getAttributeNames()
    {
	return (new Enumeration () {
	    private Enumeration config = 
		Loco.getConfig().getConfigNamesEnumeration();
	    private Enumeration attr = attributes.keys();

	    public boolean hasMoreElements()
	    {
		return (config.hasMoreElements() || attr.hasMoreElements());
	    }

	    // returns the attributes first. 
	    public Object nextElement()
	    {
		if (attr.hasMoreElements()) return ( attr.nextElement() );
		else return ( config.nextElement() );
	    }
	});
    }
		    
    /** 
     * Sets a server attribute.  Does not do so persistently, however, and 
     * the change is recognized only by other servlets, not by the main
     * servlet config.  To change the main server config object, use
     * Loco.getConfig().changeConfig().
     */
    public void setAttribute(String name, Object o)
    {
	attributes.put(name, o);
    }
    
    /**
     * Removes a server attribute.  Only removes attributes which are not
     * part of the main server however- attributes which are part of the
     * main server cannot be removed.  Removing main server attributes will
     * cause those attributes to be reset to the values that were in the
     * file when the Locomotive was started up.
     */
    public void removeAttribute(String name)
    {
	attributes.remove(name);
    }

    /**
     * Returns the ServletContext for the specified uri.  Presently this
     * will always return this object
     */
    public ServletContext getContext(String uri)
    {
	return ( (ServletContext) this);
    }
    
    /**
     * Returns the major version of the Servlet API this engine implements.
     */
    public int getMajorVersion()
    {
	return (2);
    }

    /**
     * Returns the minor version of the Servlet API this engine implements.
     */
    public int getMinorVersion()
    {
	return (1);
    }
     
    /**
     * returns the mime type for a particular file.  Currently just uses a 
     * bunch of extension heruristics.
     */
    public String getMimeType(String file)
    {
	int index = file.lastIndexOf(".");
	if (index != -1)
	{
	    String suffix = file.substring(index + 1);
	    return ( (String) mimes.get(suffix) );
	}
	else
	{
	    return (null);
	}
    }

    /** 
     * Returns a String containing the real path that corresponds to a virtual
     * path. A virtual path contains a servlet name followed by the name of a
     * file the servlet should act upon, in the form
     * <I>/dir/dir/servlet/file.ext</I>. In this form, <I>file.ext</I> is a
     * filename used instead of the path to the file. The servlet locates the
     * file and translates the file name to the path that locates the file.
     *                                                              <BR><BR>
     * The real path the servlet returns is in a form appropriate to the
     * computer and operating system on which the servlet engine is running,
     * including the proper path separators. This method returns null if the
     * servlet engine cannot translate the virtual path to a real path for
     * any reason.                                                  <BR><BR>
     * @param path a String specifying a virtual path, in the form
     *             /dir/dir/servlet/file.ext
     * @return a String specifying the real path, with path separators
     * appropriate for the system on which the servlet engine is running.
     */
    public String getRealPath(String path)
    {
	Loco.log(9, "LocoServletContext.getRealPath: Called with \"" + path +
                 "\"");

        String real_path = path;

	// Get the ServletRequest for this Thread from the ServletRoutingTable
	LocoServletRequest req = 
	    my_servlet_routing_table.getLocoServletRequest();

	// If we got a null path, return null.
	if (null == path)
        {
	    // We got a null..
	    return null;
	}

	// If we have an empty path, then just return the document root path
	if (path.length() < 1)
	{
	    return req.getDocumentRoot();
	}

        if (null != path && path.length() > 1)
        {
            // Get the "script name" String, like "/locolink"..
            String script_name = req.getScriptName();

            // If the path we got starts with the script_name String,
            // chop that part off..
            if (path.startsWith(script_name) &&
                path.length() >= script_name.length())
            {
                real_path = path.substring(script_name.length());
            }
        }

        // Prepend the web server's document root
        real_path = req.getDocumentRoot() + real_path;

	Loco.log(9, "LocoServletContext.getRealPath: Returning \"" +
		 real_path + "\"");

	return(real_path);
    }

    /**
     * Returns a URL to the resource that is mapped to a specified
     * path. The path must begin with a "/" and is interpreted
     * as relative to the current context root.
     *                                                              <BR><BR>
     * This method allows the servlet container to make a resource 
     * available to servlets from any source. Resources can be
     * located on a local or remote file system, in a database, or
     * in a <code>.war</code> file. 
     *                                                              <BR><BR>
     * The servlet container must implement the URL handlers
     * and <code>URLConnection</code> objects that are necessary
     * to access the resource.
     *                                                              <BR><BR>
     * Currently, there are no custom URLStreamHandlers nor
     * URLConnection classes that come with the Locomotive.  There
     * are, however, at least a couple "standard" ones that the
     * Locomotive can use that come with the JDK (or JRE).  These
     * include URLStreamHandlers for reading local files, files
     * stored in .zip or .jar archive files (including the Servlet
     * 2.2 .war files), HTTP-accessable files, FTP-able files.  Not
     * all JDKs or JREs come with all of these, and it's rather
     * tough to find out exactly which ones are available to any
     * given JDK/JRE.  It looks as if file and http URLStreamHandlers
     * are always available, though, so we should at least be able
     * to use those.
     *                                                              <BR><BR>
     * This method returns <code>null</code> if no resource is
     * mapped to the pathname.
     *                                                              <BR><BR>
     * Some containers may allow writing to the URL returned by
     * this method using the methods of the URL class.  Currently,
     * the Locomotive supports this as long as the URLStreamHandler
     * you try it on supports writing.
     *                                                              <BR><BR>
     * The resource content is returned directly, so be aware that 
     * requesting a <code>.jsp</code> page returns the JSP source
     * code.  Use a <code>RequestDispatcher</code> instead to
     *include results of an execution.
     *                                                              <BR><BR>
     * This method has a different purpose than
     * <code>java.lang.Class.getResource</code>, which looks up
     * resources based on a class loader. This method does not use
     * class loaders.
     *                                                              <BR><BR>
     * @param uri_path a <code>String</code> specifying the path to
     *                 the resource (like "/dir/dir/file.ext")
     * @return the resource located at the named path, or
     *         <code>null</code> if there is no resource at that
     *         path
     * @exception java.net.MalformedURLException if the pathname is
     *            not given in the correct form
     */
    public URL getResource(String uri_path) throws MalformedURLException
    {
	// First, check to make sure that the path starts with a slash
	int index = uri_path.indexOf('/');
	if (index != 0)
	    throw new MalformedURLException("Path must begin with a slash.");

	//
	// Next, check to see if we have an attribute value defined for
	// the attribute named "RESOURCE_BASE_URL".  If so, build a URL out
	// of it and the uri_path and return the URL.
	//
	String base_url = (String) getAttribute("RESOURCE_BASE_URL");
	if (null != base_url) return new URL(base_url + uri_path);

	//
	// Next, check to see if it's a file inside the document root dir..
	//

	// Get the ServletRequest for this Thread from the ServletRoutingTable
	LocoServletRequest req = 
	    my_servlet_routing_table.getLocoServletRequest();

	// Create a File object for the file's absolute filesystem pathname
	File file = new File(req.getDocumentRoot() + uri_path);

	// Check to see if it exists on the filesystem, that it's a regular
	// file (not a directory), and that we can read it
	if (file.exists() && file.isFile() && file.canRead())
	{
	    try
	    {
		return new URL("file", "localhost", -1,
			       file.getCanonicalPath());
	    }
	    catch (IOException e)
	    {
		Loco.log(0, "LocoServletContext.getResource: Couldn't get " +
			 "the cannonical path for file \"" +
			 file.getPath() + "\"");
	    }

	    // Couldn't make a file URL to the file.. problem is logged, so
	    // now we'll return null (saying no resource is mapped to the path)
	    return null;
	}

	//
	// If we still haven't found it, see if we can find it in the
	// CLASSPATH or in a .jar, .zip, or .war file.
	//

	// NOTE: this part is still unimplemented, but basically what
	// we would do here is:
	//
	// 1) Make sure we have the URLStreamHandler that would allow
	//    the creation and use of zip/jar/war file URLs..  to do this
	//    we could simply try to create those kinds of URLs, and if
	//    the java.net.URL constructor doesn't throw a
	//    MalformedURLException then that means that we do have the
	//    necessary URLStreamHandlers.
	// 2) Get the java.class.path (the CLASSPATH) String, and for
	//    each path it contains, do the following if the path is a
	//    .jar file, a .zip file, or a .war file:
	//    - Verify that the file exists, and is readable.
	//    - Verify that the file is a readable archive file (can
	//      look through the entries inside the file).
	//    - Test to see if the archive file contains a file matching
	//      the path that this method was sent (uri_path).  If it does,
	//      then create a proper URL object that points to that file,
	//      and return it.  Otherwise:
        // 3) Move on to the next archive file and try the same things.
	//
	// This stuff should probably be in its own method, since it's
	// going to take quite a bit of code to do it!  Also, for best
	// performance, maybe the Locomotive should cache information
	// about its CLASSPATH's archive files' contents on startup, and
	// then in this method's code just re-use that information every
	// time (plus look for new or modified, not-yet-cached archive
	// files in dirs that may contain new archive files, and cache
	// the new archive files' contents if it finds any).  That may
	// help us in more ways than one.
	//
	// Until we implement all of this, admins and developers who use
	// the Locomotive can make this method return URLs to files stored
	// in archive files by putting a line into the Locomotive's loco.conf
	// that defines RESOURCE_BASE_URL to be a correct base URL for the
	// archive file that they'd like this method to return URLs for.
	// I don't yet know the syntax for these URLs, but I believe it's
	// something like this:
	//
	// If your archive file is called "/usr/local/loco/lib/myjar.jar",
	// then you'd put the following line into your loco.conf:
	// 
	// RESOURCE_BASE_URL     jar://usr/local/loco/lib/myjar.jar/+
	//
	// And, when this method gets called, it will simply append the
	// uri_path String to that base URL, instantiate a new URL object
	// and return it.  The resulting URL String would look something
	// like this:
	//
	// jar://usr/local/loco/lib/myjar.jar/+/archived_dir/somefile.ext
	//
	// If the base URL isn't a useable one, then the caller of this
	// method will get a MalformedURLException.


	// We're giving up here, we found no resource mapped to the
	// uri_path that we were sent..
	return null;
    }

    /**
     * Returns the resource located at the named path as an
     * <code>InputStream</code> object.
     *                                                               <BR><BR>
     * The data in the <code>InputStream</code> can be of any type
     * or length. The path must be specified according to the rules
     * given in <code>getResource</code>.  This method returns
     * <code>null</code> if no resource exists atthe specified path. 
     *                                                               <BR><BR>
     * Meta-information such as content length and content type that
     * is available via <code>getResource</code> method is lost when
     * using this method.
     *                                                               <BR><BR>
     * The servlet container must implement the URL handlers and
     * <code>URLConnection</code> objects necessary to access the
     * resource.
     *                                                               <BR><BR>
     * This method is different from
     * <code>java.lang.Class.getResourceAsStream</code>,
     * which uses a class loader. This method allows servlet
     * containers to make a resource available to a servlet from any
     * location, without using a class loader.
     *                                                               <BR><BR>
     * @param uri_path a <code>String</code> specifying the path to
     *                 the resource (like "/dir/dir/file.ext")
     * @return the <code>InputStream</code> taken from the
     *         URLConnection to the resource, or <code>null</code>
     *         if no resource exists at the specified path.
     */
    public InputStream getResourceAsStream(String uri_path)
    {
	URL resource_url = null;

	try
	{
	    // Call getResource to create a URL Object for this path
	    resource_url = getResource(uri_path);
	}
	catch (MalformedURLException e)
	{
	    // getResource either says that the path doesn't start with a
	    // slash, or that it couldn't create a URL Object for the path..
	    Loco.log(0, "LocoServletContext.getResourceAsStream: Couldn't " +
		     "create a URL Object for path \"" + uri_path + "\"");

	    return null;
	}

	// Now we'll create a URLConnection using the URL that we got from
	// the getResource method, then we'll open its InputStream and
	// return it..
	URLConnection resource_url_connection = null;

	try
	{
	    resource_url_connection = resource_url.openConnection();
	    return resource_url_connection.getInputStream();
	}
	catch (IOException e)
	{
	    // We got an IOException either trying to open the connection,
	    // or trying to get the InputStream from the connection.  Either
	    // way we should log the problem.
	    Loco.log(0, "LocoServletContext.getResourceAsStream: " +
                     StringUtil.exceptionStackTrace(e));
	}

	// For one reason or another, we weren't able to get the InputStream.
	return null;
    }
    
    /**
     * This returns a requestDispatcher object that will allow
     * forwarding to another servlet or handler.  
     * @param uripath the path that denotes the Service to dispatch
     *        the request to
     */
    public RequestDispatcher getRequestDispatcher(String uripath)
    {
	return ((RequestDispatcher) new LocoRequestDispatcher(uripath, this));
    }

    /**
     * Returns the following information about the current server, in the
     * following format: <PRE>
     * SYSTEM_TAG: SYSTEM_INSTANCE_TAG, Locomotive v (loco version), 
     * Servlet engine v (major version).(minor version).
     * </PRE>
     * Here's an example: <PRE>
     * COMPANY_X: sv_one, Locomotive v4.1.0, Servlet Engine v2.1
     * </PRE>
     */
    public String getServerInfo()
    {
	LocoConfig lc = Loco.getConfig();
	return ( lc.getString("SYSTEM_TAG") + ": " + 
		lc.getString("SYSTEM_INSTANCE_ID") + ", " + 
		 "Locomotive v" + Loco.getVersionString() + 
		", Servlet engine v" + getMajorVersion() + "." + 
		 getMinorVersion());
    }

    /** Currently logs to the server log- log level 5 */
    public void log(String msg)
    {
	Loco.log(5, msg);
    }

    /** Logs the message and the stack trace at log level 0 */
    public void log(String msg, Throwable t)
    {
	Loco.log(0, "ERROR: " + msg + ": " + 
		 StringUtil.exceptionStackTrace(t));
    }

    /** Logs the message and the stack trace at log level 0 */
    public void log(Exception e, String msg)
    {
	log(msg, e);
    }

    /*----- deprecated methods --------*/

    /** 
     * returns a servlet keyed by the name.  Deemed too dangerous
     * by the new Servlet API- always returns null.
     * @deprecated - too dangerous
     */
    public Servlet getServlet(String name)
    {
	return ( null );
    }

    /**
     * returns the names of all the servlets presently registered.  Also
     * consided too dangerous.  Returns an empty enumeration.
     * @deprecated because it's too dangerous
     */
    public Enumeration getServletNames()
    {
	return ( new Vector().elements() );
    }
    
    /**
     * returns an enumeration of the registered servlets.  Deemed too dangerous.
     * Returns an empty enumeration.
     * @deprecated because it's too dangerous
     */
    public Enumeration getServlets()
    {
	return ( new Vector().elements() );
    }

    public ServletRoutingTable getServletRoutingTable()
    {
	return my_servlet_routing_table;
    }

    public void setServletRoutingTable(
        ServletRoutingTable servlet_routing_table)
    {
	my_servlet_routing_table = servlet_routing_table;
    }
}
