/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.server.*;
import org.locomotive.util.*;
import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;

import java.io.*;
import java.net.*;
import java.sql.*;
import java.text.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;


/**
 * The Locomotive Servlet Request class provides all the methods in 
 * ServletRequest and HttpServletRequest, and also provides a number
 * of useful attributes.  See the getAttribute() method for more information.
 */
public class LocoServletRequest implements ServletRequest, HttpServletRequest
{
    private int request_id;          // holds the request id from the loco
    private Socket sock;
    private Session session;
    private Hashtable cgi_envars;
    private InputStream is;
    private Hashtable attributes;
    private LocoServletResponse resp;
    private Cookie [] cookies;
    private long original_sid = -1;  // the sid on the cookie- not necessarily
                                     // actual sid
    private Hashtable form_hash = null;
    private boolean used_reader = false;
    private boolean used_is = false;


/*
    public Connection conn = null;
    public Hashtable steam_vars = null;
    public String [] url_tokens = null;
    public long bstamp_id = -1;
    public User user = null;
    public Session session = null;
    */
    

    /** 
     * Package protected constructor should be called only by 
     * ServletRoutingTable
     * @param request_id the RM's unique id for this request
     * @param cgi_envars_hash the CGI variables for this request
     * @param attibutes the attributes for this request- set up by
     *        ServletRoutingTable.routeRequest
     * @param is the inputStream to read the request body from, if any
     * @param resp the resp to use for sending error messages
     * @param form_hash the parsed form data- leave as null if it isn't parsed
     *        yet
     * @param session the session object for this request
     * @param url_tokens the request URI, tokenified
     */
    protected LocoServletRequest(int request_id, Socket sock,
				 Hashtable cgi_envars_hash,
				 Hashtable attributes, InputStream is,
				 LocoServletResponse resp, Hashtable form_hash,
				 Session session)
    {
	this.request_id = request_id;
	this.sock = sock;
	this.cgi_envars = cgi_envars_hash;
	this.attributes = attributes;
	this.is = is;
	this.resp = resp;
	this.form_hash = form_hash;
	this.session = session;
    }


    /** 
     * Returns the unique id for this request.  Ids may be reused,
     * but the IDs for any two concurrent requests will never be identical.
     */
    public int getRequestID()
    {
	return ( request_id );
    }


    /*-------------- ServletRequest Methods -------------------*/

    /**
     * returns the attribute with the specified name, if any.  The following
     * attributes are provided with the Locomotive SerlvetRequest: <PRE>
     *   locomotive.bstamp_id       A Long containing this browser's unique
     *                                identity number.
     *   locomotive.steam_vars      A hashtable which contains STEAM insertion
     *                                variables
     *   locomotive.url_tokens      The request URI, in a tokenized format.
     *   locomotive.connection      A database connection to the default loco DB
     *   locomotive.form_data       A FormData object containing any form
     *                              information posted with this request.
     * </PRE>
     * <P>The Following are supposed to be included with each Servlet manager, 
     * but are not presently part of the Locomotive Servlet Engine: <PRE>
     *   javax.net.ssl.cipher_suite       The string name of the SSL cipher 
     *                                      suite in use, if the request was 
     *                                      made using SSL
     *   javax.net.ssl.peer_certificates  The chain of X.509 certificates which
     *                                      authenticates the client. This is 
     *                                      only available when SSL is used 
     *                                      with client authentication is used.
     *   javax.net.ssl.session            An SSL session object, if the request
     *                                       was made using SSL.
     * </PRE>
     */
    public Object getAttribute(String name)
    {
	// let's make sure our form_data attribute is alright
	if (form_hash == null && name.equals("locomotive.form_data"))
	{
	    parseForm();
	}
		
	return (attributes.get(name));
    }

    
    /**
     * returns an Enumeration of the names of all the attributes associated with
     * this request.
     */
    public Enumeration getAttributeNames()
    {
	return (attributes.keys());
    }


    /**
     * Sets an attribute with the specified name to the specified object.
     * If an attribute already exists with that name, it is overwritten.
     */
    public void setAttribute(String name, Object obj)
    {
	if (obj != null)
	{
	    attributes.put(name, obj);
	}
	else  // if the object is null, at least look like we overwrote 
	{     // whatever was there before
	    attributes.remove(name);
	}
    }


    /**
     * Returns a ServletInputStream for the request body data.  
     * @throw IllegalStateException if getReader() has been called already
     */
    public ServletInputStream getInputStream() throws IOException
    {
	if (used_reader)
	{
	    throw new IllegalStateException("Input reader already initialized");
	}
	
	used_is = true;
	return ( new RequestInputStream(is) );
    }


    /**
     * Returns a BufferedReader of the request body data.
     * @throw IllegalStateException if getInputStream() has been called already
     */
    public BufferedReader getReader() throws IOException
    {
	if (used_is)
	{
	    throw new IllegalStateException("Input Stream already initialized");
	}

	used_reader = true;
	return (new BufferedReader(new InputStreamReader(is)));
    }


    /**
     * Returns the value of the input parameter with the specified name
     */
    public String getParameter(String name)
    {
	if (form_hash == null)
	{
	    parseForm();
	}

	if (form_hash == null)
	{
	    return (null);
	}

	Object o = form_hash.get(name);
	if (o == null)
	{
	    return (null);
	}
	else if (o instanceof String [])  // more than one
	{
	    return ( ((String [] ) o) [0] );
	}
	else
	{
	    return ( (String) o);
	}
    }


    /**
     * Returns an Enumeration of all the form input names.
     */
    public Enumeration getParameterNames()
    {
	if (form_hash == null)
	{
	    parseForm();
	}

	if (form_hash == null)
	{
	    return ( new Vector().elements() );
	}
	else
	{
	    return form_hash.keys();
	}
    }
	
    
    /**
     * Returns an array of all the values for a specified name
     */
    public String [] getParameterValues(String name)
    {
	if (form_hash == null)
	{
	    parseForm();
	}

	if (form_hash == null)
	{
	    return (null);
	}

	Object o = form_hash.get(name);
	if (o == null)
	{
	    return (null);
	}
	else if (o instanceof String [])  // more than one
	{
	    return ( (String []) o );
	}
	else
	{
	    return ( new String [] { (String) o });
	}
    }	


    /**
     * Returns the character encoding for the input body of the request, or
     * null if none has been specified.
     */
    public String getCharacterEncoding()
    {
	String content = (String) cgi_envars.get("CONTENT_TYPE");
	if (content == null)
	{
	    return (null);
	}
	
	int index = content.indexOf("charset=");
	if (index == -1)
	{
	    return (null);
	}

	int end = content.indexOf(";", index+1);
	if (end > 0)
	{
	    return (content.substring (index+8, end));
	}
	else
	{
	    return (content.substring(index+8));
	}
    }


    /**
     * Returns the length of the input content, or -1 if not known 
     */
    public int getContentLength()
    {
	int len = -1;
	try
	{
	    len = Integer.parseInt((String) cgi_envars.get("CONTENT_LENGTH"));
	}
	catch (NumberFormatException nfe) {}

	return (len);
    }


    /**
     * Returns the content (MIME) type of the request body data, or null if
     * not available.  Same as the CGI variable CONTENT_TYPE.
     */
    public String getContentType()
    {
	return ( (String) cgi_envars.get("CONTENT_TYPE") );
    }


    /**
     * Returns the request Protocol- for example, HTTP/1.1.
     * Same as CGI variable SERVER_PROTOCOL
     * Returns null if not available.
     */
    public String getProtocol()
    {
	return ( (String) cgi_envars.get("SERVER_PROTOCOL") );
    }


    /**
     * Returns the IP address of the requestor, or null if not available.
     * Same as the CGI variable REMOTE_ADDR.
     * Returns null if not available.
     */
    public String getRemoteAddr()
    {
	return ( (String) cgi_envars.get("REMOTE_ADDR") );
    }


    /**
     * Returns the Host name of the requestor, if available.  If the full
     * name is not available, the Request IP is returned, if that is available.
     * Same as the CGI variable REMOTE_HOST. Returns null if not available.

     */
    public String getRemoteHost()
    {
	return ( (String) cgi_envars.get("REMOTE_HOST") );
    }


    /**
     * Returns the scheme of the URL used for this request- For example:
     * http, https, ftp. 
     */
    public String getScheme()
    {
	String server_url = (String) cgi_envars.get("SERVER_URL");
	if (server_url == null)
	{
	    return ("http");
	}
	
	try 
	{ 
	    URL url = new URL(server_url);
	    return ( url.getProtocol() );
	}
	catch (MalformedURLException mue)
	{
	    return ("http");
	}
    }

    
    /**
     * Returns the server name.  Same as the CGI variable SERVER_NAME.
     * Returns null if not available.
     */
    public String getServerName()
    {
	return ( (String) cgi_envars.get("SERVER_NAME") );
    }


    /**
     * Returns the Server Port.  Same as the CGI variable SERVER_PORT
     * Returns -1 if not available.
     */
    public int getServerPort()
    {
	int port = -1;
	try
	{
	    port = Integer.parseInt( (String) cgi_envars.get("SERVER_PORT") );
	}
	catch (NumberFormatException nfe) {}

	return ( port );
    }


    /**
     * Returns the web server's document root filesystem path that was sent
     * in with the request.
     *                                                              <BR><BR>
     *
     * @return the document root String, or a null String if the document
     *         root wasn't found in the cgi environment variables.
     */
    public String getDocumentRoot()
    {
	// Get the document root out of the cgi environment variables
	String doc_root = (String) cgi_envars.get("DOCUMENT_ROOT");

	if (null != doc_root)
	{
	    // Make sure that the document root doesn't end with a slash
	    if (doc_root.endsWith("/"))
		doc_root = doc_root.substring(0, doc_root.length() - 1);
	}

	return doc_root;
    }

    /**
     * Returns the "script name".  Same as the cgi environment variable
     * called "SCRIPT_NAME".
     */
    public String getScriptName()
    {
	return ((String) cgi_envars.get("SCRIPT_NAME"));
    }
    
    /** 
     * Returns a String containing the real path that corresponds to a virtual
     * path. A virtual path contains a servlet name followed by the name of a
     * file the servlet should act upon, in the form
     * <I>/dir/dir/servlet/file.ext</I>. In this form, <I>file.ext</I> is a
     * filename used instead of the path to the file. The servlet locates the
     * file and translates the file name to the path that locates the file.
     *                                                              <BR><BR>
     * The real path the servlet returns is in a form appropriate to the
     * computer and operating system on which the servlet engine is running,
     * including the proper path separators. This method returns null if the
     * servlet engine cannot translate the virtual path to a real path for
     * any reason.                                                  <BR><BR>
     * @deprecated Use ServletContext.getRealPath() instead!
     * @param path a String specifying a virtual path, in the form
     *             /dir/dir/servlet/file.ext
     * @return a String specifying the real path, with path separators
     * appropriate for the system on which the servlet engine is running.
     */
    public String getRealPath(String path)
    {
	// Since this method is deprecated, the Servlet spec says that new
	// implementations of this method should just return the same value
	// that ServletContext.getRealPath(String) would return.

	Loco.log(9, "LocoServletRequest.getRealPath: Called with \"" + path +
		 "\"");

        String real_path = path;

	// If we got a null path, return null.
	if (null == path)
        {
	    // We got a null..
	    return null;
	}

	// If we have an empty path, then just return the document root path
	if (path.length() < 1)
	{
	    return getDocumentRoot();
	}

        if (null != path && path.length() > 1)
        {
            // Get the "script name" String, like "/locolink"..
            String script_name = getScriptName();

            // If the path we got starts with the script_name String,
            // chop that part off..
            if (path.startsWith(script_name) &&
                path.length() >= script_name.length())
            {
                real_path = path.substring(script_name.length());
            }
        }

        // Prepend the web server's document root
        real_path = getDocumentRoot() + real_path;

	Loco.log(9, "LocoServletRequest.getRealPath: Returning \"" +
		 real_path + "\"");

	return(real_path);
    }




    /*----------- HttpServletRequest methods ------------*/


    /**
     *  Returns an array of Cookie objects that came in with this request.
     */
    public Cookie [] getCookies()
    {
	if (cookies != null)  // if we did this before, return the result
	{
	    return (cookies);
	}

	// otherwise, parse out the cookies and create them
	String http_cookies = (String) cgi_envars.get("HTTP_COOKIE");
	
	if (http_cookies != null)
	    {
		Vector temp = new Vector();
		StringTokenizer st = new StringTokenizer(http_cookies, " ;");
	    
		while (st.hasMoreTokens())
		    {
			String name_value = st.nextToken();
			int index = name_value.indexOf("=");
			if (index > 0)
			    {
				Cookie c = new Cookie(name_value.substring(0, index), 
						      name_value.substring(index+1));
				temp.addElement(c);
			    }
		    }

		cookies = new Cookie [temp.size()];

		temp.copyInto(cookies);
	    }

	return (cookies);
    }

    
    /**
     * Returns, from the first line of the HTTP request, the part of this
     * request's URL that defines the resource being requested. If there is
     * a query string, it is not included in the return value. For example
     * a request accessed via the URL path of /catalog/books?id=1 would
     * return /catalog/books. The return value of this method contains both
     * the servlet path and the path info. 
     *                                                               <BR><BR>
     *
     * If any part of the URL path was URL encoded, the path must be decoded
     * before being returned by this method. 
     *
     *                                                               <BR><BR>
     * @see javax.servlet.http.HttpRequest#getRequestURI
     */
    public String getRequestURI()
    {
	String request_uri = (String) cgi_envars.get("SCRIPT_NAME") +
	                     cgi_envars.get("REQUEST_URI");

	// If there's a query string on the end, chop it off.
	int index = request_uri.indexOf('?');
	if (index != -1)
	    request_uri = request_uri.substring(0, index);

	// URL-decode the String.
	StringUtil.URLDecode(request_uri);

	Loco.log(9, "LocoServletRequest.getRequestURI: Returning \"" +
		 request_uri + "\"");

	return request_uri;
    }


    /**
     * Returns the part of the request URI that is a query String, if any.
     * Same as the CGI variable QUERY_STRING
     */
    public String getQueryString()
    {
	return ( (String) cgi_envars.get("QUERY_STRING") );
    }


    /**
     * Returns the part of this request's URI that identifies this the servlet
     * to be invoked. This will not return any extra path info information
     * for the request- use getPathInfo() or getRequestURI() for that
     */
    public String getServletPath()
    {
	String tok = ((String []) attributes.get("locomotive.url_tokens")) [0];

	String uri = getRequestURI();
	    
	//        if (null != uri && uri.endsWith(".jsp")) return null;

	int index = uri.indexOf(tok);

	Loco.log(9, "LocoServletRequest.getServletPath: Returning \"" + 
		 uri.substring(0, index + tok.length()) + "\"");

	return (uri.substring(0, index + tok.length()));
    }


    /**
     * Returns the path information that preceded the servlet path of this 
     * request's URI, if any.  Same as the CGI variable PATH_INFO.
     */
    public String getPathInfo()
    {
	String tok = ((String []) attributes.get("locomotive.url_tokens")) [0];
	String uri =  (String) cgi_envars.get("PATH_INFO");
	int place = uri.indexOf(tok) + tok.length();
	if (place > uri.length() + 1)
	{
	    Loco.log(9, "LocoServletRequest.getPathInfo: info = \"\"");
	    return ("");
	}
	else
	{
	    Loco.log(9, "LocoServletRequest.getPathInfo: info = " +
		     uri.substring(place));
	    return (uri.substring(place));
	}
    }


    /**
     * This method gets any extra path information following the servlet
     * path of this request's URL and translates it into a real path. The
     * request URL must be URL decoded before the translation is attempted.
     * If there is no extra path information following the servlet path of
     * the URL, this method returns null. 
     */
    public String getPathTranslated()
    {
        String uri = getRequestURI();


	// Remove the "/locolink" (same as the CGI variable SCRIPT_NAME)
        if (null != uri && uri.length() > 1)
        {
            // Get the "script name" String, like "/locolink"..
            String script_name = getScriptName();

            // If the path we got starts with the script_name String,
            // chop that part off..
            if (uri.startsWith(script_name) &&
                uri.length() >= script_name.length())
            {
                uri = uri.substring(script_name.length());
            }
        }

	// Clip off any query string on the end..
	int index = uri.indexOf('?');
	if (index != -1)
	    uri = uri.substring(0, index);

	// Prepend the document root
	uri = getDocumentRoot() + uri;

	// Replace slash with this OS's File.separator, if necessary.
	String file_separator = File.separator;
	if (!file_separator.equals("/"))
	{
	    index = -1;
	    while ((index = uri.indexOf('/')) != -1)
	    {
		if (index < uri.length() - 1)
		    uri = uri.substring(0, index) + file_separator +
			  uri.substring(index + 1);
		else
		    uri = uri.substring(0, uri.length() - 1) + file_separator;
	    }
	}

	// URL-decode the String.
	StringUtil.URLDecode(uri);

	Loco.log(9, "LocoServletRequest.getPathTranslated: Returning \"" + 
		 uri + "\"");

	return uri;
    }
    
    
    /**
     * Returns the name of the user making this request, if it has been set 
     * via HTTP authentification.  Same as the CGI variable REMOTE_USER
     */
    public String getRemoteUser()
    {
	return ( (String) cgi_envars.get("REMOTE_USER") );
    }

    
    /**
     * Returns the authentification type, if any.  Same as the CGI variable
     * AUTH_TYPE.
     */
    public String getAuthType()
    {
	return ( (String) cgi_envars.get("AUTH_TYPE") );	
    }


    /**
     * Returns the HTTP method for this request. Same as CGI variable 
     * REQUEST_METHOD
     */
    public String getMethod()
    {
	return ( (String) cgi_envars.get("REQUEST_METHOD") );
    }


    /**
     * Gets the value of the Header field with the requested name.
     * The match between the given name and the request header is 
     * case-insensitive. 
     */
    public String getHeader(String headerName)
    {
	Loco.log(9, "LocoServletRequest.getHeader: Called with \"" +
		 headerName + "\"");

	// all of our headers are upper case anyway
	return ( (String) cgi_envars.get(headerName.toUpperCase()) );
    }

    
    /**
     * Returns an integer representation of the value of the header
     * field with the specified name.
     * @throw NumberFormatException if the value can 't be converted
     */
    public int getIntHeader(String headerName)
    {
	String value = (String) cgi_envars.get(headerName);
	if (value == null)
	{
	    return ( -1 );
	}
	else
	{
	    return ( Integer.parseInt(value) );
	}
    }
    

    /** 
     * Returns the value of the requested header converted to a
     * long representing a date expressed in milliseconds since January 1,
     * 1970, 00:00:00GMT. Accepts one of the following three formats, as
     * specified by the HTTP/1.1 spec (RFC 2068, p 20): <PRE>
     *      Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
     *      Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036
     *      Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
     *</PRE>
     * @return the date expressed in milliseconds since January 1,
     * 1970, 00:00:00GMT.  Returns -1 if the header doesn't exist.
     * @throw IllegalArgumentException if the header can't be converted.
     */
    public long getDateHeader(String name)
    {
	String value = (String) cgi_envars.get(name);
	if (value == null)
	{
	    return ( -1 );
	}
	else
	{
	    String [] dates =  { "EEE, dd MMM yyyy HH:mm:ss zzz",
				 "EEEEEE, dd-MMM-yy HH:mm:ss zzz",
				 "EEE MMMM d HH:mm:ss yyyy" };
	    
	    java.util.Date d = null;
	    for (int i = 0; i < dates.length; i++)
	    {
		SimpleDateFormat sdf = new SimpleDateFormat(dates [i]);

		try { sdf.parse(value); }
		catch (ParseException pe) {}
	    }

	    if (d == null)
	    {
		throw new IllegalArgumentException( 
		    "Could not parse field as date");
	    }
	    else
	    {
		return ( d.getTime() );
	    }
	}
    }


    /** 
     * This method returns an enumeration of String objects representing 
     * the header names for this request.  
     */
    public Enumeration getHeaderNames()
    {
	return ( cgi_envars.keys() );
    }

    
    /** 
     * Returns the session id specified with this request. This may
     * differ from the session id in the current session if the session id
     * given by the client was invalid for whatever reason and a new
     * session was created.  
     */
    public String getRequestedSessionId()
    {
	if (original_sid == -1)
	{
	    getSIDFromCookie();
	}

	return (String.valueOf(original_sid));
	
    }


    /**
     * Gets the current valid session associated with this request
     */
    public HttpSession getSession()
    {
	return ( (HttpSession) session );
    }


    /**  
     * Gets the current valid session associated with this request, if
     * create is false or, if necessary, creates a new session for the
     * request, if create is true.  Returns the present session
     * regardless of the parameter, as new sessions are generated automatically
     * when needed.
     */
    public HttpSession getSession(boolean create)
    {
	// we've already created it anyway
	return ( (HttpSession) session );
    }


    /** 
     * Checks whether the session id specified by this request came
     * in as a cookie. (The requested session may not be one returned by
     * the getSession method.)
     * @return true always- Locomotive always uses cookies
     */
    public boolean isRequestedSessionIdFromCookie()
    {
	return (true);
    }
    
    /** 
     * Checks whether the session id specified by this request came
     * in as part of the URL. (The requested session may not be the
     * one returned by the getSession method.)
     * @return false always- Locomotive always uses cookies
     * @deprecated in favor of isRequestSessionIdFromURL()
     */
    public boolean isRequestedSessionIdFromUrl()
    {
	return (false);
    }

    
    /** 
     * Checks whether the session id specified by this request came
     * in as part of the URL. (The requested session may not be the
     * one returned by the getSession method.)
     * @return false always- Locomotive always uses cookies
     */
    public boolean isRequestedSessionIdFromURL()
    {
	return (false);
    }


    /**
     *  Checks whether this request is associated with a session that is 
     * valid in the current session context. 
     */
    public boolean isRequestedSessionIdValid()
    {
	if (original_sid == -1)
	{
	    getSIDFromCookie();
	}
	
	return (session.getID() == original_sid);
    }



    /*------------------ Utility ----------------------*/


    /** We'll use the HttpUtils methods in all cases except where 
     * we have a multipart form- there we'll use the FormParser.
     */
    private void parseForm()
    {
	String method = getMethod();
	String query_str = getQueryString();

	if (used_reader || used_is)
	{
	    return;  // already gave the iinput stream away
	}

	try
	{
	    if (method.equals("GET"))
	    {
		try 
		{
                    form_hash = HttpUtils.parseQueryString(getQueryString());
		}
		catch (IllegalArgumentException iae) 
		{
		    // no form data- create an empty hash
		    form_hash = new Hashtable();
		}

	    }
	    else if (method.equals("POST"))
	    {
		String content = getContentType();
		
		ServletInputStream sis = getInputStream();

		// should we just use ours?
		if (content.indexOf("multipart") != -1)
		{
		    FormParser f = new FormParser(sis, cgi_envars, 
						  Loco.getConfig(),
						  Loco.getServerLog());
		    form_hash = f.parseForm();
		}
		else
		{
		    form_hash = 
			HttpUtils.parsePostData(getContentLength(), sis);
		}
	    }
	    else
	    {
		// weird request- just create the form_hash
		form_hash = new Hashtable();
	    }

	    // stick in the form_data object
	    if (form_hash != null && ! form_hash.isEmpty())
	    {
		attributes.put("locomotive.form_data", new FormData(form_hash));
	    }
	    else
	    {
		attributes.put("locomotive.form_data", 
			      new FormData(new Hashtable()));
	    }

	}
	catch (ParseFormException pfe)
	{
	    Loco.log(0, "ERROR: could not parse form: " + pfe.toString());
	    ServletRoutingTable.sendErrorMsg(
		HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
		"Form parsing error: "+ pfe.getMessage(), this, resp);
	}
	catch (IllegalArgumentException iae)
	{
	    Loco.log(0, "ERROR: could not parse form: " + iae.toString());
	    ServletRoutingTable.sendErrorMsg(
		HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
		"Form parsing error: "+ iae.getMessage(), this, resp);
	}	    
	catch (IOException ioe)
	{
	    Loco.log(0, "ERROR: could not parse form: " + ioe.toString());
	    ServletRoutingTable.sendErrorMsg(
		HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
		"Form parsing error: "+ ioe.getMessage(), this, resp);
	}	    
	    
    }
	
    /** sets the original_sid- get's it from the cookie.  This could be 
     * more efficient- we already got the sid once in the RequestManager...
     */
    private void getSIDFromCookie()
    {
	Cookie [] cookies = getCookies();
	for (int i = 0; i < cookies.length; i++)
	{
	    if (cookies[i].getName() .equals(
		"LOCOSID_" + Loco.getConfig().getString("SYSTEM_TAG")))
	    {
		String cs = cookies[i].getValue();
		StringTokenizer st = new StringTokenizer (cs, "_");
		try { original_sid = Long.parseLong (st.nextToken()); }
		catch (NumberFormatException nfe) 
		{ original_sid = -1; }
		return;
	    }
	}
    }



    /**
     * Will create a HandlerData object with this Request's fields.
     * Used for the RequestDispatcher.forward() and include() methods
     */
    protected HandlerData generateForward()
    {
	HandlerData hd = new HandlerData(request_id);
	hd.sock = sock;
	hd.is = is;
	hd.form_hash = form_hash;
	hd.cgi_envars_hash = cgi_envars;
	hd.handler_uri = getRequestURI();
	hd.session = session;

	hd.os = null; // ?? ach...  Why did we include this in the first place?
	
	hd.user = (User) session.getValue("locomotive.user");
	hd.conn = (Connection) getAttribute("locomotive.connection");
	hd.subs = (Hashtable) getAttribute("locomotive.steam_vars");
	hd.form_data = (FormData) getAttribute("locomotive.form_data");
	hd.bstamp_id = 
	    ( (Number) getAttribute("locomotive.bstamp_id") ).longValue();

	hd.setAttributes(attributes);
	return (hd);
    }
    

    /**
     * Simple InputStream class that implements ServletInputStream
     */
    protected class RequestInputStream extends ServletInputStream 
    {
	private InputStream is;

	protected RequestInputStream(InputStream is)
	{
	    this.is = is;
	}

	/**
	 * our required method
	 */
	public int read() throws IOException
	{
	    return (is.read());
	}

	public int available() throws IOException
	{
	    return (is.available());
	}

	public void close() throws IOException
	{
	    is.close();
	}

	public void mark(int limit)
	{
	    is.mark(limit);
	}
	
	public boolean markSupported()
	{
	    return (is.markSupported());
	}

	public int read(byte [] b) throws IOException
	{
	    return (is.read(b));
	}

	public int read(byte [] b, int off, int len) throws IOException
	{
	    return (is.read(b, off, len));
	}

	public void reset() throws IOException
	{
	    is.reset();
	}

	public long skip(long n) throws IOException
	{
	    return (is.skip(n));
	}
    }
}

