/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import javax.servlet.*;
import javax.servlet.http.*;

import org.locomotive.server.*;
import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;
import org.locomotive.util.*;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;


/**
 * The response class for the servlet interface provides us all the
 * methods required for both ServletResponses and HttpServletResponses.
 * It uses the Response outputStream for its ServletOutputStream, 
 * and its own internal printwriter for the printWriter
 */
public class LocoServletResponse extends Response 
    implements ServletResponse, HttpServletResponse
{

    // servlet vars
    protected boolean used_writer = false;
    protected boolean used_os = false;
    protected int status = SC_OK;
    protected PrintWriter resw = null;


    public LocoServletResponse(OutputStream os, int request_id, int reply_type)
    {
	super(os, request_id, reply_type);
	content_type = "text/plain";
    }


    /*----------------- ServletResponse methods -------------------*/
    

    /** Here for the servlet implementation- don't like this, though.  
     * Doesn't necessarily reflect the size of the content 
     */
    public void setContentLength(int len)
    {
	addHTTPHeader("Content-length", String.valueOf(len));
    }


    /** Retrieve our friendly ServletOutputStream. */
    public ServletOutputStream getOutputStream() throws IOException
    {
	if (used_writer)
	{
	    throw new IllegalStateException (
		"Writer has already been created for this response.");
	}

	if (! header_was_flushed)
	{
	    writeHeader(null);  // throws the IOException
	}

	used_os = true;
	return (resos);
    }

    
    /** Retrieve the Response PrintWriter */
    public PrintWriter getWriter() throws IOException
    {
	if (used_os)
	{
	    throw new IllegalStateException ( 
		"ServletOutputStream has already been created for " +
		"this response.");
	}

	if (! header_was_flushed)
	{
	    writeHeader(null);  // throws the IOException
	}

	if (resw == null)
	{
	    resw = new ResponseWriter (os);
	}
	
	used_writer = true;
	return (resw);
    }

    
    
    /*-------------- HttpServletResponse methods --------------------*/



    /**
     * Encodes a url for use by the sendRedirect method.  As we do not
     * currently keep any information in urls, we return the url unchanged.
     */
    public String encodeRedirectURL(String url)
    {
	return (url);
    }


    /**
     * Encodes a url with any information necessary for session persistence.
     * As we do not currently keep any information in urls, we return the url
     * unchanged.
     */
    public String encodeURL(String url)
    {
	return (url);
    }

    
    /**
     * Sends an error back to the requestor with the specified status code.
     * Status codes are defined in HttpServletResponse.
     * @see javax.http.HttpServletResponse
     */
    public void sendError(int status_code) throws IOException
    {
	sendError(status_code, "");
    }


    /**
     * Sends an error back to the requestor with the specified status code
     * and error message.  Status codes are defined in HttpServletResponse.
     * @see javax.http.HttpServletResponse
     */
    public void sendError(int status_code, String error_msg) 
	throws IOException
    {
	setContentType("text/html");

	status = status_code;

	if (error_msg == null) error_msg = "";
	String response = 
	    "<HTML>\n<HEAD><TITLE>Error</TITLE></HEAD>\n" + 
	    "<BODY BGCOLOR=\"#FFFFFF\">\n\n" +
	    "<H1>There has been an error</H1>\n" + 
	    error_msg + "<BR>\n" +
	    "Please notify the server administrator " +
	    "and send the details of this message.<BR>\n" + 
	    "</BODY>\n</HTML>";

	PrintWriter pw = getWriter();
	pw.println(response);
	pw.flush();
	pw.close();
    }


    /**
     * sends a redirect response to client.  Location must be an absolute
     * URL.
     * @exception IllegalArgumentException if the redirect URL is not absolute
     */
    public void sendRedirect(String location) throws IOException
    {
	try { java.net.URL loc = new java.net.URL(location); }
	catch (java.net.MalformedURLException mue)
	{
	    throw new IllegalArgumentException(mue.getMessage());
	}
	    
	status = SC_MOVED_TEMPORARILY;
	addHTTPHeader("Location", location);
	writeHeader(null);
    }


    /**
     * Sets a response header with the specified name and date.  The date
     * should be specified as the number of seconds since Jan 1, 1970.
     * Sets the date in the following format: <PRE>
          Sun, 06 Nov 1994 08:49:37 GMT  
     * </PRE>
     */
    public void setDateHeader(String name, long date)
    {
	SimpleDateFormat sdf = 
	    new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz");
	java.util.Date d = new java.util.Date(date);
	String date_str = sdf.format(d);
	addHTTPHeader(name, date_str);
    }


    /**
     * Sets a response header with the specified name and integer value.
     */
    public void setIntHeader(String name, int value)
    {
	addHTTPHeader(name, String.valueOf(value));
    }


    /**
     * Sets a response header with the specified name and value.  This
     * will overwrite a header with the same name, if it exists.
     */
    public void setHeader(String name, String value)
    {
	addHTTPHeader(name, value);
    }


    /**
     * Encodes a url with any information necessary for session persistence.
     * As we do not currently keep any information in urls, we return the url
     * unchanged.
     * @deprecated in favor of encodeRedirectURL
     */
    public String encodeRedirectUrl(String url)
    {
	return (encodeRedirectURL(url));
    }


    /*----------------- deprecated --------------------*/


    /**
     * Encodes a url with any information necessary for session persistence.
     * As we do not currently keep any information in urls, we return the url
     * unchanged.
     * @deprecated in favor of encodeURL
     */
    public String encodeUrl(String url)
    {
	return (encodeURL(url));
    }

    
    /**
     * sets the status for this response and sends the response off.
     * @deprecated no no apparent reason...
     */
    public void setStatus(int status_code)
    {
	setStatus(status_code, null);
    }

    /**
     * sets the status for this reponse, and sends the specified message
     * as the body of the response.
     * @deprecated for no apparent reason
     */
    public void setStatus(int status_code, String message)
    {
	status = status_code;
	try
	{
	    PrintWriter pw = getWriter(); // this will flush the header
	    if (null != message)
		pw.println("<HTML>\n<BODY>\n" + message+ "</BODY>\n</HTML>");
	    pw.flush();
	}
	catch (IOException ioe)
	{
	    Loco.log(0, "ERROR: ServletResponse: couldn't write status: " +
		     ioe.toString());
	}
    }



    /**
     * makes sure the status makes it into the header
     */
    public void writeHeader(Log server_log) throws IOException
    {
	addStatusToHeaders();
	super.writeHeader(server_log);
    }


    /** 
     * we don't currently have access to the status line via our tunnels,
     * so we really can't do anything here...
     */
    protected void addStatusToHeaders()
    {
	String status_str = null;
	switch (status)
	{
	case SC_CONTINUE:
	    status_str = "Continue"; break;
	case SC_SWITCHING_PROTOCOLS:
	    status_str = "Switching Protocols"; break;
	case SC_OK:
	    status_str = "OK"; break;
	case SC_CREATED:
	    status_str = "Created"; break;
	case SC_ACCEPTED:
	    status_str = "Accepted"; break;
//	case SC_NON_AUTHORATIVE_INFORMATION:
//	    status_str = "Non Authorative Information"; break;
	case SC_NO_CONTENT:
	    status_str = "No Content"; break;
	case SC_RESET_CONTENT:
	    status_str = "Reset Content"; break;
	case SC_PARTIAL_CONTENT:
	    status_str = "Partial Content"; break;
	case SC_MULTIPLE_CHOICES:
	    status_str = "Multiple Choices"; break;
	case SC_MOVED_PERMANENTLY:
	    status_str = "Moved Permanently"; break;
	case SC_MOVED_TEMPORARILY:
	    status_str = "Moved Temporarily"; break;
	case SC_SEE_OTHER:
	    status_str = "See Other"; break;
	case SC_NOT_MODIFIED:
	    status_str = "Not Modified"; break;
	case SC_USE_PROXY:
	    status_str = "Use Proxy"; break;
	case SC_BAD_REQUEST:
	    status_str = "Bad Request"; break;
	case SC_UNAUTHORIZED:
	    status_str = "Unauthorized"; break;
//	case SC_PAYMENT_REQUESTED:
//	    status_str = "Payment Requested"; break;
	case SC_FORBIDDEN:
	    status_str = "Forbidden"; break;
	case SC_NOT_FOUND:
	    status_str = "Not Found"; break;
	case SC_METHOD_NOT_ALLOWED:
	    status_str = "Method Not Allowed"; break;
	case SC_NOT_ACCEPTABLE:
	    status_str = "Not Acceptable"; break;
//	case SC_PROXY_AUTHENTIFICATION_REQUIRED:
//	    status_str = "Proxy Authentification Required"; break;
	case SC_REQUEST_TIMEOUT:
	    status_str = "Request Timeout"; break;
	case SC_CONFLICT:
	    status_str = "Conflict"; break;
	case SC_GONE:
	    status_str = "Gone"; break;
	case SC_LENGTH_REQUIRED:
	    status_str = "Length Required"; break;
	case SC_PRECONDITION_FAILED:
	    status_str = "Precondition Failed"; break;
	case SC_REQUEST_ENTITY_TOO_LARGE:
	    status_str = "Request Entity Too Large"; break;
	case SC_REQUEST_URI_TOO_LONG:
	    status_str = "Request URI Too Long"; break;
	case SC_UNSUPPORTED_MEDIA_TYPE:
	    status_str = "Unsupported Media Type"; break;
	case SC_INTERNAL_SERVER_ERROR:
	    status_str = "Internal Server Error"; break;
	case SC_NOT_IMPLEMENTED:
	    status_str = "Not Implemented"; break;
	case SC_BAD_GATEWAY:
	    status_str = "Bad Gateway"; break;
	case SC_SERVICE_UNAVAILABLE:
	    status_str = "Service Unavailable"; break;
	case SC_GATEWAY_TIMEOUT:
	    status_str = "Gateway Timeout"; break;
	case SC_HTTP_VERSION_NOT_SUPPORTED:
	    status_str = "HTTP Version Not Supported"; break;
	default:
	    status_str = "Response";
	}


	addHTTPHeader("status", String.valueOf(status) + " " + status_str);
    }

	    
    /**
     * Generates a crippled version of this response which will not 
     * allow headers to be set.
     * It is used by the RequestDispatcher.include() method to provide
     * a safe way to include content of other servlets without adversely
     * affecting the state of the request.  The headers are written before this
     * object is created, so no headers can be created here.
     * If errors are sent here, then
     * the printStream will be closed, and any attempts byt the parent
     * servlet to write responses will result in errors.
     */
    protected Response generateInclude() throws IOException
    {
	if (used_writer)
	{
	    resw.flush();
	}
	else if (used_os)
	{
	    resos.flush();
	}

	/** Little class to be used for the above purpose */
	class CrippledResponse extends LocoServletResponse
	{
	    CrippledResponse(OutputStream os, int request_id, int reply_type,
			     boolean used_os, boolean used_writer)
		throws IOException
	    {
		super(os, request_id, reply_type);
		this.used_os = used_os;
		this.used_writer = used_writer;
	    }
	
	    /** Does nothing */
	    public void writeHeader(Log server_log) throws IOException {}
	}

	
	return ( new CrippledResponse(os, request_id, reply_type,
				      used_os, used_writer) );
    }




    /** 
     * Our little writer is just like the ServletOutputStream in our super 
     * class.
     * It stores everything in a CharArrayWriter, and sends it all
     * out when we flush.  Note that flush will automatically send the
     * headers out, so we don't need to worry about that.
     */
    class ResponseWriter extends PrintWriter
    {
	CharArrayWriter caw;
	private Writer out;

	ResponseWriter(OutputStream os)
	{
	    super(os);
	    out = new OutputStreamWriter(os);
	    caw = new CharArrayWriter (8192);
	}

	/** required */
	public void write(char c)
	{
	    caw.write(c);
	}

	public void write(char [] c)
	{
	    caw.write(c, 0, c.length);
	}

	public void write(char [] c, int start, int len)
	{
	    caw.write(c, start, len);
	}
	
	public void flush() 
	{
	    try
	    {
		caw.writeTo(out);
		try
                {
		    out.flush();
		}
		catch (IOException ioe)
		{
		    Loco.log(0, "ERROR: LocoServletResponse$ResponseWriter: " +
			     "Exception flushing OutputStreamWriter: " +
			     ioe.toString());
		}

		super.flush();
		caw.reset();
	    }
	    catch (IOException ioe)
	    {
		setError();
	    }
	}

	public void reset()
	{
	    caw.reset();
	}
	

	public int size()
	{
	    return (caw.size());
	}

	public void close() 
	{
	    flush();

	    // Close our own OutputStreamWriter (and likely flushing/sending
	    // the servlet's response data).
	    try
	    {
		out.close();
	    }
	    catch (IOException ioe)
	    {
		Loco.log(0, "ERROR: LocoServletResponse$ResponseWriter: " +
			 "Exception closing OutputStreamWriter: " +
			 ioe.toString());
	    }

	    super.close();
	}
    }
}


