/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 2000 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.servlets.MultiThreadedHttpServlet;
import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.server.Log;
import org.locomotive.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import java.io.*;
import java.sql.*;
import java.util.Hashtable;

/**
 * The MultiThreadedLocoServlet class is a Servlet class you can subclass
 * if you want to write a servlet for use with the Locomotive (maybe one
 * that uses STEAM).
 *                                                                   <BR><BR>
 * <I>WARNING: Servlets that are written as subclasses of this class
 * will <B>not</B> run on any servlet containers other than the Locomotive!
 * Only subclass this class if you're sure that you (nor anyone else) will
 * ever need/want to run your servlet in another servlet container.  </I>
 *                                                                   <BR><BR>
 * Notice that as a subclass of MultiThreadedHttpServlet, this class
 * does not implement SingleThreadModel, which means that subclasses
 * of this class will be treated as multithreaded servlets.  Only one
 * instance of each kind of multithreaded servlet will be instantiated
 * in the Locomotive, and that one instance will serve any/all requests
 * simultaneously.
 *
 * @see MultiThreadedHttpServlet
 *
 * @author Jason Brittain <brittainj@diamtech.com>
 */
public class MultiThreadedLocoServlet extends MultiThreadedHttpServlet
{
    /**
     * Log level to use for serious errors; e.g. log(ERROR, "Wait! There's an 
     * error!");
     */
    public static final int ERROR = 0;

    /** Log level to use for warnings */
    public static final int WARNING = 1;

    /** Log level to use for normal activities */
    public static final int NORMAL = 6;

    /** Log level to use for debugging */
    public static final int DEBUG = 9;

    /**
     * The service method is the main entry point for Servlets.  All this
     * method does is call the setup() method to get everything ready for
     * the request, and then call super.service() to send the request
     * off to the appropriate do() method (i.e. doGet(), doPost(), etc).
     * <P>If you choose to override this method in a sub class, don't forget
     * to call the setup() method at the beginning of your version.
     * @see #setup
     * @see javax.servlet.http.HttpServlet.html#service
     */ 
    public void service(HttpServletRequest request,
			HttpServletResponse response)
	throws IOException, ServletException
    {
	StandardInsertions.insertAll(request, getSteamVars(request));

	super.service(request, response);
    }

    /**
     * Writes a message to the Locomotive's global server log.
     *                                                               <BR><BR>
     * @param log_level 0 is most severe.  The higher, the less severe.
     *        Messages with log_levels higher than the config variable
     *        LOG_LEVEL will be suppressed from being written to the log.
     * @param log_message the message to write to the log.
     */
    public void log(int log_level, String log_message)
    {
	Loco.log(log_level, getClass().getName() + ": " + log_message);
    }

    /**
     * Write one event to the global Locomotive event log.
     * (this method is just a wrapper around Loco.logEvent().)
     * <P>example:
     *                                                           <BLOCKQUOTE>
     *   logEvent(5, "login", "submit_form", user.getUsername());
     *                                                          </BLOCKQUOTE>
     * @param major the major event region or code section
     * @param minor the minor event code
     * @param data  event specific data
     * @param request the HttpServletRequest for this Thread
     */
    public void logEvent(String major, String minor, String data,
			 HttpServletRequest request)
    {
	Loco.logEvent(5, ((LocoServletRequest) request).getRequestID(),
		      (Session) request.getSession(true), getUser(request),
		      major, minor, data);
    }

    /**
     * Loads a Steam document, inserts the values of Steam variables into that
     * document, and then writes the result back to the Web server.
     *                                                              <BR><BR>
     * @param document_path the file path of that document, relative
     *        to the template root.
     * @param steam_vars a Hashtable containing STEAM variables to insert
     *        into the template.
     * @param response the HttpServletResponse for this Thread
     * @return a String containing the fully evaluated Steam
     *         document, after inserting all the variables.  If the
     *         document with that file path does not exist, then
     *         returns "", an empty String.  If another page has
     *         already been displayed, will return "".
     */
    public String displayPage(String document_path, Hashtable steam_vars,
			      HttpServletResponse response)
	throws IOException
    {	
	String result = Loco.evaluateTemplate(document_path, steam_vars);
	if (result.equals(""))
	{
	    Loco.log(0, "ERROR: GenericHandler.displayPage(" + document_path +
		     ") was empty!");
	}
	else
	{
	    PrintWriter print_writer = response.getWriter();
	    print_writer.print(result);
	    print_writer.flush();
	}
	
	return result;
    }

    /**
     * Pass this method the ServletRequest Object, and it returns the
     * Locomotive's URL tokens (if they are defined in the request
     * attributes under the key "locomotive.url_tokens").
     * Otherwise, this method returns null.
     *                                                               <BR><BR>
     * @param request the HttpServletRequest for this Thread
     */
    public String[] getURLTokens(HttpServletRequest request)
    {
	return (String []) request.getAttribute("locomotive.url_tokens");
    }

    /**
     * Pass this method the ServletRequest Object, and it returns the
     * Locomotive's FormData for this request (if it is defined in the
     * request attributes under the key "locomotive.form_data").
     * Otherwise, this method returns null.
     *                                                               <BR><BR>
     * @param request the HttpServletRequest for this Thread
     */
    public FormData getFormData(HttpServletRequest request)
    {
	return (FormData) request.getAttribute("locomotive.form_data");
    }

    /**
     * Pass this method the ServletRequest Object, and it returns the
     * Locomotive's SteamVars Hashtable for this request (if it is defined
     * in the request attributes under the key "locomotive.steam_vars").
     * Otherwise, this method returns null.
     *                                                               <BR><BR>
     * @param request the HttpServletRequest for this Thread
     */
    public Hashtable getSteamVars(HttpServletRequest request)
    {
	return (Hashtable) request.getAttribute("locomotive.steam_vars");
    }

    /**
     * Pass this method the ServletRequest Object, and it returns the
     * Locomotive's database Connection for this request (if it is defined
     * in the request attributes under the key "locomotive.connection").
     * Otherwise, this method returns null.
     *                                                               <BR><BR>
     * @param request the HttpServletRequest for this Thread
     */
    public Connection getConnection(HttpServletRequest request)
    {
	return (Connection) request.getAttribute("locomotive.connection");
    }

    /**
     * Pass this method the ServletRequest Object, and it returns the
     * Locomotive's User Object for this request (if it is defined
     * in the request's session values under the key "locomotive.user").
     * Otherwise, this method returns null.
     *                                                               <BR><BR>
     * @param request the HttpServletRequest for this Thread
     */
    public User getUser(HttpServletRequest request)
    {
	HttpSession session = request.getSession(true);
	return (User) session.getValue("locomotive.user");
    }
}

