/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 2000 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.loco.Loco;
import org.locomotive.server.Log;
import org.locomotive.util.ObjectPool;
import org.locomotive.util.StringUtil;

import javax.servlet.http.*;
import javax.servlet.*;

import java.util.Enumeration;
import java.util.Vector;

/**
 * This class is the Thread subclass that makes Servlets for another
 * Object (like ServletEntry or ServletPool) that needs new servlet
 * instances.  ServletFactory has the job of instantiating and
 * initializing servlets, and not doing so when a servlet signals
 * that new servlets from its class are temporarily or permanently
 * unavailable.  Once a ServletFactory instantiates and initializes
 * a servlet, the ServletFactory distributes that servlet by
 * encapsulating it in a ServletFactoryEvent and firing the event
 * off to each of the ServletFactoryListeners.
 *                                                                   <BR><BR>
 * During initialization, the servlet instance can signal that
 * it is not to be placed into active service by throwing an
 * UnavailableException or ServletException.  If a servlet
 * instance throws an exception of this type, it must not be
 * placed into active service, and the instance must be
 * immediately released by the servlet container.  The
 * servlet's destroy() method is not called in this case as
 * initialization was not considered to be successful.
 *                                                                   <BR><BR>
 * After the instance of the failed servlet is released, a
 * new instance may be instantiated and initialized by the
 * container at any time.  The only exception to this rule is
 * if the UnavailableException thrown by the failed servlet
 * which indicates the minimum time of unavailability.  In
 * this case, the container must wait for the minimum time of
 * unavailablity to pass before creating and initializing a
 * new servlet instance.
 *                                                                   <BR><BR>
 * @author Jason Brittain <brittainj@diamtech.com>
 */
public class ServletFactory extends Thread
{
    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>AVAILABLE</I></CODE>, that means that servlets of the
     * class <CODE><I>servlet_class</I></CODE> can be created and initialized.
     */
    public static final int AVAILABLE = 0;

    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>UNAVAILABLE_AND_WAITING</I></CODE>, that means that a
     * ServletFactory won't try to create servlets of the class
     * <CODE><I>servlet_class</I></CODE> until <CODE><I>unavailable_seconds
     * </I></CODE> have elapsed.
     */
    public static final int UNAVAILABLE_AND_WAITING = 1;

    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>PERMANENTLY_UNAVAILABLE</I></CODE>, that means that a
     * ServletFactory won't try to create servlets of the class
     * <CODE><I>servlet_class</I></CODE>.
     */
    public static final int PERMANENTLY_UNAVAILABLE = 2;

    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>SHUTDOWN</I></CODE>, that means that this ServletFactory
     * has been shut down.
     */
    public static final int SHUTDOWN = 255;

    /**
     * If the instance member <CODE><I>default_wait_retries</CODE></I> is
     * set to <CODE><I>UNLIMITED_RETRIES</I></CODE>, that means that a
     * ServletFactory will never stop trying to make servlets.
     */
    public static final int UNLIMITED_RETRIES = -1;

    /**
     * This is the default waiting period that ServletFactories will wait
     * before continuing to instantiate, initialize, and offer new servlet
     * instances after a servlet throws a ServletException while the servlet
     * is either being instantiated or initialized.
     *                                                               <BR><BR>
     * If the servlet specifies permanent unavailability or a specified
     * amount of temporary unavailability by throwing an appropriate
     * UnavailableException (a subclass of ServletException), ServletFactories
     * will <I>not</I> use this <CODE><I>default_waiting_period.  </I></CODE>
     *                                                               <BR><BR>
     * If a servlet throws an UnavailableException during initialization,
     * and the UnavailableException's isPermanent() method returns false,
     * and its getUnavailableSeconds() method returns -1 (signifying a
     * temporary but unspecified duration of unavailability), then the
     * ServletFactory sleeps for this <CODE><I>default_waiting_period
     * </I></CODE> before trying again to instantiate and initialize any
     * more servlets from the class for which the ServletFactory is
     * responsible.  The same thing happens when a servlet throws an
     * instance of ServletException during initialization, since there's
     * nothing in the ServletException to tell the servlet container when
     * to try again.
     *                                                               <BR><BR>
     * This <CODE><I>default_waiting_period</I></CODE> can be specified
     * in the Locomotive's configuration information.
     */
    public static int default_waiting_period = 0;

    /**
     * The default number of retries that ServletFactories will attempt when
     * trying to instantiate and initialize servlets of any given class.
     */
    public static int default_wait_retries = UNLIMITED_RETRIES;

    /**
     * Each class's ServletFactory tries to create servlets this number of
     * times before finally giving up.
     */
    protected int wait_retries = ServletFactory.default_wait_retries;

    /**
     * This variable is used to count how many times this ServletFactory
     * has waited after trying unsuccessfully to instantiate and initialize
     * a servlet.
     */
    protected int wait_count = 0;

    /**
     * If this instance member is non-zero and <CODE><I>availability
     * </I></CODE> has the value <CODE><I>UNAVAILABLE_AND_WAITING</I></CODE>,
     *  ServletFactory will sleep for <CODE><I>unavailable_seconds
     * </I></CODE> seconds before trying again to make a new servlet.
     */
    protected int unavailable_seconds = ServletFactory.default_waiting_period;

    /**
     * This member variable stores the text reason for the unavailability
     * of these servlets.  This reason will be shown to clients.
     */
    protected String unavailability_reason = null;

    /** A collection of Objects listening for ServletFactoryEvents. */
    protected Vector servlet_factory_listeners = new Vector();

    /**
     * This is the Class of servlet(s) that this ServletFactory is
     * responsible for instantiating, initializing, and handing out.
     */
    protected Class servlet_class = null;

    /**
     * This member variable holds the current creation availability status
     * of this ServletFactory's class of servlets.  Its value can be one of
     * the following: <CODE><I>ServletFactory.AVAILABLE, 
     * ServletFactory.UNAVAILABLE_AND_WAITING, 
     * ServletFactory.PERMANENTLY_UNAVAILABLE</I></CODE>.  By default,
     * this field is set to <CODE><I>ServletFactory.AVAILABLE</I></CODE>.
     */
    protected int availability = AVAILABLE;

    /**
     * We need to keep a reference to the ServletConfig with which we're
     * going to initialize servlets.
     */
    private ServletConfig servlet_config = null;

    /**
     * This member variable stores a reference to this ServletFactory's
     * "primary listener" -- an instance of an Object (like ServletEntry or
     * ServletPool) that this ServletFactory gives the ready-to-serve
     * servlets to.  The primary listener makes the servlet available
     * to handle requests.  Regardless of the Object's type, it implements
     * the ServletFactoryListener interface, and this ServletFactory can
     * call that interface's methods on the <CODE><I>primary_listener
     * </I></CODE> Object.
     */
    protected ServletFactoryListener primary_listener = null;

    /**
     * This member variable stores the minimum number of servlets that
     * this ServletFactory's primary listener needs to keep on hand.
     */
    protected int min_pool_size = 0;

    /**
     * This member variable stores the maximum number of servlets that
     * this ServletFactory's primary listener can hold or use.
     */
    protected int max_pool_size = 0;

    /**
     * This member variable stores the current number of servlets that
     * this ServletFactory's primary listener currently has.
     */
    protected int cur_pool_size = 0;

    /**
     * This member variable simply holds a reference to this ServletFactory's
     * own Thread instance.
     */
    protected Thread servlet_factorys_thread = null;

    /**
     * This member variable is used when we've created the minimum
     * number of servlets and they're all in use, and more requests
     * come in needing more servlets.  As long as we're not going
     * above our max we'll create more servlets.  This variable holds
     * the number of new servlets that this ServletFactory should make
     * if it can.
     */
    protected int need_more_servlets = 0;

    /**
     * This ServletFactory Thread will keep looping (running) as long as
     * nothing has called our shutdown() method.  Shutdown causes this
     * <I>finished</I> boolean to become true.  This Thread's run() method
     * will just return, ending the execution of the Thread.
     */
    protected boolean finished = false;
    
    /**
     * Creates a new ServletFactory Object (not yet started, though).
     *
     * @param listener              the Object that this ServletFactory is
     *                              making servlets for.  As of this writing,
     *                              this listener should either be a
     *                              ServletPool Object, or a ServletEntry
     *                              Object.  It's usually the Object that
     *                              instantiated this ServletFactory.
     * @param servlet_class         the Class object for the servlet we're
     *                              supposed to create.  It should implement
     *                              the javax.servlet.Servlet interface.
     * @param servlet_config        a reference to the ServletConfig -- used
     *                              for configuration information and for
     *                              looking up some request information.
     */
    protected ServletFactory(ServletFactoryListener listener,
			     Class servlet_class, ServletConfig servlet_config)
    {
	super();

	if (null != listener) addServletFactoryListener(listener);

	if (null != servlet_class)
	{
	    this.servlet_class = servlet_class;

	    // Set the name of this Thread so anyone looking from a debugger
	    // will know what this Thread is for.
	    setName("ServletFactory-" + servlet_class.getName());

	    // Make this Thread a daemon Thread -- it should exit if all
	    // the other Threads did.
	    this.setDaemon(true);
	}

	this.servlet_config = servlet_config;
    }

    /**
     * After another Object calls the ServletFactory.start() method,
     * this class's run() method is called.  This marks the starting
     * point of the ServletFactory's own Thread.
     */
    public void run()
    {
	//Loco.log(Log.DEBUG,
	//    "ServletFactory::run() for " + servlet_class.getName());

	//
	// If we don't have at least one listener, or if we don't have
	// a servlet class to work with, then we can't do our job.
	//

	// First, we'll check to make sure we have at least one listener
	if (servlet_factory_listeners.isEmpty())
	{
	    // We couldn't run because we don't have even one listener..
	    return;
	}

	// Next, we'll check to make sure we have a servlet class
	if (null == servlet_class)
	{
	    // We don't have a servlet class to instantiate from..
	    return;
	}

	// We'll find out how many servlets we're supposed to make
	// initially, how many is the max we should make, and how
	// many our primary_listener says it currently has..
	getCounts();

	Loco.log(Log.DEBUG, "ServletFactory::run(): for " +
		 servlet_class.getName() + " min: " + min_pool_size +
		 " max: " + max_pool_size + " cur: " + cur_pool_size);

	// We'll also keep a copy of the reference to our own thread.
	// Some methods below need this reference to know which thread is
	// using the method.
	servlet_factorys_thread = Thread.currentThread();

	// Start the main loop that makes servlets.
        runFactory();
    }

    /**
     * This method calls some ServletFactoryListener methods on our
     * primary listener to find out the current values for the minimum
     * number of servlets we should keep on hand, the maximum number
     * of servlets we should keep on hand, and the current number of
     * servlets that are in service.  The ServletFactory needs these
     * servlet counts so that it knows when it should make more servlets
     * and when it should stop and sleep.  These values are kept up to
     * date so that if some other Object modified these values in the
     * primary listener at runtime, then the ServletFactory may take
     * action on those new values right away.
     */
    protected void getCounts()
    {
	min_pool_size = primary_listener.getMinServletCount();
	max_pool_size = primary_listener.getMaxServletCount();
	cur_pool_size = primary_listener.getCurrentServletCount();
    }

    /**
     * This method is called by the run() method, and should never return.
     * It contains the Factory's main loop algorithm, and knows when to
     * make more servlets and when not to.  When it is done making servlets
     * (enough are made -- no more are needed), or when it can't make any
     * more, it simply wait()s until it is notified that it needs to
     * perform more work.  In a healthy environment, that would probably
     * never happen, and the Factory will stay wait()ing forever, but
     * that doesn't take up valueable CPU time since its thread simply
     * stays blocked (not running on the CPU until it's needed again).
     */
    void runFactory()
    {
	Loco.log(Log.DEBUG, "ServletFactory::runFactory() for " +
		 servlet_class.getName());

	// Here's the top of the main loop:
	while (!finished)  // loop until someone calls shutdown()
	{
	    // Reset the wait_count
	    wait_count = 0;

	    // Now we'll check to see if we need to make more servlets.
	    // If we're below our minimum, we'll make more.
	    // This next loop runs until either we have enough servlets,
	    // or until they are PERMANENTLY_UNAVAILABLE.
	    while (availability != PERMANENTLY_UNAVAILABLE &&
                   (cur_pool_size < min_pool_size ||
		    (need_more_servlets > 0 && cur_pool_size < max_pool_size)))
	    {
		Loco.log(Log.DEBUG,
			 "ServletFactory::runFactory(): Looping for " +
			 servlet_class.getName() + " min: " +
			 min_pool_size + " max: " + max_pool_size +
			 " cur: " + cur_pool_size);

		// Try to make a new Servlet
		Servlet servlet = makeServlet();

		// If we successfully made a servlet, send it to the
		// listeners.
		if (null != servlet)
		{
		    // Success.

		    // Broadcast the servlet (inside an event) to all
		    // of the ServletFactoryListeners.
		    fireServletFactoryEvent(
		        new ServletFactoryEvent(this, servlet));

		    // If we should decrement our need_more_servlets variable
		    // (since we just made and gave out a servlet), then
		    // decrement it in a multithreaded way.
		    synchronized (this)
		    {
			if (need_more_servlets > 0) need_more_servlets--;
		    }

		    //Loco.log(Log.DEBUG, "ServletFactory.init: made " +
		    //         cur_pool_size + " servlet(s) of type " +
		    //         servlet_class.getName());
		}
		else
		{
		    // We were unsuccessful in making the new servlet,
		    // and the listeners have already been notified.

		    // If we've now got temporary unavailability, then
		    // we need to wait..
                    if (availability == UNAVAILABLE_AND_WAITING)
		    {
			// If we haven't retried enough times to make
			// servlets, or if we're supposed to just keep
			// trying, then we will sleep and then try again.
			if (wait_retries == UNLIMITED_RETRIES ||
			    wait_count < wait_retries)
			{
			    try
			    {
				// Try to sleep for the number of seconds
				// stored in the unavailable_seconds variable.
				if (unavailable_seconds > 0)
				    Thread.sleep(unavailable_seconds * 1000);

				// We increment the number of times we've
				// waited to retry..
				wait_count++;
			    }
			    catch (InterruptedException e)
			    {
				// Don't you just hate it when you don't
				// get enough sleep?  Aargh!

				// We won't count this as a wait_count, since
				// we didn't get to sleep for the entire
				// unavailability duration.

				// Program flow could end up here if another
				// Object is making these servlets available
				// again after being temporarily unavailable..
			    }
			}
                        else
			{
			    // Okay, we've retried enough times, we'll give
			    // up on these servlets.  We'll mark them as
			    // PERMANENTLY_UNAVAILABLE (causing this
			    // servlet-making loop to end).
			    setAvailability(PERMANENTLY_UNAVAILABLE);
			}
		    }
		}

		// Now get the current sevlet counts again..
		getCounts();

	    } // end of servlet-making loop

	    // If request Threads have asked for more servlets, and we're
	    // already at the maximum servlet pool size, then we don't need
	    // to make any more.  We'll reset the variable to zero..
	    if (need_more_servlets > 0 && cur_pool_size >= max_pool_size)
		need_more_servlets = 0;

	    // As long as we have enough of these servlets on hand, or
	    // if these servlets are permanently unavailable, we wait.
	    while ((availability == PERMANENTLY_UNAVAILABLE ||
		    (cur_pool_size >= min_pool_size &&
		     need_more_servlets == 0)) && !finished)
	    {
		synchronized (this)
		{
		    try
		    {
			Loco.log(Log.DEBUG, "ServletFactory for class " +
				 servlet_class.toString() + " waiting...");

			// Wait here until some other Thread tells this
			// ServletFactory that new servlets of this class are
			// available again (if that ever happens..)  This
			// ServletFactory Thread should block here and not
			// use up more CPU time..
			wait();

			Loco.log(Log.DEBUG, "ServletFactory for class " +
				 servlet_class.toString() + " awakened...");
		    }
		    catch (Exception exception)
		    {
			Loco.log(Log.WARNING,
				 "Error: ServletFactory for servlet(s) " +
				 "of type " + servlet_class.getName() + ": " +
				 StringUtil.exceptionStackTrace(exception));
		    }
		}
	    } // end of PERMANENTLY_UNAVAILABLE or full loop
	} // end of the main loop

	//
	// If we get here, then this ServletFactory was shut down.
	//
	synchronized(this)
	{
	    // The status is now SHUTDOWN.
	    setAvailability(ServletFactory.SHUTDOWN);

	    // We'll send the listeners an event so they are notified of the
	    // shutdown (if they need more servlets, they're just going to
	    // have to make a new ServletFactory!).
	    setUnavailableSeconds(0);
	    setUnavailabilityReason("ServletFactory shutdown complete.");
            makeAndFireStatusEvent();
	}

        Loco.log(Log.DEBUG, "ServletFactory for " + servlet_class.getName() +
                 " shut down.");

	// The Thread ends here.
    }

    /**
     * Tries to instantiate and initialize a servlet.  If it is successful,
     * the servlet is passed back to the caller of this method (at which
     * time the servlet can be handed out to a listener).  If a problem
     * occurs, this method fires an unavailability ServletFactoryEvent to
     * the listeners, and returns null.
     */
    protected synchronized Servlet makeServlet()
    {
	Servlet servlet = null;
	ServletFactoryEvent servlet_factory_event = null;

	// First create a new instance of the servlet
	try
	{
	    servlet = instantiateServlet();
	}
	catch (Exception e)
	{
	    // Servlets of this type are currently unavailable due to this
	    // Exception (regardless of which exception it is).
	    Loco.log(Log.WARNING, "Couldn't instantiate servlet from class " + 
		     servlet_class.getName() + ": " + 
		     StringUtil.exceptionStackTrace(e));

	    // If the servlet is multithreaded then that was our only
	    // instance that just failed to instantiate..
	    if (ServletRoutingTable.single_thread_model_class.isAssignableFrom(
	        servlet_class))
	    {
		// Set the availability to UNAVAILABLE_AND_WAITING.. we'll try
		// again pretty soon just in case this problem is temporary.
		setAvailability(UNAVAILABLE_AND_WAITING);

		// Set the unavailability timeout (for now, we'll try this
		// again every couple minutes -- it's not likely to work
		// anyway!)
		// FIXME: We need to calculate the unavailable seconds here!
		setUnavailableSeconds(120);

		// Set the reason.
		setUnavailabilityReason(
		    "Couldn't create an instance of the servlet");

		// Tell the listeners what happened.
		makeAndFireStatusEvent();
	    }

	    // Don't return a servlet (they're not available now)
	    return null;
	}

	// Then, initialize the servlet
	try
	{
	    initializeServlet(servlet);
	}
	// During initialization, the servlet instance can signal that
	// it is not to be placed into active service by throwing an
	// UnavailableException or ServletException.  If a servlet
	// instance throws an exception of this type, it must not be
	// placed into active service, and the instance must be
	// immediately released by the servlet container.  The
	// servlet's destroy() method is not called in this case as
	// initialization was not considered to be successful.
	//
	// After the instance of the failed servlet is released, a
	// new instance may be instantiated and initialized by the
	// container at any time.  The only exception to this rule is
	// if the UnavailableException thrown by the failed servlet
	// which indicates the minimum time of unavailability.  In
	// this case, the container must wait for the minimum time of
	// unavailablity to pass before creating and initializing a
	// new servlet instance.
	catch (UnavailableException unavailable_exception)
	{
	    // Servlets of this type are currently unavailable.
	    Loco.log(Log.WARNING, "Couldn't initialize servlet from class " + 
	        servlet_class.getName() + ": UnavailableException: " +
		(unavailable_exception.isPermanent() ? "permanent" :
		 "temporary(" + unavailable_exception.getUnavailableSeconds() +
		 ")") + ", Reason: " + unavailable_exception.getMessage());

	    // Use this UnavailableException to change state to unavailability
	    // and notify the listeners of the change..
	    triggerUnavailability(unavailable_exception);

	    // Don't return a servlet (they're not available now)
	    return null;
	}
	catch (ServletException servlet_exception)
	{
	    // Unspecified unavailability init problem, servlet threw a
	    // ServletException.
	    Loco.log(Log.WARNING, "Couldn't initialize servlet from class " + 
		     servlet_class.getName() + ": " + 
		     StringUtil.exceptionStackTrace(servlet_exception));

	    // These servlets are now temporarily unavailable.  There
	    // is no timeout because this is only a ServletException,
	    // not an UnavailableException.  So, we'll wait some amount
	    // of time, after which these servlets will be available again..

	    // Use this ServletException to change state to unavailability
	    // and notify the listeners of the change..
	    triggerUnavailability(servlet_exception);

	    // Don't return a servlet (they're not available now)
	    return null;
	}

	return servlet;
    }

    /**
     * Instantiates a new servlet and returns a reference to it.
     *                                                               <BR><BR>
     * @exception ClassCastException     if the class wasn't a servlet.
     * @exception InstantiationException if the class couldn't be instantiated.
     * @exception IllegalAccessException 
     */
    protected synchronized Servlet instantiateServlet()
        throws ClassCastException, InstantiationException,
	       IllegalAccessException
    {
	Servlet s = (Servlet) servlet_class.newInstance();

	return s;
    }    

    /**
     * Initializes the servlet by calling the init method of the servlet,
     * passing in this ServletFactory's servlet_config Object.
     *                                                               <BR><BR>
     * @param servlet_to_init the servlet that this method should
     *        initialize.
     * @exception ServletException if the servlet couldn't be initialized.
     */
    protected synchronized void initializeServlet(Servlet servlet_to_init)
        throws ServletException
    {
	servlet_to_init.init(servlet_config);
    }

    /**
     * Returns this ServletFactory's copy of the value of the primary
     * ServletFactoryListener's minimum pool size.
     */
    public int getMinPoolSize()
    {
	return min_pool_size;
    }

    /**
     * Sets the ServletFactory's copy of the value of the primary
     * ServletFactoryListener's minimum pool size.
     *                                                               <BR><BR>
     * @param new_size the new minimum size of the pool.
     */
    public synchronized void setMinPoolSize(int new_size)
    {
	min_pool_size = new_size;
    }

    /**
     * Returns this ServletFactory's copy of the value of the primary
     * ServletFactoryListener's maximum pool size.
     */
    public int getMaxPoolSize()
    {
	return max_pool_size;
    }

    /**
     * Sets the ServletFactory's copy of the value of the primary
     * ServletFactoryListener's maximum pool size.
     *                                                               <BR><BR>
     * @param new_size the new maximum size of the pool.
     */
    public synchronized void setMaxPoolSize(int new_size)
    {
	max_pool_size = new_size;
    }

    /**
     * Returns this ServletFactory's copy of the value of the primary
     * ServletFactoryListener's current pool size.
     */
    public int getCurPoolSize()
    {
	return cur_pool_size;
    }

    /**
     * Sets the ServletFactory's copy of the value of the primary
     * ServletFactoryListener's current pool size.
     *                                                               <BR><BR>
     * @param new_size the new current size of the pool.
     */
    public synchronized void setCurPoolSize(int new_size)
    {
	cur_pool_size = new_size;
    }

    /**
     * Adds the specified ServletFactoryListener to the list of
     * listeners to which this ServletFactory will broadcast events.
     */
    public synchronized void addServletFactoryListener(
        ServletFactoryListener listener)
    {
	if (null == listener) return;

	// If this listener is the first listener we're adding, then
	// it's also the primary_listener
	if (servlet_factory_listeners.isEmpty())
	    primary_listener = listener;

	// Add a listener if it is not already registered
	if (!servlet_factory_listeners.contains(listener))
	    servlet_factory_listeners.addElement(listener);
    }

    /**
     * Removes the specified ServletFactoryListener from the list of
     * listeners to which this ServletFactory will broadcast events.
     */
    public synchronized void removeServletFactoryListener(
        ServletFactoryListener listener)
    {
	if (null == listener) return;

	// Remove the listener if it is registered
	if (servlet_factory_listeners.contains(listener))
	    servlet_factory_listeners.removeElement(listener);
    }

    /**
     * This method takes a ServletException or an UnavailableException
     * and changes the state of the ServletFactory to an unavailable
     * state (as specified by the exception), and creates and fires an
     * event about the state change to the ServletFactoryListeners.
     *                                                               <BR><BR>
     * @param servlet_exception either a ServletException or an
     *        UnavailableException that contains information about
     *        the unavailability.
     */
    public synchronized void triggerUnavailability(
        ServletException servlet_exception)
    {
	// First, we need to know if it's an UnavailableException or not
	if (servlet_exception instanceof UnavailableException)
	{
	    UnavailableException unavailable_exception =
		(UnavailableException) servlet_exception;

	    if (unavailable_exception.isPermanent())
	    {
		// These servlets are now permanently unavailable, and will
		// stay that way until either the container is restarted,
		// or until some other Object sets a different availability
		// value on this ServletFactory via the setAvailability()
		// method..

		// Set the current availability to PERMANENTLY_UNAVAILABLE
		setAvailability(PERMANENTLY_UNAVAILABLE);
	    }
	    else
	    {
		// These servlets are now temporarily unavailable.  The
		// UnavailableException may or may not specify a timeout.
		// Either way, we'll wait some amount of time, after which
		// these servlets will be available again..

		// Set the unavailability timeout from the value inside the
		// UnavailableException..
		setUnavailableSeconds(
	            unavailable_exception.getUnavailableSeconds());

		// Set the current availability to UNAVAILABLE_AND_WAITING
		setAvailability(UNAVAILABLE_AND_WAITING);
	    }
	}
	else
	{
	    // It was just a ServletException, not an UnavailableException,
	    // so we need to just set the defaults..
	    // FIXME: We need to calculate the unavailable seconds here!
            setUnavailableSeconds(ServletFactory.default_waiting_period);
	    setAvailability(UNAVAILABLE_AND_WAITING);
	}

	// Set the reason.
	String message = servlet_exception.getMessage();
	if (null == message) message =
	    "Servlet initialization failed for an unknown reason.";
	setUnavailabilityReason(message);

	// We'll create and fire a ServletFactoryEvent that tells the
	// listeners about this..
	makeAndFireStatusEvent();
    }

    /**
     * Broadcasts the specified ServletFactoryEvent to all of the known
     * ServletFactoryListener Objects.
     */
    public synchronized void fireServletFactoryEvent(
        ServletFactoryEvent servlet_factory_event)
    {
	// Loop through all of the listeners and send them each a reference
	// to this servlet_factory_event..
	for (Enumeration e = servlet_factory_listeners.elements();
	     e.hasMoreElements();)
	{
	    ServletFactoryListener listener =
		(ServletFactoryListener) e.nextElement();
	    listener.handleServletFactoryEvent(servlet_factory_event);
	}
    }

    /**
     * Regardless of the current availability status of the servlets,
     * this method instantiates a new ServletFactoryEvent with the
     * servlet_class, availability, unavailable_seconds, and
     * unavailability_reason values, and then fires that event to the
     * listeners.  This is usually used for firing events that signal
     * unavailability, but it can also be used to show that the servlets
     * are currently available.
     */
    public synchronized void makeAndFireStatusEvent()
    {
	// Create a ServletFactoryEvent for the type of availability
	// and timeout that we currently have..
	ServletFactoryEvent event = new ServletFactoryEvent(this,
            servlet_class, availability, unavailable_seconds,
	    unavailability_reason);

	// Broadcast the event to all listeners
	fireServletFactoryEvent(event);
    }

    /**
     * Returns the Class of servlets that this ServletFactory is managing.
     */
    public Class getServletClass()
    {
	return servlet_class;
    }

    /**
     * Gets the number of seconds that this ServletFactory will wait (or is
     * currently waiting) before trying to make more servlets.
     */
    public int getUnavailableSeconds()
    {
	return unavailable_seconds;
    }

    /**
     * Sets the number of seconds that this ServletFactory should wait before
     * trying to make more servlets.  This method should be called before
     * calling <CODE><I>setAvailability(ServletFactory.UNAVAILABLE_AND_WAITING)
     * </I></CODE>.
     */
    public synchronized void setUnavailableSeconds(int unavailable_seconds)
    {
	this.unavailable_seconds = unavailable_seconds;
    }

    /**
     * Returns the availability status of new, initialized instances of
     * servlets whose Class is <CODE><I>servlet_class</I></CODE>.
     */
    public int getAvailability()
    {
	return availability;
    }

    /**
     * Sets the availability code for this ServletFactory's class of
     * servlets.  Calling this method and changing the availability
     * state for a class of servlets will automatically notify the
     * ServletFactory's Thread of the state change.
     *                                                               <BR><BR>
     * @param availability The new availability state, should be one
     *        of the following: <CODE><I>ServletFactory.AVAILABLE,
     *        ServletFactory.UNAVAILABLE_AND_WAITING,
     *        ServletFactory.PERMANENTLY_UNAVAILABLE,             
     *        ServletFactory.SHUTDOWN.                            </I></CODE>
     */
    public synchronized void setAvailability(int availability)
    {
	boolean should_notify = false;

	// Check to see if the state is changing..
	if (this.availability != availability) should_notify = true;

	// Don't notify this same Thread if it's a SHUTDOWN.
	if (availability == ServletFactory.SHUTDOWN) should_notify = false;

	// Set the new availability value
	this.availability = availability;

	// Notify the ServletFactory's Thread if we should..
	if (should_notify &&
	    Thread.currentThread() != servlet_factorys_thread)
	{
	    // Notify the ServletFactory's Thread that the availability has
	    // changed.
	    try
	    {
		notify();
	    }
	    catch (IllegalMonitorStateException e)
	    {
		// This should never happen, since this method is synchronized
		// and we should already be the owner of the Object's monitor.
	    }
	}
    }

    /**
     * Returns the reason for the unavailability.
     */
    public String getUnavailabilityReason()
    {
	return unavailability_reason;
    }

    /**
     * Sets the reason for the unavailability.
     *                                                               <BR><BR>
     * @param reason the text reason to use.
     */
    public synchronized void setUnavailabilityReason(String reason)
    {
	unavailability_reason = reason;
    }

    /**
     * Increments the <CODE><I>need_more_servlets</I></CODE> int member
     * variable, and then notifies the ServletFactory's Thread that the value
     * has changed.  The ServletFactory will then create another servlet,
     * and then decrement the <CODE><I>need_more_servlets</I></CODE> member
     * variable.
     */
    public synchronized void needMoreServlets()
    {
	// Increment our counter for the number of additional servlets we need.
	need_more_servlets++;

	// We need to notify the ServletFactory's Thread that it needs to
	// make more servlets.
	notify();
    }

    /**
     * Call this method to shut down the ServletFactory (it can't be
     * restarted.. if you want it to work again you'll have to instantiate
     * a new one).
     */
    public synchronized void shutdown()
    {
	// Set the boolean that causes the runFactory() method to end.
	finished = true;

	// Notify the ServletFactory's Thread to wake up and smell the boolean!
	notify();
    }
}
