/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 2000 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import javax.servlet.Servlet;

import java.util.Date;
import java.util.EventObject;

/**
 * Instances of this class are used by ServletFactory Threads to make
 * Servlet instances available to ServletEntry and to ServletPool.
 * The Servlets have been instantiated and properly initialized by the
 * ServletFactory, and they are ready to be used to handle requests.
 *
 * @author Jason Brittain <brittainj@diamtech.com>
 */
public class ServletFactoryEvent extends EventObject
{
    /**
     * A reference to a servlet that the ServletFactory is passing to
     * the listener(s).. it is supposed to be ready to serve requests.
     */
    protected Servlet servlet = null;

    /**
     * The Class of servlets that the ServletFactory creates.
     * FIXME: This will have to change soon to handle more than one
     *        differently-configured servlet of the same class in the
     *        same container.
     */
    protected Class servlet_class = null;

    /**
     * The availability status of the servlets.
     */
    protected int availability = ServletFactory.AVAILABLE;

    /**
     * The number of seconds that this kind of servlets are unavailable for.
     */
    protected int unavailable_seconds = 0;

    /**
     * The reason for the unavailability.
     */
    protected String unavailability_reason = null;

    /**
     * The date/timestamp of when this event occurred.  It is stored as
     * the number of milliseconds elapsed between Jan 1, 1970 00:00:00 GMT
     * and the actual time this event occurred.
     */
    protected long event_timestamp = 0L;

    /**
     * Constructs a ServletFactoryEvent with a Servlet that is ready to
     * serve requests.
     * @param Object source the Object that created this event, probably
     *        a ServletFactory.
     * @param servlet a Servlet instance that has been initialized and
     *        is ready to serve requests.
     */
    public ServletFactoryEvent(Object source, Servlet servlet)
    {
	super(source);
	this.servlet = servlet;
	if (null != servlet) this.servlet_class = servlet.getClass();
	this.event_timestamp = new Date().getTime();
    }

    /**
     * Constructs a ServletFactoryEvent for specifying that servlets of
     * the specified class are unavailable.
     *
     * @param Object source the Object that created this event, probably
     *        a ServletFactory.
     * @param servlet_class the class of servlets that are currently
     *        unavailable.
     * @param availability the current availability status for the class
     *        of servlets that this ServletFactoryEvent's ServletFactory
     *        makes.  Note that this value reflects the availability
     *        status of the servlets as of the time that this event was
     *        created, and that the value could be different by the time
     *        the listener acts on it.
     * @param unavailable_seconds the number of seconds that this
     *        ServletFactoryEvent's ServletFactory is saying that
     *        its servlets are unavailable for.
     * @param unavailability_reason the reason for the unavailability.
     */
    public ServletFactoryEvent(Object source, Class servlet_class,
			       int availability, int unavailable_seconds,
			       String unavailability_reason)
    {
	super(source);
	this.servlet_class = servlet_class;
	this.availability = availability;
	this.unavailable_seconds = unavailable_seconds;
	this.unavailability_reason = unavailability_reason;
	this.event_timestamp = new Date().getTime();
    }

    /**
     * If this method returns null, then this ServletFactoryEvent was
     * sent by the ServletFactory to tell the listeners that the servlets
     * that it creates are currently unavailable.
     */
    public Servlet getServlet()
    {
        return servlet;
    }

    /**
     * Returns the class of servlets that this ServletFactoryEvent affects.
     */
    public Class getServletClass()
    {
	return servlet_class;
    }

    /**
     * Returns the availability status for the class of servlets that are
     * made by the ServletFactory that created this ServletFactoryEvent.
     * This value should be one of the following:                   <CODE><I>
     *          ServletFactory.AVAILABLE,
     *          ServletFactory.UNAVAILABLE_AND_WAITING,
     *          ServletFactory.PERMANENTLY_UNAVAILABLE.           </I></CODE>
     */
    public int getAvailability()
    {
	return availability;
    }

    /**
     * Gets the number of seconds that this ServletFactoryEvent's
     * ServletFactory will wait (or is currently waiting) before trying
     * to make more servlets.
     */
    public int getUnavailableSeconds()
    {
	return unavailable_seconds;
    }

    /**
     * Get the reason for the unavailability.  In the case where the
     * event is an availability event, this method returns null.
     */
    public String getUnavailabilityReason()
    {
	return unavailability_reason;
    }

    /**
     * Gets the timestamp for when this event occurred, according to the
     * object that created this ServletFactoryEvent.
     */
    public long getEventTimestamp()
    {
	return event_timestamp;
    }
}
