/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 2000 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import javax.servlet.Servlet;

import java.util.EventListener;

/**
 * A ServletFactoryListener is any Object that implements this interface
 * and is interested in the events (and Objects) generated by a
 * ServletFactory.  Usually, a ServletFactoryListener is an Object that
 * is using a ServletFactory to instantiate and initialize a servlet (or
 * servlets) for it.  We'll call that listener a "primary listener".  But,
 * ServletFactory and this Listener interface were written so any Object
 * could also listen in to the events generated by a ServletFactory -- 
 * for monitoring reasons.  Any listener that is simply monitoring the
 * ServletFactory's activity in this way should be thought of as a
 * "secondary listener".  Primary listeners are instances of classes like
 * ServletPool, that actually use the servlets that the ServletFactory is
 * making.  Any/all secondary listeners are just listening in.  A 
 * ServletFactory will broadcast its events to all of its listeners.
 *
 * @author Jason Brittain <brittainj@diamtech.com>
 */
public interface ServletFactoryListener extends EventListener
{
    /**
     * The ServletFactory calls this method when it needs to know how
     * many servlets that the primary listener (like a ServletPool or
     * a ServletEntry) needs to keep active and ready to serve.
     */
    public int getMinServletCount();

    /**
     * The ServletFactory calls this method when it needs to know the
     * maximum number of servlets that the primary listener (like a
     * ServletPool or a ServletEntry) will hold..  If the listener
     * has all the servlets it needs, then the ServletFactory doesn't
     * need to make any more unless the primary listener asks for more.
     */
    public int getMaxServletCount();

    /**
     * The ServletFactory calls this method when it needs to know how
     * many servlets that the primary listener (like a ServletPool or
     * a ServletEntry) currently holds.
     */
    public int getCurrentServletCount();

    /**
     * The ServletFactory calls this method when it has either successfully
     * instantiated and initialized a servlet (and the new servlet is
     * ready to serve requests), or when the ServletFactory needs to signal
     * the ServletFactoryListener(s) that some other event has occurred.  In
     * the case that a new servlet instance is ready, the new servlet is
     * encapsulated inside the ServletFactoryEvent that is passed to this
     * method.  It is up to the ServletFactoryListener to make the servlet
     * available to clients.
     *                                                              <BR><BR>
     * @param servlet_factory_event an EventObject subclass from the
     *        ServletFactory that (in this case) encapsulates a new
     *        servlet that is ready to serve requests.
     * @see ServletFactoryEvent
     */
    public void handleServletFactoryEvent(
        ServletFactoryEvent servlet_factory_event);
}
