/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.loco.*;
import org.locomotive.server.*;
import org.locomotive.util.ObjectPool;

import javax.servlet.http.*;
import javax.servlet.*;

import java.util.*;

/**
 * Holds a pool of servlets.  It's used in the ServletRoutingTable for
 * Servlets that implement SingleThreadModel.  This creates some
 * servlets (currently, the number of minimum Threads that are created
 * at startup), and scales to a maximum number of servlets (currently
 * the maximum number of request Threads that are allowed by the server).
 * This ServletPool class works in conjunction with the ServletFactory
 * to create and maintain a pool of servlets that are ready to serve
 * requests.
 */
public class ServletPool extends ObjectPool implements ServletFactoryListener
{
    /**
     * The Class of servlets that this pool holds.
     */
    private Class servlet_class;

    /**
     * The ServletConfig Object to use to configure new servlets.
     */
    private ServletConfig servlet_config;

    /**
     * Our ServletFactory -- it will make servlets for us and hand them over.
     */
    private ServletFactory servlet_factory = null;

    /**
     * The ServletEntry that encapsulates this ServletPool.  We need a
     * reference to it so that we may propogate unavailability info if
     * we need to.
     */
    ServletRoutingTable.ServletEntry servlet_entry;

    /**
     * Creates the new ServletPool.  Use LOCO_MIN_POOLED_THREADS for
     * our min, and LOCO_MAX_POOLED_THREADS for our max (for now.. but
     * this should be configurable.  See the inline comments in the
     * source code for this constructor).  Logs stuff to the server log
     * at log level 5.
     * @param servlet_class the class of servlets that will populate this
     *        ServletPool.
     * @param servlet_config the ServletConfig Object that holds any config
     *        information for the servlets to use.
     */
    protected ServletPool(Class servlet_class, ServletConfig servlet_config,
			  ServletRoutingTable.ServletEntry servlet_entry)
	throws InitializationException
    {
	// FIXME: Later, we should allow an administrator to set the
	//        ServletPool's min and max sizes to something other
	//        than the same sizes as the request Thread pool sizes.
	//        What if one request uses multiple servlets of the
	//        same kind?  What if very few request Threads use
	//        one kind of servlet (like an admin servlet for example)?
	//        It should be configurable.  -JB
	super(Loco.getConfig().getInt("LOCO_MIN_POOLED_THREADS"),
	      Loco.getConfig().getInt("LOCO_MAX_POOLED_THREADS"),
	      Loco.getServerLog(), 5);

	this.servlet_class = servlet_class;
	this.servlet_config = servlet_config;
	this.servlet_entry = servlet_entry;
    }

    /**
     * This method overrides ObjectPool.initialize().  This needs to be
     * called after the pool is created.
     *                                                               <BR><BR>
     * ServletPools create their own ServletFactory Thread that does
     * the job of instantiating and initializing servlets for this pool,
     * asynchronously, as well as waiting and retrying to instantiate and
     * initialize more servlets if there was any temporary unavailability.
     *                                                               <BR><BR>
     * While this ServletPool's ServletFactory does its work, the
     * ServletRoutingTable tries to get servlets from this ServletPool.
     * If it calls ServletPool.getObject() and there aren't any servlets
     * in the pool, getObject() returns a reference to the only instance
     * of the UnavailableServlet.  The UnavailableServlet is a multithreaded
     * servlet that will serve the client a 503 - Service Unavailable
     * message.
     */
    public synchronized void initialize() throws InitializationException
    {
	// Get a new ServletFactory Thread and start it up, instantiating
	// and initializing servlets for us in another Thread..  This Object
	// gets the servlets delivered to it via the servletReady() method.
        getServletFactory();

	// Start up the ServletFactory if it isn't already running
	if (!servlet_factory.isAlive()) servlet_factory.start();
    }

    /**
     * ServletPool does not need an implementation of this method, since
     * the ServletFactory creates them for ServletPool and hands them over
     * when they're ready to serve.
     */
    public Object createNewObject() throws InitializationException
    {
	return null;
    }

    /**
     * Use this method to get a servlet from the ServletPool.
     *                                                               <BR><BR>
     * Under normal conditions, this method will simply get a servlet from
     * the pool and return it right away.  Regardless, this method should
     * always return a servlet Object - though it may be the unavailable
     * servlet.
     *                                                               <BR><BR>
     * While this ServletPool's ServletFactory does its work, the
     * ServletRoutingTable tries to get servlets from this ServletPool.
     * This method is called by a request Thread, and tries to get a
     * servlet from the pool, and if there isn't one waiting in the
     * pool it notifies the ServletFactory's Thread that it should make
     * one.  After it notifies the ServletFactory, it will wait a short
     * period of time for the new servlet to arrive in the pool so that
     * the request can be served correctly.  Request Threads can't wait
     * forever, so if no new servlet shows up then this method times
     * out and returns the unavailable servlet (hopefully sending a
     * <I>503 - Service Unavailable</I> message to the client).  
     *                                                               <BR><BR>
     * If this method calls ServletPool.getServletFromPool() and there
     * aren't any servlets in the pool, getObject() eventually returns
     * a reference to the only instance of the UnavailableServlet.  The
     * UnavailableServlet is a multithreaded servlet that will serve the
     * client a <I>503 - Service Unavailable message</I>.
     *                                                               <BR><BR>
     * Other Objects using a ServletPool should call this method to
     * get a servlet from the pool instead of calling getServletFromPool()!
     * This method is not synchronized, and it tries harder to return a
     * servlet of the correct type without causing too much synchronization
     * lock waiting.  Also, getServletFromPool() does not contain any
     * code to grow the pool up to its maximum pool size -- this method
     * does that.
     *                                                               <BR><BR>
     * @return a servlet Object.
     */
    public Object getObject()
    {
	// Get a servlet from the pool, if there are any..
	Servlet servlet = getServletFromPool();

	// If we got the unavailable servlet, then maybe more are being made..
	if (servlet == ServletRoutingTable.unavailable_servlet)
	{
	    // If the servlets are available, maybe more should be made..
	    if (servlet_entry.getAvailability() == ServletFactory.AVAILABLE)
	    {
		// They're available, so if the stack is empty we should
		// ask the ServletFactory to make one for us, then wait for
		// a small amount of time for one while it's being made.
		if (pe_stack.empty())
		{
		    // The stack was empty.. if that's because they're all in
		    // use, and we're not yet at the max, then we'll prod the
		    // ServletFactory to make more.
		    if (pe_in_use_hash.size() == current_open &&
			current_open < max_open)
		    {
			// We'll make sure that our ServletFactory gets busy
			// making at least one more servlet for us.
			getServletFactory().needMoreServlets();
		    }

		    // Then, we'll wait for a little bit to see if we get
		    // a servlet that we can use.  We'll wait in quarter-
		    // second intervals for a total of 2 and a half seconds
		    // before giving up and sending a 503 - Service Unavailable
		    // message to the client.  If we don't get a new servlet
		    // within 2.5 seconds, chances are that we're not going
		    // to get one soon, and we can't have the user waiting
		    // very long for a response.  Feel free to adjust this
		    // timout to your own values, but remember that this
		    // code is likely to time out only during conditions of
		    // heavy request load, and that the delay could be even
		    // longer than 2.5 seconds if lots of threads are calling
		    // getObject() and causing synchronization lock waits on
		    // this ServletPool Object.
		    for (int i = 0; i < 10; i++)
		    {
			try
			{
			    // Sleep one quarter of a second
			    Thread.sleep(250);

			    // If we can get a servlet, then return it.
			    Servlet new_servlet = getServletFromPool();
			    if (new_servlet !=
				ServletRoutingTable.unavailable_servlet)
				return new_servlet; // Good servlet, return it!
			}
			catch (InterruptedException e)
			{
			    // We don't need to do anything here..
			}
		    }

		    // If we got here, then we timed out waiting for a
		    // servlet to be made.  We should probably warn the
		    // system administrator that servlet resources were
		    // scarce..
		    Loco.log(Log.WARNING, "ServletPool::getObject(): timed " +
		        "out waiting for ServletFactory to make a new " +
			"servlet of type " + servlet_class.getName() + " -- " +
			"set the ServletPool's minimum size to a larger " +
			"value or get a faster server machine (or both).");
		}
	    }
	}

	return servlet;  // Return whatever we have, unavailable or not..
    }

    /**
     * Gets a servlet from this pool.
     *                                                               <BR><BR>
     * This Object cannot simply create and return a new servlet instance,
     * since only the ServletFactory can instantiate and initialize
     * servlets.  This is because servlet instances of this pool's type
     * can't be created during servlet-specified downtime, so the
     * ServletFactory Thread keeps track of when servlet instances can
     * (and can't) be created.
     *                                                               <BR><BR>
     * Only getObject() should make calls to this method.  This method
     * is synchronized (for multithreaded use) and should only be
     * called when necessary -- too many calls could cause lots of
     * synchronization lock waiting, and slow down the Locomotive.
     *                                                               <BR><BR>
     * @return a servlet from the pool, or the UnavailableServlet if
     *         there were no servlets waiting in the pool.
     */
    public synchronized Servlet getServletFromPool() 
    {
	PoolEntry pe;
	
	// If a servlet is available on the Stack
	if (! pe_stack.empty())
	{
	    pe = (PoolEntry) pe_stack.pop();

	    log("pe " + pe.id + " : " + pe.usage_count);

	    pe_in_use_hash.put(pe.object, pe);

	    pe.usage_count++;
	    return ((Servlet) pe.object);
	}

	// if we don't have one, return a reference to the UnavailableServlet
	return ServletRoutingTable.unavailable_servlet;
    }

    /**
     * This method overrides ObjectPool.wipeFromPool(Object).  When a
     * ServletPool wipes a servlet from the pool, it must notify its
     * ServletFactory that it might need to make another servlet.
     * This method removes the specified Object from the pool.  Usually
     * used if an Object has been put into some bad state.
     *                                                              <BR><BR>
     * @param obj the Object to remove from the pool.
     */
    public synchronized void wipeFromPool (Object obj)
    {
	PoolEntry pe = (PoolEntry) pe_in_use_hash.get (obj);
	pe_in_use_hash.remove (obj);

	if (pe != null)
	{
	    current_open--;

	    log ("removed pe " + pe.id + " after " + pe.usage_count + 
		 " uses");   
	}

	destroyObject(obj);

	// Now we need to notify our ServletFactory that it may need to
	// make a replacement servlet for the one we just destroyed.
	synchronized (getServletFactory())
        {
	    servlet_factory.notify();
	}
    }

    /** Calls destroy() */
    public void destroyObject(Object obj)
    {
	if (obj instanceof Servlet && !(obj instanceof UnavailableServlet))
	{
	    ( (Servlet) obj).destroy();
	}

    }

    /**
     * Checks to make sure that the Object you're trying to return is
     * an instance of a Servlet, and if so this method returns that
     * Servlet instance to the pool.
     */
    public void returnToPool(Object obj)
    {
	if (obj instanceof Servlet)
	{
	    super.returnToPool(obj);
	}
    }
    
    /** Basic exception handler. */
    private void handleException(String err_msg, Throwable t) 
	throws InitializationException
    {
	throw new InitializationException (err_msg + ": " + t.toString());
    }

    /**
     * Set the minimum size of the ServletPool.  This method overrides
     * the method <CODE><I>ObjectPool.setMinSize()</I></CODE>.  This
     * implementation also notifies the ServletFactory of the change
     * (if there is one).
     *                                                              <BR><BR>
     * @param minSize The minimum size of the object pool.
     */
    public void setMinSize(int minSize)
    {
	this.min_open = minSize;

	// If we have a ServletFactory at the time that this method is
	// invoked, then we'll notify ServletFactory of the change as well..
	if (null != servlet_factory)
	{
	    servlet_factory.setMinPoolSize(minSize);
	}
    }

    /**
     * Set the maximum size of the object pool.  This method overrides
     * the method <CODE><I>ObjectPool.setMaxSize()</I></CODE>.  This
     * implementation also notifies the ServletFactory of the change
     * (if there is one).
     *                                                              <BR><BR>
     * @param maxSize The maximum size of the object pool.
     */
    public void setMaxSize(int maxSize)
    {
	this.max_open = maxSize;

	// If we have a ServletFactory at the time that this method is
	// invoked, then we'll notify ServletFactory of the change as well..
	if (null != servlet_factory)
	{
	    servlet_factory.setMaxPoolSize(maxSize);
	}
    }

    /**
     * Returns a reference to this ServletPool's ServletFactory Object.
     * If one doesn't already exist, a new one is instantiated, and a
     * reference to it is returned.
     *                                                              <BR><BR>
     * @return a reference to this ServletPool's ServletFactory.
     */
    public ServletFactory getServletFactory()
    {
	if (null == servlet_factory)
	{
	    servlet_factory =
		new ServletFactory(this, servlet_class, servlet_config);
	}

	return servlet_factory;
    }

    public void setServletFactory(ServletFactory new_servlet_factory)
    {
	servlet_factory = new_servlet_factory;
    }


    //--------------- ServletFactoryListener methods ------------------

    /**
     * The ServletFactory calls this method when it needs to know how
     * many servlets that the primary listener (like a ServletPool or
     * a ServletEntry) needs to keep active and ready to serve.
     */
    public int getMinServletCount()
    {
	return min_open;
    }

    /**
     * The ServletFactory calls this method when it needs to know the
     * maximum number of servlets that the primary listener (like a
     * ServletPool or a ServletEntry) will hold..  If the listener
     * has all the servlets it needs, then the ServletFactory doesn't
     * need to make any more unless the primary listener asks for more.
     */
    public int getMaxServletCount()
    {
	return max_open;
    }

    /**
     * The ServletFactory calls this method when it needs to know how
     * many servlets that the primary listener (like a ServletPool or
     * a ServletEntry) currently holds.
     */
    public int getCurrentServletCount()
    {
	return current_open;
    }

    /**
     * The ServletFactory calls this method when it has either
     * successfully instantiated and initialized a servlet (and the
     * new servlet is ready to serve requests), or when the
     * ServletFactory needs to signal the ServletFactoryListener(s)
     * that some other event has occurred.  In the case that a new
     * servlet instance is ready, the new servlet is encapsulated
     * inside the ServletFactoryEvent that is passed to this method.
     * It is up to the ServletFactoryListener to make the servlet
     * available to clients.
     *                                                              <BR><BR>
     * @param servlet_factory_event an event that tells this
     *        ServletFactoryListener Object that a servlet is
     *        ready to serve requests (in which case the servlet
     *        is encapsulated in the event Object), or that the
     *        servlet is unavailable.
     * @see ServletFactoryEvent
     */
    public synchronized void handleServletFactoryEvent(
        ServletFactoryEvent servlet_factory_event)
    {
        // If we got an event from a ServletFactory other than our
        // current ServletFactory (maybe one that we just replaced)
        // then ignore events from it!
        if (servlet_factory_event.getSource() != servlet_factory) return;

	// First, we'll store the event timestamp as the date/time
	// of the last event that occurred.
	servlet_entry.setLastAvailabilityEventTimestamp(
            servlet_factory_event.getEventTimestamp());

	int availability = servlet_factory_event.getAvailability();
	int unavailable_seconds = 0;
	String unavailability_reason = null;

	// Find out the availability from the event
	if (availability == ServletFactory.AVAILABLE)
	{
	    // Grab the servlet out of the event, and put it into the pool!
	    Servlet new_servlet = servlet_factory_event.getServlet();

	    // If the servlet isn't a null, we'll use it.
	    if (null != new_servlet)
	    {
		// Put it into the ServletPool..
		pe_stack.push(new PoolEntry(new_servlet));

		current_open++;
	    }
	}
	else if (availability == ServletFactory.UNAVAILABLE_AND_WAITING)
	{
	    // We now know that the servlet is currently unavailable,
	    // and that the ServletFactory is going to wait and then
	    // retry to make it.

	    // We'll keep a copy of how long the ServletFactory
	    // is going to wait before retrying..
	    unavailable_seconds =
		servlet_factory_event.getUnavailableSeconds();

	    // Get the reason why the servlet is temporarily unavailable.
	    unavailability_reason =
		servlet_factory_event.getUnavailabilityReason();
	}
	else if (availability == ServletFactory.PERMANENTLY_UNAVAILABLE)
	{
	    // Get the reason why the servlet is temporarily unavailable.
	    unavailability_reason =
		servlet_factory_event.getUnavailabilityReason();
	}
	else if (availability == ServletFactory.SHUTDOWN)
	{
	    // So far, we don't need to do anything with shutdown events.
	    // That's because all we would do is de-reference the
	    // ServletFactory and allow it to be garbage collected.
	    // The shutdown method of this Object already does that.
	    // But, I suppose we should log the shutdown.
	    Loco.log(Log.INFO, "ServletEntry for " + servlet_class.getName() +
		     ": " +
		     servlet_factory_event.getUnavailabilityReason());
	}

	// Set the availability in our servlet_entry
	servlet_entry.setAvailability(availability);

	// Set the unavailable_seconds
	servlet_entry.setUnavailableSeconds(unavailable_seconds);

	// Set the unavailability_reason
	if (availability != ServletFactory.AVAILABLE &&
	    null == unavailability_reason) unavailability_reason = "Unknown.";
	servlet_entry.setUnavailabilityReason(unavailability_reason);
    }
}
