/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.server.*;
import org.locomotive.util.*;
import org.locomotive.loco.*;

import java.net.*;
import java.io.*;
import java.util.*;
import java.sql.*;

import javax.servlet.*;
import javax.servlet.http.*;


/** 
 * This class controls servlet activity for a particular servlet routing
 * table.  An instance should be created at startup and maintained
 * throughout the life of the server.  The send a servlet on it's
 * way, use the routeRequest method.
 */
public class ServletRoutingTable extends ServiceRoutingTable 
    implements HttpSessionContext
{
    private Class singleThreadModel;
    private Hashtable requests;

    // We'll make one static final copy of the name of SingleThreadModel Class
    private static final String single_thread_model_class_name =
	"javax.servlet.SingleThreadModel";

    // We'll keep a static reference to the SingleThreadModel Class
    public static Class single_thread_model_class = null;

    // We'll keep a reference to the one and only instance of
    // UnavailableServlet..
    public static Servlet unavailable_servlet = null;

    static
    {
	// We need to hold one copy of the SingleThreadModel class for
	// all ServletEnry instances to use..
	try
	{
	    single_thread_model_class =
		Class.forName(single_thread_model_class_name);
	}
	catch (Exception e)
	{
	    Loco.log(Log.CRITICAL_ERROR, "ERROR: ServletEntry - " +
		     "Exception caught while trying to load " +
		     "SingleThreadModel class:" +
		     StringUtil.exceptionStackTrace(e));
	}

	// Keep a single reference to the one and only UnavailableServlet
	// instance (it's multithreaded) so we can reuse it.
  	unavailable_servlet = new UnavailableServlet();
    }

    /**
     * Initializes the servletRoutingTable and loads the servlets.
     */
    public ServletRoutingTable(String path, Log server_log, Vector classpath,
			       boolean auto_reload) 
	throws InitializationException
    {
	super(path, server_log, classpath, auto_reload);

	requests = new Hashtable();

	// get our SingleThreadModel class for later use...
	try
	{
	    singleThreadModel = 
		Class.forName("javax.servlet.SingleThreadModel");
	}
	catch (ClassNotFoundException snfe) 
	{
	    throw new InitializationException(
		"could not load SingleThreadModel");
	}
    }	

    /**
     * This method overrides ServiceRoutingTable::initialize().  This
     * implementation first shuts down the ServletRoutingTable's services
     * (if any), and then it calls super.initialize().
     */
    public void initialize() throws InitializationException
    {
        // First, shut everything down properly..
        shutdown();

        // Then, start everything back up by re-reading servlet.conf and
        // reloading/re-instantiating classes.
        super.initialize();
    }

    /**
     * Adds the a ServletEntry to the service_tokens_hash or the
     * service_extensions_hash.
     */
    protected ServiceEntry createServiceEntry(String pattern, String classname,
					      String config_path)
    {
	return ( new ServletEntry(pattern, classname, config_path) );
    }


    /** Overrides the superclass to put in the default handlers.
     * here's what we put in: <PRE>
     *   default      SimpleServlet
     *   unknown      SimpleServlet
     *   srv          ServerServlet
     * </PRE>
     */
    protected void putInDefaults() throws InitializationException
    {
	ServiceEntry se;

	// We'll make one instance of UnavailableServlet (it is a
	// MultithreadedLocoServlet subclass, we only need one), and make
	// it available.
	addService("unavailable",
		   "org.locomotive.loco.servlet.UnavailableServlet", null);

	// Allow external file to override default servlet,
	// but install the default if one is missing.
	if (! containsService("default"))
	{
	    addService("default",
			"org.locomotive.loco.servlet.DefaultServlet", null);
	}

	// The servlet to be called if no other one can be found
	// may also be overridden in the config file
	if (! containsService("unknown"))
	{
	    addService("unknown", 
		       "org.locomotive.loco.servlet.UnknownServlet", null);
	}

	// always install the server handler.
	addService("srv", "org.locomotive.loco.ServerServlet", null);
    }


    /** 
     * This gets called on every request by the loco, once it
     * determines it has a servlet in its hands.  It sets up everything
     * the servlet needs, then sends it on its way.
     * 
     * @param request_id the unique id for this request
     * @param sock the socket this request connected on
     * @param is the inputstream from that socket
     * @param os the outputstream from that socket
     * @param cgi_envars the CGI variables passed in with the request
     * @param form_hash the form data, if there was any
     * @param url_tokens a tokenized form of the url
     * @param request_uri the unparsed url tokens
     * @param bstamp_id the browser stamp id that came in from the
     *        bstamp cookie for this request
     * @param response the response for this request
     * @param conn the Database connection for this request
     * @param session the Session object for this request
     * @param attrib attributes for this request
     */
    public void routeRequest(int request_id,
			     Socket sock, 
			     InputStream is, 
			     OutputStream os,
			     Hashtable cgi_envars,
			     Hashtable form_hash, 
			     Hashtable steam_vars,
			     String [] url_tokens,
			     String request_uri, 
			     long bstamp_id,
			     Response response,
			     Connection conn, 
			     Session session,
                             Hashtable attrib)
    {

        Hashtable attributes = null;
        
	// If the attributes Hashtable that we were given is null, then
	// we'll build a new one..
        if (attrib == null)
        {
            attributes = new Hashtable();

	    // Put in some default attributes for this request
            attributes.put("locomotive.bstamp_id", new Long(bstamp_id));
            attributes.put("locomotive.steam_vars", steam_vars);
            attributes.put("locomotive.url_tokens", url_tokens);
	    
            // this can be null- no DB Connection
            if (conn != null) attributes.put("locomotive.connection", conn);

            if (null != form_hash)
                attributes.put("locomotive.form_data",new FormData(form_hash));
        }
        else
        {
	    // Otherwise, we'll use the one we were given.
	    attributes = attrib;
        }
        
	// Create the Request Object
	LocoServletRequest request = new LocoServletRequest(
	    request_id, sock, cgi_envars, attributes, is,
	    (LocoServletResponse) response, form_hash, session);
	
	// Store an association between this Thread and its LocoServletRequest
	// object.  This comes in handy later when LocoServletContext needs
	// information about the request.
	requests.put(Thread.currentThread(), request);

	Log server_log = Loco.getServerLog();
	    
	// now, are we handling a servlet or what?
	// right now, we only support a servlet token to identify servlets.
	// if there's a servlet that maps to our token, we start the
	// servlet.  Otherwise, we start the default servlet.

  	Servlet s = null;

  	String first_token = "";
  	if (url_tokens != null && url_tokens.length > 0 && 
  	    url_tokens[0] != null)
  	{
  	    first_token = url_tokens [0];
  	}

	// Use the request_uri to route the request to the correct servlet.
	Object service = routeToService(Loco.getConfig(), request_uri,
					first_token);

	// Does the servlet implement SingleThreadModel?
	boolean single_thread = false;

	// Did we intentionally shut down and destroy the servlet?
	boolean servlet_destroyed = false;

	if (service == null)
	{
	    server_log.log(0, "ERROR: ServletRoutingTable.routeRequest: " +
			   "no servlet returned for URI: " + request_uri);
	    sendErrorMsg(HttpServletResponse.SC_NOT_FOUND,
			 "Unknown Server Function: " +
			 request_uri, request, (LocoServletResponse) response);
	    return;
	}
	else if (service instanceof ServletPool)
	{
	    s = (Servlet) ( (ServletPool) service).getObject();
	    single_thread = true;
	}	    
	else  // s is a servlet
	{
	    s = (Servlet) service;
	}
	
	try
	{
	    // ach - is there any way of avoiding this, short of
	    // subclassing our Response object to implement
	    // HttpServletResponse?
	    if (s instanceof HttpServlet)
	    {
		response.setContentType("text/html");
	    }

	    // ========================================================
	    // Call the service method, passing control to the servlet!
	    // ========================================================
	    s.service(request, (LocoServletResponse) response);
	}
	//
	// The Servlet 2.2 spec says the following:
	//
	// A servlet may throw either a ServletException or an
	// UnavailableException during the service of a request. A
	// ServletException signals that some error occurred during
	// the processing of the request and that the container
	// should take appropriate measures to clean up the request.
	// An UnavailableException signals that the servlet is
	// unable to handle requests either temporarily or
	// permanently.
	//
	// If a permanent unavailability is indicated by the
	// UnavailableException, the servlet container must remove
	// the servlet from service, call its destroy method, and
	// release the servlet instance.
	//
	// If temporary unavailability is indicated by the
	// UnavailableException, then the container may choose to not
	// route any requests through the servlet during the time
	// period of the temporary unavailability. Any requests
	// refused by the container during this period must be
	// returned with a SERVICE_UNAVAILABLE (503) response status
	// along with a Retry-After header indicating when the
	// unavailability will terminate. The container may choose to
	// ignore the distinction between a permanent and temporary
	// unavailability and treat all UnavailableExceptions as
	// permanent, thereby removing a servlet that throws any
	// UnavailableException from service.
	catch (UnavailableException ue)
	{
	    // This is getting too long, and should really call another
	    // method so to not clutter up this one, but that's better
	    // left for when we do a major architectural overhaul (soon!!)

	    // Servlets of this type are currently unavailable.
	    server_log.log(Log.WARNING, "routeRequest(): in service() " +
	        s.getClass().getName() +
	        " instance threw UnavailableException: " +
		(ue.isPermanent() ? "permanent" : "temporary(" +
		 ue.getUnavailableSeconds() + ")") + ", Reason: " +
	         ue.getMessage());

	    // Get a reference to the servlet's ServletEntry Object.
	    ServletEntry the_servlet_entry = null;
	    ServiceEntry the_service_entry =
		routeToServiceEntry(Loco.getConfig(), request_uri,first_token);
	    if (null != the_service_entry)
		the_servlet_entry = (ServletEntry) the_service_entry;

	    // If we didn't get ahold of the ServletEntry, then we'll log
	    // this problem, and just move on..
	    if (null == the_servlet_entry)
	    {
		server_log.log(0, "ERROR: routeRequest(): " + 
		    "Can't get ahold of ServletEntry for URI " + request_uri);

		// We can't really signal about the unavailability, nor
		// wait (if it's temporary), since we can't get the
		// ServletEntry..  Oh well.  As long as the problem's logged.
	    }
	    else
	    {
		// First we'll check to see if the UnavailableException
		// is permanent.
		if (ue.isPermanent())
		{
		    // It's permanent, so we'll destroy the servlet..
		    // Destroy and remove the instance (this also notifies
		    // the ServletFactory that it may need to make more).
		    the_servlet_entry.shutdown(s, server_log);
		    servlet_destroyed = true;

		    // Now, send the client an error about this..  Right
		    // here we're hoping that the servlet didn't already
		    // send anything to the client..  But, even if they
		    // did, there's nothing we could do about it here, and
		    // they're going to get the error message text anyway.
		    try
		    {
			((LocoServletResponse) response).sendError(
			    HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
			    "Caught UnavailableException during " +
			    "service().\n<BR>" +
			    "Please try your request again.\n<BR>");
		    }
		    catch (IOException ioe)
		    {
		    }
		}
		else
		{
		    // It's temporarily unavailable..  If it's not a
		    // multithreaded servlet, then we'll ditch it because
		    // we have more servlets in the pool (probably) and
		    // because so far we have no way of tracking
		    // unavailability per instance in the ServletPool.
		    if (single_thread)
		    {
			// It's in a ServletPool, so we'll destroy the
			// servlet..
			// Destroy and remove the instance (this also notifies
			// the ServletFactory that it may need to make more).
			the_servlet_entry.shutdown(s, server_log);
			servlet_destroyed = true;

			// Now, send the client an error about this..  Right
			// here we're hoping that the servlet didn't already
			// send anything to the client..  But, even if they
			// did, there's nothing we could do about it here, and
			// they're going to get the error message text anyway.
			try
			{
			    ((LocoServletResponse) response).sendError(
			        HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
				"Caught UnavailableException during " +
				"service().\n<BR>" +
				"Please try your request again.\n<BR>");
			}
			catch (IOException ioe)
			{
			}
		    }
		    else
		    {
			// It's multithreaded.  We'll mark it as being
			// temporarily unavailable.

			// We'll tell the ServletFactory that it's unavailable.
			ServletFactory sf =
			    the_servlet_entry.getServletFactory();

			// Use this UnavailableException to change state to
			// unavailability and notify the listeners of the
			// change..
			sf.triggerUnavailability(ue);

			// Now we need to show a SERVICE UNAVAILABLE error
			// page (from the UnavailableServlet), and the
			// best way to do that from right here is to call
			// routeToService() again now.
			Servlet unavail_servlet = (Servlet)
			    routeToService(Loco.getConfig(), request_uri,
					   first_token);

			// We should have gotten back the unavailable_servlet
			if (unavail_servlet == unavailable_servlet)
			{
			    try
			    {
				// Invoke the UnavailableServlet.
				unavail_servlet.service(request,
			            (LocoServletResponse) response);
			    }
			    catch (Exception e)
			    {
				server_log.log(0, "ERROR: routeRequest(): " + 
				    s.getClass().getName() + ": " + 
				    StringUtil.exceptionStackTrace(e));

				// Send an error page.
				try
				{
				    ((LocoServletResponse) response).sendError(
				        HttpServletResponse.
					SC_INTERNAL_SERVER_ERROR, "");
				}
				catch (IOException ioe)
				{
				}
			    }
			}
		    }
		}
	    }
	}
	catch (Throwable t)
	{
	    server_log.log(0, "ERROR: handoff(): " + 
			   s.getClass().getName() + ": " + 
			   StringUtil.exceptionStackTrace(t));
	    sendErrorMsg(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
			 "", request, (LocoServletResponse) response);
	}
	    
	// now, clean up a little
	// return the servlet to the pool if that's where it came from
	if (single_thread && !servlet_destroyed)
	{
	    ((ServletPool) service).returnToPool(s);
	}
	
	// if the servlet connection isn't the one we gave them, let's 
	// put that back in the pool too.  The requestManager handles the
	// normal conn.
	if (request != null)
	{
	    Connection temp = 
		(Connection) request.getAttribute("locomotive.connection");
	    if (temp != null && temp != conn)
	    {
		Loco.putBackDBConnection(temp);
	    }
	}
    }
    
    /**
     * Finds and returns the service Object associated with the specified
     * request uri.  This method first tries to match the first token to
     * a token mapping (like "base"), and then tries to match the request uri
     * to an extension mapping (like "*.jsp").
     *                                                               <BR<BR>
     * This method overrides the method routeToService(LocoConfig,
     * String, String) in ServiceRoutingTable, only this method also
     * passes through any unavailability information to present to
     * the user if the ServiceEntry's availability status shows either
     * temporary or permanent unavailability.
     *                                                               <BR><BR>
     * @param config a reference to the LocoConfig object
     * @param request_uri the URI of the request
     * @param first_token the first URL token of the request.  It is
     *        used to find the correct service to map the request to.
     * @see ServiceRoutingTable#routeToService
     */
    public Object routeToService(LocoConfig config, String request_uri,
				 String first_token)
    {
	// Get the ServiceEntry by calling routeToServiceEntry()..
	ServiceEntry service_entry = routeToServiceEntry(config, request_uri,
							 first_token);
	if (null != service_entry)
	{
	    // Get the service Object from the service_entry, it's likely
	    // to be a ServletPool, or a Servlet of some kind.
	    Object service = service_entry.getService();

	    // If it's currently unavailable, then we'll pass through the
	    // unavailability information.
	    int availability = service_entry.getAvailability();
	    if (availability != ServletFactory.AVAILABLE)
	    {
		// Set the status String, either "temporarily" or "permanently"
		String status = null;
		if (availability == ServletFactory.UNAVAILABLE_AND_WAITING)
		    status = UnavailableServlet.TEMPORARY_STRING;
		else status = UnavailableServlet.PERMANENT_STRING;

		// Get the unavailability info from the service_entry
		Integer seconds = new Integer(
		    service_entry.getUnavailableSeconds());
		String reason = service_entry.getUnavailabilityReason();
		Long timestamp = new Long(
		    service_entry.getLastAvailabilityEventTimestamp());

		//
		// Set each of these in the request as attributes
		//
		    
		// First, we have to get the request for this Thread
		LocoServletRequest request = getLocoServletRequest();

		// Set the request attributes
		request.setAttribute(UnavailableServlet.STATUS_ATTRIBUTE_NAME,
				     status);
		request.setAttribute(UnavailableServlet.SECONDS_ATTRIBUTE_NAME,
				     seconds);
		request.setAttribute(UnavailableServlet.REASON_ATTRIBUTE_NAME,
				     reason);
		request.setAttribute(
		    UnavailableServlet.TIMESTAMP_ATTRIBUTE_NAME, timestamp);

		// Since it's unavailable, we'll be sure to return the
		// UnavailableServlet..
		return unavailable_servlet;
	    }

	    return service;
	}

	// We couldn't find a service for this request.
	return null;
    }

    /**
     * Returns the session context object for this server.  Context should be
     * created in initialize()
     */
    public HttpSessionContext getSessionContext()
    {
	return ((HttpSessionContext) this );
    }

    /**
     * This method returns the LocoServletRequest for the current Thread
     */
    public LocoServletRequest getLocoServletRequest()
    {
	return (LocoServletRequest) requests.get(Thread.currentThread());
    }


    /**
     * Displays and error page with Standard insertions, plus
     * the error string for the class.  The error string is set in
     * the constructor.  Evaluates the template
     * LOCO_TEMPLATE_ROOT_PATH/error/generic_error.tmpl.
     * LOCO_TEMPLATE_ROOT_PATH is the template directory, as specified in
     * the Loco configuration file.
     * If this template is missing, then it will send a generic html file
     * including the message.
     * @param status_code one of the status codes defined in
     *                    HttpServletResponse.
     * @param error_msg The message to display.  If left blank or null,
     *        then the following will message will be displayed: <BR>
     *        <CODE>Your request could not be processed due to an
     *		     internal servlet error.</CODE>
     * @param req the request object for this request
     * @param resp the response object for this request
     */
    public static void sendErrorMsg(int status_code,
				    String error_msg, 
				    ServletRequest req, 
				    HttpServletResponse resp)
    {
	if (error_msg == null || error_msg.length() < 1)
	{
	    error_msg = "Your request could not be processed " +
		"due to an internal servlet error. ";
	}

	resp.setContentType("text/html");

	Hashtable subs = new java.util.Hashtable();
	subs = StandardInsertions.insertAll(req, subs);
	subs.put("ERROR_MSG", error_msg);

	String response = 
	    Loco.evaluateTemplate("error/generic_error.tmpl", subs);

	try
	{
	    // the template is missing
	    if (response == null || response.length() < 1) 
	    {
		Loco.log(Log.NOTICE,
			 "ERROR: could not find generic error template at " +
			 "error/generic_error.tmpl.");
		resp.sendError(resp.SC_INTERNAL_SERVER_ERROR, 
			       error_msg);
	    }
	    else
	    {
		resp.setStatus(resp.SC_INTERNAL_SERVER_ERROR);
		try
		{
		    PrintWriter pw = resp.getWriter();
		    pw.print(response);
		    pw.flush ();
		    pw.close();  // that way no one else can use it
		}
		catch (IllegalStateException ise)
		{
		    ServletOutputStream sos = resp.getOutputStream();
		    sos.print(response);
		    sos.flush();
		    sos.close();
		}
	    }
	}
	catch (IOException e) 
	{
	    Loco.log(Log.NOTICE,
		     "ERROR: could not write error message: " + response + 
		     "Error: " + e.toString());
	}
    }	



    //-------------- HttpSessionContext methods --------------------


    /** 
     * Currently returns all the ids of all the sessions
     * in the database.  
     * @return an Enumeration of Long objects, or null if we had 
     *         database problems
     * @deprecated for security reasons
     */
    public Enumeration getIds()
    {
	return ( new Vector().elements() );

/*
	String query = "select sid from loco_sessions order by sid asc";
	Statement stmt = null;
	ResultSet rs = null;
	Connection conn = Loco.getDBConnection();
	Vector ids = new Vector();
	
	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);
	    while (rs.next())
	    {
		ids.addElement(new Integer(rs.getInt(1)));
	    }
	    
	    return ( ids.elements() );
	}
	catch (SQLException se)
	{
	    Loco.log(Log.NOTICE, "ERROR: ServletSession.getIds: " + 
		     StringUtil.exceptionStackTrace(se));
	    
	    return ( null );
	}
	finally
	{
	    try 
	    {
		if (rs != null) rs.close();
		if (stmt != null) stmt.close();
	    }
	    catch (SQLException sse) {}
	    
	    Loco.putBackDBConnection(conn);
	}

	*/
    }
		    

    /** 
     * Presently goes to the database and gets the session with the
     * specified Session ID.  The session id has to be a number that exists,
     * otherwise this will null.
     * @deprecated for security reasons
     */
    public HttpSession getSession(String sessionId)
    {
	return ( (HttpSession) new Session() );

/*
	

	Session sess = null;
	Connection conn = Loco.getDBConnection();
	long sid = 0;
	
	try 
	{ 
	    sid = Long.parseLong(sessionId); 
	    sess = new Session(conn, sid);
	    return ( (HttpSession) sess );
	}
	catch (FailedSessionException se)
	{
	    Loco.log(Log.NOTICE, "ERROR: LocoSessionContext: getSession: " + 
		     StringUtil.exceptionStackTrace(se));
	    return ( null );
	}
	catch (NumberFormatException ne) 
	{ 
	    return (null); 
	}
	finally
	{
	    Loco.putBackDBConnection(conn);
	}

	*/

    }



    //----- Inner Class required for extending ServiceRoutingTable -----



    public class ServletEntry extends ServiceRoutingTable.ServiceEntry
	implements ServletFactoryListener
    {
	/**
	 * We keep a reference to our ServletFactory -- it makes servlets
	 * for us in another Thread.
	 * @see ServletFactory
	 */
	protected ServletFactory servlet_factory = null;

	ServletEntry (String pat, String cname, String c_path)
	{
	    super(pat, cname, c_path);

	    // Start off with ServletEntries being temporarily unavailable
	    // while they're created in another Thread (the ServletFactory).
	    setAvailability(UNAVAILABLE_AND_WAITING);
	}

	protected void initialize() throws InitializationException
	{
	    ServletConfig sc = 
		new LocoServletConfig(props, ServletRoutingTable.this);

	    try
	    {
		// Find out if the servlet class is an instance of
		// SingleThreadModel
		if (single_thread_model_class.isAssignableFrom(theclass))
		{
		    //
		    // If it's a SingleThreadModel, then we need to make a
		    // ServletPool of them..
		    //
		    ServletPool sp = new ServletPool(theclass, sc, this);
		    Loco.log(4, "Creating Pool for " + theclass);
		    sp.initialize();
		    service = sp;

		    cached = true;
		}
		else
		{
		    //
		    // If it's not a SingleThreadModel servlet, then we need
		    // to just instantiate a ServletFactory and it will make
		    // one new instance for us.
		    //
		    getServletFactory();

		    // Start up the ServletFactory if it isn't already running
		    if (!servlet_factory.isAlive()) servlet_factory.start();

		    // It'll get marked as cached when we get it from
		    // the ServletFactory..
		    cached = true;
		}
	    }
	    catch (Exception e)
	    {
		Loco.log(Log.WARNING, "ERROR: ServletEntry.initialize(): " +
			 StringUtil.exceptionStackTrace(e));
	    }
	}
	
	/**
	 * Shuts down the service for this ServletEntry.
	 */
	public boolean shutdown(Log server_log)
	{
	    if (null != service)
	    {
		// Depending on what kind of service Object we have,
		// we shut it down differently..
		if (service instanceof Servlet)
		{
		    // Destroy the Servlet
		    Servlet s = (Servlet) service;
		    s.destroy();

		    // Shut down our ServletFactory Thread
		    getServletFactory().shutdown();

		    // De-reference the shut-down factory.
		    servlet_factory = null;
		}
		else
	        {
		    // It must be a ServletPool
		    ServletPool sp = (ServletPool) service;
		    sp.closeCurrentPoolMembers();

		    // Shut down the ServletPool's ServletFactory Thread.
		    ServletFactory sf = sp.getServletFactory();
		    sf.shutdown();

		    // De-reference the ServletPool's ServletFactory.
		    sp.setServletFactory(null);
		}
	    }

	    // For now, always return true -- what a happy method!
	    return (true);
	}
	
	/**
	 * Shuts down just the specified servlet instance for this
	 * ServletEntry.  Call this method when you want to destroy
	 * and de-reference just an individual servlet instance.
	 */
	public boolean shutdown(Servlet servlet_instance, Log server_log)
	{
	    if (null != service && null != servlet_instance)
	    {
		// Depending on what kind of service Object we have,
		// we shut the servlet_instance down differently..
		if (service instanceof Servlet)
		{
		    // It's a multithreaded servlet
		    servlet_instance.destroy();

		    // De-reference the servlet instance
                    service = null;

		    // Notify our ServletFactory that it may need to make
		    // another Servlet to replace the destroyed one.
                    ServletFactory sf = getServletFactory();
                    synchronized (sf)
                    {
                        sf.notify();
                    }

		    // Log what happened.
		    server_log.log(Log.WARNING, "ServletEntry::shutdown(): " +
				   "instance of " + theclass.getName() +
				   " shut down.");
		}
		else
	        {
		    // It must be a ServletPool
		    ServletPool sp = (ServletPool) service;

		    // Wipe the instance from the pool, and notify the
		    // ServletFactory that it may need to make a replacement.
		    sp.wipeFromPool(servlet_instance);

		    // Log what happened.
		    server_log.log(Log.WARNING, "ServletEntry::shutdown(): " +
				   "instance of " + theclass.getName() +
				   " in ServletPool shut down.");
		}
	    }

	    // For now, always return true -- what a happy method!
	    return (true);
	}
	
	private void handleException(String err_msg, Throwable t) 
	    throws InitializationException
	{
	    props = null;
	    throw new InitializationException (err_msg + ": " + t.toString());
	}

	//-------------- ServletFactoryListener methods ---------------------

        /**
         * The ServletFactory calls this method when it needs to know how
         * many servlets that the primary listener (like a ServletPool or
	 * a ServletRoutingTable) needs to keep active and ready to serve.
	 *                                                           <BR><BR>
	 * This implementation in ServletEntry will just return a value
	 * of one, since ServletEntry just needs to hold just one servlet
	 * if the ServletEntry isn't using a ServletPool.
	 */
	public int getMinServletCount()
	{
            // FIXME:  Later, we may want to return 0 or 1 -- 0 when the
            //         admin wants none of a particular multithreaded
            //         servlet to start up ahead of time.  Administrators
            //         should have the option to make the Locomotive only
            //         instantiate servlets at request time.
	    return 1;
	}

	/**
	 * The ServletFactory calls this method when it needs to know the
	 * maximum number of servlets that the primary listener (like a
	 * ServletPool or a ServletRoutingTable) will hold..  If the listener
	 * has all the servlets it needs, then the ServletFactory doesn't need
	 * to make any more unless the primary listener asks for more.
	 *                                                           <BR><BR>
	 * This implementation in ServletEntry will just return a value
	 * of one, since ServletEntry just needs to hold a maximum of one
	 * multithreaded servlet (the ServletEntry isn't using a
	 * ServletPool).
	 */
	public int getMaxServletCount()
	{
	    return 1;
	}

	/**
	 * The ServletFactory calls this method when it needs to know how
	 * many servlets that the primary listener (like a ServletPool or
	 * a ServletRoutingTable) currently holds.
	 */
	public int getCurrentServletCount()
	{
	    // Basically if it's not the unavailable servlet and it's not
            // null, then we have one.
	    if (service != unavailable_servlet && null != service) return 1;
	    else return 0;
	}

	/**
	 * The ServletFactory calls this method when it has either
	 * successfully instantiated and initialized a servlet (and the
	 * new servlet is ready to serve requests), or when the
	 * ServletFactory needs to signal the ServletFactoryListener(s)
	 * that some other event has occurred.  In the case that a new
	 * servlet instance is ready, the new servlet is encapsulated
	 * inside the ServletFactoryEvent that is passed to this method.
	 * It is up to the ServletFactoryListener to make the servlet
	 * available to clients.
	 *                                                           <BR><BR>
	 * @param servlet_factory_event an event that tells this
	 *        ServletFactoryListener Object that a servlet is
	 *        ready to serve requests (in which case the servlet
	 *        is encapsulated in the event Object), or that the
	 *        servlet is unavailable.
	 * @see ServletFactoryEvent
	 */
	public synchronized void handleServletFactoryEvent(
	    ServletFactoryEvent servlet_factory_event)
	{
            // If we got an event from a ServletFactory other than our
            // current ServletFactory (maybe one that we just replaced)
            // then ignore events from it!
            if (servlet_factory_event.getSource() != servlet_factory) return;

	    // First, we'll store the event timestamp as the date/time
	    // of the last event that occurred.
	    setLastAvailabilityEventTimestamp(
	        servlet_factory_event.getEventTimestamp());

	    availability = servlet_factory_event.getAvailability();

	    // Find out the availability from the event
	    if (availability == ServletFactory.AVAILABLE)
	    {
		// Grab the servlet out of the event, and put it into service!
		Servlet new_servlet = servlet_factory_event.getServlet();

		// If the servlet isn't a null, we'll use it.
		if (null != new_servlet)
		{
		    service = new_servlet;

		    // We'll mark it cached too.
		    cached = true;
		}
	    }
	    else if (availability == ServletFactory.UNAVAILABLE_AND_WAITING)
	    {
		// We now know that the servlet is currently unavailable,
		// and that the ServletFactory is going to wait and then
		// retry to make it.

		// We'll keep a copy of how long the ServletFactory
		// is going to wait before retrying..
		unavailable_seconds =
		    servlet_factory_event.getUnavailableSeconds();

		// Get the reason why the servlet is temporarily unavailable.
		unavailability_reason =
		    servlet_factory_event.getUnavailabilityReason();
	    }
	    else if (availability == ServletFactory.PERMANENTLY_UNAVAILABLE)
	    {
		unavailable_seconds = 0;

		// Get the reason why the servlet is temporarily unavailable.
		unavailability_reason =
		    servlet_factory_event.getUnavailabilityReason();
	    }
	    else if (availability == ServletFactory.SHUTDOWN)
	    {
		// So far, we don't need to do anything with shutdown events.
		// That's because all we would do is de-reference the
		// ServletFactory and allow it to be garbage collected.
		// The shutdown method of this Object already does that.
		// But, I suppose we should log the shutdown.
		Loco.log(Log.INFO, "ServletEntry for " + theclass.getName() +
			 ": " +
			 servlet_factory_event.getUnavailabilityReason());
	    }
	}

	//----------------------- Utility methods -----------------------

        /**
	 * Returns a reference to this ServletPool's ServletFactory Object.
	 * If one doesn't already exist, a new one is instantiated, and a
	 * reference to it is returned.
	 *                                                          <BR><BR>
	 * @return a reference to this ServletPool's ServletFactory.
	 */
	public ServletFactory getServletFactory()
	{
	    if (null == servlet_factory)
	    {
		// Before creating a new ServletFactory, we need to make
		// a ServletConfig Object to instantiate the ServletFactory
		// with..
		ServletConfig sc = 
		    new LocoServletConfig(props, ServletRoutingTable.this);

		// Create our ServletFactory, passing it a reference to this
		// ServletEntry, the servlet class to instantiate, and a
		// ServletConfig (used for configuration information, and to
		// look up some request information).
		servlet_factory = new ServletFactory(this, theclass, sc);
	    }

	    return servlet_factory;
	}

	public void setServletFactory(ServletFactory new_servlet_factory)
	{
	    servlet_factory = new_servlet_factory;
	}

	/**
	 * Give out the Servlet, but only to classes within the same package.
	 */
	protected Servlet getServlet()
	{
	    try
	    {
		// Try to return the servlet, if that fails then we
		// didn't have one in the first place.
		return (Servlet) service;
	    }
	    catch (Exception e)
	    {
		// This is probably a ClassCastException, and we don't need
		// to do anything about it.
	    }

	    // We don't have a servlet to return.
	    return null;
	}
    }
}    
