/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 2000 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.servlet;

import org.locomotive.loco.*;
import org.locomotive.server.Log;

import javax.servlet.*;
import javax.servlet.http.*;

import java.io.*;
import java.util.Date;
import java.util.Hashtable;


/** 
 * This servlet is invoked when a servlet being requested is currently
 * unavailable (temporarily or permanently).  It returns the template
 * LOCO_TEMPLATE_ROOT_PATH/error/service_unavailable.tmpl, or, if
 * that template is missing, a similar warning message is sent.  This
 * servlet is a multithreaded servlet.
 *
 * @author Jason Brittain <brittainj@diamtech.com>
 */
public class UnavailableServlet extends MultiThreadedLocoServlet
{
    /** The request attribute name String for the unavailability status. */
    public static final String STATUS_ATTRIBUTE_NAME =
	"locomotive.servlet.unavailability_status";

    /** The request attribute name String for the unavailability seconds. */
    public static final String SECONDS_ATTRIBUTE_NAME =
	"locomotive.servlet.unavailability_seconds";

    /** The request attribute name String for the unavailability reason. */
    public static final String REASON_ATTRIBUTE_NAME =
	"locomotive.servlet.unavailability_reason";

    /** The request attribute name String for the event timestamp. */
    public static final String TIMESTAMP_ATTRIBUTE_NAME =
	"locomotive.servlet.unavailability_timestamp";

    public static final String TEMPORARY_STRING = "temporarily";
    public static final String PERMANENT_STRING = "permanently";

    /**
     * This method is run for each request to this servlet.  It uses the
     * LOCO_TEMPLATE_ROOT_PATH/error/service_unavailable.tmpl template.
     * LOCO_TEMPLATE_ROOT_PATH is the template directory, as specified in
     * the Locomotive's configuration.
     */
    public void doGet(HttpServletRequest request, HttpServletResponse response)
	throws ServletException, IOException
    {
	//
	// This unavailability status String should say either "permanently"
	// or "temporarily".
	//
	String status = (String) request.getAttribute(STATUS_ATTRIBUTE_NAME);

	// If it's empty, we'll say "temporarily", it's probably just not
	// started up yet.
	if (null == status) status = TEMPORARY_STRING;

	int minutes = 0;

	if (status.equals(TEMPORARY_STRING))
	{
	    //
	    // Get the number of seconds for which we'll show unavailability.
	    //
	    Long last_event_timestamp = (Long) request.getAttribute(
	        TIMESTAMP_ATTRIBUTE_NAME);
	    Integer seconds_integer = (Integer) request.getAttribute(
	        SECONDS_ATTRIBUTE_NAME);

	    // If timestamp isn't defined, then we'll use the current date.
	    long time_now = new Date().getTime();
	    if (null == last_event_timestamp)
		last_event_timestamp = new Long(time_now);
	    long time_then = last_event_timestamp.longValue();

	    // If seconds is not defined, we'll use the default_waiting_period.
	    if (null == seconds_integer) seconds_integer =
	        new Integer(ServletFactory.default_waiting_period);
	    int seconds = seconds_integer.intValue();

	    // We're assuming the unavailability hasn't yet expired.. so the
	    // remaining wait time is the expiration time in millis minus the
	    // current time in millis.
	    long remaining_unavailability_milliseconds =
		(time_then + seconds * 1000) - time_now;

	    // If the remaining wait time is less than zero, then there is no
	    // more wait time, the unavailability should be timed out
	    // (and the ServletFactory should be retrying).
	    if (remaining_unavailability_milliseconds < 0L)
		remaining_unavailability_milliseconds = 0L;

	    // Set our Retry-After header to the number of unavailability
	    // seconds left.
	    response.setHeader("Retry-After", "" +
	        (int) (remaining_unavailability_milliseconds / 1000L));

	    // Round to the nearest minute
	    minutes = (int)
		(((float) (remaining_unavailability_milliseconds / 1000L) /
		  60f) + 0.5f);

	    // Don't say "try again in 0 minutes"..
	    if (minutes < 1) minutes = 1;
	}

	//
	// Get the reason why this service is unavailable.
	//
	String reason = (String) request.getAttribute(REASON_ATTRIBUTE_NAME);

	// If there's no reason, it's probably because no instances have
	// yet been made and initialized.
	if (null == reason) reason = "The service is currently being started.";

	// Put the values into STEAM
	Hashtable steam_vars = getSteamVars(request);
	steam_vars.put("SERVLET_UNAVAILABILITY_STATUS", status);
	steam_vars.put("SERVLET_UNAVAILABILITY_MINUTES", "" + minutes);
	steam_vars.put("SERVLET_UNAVAILABILITY_REASON", reason);

	// Set the HTTP status code of the response to be
	// 503 - Service Unavailable
	response.setStatus(HttpServletResponse.SC_SERVICE_UNAVAILABLE);

	// Generate the response content from the template and the STEAM vars
  	String response_content = Loco.evaluateTemplate(
  	    "error/service_unavailable.tmpl", steam_vars);

	// is the template is missing ?
	if (response_content == null || response_content.length() < 1)
	{
	    response_content = 
		"<HTML>\n<HEAD><TITLE>Locomotive - 503 Service Unavailable " +
		"</TITLE></HEAD>\n" +
		"<BODY>\n\n<H1>503 - Service Unavailable</H1>\n" +
		"The service specified is " + status + " unavailable.<BR>\n";
            if (!status.equals(PERMANENT_STRING))
		response_content = response_content +
		    "Please try again after waiting " + minutes +
		    " minute(s).<BR>\n<BR>\n";
	    response_content = response_content + "Reason: " + reason +
		"<BR>\n</BODY>\n</HTML>";
	}

	PrintWriter print_writer = response.getWriter();
	print_writer.print(response_content);
	print_writer.flush();
    }
    
    /**
     * This method just calls doGet(request, response).
     */
    public void doPost(HttpServletRequest request,
		       HttpServletResponse response)
	throws ServletException, IOException
    {
	doGet(request, response);
    }
}
