/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.store.session;

import org.locomotive.loco.handler.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.util.*;
import java.io.*;
import java.util.*;
import java.sql.*;
import java.math.BigInteger;
import org.locomotive.loco.User;
import org.locomotive.loco.FailedSessionException;
import org.locomotive.loco.store.id.IdException;

public abstract class SessionStore {

  /**
   * associate a user with this session, persistantly.  Authentication
   * should be handled externally
   * @param conn    a JDBC connection with access to the Loco tables
   * @param user    the user to be associated
   * @param sid     the session id
   * @exception FailedSessionException if there is a database storage problem
   */
  public abstract void associateUser(Connection conn, User user, long sid) throws FailedSessionException;


  /**
   * expires this session, in the database
   * returns false if the operation failed
   * @param conn    a JDBC connection with access to the Loco session tables
   * @param sid     the session id
   * @param expireTime The timestamp
   * @exception FailedSessionException if the session cannot be marked in
   *            the database
   */
  public abstract void expire (Connection conn, long sid, Timestamp expireTime) throws FailedSessionException;

  /**
   * Stores the session_objects table in the database.
   * Should be called to store the session's resources after the session
   * has been used for a request, if caching is not being used.
   * @param conn    a JDBC connection with access to the Loco session tables
   * @param sid     the session id
   * @param objects the hashtable of objects to be saved
   * @throw SQLException if there was a database error
   */

  //dwq
  // for consistancy should this not throw FailedSessionException not SQLException
  public abstract void storeObjects(Connection conn, long sid, Hashtable objects) throws SQLException;


  /**
   * Removes the objects for this session, either from the cache, if
   * it's being used, or from the database
   * @param conn    a JDBC connection with access to the Loco session tables
   * @param sid     the session id
   * @exception FailedSessionException if the session cannot be marked in
   *            the database
   */
  public abstract void removeObjects(Connection conn, long sid) throws FailedSessionException;

  /**
   * Advances the expiration time of the session by exp_seconds
   * if the session is persistant, the new time will be committed to the
   * session database
   * @param conn          a JDBC connection with access to
   *	                    the Loco session tables
   * @param sid     the session id
   * @param expireTime The timestamp
   * @exception FailedSessionException if there is a database error
   */
  public abstract void resetExpiration(Connection conn, long sid, Timestamp expireTime) throws FailedSessionException;

  /**
   * Returns an individual session row if it can be found.
   * Returns an object because there are multiple return values
   * The LocoSessionRow is a prime candidate for a pool to reduce object
   * creation/destruction
   * @param conn          a JDBC connection with access to
   *	                    the Loco session tables
   * @param sid     the session id
   * @exception FailedSessionException if there is a database error or the session cannot be found
   */
  public abstract LocoSessionRow getLocoSessionRow(Connection conn, long sid) throws FailedSessionException;

  /**
   * Returns the objects for the individual session
   * @param conn          a JDBC connection with access to
   *	                    the Loco session tables
   * @param sid     the session id
   * @exception FailedSessionException if there is a database error or the session cannot be found
   */
  public abstract Hashtable getSessionObjects(Connection conn, long sid) throws FailedSessionException;

  /**
   * Saves a new Session and returns its sid
   * @param conn          a JDBC connection with access to
   *	                    the Loco session tables
   * @param srid    the random number for cookie checks
   * @param userid  the user
   * @param expire  the timestamp
   * @exception FailedSessionException if there is a database error or the session cannot be found
   */
  public abstract long createNewSession (Connection conn, int srid, int userid, Timestamp expire) throws FailedSessionException;
	
  /**
   * Saves a new Session and returns its sid
   * @param conn          a JDBC connection with access to
   *	                    the Loco session tables
   * @param sid     the session id
   * @param objects the hashtable of objects to be saved
   * @exception FailedSessionException if there is a database error or the session cannot be found
   */
  public abstract void storeSessionObjects(Connection conn, Hashtable objects);
	

  /**
   * Returns the next available id to use when saving a new session
   * @param conn   a JDBC connection with access to the Loco id table
   * @exception IdException if the id could not be generated
   */
  protected abstract BigInteger getNextSessionId(Connection conn) throws IdException;



}


