/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.store.session;

import java.io.*;
import java.util.*;
import java.sql.*;
import java.math.BigInteger;
import org.locomotive.loco.User;
import org.locomotive.loco.FailedSessionException;
import org.locomotive.loco.Loco;
import org.locomotive.loco.store.id.LocoId;
import org.locomotive.loco.store.id.IdException;



public class SessionStoreJdbc extends SessionStore {


  protected BigInteger getNextSessionId(Connection conn) throws IdException {
    return LocoId.getInstance().getNextId(LocoId.LOCO_SESSIONS, conn);
  }


  public void associateUser(Connection conn, User user, long sid) throws FailedSessionException {
    String statement = null;
    PreparedStatement pstmt = null;


    try {
      statement = "update loco_sessions set userid = ? " +
                  ", " + "active = ? where sid =  ?";

      pstmt = conn.prepareStatement(statement);
      pstmt.setLong(1, user.getUserid());
      pstmt.setTimestamp(2, new Timestamp(System.currentTimeMillis()));
      pstmt.setLong(3, sid);
      pstmt.executeUpdate ();
    } catch (SQLException se) {
      Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      throw new FailedSessionException(se.toString());
    } finally {
      try {
        if (pstmt != null) pstmt.close();
      } catch (SQLException se) {
	  Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      }
    }
  }

  public void expire (Connection conn, long sid, Timestamp expireTime) throws FailedSessionException
  {
    String statement = null;

    statement = "update loco_sessions set expire = ?, " +
                "active = ? where sid = " + sid;

    PreparedStatement pstmt = null;

    try {
	pstmt = conn.prepareStatement (statement);
	pstmt.setTimestamp(1, expireTime);
	pstmt.setTimestamp(2, new Timestamp(System.currentTimeMillis()));
	pstmt.executeUpdate ();
	pstmt.close();
    } catch (SQLException se) {
	Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
	throw new FailedSessionException (se.toString());
    } 
    finally {
	try {
	    if (pstmt != null) pstmt.close();
	} catch (SQLException se) {
	  Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      }
    }
  }

  public void storeObjects(Connection conn, long sid, Hashtable objects) throws SQLException
  {

    PreparedStatement pstmt = null;
    String query = null;

    try {
      conn.setAutoCommit(false);
      pstmt = conn.prepareStatement("delete from loco_session_objects where sid = " + sid);
      pstmt.executeUpdate();
      pstmt.close();
      query = "insert into loco_session_objects (sid, objects) " +
              "values (?, ?)";

      pstmt = conn.prepareStatement(query);
      pstmt.setLong(1, sid);

      // is 1 k enough?
      ByteArrayOutputStream bos = new ByteArrayOutputStream(1024);
      ObjectOutputStream oos = new ObjectOutputStream(bos);
      oos.writeObject(objects);
      oos.flush();
      ByteArrayInputStream bis = new ByteArrayInputStream(bos.toByteArray());
      pstmt.setBinaryStream(2, bis, bis.available());

      pstmt.executeUpdate();

      if (! conn.getAutoCommit()) {
        conn.commit();
        conn.setAutoCommit(true);
      }
    } catch (SQLException se) {
      if (! conn.getAutoCommit()) {
        conn.rollback();
        conn.setAutoCommit(true);
      }
      Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      throw se;
    } catch (IOException ioe) {
	conn.setAutoCommit(true);
	ioe.printStackTrace();
	throw new SQLException ("couldn't write session objects: " +
				ioe.toString() + " stacktrace " ); 
  } finally {
      conn.setAutoCommit(true);
      if (pstmt != null) pstmt.close();
    }
  }

  public void removeObjects(Connection conn, long sid) throws FailedSessionException
  {

    String query = "delete from loco_session_objects where sid = " + sid;
    Statement stmt = null;
    try {
	stmt = conn.createStatement();
	stmt.executeUpdate(query);
        } 
    catch (SQLException se) {
      throw new FailedSessionException (se.toString());
    } finally {
      try {
        if (stmt != null) stmt.close();
      } catch (SQLException se) {
	  Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      }
    }
  }

  public void resetExpiration (Connection conn, long sid, Timestamp expireTime) throws FailedSessionException
  {
    // write it to the db too
    String statement = null;

    statement = "update loco_sessions set expire = ?, " +
                "active = ? where sid = " + sid;

    PreparedStatement pstmt = null;

    try {
	pstmt = conn.prepareStatement (statement);
	pstmt.setTimestamp(1, expireTime);
	pstmt.setTimestamp(2, new Timestamp(System.currentTimeMillis()));
	pstmt.executeUpdate ();
	pstmt.close();
    } catch (SQLException se) {
      Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      throw new FailedSessionException (se + "- sessionId: " + sid);
    }
    finally
    {
	try
	{
	    if (pstmt != null) pstmt.close();
	} catch (SQLException se) {
	    Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
	}
    }
  }

  public LocoSessionRow getLocoSessionRow(Connection conn, long sid) throws FailedSessionException
  {
    Statement stmt = null;
    ResultSet rs = null;
    LocoSessionRow r = null;


    try {
      String statement = "select srid, userid, expire, added, active " +
                         "from loco_sessions where sid = " + sid;

      stmt = conn.createStatement();
      rs = stmt.executeQuery (statement);

      if (! rs.next()) {
        rs.close();
        stmt.close();
        throw new FailedSessionException( "LocoSessionRow:session not found in db: " +
                                          sid);
      }

      r = new LocoSessionRow( rs.getInt(1), rs.getInt(2), rs.getTimestamp(3), rs.getTimestamp(4), rs.getTimestamp(5) );
    } catch (SQLException se) {
      throw new FailedSessionException("Session " + sid + ": " + se.toString());
    } finally {
      try {
        if (rs != null) rs.close();
        if (stmt != null) stmt.close();
      } catch (SQLException se) {
	  Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      }
    }
    return r;
  }

  public Hashtable getSessionObjects(Connection conn, long sid) throws FailedSessionException
  {
      Statement stmt = null;
      ResultSet rs = null;
      Hashtable objects = null;

        try {
	  stmt = conn.createStatement();
	  
	  rs = stmt.executeQuery("select objects from loco_session_objects " + 
				 "where sid = " + sid);

	  if (rs.next()) {
	      ObjectInputStream obis = new 
		  ObjectInputStream(new BufferedInputStream(rs.getBinaryStream(1)));
	      objects = (Hashtable) obis.readObject();
	      obis.close();
	  }
	  
      } catch (SQLException se) {
	  se.printStackTrace();
	  throw new FailedSessionException("getSessionObjects: Session " + sid + ": " +
					   se.toString() ); 
      } catch (IOException ioe) {
	  ioe.printStackTrace();
	  throw new FailedSessionException("getSessionObjects: Session " + sid + ": " +
					   ioe.toString());
      } catch (ClassNotFoundException cnfe) {
	  cnfe.printStackTrace();
	  throw new FailedSessionException("getSessionObjects: Session " + sid + ": " +
					   cnfe.toString());
      } finally {
	  try {
	      if (rs != null) rs.close();
	      if (stmt != null) stmt.close();
	  } catch (SQLException se) {
	      Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
	  }
    }

    return objects;
  }

  //dw
  // not going to provide this by default until we have the sid created by loco
  //  public long createNewSession (Connection conn, int srid, int userid, Timestamp expire) throws FailedSessionException

  public void storeSessionObjects(Connection conn, Hashtable objects)
  {
    int rows = 0;
    PreparedStatement pstmt1 = null;
    PreparedStatement pstmt2 = null;
    String query1 = null;
    String query2 = null;

    query1 = "delete from loco_session_objects where sid = ?";
    query2 = "insert into loco_session_objects (sid, objects) " +
             " values (?, ?)";

    try {
      pstmt1 = conn.prepareStatement(query1);
      conn.setAutoCommit(false);

      pstmt2 = conn.prepareStatement(query2);

      // for some reason this, combined with the reset() calls, isn't
      // working correctly.  It seems to lose about 2-3 bytes
      // or so every time reset is called.
//    ByteArrayOutputStream bos = new ByteArrayOutputStream(8192);
//    ObjectOutputStream oos = new ObjectOutputStream(bos);

      Enumeration e = objects.keys();
      while (e.hasMoreElements()) {
        Long sessid = (Long) e.nextElement();

        try {
          // delete the row from the table
          pstmt1.setLong(1, sessid.longValue());
          pstmt1.executeUpdate();

          // add the new row
          pstmt2.setLong(1, sessid.longValue());

          ByteArrayOutputStream bos =
          new ByteArrayOutputStream(1024);
          ObjectOutputStream oos = new ObjectOutputStream(bos);
          oos.writeObject(objects.get(sessid));
          oos.flush();
          ByteArrayInputStream bis =
          new ByteArrayInputStream(bos.toByteArray());
          pstmt2.setBinaryStream(2, bis, bis.available());
//			bos.reset();
//			oos.reset();

          rows += pstmt2.executeUpdate();

          if (! conn.getAutoCommit()) {
            conn.commit();
          }
        } catch (SQLException se) {
          Loco.log(0, "ERROR: session shutdown: could not save " +
                   "session objects for " + sessid + ": " + se);
	  Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();

        } catch (IOException ioe) {
          Loco.log(0, "ERROR: session shutdown: could not save " +
                   "session objects for " + sessid + ": " + ioe);
        }
      }

    } catch (SQLException se) {
      Loco.log(0, "ERROR: session shutdown: could not save session " +
               "objects: " + se + "stacktrace " ); se.printStackTrace();
    } finally {
      try {
        if (pstmt1 != null) pstmt1.close();
        if (pstmt2 != null) pstmt2.close();
        if (! conn.getAutoCommit()) {
          conn.setAutoCommit(true);
        }
      } catch (SQLException se) {
	  Loco.log(0, "SQL Exception " + se + "stacktrace " ); 
	  se.printStackTrace();
      }
  }
  }




  public long createNewSession (Connection conn, int srid, int userid, Timestamp expire) throws FailedSessionException {

      
    // get new sid, create new row in LOCO_SESSIONS


    long sid = 0;

    PreparedStatement pstmt = null;

    try {

      sid = getNextSessionId(conn).longValue();

      // write everything in, along with current date/time
      String statement = "insert into LOCO_SESSIONS " +
        "(sid, srid, userid, added, active, expire) values ( " +
        sid + ", " + srid + ", " + userid +
        ", ?, ?, ?)";
      pstmt = conn.prepareStatement (statement);
      Timestamp now = new Timestamp(System.currentTimeMillis());
      pstmt.setTimestamp (1, now); // set in constructor
      pstmt.setTimestamp (2, now);
      pstmt.setTimestamp (3, expire);
      pstmt.executeUpdate();
      pstmt.close();

    } catch (SQLException se) {
      throw new FailedSessionException("CreateNewSession: " + se.toString());
    } catch (IdException ie) {
      throw new FailedSessionException("CreateNewSession: " + ie.toString());
    } finally {
      try {
        if (pstmt!= null) pstmt.close();
      } catch (SQLException se) {
	  Loco.log(0, "SQL Exception " + se + "stacktrace " ); se.printStackTrace();
      }
    }
    return sid;
  }




}








