/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.store.user;

import java.io.*;
import java.util.*;
import java.sql.*;
import java.math.BigInteger;
import javax.servlet.http.*;
import org.locomotive.util.*;
import org.locomotive.loco.LoadUserException;
import org.locomotive.loco.User;
import org.locomotive.loco.store.id.IdException;

public abstract class UserStore {
  /**
   * Persistantly stores User's data.  Will fail
   * if uid is not unique (uid <= 0 will cause system to pick a userid).
   * Will also fail if un (intended username) is already used.
   *
   * @param conn   a JDBC connection with access to the Loco user database
   * @param uid    the userid, system will pick if <= 0
   * @param un     username of the new user
   * @param pword  password of the new user
   * @param locoflg flags of the new user for Loco-specific functions
   * @exception LoadUserException if there is a database storage error
   */
  public abstract int createUser(Connection conn, int uid, String un, String pword,
              int locoflg) throws LoadUserException;


  /**
   * returns true if a user with this userid exists in the user database
   * @param conn   a JDBC connection with access to the Loco user database
   * @param id     the userid of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public abstract boolean exists (Connection conn, int id)
  throws LoadUserException;

  /**
   * returns true if a user in the user database has the same name
   * @param conn   a JDBC connection with access to the Loco user database
   * @param un     the username of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public abstract boolean exists (Connection conn, String un)
  throws LoadUserException;

  /**
   * returns true if a user in the user database has the same name,
   * regardless of case.  e.g. user "bob" exists, even though the
   * userbase only contains a user "Bob"
   * @param conn   a JDBC connection with access to the Loco user database
   * @param un     the username of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public abstract boolean existsAnyCase (Connection conn, String un)
  throws LoadUserException;

  /**
   * Updates the user.last_login field in the database.
   * silently ignored on anonymous users
   * @param conn an active JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public abstract void saveLoginTime (Connection conn, int userid) throws LoadUserException;

  /**
   * Updates the User data in the database if changes have been made,
   * silently ignored on anonymous users
   * @param conn an active JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public abstract void save (Connection conn, int userid, String username, String password, int loco_flags) throws LoadUserException;

  /**
   * used by constructor above - loads User based on un from db
   */
  public abstract LocoUserRow load (Connection conn, String un, boolean sensitiveToCase)
  throws LoadUserException;

  /**
   * used by constructor above - loads User based on uid from db
   */
  public abstract LocoUserRow load (Connection conn, int uid) throws LoadUserException;

  /**
   * Searches the user database and returns a list of Users, based upon
   * userid.
   * @param conn  a JDBC connection
   * @param userids  a list of userids to retrieve
   * @exception LoadUserException if there is a database storage error
   */
  public abstract User[] loadUsers (Connection conn, int userids[])
  throws LoadUserException;

  /**
   * Returns an array of all the user names in the database.
   * @param conn  a JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public abstract String[] loadAllUserNames (Connection conn)
  throws LoadUserException;

  /**
   * Given a vector of Integer-ized userids, look up each username for
   * them.
   */
  public abstract Vector getUsernames(Connection conn, Vector userids)
  throws SQLException;

  /**
   * Given a userid, return the username
   */
  public abstract String getUsername(Connection conn, int userid)
  throws SQLException;


  /**
   * Returns the next available id to use when saving a new user
   * @param conn   a JDBC connection with access to the Loco id table
   * @exception IdException if the id could not be generated
   */
  protected abstract BigInteger getNextUserId(Connection conn) throws IdException;

}
