/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.loco.store.user;

import java.io.*;
import java.util.*;
import java.sql.*;
import java.math.BigInteger;
import javax.servlet.http.*;
import org.locomotive.util.*;
import org.locomotive.loco.LoadUserException;
import org.locomotive.loco.User;
import org.locomotive.loco.store.id.LocoId;
import org.locomotive.loco.store.id.IdException;


public class UserStoreJdbc extends UserStore {


  /**
   * returns the next free user id
   * @param conn   a JDBC connection with access to the Loco user database
   * @exception IdException if the id could not be generated
   */
  protected BigInteger getNextUserId(Connection conn) throws IdException {
    return LocoId.getInstance().getNextId(LocoId.LOCO_USERS, conn);
  }



  /**
   * Persistantly stores User's data.  Will fail
   * if uid is not unique (uid <= 0 will cause system to pick a userid).
   * Will also fail if un (intended username) is already used.
   *
   * @param conn   a JDBC connection with access to the Loco user database
   * @param uid    the userid, system will pick if <= 0
   * @param un     username of the new user
   * @param pword  password of the new user
   * @param locoflg flags of the new user for Loco-specific functions
   * @exception LoadUserException if there is a database storage error
   */
  public int createUser(Connection conn, int uid, String un, String pword,
                        int locoflg) throws LoadUserException
  {
    int userid = uid;
    String username = un;
    String password = pword;
    int loco_flags = locoflg;

    // order in the table is as follows:
    // userid, username, password, loco_flags, added, modified, last_login
    // HEY!  LETS GET THOSE FINALLY CLAUSES IN THERE!
    
    PreparedStatement pstmt = null;

    try {
      if (userid <= 0)
        userid = getNextUserId(conn).intValue();

      // enter the user into the table
      String statement = "insert into LOCO_USERS (userid, username, " +
        "password, loco_flags, added, " +
        "modified) values (?, ?, ?, ?, ?, ?)";
      pstmt = conn.prepareStatement(statement);
      Timestamp t = new Timestamp(System.currentTimeMillis());
      pstmt.setInt (1, userid);
      pstmt.setString (2, username);
      pstmt.setString (3, password);
      pstmt.setInt(4, loco_flags);
      pstmt.setTimestamp(5, t);
      pstmt.setTimestamp(6, t);
      pstmt.executeUpdate ();
      pstmt.close();
      return userid;

    } catch (SQLException se) {
      throw new LoadUserException ("storage failure: " + se.toString());
    } catch (IdException ie) {
      throw new LoadUserException ("storage failure: " + ie.toString());
    } finally {
      try {
        if (pstmt!= null) pstmt.close();
      } catch (SQLException se) {

      }
    }
  }






  /**
   * returns true if a user with this userid exists in the user database
   * @param conn   a JDBC connection with access to the Loco user database
   * @param id     the userid of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public boolean exists (Connection conn, int id)
  throws LoadUserException
  {
    try {
      boolean exists = true;
      Statement stmt = conn.createStatement();
      ResultSet rs =
      stmt.executeQuery("select userid from loco_users" +
                        " where userid = " + id);
      if (rs.next()) {
        exists = true;
      } else {
        exists = false;
      }

      rs.close();
      stmt.close();
      return (exists);
    } catch (SQLException se) {
      throw new LoadUserException (se.toString());
    }
  }

  /**
   * returns true if a user in the user database has the same name
   * @param conn   a JDBC connection with access to the Loco user database
   * @param un     the username of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public boolean exists (Connection conn, String un)
  throws LoadUserException
  {
    try {
      boolean exists = true;
      Statement stmt = conn.createStatement();
      ResultSet rs =
      stmt.executeQuery ("select userid from loco_users" +
                         " where username = " +
                         org.locomotive.util.SQL.quoteString (un));
      if (rs.next()) {
        exists = true;
      } else {
        exists = false;
      }

      rs.close();
      stmt.close();
      return (exists);
    } catch (SQLException se) {
      throw new LoadUserException (se.toString());
    }
  }

  /**
   * returns true if a user in the user database has the same name,
   * regardless of case.  e.g. user "bob" exists, even though the
   * userbase only contains a user "Bob"
   * @param conn   a JDBC connection with access to the Loco user database
   * @param un     the username of the user to look for
   * @exception LoadUserException if there is a database access error
   */
  public boolean existsAnyCase (Connection conn, String un)
  throws LoadUserException
  {
    String query = null;

    query ="select userid from loco_users" +
           " where Lower(username) = " +
           org.locomotive.util.SQL.quoteString (un.toLowerCase());

    try {
      boolean exists = true;
      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery (query);

      if (rs.next()) {
        exists = true;
      } else {
        exists = false;
      }

      rs.close();
      stmt.close();
      return (exists);
    } catch (SQLException se) {
      throw new LoadUserException (se.toString());
    }
  }





  /**
   * Updates the user.last_login field in the database.
   * silently ignored on anonymous users
   * @param conn an active JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public void saveLoginTime (Connection conn, int userid) throws LoadUserException {
    if (userid <= 0)
      return;

    PreparedStatement pstmt = null;
    String statement = "update LOCO_USERS set last_login = " +
      "? where userid = " + userid;

    int result = 0;
    try {
      pstmt = conn.prepareStatement(statement);
      pstmt.setTimestamp(1, new Timestamp(System.currentTimeMillis()));
      result = pstmt.executeUpdate();
      pstmt.close();
    } catch (SQLException se) {
      throw new LoadUserException (se.toString());
    } finally {
      try {
        if (pstmt!= null) pstmt.close();
      } catch (SQLException se) {

      }
    }

    if (result == 0) {
      throw new LoadUserException ("User " + userid + " does not exist.");
    }
  }


  /**
   * Updates the User data in the database if changes have been made,
   * silently ignored on anonymous users
   * @param conn an active JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public void save (Connection conn, int userid, String username, String password, int loco_flags) throws LoadUserException {

    PreparedStatement pstmt = null;
    String statement = "update LOCO_USERS set " +
      "username = ?, password = ?, loco_flags = " + loco_flags +
      ", modified = ? where userid = " + userid;

    try {
      pstmt = conn.prepareStatement (statement);
      pstmt.setString(1, username);
      pstmt.setString(2, password);
      pstmt.setTimestamp(3, new Timestamp(System.currentTimeMillis()));
      pstmt.executeUpdate();
      pstmt.close();
    } catch (SQLException se) {
      throw new LoadUserException (se.toString());
    } finally {
      try {
        if (pstmt!= null) pstmt.close();
      } catch (SQLException se) {

      }
    }
  }







  /**
   * used by constructor above - loads User based on un from db
   */
  public LocoUserRow load (Connection conn, String un, boolean sensitiveToCase)
  throws LoadUserException
  {
    try {
      String statement;
      if (sensitiveToCase) {
        statement = "select userid, password, loco_flags, username " +
                    "from loco_users where username = " +
                    org.locomotive.util.SQL.quoteString (un);
      } else {
        statement = "select userid, password, loco_flags, username " +
                    "from loco_users where Lower(username) = " +
                    org.locomotive.util.SQL.quoteString (un.toLowerCase());
      }

      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery (statement);

      if (! rs.next()) {   // no user exists
        rs.close();
        stmt.close();
        return (null);
      }


      int userid = rs.getInt (1);
      String password = rs.getString (2);
      int loco_flags = rs.getInt (3);
      String username;
      username = rs.getString(4);

      LocoUserRow ur = new LocoUserRow(userid, username, password, loco_flags);

      rs.close();
      stmt.close();

      return ur;
    } catch (SQLException se) {
      throw new LoadUserException(se.toString());
    }
  }

  /**
   * used by constructor above - loads User based on uid from db
   */
  public LocoUserRow load (Connection conn, int uid) throws LoadUserException
  {

    try {
      String statement = "select username, password, loco_flags " +
                         " from loco_users where userid = " + uid;

      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery (statement);

      if (! rs.next()) {   // no user exists
        rs.close();
        stmt.close();
        return (null);
      }

      String username = rs.getString(1);
      String password = rs.getString (2);
      int loco_flags = rs.getInt (3);

      LocoUserRow ur = new LocoUserRow(uid, username, password, loco_flags);

      rs.close();
      stmt.close();
      return ur;
    } catch (SQLException se) {
      throw new LoadUserException(se.toString());
    }
  }

  /**
   * Searches the user database and returns a list of Users, based upon
   * userid.
   * @param conn  a JDBC connection
   * @param userids  a list of userids to retrieve
   * @exception LoadUserException if there is a database storage error
   */
  public User[] loadUsers (Connection conn, int userids[])
  throws LoadUserException
  {
    try {
      String statement =
      "select userid, username, password, loco_flags " +
      " from loco_users where " +
      org.locomotive.util.SQL.buildINClause ("userid", userids);

      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery (statement);

      Vector v = new Vector();

      User u;
      while (rs.next()) {
        // uses private constructor (above)
        u = new User (rs.getInt (1),
                      rs.getString (2),
                      rs.getString (3),
                      rs.getInt (4));
        v.addElement (u);
      }

      stmt.close();
      rs.close();

      User ua[] = new User[v.size()];
      if (ua.length > 0) {
        v.copyInto (ua);
      }
      return (ua);
    } catch (SQLException se) {
      throw new LoadUserException (se.toString());
    }
  }

  /**
   * Returns an array of all the user names in the database.
   * @param conn  a JDBC connection
   * @exception LoadUserException if there is a database storage error
   */
  public String[] loadAllUserNames (Connection conn)
  throws LoadUserException
  {
    try {
      String statement =
      "select username from loco_users order by username";

      Statement stmt = conn.createStatement();
      ResultSet rs = stmt.executeQuery (statement);

      Vector v = new Vector();

      while (rs.next()) {
        v.addElement (rs.getString(1));
      }

      stmt.close();
      rs.close();

      String un[] = new String[v.size()];
      if (un.length > 0) {
        v.copyInto (un);
      }
      return (un);
    } catch (SQLException se) {
      throw new LoadUserException (se.toString());
    }
  }

  /**
   * Given a vector of Integer-ized userids, look up each username for
   * them.
   */
  public Vector getUsernames(Connection conn, Vector userids)
  throws SQLException
  {
    Vector names = new Vector();
    Enumeration e = userids.elements();
    String cur_name;
    Object o;
    int current_userid;

    try {
      String statement = "select username from loco_users where " +
                         "userid = ?";
      PreparedStatement pstmt = conn.prepareStatement(statement);
      ResultSet results = null;

      while (e.hasMoreElements()) {
        o = e.nextElement();

        if (o == null) {
          pstmt.close();
          return (null);
        }

        current_userid = ((Number) o).intValue();

        if (current_userid == User.ANONYMOUS_ID) {
          names.addElement(User.ANONYMOUS_NAME);
        } else if (current_userid < User.MIN_ID) {
          pstmt.close();
          return (null);
        } else {
          pstmt.setInt(1, ((Number) o).intValue());
          results = pstmt.executeQuery();

          if (results.next()) {
            cur_name = results.getString(1);
            if (cur_name == null) cur_name = "";

            names.addElement(cur_name);
          } else {
            results.close();
            pstmt.close();
            return (null);
          }

          results.close();
        }
      }

      pstmt.close();
    } catch (SQLException se) {
      throw se;
    }

    return (names);
  }

  /**
   * Given a userid, return the username
   */
  public String getUsername(Connection conn, int userid)
  throws SQLException
  {
    String username = null;

    try {
      String statement = "select username from loco_users where " +
                         "userid = " + userid;

      Statement stmt = conn.createStatement();
      ResultSet results = stmt.executeQuery(statement);

      if (results.next())
        username = results.getString(1);
      else
        username = null;

      results.close();
      stmt.close();
    } catch (SQLException se) {
      throw se;
    }

    return (username);
  }

}
