/*
 * LoginServlet.java
 *
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.base;

import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;

import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;


/**
 * A Sample login handling Servlet
 */
public class LoginServlet extends LocoServlet
{
    /**
     * A GET simply shows the login page.
     */
    public void doGet(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	// if they are already logged in, then tell them
	if (!user.isAnonymous())
	{
	    logEvent ("login", "already", "");
	    displayPage ("error/already_logged_in.tmpl", steam_vars);
	    return;
	}

	logEvent ("login", "form", "");
	displayPage ("base/login/login.tmpl", steam_vars);
    }


    /**
     * Takes the username and password entered from the HTML form and
     * tries to log the user in.
     */
    public void doPost(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	// if they are already logged in, then tell them
	if (!user.isAnonymous())
	{
	    logEvent ("login", "already", "");
	    displayPage ("error/already_logged_in.tmpl", steam_vars);
	    return;
	}

        // if the form contains no data, then we'll pretend that we had an
        // HTTP 'GET' request, and simply show the form.
        if (form_data.isEmpty())
        {
	    doGet(req, resp);
            return;
        }

	// only if their session is persistent do we proceed.
	// if it is not, then attempting to associate a new user with
	// the session will be ignored.
	//
	// This currently means that their browser must support cookies
	// and that they need to have hit the Locomotive at least once
	// before getting this document.  That's fine, because they
	// almost certainly had to visit the Login page before they
	// submitted the Login HTML form.
	if (session.isNew())
	{
	    logEvent ("login", "failed", "no_cookie_support");
	    displayPage ("error/no_cookie_support.tmpl",
			 steam_vars);
	    return;
	}
	
	// everything's cool, now let's process the HTML form.
	
        handleFormSubmission();
    }

    
    /**
     * handle the login form submission.  At the end, we should have a
     * logged-in user, or else we'll redisplay the login form with
     * error messages.
     */
    protected void handleFormSubmission()
	throws ServletException, IOException
    {
	// if the username does not exist, or the password does not match,
	// then here's the message to give to the user:
	String generic_error =	"Your username and password are invalid, " +
	    "please re-enter. <BR>";

	// pull the values from the form so we can work with them.

	String form_username =
	    form_data.getAlphanum ("LOGIN_USERNAME", "username", 2, 12, true);
	String form_password =
	    form_data.getString ("LOGIN_PASSWORD", "password", 6, 12);

	String form_errors = form_data.allErrorMsgs();
	if (form_errors.equals(""))
	{
	    // if there were no errors, then log this user into the system.
	    try
	    {
		// load this user from the db.
		user = new User (conn, form_username, false);
		if (! user.hasPassword (form_password))
		{
		    form_errors = generic_error;
		}
	    }
	    catch (LoadUserException lue)
	    {
		handleException (lue, "There was a Locomotive error, code: " +
				 "login_lue.");
		return;
	    }
	    catch (UserNotFoundException unfe)
	    {
		// if the user was not found, that's okay; it's not a
		// serious error.  Display a normal error message.
		form_errors = generic_error;
	    }
	}

	// if there are errors, set the error string and the value of
	// the LOGIN_USERNAME variable so that they can be inserted
	// into the HTML template containing the login form.
	if (! form_errors.equals (""))
	{
 	    log(8, "login errors: " + form_errors);
	    
	    steam_vars.put ("LOGIN_ERRORS", form_errors);
	    steam_vars.put ("LOGIN_USERNAME", form_username);

	    logEvent ("login", "failed", form_errors);
	    displayPage ("base/login/login.tmpl", steam_vars);
	    return;
	}

	// if we get here, we know that the user is valid and loaded.
	// Associate the user with the current session persistently
	// (by modifying the loco_session table in the database).
	// this is done automatically when you replace the user in the
	// session object.
	session.putValue("locomotive.user", user);
	
	// run the default insertions again.  This was done in
	// BaseServlet earlier, but we've authenticated a user since then,
	// so we might need to reset some user-specific variables.
	StandardInsertions.insertAll (req, steam_vars);

	// log the event
	logEvent ("login", "success", "");
	steam_vars.put ("LOGIN_MESSAGE", "Login Successful, welcome!");

	// Now, let's send them back to the BaseServlet and let it
	// display its default page.  This lets the BaseServlet
	// perform any of the personalization logic it normally
	// performs before it displays the page.

//	RequestDispatcher reqd = getServletContext().getRequestDispaAtcher(
//	    Loco.getConfig().getString("SERVLET_URI_TOKEN_PREFIX") + "/base");
//	reqd.forward(req, resp);

//	 alternatively, you could use a server-side redirect to send them
//	 to the base page, thus:
	
	 String servlet_path = req.getServletPath();
 	 String path_to_servlet_engine = 
	     servlet_path.substring(0, servlet_path.indexOf(url_tokens[0]));
	 java.net.URL url = new java.net.URL( 
	     req.getScheme(), req.getServerName(), req.getServerPort(), 
	     path_to_servlet_engine + "base");
	 resp.sendRedirect(url.toString());
    
    }

}
