/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.base;

import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;
import org.locomotive.util.StringUtil;

/**
 * A sample Registration form handler.  Allows new users to sign up
 * on the User Management System managed by the Locomotive.  To use this
 * as a separate module add it to your hrt.conf file.  An example would
 * be: <PRE>
 *      reg           org.locomotive.module.base.RegistrationHandler
 * </PRE>
 */
public class RegistrationHandler extends GenericHandler
{
    /**
     * called by BaseHandler.handleRequest () if routed here by url_tokens
     */
    public void handleRequest (HandlerData hd_in)
    {
	super.handleRequest(hd_in);
	
	hd = hd_in;

	// if they are already logged in 
	if (! user.isAnonymous())
	{
	    sendAlreadyLoggedInMsg ();
	    return;
	}
	
	// only if their session is persistent do we proceed.  
	// if it is not, then attempting to associate a new user with 
	// the session will be ignored.
	// 
	// This currently means they must support cookies and that they
	// need to have hit the Locomotive at least once before getting this 
	// document.
	if (session.isNew ())
	{
	    sendBrowserNotSupportedMsg ();
	    return;
	}
	
	// if form is null, then we had an HTTP 'GET' request, so
	// we'll just display the Registration Form again.  Otherwise,
	// it was a 'POST', and we'll handle the Registration form
	// submission.
	
	if (form_data.isEmpty())
	{
	    displayPage ("base/reg/reg_form.tmpl", steam_vars);
	    logEvent ("reg", "form", "");
	}
	else 
	{
	    handleFormSubmit ();
	}
    }

    private void handleFormSubmit ()
    {
	String form_reg_username = null;
	String form_reg_password = null;
	String form_reg_password_confirm = null;
	String form_reg_firstname = null;
	String form_reg_lastname = null;
	
	String required_fields[] = { "REG_USERNAME", "REG_PASSWORD",
				     "REG_PASSWORD_CONFIRM",
				     "REG_FIRSTNAME", "REG_LASTNAME" };
	
	if (! form_data.containsAll (required_fields))
	{
	    steam_vars.put ("FORM_CONTENTS_TRANSMIT_ERROR",
			    "One or more fields in your Form were empty.  " +
			    "Please try re-submitting.");
	    displayPage ("base/reg/reg_form.tmpl", steam_vars);
	    logEvent ("reg", "form", "null_fields");
	    return;
	}

	// here we are actually writing the output text for the
	// template and embedding it in code.  another way to do it
	// would be to have error codes, and let the template contain
	// the actual message text (this is recommended because it's
	// more flexible, but more work in the code and template)

	// i.e.  set variable ERROR_USERNAME_TOO_SHORT in the environment
	// via code then insert something like this in the template:
	// blah blah html
	// 
	// [if [ERROR_USERNAME_TOO_SHORT] [quote 
	// <B>Your username is too short.  Please make it longer.</B>] ]
	// 
	// blah blah html

	// this variable stores miscellaneous errors that are not
	// automatically caught by the form_data form checker.
	String other_errors = "";

	// first, copy all the fields back into the steam variables,
	// so that the fields are filled out for the user the next time
	// the form is displayed.
	form_data.copyAllIntoSteam (steam_vars);
	
	// however, get rid of the password fields, because they have to
	// type it in again anyways if there's a mistake.
	steam_vars.remove ("REG_PASSWORD");
	steam_vars.remove ("REG_PASSWORD_CONFIRM");
	
	form_reg_username = form_data.getAlphanum ("REG_USERNAME", "username",
						   2, 12, true);

	form_reg_password = form_data.getString ("REG_PASSWORD", "password",
						 6, 12);

	form_reg_password_confirm = form_data.getString (
	    "REG_PASSWORD_CONFIRM", "password confirmation", 6, 12);

	if (! form_reg_password.equals (form_reg_password_confirm))
	{
	    other_errors += "Your two password fields do not match.  " + 
		"Please type them in again.<BR>\n";
	}
	
	form_reg_firstname =
	    form_data.getString ("REG_FIRSTNAME", "first name", 1, 31);

	form_reg_lastname =
	    form_data.getString ("REG_LASTNAME", "last name", 1, 31);

	String form_errors = form_data.allErrorMsgs();
	
	try
	{
	    if (User.existsAnyCase (conn, form_reg_username))
	    {
		other_errors += "The username you chose is already " +
		    "taken. Please choose another one.<BR>";
	    }
	}
	catch (LoadUserException lue)
	{
	    handleException(lue, "basereg_lue_exists");
	    return;
	}

	if (other_errors.length() > 0 || form_errors.length() > 0)
	{
	    steam_vars.put ("REG_CODE_GENERATED_ERRORS",
			    form_errors + other_errors);
	    displayPage ("base/reg/reg_form.tmpl", steam_vars);
	    logEvent ("reg", "submit_errors", "");
	    return;
	}
	
	//------------------ they've passed all checks. ------------------
	
	// We can now safely add them as a new user intto the database.
	
	User newuser;

	try
	{
	    newuser = new User (conn, 0, form_reg_username, 
				form_reg_password, User.LOCO_FLAGS_DEFAULT);
	}
	catch (LoadUserException lue)
	{
	    // some kind of nasty db error.  
	    // in rare cases (on a browser double-submit or just plain rare)
	    // an exception could be thrown here where there has been 
	    // a race condition between User.existsAnyCase (above) and 
	    // database level unique constraint username and lowerusername
	    
	    handleException(lue, "reg_new_user_lue");
	    return;
	}

	// mark the session as authenticated for the newly created user
	// (otherwise they have to go log in!)
	try
	{
	    session.associateUser (conn, newuser);
	}
	catch (FailedSessionException fse)
	{
	    handleException(fse, "reg_enable_session_fse");
	    return;
	}
	
	user = newuser;

	// remember: it's not longer true that hd.user != user.
	// so, put the new user into hd.user.
	hd.user = newuser;
	
	// get the user-associated info in subs
	StandardInsertions.insertAll(hd);
	
	logEvent ("reg", "success", "");
	displayPage ("base/reg/reg_success.tmpl", steam_vars);
    }

    /**
     * send back the already_logged_in template
     */
    private void sendAlreadyLoggedInMsg ()
    {
	logEvent("reg", "already", "");

	steam_vars.put ("TRIED_TO_REGISTER", "true");
	displayPage ("error/already_logged_in.tmpl", steam_vars);
    }

    /**
     * like the name implies...
     */
    private void sendBrowserNotSupportedMsg ()
    {
	logEvent ("reg", "failed", "no_cookie_support");

	steam_vars.put ("TRIED_TO_REGISTER", "true");
	displayPage ("error/no_cookie_support.tmpl", steam_vars);
    }
}
