/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.btrack;


import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;
import org.locomotive.server.Log;
import org.locomotive.util.*;

import java.util.Hashtable;
import java.util.Date;
import java.util.Vector;
import java.sql.*;

/**
 * BugTracker is a bug tracking system that allows the creation and editing
 * of projects and bugs associated with them. You can list or edit the
 * projects or the bugs. Listing the bugs allows filtering and per-package
 * display.
 *
 * <pre>
 *
 * All comands here are preceded by "btrack?":
 *
 * URL format (# = integer)    Command
 * ----------------------------------------------------------------------------
 *
 * (nothing)                   i.e. "btrack" - go to main page
 * listbugs                    go to bug-listing filter page
 * listbugs+#                  # = project id for which to list bugs
 * listbugs+all                list all bugs in the system
 * listprojects                list all projects in the system
 * edit+project+#              # = project id to edit
 * edit+project+#              this is also the url to use to submit the
 *                               editing form from the project edit page
 * new+project                 create a new project
 * new+bug                     create a new bug
 * edit+bug+#                  # = bug id to edit
 * edit+bug+#                  this is also the url to use to submit the
 *                               editing form from the bug edit page
 * gen+S                       S = name of generic (&quot;static&quot;) page
 *                                 to show
 * </pre>
 */
public class BugTrackerHandler extends GenericHandler
{

  /**
   * This message we will give to the user whenever there's a problem
   * not worth trying to explain to them (or not something they should
   * know about.) It is here only for coding convenience.
   */
  private static String generic_error_message =
  "Server error! Please contact the administrator.";

  /**
   * These log levels are set here for easy changing. You can set up your
   * own conventions. Here, 2 indicates some relatively harmless (but real)
   * error - like a broken link that freaks out the bug tracker. 0 represents
   * a grave error - like an SQL statement doing things we don't expect,
   * or a database interaction throwing an SQLException.
   */
  private static int GENERIC_LOG_LEVEL = 2;
  private static int URGENT_LOG_LEVEL = 0;

  /**
   * Some constants that remind us what integer values we have selected
   * to be "special":
   * had to make these non-private so that BugTrackerStore... can see them
   */
  static int NULL_BUG_ID = -1;
  static int NULL_PROJECT_ID = -1;
  static int NULL_PRIORITY_VALUE = -1;
  static int NULL_VERSION_VALUE = -1;
  static int NULL_SEVERITY_VALUE = -1;
  static int NULL_STATUS_VALUE = -1;

  /**
   * The database has restricitions on the size of strings. Here we
   * set up a convenient list of those maximums so we can tell the user
   * what they are but only change them in one place if the database
   * implementation changes around.
   */
  private static int MAX_SHORT_TITLE_LENGTH = 64;
  private static int MAX_LONG_TITLE_LENGTH = 256;
  private static int MAX_DESCRIPTION_LENGTH = 2000;
  private static int MAX_COMMENT_LENGTH = 2000;
  private static int MAX_STATUS_LENGTH = 128;


  /**
   * An estimated max number of projects in the db - for displaying
   * a menu for bug assignment on a template.
   */
  private static int EST_MAX_PROJECTS = 300;

  /**
   * Create a set of bug status selections and a set of
   * bug priorities, so that when a bug is created/edited, its status
   * and priority value will be checked against these for validity.
   * the values are set in the init(Log l) method
   */

  static Hashtable BUG_STATUS_CHOICES = new Hashtable(5);
  static Hashtable BUG_SEVERITY_CHOICES = new Hashtable(6);
  static Hashtable BUG_PRIORITY_CHOICES = new Hashtable(5);

  /**
   * This region of code is executed when the static class is initialized,
   * in case you weren't aware.
   */
  static
  {
    // init status choices
    BUG_STATUS_CHOICES.put("new","new");
    BUG_STATUS_CHOICES.put("open","open");
    BUG_STATUS_CHOICES.put("closed","closed");
    BUG_STATUS_CHOICES.put("reopened","reopened");
    BUG_STATUS_CHOICES.put("resolved","resolved");

    // init severity choices
    BUG_SEVERITY_CHOICES.put("critical","critical");
    BUG_SEVERITY_CHOICES.put("major","major");
    BUG_SEVERITY_CHOICES.put("normal","normal");
    BUG_SEVERITY_CHOICES.put("minor","minor");
    BUG_SEVERITY_CHOICES.put("trivial","trivial");
    BUG_SEVERITY_CHOICES.put("enhancement","enhancement");

    // init priority choices
    BUG_PRIORITY_CHOICES.put("1", "1");
    BUG_PRIORITY_CHOICES.put("2", "2");
    BUG_PRIORITY_CHOICES.put("3", "3");
    BUG_PRIORITY_CHOICES.put("4", "4");
    BUG_PRIORITY_CHOICES.put("5", "5");
  }

  /**
   * The version string for the entire BugTracking system - this
   * gets inserted into the client's environment every time a hit comes
   * in to this handler.
   */
  private static String version_string = "1.1";


  /**
   * This entry point is called by RequestManager for each new request.
   * This basically figures out what the user wants and passes off to
   * handleEdit(), handleListBugs(), or handleListProjects(),
   * where the real work happens.
   */
  public void handleRequest(HandlerData hd)
  {
    super.handleRequest(hd);


    // Call the Loco library call for standard insertions to the client
    // environment. Changing of this data is out of our control in the
    // bug tracking area, so we don't worry about synchronization.
    StandardInsertions.insertAll(hd);

    // If security concers ever made us want to pretend that the bug
    // tracker area was never reached, this variable might want to be
    // removed from the hash or not inserted here. For example, some
    // hypothetical secret management area, secrt?xxx, isn't supposed to
    // be known about, so if people without permissions go there it
    // pretends that it doesn't know what they're talking about (as if
    // they had entered garbage for the url). We don't have such concerns
    // here, really, but it's good to be aware of what we're putting into
    // the user's environment and to remember that the user can see all
    // of it if you leave it in when you evaluate a template in some method
    // call down below.
    steam_vars.put("BTRACK_VERSION", version_string);

    // Parse url tokens for proper function.
    // This can get hairy, but we need to decide what cases we want
    // to support and make sure they work.
    if (url_tokens.length < 2)
    {
      // i.e. "btrack" or "btrack?" came in. - length is probably 1.
      // There are no URL tokens, so GenericHandler will simply
      // return the 'main.tmpl' file found in the top level of the
      // specified generic directory
      displayPage ("btrack/main.tmpl", steam_vars);
    } else if (url_tokens[1].equals("gen"))
    {
      // i.e. "btrack?gen" - we pass off to the generic request.
      displayPage ("btrack/main.tmpl", steam_vars);
    } else if (url_tokens[1].equals("edit"))
    {
      // for all "btrack?edit..." style commands we let the next
      // method decide what to do with it, but if
      // all they've given us is "btrack?edit" we give an error here.
      if (url_tokens.length < 3)
      {
        displayPage ("btrack/error/unknown_handler_option.tmpl", steam_vars);
      } else
      {
        handleEdit();
      }
    } else if (url_tokens[1].equals("create"))
    {
      // for all "btrack?create..." style commands we let the next
      // method decide what to do with it, but if
      // all they've given us is "btrack?create" we give an error here.
      if (url_tokens.length < 2)
      {
        displayPage("btrack/error/unknown_handler_option.tmpl", steam_vars);
      } else
      {
        handleCreate();
      }
    } else if (url_tokens[1].equals("listbugs"))
    {
      // "btrack?listbugs...", like the edit above, gets passed off
      // to BugTracker.java to figure out.
      handleListBugs();
    } else if (url_tokens[1].equals("listprojects"))
    {
      // "btrack?listprojects", however, is pretty straightforwad.
      handleListProjects();
    } else
    {
      // Otherwise, they've entered something we don't understand.
      displayPage ("btrack/error/unknown_handler_option.tmpl", steam_vars);
    }
  }

  /**
   * Called to handle any URLs of the form
   * &quot;btrack?edit+... &quot; - will parse it out, figure out what
   * the user wants, return appropriate errors, etc.
   */
  protected void handleEdit()
  {
    // make sure they have enough tokens:
    if (url_tokens.length < 4)
    {
      log (GENERIC_LOG_LEVEL, "ERROR: BugTracker: handleEdit: got " +
           "bad url toks, size less than 4.");
      displayPage ("btrack/error/unknown_handler_option.tmpl", steam_vars);
    } else if (url_tokens[2].equals("bug"))
      handleEditBug();
    else if (url_tokens[2].equals("project"))
      handleEditProject();
    else
    {
      // then they've given us garbage:

      log (GENERIC_LOG_LEVEL, "ERROR: BugTracker: handleEdit: got " +
           "unrecognized url token: " + url_tokens[2]);

      displayPage ("btrack/error/unknown_handler_option.tmpl", steam_vars);
    }
  }

  /**
   * Called to handle any URLs of the form
   * &quot;btrack?create+... &quot; - will parse it out, figure out what
   * the user wants, return appropriate errors, etc.
   */
  protected void handleCreate()
  {
    // make sure they have enough tokens:
    if (url_tokens.length < 3 || url_tokens.length > 3)
    {
      log (GENERIC_LOG_LEVEL, "ERROR: BugTracker: handleCreate: got " +
           "bad url toks, size less than 3 or greater than 3.");
      displayPage ("btrack/error/unknown_handler_option.tmpl", steam_vars);
    } else if (url_tokens[2].equals("bug"))
      handleCreateBug();
    else if (url_tokens[2].equals("project"))
      handleCreateProject();
    else
    {
      // then they've given us garbage:

      log (GENERIC_LOG_LEVEL, "ERROR: BugTracker: handleCreate: got " +
           "unrecognized url token: " + url_tokens[2]);

      displayPage ("btrack/error/unknown_handler_option.tmpl", steam_vars);
    }
  }
  /**
   * Called internally by handleCreate to handle any URLs of the form
   * &quot;btrack?create+project...&quot; will give appropriate
   * errors or call other functions as needed. That is also the url
   * to use for submitting created Projects, so we will check form_hash to
   * see if that's the case.
   */
  private void handleCreateProject()
  {
    // check if this is a submission or a page display
    // if the form_hash is null, it means
    // they just want to bring up the project creation form. If form_hash is
    // not null they are submitting a form to save the new project. Call
    // the appropriate function:
    if (form_data.isEmpty())
      handleCreateProjectFormGen();
    else
      handleCreateProjectSubmission();
  }

  private void handleCreateProjectFormGen()
  {
    // display the project creation page
    displayPage ("btrack/create_project.tmpl", steam_vars);

    // and log the event.
    logEvent("btrack", "create_project", "new-proj-formgen");
  }



  /**
   *
   */
  private void handleCreateProjectSubmission()
  {
    int project_id = 0;
    StringBuffer errors = new StringBuffer("");

    logEvent("DEBUG", "about to do creation of proj", "go!");
    // check input from the user
    String short_title = form_data.getField("BTRACK_PROJECT_SHORT_TITLE");
    String long_title =  form_data.getField("BTRACK_PROJECT_LONG_TITLE");
    String text = form_data.getField("BTRACK_PROJECT_TEXT");

    if (form_data.hasAnyErrors())
    {
      errors.append(form_data.allErrorMsgs());
    }


    BugTrackerStore store = null;
    try {
      store = BugTrackerStore.getInstance();
    } catch (BugTrackerException be) {
      handleException(be, "can't instantiate store");
      return;
    }

    try {
      if (store.projectNameExists(short_title, conn))
        errors.append("That short_title name is already taken - try again");
    } catch (BugTrackerException be) {
      handleException(be, "projectNameExists check");
    }

    if (!errors.toString().equals(""))
    {
      errors.append(form_data.allErrorMsgs());
      steam_vars.put("BTRACK_CREATE_PROJECT_ERRORS", errors);
      form_data.copyAllIntoSteam(steam_vars);

      logEvent("DEBUG", "proj creation: constraint errors", errors.toString());

      displayPage("btrack/create_project.tmpl", steam_vars);
      return;
    }

    try {
      // Use the handy function call to get the next project id
      // from the database:
      project_id = store.getNextProjectId(conn);
    } catch (BugTrackerException be) {
      handleException(be, "Can't get next project id");
      return;
    }


    // Now uses the store to insert the project into the database - ee
    try {
      store.insertProject(project_id, short_title, long_title,
                          text, user.getUserid(), conn);
    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
    }

    // save success steam vars to display to user
    steam_vars.put("BTRACK_SUCCESSFUL_PROJECT_ADD", "1");
    steam_vars.put("BTRACK_NEW_PROJECT_NAME", short_title);
    displayPage("btrack/mod_complete.tmpl", steam_vars);

  }


  /**
   * Called internally by handleCreate to handle any URLs of the form
   * &quot;btrack?create+bug...&quot; will give appropriate
   * errors or call other functions as needed. That is also the url
   * to use for submitting created bugs, so we will check form_hash to
   * see if that's the case.
   */
  private void handleCreateBug()
  {
    // check if this is a submission or a page display
    // if the form_hash is null, it means
    // they just want to bring up the bug creation form. If form_hash is
    // not null they are submitting a form to save the new bug. Call
    // the appropriate function:
    if (hd.form_hash == null)
      handleCreateBugFormGen();
    else
      handleCreateBugSubmission();
  }

  /**
   * display page to user for new bug submission.  requires accessing the
   * db to get the project list.
   */
  private void handleCreateBugFormGen()
  {
    // get list of projects
    try
    {
      String[] projects = new String[EST_MAX_PROJECTS];
      int num_of_projects = BugTrackerStore.getInstance().
        getProjectNamesFromDB(projects, conn);
      // make list of project names available to template
      // through STEAM's nifty array display
      steam_vars.put("BTRACK_PROJECT_COUNT", String.valueOf(num_of_projects));
      steam_vars.put("BTRACK_PROJECT_NAMES", projects);
    } catch (BugTrackerException be)
    {
      // serious error.
      handleException(be, "Retrieving project names");
      return;
    }

    // display the bug creation page
    displayPage ("btrack/create_bug.tmpl", steam_vars);

    // and log the event.
    logEvent("btrack", "create_bug", "new-bug-formgen");
  }



  /**
   * creates a new bug in the db, first checking that form values are valid,
   * then attempting insertion.  if a non-sql error occurs, the page is
   * redisplayed with the saved form values and an error message.
   */
  private void handleCreateBugSubmission()
  {
    // holds errors if form_data retrieval has any
    StringBuffer errors = new StringBuffer("");
    // holds generated, and looked up ids necessary for db insert
    int bug_id = 0;
    int project_id = 0;

    // get all the form values
    String severity = form_data.getField("BTRACK_BUG_SEVERITY");
    String priority = form_data.getField("BTRACK_BUG_PRIORITY");
    String version = form_data.getField("BTRACK_LOCO_VERSION");
    String project_name = form_data.getField("BTRACK_PROJECT_NAME");
    String summary = form_data.getField("BTRACK_BUG_SUMMARY");
    String comments = form_data.getField("BTRACK_BUG_COMMENTS");

    // confirms the above values conform to string constraints
    if (!form_data.hasAnyErrors())
    {
      // no errors so far - keep plugging along...
      // check that the values for status and priority are valid

      if (!BUG_PRIORITY_CHOICES.contains(priority))
        errors.append("\nPriority value is not a valid choice");
      if (!BUG_SEVERITY_CHOICES.contains(severity))
        errors.append("\nSeverity value is not a valid choice");

    } else
    { // save the errors from form_data
      errors.append(form_data.allErrorMsgs());
    }

    // Now uses the store to get ids and insert the bug and initial comment - ee
    BugTrackerStore store = null;
    try {
      store = BugTrackerStore.getInstance();
    } catch (BugTrackerException be) {
      handleException(be, "can't instantiate store");
      return;
    }

    // if there are errors at this point, return the page with the error
    // information.
    if (!errors.toString().equals(""))
    {
      steam_vars.put("BTRACK_CREATE_ERRORS", errors);

      // save form values for redisplay	
      // project list must be regenerated

      String[] projects = new String[EST_MAX_PROJECTS];
      try{
        int num_of_projects = store.getProjectNamesFromDB(projects, conn);
        steam_vars.put("BTRACK_PROJECT_COUNT", String.valueOf(num_of_projects));
      } catch (BugTrackerException be) {
        handleException(be, "getProjectNamesFromDB for redisplay");
      }
      steam_vars.put("BTRACK_PROJECT_NAMES", projects);
      form_data.copyAllIntoSteam(steam_vars);

      displayPage("btrack/create_bug.tmpl", steam_vars);
      return;
    }

    // get a unique id for this new bug
    try{
      bug_id = store.getNextBugId(conn);
    } catch (BugTrackerException be) {
      handleException(be, "Retrieving next bug id");
      return;
    }

    // get the project id given the project_name
    try {
      project_id = store.getProjectIdGivenShortName(project_name,conn);
    } catch (BugTrackerException be) {
      handleException(be, "Retrieving project_id given short_name");
      return;
    }

    // get a unique id for this new bug's initial comment
    int bug_text_id;
    try{
      bug_text_id = store.getNextBugTextId(conn);
    } catch (BugTrackerException be) {
      handleException(be, "Retrieving next bug id");
      return;
    }

    // insert the bug itself
    try {
      store.insertBug(bug_id, project_id, version, priority,
                      severity, summary, user.getUserid(), conn);
    } catch (BugTrackerException be) {
      handleException(be, "Can't insert bug");
      return;
    }

    // insert an initial comment for the bug
    try {
      store.insertBugText(bug_text_id, bug_id, user.getUserid(),
                          comments, conn);
    } catch (BugTrackerException be) {
      handleException(be, "Can't insert bug comment");
      return;
    }


    // save success steam vars to display to user
    steam_vars.put("BTRACK_SUCCESSFUL_BUG_ADD", "1");
    steam_vars.put("BTRACK_NEW_BUG_NAME", summary);
    displayPage("btrack/mod_complete.tmpl", steam_vars);

  }

  /**
   * Called internally by handleEdit to handle any URLs of the form
   * &quot;btrack?edit+bug...&quot; - will determine if the url asks
   * to edit a new bug or an existing ID and make sure that a new bug
   * request supplies the project ID as well - will give appropriate
   * errors or call other functions as needed. That is also the url
   * to use for submitting edits, so we will check form_hash to see if
   * that's the case.
   */
  private void handleEditBug()
  {
    int bug_id;
    int project_id = -1;

    boolean isnewbug = false;

    if (url_tokens[3].equals("new"))
    {
//GET RID OF THIS AFTER CHECK!!
      log (GENERIC_LOG_LEVEL, "ERROR: invalid token string; " +
           "edit new bug via edit+new");
    } else
    { // otherwise they edit a pre-existing bug, not a new bug
      // first we figure out what bug id they want, error if we don't
      // understand it
      try
      {
        bug_id = Integer.parseInt(url_tokens[3]);
      } catch (NumberFormatException nfe)
      {
        log (GENERIC_LOG_LEVEL, "ERROR: BugTracker got bad " +
             "edit-bug id token: " + url_tokens[3]);
        Loco.sendErrorMsg(generic_error_message, hd);
        return;
      }

      // now we have the bug id - if the form_hash is null, it means
      // they just want to bring up the editing form. If form_hash is
      // not null they are submitting a form to update a bug. Call
      // the appropriate function:
      if (hd.form_hash == null)
        handleEditBugFormGen(bug_id);
      else
        handleEditBugSubmission(bug_id);
    }
  }

  /**
   * Given a determined bug id, show the page to edit it (after getting
   * the applicable information out of the database.)
   *
   * @param isnewbug is a boolean that controls the event logging - if true
   * the logging statement will reflect the fact that this was a bug being
   * edited for the first time.
   */
  private void handleEditBugFormGen(int bug_id)
  {

    Vector project_ids = new Vector();
    Vector project_names = new Vector();
    String comment_list;

    BugTrackerStore store = null;
    try {
      store = BugTrackerStore.getInstance();
    } catch (BugTrackerException be) {
      handleException(be, "can't instantiate store");
      return;
    }


    try {
      boolean b = store.getBugData(bug_id, steam_vars, conn);

      if (!b) {
        log (GENERIC_LOG_LEVEL, "ERROR: BugTracker: " +
             "handleEditBugFormGen: got non-existant bug id " +
             bug_id);
        Loco.sendErrorMsg("No such bug exists: " + bug_id +" ", hd);
        return;
      }

    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
    }



    try {
      store.populateProjectIDsAndNames(project_ids, project_names, conn);
    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
    }

    // Bad values for the project_name are screened out for us by
    // the above call.

    // Given the bug id, we need to assemble the list of comments
    // (if any) that the users have submitted so we can show it
    // (for context) on the edit bug screen.
    try {
      comment_list = store.assembleCommentChunk(bug_id, conn);
    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }

    int length = project_ids.size();
    String[] project_ids_strings = new String[length];
    String[] project_names_strings = new String[length];

    int project_id = 0;
    try {
      project_id = store.getProjectIdGivenBugId(bug_id, conn);

      String project_name = populateProjectIDsAndNamesForArray(project_ids_strings,
                                                               project_names_strings,
                                                               project_ids,
                                                               project_names,
                                                               project_id);
    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }

    // Put everything into the environment for the user:
    steam_vars.put("BTRACK_BUG_COMMENTS", comment_list);
    steam_vars.put("BTRACK_PROJECT_IDS_STRINGS", project_ids_strings);
    steam_vars.put("BTRACK_PROJECT_NAMES", project_names_strings);
    steam_vars.put("BTRACK_PROJECT_COUNT", "" + project_names_strings.length);

    // Now we're ready to show the template:
    displayPage ("btrack/edit_bug.tmpl", steam_vars);
    logEvent("btrack", "edit_bug", "formgen");
  }



  /**
   * This is called internally by handleEditBug when it is determined that
   * the user is submitting a form from the bug editing page. It will
   * attempt to save the data to the database and send the user to the
   * list-bugs screen for the associated project. If there are errors,
   * it will either display the edit page again with the same data the user
   * had tried to submit or give some other error screen.
   */
  private void handleEditBugSubmission(int bug_id)
  {
    int project_id = 1;
    StringBuffer errors = new StringBuffer("");

    // get all the form values
    String status = form_data.getField("BTRACK_STATUS");
    String severity = form_data.getField("BTRACK_SEVERITY");
    String priority = form_data.getField("BTRACK_PRIORITY");
    String version = form_data.getField("BTRACK_VERSION");
    String project_name = form_data.getField("BTRACK_PROJECT_NAME");
    String summary = form_data.getField("BTRACK_SUMMARY");
    String comments = form_data.getField("BTRACK_BUG_COMMENTS");
    String added_comment = form_data.getField("BTRACK_BUG_NEW_COMMENT");

    BugTrackerStore store = null;
    try {
      store = BugTrackerStore.getInstance();
    } catch (BugTrackerException be) {
      handleException(be, "can't instantiate store");
      return;
    }

    // confirms the above values conform to string constraints
    if (!form_data.hasAnyErrors())
    {
      logEvent("DEBUG", "status:", status);

      // no errors so far - keep plugging along...
      // check that the values for status and priority are valid
      if (!BUG_STATUS_CHOICES.contains(status))
        errors.append("\nStatus value is not a valid choice");
      if (!BUG_PRIORITY_CHOICES.contains(priority))
        errors.append("\nPriority value is not a valid choice");
      if (!BUG_SEVERITY_CHOICES.contains(severity))
        errors.append("\nSeverity value is not a valid choice");

      if (added_comment.length() > MAX_DESCRIPTION_LENGTH)
        errors.append("New comment can be a maximum of " +
                      MAX_COMMENT_LENGTH + " characters.<br>\n");

      // Make sure that the project actually exists in db
      try {
        if (!store.projectExists(store.getProjectIdGivenShortName(
          project_name,conn), conn))
        {
          errors.append("The parent project does not exist.");
        }
      } catch (BugTrackerException be) {
        handleException(be, generic_error_message);
        return;
      }

    } else
    { // save the errors from form_data
      errors.append(form_data.allErrorMsgs());
    }

    // if there are errors at this point, return the page with the error
    // information.
    if (!errors.toString().equals(""))
    {
      steam_vars.put("BTRACK_EDIT_ERRORS", errors);

      // save form values for redisplay	
      // project list must be regenerated

      String[] projects = new String[EST_MAX_PROJECTS];
      try{
        int num_of_projects = store.getProjectNamesFromDB(projects, conn);
        steam_vars.put("BTRACK_PROJECT_COUNT", String.valueOf(num_of_projects));
      } catch (BugTrackerException be)
      {
        handleException(be, "getProjectNamesFromDB for redisplay");
      }
      steam_vars.put("BTRACK_PROJECT_NAMES", projects);
      steam_vars.put("BTRACK_BUG_ID", String.valueOf(bug_id));
      form_data.copyAllIntoSteam(steam_vars);

      displayPage("btrack/edit_bug.tmpl", steam_vars);
      return;
    }

    // get the project id given the project_name
    try {
      project_id = store.getProjectIdGivenShortName(project_name,conn);
    } catch (BugTrackerException be) {
      handleException(be, "Retrieving project_id given short_name");
      return;
    }



    // Ok, we've got the data and things look good. Let's try to write
    // everything to the database:

    // All the database specific stuff moved to the relevant store - ee
    try {
      store.updateBug(bug_id, project_id, version, priority, status, severity, summary, conn);

      if (!added_comment.equals(""))
      {
        // If they submitted a comment then we need to get the
        // userid.

        // We should say user.isAnonymous() and if so,
        // set userid to User.ANONYMOUS_ID, but coincidentally
        // that's what it'll be if the user is anonymous, so just
        // to save some time we'll do this. Later, if we restrict
        // anonymous users and don't allow them to make new bugs
        // and whatnot, it'll be moot, since we'll be checking
        // isAnonymous anyway.
        int creator_userid = user.getUserid();
        int bug_text_id = store.getNextBugTextId(conn);
        store.insertBugText(bug_text_id, bug_id, creator_userid, added_comment, conn);
      }
    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }

    // inform the user that his/her changes have been made successfully.
    steam_vars.put("BTRACK_SUCCESSFUL_BUG_EDIT", "1");
    steam_vars.put("BTRACK_EDIT_BUG_NAME", summary);
    displayPage("btrack/mod_complete.tmpl", steam_vars);

    // Then we log the event here to indicate that it was a form submit
    // moving to the list projects screen.
    logEvent("btrack", "edit_bug", "formsubmit_to_projlist");

    return;
  }


  /**
   * Called internally by handleEdit to deal with editing projects. Will
   * figure out what the url indicates - either editing an existing
   * project or making a new project and editing that.
   */
  private void handleEditProject()
  {
    int project_id;

    try
    {
      project_id = Integer.parseInt(url_tokens[3]);
    } catch (NumberFormatException nfe)
    {
      log (GENERIC_LOG_LEVEL, "ERROR: BugTracker got bad " +
           "edit-project id token: " + url_tokens[3]);
      Loco.sendErrorMsg(generic_error_message, hd);
      return;
    }

    // Now the project id is figured out... we call
    // handleEditProjectFormGen if they are just wanting to edit
    // the project to begin with or handleEditProjectSubmission if
    // they are actually submitting a form edit that they want saved
    // to the database:
    if (hd.form_hash == null)
      handleEditProjectFormGen(project_id);
    else
      handleEditProjectSubmission(project_id);

  }

  /**
   * Called internally by handleEditProject to show the edit form for
   * a known project id.
   * @param isnewproject controls how the event logging happens - if this
   * variable is true it will show that the user is editing a project for
   * the first time.
   */
  private void handleEditProjectFormGen(int project_id)
  {

    try {
      boolean b = BugTrackerStore.getInstance().
        getProjectData(project_id, steam_vars, conn);

      if (!b) {
        log (GENERIC_LOG_LEVEL, "ERROR: BugTracker: " +
             "handleEditProjectFormGen got non-existant " +
             "project id " + project_id);
        Loco.sendErrorMsg(
                         "No such project exists: " + project_id + "... " +
                         "someone may have recently deleted it or the link " +
                         "may be corrupt.", hd);
      }

    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }



    // Display the page:
    displayPage ("btrack/edit_project.tmpl", steam_vars);

    logEvent("btrack", "edit_project", "formgen");
  }


  /**
   * Called internally to save the user's input to the database when
   * editing a project. Will check out the form, attempt to save the data,
   * and then redisplay the edit form when done. If anything is wrong with
   * the submitted data it will give an error back to the user along with the
   * form containing the erroneous data for correction. If anything else goes
   * wrong, appropriate errors will be sent.
   */
  private void handleEditProjectSubmission(int project_id)
  {
    // First let's get the data from the form hash that pertains
    // to the editing process (there's lots more in there in case we
    // need to redisplay the form but we don't need it right away):
    String short_title = (String) (hd.form_hash.get("BTRACK_SHORT_TITLE"));
    String long_title = (String) (hd.form_hash.get("BTRACK_LONG_TITLE"));
    String status = (String) (hd.form_hash.get("BTRACK_STATUS"));
    String text = (String) (hd.form_hash.get("BTRACK_TEXT"));
    String project_name = (String) (hd.form_hash.get("BTRACK_PROJECT_NAME"));

    // Error buffer will record any errors we encounter and bunch them up
    // for batch notification upon form regeneration. Hopefully we won't
    // need this variable but sadly we often do:
    StringBuffer error_buffer = new StringBuffer();

    error_buffer.append(form_data.allErrorMsgs());
    // Check all the string values for distasteful lengths:
    if (short_title.length() > MAX_SHORT_TITLE_LENGTH)
      error_buffer.append("Short name can be a maximum of " +
                          MAX_SHORT_TITLE_LENGTH + " characters.<br>\n");
    if (long_title.length() > MAX_LONG_TITLE_LENGTH)
      error_buffer.append("Long name can be a maximum of " +
                          MAX_LONG_TITLE_LENGTH + " characters.<br>\n");
    if (!BUG_STATUS_CHOICES.contains(status))
      error_buffer.append("\nStatus value is not a valid choice");
    if (text.length() > MAX_DESCRIPTION_LENGTH)
      error_buffer.append("Description can be a maximum of " +
                          MAX_DESCRIPTION_LENGTH + " characters.<br>\n");


    BugTrackerStore store = null;
    try {
      store = BugTrackerStore.getInstance();
    } catch (BugTrackerException be) {
      handleException(be, "can't instantiate store");
      return;
    }

    // If there were any errors a'tall, show the edit form again with
    // the same data the user submitted so they can correct things:
    if (error_buffer.length() != 0)
    {

      steam_vars.put("BTRACK_EDIT_ERRORS", error_buffer);

      // save form values for redisplay	
      // project list must be regenerated

      String[] projects = new String[EST_MAX_PROJECTS];
      try{
        int num_of_projects = store.getProjectNamesFromDB(projects, conn);
        steam_vars.put("BTRACK_PROJECT_COUNT", String.valueOf(num_of_projects));
      } catch (BugTrackerException be) {
        handleException(be, "getProjectNamesFromDB for redisplay editproject");
      }
      steam_vars.put("BTRACK_PROJECT_NAMES", projects);
      form_data.copyAllIntoSteam(steam_vars);

      displayPage("btrack/edit_project.tmpl", steam_vars);
      return;

    }

    // Ok, everything went peachy, so let's go ahead and write the
    // stuff into the database:
    // Moved database specific code to the store classes - ee
    try {

      int results = store.updateProject(project_id, short_title, long_title,
                                        text, status, conn);

      // results is the number of rows updated, which should be
      // exactly 1
      if (results != 1)
      {
        log (URGENT_LOG_LEVEL, "ERROR: BugTracker: " +
             "handleEditProjectSubmission got non-1 result " +
             "making new project entry!");
        Loco.sendErrorMsg(generic_error_message, hd);
        return;
      }

    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }

    // Now we have saved the data to the database for the project, but
    // we need to show the edit page again with the "Update successful"
    // message and all the new data.

    // At this point, we really should read the modified date from the
    // database (to know exactly what it put in automatically for
    // SYSDATE) but as a shortcut we will approximate with the current
    // java time. We put it back into the form hash so
    // displayEditProjectAgain will pick up the change when it
    // redisplays. This is no big deal since people aren't going to be
    // intently cross-corellating the dates to the millisecond, but if
    // it were some system-critical thing where the user might be
    // looking up things based on some exact date you give back, you
    // should read the date back out with ResultSet.getTimestamp(). One
    // interesting thing is that getTimestamp and new Date() don't seem
    // to return exactly the same type of Date, so this date is
    // formatted a little differently.
    hd.form_hash.put("BTRACK_MODIFIED", "" + new Date() +
                     " (approximate)");



    // save success steam vars to display to user
    steam_vars.put("BTRACK_SUCCESSFUL_PROJECT_EDIT", "1");
    steam_vars.put("BTRACK_EDIT_PROJECT_NAME", short_title);
    displayPage("btrack/mod_complete.tmpl", steam_vars);


    // And last and least log the event:
    logEvent("btrack", "edit_project", "update_formsubmit");

    return;
  }

  /**
   * Lists bugs.  Will determine from the url what the user is
   * looking to do and send them on their way through the method
   * calls.
   */
  protected void handleListBugs()
  {
    // Check for the first case, the listbugs+all command or the
    // listbugs+# command (see top):
    if (url_tokens.length == 3)
    {
      int project_id;

      // First we will try to get the project id by parsing. Failing
      // that, we will subsequently check to see if it's "all" instead
      // of a number.
      try
      {
        project_id = Integer.parseInt(url_tokens[2]);
      } catch (NumberFormatException nfe)
      {
        // Argh, it's not a number. Let's hope that it's "all":

        if (url_tokens[2].equals("all"))
        {
          // They want to list all of the bugs, eh? Well, they asked
          // for it:
          handleListBugs(-1, false, false, true);
          return;
        } else
        { // then they gave us garbage for a listbugs parameter
          log (GENERIC_LOG_LEVEL, "ERROR: BugTracker got bad " +
               "list_bugs pid token: " + url_tokens[2]);
          Loco.sendErrorMsg(generic_error_message, hd);
          return;
        }
      }

      // If the project id was obtained sucessfully, then we list them
      // some bugs. Even though it's not exactly proper, we allow them
      // to have a non-null form hash here to mean that they are
      // submitting the form from the list filter - they should have
      // used the btrack?listbugs url, but we'll go easy on them in the
      // interest of easy html editing. The filter submission for the
      // project id will override anything they have in the url:
      if (hd.form_hash == null)
      {
        // list the bugs just by the project id given in the url:
        handleListBugs(project_id, false, true, true);
        return;
      } else
      {
        // list the bugs by the submitted filter, even though they
        // didn't really use the right url:
        handleListBugs(project_id, true, true, true);
        return;
      }
    } else if (url_tokens.length == 2) // the "btrack?listbugs" case
    {
      // If no form_hash is present, they've just asked for the
      // filter form, so we gives them what they wants:
      if (hd.form_hash == null)
        displayFilterForm();
      else // otherwise list the bugs according to the submitted filter:
        handleListBugs(-1, true, false, true);

      return;
    } else
    {
      // then they gave us garbage. Therefore, we give them a meaningless
      // error message as punishment:
      Loco.sendErrorMsg(generic_error_message, hd);
      return;
    }
  }

  /**
   * Called internally by handleListBugs above - this version takes some
   * variables to indicate exactly how it should use the data that came in -
   * i.e. all the parsing of the url has been done and whatnot.
   *
   * No more words! Time for action!
   *
   * @param project_id the project_id for which we wish to list bugs - will
   *   be ignored if useFilter is true or userProjectID is false.
   * @param useFilter whether or not to use a filter to be found in form_hash
   *   to find bugs in the database
   * @param useProjectID if useFilter is false, whether or not to use the
   *   project_id passed into this call to determine which bugs to show - if
   *   useFilter is true and no project id is given from the filter form or
   *   NULL_PROJECT_ID is given then this variable will determine if the
   *   passed-in project_id should be used.
   * @param log_event whether or not to log an event at the end of this
   *   call - sometimes you want to do it yourself, you know?
   */
  private void handleListBugs(int project_id, boolean useFilter,
                              boolean useProjectID, boolean log_event) {

    //String statement = null;

    // Initialize some variables so we can try to get the integer
    // stuff from the form:
    int filter_project_id = NULL_PROJECT_ID;
    int filter_bug_id = NULL_BUG_ID;
    int filter_creator_userid = User.NULL_USERID;
    int filter_priority = NULL_PRIORITY_VALUE;

    String filter_status = null;
    String filter_creator_username = null;
    String filter_version = null;
    String filter_severity = null;
    String filter_orderby = null;



    if (useFilter) {
      // Ok, so we're in filter land now. We must get everything from
      // the form, make sure it makes sense, and then construct a clever
      // SQL statement that will give them what they want, and all for
      // under a dollar.

      // Get the string stuff from the form:
      filter_status = (String) (hd.form_hash.get("BTRACK_FILTER_STATUS"));
      filter_creator_username =
        (String) (hd.form_hash.get("BTRACK_FILTER_CREATOR_USERNAME"));
      filter_version = (String) (hd.form_hash.get("BTRACK_FILTER_VERSION"));
      filter_severity = (String) (hd.form_hash.get("BTRACK_FILTER_SEVERITY"));

      filter_orderby = (String) (hd.form_hash.get("BTRACK_FILTER_ORDERBY"));

      String filter_project_id_s =
      (String) (hd.form_hash.get("BTRACK_FILTER_PROJECT_SELECT"));
      String filter_bug_id_s =
      (String) (hd.form_hash.get("BTRACK_FILTER_BUG_ID"));
      String filter_creator_userid_s =
      (String) (hd.form_hash.get("BTRACK_FILTER_CREATOR_USERID"));
      String filter_priority_s =
      (String) (hd.form_hash.get("BTRACK_FILTER_PRIORITY"));


      // Screen any problem variables out just to guard against null
      // exceptions:
      if (filter_status == null) filter_status = "";
      if (filter_creator_username == null) filter_creator_username = "";
      if (filter_version == null) filter_version = "";
      if (filter_severity == null) filter_severity = "";

      if (filter_orderby == null) filter_orderby = "";

      if (filter_project_id_s == null) filter_project_id_s = "";
      if (filter_bug_id_s == null) filter_bug_id_s = "";
      if (filter_creator_userid_s == null) filter_creator_userid_s = "";
      if (filter_priority_s == null) filter_priority_s = "";



      // Get the integer stuff from the form if the user typed anything
      // into the field:
      try
      {
        if (!(filter_project_id_s.equals("")))
          filter_project_id = Integer.parseInt(filter_project_id_s);

        if (!(filter_bug_id_s.equals("")))
          filter_bug_id = Integer.parseInt(filter_bug_id_s);

        if (!(filter_creator_userid_s.equals("")))
          filter_creator_userid =
          Integer.parseInt(filter_creator_userid_s);

        if (!(filter_priority_s.equals("")))
          filter_priority =
          Integer.parseInt(filter_priority_s);
      } catch (NumberFormatException nfe)
      {
        // Uh oh, we couldn't parse the values they entered in the
        // form, ergo our destiny lay clear before us: we must show
        // ye olde forme once more.

        // display error message
        steam_vars.put("BTRACK_ERRORS", "One of the text fields was "+
                       "an invalid type.  Try again.");

        displayFilterForm();

        return;
      }


    } // if there's no filter just use all the null values
    else {
      filter_project_id = project_id;
    }

    BugTrackerStore store = null;
    try {
      store = BugTrackerStore.getInstance();
    } catch (BugTrackerException be) {
      handleException(be, "can't instantiate store");
      return;
    }

    // get a sortable table from the store
    SortableTable st = null;
    try {
      st = store.getBugs(filter_bug_id, filter_project_id, filter_status,
                         filter_priority, filter_severity, filter_version,
                         filter_creator_userid, filter_creator_username, filter_orderby,
                         conn);
    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }


    int length = st.getNumRows();

    // Hallelujah, all the data is here, now let's just insert it into the
    // subs environment and go for it:

    steam_vars.put("BTRACK_BUG_LIST_TABLE", st);
    /*
    steam_vars.put("BTRACK_BUG_LIST_TABLE_PROJECT_NAMES_ARRAY",
                   project_names);
    steam_vars.put("BTRACK_BUG_LIST_TABLE_BUG_DESCRIPTIONS_ARRAY",
                   bug_descriptions);
                   */

    // Display the template with our new data:
    displayPage ("btrack/list_bugs.tmpl", steam_vars);

    // If the caller has requested an event to be logged, do that one
    // kind favour they ask of us:
    if (log_event)
      logEvent("btrack", "list_bugs", "f_" + useFilter +
               "_p_" + useProjectID);
  }


  /**
   * Take the passed in vectors and put their elements into the supplied
   * arrays along with some string padding to make it work in the <select>
   * tag we're working up (html quoting, mainly.) Also, according to the
   * id passed in, tell us what the name of the currently-chosen value
   * is and set the <select> tag to be pre-selected for that value.
   */
  private String populateProjectIDsAndNamesForArray(
                                                   String[] project_ids_strings,
                                                   String[] project_names_strings,
                                                   Vector project_ids,
                                                   Vector project_names,
                                                   int current_project_id)
  {
    // We set to this value just in case we don't find a match between
    // our parent project id and a name - though that should never happen.
    String project_name = "<undefined>";

    int length = project_ids_strings.length;

    for (int i = 0; i < length; i++)
    {
      if (current_project_id !=
          ((Number) (project_ids.elementAt(i))).intValue())
        project_ids_strings[i] = "\"" + project_ids.elementAt(i) +
                                 "\"";
      else
      {
        project_name = (String) project_names.elementAt(i);
        project_ids_strings[i] = "\"" + project_ids.elementAt(i) +
                                 "\" selected";
      }

      project_names_strings[i] = StringUtil.htmlQuoteString(
                                                           ((String) project_names.elementAt(i)));
    }

    return(project_name);
  }


  /**
   * display to the client the "list bugs filter" tmpl file. Will
   * also log an event.
   */
  private void displayFilterForm()
  {
    Vector project_ids = new Vector();
    Vector project_names = new Vector();

    // We artificially insert an option into the vectors so the user
    // can select a "don't care" option. The populateProjectIDsAndNames
    // call will append to the end of the supplied vectors, so our info
    // will stay in there. We give it NULL_PROJECT_ID so that if the user
    // selects it (and it will be selected by default) we will be able
    // to discern that value from the valid id's they might have selected.

    // The populateProjectIDsAndNames call is actually a single-step
    // thing so we don't really need to synchronize here, but just for
    // extra safety we will since it may change in the future.
    try {
      BugTrackerStore.getInstance().
        populateProjectIDsAndNames(project_ids, project_names, conn);
    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }


    // Now we prepare to translate the vectors that have been populated
    // into arrays for the tempalte evaluator.
    int length = project_ids.size();
    String[] project_ids_strings = new String[length];
    String[] project_names_strings = new String[length];

    // We will ignore what comes back since we are defaulting the select
    // option to "don't care" and don't need to know what the project
    // name is. We send in "NULL_PROJECT_ID" to ensure that our
    // "don't care" value is the one selected by default, even though
    // most browsers will use the first option in the <select> as the
    // default anyway.
    populateProjectIDsAndNamesForArray(project_ids_strings,
                                       project_names_strings,
                                       project_ids,
                                       project_names,
                                       NULL_PROJECT_ID);

    steam_vars.put("BTRACK_FILTER_PROJECT_IDS_STRINGS", project_ids_strings);
    steam_vars.put("BTRACK_FILTER_PROJECT_NAMES", project_names_strings);
    steam_vars.put("BTRACK_FILTER_PROJECT_COUNT", "" +
                   project_names_strings.length);

    displayPage ("btrack/list_bugs_filter.tmpl", steam_vars);

    logEvent("btrack", "list_bugs_filter", "formgen");
  }




  /**
   * display to the client the list of all the projects. See how
   * short this function is? It shows you what utility functions can
   * do for you.
   */

  protected void handleListProjects()
  {
    // We'll use a SortableTable object to make quick work of those
    // pesky database rows:
    SortableTable st = new SortableTable(10, 8, 5, 0);
    try {

      // This is all done in the store now - ee
      st = BugTrackerStore.getInstance().getProjects(conn);

    } catch (BugTrackerException be) {
      handleException(be, generic_error_message);
      return;
    }


    // Now that we have the sortable table all set up, let's display
    // the actual page.
    steam_vars.put("BTRACK_PROJECT_LIST_TABLE", st);
    displayPage ("btrack/list_projects.tmpl", steam_vars);

    // And finally we log an event:
    logEvent("btrack", "list_projects", "");
  }



}
