/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.codegen;

import org.locomotive.loco.*;
import org.locomotive.server.*;
import org.locomotive.util.*;
import org.locomotive.loco.handler.*;
import java.io.*;
import java.util.*;

/**
 * CodeGenerator is a Java program which generates source code for other
 * Java programs.  It can take a class declaration from an HTML Form field,
 * as a Handler, or it can take it from the standard input.
 * The format of the class should look like:
 * <PRE>
 * public class MyClass
 * {
 *     int id;
 *     String str;
 * }
 * </PRE>
 * The first field in the class is assumed to be the primary identifier --
 * in other words, its PRIMARY KEY in the database.
 */
public class CodeGenerator extends GenericHandler
{
    static String correct_format =
        "<PRE>public class MyClass\n{    int id;\n    String name (5..30);}</PRE>";
    
    String class_name = null;
    
    public void handleRequest(HandlerData hd)
    {
	super.handleRequest(hd);

	if (form_data.isEmpty())
	{
	    displayPage ("codegen/form.tmpl", steam_vars);
	    logEvent ("codegen", "disp_form", "");
	    return;
	}
	
	String class_definition = form_data.getString ("CLASS_DEFINITION",
						       "class definition",
						       0, 2000);

	if (class_definition == null || class_definition.length() == 0) 
	{
	    displayString ("The form was empty.  Please try re-submitting.");
	    logEvent ("codegen", "submit_empty", "");
	    return;
	}
	
	String result = parseClassDefinition (class_definition);
	
	if (! result.equals("success"))
	{
	    displayString (result);
	    logEvent ("codegen", "submit_errors", "");
	    return;
	}
    }
    
    public static void main(String argv[])
    {
	new CodeGenerator().parseClassDefinition("");
    }
	
    protected String parseClassDefinition (String class_definition)
    {
//	class_definition =
//	    "public class MyClass\n{    int id_for_me (5..30);\n    String name_of_this;\n}";
log(0, "class_definition = '" + class_definition + "'.");
	StringTokenizer st = new StringTokenizer(class_definition, "\n{}",
						 false);
	String line = null;

	SortableTable instance_variables =
	    new SortableTable(st.countTokens(), 4);

	instance_variables.setColumnName(0, "VAR_NAME");
	instance_variables.setColumnName(1, "VAR_NAME_MIXED_CASE");
	instance_variables.setColumnName(2, "VAR_TYPE");
	instance_variables.setColumnName(3, "VAR_TYPE_MIXED_CASE");
	instance_variables.setColumnName(4, "VAR_MINLENGTH");
	instance_variables.setColumnName(5, "VAR_MAXLENGTH");
	
	if (st.hasMoreTokens())
	{
	    line = st.nextToken();
log(0, "first line = '" + line + "'.");
	    if (line.startsWith("public class "))
		line = line.substring(13).trim();
	    else if (line.startsWith("class "))
		line = line.substring(6).trim();
	    
	    if (line.length() > 0)
	    {
		class_name = line;
		steam_vars.put ("CLASS_NAME", class_name);
	    }
	    else
	    {
		return ("No class name was specified.  Please use the "+
			"format: " + correct_format);
	    }
	}

	int num_instance_vars = 0;

	if (! st.hasMoreTokens())
	{
	    return ("No instance variables were specified.  Please use the "+
		    "format: " + correct_format);
	}

	while (st.hasMoreTokens())
	{
	    line = st.nextToken().trim();
	    if (line.length() == 0)
		continue;
log (0, "line = " + line);
	    StringTokenizer st1 = new StringTokenizer(line, " (.);", false);

	    String var_type = null;
	    String var_name = null;
	    int var_minlength = 0;
	    int var_maxlength = 0;
	    
	    if (st1.countTokens() > 1)
	    {
		var_type = st1.nextToken();
		if (var_type.equalsIgnoreCase("date"))
		    var_type = "Timestamp";
log (0, "var_type = " + var_type);
		var_name = st1.nextToken();
log (0, "var_name = " + var_name);
	    }
	    else
	    {
		return ("Please specify the correct type and name for each " +
			"instance variable, for example: " + correct_format);
	    }
	    
	    if (st1.countTokens() > 1)
	    {
		try {
		    var_minlength = Integer.parseInt(st1.nextToken());
log (0, "var_minlength = " + var_minlength);
		    var_maxlength = Integer.parseInt(st1.nextToken());
log (0, "var_maxlength = " + var_maxlength);
		}
		catch (NumberFormatException nfe)
		{
		    var_minlength = 0;
		    var_maxlength = 0;
		}
	    }
	    
	    insertInstanceVariable (instance_variables, var_type, var_name,
				    var_minlength, var_maxlength,
				    num_instance_vars);
	    num_instance_vars++;
	}

	steam_vars.put ("INSTANCE_VARIABLES", instance_variables);
log (0, "INSTANCE_VARIABLES = " + instance_variables);
	
	displayPage ("codegen/java_class.tmpl", steam_vars);
	logEvent ("codegen", "success", "");
	
	return "success";
    }
    
    void insertInstanceVariable (SortableTable instance_variables,
				 String var_type, String var_name,
				 int var_minlength, int var_maxlength,
				 int num_instance_vars)
    {
	if (num_instance_vars == 0)
	{
	    steam_vars.put ("FIRST_VAR_TYPE", var_type);
	    steam_vars.put ("FIRST_VAR_NAME", var_name);
	}
	
	instance_variables.insert (var_type, num_instance_vars, "VAR_TYPE");
	instance_variables.insert (changeToMixedCase(var_type),
				   num_instance_vars, "VAR_TYPE_MIXED_CASE");
	
	instance_variables.insert (var_name, num_instance_vars, "VAR_NAME");
	instance_variables.insert (changeToMixedCase (var_name),
				   num_instance_vars, "VAR_NAME_MIXED_CASE");
	
	instance_variables.insert (new Integer(var_minlength),
				   num_instance_vars, "VAR_MINLENGTH");
	instance_variables.insert (new Integer(var_maxlength),
				   num_instance_vars, "VAR_MAXLENGTH");
    }

    
    String changeToMixedCase (String str)
    {
	StringBuffer sb = new StringBuffer(str.length());
	
	StringTokenizer st = new StringTokenizer(str, "_");
	while (st.hasMoreTokens())
	{
	    String word = st.nextToken();
	    sb.append (word.substring(0,1).toUpperCase());
	    sb.append (word.substring(1));
	}
	
	return sb.toString();
    }
}
