/**
 * CodeGeneratorServlet.java
 *
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License. 
 *
 * The initial develop of this code under the LPL Leverage Information Systems.
 * Portions created by Leverage are Copyright (C) 1998 Leverage Information Systems.
 * All Rights reserved.
 */

package org.locomotive.module.codegen;

import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.server.*;
import org.locomotive.util.*;
import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;
import java.util.*;

/**
 * CodeGenerator is a Java program which generates source code for other
 * Java programs.  CodeGenerator can read a class declaration from an
 * HTML Form field, and fill in some Steam variables which allow the template
 * java_class.tmpl to fill in many details of a full LocoServlet class.
 * The format of the class declaration should look like:
 * <PRE>
 * public class MyClass
 * {
 *     String str;
 * }
 * </PRE>
 * CodeGenerator will create a primary identifier field, called "id", and
 * make it the first field.  This field will be this class's PRIMARY KEY
 * in the database.
 */
public class CodeGeneratorServlet extends LocoServlet
{
    static String correct_format =
        "<PRE>public class MyClass\n{    int id;\n    String name (5..30);}</PRE>";
    
    String class_name = null;
    
    public void doGet(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	StandardInsertions.insertAll(req, steam_vars);

	displayPage ("codegen/java_class.tmpl", steam_vars);
	logEvent ("codegen", "disp_form", "");
    }

    public void doPost(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	if (form_data.isEmpty())
	{
	    doGet(req, resp);
	    return;
	}

	form_data.copyAllIntoSteam (steam_vars);
	
	String class_definition = form_data.getString ("CLASS_DEFINITION",
						       "class definition",
						       0, 2000);

	if (class_definition == null || class_definition.length() == 0) 
	{
	    steam_vars.put ("FORM_ERRORS",
			    "The form was empty.  Please try re-submitting.");
	    displayPage ("codegen/java_class.tmpl", steam_vars);
	    logEvent ("codegen", "submit_empty", "");
	    return;
	}
	
	String result = parseClassDefinition (class_definition);
	
	if (! result.equals("success"))
	{
	    steam_vars.put ("FORM_ERRORS", result);
	    displayPage ("codegen/java_class.tmpl", steam_vars);
	    logEvent ("codegen", "submit_errors", "");
	    return;
	}
    }

    /**
     * To be implemented in the future: read a class definition from the
     * standard input and generate a Java class and HTML form on the output.
     */
    public static void main(String argv[])
    {
	new CodeGenerator().parseClassDefinition("");
    }
    
    protected String parseClassDefinition (String class_definition)
	throws ServletException, IOException
    {
//	class_definition =
//	    "public class MyClass\n{    int id_for_me (5..30);\n    String name_of_this;\n}";
log(0, "class_definition = '" + class_definition + "'.");
	StringTokenizer st = new StringTokenizer(class_definition, "\n{}",
						 false);
	String line = null;

	SortableTable instance_variables =
	    new SortableTable(st.countTokens(), 7);

	instance_variables.setColumnName(0, "VAR_NAME");
	instance_variables.setColumnName(1, "VAR_NAME_MIXED_CASE");
	instance_variables.setColumnName(2, "FORM_TYPE");
	instance_variables.setColumnName(3, "FORM_TYPE_MIXED_CASE");
	instance_variables.setColumnName(4, "JAVA_TYPE");
	instance_variables.setColumnName(5, "JAVA_TYPE_MIXED_CASE");
	instance_variables.setColumnName(6, "VAR_MINLENGTH");
	instance_variables.setColumnName(7, "VAR_MAXLENGTH");
	
	if (st.hasMoreTokens())
	{
	    line = st.nextToken();
log(0, "first line = '" + line + "'.");
	    if (line.startsWith("public class "))
		line = line.substring(13).trim();
	    else if (line.startsWith("class "))
		line = line.substring(6).trim();
	    
	    if (line.length() > 0)
	    {
		class_name = line;
		steam_vars.put ("CLASS_NAME", class_name);
	    }
	    else
	    {
		return ("No class name was specified.  Please use the "+
			"format: " + correct_format);
	    }
	}

	int num_instance_vars = 0;

	if (! st.hasMoreTokens())
	{
	    return ("No instance variables were specified.  Please use the "+
		    "format: " + correct_format);
	}

	Vector option_names = new Vector();
	
	while (st.hasMoreTokens())
	{
	    line = st.nextToken().trim();
	    if (line.length() == 0)
		continue;
log (0, "line = " + line);
	    StringTokenizer st1 = new StringTokenizer(line, " (.);,", false);

	    String form_type = null;
	    String java_type = null;
	    String var_name = null;
	    int var_minlength = 0;
	    int var_maxlength = 0;
	    
	    if (st1.countTokens() > 1)
	    {
		form_type = st1.nextToken();
		java_type = form_type;
		var_name = st1.nextToken();

		if (form_type.equalsIgnoreCase("date"))
		{
		    form_type = changeToMixedCase(form_type);
		    java_type = "Timestamp";
		}
		else if (form_type.equalsIgnoreCase("pulldown") ||
			 form_type.equalsIgnoreCase("radio"))
		{
		    java_type = "int";
		    form_type = changeToMixedCase(form_type);
		    var_minlength = option_names.size();
		    
		    int option_value = 0;
		    while (st1.hasMoreTokens())
		    {
			option_names.addElement (st1.nextToken());
			option_value++;
		    }
		    var_maxlength = option_value;
		}
		else if (form_type.equalsIgnoreCase("checkbox") ||
			 form_type.equalsIgnoreCase("boolean"))
		{
		    java_type = "boolean";
		    form_type = "Checkbox";
		    var_minlength = 0;
		    var_maxlength = 1;
		}
		else if (form_type.equalsIgnoreCase("yesornoradio"))
		{
		    java_type = "boolean";
		    form_type = "YesOrNoRadio";
		    var_minlength = 0;
		    var_maxlength = 1;
		}
		else if (form_type.equalsIgnoreCase("alpha") ||
			 form_type.equalsIgnoreCase("alphanum") || 
			 form_type.equalsIgnoreCase("digits") ||
			 form_type.equalsIgnoreCase("email") ||
			 form_type.equalsIgnoreCase("string"))
		{
		    java_type = "String";
		    form_type = changeToMixedCase(form_type);

		    if (st1.countTokens() > 1)
		    {
			try {
			    var_minlength = Integer.parseInt(st1.nextToken());
			    log (0, "var_minlength = " + var_minlength);
			    var_maxlength = Integer.parseInt(st1.nextToken());
			    log (0, "var_maxlength = " + var_maxlength);
			}
			catch (NumberFormatException nfe)
			{
			    var_minlength = 1;
			    var_maxlength = 30;
			}
		    }
	    
		}
		else if (form_type.equalsIgnoreCase("int") ||
			 form_type.equalsIgnoreCase("long"))
		{
		    if (st1.countTokens() > 1)
		    {
			try {
			    var_minlength = Integer.parseInt(st1.nextToken());
			    log (0, "var_minlength = " + var_minlength);
			    var_maxlength = Integer.parseInt(st1.nextToken());
			    log (0, "var_maxlength = " + var_maxlength);
			}
			catch (NumberFormatException nfe)
			{
			    var_minlength = 0;
			    var_maxlength = 0;
			}
		    }
	    
		}
		
log (0, "form_type = " + form_type);
log (0, "java_type = " + java_type);
log (0, "var_name = " + var_name);
	    }
	    else
	    {
		return ("Please specify the correct type and name for each " +
			"instance variable, for example: " + correct_format);
	    }
	    
	    insertInstanceVariable (instance_variables, form_type, java_type,
				    var_name, var_minlength, var_maxlength,
				    num_instance_vars);
	    num_instance_vars++;
	}
	
	String[] option_names_list = new String [option_names.size()];
	option_names.copyInto (option_names_list);
	steam_vars.put ("OPTION_NAMES", option_names_list);

	steam_vars.put ("INSTANCE_VARIABLES", instance_variables);
	steam_vars.put ("NUM_INSTANCE_VARIABLES",
			new Integer(num_instance_vars));

log (0, "INSTANCE_VARIABLES = " + instance_variables);
	
	displayPage ("codegen/java_class.tmpl", steam_vars);
	logEvent ("codegen", "success", "");
	
	return "success";
    }
    
    void insertInstanceVariable (SortableTable instance_variables,
				 String form_type, String java_type,
				 String var_name,
				 int var_minlength, int var_maxlength,
				 int num_instance_vars)
    {
	if (num_instance_vars == 0)
	{
	    steam_vars.put ("FIRST_FORM_TYPE", form_type);
	    steam_vars.put ("FIRST_VAR_NAME", var_name);
	}
	
	instance_variables.insert (form_type, num_instance_vars, "FORM_TYPE");
	instance_variables.insert (changeToMixedCase(form_type),
				   num_instance_vars, "FORM_TYPE_MIXED_CASE");
	
	instance_variables.insert (java_type, num_instance_vars, "JAVA_TYPE");
	instance_variables.insert (changeToMixedCase(java_type),
				   num_instance_vars, "JAVA_TYPE_MIXED_CASE");
	
	instance_variables.insert (var_name, num_instance_vars, "VAR_NAME");
	instance_variables.insert (changeToMixedCase (var_name),
				   num_instance_vars, "VAR_NAME_MIXED_CASE");
	
	instance_variables.insert (new Integer(var_minlength),
				   num_instance_vars, "VAR_MINLENGTH");
	instance_variables.insert (new Integer(var_maxlength),
				   num_instance_vars, "VAR_MAXLENGTH");
    }

    
    String changeToMixedCase (String str)
    {
	StringBuffer sb = new StringBuffer(str.length());
	
	StringTokenizer st = new StringTokenizer(str, "_");
	while (st.hasMoreTokens())
	{
	    String word = st.nextToken();
	    sb.append (word.substring(0,1).toUpperCase());
	    sb.append (word.substring(1));
	}
	
	return sb.toString();
    }
}
