/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.discuss;

import java.util.*;
import java.sql.*;
import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;
import org.locomotive.util.*;
import org.locomotive.loco.perm.*;
import org.locomotive.server.Log;


/** 
 * This class handles the administration for the communities.  It provides
 * screens which allow a user to modify discussion, topics, and posts,
 * as well as add them.  It is desgined to allow quick and easy moderation
 * of discussions; at the possible expense of some performance.  This
 * should be okay because only a few administrators should be seeing these
 * pages.  Permission to view these pages is based on whether or
 * not the user has org.locomotive.module.discuss.DiscussionAdminHandler.COMM_ADMIN
 * permission, listed below.
 */
public class DiscussionAdminHandler extends GenericHandler
{
    /**
     * String constants for permission names used in this handler:
     */

    /**
     * Allows a user to administer user access control for discussions and
     * topics:
     */
    public static final String COMM_ADMIN = "COMM_ADMIN";

    /**
     * Allows a user read access to a discussion or topic:
     */
    public static final String COMM_READ = "COMM_READ";

    /**
     * Since categories and topics share the same ids since they both start at
     * 1, we must make a distinction between category and topic ids for
     * the acls.  Therefore acl ids for categories are the same as the
     * category ids, while acl ids for topics are the topic ids plus this
     * filter.  Please note that we cannot have more than 1000000 categories!
     */
    public static final int TOPIC_TO_ACL_FILTER = 1000000;

    private HandlerData hd;

    /**
     * This handles the administration for communities discussions, topics, and
     * posts.  Shows an unknown handler error page if the user does not have
     * the correct admin permission to be here.
     * All templates for communities admin should be in the 
     * 'discuss/admin/commmgr' directory. 
     * The tokens and their corresponding methods are:
     * <BR><PRE>
     *  Token            Method               Template
     *  MOD              modifyDiscussion     comm_disc_filledtempl.tmpl
     *  TOP              handleAdminTopics()  comm_topics_list.tmpl
     *  TOP+MOD          modifyTopic()        comm_topics_filledtempl.tmpl
     *  POST             handleAdminPosts()   comm_posts_list.tmpl
     *  POST+MOD         modifyPost()         comm_posts_filledtempl.tmpl
     *  (default)        handleAdminDefault() comm_disc_list.tmpl
     *  USER             modifyDiscUsers      comm_disc_users.tmpl
     *  TOP+USER         modifyTopicUsers     comm_topic_users.tmpl
     *  
     * </PRE>
     */
    public void handleRequest(HandlerData hd_in)
    {
	super.handleRequest(hd_in);
	hd = hd_in;

	if(user.isAnonymous())
	{
	    // Not allowed here; pretend they got a unknown handler
	    displayPage ("error/unknown_handler_option.tmpl", steam_vars);
	    return;
	}
	else
	{
	    PermissionSet pSet = null;
	    
	    try
	    {
		pSet = UserPerm.getPermissions(user.getUserid(), conn);
	    }
	    catch(SQLException se)
	    {
		log (0, "An error occured accessing permissions" +
			" for Userid " + user.getUserid() + ": " +
			StringUtil.exceptionStackTrace(se));
		displayPage ("error/generic_error.tmpl", steam_vars);
		return;
	    }
	    	
	    // See if the user has permission to be here
	    if (!pSet.isSet(COMM_ADMIN))
	    {
		// Is not an administrator.  Show them the unknown handler
		// error page.
		displayPage ("error/unknown_handler_option.tmpl", steam_vars);
		return;
	    }
	}

	if (url_tokens.length < 4)
	{
	    if (hd.form_hash != null)
	    {
		deleteDiscussions();
	    }

	    showAdminDefault();
	}
	else if (url_tokens[2].equals("MOD"))
	{
	    try
	    {
		int catId = Integer.parseInt(url_tokens[3]);
		modifyDiscussion(catId);
	    }
	    catch (NumberFormatException e)
	    {
		showAdminDefault();
	    }
	}
	else if (url_tokens[2].equals("TOP"))
	{
	    handleAdminTopics();
	}
	else if (url_tokens[2].equals("POST"))
	{
	    handleAdminPosts();
	}
	else if (url_tokens[2].equals("USER"))
	{
	    int catId = Integer.parseInt(url_tokens[3]);
	    modifyDiscUsers(catId);
	}
	else
	{
	    showAdminDefault();
	}
    }


    /** use this to display the communities admin home page.  */
    private void showAdminDefault()
    {
	SortableTable categories = 
	    DiscussionHandler.lib.getCategoryList(conn, "LIVE");
	if (categories != null)
	    steam_vars.put("CAT_LIST", categories);

	DiscussionHandler.insertTemplate(hd, "Admin");
    }


    
    private void deleteDiscussions()
    {
	Vector v = new Vector(hd.form_hash.size());
	Enumeration e = hd.form_hash.keys();
	for(int i = 0; e.hasMoreElements(); i++)
	{
	    try
	    {
		int catId = Integer.parseInt(e.nextElement().toString());
		DiscussionHandler.lib.changeCategoryStatus(conn, catId, "DEAD");
		v.addElement(hd.form_hash.get(String.valueOf(catId)));
	    }
	    catch (NumberFormatException ne) {}
	}

	StringBuffer message = new StringBuffer();
	if (v.size() > 0)
	{
	    message.append("The following categories have been deleted: ");
	    message.append("\n\t<BLOCKQUOTE>\n");
	    for (int i = 0; i < v.size(); i++)
	    {
		message.append("\t\t");
		message.append(v.elementAt(i).toString());
		message.append("<BR>\n");
	    }
	    message.append("\t</BLOCKQUOTE>\n");
	}
	else
	{
	    message.append("No categories were deleted");
	}
		
	steam_vars.put("message", message.toString());
    }


    /** handle modification of a discussion   if no post, display the
     * page with the info.  Otherwise, update the info with what was
     * posted.  If the number is less than 1, then create a new discussion.
     */
    private void modifyDiscussion(int catId)
    {
	if (hd.form_hash != null)
	{
	    String title = (String) hd.form_hash.get("COMMUNITYNAME");
	    String body = (String) hd.form_hash.get("COMMUNITYDESC");
	    String status = (String) hd.form_hash.get("COMMUNITYSTATUS");

	    if (title == null || body == null || title.length() < 1)
	    {
		steam_vars.put("message", "You must enter a valid title. " +
			    "Please try again");
		steam_vars.put("NEW_CAT", "true");
		steam_vars.put("COMMUNITYID", String.valueOf(catId));
		if (body != null) 
		    steam_vars.put("COMMUNITYDESC", body);
		if (title != null)
		    steam_vars.put("COMMUNITYNAME", title);
		if (status != null)
		    steam_vars.put("COMMUNITYSTATUS", status);

		DiscussionHandler.insertTemplate(hd, "Admin+MOD");
		return;
	    }
		
	    title = title.trim();
	    body = body.trim();
	    if (status == null) 
		status = "LIVE";
	    else
		status = status.trim();

	    if (catId == 0)
	    {
		SortableTable cat = 
		    DiscussionHandler.lib.addCategory(conn, title, body, status);
		if (cat == null)
		{
		    steam_vars.put("message", "There was a problem creating " +
				"your new discussion. Please try again");
		    steam_vars.put("NEW_CAT", "true");
		    steam_vars.put("COMMUNITYID", String.valueOf(catId));
		    DiscussionHandler.insertTemplate(hd, "Admin+MOD");
		    return;
		}
		    
		catId = ((Number) cat.get(0,0)).intValue();

		// alright.  We created the discussion. Now take them to the
		// topics page, so they can add topics.
		// do we want to give them this as an option?
//		url_tokens[3] = String.valueOf(catId);
//		hd.form_hash = null;
//		handleAdminTopics();
//		return;
	    }
	    else
	    {
		DiscussionHandler.lib.modifyCategory(conn, catId, title, 
					     body, status);
	    }
	}


	if (catId != 0)
	{
	    SortableTable s = DiscussionHandler.lib.getCategoryInfo(conn, catId);
	    if (s == null)
	    {
		steam_vars.put("message", "That category does not exist.");
		showAdminDefault();
		return;
	    }
	    else
	    {
		// put in the values of the first row of the table with
		// the column names as their keys
		for (int i = 0; i < 4; i++)
		{
		    Object o = s.get(0, i);
		    log (0, "COMMUNITY" + s.getColumnName(i) + " : " + o);
		    if (o != null)
		    {
			if (s.getColumnName(i).equals("DESCRIPTION"))
			    steam_vars.put("COMMUNITYDESC", o);
			else
			    steam_vars.put("COMMUNITY" + s.getColumnName(i), o);
		    }
		}
	    }
	}
	else
	{
	    steam_vars.put("NEW_CAT", "true");
	}

	steam_vars.put("COMMUNITYID", String.valueOf(catId));
	DiscussionHandler.insertTemplate(hd, "Admin+MOD");
    }

    public void modifyDiscUsers(int catId)
    {
	// See if this is a valid discussion...
	if (catId < 1)
	{
	    steam_vars.put("message", "That category does not exist.");
	    showAdminDefault();
	    return;
	}

	if (catId > TOPIC_TO_ACL_FILTER)
	{
	    steam_vars.put("message", "Cannot perform access control on this " +
			"category.");
	    log (0, "Too high an id for category: Cannot surpass " +
		    "DiscussionAdminHandler.TOPIC_TO_ACL_FILTER = " +
		    TOPIC_TO_ACL_FILTER);	    
	    showAdminDefault();
	    return;
	}

	// Get the discussion (category) info for general substitutions
	SortableTable s = DiscussionHandler.lib.getCategoryInfo(conn, catId);
	if (s == null)
	{
	    steam_vars.put("message", "That category does not exist.");
	    showAdminDefault();
	    return;
	}
	else
	{
	    // put in the values of the first row of the table with
	    // the column names as their keys
	    for (int i = 0; i < 4; i++)
	    {
		Object o = s.get(0, i);
		log (0, "COMMUNITY" + s.getColumnName(i) + " : " + o);
		if (o != null)
		{
		    if (s.getColumnName(i).equals("DESCRIPTION"))
			steam_vars.put("COMMUNITYDESC", o);
		    else
			steam_vars.put("COMMUNITY" + s.getColumnName(i), o);
		}
	    }

	    if (((String)steam_vars.get("COMMUNITYNAME")).equals("(No name)"))
	    {
		// Not a valid discussion...
		steam_vars.put("message", "That category does not exist.");
		showAdminDefault();
		return;
	    }
	}
   
	steam_vars.put("COMMUNITYID", String.valueOf(catId));

	// Fill in the access control tables
	createAccessControlTables(catId);

	// Display the page
	DiscussionHandler.insertTemplate(hd, "Admin+USER");
    }
    

/*--------------- TOPICS ---------------------*/


    /** Handles all topic administration. This includes: 1. display the page
     * 2. handle a delete topics post  3. handle an add topic post
     * 4. handle a modify topic and now 5. handle user access control for a
     * topic
     */
    public void handleAdminTopics()
    {
	// lets do (4) first
	if (url_tokens[3].equals("MOD"))
	{
	    handleModifyTopic();
	    return;
	}

	if (url_tokens[3].equals("USER"))
	{
	    handleModifyTopicUsers();
	    return;
	}

	int catId;
	try
	{
	    catId = Integer.parseInt(url_tokens[3]);
	}
	catch (NumberFormatException e)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed community id: " + url_tokens[3]);
	    return;
	}

	// here's the default page
	if (hd.form_hash == null)
	{
	    displayTopicList(catId);
	}
	else if (hd.form_hash.containsKey("Delete"))
	{
	    deleteTopics(catId);
	}
	else if (hd.form_hash.containsKey("Add"))
	{
	    addTopic(catId);
	} 
	else
	{
	    showAdminDefault();
	}
    }


    /** displays the generic topic page*/
    private void displayTopicList(int catId)
    {
	SortableTable community = 
	    DiscussionHandler.lib.getCategoryInfo(conn, catId);
	SortableTable topics = 
	    DiscussionHandler.lib.getTopicList(conn, catId, "LIVE");

	if (community == null)
	{
	    sendErrorMsg("ERROR: category " + catId + " had no info");
	    return;
	}
/*	if (community == null || topics == null)
	{
	    sendErrorMsg("ERROR: category " + catId + 
		" had no info or had no topics");
	    return;
	}
	*/
	steam_vars.put("community", community);
	if (topics != null) 
	    steam_vars.put("topics", topics);
	DiscussionHandler.insertTemplate(hd, "Admin+TOP");
    }


    /** add a topic and the first post */
    private void addTopic(int catId)
    {
	String topicTitle =  null;
	String postTitle =   null;
	String postBody =    null;
	String firstname =   null;
	String lastname  =   null;
	String email     =   null;
	    
	try
	{
	    topicTitle = hd.form_hash.get("topic_title").toString();
	    postTitle = hd.form_hash.get("post_title").toString();
	    postBody = hd.form_hash.get("post_body").toString();
	    firstname = hd.form_hash.get("firstname").toString();
	    lastname = hd.form_hash.get("lastname").toString();
	    email = hd.form_hash.get("email").toString();
	}
	catch (NullPointerException ne)
	{
	    sendErrorMsg("Your request to create a new post " +
		" could not be processed due to an internal error");
	    return;
	}

	SortableTable topic = 
	    DiscussionHandler.lib.addTopic(conn, catId, topicTitle);
	int topicId = ((Number) topic.get(0,0)).intValue();
	boolean success = 
	    DiscussionHandler.lib.addPost(conn, topicId, postTitle,
				  postBody, firstname, lastname, email);
	DiscussionHandler.lib.changeTopicStatus(conn, topicId, "LIVE");

	String message = null;
	if (!success)
	{
	    message = "The new topic could not be created.  Please" +
		" try again";
	}
	else
	{
	    message = "The topic " + topicTitle + " has been added.";
	}

	steam_vars.put("message", message);
	displayTopicList(catId);
    }	


    /** deletes a list of topics if their ids are in the form hash */
    private void deleteTopics(int catId)
    {
	Vector v = new Vector(hd.form_hash.size());
	Enumeration e = hd.form_hash.keys();
	for(int i = 0; e.hasMoreElements(); i++)
	{
	    try
	    {
		int topicId = Integer.parseInt(e.nextElement().toString());
		DiscussionHandler.lib.changeTopicStatus(conn, topicId, "DEAD");
		v.addElement(hd.form_hash.get(String.valueOf(topicId)));
	    }
	    catch (NumberFormatException ne) {}
	}

	StringBuffer message = new StringBuffer();
	if (v.size() > 0)
	{
	    message.append("The following topics have been deleted: ");
	    message.append("\n\t<BLOCKQUOTE>\n");
	    for (int i = 0; i < v.size(); i++)
	    {
		message.append("\t\t");
		message.append(v.elementAt(i).toString());
		message.append("<BR>\n");
	    }
	    message.append("\t</BLOCKQUOTE>\n");
	}
	else
	{
	    message.append("No topics were deleted");
	}
		
	steam_vars.put("message", message.toString());
	displayTopicList(catId);
    }


    /** display the modify page, or modify the topic with post info
     */
    private void handleModifyTopic()
    {
	if (url_tokens.length < 5)
	{
	    // no category id- show the default page
	    sendErrorMsg("ERROR: Communities Admin: No topic id: ");
	    return;
	}

	int topicId = 0;
	try
	{
	    topicId = Integer.parseInt(url_tokens[4]);
	}
	catch (NumberFormatException ne)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed topic id: " + url_tokens[4]);
	    return;
	}
	catch (ArrayIndexOutOfBoundsException ae)
	{
	    sendErrorMsg("ERROR: URL is not valid");
	    return;
	}
		
	// phew.  Through the errors- onwards
	if (hd.form_hash == null)
	{
	    // just display the mod page
	    SortableTable topic = 
		DiscussionHandler.lib.getTopicInfo(conn, topicId);
	    
	    int catId = ((Number) topic.get(0,1)).intValue();
	    SortableTable community = 
		DiscussionHandler.lib.getCategoryInfo(conn, catId);
	    
	    steam_vars.put("topic", topic);
	    steam_vars.put("community", community);
	    
	    DiscussionHandler.insertTemplate(hd, "Admin+TOP+MOD");
	    return;
	}
	else   // we have a POST- get the info and update the topic
	{
	    String title = hd.form_hash.get("topic_title").toString();
	    String old_title = 
		hd.form_hash.get("old_topic_title").toString();

	    if (title == null || old_title == null)
	    {
		sendErrorMsg("ERROR: Community: modifyPost: no title or " +
			     "old_title in form");
		return;
	    }

	    // get the community id
	    int catId;
	    Object cat = hd.form_hash.get("community_id");
	    if (cat == null)
	    {
		sendErrorMsg( "ERROR: Community: modifyPost: "+ 
			      "no community id in form");
		return;
	    }
	    else
	    {
		try
		{	
		    catId = Integer.parseInt(cat.toString());
		}
		catch (NumberFormatException ne)
		{
		    sendErrorMsg( "ERROR: Community: modifyPost: " +
				  "malformed community id");
		    return;
		}
	    }		    
	    
	    String message = null;
 	    if (title != null && title.length() > 0 && 
		!title.equals(old_title))
	    {
		DiscussionHandler.lib.modifyTopic(conn, topicId, 
					  title, "LIVE");
		message = "Topic with title '" + old_title +
		    "' has been changed to '" + title + "'.";
	    }
	    else
	    {
		message = "The title you entered was either blank or " +
		    "identical to the present title.  Please try again.";
	    }
	    steam_vars.put("message", message);
	    displayTopicList(catId);
	    return;
	}
    }   // end of modifying topics...

    private void handleModifyTopicUsers()
    {
	if (url_tokens.length < 5)
	{
	    // no category id- show the default page
	    sendErrorMsg("ERROR: Communities Admin: No topic id: ");
	    return;
	}

	int topicId = 0;
	try
	{
	    topicId = Integer.parseInt(url_tokens[4]);
	}
	catch (NumberFormatException ne)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed topic id: " + url_tokens[4]);
	    return;
	}
	catch (ArrayIndexOutOfBoundsException ae)
	{
	    sendErrorMsg("ERROR: URL is not valid");
	    return;
	}

	// Get the discussion (category) info for general substitutions
	SortableTable topic = DiscussionHandler.lib.getTopicInfo(conn, topicId);
	if (topic == null)
	{
	    // Display an error
	    sendErrorMsg(
		"ERROR: Communities: Admin: modifyTopicUsers Handler: " +
		"Topic Id not valid");
	    return;
	}
	else
	{
	    // Make the general substitutions from the topic info
	    int catId = ((Number) topic.get(0,1)).intValue();
	    SortableTable community = 
		DiscussionHandler.lib.getCategoryInfo(conn, catId);
	    
	    steam_vars.put("topic", topic);
	    steam_vars.put("community", community);
	}

	// Get the database equivalent of this topic id
	topicId += TOPIC_TO_ACL_FILTER;

	// Fill in the access control tables
	createAccessControlTables(topicId);

	// Display the page
	DiscussionHandler.insertTemplate(hd, "Admin+TOP+USER");	
    }


/*--------------- POSTS ----------------*/


    /** Handles all post administration.  This inlcudes: 
     * 1. display the post summary page
     * 2. handle a delete post  3. handle an add post
     * 4. handle modifying a post
     */
    public void handleAdminPosts()
    {
	// let's do (4) first
	if (url_tokens[3].equals("MOD"))
	{
	    handleModifyPost();
	    return;
	}

	// get the topic id
	int topicId;
	try
	{
	    topicId = Integer.parseInt(url_tokens[3]);
	}
	catch (NumberFormatException e)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed community id: " + url_tokens[3]);
	    return;
	}

	// here's the default page
	if (hd.form_hash == null)
	{
	    displayPostList(topicId);
	}
	else if (hd.form_hash.containsKey("Delete"))
	{
	    deletePosts(topicId);
	}
	else if (hd.form_hash.containsKey("Add"))
	{
	    addPost(topicId);
	} // end of add
	else
	{
	    showAdminDefault();
	}
    }

    /** displays the generic topic page*/
    private void displayPostList(int topicId)
    {
	SortableTable topic = 
	    DiscussionHandler.lib.getTopicInfo(conn, topicId);

	SortableTable posts = 
	    DiscussionHandler.lib.getPostSummaries(conn, topicId, 
					   DiscussionLib.TITLE);

	SortableTable community = null;

	if (topic != null)
	{
	    int catId = ((Number) topic.get(0, 1)).intValue();
	    community = DiscussionHandler.lib.getCategoryInfo(conn, catId);
	}

	if (topic == null || community == null) // || posts == null)
	{
	    sendErrorMsg("ERROR: topic " + topicId + 
			 " had no info or had no posts: null topic? " + 
			 (topic == null) + " null community? " + 
			 (community == null) + " null posts? " + 
			 (posts == null));
	    return;
	}

	if (posts != null) steam_vars.put("posts", posts);
	steam_vars.put("topic", topic);
	steam_vars.put("community", community);
	DiscussionHandler.insertTemplate(hd, "Admin+POST");
    }

    
    /** delete the posts whose id are in the form hash */
    private void deletePosts(int topicId)
    {
	Vector v = new Vector(hd.form_hash.size());
	Enumeration e = hd.form_hash.keys();
	for(int i = 0; e.hasMoreElements(); i++)
	{
	    try
	    {
		int postId = Integer.parseInt(e.nextElement().toString());
		DiscussionHandler.lib.changePostStatus(conn, postId, "DEAD");
		v.addElement(hd.form_hash.get(String.valueOf(postId)));
	    }
	    catch (NumberFormatException ne) {}
	}

	StringBuffer message = new StringBuffer();
	if (v.size() > 0)
	{
	    message.append("The following posts have been deleted: ");
	    message.append("\n\t<BLOCKQUOTE>\n");
	    for (int i = 0; i < v.size(); i++)
	    {
		message.append("\t\t");
		message.append(v.elementAt(i).toString());
		message.append("<BR>\n");
	    }
	    message.append("\t</BLOCKQUOTE>\n");
	}
	else
	{
	    message.append("No posts were deleted");
	}
		
	steam_vars.put("message", message.toString());
	displayPostList(topicId);
    }


    /** add a post to the topic whose id is the parameter */
    private void addPost(int topicId)
    {

	String postTitle =   null;
	String postBody  =   null;
	String firstname =   null;
	String lastname  =   null;
	String email     =   null;
	    
	try
	{
	    postTitle = hd.form_hash.get("post_title").toString();
	    postBody = hd.form_hash.get("post_body").toString();
	    firstname = hd.form_hash.get("firstname").toString();
	    lastname = hd.form_hash.get("lastname").toString();
	    email = hd.form_hash.get("email").toString();
	}
	catch (NullPointerException ne)
	{
	    sendErrorMsg("Your request to create a new post could not " +
			 "be processed due to an internal error");
	    return;
	}

	boolean success = 
	    DiscussionHandler.lib.addPost(conn, topicId, postTitle,
				  postBody, firstname, lastname, email);

	String message = null;
	if (!success)
	{
	    message = "The newpost could not be created.  Please" +
		" try again";
	}
	else
	{
	    message = "The post " + postTitle + " has been added.";
	}

	steam_vars.put("message", message);
	displayPostList(topicId);
    }
    

    /** display the modify page, or modify the post if we have a post */
    private void handleModifyPost()
    {
	if (url_tokens.length < 5)
	{
	    // no category id- show the default page
	    sendErrorMsg("ERROR: Communities Admin: " +
		"No post id: ");
	    return;
	}
	
	int postId = 0;
	try
	{
	    postId = Integer.parseInt(url_tokens[4]);
	}
	catch (NumberFormatException ne)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		    "Malformed topic id: " + url_tokens[4]);
	    return;
	}
	catch (ArrayIndexOutOfBoundsException ae)
	{
	    sendErrorMsg(
		"ERROR: Communities: AdminPostHandler: URL is not valid");
	    return;
	}
		
	// phew.  Through the errors- onwards
	if (hd.form_hash == null)
	{
	    // just display the mod page
	    SortableTable post = DiscussionHandler.lib.getPost(conn, postId);

	    SortableTable topic = null;
	    SortableTable community = null;

	    if (post != null)
	    {
		int topicId = ((Number) post.get(0, 7)).intValue();
		topic = DiscussionHandler.lib.getTopicInfo(conn, topicId);

		if (topic != null)
		{
		    int catId = ((Number) topic.get(0, 1)).intValue();
		    community = DiscussionHandler.lib.getCategoryInfo(conn, catId);
		}
	    }

	    if (topic == null || post == null || community == null)
	    {
		sendErrorMsg("ERROR: post " + postId + 
		    " had no info or had no parent topic or community");
		return;
	    }

	    steam_vars.put("post", post);
	    steam_vars.put("community", community);
	    steam_vars.put("topic", topic);

	    DiscussionHandler.insertTemplate(hd, "Admin+POST+MOD");
	    return;
	}
	else
	{
	    int topicId;
	    Object top = hd.form_hash.get("topic_id");
	    if (top == null)
	    {
		sendErrorMsg( "ERROR: Community: modifyPost: " +
			      "no topic id in form");
		return;
	    }
	    else
	    {
		try
		{	
		    topicId = Integer.parseInt(top.toString());
		}
		catch (NumberFormatException ne)
		{
		    sendErrorMsg( "ERROR: Community: modifyPost: " +
				  "malformed topic id");
		    return;
		}
	    }		    

	    // modify the post
	    String postTitle =   null;
	    String postBody  =   null;
	    String firstname =   null;
	    String lastname  =   null;
	    String email     =   null;
	    
	    try
	    {
		postTitle = hd.form_hash.get("post_title").toString();
		postBody = hd.form_hash.get("post_body").toString();
		firstname = hd.form_hash.get("firstname").toString();
		lastname = hd.form_hash.get("lastname").toString();
		email = hd.form_hash.get("email").toString();
	    }
	    catch (NullPointerException ne)
	    {
		sendErrorMsg("Your request to create a new post " +
		    " could not be processed due to an internal error");
		return;
	    }
	    catch (NumberFormatException ne)
	    {
		sendErrorMsg("Your request to create a new post " +
		    " could not be processed due to an internal error");
		return;
	    }

	    boolean success = 
		DiscussionHandler.lib.modifyPost(
		    conn, postId, topicId, postTitle, postBody, 
		    firstname, lastname, email, "LIVE");
	    
	    String message = null;
	    if (success)
	    {
		message = "The post has been updated";
	    }
	    else
	    {
		message = "The post could not be updated.  Please try again";
	    }
	    steam_vars.put("message", message);
	    displayPostList(topicId);
	    return;
	}
    }   // end of modifying posts...


    /**
     * Used for filling in the select tables for showing users with access
     * and without access to either a discussion or topic, depending on the id.
     */
    private void createAccessControlTables(int id)
    {
	AccessControlList acl = null;

	if (hd.form_hash != null)
	{
	    if(hd.form_hash.containsKey("PUBLIC"))
	    {
		// Create a deny read access acl containing no users:
		acl = new AccessControlList(id);
		acl.addPermission(COMM_READ);
		acl.setPolarity(false, COMM_READ);
	    }
	    else if (hd.form_hash.containsKey("PRIVATE"))
	    {
		// Create an allow read access acl containing no users:
		acl = new AccessControlList(id);
		acl.addPermission(COMM_READ);
		acl.setPolarity(true, COMM_READ);		
	    }
	    else
	    {
		// Get the ACL associated with this id
		try
		{
		    acl = ACLManager.getACL(id, conn);
		}
		catch(Exception e)
		{
		    log (0, "An error occured loading an ACL: " +
			    StringUtil.exceptionStackTrace(e));
		    displayPage ("error/generic_error.tmpl", steam_vars);
		    return;
		}

		// if acl does not exist, create a default one which grants
		// all users read access
		if (acl == null)
		{
		    acl = new AccessControlList(id);
		    acl.addPermission(COMM_READ);
		    acl.setPolarity(false, COMM_READ);		    
		}
		
		if (hd.form_hash.containsKey("ADD"))
		{
		    String added = (String)hd.form_hash.get(
			"WITHOUT_ACCESS_SELECT");
		    if(added != null && added.length() > 0)
		    {
			StringTokenizer toker = new StringTokenizer(added,
								    ",",
								    false);
			if(acl.getPolarity(COMM_READ))
			{
			    while(toker.hasMoreTokens())
			    {
				acl.addUser(toker.nextToken(),
					    COMM_READ);
			    }
			}
			else
			{
			    while(toker.hasMoreTokens())
			    {
				acl.removeUser(toker.nextToken(),
					       COMM_READ);
			    }
			}
		    }
		}
		else if (hd.form_hash.containsKey("REMOVE"))
		{
		    String removed = (String)hd.form_hash.get(
			"WITH_ACCESS_SELECT");
		    if(removed != null && removed.length() > 0)
		    {
			StringTokenizer toker = new StringTokenizer(removed,
								    ",",
								    false);
			if(acl.getPolarity(COMM_READ))
			{
			    while(toker.hasMoreTokens())
			    {
				acl.removeUser(toker.nextToken(),
					       COMM_READ);
			    }
			}
			else
			{
			    while(toker.hasMoreTokens())
			    {
				acl.addUser(toker.nextToken(),
					    COMM_READ);
			    }
			}
		    }
		}
	    }

	    // Update acl in the database
	    try
	    {
		ACLManager.updateACL(acl, conn);
	    }
	    catch(Exception e)
	    {
		log (0, "An error occured updating an ACL: " +
			StringUtil.exceptionStackTrace(e));
		displayPage ("error/generic_error.tmpl", steam_vars);
		return;
	    }
	}
	else
	{
	    // No form hash, so get the acl
	    try
	    {
		acl = ACLManager.getACL(id, conn);
	    }
	    catch(Exception e)
	    {
		log (0, "An error occured loading an ACL: " +
			StringUtil.exceptionStackTrace(e));
		displayPage ("error/generic_error.tmpl", steam_vars);
		return;
	    }
	    
	    // Create a default "grant all users read access" acl
	    // (A deny acl with no users allows all users).
	    if (acl == null)
	    {
		acl = new AccessControlList(id);
		acl.addPermission(COMM_READ);
		acl.setPolarity(false, COMM_READ);
	    }
	}

	// Now fill out the "With Access" and "Without Access" lists for this
	// discussion...

	// First get all the user names in the system
	String[] all_users = null;
	try
	{
	    all_users = User.loadAllUserNames(conn);
	}
	catch (LoadUserException lue)
	{
	    log (0, "An error occured loading users: " +
		    StringUtil.exceptionStackTrace(lue));
	    displayPage ("error/generic_error.tmpl", steam_vars);
	    return;
	}

	// Then create two user vectors: one containing the "With Access"
	// users, and the other containing the "Without Access" users.
	Vector with_users = new Vector();
	Vector without_users = new Vector();
	    
	// Get the list of users associated with the read permission in this
	// acl.
	Vector acl_users = acl.getUsers(COMM_READ);

	if(acl_users != null && !acl_users.isEmpty())
	{
	    if(acl.getPolarity(COMM_READ))
	    {
		// Users in acl have read access
		for(int i=0; i<all_users.length; i++)
		{
		    if (acl_users.contains(all_users[i]))
		    {
			with_users.addElement(all_users[i]);
		    }
		    else
		    {
			without_users.addElement(all_users[i]);
		    }
		}
		// Special anonymous user case:
		if (acl_users.contains(User.ANONYMOUS_NAME))
		{
		    with_users.insertElementAt(User.ANONYMOUS_NAME, 0);
		}
		else
		{
		    without_users.insertElementAt(User.ANONYMOUS_NAME,
						  0);
		}
	    }
	    else
	    {
		// Users in acl do not have read access
		for(int i=0; i<all_users.length; i++)
		{
		    if (acl_users.contains(all_users[i]))
		    {
			without_users.addElement(all_users[i]);
		    }
		    else
		    {
			with_users.addElement(all_users[i]);
		    }
		}
		// Special anonymous user case:
		if (acl_users.contains(User.ANONYMOUS_NAME))
		{
		    without_users.insertElementAt(User.ANONYMOUS_NAME,
						  0);
		}
		else
		{
		    with_users.insertElementAt(User.ANONYMOUS_NAME, 0);
		}
	    }
	}
	else
	{
	    // There are no users for the read permission in this acl.
	    // If the acl for the permission is a deny list, allow all users
	    // access, and vice-vesra.

	    if(!acl.getPolarity(COMM_READ))
	    {
		// Put all users in the "With Access" list
		for(int i=0; i<all_users.length; i++)
		{		
		    with_users.addElement(all_users[i]);
		}
		with_users.insertElementAt(User.ANONYMOUS_NAME, 0);
	    }
	    else
	    {
		// Put all users in the "Without Access" list
		for(int i=0; i<all_users.length; i++)
		{		
		    without_users.addElement(all_users[i]);
		}
		without_users.insertElementAt(User.ANONYMOUS_NAME, 0);
	    }
	}

	// Now put the user names into their SELECT lists as options:
	String WITH_ACCESS = "";
	Enumeration with = with_users.elements();
	while(with.hasMoreElements())
	{
	    WITH_ACCESS = WITH_ACCESS + "<option>" +
		(String)with.nextElement() + "</option>";
	}

	String WITHOUT_ACCESS = "";
	Enumeration without = without_users.elements();
	while(without.hasMoreElements())
	{
	    WITHOUT_ACCESS = WITHOUT_ACCESS + "<option>" +
		(String)without.nextElement() + "</option>";
	}
	
	steam_vars.put("WITH_ACCESS", WITH_ACCESS);
	steam_vars.put("WITHOUT_ACCESS", WITHOUT_ACCESS);
    }
    
	/** display a generic error message. Just a wrapper for whatever
	 * method of error display the site uses.  */
    private void sendErrorMsg(String errmsg)
    {
	log (0,  errmsg);
	Loco.sendErrorMsg(errmsg, hd);
//	(new ErrorHandler(errmsg)).handleRequest(hd);
    }


    /** our templates are loaded by DiscussionHandler for now, 
     * so we don't need to do anything
     */
    public boolean init(Log log)
    {
	return (true);
    }

    /**
     * Nothing to do here either, presently.
     * @see org.locomotive.loco.Handler#shutdown
     */
    public boolean shutdown (Log log)
    {
	return (true);
    }

}
