/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.discuss;

import java.io.*;
import java.util.*;
import java.sql.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.util.*;
import org.locomotive.loco.perm.*;


/** 
 * This class handles the administration for the communities.  It provides
 * screens which allow a user to modify discussion, topics, and posts,
 * as well as add them.  It is desgined to allow quick and easy moderation
 * of discussions; at the possible expense of some performance.  This
 * should be okay because only a few administrators should be seeing these
 * pages.  Permission to view these pages is based on whether or
 * not the user has 
 * org.locomotive.module.discuss.DiscussionAdminHandler.COMM_ADMIN
 * permission, listed below.
 */
public class DiscussionAdminServlet extends LocoServlet
{
    protected DiscussionLib lib;

    /**
     * Allows a user to administer user access control for discussions and
     * topics:
     */
    public static final String COMM_ADMIN = "COMM_ADMIN";

    /**
     * Allows a user read access to a discussion or topic:
     */
    public static final String COMM_READ = "COMM_READ";

    /**
     * Since categories and topics share the same ids since they both start at
     * 1, we must make a distinction between category and topic ids for
     * the acls.  Therefore acl ids for categories are the same as the
     * category ids, while acl ids for topics are the topic ids plus this
     * filter.  Please note that we cannot have more than 1000000 categories!
     */
    public static final int TOPIC_TO_ACL_FILTER = 1000000;


    /**
     * This handles the administration for communities discussions, topics, and
     * posts.  Shows an unknown handler error page if the user does not have
     * the correct admin permission to be here.
     * All templates for communities admin should be in the 
     * 'discuss/admin/commmgr' directory. 
     * The tokens and their corresponding methods are:
     * <BR><PRE>
     *  Token            Method               Template
     *  MOD              modifyDiscussion     comm_disc_filledtempl.tmpl
     *  TOP              handleAdminTopics()  comm_topics_list.tmpl
     *  TOP+MOD          modifyTopic()        comm_topics_filledtempl.tmpl
     *  POST             handleAdminPosts()   comm_posts_list.tmpl
     *  POST+MOD         modifyPost()         comm_posts_filledtempl.tmpl
     *  (default)        handleAdminDefault() comm_disc_list.tmpl
     *  USER             modifyDiscUsers      comm_disc_users.tmpl
     *  TOP+USER         modifyTopicUsers     comm_topic_users.tmpl
     *  
     * </PRE>
     */
    public void service(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, java.io.IOException
    {
	setup(req, resp);

	if(user.isAnonymous())
	{
	    // Not allowed here; pretend they got a unknown handler
	    displayPage ("error/unknown_handler_option.tmpl", steam_vars);
	    return;
	}
	else
	{
	    PermissionSet pSet = null;
	    
	    try
	    {
		pSet = UserPerm.getPermissions(user.getUserid(), conn);
	    }
	    catch(SQLException se) 
	    { 
		handleException(se, "discuss_adm_perm");
		return;
	    }
	    	
	    // See if the user has permission to be here
	    if (!pSet.isSet(COMM_ADMIN))
	    {
		// Is not an administrator.  Show them the unknown handler
		// error page.
		displayPage ("error/unknown_handler_option.tmpl", steam_vars);
		return;
	    }
	}

	if (url_tokens.length < 4)
	{
	    if (req.getMethod().equals("POST"))
	    {
		deleteDiscussions();
	    }

	    showAdminDefault();
	}
	else if (url_tokens[2].equals("MOD"))
	{
	    try
	    {
		int catId = Integer.parseInt(url_tokens[3]);
		modifyDiscussion(catId);
	    }
	    catch (NumberFormatException e)
	    {
		showAdminDefault();
	    }
	}
	else if (url_tokens[2].equals("TOP"))
	{
	    handleAdminTopics();
	}
	else if (url_tokens[2].equals("POST"))
	{
	    handleAdminPosts();
	}
	else if (url_tokens[2].equals("USER"))
	{
	    int catId = Integer.parseInt(url_tokens[3]);
	    modifyDiscUsers(catId);
	}
	else
	{
	    showAdminDefault();
	}
    }



    /** use this to display the communities admin home page.  */
    private void showAdminDefault() throws IOException
    {
	SortableTable categories = lib.getCategoryList(conn, "LIVE");
	if (categories != null)
	    steam_vars.put("CAT_LIST", categories);

	DiscussionServlet.insertTemplate("Admin", resp, steam_vars);
    }


    
    private void deleteDiscussions() throws IOException
    {
	Hashtable form_hash = form_data.getHashtable();
	Vector v = new Vector(form_hash.size());
	Enumeration e = form_hash.keys();
	for(int i = 0; e.hasMoreElements(); i++)
	{
	    try
	    {
		int catId = Integer.parseInt(e.nextElement().toString());
		lib.changeCategoryStatus(conn, catId, "DEAD");
		v.addElement(form_hash.get(String.valueOf(catId)));
	    }
	    catch (NumberFormatException ne) {}
	}

	StringBuffer message = new StringBuffer();
	if (v.size() > 0)
	{
	    message.append("The following categories have been deleted: ");
	    message.append("\n\t<BLOCKQUOTE>\n");
	    for (int i = 0; i < v.size(); i++)
	    {
		message.append("\t\t");
		message.append(v.elementAt(i).toString());
		message.append("<BR>\n");
	    }
	    message.append("\t</BLOCKQUOTE>\n");
	}
	else
	{
	    message.append("No categories were deleted");
	}
		
	steam_vars.put("message", message.toString());
    }


    /** handle modification of a discussion   if no post, display the
     * page with the info.  Otherwise, update the info with what was
     * posted.  If the number is less than 1, then create a new discussion.
     */
    private void modifyDiscussion(int catId) throws IOException
    {
	if (req.getMethod().equals("POST"))
	{
	    String title = form_data.getNonNullField("COMMUNITYNAME", "");
	    String body = form_data.getNonNullField("COMMUNITYDESC", "");
	    String status = form_data.getTrimmedString("COMMUNITYSTATUS",
						       "title", 1, 150);

	    if (form_data.allErrorMsgs().length() > 0)
	    {
		form_data.copyAllIntoSteam(steam_vars);
		steam_vars.put("message", "You must enter a valid title. " +
			    "Please try again");
		steam_vars.put("NEW_CAT", "true");
		steam_vars.put("COMMUNITYID", String.valueOf(catId));

		DiscussionServlet.insertTemplate("Admin+MOD", resp, steam_vars);
		return;
	    }
		
	    title = title.trim();
	    body = body.trim();
	    if (status == null) 
		status = "LIVE";
	    else
		status = status.trim();

	    if (catId == 0)
	    {
		SortableTable cat = lib.addCategory(conn, title, body, status);
		if (cat == null)
		{
		    steam_vars.put("message", "There was a problem creating " +
				"your new discussion. Please try again");
		    steam_vars.put("NEW_CAT", "true");
		    steam_vars.put("COMMUNITYID", String.valueOf(catId));
		    DiscussionServlet.insertTemplate("Admin+MOD", 
						     resp, steam_vars);
		    return;
		}
		    
		catId = ((Number) cat.get(0,0)).intValue();

		// alright.  We created the discussion. Now take them to the
		// topics page, so they can add topics.
		// do we want to give them this as an option?
//		url_tokens[3] = String.valueOf(catId);
//		hd.form_hash = null;
//		handleAdminTopics();
//		return;
	    }
	    else
	    {
		lib.modifyCategory(conn, catId, title, body, status);
	    }
	}


	if (catId != 0)
	{
	    SortableTable s = lib.getCategoryInfo(conn, catId);
	    if (s == null)
	    {
		steam_vars.put("message", "That category does not exist.");
		showAdminDefault();
		return;
	    }
	    else
	    {
		// put in the values of the first row of the table with
		// the column names as their keys
		for (int i = 0; i < 4; i++)
		{
		    Object o = s.get(0, i);
		    if (o != null)
		    {
			if (s.getColumnName(i).equalsIgnoreCase("DESCRIPTION"))
			    steam_vars.put("COMMUNITYDESC", o);
			else
			    steam_vars.put("COMMUNITY" + 
					   s.getColumnName(i).toUpperCase(), o);
		    }
		}
	    }
	}
	else
	{
	    steam_vars.put("NEW_CAT", "true");
	}

	steam_vars.put("COMMUNITYID", String.valueOf(catId));
	DiscussionServlet.insertTemplate("Admin+MOD", resp, steam_vars);
    }

    public void modifyDiscUsers(int catId) throws IOException
    {
	// See if this is a valid discussion...
	if (catId < 1)
	{
	    steam_vars.put("message", "That category does not exist.");
	    showAdminDefault();
	    return;
	}

	if (catId > TOPIC_TO_ACL_FILTER)
	{
	    steam_vars.put("message", "Cannot perform access control on this " +
			"category.");
	    log (0, "Too high an id for category: Cannot surpass " +
		    "DiscussionAdminHandler.TOPIC_TO_ACL_FILTER = " +
		    TOPIC_TO_ACL_FILTER);	    
	    showAdminDefault();
	    return;
	}

	// Get the discussion (category) info for general substitutions
	SortableTable s = lib.getCategoryInfo(conn, catId);
	if (s == null)
	{
	    steam_vars.put("message", "That category does not exist.");
	    showAdminDefault();
	    return;
	}
	else
	{
	    // put in the values of the first row of the table with
	    // the column names as their keys
	    for (int i = 0; i < 4; i++)
	    {
		Object o = s.get(0, i);
		if (o != null)
		{
		    if (s.getColumnName(i).equalsIgnoreCase("DESCRIPTION"))
			steam_vars.put("COMMUNITYDESC", o);
		    else
			steam_vars.put("COMMUNITY" + s.getColumnName(i).toUpperCase(), o);
		}
	    }

	    if (((String)steam_vars.get("COMMUNITYNAME")).equals("(No name)"))
	    {
		// Not a valid discussion...
		steam_vars.put("message", "That category does not exist.");
		showAdminDefault();
		return;
	    }
	}
   
	steam_vars.put("COMMUNITYID", String.valueOf(catId));

	// Fill in the access control tables
	createAccessControlTables(catId);

	// Display the page
	DiscussionServlet.insertTemplate("Admin+USER", resp, steam_vars);
    }
    

/*--------------- TOPICS ---------------------*/


    /** Handles all topic administration. This includes: 1. display the page
     * 2. handle a delete topics post  3. handle an add topic post
     * 4. handle a modify topic and now 5. handle user access control for a
     * topic
     */
    public void handleAdminTopics() throws IOException
    {
	// lets do (4) first
	if (url_tokens[3].equals("MOD"))
	{
	    handleModifyTopic();
	    return;
	}

	if (url_tokens[3].equals("USER"))
	{
	    handleModifyTopicUsers();
	    return;
	}

	int catId;
	try
	{
	    catId = Integer.parseInt(url_tokens[3]);
	}
	catch (NumberFormatException e)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed community id: " + url_tokens[3]);
	    return;
	}

	// here's the default page
	if (req.getMethod().equals("GET"))
	{
	    displayTopicList(catId);
	}
	else if (form_data.getField("Delete") != null)
	{
	    deleteTopics(catId);
	}
	else if (form_data.getField("Add") != null)
	{
	    addTopic(catId);
	} 
	else
	{
	    showAdminDefault();
	}
    }


    /** displays the generic topic page*/
    private void displayTopicList(int catId) throws IOException
    {
	SortableTable community = lib.getCategoryInfo(conn, catId);
	SortableTable topics = lib.getTopicList(conn, catId, "LIVE");

	if (community == null)
	{
	    sendErrorMsg("ERROR: category " + catId + " had no info");
	    return;
	}
/*	if (community == null || topics == null)
	{
	    sendErrorMsg("ERROR: category " + catId + 
		" had no info or had no topics");
	    return;
	}
	*/
	steam_vars.put("community", community);
	if (topics != null) 
	    steam_vars.put("topics", topics);
	DiscussionServlet.insertTemplate("Admin+TOP", resp, steam_vars);
    }


    /** add a topic and the first post */
    private void addTopic(int catId) throws IOException
    {
	String topicTitle =  form_data.getNonNullField("topic_title", "");
	String postTitle =   form_data.getNonNullField("post_title", "");
	String postBody =    form_data.getNonNullField("post_body", "");
	String firstname =   form_data.getNonNullField("firstname", "");
	String lastname  =   form_data.getNonNullField("lastname", "");
	String email     =   form_data.getNonNullField("email", "");
	    
	if (form_data.allErrorMsgs().length() > 0)
	{
	    sendErrorMsg("Your request to create a new post " +
		" could not be processed due to an internal error: " +
			 "form fields missing");
	    return;
	}

	SortableTable topic = lib.addTopic(conn, catId, topicTitle);
	int topicId = ((Number) topic.get(0,0)).intValue();
	boolean success = lib.addPost(conn, topicId, postTitle,
				      postBody, firstname, lastname, email);
	lib.changeTopicStatus(conn, topicId, "LIVE");

	String message = null;
	if (!success)
	{
	    message = "The new topic could not be created.  Please" +
		" try again";
	}
	else
	{
	    message = "The topic " + topicTitle + " has been added.";
	}

	steam_vars.put("message", message);
	displayTopicList(catId);
    }	


    /** deletes a list of topics if their ids are in the form hash */
    private void deleteTopics(int catId) throws IOException
    {
	Hashtable form_hash = form_data.getHashtable();
	Vector v = new Vector(form_hash.size());
	Enumeration e = form_hash.keys();
	for(int i = 0; e.hasMoreElements(); i++)
	{
	    try
	    {
		int topicId = Integer.parseInt(e.nextElement().toString());
		lib.changeTopicStatus(conn, topicId, "DEAD");
		String [] value = 
		    (String[]) form_hash.get(String.valueOf(topicId));
		v.addElement(value [0]);
	    }
	    catch (NumberFormatException ne) {}
	}

	StringBuffer message = new StringBuffer();
	int size = v.size();
	if (size > 0)
	{
	    message.append("The following topics have been deleted: ");
	    message.append("\n\t<BLOCKQUOTE>\n");
	    for (int i = size - 1; i >= 0; i--)
	    {
		message.append("\t\t");
		message.append(v.elementAt(i).toString());
		message.append("<BR>\n");
	    }
	    message.append("\t</BLOCKQUOTE>\n");
	}
	else
	{
	    message.append("No topics were deleted");
	}
		
	steam_vars.put("message", message.toString());
	displayTopicList(catId);
    }


    /** display the modify page, or modify the topic with post info
     */
    private void handleModifyTopic() throws IOException
    {
	if (url_tokens.length < 5)
	{
	    // no category id- show the default page
	    sendErrorMsg("ERROR: Communities Admin: No topic id: ");
	    return;
	}

	int topicId = 0;
	try
	{
	    topicId = Integer.parseInt(url_tokens[4]);
	}
	catch (NumberFormatException ne)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed topic id: " + url_tokens[4]);
	    return;
	}
	catch (ArrayIndexOutOfBoundsException ae)
	{
	    sendErrorMsg("ERROR: URL is not valid");
	    return;
	}
		
	// phew.  Through the errors- onwards
	if (req.getMethod().equals("GET"))
	{
	    // just display the mod page
	    SortableTable topic = lib.getTopicInfo(conn, topicId);
	    
	    int catId = ((Number) topic.get(0,1)).intValue();
	    SortableTable community = lib.getCategoryInfo(conn, catId);
	    
	    steam_vars.put("topic", topic);
	    steam_vars.put("community", community);
	    
	    DiscussionServlet.insertTemplate("Admin+TOP+MOD", resp, steam_vars);
	    return;
	}
	else   // we have a POST- get the info and update the topic
	{
	    if (! form_data.containsAllNonEmpty( new String [] { 
		"community_id,", "old_topic_title", "topic_title" }))
	    {
		sendErrorMsg("ERROR: Community: modifyPost: " + 
			     form_data.allErrorMsgs());
		return;
	    }


	    // get the community id
	    int catId = form_data.getInt("community_id", "community id");

	    if (form_data.hasAnyErrors())
	    {
		sendErrorMsg( "ERROR: Community: modifyPost: " +
			      "malformed community id");
		return;
	    }
	    
	    String old_title = form_data.getField("old_topic_title");
	    String title = 
		form_data.getTrimmedString("topic_title", "", 1, 150);
	    String message = null;

 	    if (form_data.hasAnyErrors() || title.equals(old_title))
	    {
		message = "The title you entered was either blank or " +
		    "identical to the present title.  Please try again.";
	    }
	    else
	    {
		lib.modifyTopic(conn, topicId, title, "LIVE");
		message = "Topic with title '" + old_title +
		    "' has been changed to '" + title + "'.";
	    }

	    steam_vars.put("message", message);
	    displayTopicList(catId);
	    return;
	}
    }   // end of modifying topics...


    private void handleModifyTopicUsers() throws IOException
    {
	if (url_tokens.length < 5)
	{
	    // no category id- show the default page
	    sendErrorMsg("ERROR: Communities Admin: No topic id: ");
	    return;
	}

	int topicId = 0;
	try
	{
	    topicId = Integer.parseInt(url_tokens[4]);
	}
	catch (NumberFormatException ne)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed topic id: " + url_tokens[4]);
	    return;
	}
	catch (ArrayIndexOutOfBoundsException ae)
	{
	    sendErrorMsg("ERROR: URL is not valid");
	    return;
	}

	// Get the discussion (category) info for general substitutions
	SortableTable topic = lib.getTopicInfo(conn, topicId);
	if (topic == null)
	{
	    // Display an error
	    sendErrorMsg(
		"ERROR: Communities: Admin: modifyTopicUsers Handler: " +
		"Topic Id not valid");
	    return;
	}
	else
	{
	    // Make the general substitutions from the topic info
	    int catId = ((Number) topic.get(0,1)).intValue();
	    SortableTable community = lib.getCategoryInfo(conn, catId);
	    
	    steam_vars.put("topic", topic);
	    steam_vars.put("community", community);
	}

	// Get the database equivalent of this topic id
	topicId += TOPIC_TO_ACL_FILTER;

	// Fill in the access control tables
	createAccessControlTables(topicId);

	// Display the page
	DiscussionServlet.insertTemplate("Admin+TOP+USER", resp, steam_vars);	
    }


/*--------------- POSTS ----------------*/


    /** Handles all post administration.  This inlcudes: 
     * 1. display the post summary page
     * 2. handle a delete post  3. handle an add post
     * 4. handle modifying a post
     */
    public void handleAdminPosts() throws IOException
    {
	// let's do (4) first
	if (url_tokens[3].equals("MOD"))
	{
	    handleModifyPost();
	    return;
	}

	// get the topic id
	int topicId;
	try
	{
	    topicId = Integer.parseInt(url_tokens[3]);
	}
	catch (NumberFormatException e)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		"Malformed community id: " + url_tokens[3]);
	    return;
	}

	// here's the default page
	if (req.getMethod().equals("GET"))
	{
	    displayPostList(topicId);
	}
	else if (form_data.getField("Delete") != null)
	{
	    deletePosts(topicId);
	}
	else if (form_data.getField("Add") != null)
	{
	    addPost(topicId);
	} // end of add
	else
	{
	    showAdminDefault();
	}
    }

    /** displays the generic topic page*/
    private void displayPostList(int topicId) throws IOException
    {
	SortableTable topic = lib.getTopicInfo(conn, topicId);

	SortableTable posts = lib.getPostSummaries(conn, topicId, 
						   DiscussionLib.TITLE);

	SortableTable community = null;

	if (topic != null)
	{
	    int catId = ((Number) topic.get(0, 1)).intValue();
	    community = lib.getCategoryInfo(conn, catId);
	}

	if (topic == null || community == null) // || posts == null)
	{
	    sendErrorMsg("ERROR: topic " + topicId + 
			 " had no info or had no posts: null topic? " + 
			 (topic == null) + " null community? " + 
			 (community == null) + " null posts? " + 
			 (posts == null));
	    return;
	}

	if (posts != null) steam_vars.put("posts", posts);
	steam_vars.put("topic", topic);
	steam_vars.put("community", community);
	DiscussionServlet.insertTemplate("Admin+POST", resp, steam_vars);
    }

    
    /** delete the posts whose id are in the form hash */
    private void deletePosts(int topicId) throws IOException
    {
	Hashtable form_hash = form_data.getHashtable();
	Vector v = new Vector(form_hash.size());
	Enumeration e = form_hash.keys();
	for(int i = 0; e.hasMoreElements(); i++)
	{
	    try
	    {
		int postId = Integer.parseInt(e.nextElement().toString());
		lib.changePostStatus(conn, postId, "DEAD");
		v.addElement(form_hash.get(String.valueOf(postId)));
	    }
	    catch (NumberFormatException ne) {}
	}

	StringBuffer message = new StringBuffer();
	if (v.size() > 0)
	{
	    message.append("The following posts have been deleted: ");
	    message.append("\n\t<BLOCKQUOTE>\n");
	    for (int i = 0; i < v.size(); i++)
	    {
		message.append("\t\t");
		message.append(v.elementAt(i).toString());
		message.append("<BR>\n");
	    }
	    message.append("\t</BLOCKQUOTE>\n");
	}
	else
	{
	    message.append("No posts were deleted");
	}
		
	steam_vars.put("message", message.toString());
	displayPostList(topicId);
    }


    /** add a post to the topic whose id is the parameter */
    private void addPost(int topicId) throws IOException
    {

	String postTitle =   form_data.getNonNullField("post_title", 
						       "post title");
	String postBody  =   form_data.getNonNullField("post_body", 
						       "post body");
	String firstname =   form_data.getNonNullField("firstname", 
						       "first name");
	String lastname  =   form_data.getNonNullField("lastname", 
						       "last name");
	String email     =   form_data.getNonNullField("email", 
						       "email address");
	    
	if (form_data.allErrorMsgs().length() > 0)
	{
	    sendErrorMsg("Your request to create a new post could not " +
			 "be processed due to an internal error: " +
		form_data.allErrorMsgs());
	    return;
	}

	boolean success = lib.addPost(conn, topicId, postTitle,
				      postBody, firstname, lastname, email);

	String message = null;
	if (!success)
	{
	    message = "The newpost could not be created.  Please" +
		" try again";
	}
	else
	{
	    message = "The post " + postTitle + " has been added.";
	}

	steam_vars.put("message", message);
	displayPostList(topicId);
    }
    

    /** display the modify page, or modify the post if we have a post */
    private void handleModifyPost() throws IOException
    {
	if (url_tokens.length < 5)
	{
	    // no category id- show the default page
	    sendErrorMsg("ERROR: Communities Admin: " +
		"No post id: ");
	    return;
	}
	
	int postId = 0;
	try
	{
	    postId = Integer.parseInt(url_tokens[4]);
	}
	catch (NumberFormatException ne)
	{
	    sendErrorMsg("ERROR: Communities Admin: " +
		    "Malformed topic id: " + url_tokens[4]);
	    return;
	}
	catch (ArrayIndexOutOfBoundsException ae)
	{
	    sendErrorMsg(
		"ERROR: Communities: AdminPostHandler: URL is not valid");
	    return;
	}
		
	// phew.  Through the errors- onwards
	if (req.getMethod().equals("GET"))
	{
	    // just display the mod page
	    SortableTable post = lib.getPost(conn, postId);

	    SortableTable topic = null;
	    SortableTable community = null;

	    if (post != null)
	    {
		int topicId = ((Number) post.get(0, 7)).intValue();
		topic = lib.getTopicInfo(conn, topicId);

		if (topic != null)
		{
		    int catId = ((Number) topic.get(0, 1)).intValue();
		    community = lib.getCategoryInfo(conn, catId);
		}
	    }

	    if (topic == null || post == null || community == null)
	    {
		sendErrorMsg("ERROR: post " + postId + 
		    " had no info or had no parent topic or community");
		return;
	    }

	    steam_vars.put("post", post);
	    steam_vars.put("community", community);
	    steam_vars.put("topic", topic);

	    DiscussionServlet.insertTemplate("Admin+POST+MOD", resp, 
					     steam_vars);
	    return;
	}
	else
	{
	    // modify the post
	    int topicId = form_data.getInt("topic_id", "topic id");
	    String postTitle =   form_data.getNonNullField("post_title", 
							   "post title");
	    String postBody  =   form_data.getNonNullField("post_body", 
							   "post body");
	    String firstname =   form_data.getNonNullField("firstname", 
							   "first name");
	    String lastname  =   form_data.getNonNullField("lastname", 
							   "last name");
	    String email     =   form_data.getNonNullField("email", 
							   "email address");

	    if (form_data.allErrorMsgs().length() > 0)
	    {
		sendErrorMsg("Your request to create a new post could not " +
			     "be processed due to an internal error: " +
			     form_data.allErrorMsgs());
		return;
	    }

	    boolean success = 
		lib.modifyPost( conn, postId, topicId, postTitle, postBody, 
				firstname, lastname, email, "LIVE");
	    
	    String message = null;
	    if (success)
	    {
		message = "The post has been updated";
	    }
	    else
	    {
		message = "The post could not be updated.  Please try again";
	    }
	    steam_vars.put("message", message);
	    displayPostList(topicId);
	    return;
	}
    }   // end of modifying posts...


    /**
     * Used for filling in the select tables for showing users with access
     * and without access to either a discussion or topic, depending on the id.
     */
    private void createAccessControlTables(int id) throws IOException
    {
	AccessControlList acl = null;

	if (req.getMethod().equals("POST"))
	{
	    if(form_data.getField("PUBLIC") != null)
	    {
		// Create a deny read access acl containing no users:
		acl = new AccessControlList(id);
		acl.addPermission(COMM_READ);
		acl.setPolarity(false, COMM_READ);
	    }
	    else if (form_data.getField("PRIVATE") != null)
	    {
		// Create an allow read access acl containing no users:
		acl = new AccessControlList(id);
		acl.addPermission(COMM_READ);
		acl.setPolarity(true, COMM_READ);		
	    }
	    else
	    {
		// Get the ACL associated with this id
		try
		{
		    acl = ACLManager.getACL(id, conn);
		}
		catch(Exception e)
		{
		    log (0, "An error occured loading an ACL: " +
			    StringUtil.exceptionStackTrace(e));
		    displayPage ("error/generic_error.tmpl", steam_vars);
		    return;
		}

		// if acl does not exist, create a default one which grants
		// all users read access
		if (acl == null)
		{
		    acl = new AccessControlList(id);
		    acl.addPermission(COMM_READ);
		    acl.setPolarity(false, COMM_READ);		    
		}
		
		if (form_data.getField("ADD") != null)
		{
		    String added = form_data.getTrimmedString(
			"WITHOUT_ACCESS_SELECT", "", 1, 1000000);
		    if(! form_data.hasAnyErrors())
		    {
			StringTokenizer toker = 
			    new StringTokenizer(added, ",", false);
			if(acl.getPolarity(COMM_READ))
			{
			    while(toker.hasMoreTokens())
			    {
				acl.addUser(toker.nextToken(),
					    COMM_READ);
			    }
			}
			else
			{
			    while(toker.hasMoreTokens())
			    {
				acl.removeUser(toker.nextToken(),
					       COMM_READ);
			    }
			}
		    }
		}
		else if (form_data.getField("REMOVE") != null)
		{
		    String removed = form_data.getTrimmedString(
			"WITH_ACCESS_SELECT", "", 0, 1000000);
		    if(! form_data.hasAnyErrors())
		    {
			StringTokenizer toker = 
			    new StringTokenizer(removed, ",", false);
			if(acl.getPolarity(COMM_READ))
			{
			    while(toker.hasMoreTokens())
			    {
				acl.removeUser(toker.nextToken(),
					       COMM_READ);
			    }
			}
			else
			{
			    while(toker.hasMoreTokens())
			    {
				acl.addUser(toker.nextToken(),
					    COMM_READ);
			    }
			}
		    }
		}
	    }

	    // Update acl in the database
	    try
	    {
		ACLManager.updateACL(acl, conn);
	    }
	    catch(Exception e)
	    {
		handleException(e, "discuss_update_acl");
		return;
	    }
	}
	else
	{
	    // No form hash, so get the acl
	    try
	    {
		acl = ACLManager.getACL(id, conn);
	    }
	    catch(Exception e)
	    {
		handleException(e, "discuss_load_acl");
		return;
	    }
	    
	    // Create a default "grant all users read access" acl
	    // (A deny acl with no users allows all users).
	    if (acl == null)
	    {
		acl = new AccessControlList(id);
		acl.addPermission(COMM_READ);
		acl.setPolarity(false, COMM_READ);
	    }
	}

	// Now fill out the "With Access" and "Without Access" lists for this
	// discussion...

	// First get all the user names in the system
	String[] all_users = null;
	try
	{
	    all_users = User.loadAllUserNames(conn);
	}
	catch (LoadUserException lue)
	{
	    handleException(lue, "discuss_acl_load_users");
	    return;
	}

	// Then create two user vectors: one containing the "With Access"
	// users, and the other containing the "Without Access" users.
	Vector with_users = new Vector();
	Vector without_users = new Vector();
	    
	// Get the list of users associated with the read permission in this
	// acl.
	Vector acl_users = acl.getUsers(COMM_READ);

	if(acl_users != null && !acl_users.isEmpty())
	{
	    if(acl.getPolarity(COMM_READ))
	    {
		// Users in acl have read access
		for(int i=0; i<all_users.length; i++)
		{
		    if (acl_users.contains(all_users[i]))
		    {
			with_users.addElement(all_users[i]);
		    }
		    else
		    {
			without_users.addElement(all_users[i]);
		    }
		}
		// Special anonymous user case:
		if (acl_users.contains(User.ANONYMOUS_NAME))
		{
		    with_users.insertElementAt(User.ANONYMOUS_NAME, 0);
		}
		else
		{
		    without_users.insertElementAt(User.ANONYMOUS_NAME,
						  0);
		}
	    }
	    else
	    {
		// Users in acl do not have read access
		for(int i=0; i<all_users.length; i++)
		{
		    if (acl_users.contains(all_users[i]))
		    {
			without_users.addElement(all_users[i]);
		    }
		    else
		    {
			with_users.addElement(all_users[i]);
		    }
		}
		// Special anonymous user case:
		if (acl_users.contains(User.ANONYMOUS_NAME))
		{
		    without_users.insertElementAt(User.ANONYMOUS_NAME,
						  0);
		}
		else
		{
		    with_users.insertElementAt(User.ANONYMOUS_NAME, 0);
		}
	    }
	}
	else
	{
	    // There are no users for the read permission in this acl.
	    // If the acl for the permission is a deny list, allow all users
	    // access, and vice-vesra.

	    if(!acl.getPolarity(COMM_READ))
	    {
		// Put all users in the "With Access" list
		for(int i=0; i<all_users.length; i++)
		{		
		    with_users.addElement(all_users[i]);
		}
		with_users.insertElementAt(User.ANONYMOUS_NAME, 0);
	    }
	    else
	    {
		// Put all users in the "Without Access" list
		for(int i=0; i<all_users.length; i++)
		{		
		    without_users.addElement(all_users[i]);
		}
		without_users.insertElementAt(User.ANONYMOUS_NAME, 0);
	    }
	}

	// Now put the user names into their SELECT lists as options:
	String WITH_ACCESS = "";
	Enumeration with = with_users.elements();
	while(with.hasMoreElements())
	{
	    WITH_ACCESS = WITH_ACCESS + "<option>" +
		(String)with.nextElement() + "</option>";
	}

	String WITHOUT_ACCESS = "";
	Enumeration without = without_users.elements();
	while(without.hasMoreElements())
	{
	    WITHOUT_ACCESS = WITHOUT_ACCESS + "<option>" +
		(String)without.nextElement() + "</option>";
	}
	
	steam_vars.put("WITH_ACCESS", WITH_ACCESS);
	steam_vars.put("WITHOUT_ACCESS", WITHOUT_ACCESS);
    }
    
    /** display a generic error message. Just a wrapper for whatever
     * method of error display the site uses.  */
    private void sendErrorMsg(String errmsg) throws IOException
    {
	log (0,  errmsg);
	resp.sendError(resp.SC_INTERNAL_SERVER_ERROR, errmsg);
//	(new ErrorHandler(errmsg)).handleRequest(hd);
    }


    /** intializes the DiscussionLib this Servlet will use */
    public void init(ServletConfig config) throws ServletException
    {
	Connection conn = Loco.getDBConnection();

	try
	{
	    lib = new DiscussionLib(conn);
	}
	catch (SQLException se)
	{
	    throw new ServletException(
		"ERROR: could not initialize DiscussionLib: " + se);
	}
	finally
	{
	    if (conn != null) Loco.putBackDBConnection(conn);
	}
    }



}
