/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.discuss;

import java.util.*;
import java.sql.*;
import java.io.*;
import java.text.*;
import org.locomotive.loco.*;
import org.locomotive.util.*;

/**
 * This class provides a series of static methods which return Database 
 * information used by the Locomotive Discussion system.
 * Each of the methods return the data in a
 * org.locomotive.util.SortableTable format, which then may be altered
 * by the handler and displayed using the [table ] expression in the
 * Steam template language.  All database connectivity for the
 * Discussion system is provided in the class; if support for new
 * databases is required, then only this class should be altered!
 * <P>Note that any reference to 'deleting' topics, posts, or
 * discussions via the handlers only changes their status to 'DEAD'.
 * To actually remove them from the database, use the
 * deleteDeadTopics(), deleteDeadPosts(), and deleteDeadDiscussions()
 * methods. The main method for this class does just this.
 */
public class DiscussionLib
{

    /** used by the getPostSummaries method specify sorting */
    public static final int DATE   = 1024;
    /** used by the getPostSummaries method specify sorting */
    public static final int AUTHOR = 1025;
    /** used by the getPostSummaries method specify sorting */
    public static final int TITLE  = 1026;

    
    // default
    private int DB_TYPE = SQL.DB_TYPE_ORACLE;


    /**
     * Initializes a DiscussionLib instance, with the correct database
     * type.  
     */
    public DiscussionLib (Connection conn) throws SQLException
    {
	DB_TYPE = SQL.getDBType(conn);
    }
	

    public void getUserInfo(Connection conn, User user, Hashtable steam_vars)
	throws SQLException
    {
	String statement = "select firstname, lastname, email_addr from " +
	    "loco_user_prefs where user_id_ref = " + user.getUserid();

	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(statement);

	    if(rs.next())
	    {
		steam_vars.put("firstname", rs.getString(1));
		steam_vars.put("lastname", rs.getString(2));
		steam_vars.put("email", rs.getString(3));
	    }

	}
	finally
	{
	    if (rs != null) { rs.close(); }
	    if (stmt != null) { stmt.close(); }
	}
    }

    /** returns the list of categories available for the community
     * @param status the status of the list of communities to return
     * @return a SortableTable with the following columns:<BR><PRE>
     *   0  id                  Integer
     *   1  name                String
     *   2  description         String
     *   3  status              string 
     *   4  topic_count         Integer 
     *   5  post_count          Integer 
     *   6  last_post_date      Integer
     *  </PRE>
     */
    public SortableTable getCategoryList(Connection conn, 
					 String status)
    {
	return getCategoryInfo(conn, -1, status);
    }


    /** retrieves a single category's information
     * @param id the id of the category to return.
     * @return a SortableTable with the following columns:<BR><PRE>
     *   0  id                  Integer
     *   1  name                String
     *   2  description         String
     *   3  status              string 
     *   4  topic_count         Integer 
     *   5  post_count          Integer 
     *   6  last_post_date      Integer
     *  </PRE>
     */
    public SortableTable getCategoryInfo(Connection conn, int id)
    {
	return getCategoryInfo(conn, id, "LIVE");
    }


    /** the workhorse for the above two.  Combined them for easier
     * handling of the sql and jdbc statements.
     */
    private SortableTable getCategoryInfo(Connection conn, int id, 
						 String status)
    {
	StringBuffer query1 = new StringBuffer();
	StringBuffer query2 = new StringBuffer();
	StringBuffer query3 = new StringBuffer();

	boolean status_dead = status.equals("DEAD");
	boolean status_live = status.equals("LIVE");
	boolean has_id = (id != -1);

	switch (DB_TYPE)
	{
	case SQL.DB_TYPE_MYSQL:
	    
	    // this mess is because MySQL doesn't support 
	    // count ( distinct <colname>) queries.  We have to
	    // seperate out into two queries.
	    query1.append("SELECT c.communityid as id, ");
	    query1.append("c.communityname as name, ");
	    query1.append("c.communitydesc as description, ");
	    query1.append("c.communitystatus as status ");
	    query1.append("FROM community c WHERE ");

	    query2.append("SELECT COUNT(ct.topicid) as topic_count ");
	    query2.append("FROM community c, communitytopic ct WHERE ");

	    query3.append("SELECT COUNT(cp.postid ) as post_count, ");
	    query3.append("MAX(cp.postDate) as last_post_date ");
	    query3.append("FROM community c, communitytopic ct, ");
	    query3.append("communitypost cp WHERE ");

	    if (has_id)
	    {
		query1.append("c.communityid = " + id);
		query2.append("c.communityid = " + id);
		query3.append("c.communityid = " + id);
	    }
	    
	    if (status_live || status_dead)
	    {
		if (has_id)
		{ 
		    query1.append(" AND ");
		    query2.append(" AND ");
		    query3.append(" AND ");
		}

		query1.append("c.communitystatus = ");
		query1.append(SQL.quoteString(status));
		
		query2.append("c.communitystatus = ");
		query2.append(SQL.quoteString(status));

		query3.append("c.communitystatus = ");
		query3.append(SQL.quoteString(status));
	    }
	    
	    query1.append(" ORDER BY c.communityid");

	    query2.append(" AND ct.communityid = c.communityid ");
	    query2.append("AND ct.topicstatus = 'LIVE' ");
	    query2.append("GROUP BY c.communityid ORDER BY c.communityid ");

	    query3.append(" AND ct.communityid = c.communityid ");
	    query3.append("AND cp.topicid = ct.topicid ");
	    query3.append("AND ct.topicstatus = 'LIVE' ");
	    query3.append("AND cp.poststatus = 'LIVE' ");
	    query3.append("GROUP BY c.communityid ORDER BY c.communityid");
	    break;

	default:

	    // we split this one up because Oracle won't let us group by
	    // a LONG datatype, thereby making this impossible as a single 
	    // query (at least as far as I know...)
	    query1.append("SELECT c.communityid as id, ");
	    query1.append("c.communityname as name, ");
	    query1.append("c.communitydesc as description, ");
	    query1.append("c.communitystatus as status ");
	    query1.append("FROM community c WHERE ");

	    query2.append("SELECT COUNT(DISTINCT ct.topicid) as topic_count, ");
	    query2.append(" COUNT(cp.postid ) as post_count, ");
	    query2.append(" MAX(cp.postDate) as last_post_date ");
	    query2.append(" FROM communitytopic ct, communitypost cp, ");
	    query2.append(" community c WHERE ");

	    if (has_id)
	    {
		query1.append("c.communityid = " + id);
		query2.append("c.communityid = " + id);
	    }
	    
	    if (status_live || status_dead)
	    {
		if (has_id)
		{ 
		    query1.append(" AND ");
		    query2.append(" AND ");
		}		

		query1.append("c.communitystatus = ");
		query1.append(SQL.quoteString(status));
		
		query2.append("c.communitystatus = ");
		query2.append(SQL.quoteString(status));
	    }		

	    query2.append(" AND ct.communityid = c.communityid");
	    query2.append(" AND cp.topicid = ct.topicid");
	    query2.append(" AND ct.topicstatus = 'LIVE'");
	    query2.append(" AND cp.poststatus = 'LIVE'");
	    query2.append(" GROUP BY c.communityid ORDER BY c.communityid");
	}

	Statement stmt = null;
	ResultSet rs = null;
	String query = query1.toString();
	SortableTable table = null;

	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);
	    table = new SortableTable();
	    table.load(rs, true);
	    
	    query = query2.toString();
	    rs = stmt.executeQuery(query);
	    table.load(rs, 0, 4, true);

	    if (query3.length() > 0)
	    {
		query = query3.toString();
		rs = stmt.executeQuery(query);
		table.load(rs, 0, 5, true);
	    }

	    if (id < 1 && DB_TYPE == SQL.DB_TYPE_MYSQL)
	    {
		// our lovely hack to get around a bug in MySQL:
		for (int i = 0; i < table.getNumRows(); i++)
		{
		    Object o = table.get(i, 6);
		    if (o != null && ! (o instanceof Timestamp))
		    {
			String s = new String((byte []) o);
			table.insert(Timestamp.valueOf(s), i, 6);
		    }
		}
	    }
		
	    table.formatDateColumn( 6, "MM/d/yyyy");
	    table.changeNullsToString( 1, "(No name)");
	    table.changeNullsToString( 2, "(no description)");
	    table.changeNullsToString( 4, "(no topics)");
	    table.changeNullsToString( 5, "(no posts)");
	    table.changeNullsToString( 6, "(no posts)");
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: getCategoryInfo: " + e + 
		": " + query);
	}
	finally
	{
	    try
	    {
		if (rs != null) rs.close();
		if (stmt != null) stmt.close();
	    } 
	    catch (SQLException e) {}
	}

	if (table.getNumRows() > 0) 
	    return table;
	else
	    return null;

    }


    /** returns a list of topics for the category whose id is specified 
     * @param catId the id of the category to return topics for
     * @return a SortableTable with the following columns:<BR><PRE>
     *   0  topicid             int
     *   1  communityid         int
     *   2  topicname           String
     *   3  post_count          Integer
     *   4  last_post_date      Integer</PRE>
     */
    public SortableTable getTopicList(Connection conn, int catId, 
					     String status) 
    {
	String query = "select ct.topicid, ct.communityid, ct.topicname, " +
	    "count(cp.topicid) as post_count, max(cp.postdate) as " +
	    "last_post_date from communitytopic ct, communitypost cp " +
	    "where ct.topicid = cp.topicid and ct.communityid = " + catId + 
	    " AND ct.topicstatus = " + SQL.quoteString(status) + 
	    " AND cp.poststatus = 'LIVE' " +
	    " group by ct.topicid, ct.communityid, ct.topicname" +
	    " order by ct.topicname ";

	SortableTable table = SortableTable.loadQuery(conn, query);
	if ( table == null)
	    return null;

	table.formatDateColumn(4, "MM/d/yyyy");
	table.changeNullsToString(2, "(no name)");
	table.changeNullsToString(3, "(no posts)");
	table.changeNullsToString(4, "(no posts)");
	return table;
    }


    /** returns the information for the topic whose id is specified
     * @param topicId the id of the topic to return. If -1, all topics 
     *        are returned.
     * @return a SortableTable with the following columns:<BR><PRE>
     *   0  topicid             int
     *   1  communityid         int
     *   2  topicname           String
     *   3  post_count          Integer
     *   4  last_post_date      Integer
     *   5  topicstatus         String
     * </PRE>
     */
    public SortableTable getTopicInfo(Connection conn, int topicId) 
    {
	StringBuffer query = new StringBuffer();
	query.append("select ct.topicid, ct.communityid, ct.topicname, ");
	query.append("count(cp.topicid) as post_count, max(cp.postdate) as ");
	query.append("last_post_date, ct.topicstatus from communitytopic ct ");

	switch (LocoConfig.DB_TYPE)
	{
	case SQL.DB_TYPE_INFORMIX:
	    query.append(", OUTER communitypost cp "); 
	    query.append(" where ct.topicid = "); query.append(topicId); 
	    query.append(" and cp.topicid = ct.topicid");
	    query.append(" and cp.poststatus = 'LIVE' ");
	    break;
	case SQL.DB_TYPE_ORACLE:
	    query.append(", communitypost cp "); 
	    query.append(" where ct.topicid = "); query.append(topicId); 
	    query.append(" and cp.topicid (+) = ct.topicid ");
	    query.append(" and cp.poststatus (+) = 'LIVE' ");
	    break;
	case SQL.DB_TYPE_MYSQL:
	    query.append(" LEFT OUTER JOIN communitypost cp on ");
	    query.append(" cp.topicid = ct.topicid");
	    query.append(" and cp.poststatus = 'LIVE' ");
	    query.append(" where ct.topicid = "); query.append(topicId); 
	    break;
	}

	query.append(" group by ct.topicid, ct.communityid, ct.topicname");

	SortableTable table = SortableTable.loadQuery(conn, query.toString());
	if (table == null) return null;

	table.formatDateColumn(4, "MM/d/yyyy");
	return table;
    }

    /** returns a list of posts for a topic specified by the id.
    * @param start the post number to start counting back from.  In
    *   other words, return the first num posts whose postid is 
    *   less than or equal to this number.  If the number is -1, 
    *   return the last num posts.
    * @param num the number of posts to return
    * @return a SortableTable with the following columns:<BR><PRE>
    *    0  postid              int
    *    1  posttitle           String
    *    2  postbody            String
    *    3  postauthorfn        String
    *    4  postauthorln        String
    *    5  postemail           String
    *    6  postdate            Date</PRE>
    */
    public SortableTable getPostList(Connection conn, int topicId,
					    int start, int num)
    {
	// get the first num rows out the result set and get rid of the
	// connection
	StringBuffer query = new StringBuffer();
	query.append("select postid, posttitle, postbody, ");
	query.append("postauthorfn, postauthorln, postemail,  ");
	query.append("postdate "); // , postdate as date
	query.append("from communitypost where topicid = ");
	query.append(topicId);
	query.append(" and poststatus = 'LIVE'");
	
	if (start != -1)
	{
	    query.append(" and postId <= ");
	    query.append(start);
	}

	query.append(" order by postdate desc, postid desc");

	Statement stmt = null;
	ResultSet rs = null;
	SortableTable table = null;

	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query.toString());
	    table = new SortableTable();

//	    table.load(rs, true);

	    // get the column names
	    ResultSetMetaData rsmd = rs.getMetaData();
	    int colCount = rsmd.getColumnCount();

	    for (int i = 0; i < colCount; i++)
	    {
		table.setColumnName(i, rsmd.getColumnLabel(i+1));
		table.setSQLType(i, rsmd.getColumnType(i+1));
	    }

	    // get the rest... up to num
	    for (int i = 0; i < num && rs.next(); i++)
	    {
		for (int j = 0; j < colCount; j++)
		{
		    table.insert(rs.getObject(j+1), i, j);
		}
	    }
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: getPostList: " + e + " " +
		    query.toString());
	    return null;
	}

	try
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	} 
	catch (SQLException e) {}

	// let's switch the order around, to make the last post at the
	// bottom
	SortableTable temp = new SortableTable();
	int row = 0;
	int colCount = table.getNumColumns();
	for (int j = 0; j < colCount; j++)
	{
	    temp.setColumnName(j, table.getColumnName(j));
	    temp.setSQLType(j, table.getSQLType(j));
	}

	for (int i = table.getNumRows() - 1; i >= 0; i--, row++)
	{
	    for (int j = 0; j < colCount; j++)
	    {
		temp.insert(table.get(i, j), row, j);
//		Loco.log (0, "\ti: " + i + "\tj: " + j + 
//			"\trows: " + row + "\tObj: " + table.get(i, j));
	    }
	}

	temp.formatDateColumn(6, "MM/d/yyyy");

	Loco.log(9, "DEBUG: posts: " + temp.toHTML());
	
	return temp;

    }


    /** returns a post
    * @param postId the post id
    * @param num the number of posts to return
    * @return a SortableTable with the following columns:<BR><PRE>
    *    0  postid              int
    *    1  posttitle           String
    *    2  postbody            String
    *    3  postauthorfn        String
    *    4  postauthorln        String
    *    5  postemail           String
    *    6  postdate            Date
    *    7  topicid             int
    * </PRE>
    */
    public SortableTable getPost(Connection conn, int postId)
    {
	// get the first num rows out the result set and get rid of the
	// connection
	String query = 
	    "select postid, posttitle, postbody, postauthorfn, " +
	    "postauthorln, postemail, postdate, " +
	    "topicid from communitypost where postid = " + postId;

	return SortableTable.loadQuery(conn, query);
    }


    /** returns the entire list of posts from the database for the
     * topic specified by the id 
     * @return a SortableTable with the following columns:<BR><PRE>
     *    0  postid              int</PRE>
     */
    public SortableTable getPostIds(Connection conn, int id)
    {
	String query = "select postid, postdate from  communitypost " +
	    "where topicid = " + id + " and poststatus = 'LIVE' " +
	    " order by postdate desc, postid";

	return SortableTable.loadQuery(conn, query);
    }


    /** this method returns a summary of all the posts related to
     * a particular topic.  These summaries may be sorted in three
     * ways: by date, by Author, and by Title.
     * @param id the topic's id to summarize the posts for
     * @param order one of either of the class constants
     *        DiscussionLib.DATE, DiscussionLib.AUTHOR, DiscussionLib.TITLE
     * @return a SortableTable with the following columns:<BR><PRE>
     *    0  postid              int
     *    1  posttitle           String
     *    2  postauthorfn        String
     *    3  postauthorln        String
     *    4  postemail           String
     *    5  postdate            Date</PRE>
     */
    public SortableTable getPostSummaries(Connection conn, 
						 int id, int order)
    {
	StringBuffer query = new StringBuffer();
	query.append("select postid, posttitle, postauthorfn, ");
	query.append("postauthorln, postemail, postdate ");
	query.append("from communitypost where topicid = ");
	query.append(id);
	query.append(" and poststatus = 'LIVE' ");

	// default order is by date
	switch (order)
	{
	case DATE: 
	    query.append(" order by postdate, postid"); 
	    break;
	case AUTHOR: 
	    query.append(
		" order by postauthorln, postauthorfn, postemail");
	    break;
	case TITLE:
	    query.append(" order by posttitle");
	    break;
	default:
	    query.append(" order by date, postid");
	}
	    
	SortableTable table = SortableTable.loadQuery(conn, query.toString());
	if (table == null) return null;

	table.formatDateColumn(5, "MM/d/yyyy");
	return table;
    }

    /** adds a post with the specified into to list of posts for the 
     * topic specified by the id.
     * @return whether the update was successful
     */
    public boolean addPost(Connection conn, int topicId, String title, 
				  String postBody, String firstname, 
				  String lastname, String email)
    {
	boolean success = false;
	PreparedStatement pstmt = null;
	String query = "insert into communitypost (postid, topicid, " +
	    "posttitle, postbody, postauthorfn, postauthorln, " +
	    "postemail, postdate, poststatus) ";

	switch (LocoConfig.DB_TYPE)
	{
	case (SQL.DB_TYPE_INFORMIX):
	case (SQL.DB_TYPE_MYSQL):
	    query +=" values (0, ?, ?, ?, ?, ?, ?, ?, 'LIVE')"; 
	    break;
	case (SQL.DB_TYPE_ORACLE):
	    query += " values (communitypost_id_sq.nextval, " +
		"?, ?, ?, ?, ?, ?, ?, 'LIVE')"; 
	    break;
	}
	    
	    
//	String query = "insert into communitypost (postid, topicid, " +
//	    "posttitle, postauthorfn, postauthorln, " +
//	    "postemail, postdate) values (0, ?, ?, ?, ?, ?, ?)";


	try 
	{
	    pstmt = conn.prepareStatement(query.toString());
	    pstmt.setInt(1, topicId);
	    pstmt.setString(2, title);
	    
	    ByteArrayInputStream pbis = 
		new ByteArrayInputStream(postBody.getBytes());
	    
	    pstmt.setAsciiStream(3, pbis, postBody.length());
	    pstmt.setString(4, firstname);
	    pstmt.setString(5, lastname);
	    pstmt.setString(6, email);
	    pstmt.setTimestamp(7, new Timestamp(System.currentTimeMillis()));
	    pstmt.executeUpdate();
	    success = true;
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: addPost: " + query + e);
	}
	finally
	{
	    try { if (pstmt != null) pstmt.close(); }
	    catch (SQLException e) {}
	}
	return success;
    }

    /** changes a post status. Use this to 'delete' a post- just set post
     * status to 'DEAD'
     */
    public boolean changePostStatus(Connection conn, int postId,
				     String status)
    {    
	boolean success = false;
	Statement stmt = null;
	String query = "update communitypost set poststatus = '" +
	    status + "' where postid = " + postId;

	try 
	{
	    stmt = conn.createStatement();
	    stmt.executeUpdate(query);
	    success = true;
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: changePostStatus: " + 
		    query + e);
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); }
	    catch (SQLException e) {}
	}
	return success;
    }


    /** modifies an existing post with the new fields passed in
     * as parameters.  To 'delete' a post- just set post
     * status to 'DEAD'
     */
    public boolean modifyPost(Connection conn, int postId, int topicId,
				     String title, String postBody, 
				     String firstname, String lastname, 
				     String email, String status)
    {    
	boolean success = false;
	PreparedStatement pstmt = null;
	String query = "update communitypost set topicid = ?, " +
	    "posttitle = ?, postbody = ?, postauthorfn = ?, " +
	    "postauthorln = ?, postemail = ?, poststatus = ? " +
	    "where postid = ?";

	try 
	{
	    pstmt = conn.prepareStatement(query.toString());
	    pstmt.setInt(1, topicId);
	    pstmt.setString(2, title);
	    
	    ByteArrayInputStream pbis = 
		new ByteArrayInputStream(postBody.getBytes());
	    
	    pstmt.setAsciiStream(3, pbis, postBody.length());
	    pstmt.setString(4, firstname);
	    pstmt.setString(5, lastname);
	    pstmt.setString(6, email);
	    pstmt.setString(7, status);
	    pstmt.setInt(8, postId);
	    pstmt.executeUpdate();
	    success = true;
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: addPost: " + query + e);
	}
	finally
	{
	    try { if (pstmt != null) pstmt.close(); }
	    catch (SQLException e) {}
	}
	return success;
    }



    /** adds a topic for the category specified by the id whose title
     * is the specified title.
     * @return a SortableTable with the following columns:<BR><PRE>
     *   0  topicid             int
     *   1  communityid         int
     *   2  topicname           String</PRE>
     */
    public SortableTable addTopic(Connection conn, 
					 int catId, String title) 
    {
	StringBuffer query = new StringBuffer();
	query.append("insert into communitytopic (topicid, ");
	query.append( "communityid, topicname, topicstatus) ");

	switch (LocoConfig.DB_TYPE)
	{
	case SQL.DB_TYPE_INFORMIX:
	case SQL.DB_TYPE_MYSQL:
	    query.append( " values (0, ");
	    break;
	case SQL.DB_TYPE_ORACLE:
	    query.append( " values (communitytopic_id_sq.nextval, ");
	    break;
	}

	query.append( catId); query.append(", ");
	query.append(SQL.quoteString(title)); 
	query.append(", 'DEAD')");

	Statement stmt = null;
	ResultSet rs = null;
	SortableTable topicInfo = null;

	try
	{
	    stmt = conn.createStatement();
	    stmt.executeUpdate(query.toString());

	    // now get the topic info
	    String queryStr =  
		"select topicid, communityid, topicname, topicstatus " +
		" from communitytopic where topicname = " +
		SQL.quoteString(title) + " and communityid = " + catId +
		" order by topicid desc";
	    
	    rs = stmt.executeQuery(queryStr);

	    topicInfo = new SortableTable();
	    topicInfo.load(rs, true);
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: AddTopic: " + query + 
		     " error: " + e);
	} 
	finally
	{
	    try
	    {
		if (stmt != null) stmt.close();
		if (rs != null) rs.close();
	    } 
	    catch (SQLException e) {}	
	}

	if (topicInfo == null || topicInfo.getNumRows() < 1) {
	    return null;
	} else {
	    return topicInfo;
	}
    }


    /** Changes a topic's status to the specified parameter.
     * @param conn the Connection to use
     * @param id the topicid of the topic to change
     * @param status the status to change the topic to.  Should
     * be either LIVE or DEAD.
     */
    public void changeTopicStatus(Connection conn, int id, 
					 String status)
    {
	modifyTopic(conn, id, null, status);
    }


    /** Modify the topic.
     * @param conn the Connection to use
     * @param id the topicid of the topic to change
     * @param title the new title for the topic.  Set this to null
     * if you don't want to update the title.
     * @param status the status to change the topic to.  Should
     * be either 'LIVE' or 'DEAD'.
     */
    public void modifyTopic(Connection conn, int id, 
				   String title, String status)
    {
	String query = null;
	if (title != null && title.length() > 0)
	{
	    query = "update communitytopic set topicname = " + 
		SQL.quoteString(title) + ", topicstatus = '" + 
		status + "' where topicid = " + id;
	}
	else
	{
	    query = "update communitytopic set topicstatus = '" + status + 
		"' where topicid = " + id;
	}

	Statement stmt = null;
	
	try
	{
	    stmt = conn.createStatement();
	    stmt.executeUpdate(query);
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: modifyTopic: " +
		    "could not update topic status: " + e );
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}
    }


    /** adds a new category.  Returns a table containing the category's info:
     * @return a SortableTable with the following columns:<BR><PRE>
     *   0  id                  int
     *   1  name                String
     *   2  description         String
     * </PRE>
     * Return null if there was an error
     */
    public SortableTable addCategory(Connection conn, String title,
					    String body, String status)
    {
	StringBuffer query = new StringBuffer();
	query.append("insert into community (communityid, ");
	query.append( "communityname, communitydesc, communitystatus) ");

	switch (LocoConfig.DB_TYPE)
	{
	case SQL.DB_TYPE_INFORMIX:
	case SQL.DB_TYPE_MYSQL:
	    query.append( " values (0, ");
	    break;
	case SQL.DB_TYPE_ORACLE:
	    query.append( " values (community_id_sq.nextval, ");
	    break;
	}

	query.append(" "); query.append( SQL.quoteString(title)); 
	query.append(", "); query.append(SQL.quoteString(body)); 
	query.append(", '"); query.append(status); query.append("')");

	Statement stmt = null;
	ResultSet rs = null;
	SortableTable topicInfo = null;

	try
	{
	    stmt = conn.createStatement();
	    stmt.executeUpdate(query.toString());

	    // now get the community info
	    query = new StringBuffer();
	    query.append("select communityid, communityname, communitydesc, ");
	    query.append("communitystatus from community ");
	    query.append(" where communityname = "); 
	    query.append(SQL.quoteString(title));
	    query.append(" order by communityid desc");
	    
	    rs = stmt.executeQuery(query.toString());

	    topicInfo = new SortableTable();
	    topicInfo.load(rs, true);
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: AddCategory:" + e + 
		    " query: " + query);
	} 
	finally
	{
	    try
	    {
		if (stmt != null) stmt.close();
		if (rs != null) rs.close();
	    } 
	    catch (SQLException e) {}	
	}

	if (topicInfo == null || topicInfo.getNumRows() < 1)
	{
	    return null;
	}
	else if (topicInfo.getNumRows() > 1)
	{
	    topicInfo.setRowCapacity(1);
	}

	return topicInfo;
    }	


    /** Changes a topic's status to the specified parameter.
     * @param conn the Connection to use
     * @param id the topicid of the topic to change
     * @param status the status to change the topic to.  Should
     * be either LIVE or DEAD.
     */
    public void changeCategoryStatus(Connection conn, int id, 
					    String status)
    {
	modifyCategory(conn, id, null, null, status);
    }


    /** Modify the topic.
     * @param conn the Connection to use
     * @param id the topicid of the topic to change
     * @param title the new title for the topic.  Set this to null
     * if you don't want to update the title.
     * @param status the status to change the topic to.  Should
     * be either 'LIVE' or 'DEAD'.
     */
    public void modifyCategory(Connection conn, int id, String title, 
				      String body, String status)
    {
	StringBuffer query = new StringBuffer();
	query.append("update community set communitystatus = '");
	query.append(status + "'");

	if (title != null && title.length() > 0)
	{
	    query.append(", communityname = '" + title + "'");
	}
	
	if (body != null && body.length() > 0)
	{
	    query.append(", communitydesc = '" + body + "'");

	}
	
	query.append(" where communityid = " + id);

	Statement stmt = null;
	
	try
	{
	    stmt = conn.createStatement();
	    stmt.executeUpdate(query.toString());
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: modifyTopic: " +
		    "could not update topic status: " + e  + " query: " + 
		    query.toString());
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}
    }

    /** Get rid of the 'DEAD' posts.  This is an administration
     * helper method that can be used to clean out deleted posts.
     * Run this before deleteDeadTopics(). 
     * @return the number of topics we deleted, or -1 if error
     */
    public int deleteDeadPosts(Connection conn)
    {
	int deleted = -1;
	String query = "delete from communitypost where poststatus = 'DEAD'";
	Statement stmt = null;

	try
	{
	    stmt = conn.createStatement();
	    deleted = stmt.executeUpdate(query);
	}
	catch (SQLException e)
	{
	    String err = "ERROR: DiscussionLib: deleteDeadPosts: " +
		    "failed!: " + e;
	    try
	    {
		Loco.log (0, err);
	    }
	    catch (NullPointerException ne)
	    {
		System.err.println(err);
	    }
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}
	return deleted;
    }


    /** Get rid of the 'DEAD' topics.  This is an administration method
     * that can be used by an administration handler to clean out
     * all deleted topics.  Run this after deleteDeadPosts().
     * @return the number of topics we deleted, or -1 if error
     */
    public int deleteDeadTopics(Connection conn)
    {
	int deleted = -1;
	String query = 
	    "delete from communitytopic where topicstatus = 'DEAD'";
	Statement stmt = null;

	try
	{
	    stmt = conn.createStatement();
	    deleted = stmt.executeUpdate(query);
	}
	catch (SQLException e)
	{
	    String err = "ERROR: DiscussionLib: deleteDeadTopics: " +
		    "failed!: " + e;
	    try
	    {
		Loco.log (0, err);
	    }
	    catch (NullPointerException ne)
	    {
		System.err.println(err);
	    }
		    
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}
	return deleted;
    }


    /** Get rid of the 'DEAD' discussions.  This is an administration method
     * that can be used by an administration handler to clean out
     * all deleted discussions.  Run this after deleteDeadPosts() and 
     * deleteDeadTopics().
     * @return the number of discussions we deleted, or -1 if error
     */
    public int deleteDeadDiscussions(Connection conn)
    {
	int deleted = -1;
	String query = "delete from community where commmunitystatus = 'DEAD'";
	Statement stmt = null;

	try
	{
	    stmt = conn.createStatement();
	    deleted = stmt.executeUpdate(query);
	}
	catch (SQLException e)
	{
	    String err = "ERROR: DiscussionLib: deleteDeadDiscussions: " +
		    "failed!: " + e;
	    try
	    {
		Loco.log (0, err);
	    }
	    catch (NullPointerException ne)
	    {
		System.err.println(err);
	    }
		    
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}
	return deleted;
    }



    /** Delete all topics which don't have any posts.  Another administration
     * method. 
     * @return a SortableTable containing all the topics we deleted,
     * or null if there were no topics to delete or if there was an error.
     */
    public SortableTable cleanTopics(Connection conn)
    {
	String query = 
	    "select topicid, topicname from communitytopic where topicid " +
	    "not in (select communitypost.topicid from communitypost, " +
	    " communitytopic where communitypost.topicid = " +
	    "communitytopic.topicid) ";
	String query2 =
	    "delete from communitytopic where topicid = ?";
	
	PreparedStatement stmt = null;
	ResultSet rs = null;
	SortableTable table = null;
	try
	{
	    stmt = conn.prepareStatement(query);
	    rs = stmt.executeQuery();
	    table = new SortableTable();
	    table.load(rs, true);
	    rs.close();
	    stmt.close();
	    stmt = conn.prepareStatement(query2);
	    for (int i = 0; i < table.getNumRows(); i++) {
		stmt.setInt(1, ((Number) table.get(i, 0)).intValue());
		stmt.executeUpdate();
	    }
	}
	catch (SQLException e)
	{
	    Loco.log (0, "ERROR: DiscussionLib: loadQuery: " + 
		": " + query);
	}
	finally
	{
	    
	    try
	    {
		if (rs != null) rs.close();
		if (stmt != null) stmt.close();
	    } 
	    catch (SQLException e) {}
	}

	if (table.getNumRows() > 0)
	    return table;
	else 
	    return null;
    }




    /** This sample app will go ahead and delete all the discussions, 
     * topics and posts marked as 'DEAD'.  It will remove all the topics 
     * without posts too.  This may remove data you want to hold on to, so
     * use it with care.
     * You'll have to go into the source yourself to edit the database
     * user, password, driver, and URL.
     * It wouldn't be bad to run this as a cron job occasionally...
     */
    public static void main(String [] args)
    {

	String user = "Loco";
 	String password = "gr0ntr33";
	String driver = "sun.jdbc.odbc.JdbcOdbcDriver";
	String url = "jdbc:odbc:GreenTree_odbc";
	Connection conn = null;
	DiscussionLib lib = null;

	long time = System.currentTimeMillis();
	try
        {
	    // get the connection.
	    try
	    {
		Class.forName(driver);
		conn = DriverManager.getConnection(url, user, password);
		lib = new DiscussionLib(conn);
	    } catch (ClassNotFoundException e) {
		throw new ClassNotFoundException (
		"Driver could not be found: " + driver);
	    } catch (SQLException e) {
		throw new SQLException("could get connection: " + 
				       e.getMessage());
	    }

	    System.out.println("Deleting Dead posts: ");
	    int deleted = lib.deleteDeadPosts(conn);
	    switch (deleted)
	    {
	    case -1: 
		System.err.println("ERROR: could not delete topics");
		System.exit(-1);
		break;
	    default:
		System.out.println(deleted + " posts deleted.");
		break;
	    }

	    System.out.println("Deleting Dead topics: ");
	    deleted = lib.deleteDeadTopics(conn);
	    switch (deleted)
	    {
	    case -1: 
		System.err.println("ERROR: could not delete topics");
		System.exit(-1);
		break;
	    default:
		System.out.println(deleted + " topics deleted.");
		break;
	    }
	    System.out.println("Deleting Dead discussions: ");
	    deleted = lib.deleteDeadDiscussions(conn);
	    switch (deleted)
	    {
	    case -1: 
		System.err.println("ERROR: could not delete discussions");
		System.exit(-1);
		break;
	    default:
		System.out.println(deleted + " discussions deleted.");
		break;
	    }
	    System.out.println("Deleting all topics without any posts");
	    SortableTable deletedTopics = lib.cleanTopics(conn);
	    if (deletedTopics == null)
	    {
		System.out.println("No topics were deleted.");
	    }
	    else
	    {
		System.out.println(deletedTopics.getNumRows() +
				   " topics deleted.");
	    }
	    System.out.println("Finished!");
	} catch (Exception e) 
	{
	    e.printStackTrace();
	    System.out.println(
		"org.locomotive.module.discuss.DiscussionLib: ERROR:\n" + e);

	}
	System.out.println( "Time: " + (System.currentTimeMillis() - time));
    }


} // end of discussionLib class

