/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.discuss;

import java.io.*;
import java.util.*;
import java.sql.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.util.*;
import org.locomotive.loco.perm.*;


/** 
 * This class is the main handler class for The Locomotive Discussion system.
 * The Locomotive Discussion system is a simple three-level web-based 
 * conferencing system for hosting non-threaded discussions.  This particular 
 * class directs the appropriate requests to the appropriate Handlers and 
 * provides a number of utility methods for use by those handlers.
 * The templates used for the various URL's handled by this class are
 *  listed under the service method.
 * <P>One config property can be set for this servlet- the number of
 * posts to display on a single list of posts.  One does this by setting
 * the num.posts.to.display property in a property file that is loaded
 * at startup for this Servlet.  For example, the following line would
 * specify 4 posts per page: <PRE>
 * num.posts.to.display = 4
 * </PRE>
 */
public class DiscussionServlet extends LocoServlet
{

    // load this in init
    private static Hashtable templateHash = new Hashtable(); 
    protected DiscussionLib lib;
    protected DiscussionAdminServlet admin;

    private PermissionSet pSet = null;

    /** SET THIS TO DETERMINE THE NUMBER OF POSTS TO DISPLAY ON
     * THE PAGE */
    public int numPostsToDisplay = 4;
    

    /** The main method to call for this Servlet.  Directs the handling of the
     * request to the appropriate places depending on the url tokens found.
     * Presently handles five different URL Tokens.  To do any custom handling
     * for these tokens, override the methods which handle the actions involved
     * with each token.  The tokens and their corresponding methods are:
     * <BR><PRE>
     *  Token      Method            Template
     *  CMCAT      showCategory()    discuss/community/comm_disc.tmpl
     *  CMTOP      showTopic()       discuss/community/comm_disc_topic.tmpl
     *  POST       addPost()         discuss/community/comm_disc_topic.tmpl
     *  TOPIC      addTopic()        discuss/community/comm_disc_topic_new.tmpl
     *  SUMMARY    showSummary()     discuss/community/comm_disc_topic_sum.tmpl
     *  Admin      see handleAdminHandler()
     *  [no token] showDefault()     discuss/community/comm_section.tmpl
     * </PRE>
     */
    public void service (HttpServletRequest req, HttpServletResponse resp)
	throws java.io.IOException, ServletException
    {
	setup(req, resp);
	
	// Permission checks:
	try
	{
	    pSet = UserPerm.getPermissions(user.getUserid(), conn);
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_getperms");
	    return;
	}
	
	StandardInsertions.insertAll(req, steam_vars); // for USERNAME
	
	try { lib.getUserInfo(conn, user, steam_vars); }
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_user_info");
	    return;
	}

	steam_vars.put("num_posts_to_display", new Integer(numPostsToDisplay));

	// set up page headers...
	if (url_tokens.length < 2) 
	{
	    showDefault();
	}
	else 
	{
	    String token = url_tokens[1];

	    if (token.equals("CMCAT")) 
	    {
		showCategory();
	    }
	    else if (token.equals("CMTOP")) 
	    {
		showTopic();
	    }
	    else if (token.equals("CMPOS")) 
		// this shouldn't be needed!
	    {
		showPost();
	    }
	    else if (token.equals("TOPIC")) 
	    {
		addTopic();
	    }
	    else if (token.equals("POST")) 
	    {
		addPost();
	    }
	    else if (token.equals("SUMMARY")) 
	    {
		showSummary();
	    }
	    else if (token.equals("Admin")) 
	    {
		// If a person is an admin, direct them to the
		// DiscussionAdminHandler.  
		if(pSet.isSet(DiscussionAdminHandler.COMM_ADMIN))
		{
		    admin.service(req, resp);
		}
		else
		{
		    // Person is not an admin; show them the default and log
		    // the incriminating little hacker!
		    log (0, "Userid " + user.getUserid() + " tried to " +
			    "access the DiscussionAdminHandler without " +
			    "permission.");
		    showDefault();
		}
	    }
	    else 
	    {
		// got a strange url token- log an error, but show the default 
		// page
		log (0, "ERROR: DiscussionHandler: malformed url: " +
			"Community:" + token);
		showDefault();
	    }
	}
    }


    /** This method displays the default Discussion Page, which contains a
     *  list of categories to choose from along with choices on where to go.
     * <P> Puts the following subs into the environment:<BR><PRE>
     *     categories_table      SortableTable with columns: id, title, 
     *                                 description
     *     num_posts_to_display  the number of posts to display on the 
     *                           topics page
     * </PRE>
     */
    protected void showDefault() throws IOException
    {
	// construct table for page (-1 id gets all)
	SortableTable categoriesTable = 
	    lib.getCategoryList(conn, "LIVE");
	
	// If the user is an admin, give them a link to the 
	// DiscussionAdminHandler:
	if(pSet.isSet(DiscussionAdminHandler.COMM_ADMIN))
	{
	    steam_vars.put("admin_link", url_tokens[0] + "?Admin");
	}
	
	if (categoriesTable == null) 
	{
	    log (0,"ERROR: DiscussionHandler: showDefault: " + 
		    "got null from DiscussionLib.getCategoryInfo(-1)");
	    insertTemplate("DEFAULT", resp, steam_vars);
	    return;
	}

	// Prune the table according to the user's permissions:
	categoriesTable = permPruneCategoriesTable(categoriesTable);

	// add the topic table to the template
	steam_vars.put("categories_table", categoriesTable);
	
	// for debugging
	steam_vars.put("categories_table_html", categoriesTable.toHTML());

	insertTemplate("DEFAULT", resp, steam_vars);
	logEvent("Community","","");
    }

    
    /** This method displays the Discussion cateogory page, which shows a list
     *  of the topics in that category.  The category to display the topics for
     * is provided by the URL token which follows the CMCAT token that calls
     * this method.
     * <P> Puts the following subs into the environment: <BR><PRE>
     *   navigation     the html for the navigation sidebar
     *   category_info  SortableTable with columns: id, title, description, 
     *                          topic_count, post_count, last_post_date
     *   topics_table   SortableTable with columns: topicid, communityid, 
     *                          topicname, post_count, last_post_date
     *   num_posts_to_display  the number of posts to display on the 
     *                          topics page
     * </PRE>
     */
    protected void showCategory() throws IOException
    {
	// get the table of all topics for this category
	int id = 0;
	try
	{
	    id = Integer.parseInt(url_tokens[2]);
	} 
	catch (NumberFormatException nfe)
	{
	    // log an error and send the user back to the default page
	    log (0, "ERROR: DiscussionHandler: showCategory: " +
		    "Malformed category id: " + url_tokens[2]);
	    showDefault();
	}

	// Check to see if the user has permission to read this category:
	try
	{
	    if(!ACLManager.permit(id, DiscussionAdminHandler.COMM_READ,
				  user.getUserid(), conn))
	    {
		// log a warning and send the user back to the default page
		log (0, "WARNING: DiscussionHandler: showCategory: " +
			"Userid " + user.getUserid() + " tried " + 
			"accessing a category that he/she does not " +
			"have read permission for: category id: " + id);
		showDefault();
	    }
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_acl");
	    return;
	}
	
	// get the category's info, plus the info of its topics
	SortableTable categoryInfo = 
	    lib.getCategoryInfo(conn, id);
	SortableTable topicsTable = 
	    lib.getTopicList(conn, id, "LIVE");
	
	// Permission prune the tables:
	categoryInfo = permPruneCategoriesTable(categoryInfo);
	topicsTable = permPruneTopicsTable(topicsTable);

	// add the topic table to the template
	steam_vars.put("category_info", categoryInfo);
	if (topicsTable != null) 
	    steam_vars.put("topics_table", topicsTable);

	// for debugging
	steam_vars.put("category_info_html", categoryInfo.toHTML());
	if (topicsTable != null) 
	    steam_vars.put("topics_table_html", topicsTable.toHTML());

	insertTemplate("CMCAT", resp, steam_vars);
	logEvent("Community","CMCAT+" + id,"");
    }


    /** This method displays the topic page, which contains a set of the 
     * posts for that topic.  The URL tokens for this function must be the
     * following: <BLOCKQUOTE><TT>
     *     Community?CMTOP+&lt;topicID&gt;+&lt;numberOfPostsToDisplay&gt;+&lt;postStartNum&gt;
     *     </TT></BLOCKQUOTE>
     * where the topicID is the unique topic id number for the topic whose
     * posts the page show display, the postStartNum is the post number to 
     * start on, and the numberOfPostsToDisplay is the number of posts to 
     * display on the page. 
     *  <P>Note that this method displays the posts in
     * descending order, so that a URL of Community?CMTOP+2+10+192 would
     * display posts 183 to 192 from top to bottom on the first page, and
     * the URL of Community?CMTOP+2+10+182 would display posts 173 to 
     * 182 from the top to the bottom of the page. 
     *  <P>The defaults for these values are the latest post for 
     * postStartNum, and 10 for numberOfPostsToDisplay, so the URL 
     * Community?CMTOP+2 evaluates to Community?CMTOP+2+10+192, if 192 is the
     * number of the latest post.
     *  <P>This method also inserts into steam_vars an array of urls to all the 
     * pages that would be generated to see all the posts for the current topic,
     * at the current numberOfPostsToDisplay token.  Thus, the first example 
     * above would provide an array of 20 urls which would point to each page 
     * needed to display all the posts.  Note that The page designer, using the 
     * [loop ] expression, can display as many or all of these as he or she 
     * would like.  See the steam documentation for more on this. 
     * <P> Puts the following subs into the environment:<BR><PRE>
     *     navigation      the html for the navigation sidebar
     *     category_info   SortableTable with columns: id, title, description, 
     *                          topic_count, post_count, last_post_date
     *     topic_info      SortableTable with columns: topicid, communityid, 
     *                          topicname, post_count, last_post_date 
     *     post_table      SortableTable with columns: postid, posttitle, 
     *                          postbody, postauthorfn, postauthorln, 
     *                          postemail, postdate
     *     post_links      Array of strings that provide the links to each
     *                          page of posts
     *     num_posts_to_display  the number of posts to display on the 
     *                            topics page
     * </PRE>
     */ 
    protected  void showTopic() throws IOException
    {
	// get the url tokens.  There has to be at least 3, and can be as many
	// as five
	int topicId = 0;          // the id of the topic to display posts for
	int postLimit = -1;        // the highest numbered post to display
	int postIterator = numPostsToDisplay; // # of posts to display on a page
	try
	{
	    topicId = Integer.parseInt(url_tokens[2]);
	    if (url_tokens.length > 3) 
	    {
		postIterator = Integer.parseInt(url_tokens[3]);
		
		if (url_tokens.length > 4)
		    postLimit = Integer.parseInt(url_tokens[4]);
	    }
	}
	catch (NumberFormatException nfe)
	{
	    StringBuffer sb = new StringBuffer();
	    for (int i = 1; i < url_tokens.length; i++)
	    {
		sb.append(url_tokens[i] + "+");
	    }
	    log (0, "DiscussionHandler: showTopic: " + sb.toString()) ;
	    showDefault();
	}

	// Check to see if the user has permission to read this topic:
	try
	{
	    int acl_id = topicId +
		DiscussionAdminHandler.TOPIC_TO_ACL_FILTER;
	    if(!ACLManager.permit(acl_id, DiscussionAdminHandler.COMM_READ,
				  user.getUserid(), conn))
	    {
		// log a warning and send the user back to the default page
		log (0, "WARNING: DiscussionHandler: showTopic: " +
			"Userid " + user.getUserid() + " tried " +
			"accessing a topic that he/she does not have " +
			" read permission for: topic id: " + topicId);
		showDefault();
	    }
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_acl_for_topic");
	    return;
	}	

	// we need three tables here: 1 the topic's name etc,
	// 2 the list of posts, and 3 the category name
	SortableTable topicInfo = lib.getTopicInfo(conn, topicId);
	SortableTable postTable = lib.getPostList(conn, topicId, 
						  postLimit, postIterator);

	// topicInfo.get(0, 1) should return the category id of the topic's
	//parent
	if (topicInfo != null)
	{
	    int catId = Integer.parseInt(topicInfo.get(0, 1).toString());

	    // Check to see if the user has permission to read this category:
	    try
	    {
		if(!ACLManager.permit(catId, DiscussionAdminHandler.COMM_READ,
				      user.getUserid(), conn))
		{
		    // log a warning and send the user back to the default page
		    log (0, "WARNING: DiscussionHandler: showTopic: " +
			    "Userid " + user.getUserid() + " tried " +
			    "accessing a category that he/she does not have " +
			    " read permission for: cat id: " + catId);
		    showDefault();
		}
	    }
	    catch(SQLException se)
	    {
		handleException (se, "discuss_get_acl_for_category");
		return;
	    }	

	    SortableTable categoryInfo = lib.getCategoryInfo(conn, catId );
	    
	    // add the topic table to the template
	    steam_vars.put("topic_info", topicInfo);
	    steam_vars.put("post_table", postTable);
	    steam_vars.put("category_info", categoryInfo);
	    
	    // for debugging
	    steam_vars.put("category_info_html", categoryInfo.toHTML());
	    steam_vars.put("topic_info_html", topicInfo.toHTML());
	    steam_vars.put("post_table_html", postTable.toHTML());

	    generatePostLinks(topicId, postIterator, postLimit);
	}
	else
	{
	    steam_vars.put("error_msg", "Information for topic with id " +
			   topicId + " could not be found...");
	    steam_vars.put("topic_id_err", new Integer(topicId));
	}
	
	insertTemplate("CMTOP", resp, steam_vars);
	String url = topicId + "+" + postIterator + "+" + postLimit;
	logEvent("Community","CMTOP+" + url, "");
    }
    

    /** This method displays a list of posts for a given topic.  The topic
     * to display the posts for is specified by the url token that follows the 
     * CMPOS token which causes this method to be called.
     *
     * Shouldn't be needed!
     */
    private void showPost() throws IOException
    {
    }
    

    /** This method handles adding a topic to a category.  The category to 
     * add the topic under is specified by the url token that follows the 
     * TOPIC token.  This will add the topic specified, and display the
     * 'add a first post template', whose actions are handled by the addPost
     *  method in this class.
     * <P> Puts the following subs into the environment:<BR><PRE>
     *     category_info  SortableTable with columns: id, title, description, 
     *                          topic_count, post_count, last_post_date
     *     topic_info      SortableTable with columns: topicid, communityid, 
     *                          topicname, post_count, last_post_date, 
     *     error_msg       Message if topic could not be added
     *     num_posts_to_display  the number of posts to display on the 
     *                           topics page
     * </PRE>
     */
    protected void addTopic() throws IOException
    {
	if (req.getMethod().equals("GET")) {  // somehow they got the wrong url
	    showCategory();
	    return;
	}

	int catId = 0;
	try
	{
	    catId = Integer.parseInt(url_tokens[2]);
	}
	catch (NumberFormatException nfe)
	{
	    log (0, "ERROR: DiscussionHandler: addTopic: malformed URL: " +
		    "TOPIC+" + url_tokens[2]);
	    showDefault();
	}

	// Check to see if the user has permission to read this category:
	try
	{
	    if(!ACLManager.permit(catId, DiscussionAdminHandler.COMM_READ,
				  user.getUserid(), conn))
	    {
		// log a warning and send the user back to the default page
		log (0, "WARNING: DiscussionHandler: addTopic: " +
			"Userid " + user.getUserid() + " tried " + 
			"accessing a category that he/she does not " + 
			"have read permission for: category id: " + catId);
		showDefault();
	    }
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_acl_for_category");
	    return;
	}

	// ------------ we should probably do just a little checking here------
	String title = form_data.getTrimmedString("topic_name", 
						  "name or messge", 1, 150);
	SortableTable topicsTable = 
	    lib.getTopicList(conn, catId, "LIVE");
	
	StringBuffer errors = new StringBuffer(form_data.allErrorMsgs());
	
	if (topicsTable != null)
	{
	    for (int i = 0; i< topicsTable.getNumRows(); i++) {
		log (8, "Topics in cat " + catId + ": " + 
			topicsTable.get(i, 2));
		if (topicsTable.get(i, 2).equals(title))
		    errors.append("\nThe title entered is identical to "+
				  "another topic.");
	    }
	}

	// if we added the topic successfully, get the category name and display
	// the first post page
        if (errors.length() < 1)
	{
	    SortableTable topicInfo =
		lib.addTopic(conn, catId, title);
	    
	    SortableTable catInfo = 
		lib.getCategoryInfo(conn, catId);
	    
	    // we have problems- do the errors below
	    if (topicInfo != null && catInfo != null)
	    {
		steam_vars.put("category_info", catInfo);
		steam_vars.put("topic_info", topicInfo);
		
		// for debugging
		steam_vars.put("topic_title", title);
		steam_vars.put("topic_id", topicInfo.get(0, 0).toString());
		steam_vars.put("category_info_html", catInfo.toHTML());
		steam_vars.put("topic_info_html", topicInfo.toHTML());
		
		insertTemplate("TOPIC", resp, steam_vars);
		logEvent("Community", "TOPIC+" + catId + "+" + title, 
			 "Topic added successfully");
		return;
	    }
	    else
		errors.append("either it is blank or identical to a " +
			      "previously existing topic.  Please " +
			      "enter another name.");
	}

	// we didn't make it through the if stmt's above- must be an error
        errors.insert(0, "<FONT SIZE=3>The topic you entered could not be " +
		      "added:</FONT><BR> "); 
	steam_vars.put("error_msg", errors.toString());
	
	showCategory();
	logEvent("Community", "TOPIC+" + catId + "+" + title, 
		 "add Topic failed: " + errors);
    }



    /** This method handles adding a post to a topic.  The topic to add the
     * This will add a post using the form data passed to it, and then display
     * The topic to add th the Topics page, displaying the post that was just 
     * added. post under is specified by the url token that follows the POST 
     * token. The last post that was displayed on the page is the last URL 
     * token. This allows us to know what the last post was before the user 
     * began to write the reply...
     * <P> Puts the following subs into the environment:<BR><PRE>
     *     These are added only if there was an error adding the post:
     *     error_msg       an error explaining why post could not be added
     *     post_title      title entered for the failed post
     *     post_body       body entered for the failed post
     *     firstname       first name entered the for the failed post
     *     lastname        last name entered for the failed post
     *     email           email entered for the failed post
     *     num_posts_to_display  the number of posts to display on the 
     *                           topics page
     * </PRE>
     */
    protected void addPost() throws IOException
    {
	if (req.getMethod().equals("GET")) {  // somehow they got the wrong url
	    showTopic();
	    return;
	}

	int topicId = -1;
	try
	{
	    topicId = Integer.parseInt(url_tokens[2]);
	}
	catch (NumberFormatException nfe)
	{
	    // log an error and send the user back to the default page
	    log (0, "ERROR: Malformed topic id: " + url_tokens[2]);
	    showDefault();
	}

	// Check to see if the user has permission to read this topic:
	try
	{
	    int acl_id = topicId +
		DiscussionAdminHandler.TOPIC_TO_ACL_FILTER;
	    if(!ACLManager.permit(acl_id, DiscussionAdminHandler.COMM_READ,
				  user.getUserid(), conn))
	    {
		// log a warning and send the user back to the default page
		log (0, "WARNING: DiscussionHandler: addPost: " +
			"Userid " + user.getUserid() + " tried " + 
			"accessing a topic that he/she does not have " +
			"read permission for: topic id: " + topicId);
		showDefault();
	    }
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_acl_for_topic");
	    return;
	}	
	
	// Now get the topic info to check access control to the category:
	SortableTable topicInfo =
	    lib.getTopicInfo(conn, topicId);

	if (topicInfo != null)
	{
	    int catId = ((Number) topicInfo.get(0, 1)).intValue();

	    // Check to see if the user has permission to read this category:
	    try
	    {
		if(!ACLManager.permit(catId, DiscussionAdminHandler.COMM_READ,
				      user.getUserid(), conn))
		{
		    // log a warning and send the user back to the default page
		    log (0, "WARNING: DiscussionHandler: addPost: " +
			    "Userid " + user.getUserid() + " tried " +
			    "accessing a category that he/she does not have " +
			    " read permission for: cat id: " + catId);
		    showDefault();
		}
	    }
	    catch(SQLException se)
	    {
		handleException (se, "discuss_get_acl_for_category");
		return;
	    }	
	}

	//get the goods from the form_hash and send them off....
	String title = 
	    form_data.getTrimmedString("post_title", "title", 1, 150);
	String post_body = 
	    form_data.getTrimmedString("post_body", "body", 1, 100000);
	String firstname = 
	    form_data.getTrimmedString("firstname", "first name", 0, 100);
	String lastname = 
	    form_data.getTrimmedString("lastname", "last name", 0, 100);
	String email = 
	    form_data.getEmailAddress("email", "email address", 1, 100);
	
	if (form_data.allErrorMsgs().length() < 1 &&
	    lib.addPost(conn, topicId, title,
			post_body, firstname, lastname, email) )
	{
	    // we just got a successful post from the new topics page- 
	    // make the topic live...
	    if (! topicInfo.get(0, 5).toString().equals("LIVE"))
	    {
		lib.changeTopicStatus(conn, topicId, "LIVE");
	    }

	    showTopic();
	    logEvent("Communty", "POST+"+ topicId, "Successful post");
	}
	else
	{
	    // show the same page, with an error message, and the
	    // info that's been filled out
	    String errors ="<FONT SIZE=3>The topic you entered could not " +
		"be added:</FONT><BR> " + form_data.allErrorMsgs(); 
	    steam_vars.put("error_msg", errors);
	    form_data.copyAllIntoSteam(steam_vars);

	    showTopic();
	    logEvent("Communty", "POST+"+ topicId, 
		     "Unsuccessful post: " + errors);
	}
    }
    
    /** This method shows a summary of posts for a given topic.  
     * The URL is as follows: <BLOCKQUOTE><TT>
     *     Community?SUMMARY+&lt;topicID&gt;+&lt;displayOrder&gt;
     *     </TT></BLOCKQUOTE>
     * The order token may be either DATE, POST, or AUTHOR. Thus an example 
     * summary URL might look like: Community?SUMMARY+174+AUTHOR.
     * <P> Puts the following subs into the environment:<BR><PRE>
     *     navigation            the html for the navigation sidebar
     *     post_summaries  SortableTable with columns: postid, posttitle, 
     *                             postauthorfn, postauthorln, postemail,
     *                             postdate
     *     topic_info      SortableTable with columns:  topicid, communityid, 
     *                             topicname
     *     category_info   SortableTable with columns: id, title, description, 
     *                          topic_count, post_count, last_post_date
     *     post_links      an array containing the post numbers to show at 
     *                          the bottom of the topics page when the
     *                          links for each post are followed
     *     num_posts_to_display  the number of posts to display on the 
     *                           topics page
     * </PRE>
     */
    protected void showSummary() throws IOException
    {
	int topicId = 0;
	int order = DiscussionLib.DATE;
	String orderStr = "date";
	int postIterator = numPostsToDisplay;

	try
	{
	    topicId = Integer.parseInt(url_tokens[2]);
	}
	catch (NumberFormatException nfe)
	{
	    // log an error and send the user back to the default page
	    log (0, "ERROR: Malformed topic id: " + url_tokens[2]);
	    showDefault();
	}

	// Check to see if the user has permission to read this topic:
	try
	{
	    int acl_id = topicId + DiscussionAdminHandler.TOPIC_TO_ACL_FILTER;
	    if(!ACLManager.permit(acl_id, DiscussionAdminHandler.COMM_READ,
				  user.getUserid(), conn))
	    {
		// log a warning and send the user back to the default page
		log (0, "WARNING: DiscussionHandler: showSummary: " +
			"Userid " + user.getUserid() + " tried " +
			"accessing a topic that he/she does not have " +
			"read permission for: topic id: " + topicId);
		showDefault();
	    }
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_acl_for_topic");
	    return;
	}	
	
	// this token can be either of two things: first, the number
	// of posts to show on the topics page, and second, the 
	// sort order.
	if (url_tokens.length > 3 && url_tokens[3] != null)
	{
	    try
	    {
		postIterator = Integer.parseInt(url_tokens[3]);
		if (url_tokens.length > 4 && url_tokens[4] != null)
		{
		    orderStr = (String) url_tokens[4];
		}
	    }
	    catch (NumberFormatException nfe)
	    {
		// use the default postIterator set in showDefaults
		orderStr = (String) url_tokens[3];
	    }

	    if (orderStr.equalsIgnoreCase("author"))
		order = DiscussionLib.AUTHOR;
	    else if (orderStr.equalsIgnoreCase("title"))
		order = DiscussionLib.TITLE;
	}

        // we need three tables here: 1 the topic's name etc,
	// 2 the list of posts, and 3 the category name
	SortableTable topicInfo = 
	    lib.getTopicInfo(conn, topicId);
	SortableTable postSummaries = 
	    lib.getPostSummaries(conn, topicId, order);

	// topicInfo.get(0, 1) should return the category id of the topic's 
	// parent
	int catId = Integer.parseInt( topicInfo.get(0, 1).toString());

	// Check to see if the user has permission to read this category:
	try
	{
	    if(!ACLManager.permit(catId, DiscussionAdminHandler.COMM_READ,
				  user.getUserid(), conn))
	    {
		// log a warning and send the user back to the default page
		log (0, "WARNING: DiscussionHandler: showSummary: " +
			"Userid " + user.getUserid() + " tried " +
			"accessing a category that he/she does not have " +
			" read permission for: cat id: " + catId);
		showDefault();
	    }
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_acl_for_category");
	    return;
	}	
	
	SortableTable categoryInfo = 
	    lib.getCategoryInfo(conn, catId);
	
	// add the topic table to the template
	steam_vars.put("topic_info", topicInfo);
	steam_vars.put("post_summaries", postSummaries);
	steam_vars.put("category_info", categoryInfo);
	steam_vars.put("sort_order", orderStr);

	// for the urls
	int len = postSummaries.getNumRows();
	int topLink = 0;
	for (int i = 0; i < len; i++) {
	    int temp =  ((Number) postSummaries.get(i, 0)).intValue();
	    if ( temp > topLink)
		topLink = temp;
	}

	Object [] links = new Object [len];
	for (int i = 0; i < links.length; i++)
	{
	    if (i + postIterator >= len)
		links [i] = new Integer(topLink);
	    else
		links [i] = postSummaries.get(i+postIterator-1, 0);
	}

	steam_vars.put("post_links", links);
	steam_vars.put("post_iterator", new Integer(postIterator));

	// for debugging
	steam_vars.put("category_info_html", categoryInfo.toHTML());
	steam_vars.put("topic_info_html", topicInfo.toHTML());
	steam_vars.put("post_summaries_html", postSummaries.toHTML());

	insertTemplate("SUMMARY", resp, steam_vars);
	logEvent("Community", "SUMMARY+" + topicId + "+" + orderStr, "");
    }


    /** Calculate and generate the post links for a topic page */
    private void generatePostLinks(int topicId, int postIterator, int postLimit)
    {
	// do the links here
	// this should give us the lowest number post presently displayed
	SortableTable linkTable = lib.getPostIds(conn, topicId);
	
	try
	{
	    steam_vars.put("link_table", linkTable.toHTML());
	    
	    int numRows = linkTable.getNumRows();
	    
	    int len = numRows/postIterator+1;
	    if (numRows% postIterator == 0)
		len --;
	    
	    String [] links = new String [len];
	    
	    if (postLimit == -1)
	    {
		postLimit = ((Number) linkTable.get(0, 0)).intValue();
	    }
		
	    // the lovely loop for the links to the other post pages
	    for (int i = 0; i < links.length; i++)
	    {
		int lastPost = 
		    ((Number) linkTable.get(i*postIterator, 0)).intValue();
		
		int temp = (i+1)*(postIterator) - 1;
		int firstPost = 0;
		if (temp > 1 && temp < numRows)
		{
		    firstPost = 
			((Number) linkTable.get(temp, 0)).intValue();
		}
		else
		{
		    firstPost = ((Number) linkTable.get(numRows - 1,
							0)).intValue();
		}
		
//              log (9, "***** [" + i + "] numrows: " + numRows + 
//                      " last post: " + lastPost + 
//                      " firstPost " + firstPost + " post limit " + postLimit
//                      + " it " + postIterator);
		
		links[links.length - i - 1] = url_tokens[0] + "?CMTOP+" + 
		    topicId + "+" + postIterator + "+" + lastPost;
		
		if (lastPost >= postLimit && firstPost < postLimit)
		{
//                  log (9, "***** present page: " + (links.length - i));
		    steam_vars.put("present_page", 
				   String.valueOf(links.length - i));
		}
	    }           
	    
	    steam_vars.put("post_links", links);
	    steam_vars.put("post_links_length", new Integer(links.length));
	}
	catch (NullPointerException npe) {}  // if linkTable was null

    }


    /** evaluate a template using the template path from the 
     * template hash and flush the template to our response object.
     * @param key the key to use to get the template path from the hash
     */
    protected static void insertTemplate(String key, HttpServletResponse resp,
					 Hashtable steam_vars)
	throws java.io.IOException
    {
	String template = (String) templateHash.get(key);
	Loco.log (9, "Community: inserting template: " + template + 
		  " key: " + key);

	if (template != null)
	{
	    String result = Loco.evaluateTemplate(template, steam_vars);
	    if (result.length() < 1)
	    {
		Loco.log(0, "ERROR: template at " + template + 
			 " does not exist!");
	    }
	    else
	    {
		PrintWriter pw = resp.getWriter();
		pw.print(result);
		pw.flush();
	    }
	}
    }

    /**
     * Permission prune the categories table, removing the categories
     * for which this user does not have read permission.
     * This is done by starting with a blank version of the table
     * and adding the rows of only permissable categories.
     *
     * @param categoriesTable SortableTable containing the categories
     */
    protected SortableTable permPruneCategoriesTable(SortableTable
						     categoriesTable)
    {
	if(categoriesTable == null)
	{
	    return(null);
	}
	
	SortableTable prunedTable = new SortableTable(categoriesTable, 0, 0);

	int pruned_row_num = 0;

	try
	{
	    for(int i = 0; i < categoriesTable.getNumRows(); i++)
	    {
//		int catId = (new Integer(
//		    ((Object)categoriesTable.get(
//			i, 0)).toString())).intValue();
		
		int catId = Integer.parseInt(categoriesTable.get(
		    i, 0).toString());

		if(ACLManager.permit(catId, DiscussionAdminHandler.COMM_READ,
				     user.getUserid(), conn))
		{
		    for(int j = 0; j < categoriesTable.getNumColumns(); j++)
		    {
			prunedTable.insert(categoriesTable.get(i, j),
					   pruned_row_num, j);
		    }
		    pruned_row_num++;
		}
	    }
	}
	catch(SQLException se)
	{
	    handleException(se, "discuss_get_cat_acl");
	    return(null); 
	}

	return(prunedTable);
    }

    protected SortableTable permPruneTopicsTable(SortableTable topicsTable)
    {
	if(topicsTable == null)
	{
	    return(null);
	}
	
	SortableTable prunedTable = new SortableTable(topicsTable, 0, 0);

	int pruned_row_num = 0;

	try
	{
	    for(int i = 0; i < topicsTable.getNumRows(); i++)
	    {
//		int topicId = (new Integer(
//		    ((Object)topicsTable.get(
//			i, 0)).toString())).intValue();

		int topicId = Integer.parseInt(topicsTable.get(
		    i, 0).toString());

		// Get the acl topic id by using the filter:
		topicId += DiscussionAdminHandler.TOPIC_TO_ACL_FILTER;

		if(ACLManager.permit(topicId, DiscussionAdminHandler.COMM_READ,
				     user.getUserid(), conn))
		{
		    for(int j = 0; j < topicsTable.getNumColumns(); j++)
		    {
			prunedTable.insert(topicsTable.get(i, j),
					   pruned_row_num, j);
		    }
		    pruned_row_num++;
		}
	    }
	}
	catch(SQLException se)
	{
	    handleException (se, "discuss_get_top_acl");
	    return(null); 
	}

	return(prunedTable);
    }

    
    /**
     * The required init method, as defined in org.locomotive.loco.Handler.  In
     * this example, we load all the templates into a hashtable for 
     * easy access.
     */
    public void init (ServletConfig config) throws ServletException
    {
	String s = (String) config.getInitParameter("num.posts.to.display");
	if (s != null) try
	{
	    numPostsToDisplay = Integer.parseInt(s);
	}
	catch (NumberFormatException nfe)
	{
	    throw new ServletException(
		"ERROR: malformed config property: " + s + 
		"num.posts.to.display must be a number!");
	}

	Connection conn = Loco.getDBConnection();

	try
	{
	    lib = new DiscussionLib(conn);
	}
	catch (SQLException se)
	{
	    throw new ServletException(
		"ERROR: could not initialize DiscussionLib: " + se);
	}
	finally
	{
	    if (conn != null) Loco.putBackDBConnection(conn);
	}

	admin = new DiscussionAdminServlet();
	admin.init(config);

	Loco.log(9, "DiscussionHandler: creating template hash...");

	String path = "discuss/community/";
	String admin_path = "discuss/admin/commmgr/";
	templateHash.put("DEFAULT", path + "comm_section.tmpl");
	templateHash.put("CMCAT", path + "comm_disc.tmpl");
	templateHash.put("CMTOP", path + "comm_disc_topic.tmpl");
	templateHash.put("TOPIC", path + "comm_disc_topic_new.tmpl");
	templateHash.put("SUMMARY", path + "comm_disc_topic_sum.tmpl");
	
	// here's the admin part
	templateHash.put("Admin", admin_path + "comm_disc_list.tmpl");
	templateHash.put("Admin+MOD", admin_path + 
			 "comm_disc_filledtempl.tmpl");
	templateHash.put("Admin+TOP", admin_path + "comm_topics_list.tmpl");
	templateHash.put("Admin+TOP+MOD", admin_path + 
			 "comm_topics_filledtempl.tmpl");
	templateHash.put("Admin+POST", admin_path + "comm_posts_list.tmpl");
	templateHash.put("Admin+POST+MOD", admin_path + 
			 "comm_posts_filledtempl.tmpl");
	templateHash.put("Admin+USER", admin_path + "comm_disc_users.tmpl");
	templateHash.put("Admin+TOP+USER", admin_path + 
			 "comm_topic_users.tmpl");
    }


    /** Makes sure the Admin servlet gets destroyed too. */
    public void destroy()
    {
	admin.destroy();
    }


}








