/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.display;
import org.locomotive.util.*;
import org.locomotive.server.*;
import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;

/**
 * DisplayPageHandler can be used to map a url into the path of a HTML 
 * template file in the file system.  This gives the Handler implementer
 * an easy way to evaluate a hierarchy of templates with a given 
 * default or &quot;generic&quot; environment.  
 * For instance, this class would
 * map urls beginning with &quot;disp?base+generic&quot; to the template root
 * sub directory &quot;base/generic&quot;  Therefore, <BR>
 * <PRE> 
 * disp?base+main          maps to $TEMPLATE_ROOT/base/main.tmpl
 * disp?base+generic+page1 maps to $TEMPLATE_ROOT/base/generic/page1.tmpl
 * disp?base+generic+page2 maps to $TEMPLATE_ROOT/base/generic/page2.tmpl
 * </PRE> 
 *
 * DisplayPageHandler will also log the Event of the generic template display, 
 * and neutralize any file system separators/operators from the url_tokens
 * list for security purposes.  In other words, all non-alphanumeric letters
 * in the URL will be converted into a '/', meaning that each alphanumeric
 * word will become a directory name except for the last, which becomes the
 * name of the file. <BR>
 * In order to use this Handler, you have to have an entry for this class
 * in your hrt.conf config file, ex:
 * <BLOCKQUOTE>disp          org.locomotive.las.DisplayPageHandler</BLOCKQUOTE>
 * or, alternatively, you can invoke this Handler from another Handler:
 * <BLOCKQUOTE><PRE>
 * DisplayPageHandler disp = new DisplayPageHandler();
 * disp.handleRequest(hd);
 * </PRE></BLOCKQUOTE>
 */
public class DisplayPageHandler extends GenericHandler
{
    /**
     * DisplayPageHandler doesn't do anything in the init() method.
     */
    public boolean init (Log server_log)
    {
	return true;
    }
    
    
    /**
     * DisplayPageHandler doesn't do anything in the shutdown() method, either.
     */
    public boolean shutdown (Log server_log)
    {
	return true;
    }
    
    
    /**
     * main entry point called by RequestManager for each new request.
     */
    public void handleRequest (HandlerData hd)
    {
        // call GenericHandler's handleRequest(hd) method,
        // to set up local variables
        super.handleRequest(hd);

        // insert some standard Steam variables, such as date and time.
        StandardInsertions.insertAll (hd);
    
        // if the URI is not null or empty, and it has only alphanumeric 
        // characters or '?' or '+', then try to convert it into a file
        // path.
        if (handler_uri != null && !handler_uri.equals(""))
        {
	    // NOTE: handler_uri, like a standard URI, always begins with
	    // a leading '/', so we strip it off with substring(1) first.
            String convert_url_to_alphanum =
                handler_uri.substring(1).replace ('?', 'a').replace ('+', 'a');
    
            if (StringUtil.isAlphanum (convert_url_to_alphanum))
            {
                // convert "base?login+form" to "base/login/form.tmpl",
                // then display that file as a HTML template.
                String result = displayPage (
		    handler_uri.substring(1).replace('?', '/').replace(
		    '+', '/') + ".tmpl",
                    steam_vars);
    
                // if there is no file in that file path, then 
                // displayPage() will return an empty string "", 
                // in which case we should
                // tell the user that the URL was faulty.

                if (result.equals(""))
                    sayBadURL();
            }
            else
            {
                sayBadURL();
            }
        }
        else
        {
            // send the user an error message saying that it was a bad URL
            sayBadURL();
        }
    }

}
