/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.graffiti;

import java.sql.*;
import java.util.*;
import org.locomotive.loco.*;  // error logging
import org.locomotive.util.*;  

/**
 *  Tag - an individual graffiti message. <BR>
 *	  Its fields mirror the columns in the db table, loco_graffiti_tags.
 *        You can store a Tag by overwriting the oldest row in the table.
 *	  This class also contains static methods for db manipulation.
 */
public class Tag  
{
    private String author;
    private Timestamp insTime;
    private String text;
    static int maxRows = 10;		 // max # of rows kept in db  
     
    private static LocoConfig config = Loco.getConfig();
    private static String DB_CURR_TIME = null;
    
    
    private Tag (String author, Timestamp insTime, String text) 
    {
	this.author = author;
	this.insTime = insTime;
	this.text = text;
    }
    
    /**
     * saves this Tag into the db table loco_graffiti_rows.
     * the db should only contain maxRows so it will delete
     * older ones if they exist.
     */
    public Tag (Connection conn, String author, String text) 
	throws SQLException
    {
	this.author = author;
	this.text = text;
	
	if (this.text == null || this.text.equals (""))
	{
	    this.text = " ";
	}
	
	overwriteRow(conn);
	
	// might not be the exact same as the date from the DB, but 
	// it will be close enough for government work.
	this.insTime = new Timestamp(System.currentTimeMillis());
    }

    /**
     * returns the Date the Tag was created.
     */
    public Timestamp getInsTime() { return insTime; }

    /**
     * returns the author of the Tag
     */
    public String getAuthor() { return author; }

    /**
     * returns the text of the Tag.
     */
    public String getText() { return text; }
    
    /**
     * saves Tag t as a new row in the db table loco_graffiti_tags.
     * overwrites the oldest Tag by using a subquery, "select min(tag_id)".
     * (!! NOTE: assumes that loco_graffiti_tags already has maxRows number
     * of rows in it!  Be sure to call Tags.prepopulate() somewhere (for
     * example, in GraffitiWallHandler.init()) to be sure.)
     */
    private void overwriteRow(Connection conn) 
        throws SQLException
    {
	String statement = null;
	Statement stmt = null;

	switch (config.DB_TYPE)
	{
	case SQL.DB_TYPE_ORACLE:
	    statement = "update loco_graffiti_tags " + 
		"set tag_id = loco_tag_id_sq.nextval, " +
		" tag_author = " + SQL.quoteString (author) +
		", tag_text = " + SQL.quoteString(text) + 
		", tag_ins_time = SYSDATE where " + 
		"tag_id = (select min(tag_id) from loco_graffiti_tags)";
	    break;

	    // Believe it or not, we are actually not in favor of using
	    // stored procedures for these types of operations.  However, 
	    // solid, informix, and cloudscape won't let you access a sequence 
	    // without being in a stored proc.  It's possible that we're
	    // missing the point big-time, but it seems rather silly.  
	    // Every time we add support for a new DB, we have to write 
	    // a new procedure in the proprietary language of the DB.   
	    // Annoying. 
	case SQL.DB_TYPE_INFORMIX:
	    statement = "execute procedure loco_ow_graf_tag ( " +
		SQL.quoteString (author) + ", " + SQL.quoteString(text) +
		" )";
	    break;

	case SQL.DB_TYPE_SOLID:
	    statement = "call loco_overwrite_graffiti_tag ( " +
		SQL.quoteString (author) + ", " + SQL.quoteString(text) +
		" )";
	    break;

	case SQL.DB_TYPE_CLOUDSCAPE:
	    statement = "update loco_graffiti_tags " +
		"set tag_id = (CLASS org.locomotive.loco.Cloudscape)." +
		"getNextGraffitiTagID( GETCURRENTCONNECTION()) " +
		", tag_author = " + SQL.quoteString (author) +
		", tag_text = " + SQL.quoteString(text) + 
		", tag_ins_time = CURRENT_TIMESTAMP where " + 
		"tag_id = (select min(tag_id) from loco_graffiti_tags)";
	    break;
	case SQL.DB_TYPE_MYSQL:
	    // now this is in serious need of a stored procedure.  4
	    // queries for this little transaction!

	    statement = "unlock tables"; // we'll do this last
	    ResultSet rs = null;

	    try
	    {
		stmt = conn.createStatement();
		stmt.executeUpdate("lock tables loco_graffiti_tags write");
		String query = "select min(tag_id) from loco_graffiti_tags";
		rs = stmt.executeQuery(query);
		rs.next();
		int min_id = rs.getInt(1);

		query = "delete from loco_graffiti_tags where " +
		    "tag_id = " + min_id;
		stmt.executeUpdate(query);
		query = "insert into loco_graffiti_tags (tag_id, tag_author, " +
		    "tag_text, tag_ins_time) values (0, " + 
		    SQL.quoteString (author) + ", " + SQL.quoteString(text) + 
		    ", SYSDATE())";
		stmt.executeUpdate(query);
	    }
	    catch (SQLException e)
	    {
		// now, let's try to unlock the table
		stmt = conn.createStatement();
		stmt.executeUpdate(statement);
	    }
	    finally
	    {
		if (rs != null) rs.close();
		if (stmt != null) stmt.close();
	    }

	    break;
	}

	try
	{
	    stmt = conn.createStatement();
	    stmt.executeUpdate(statement);
	}
	finally
	{
	    if (stmt != null) stmt.close();
	}
    }


    /**
     * loads in all rows from the db, sorted by date in desc order.
     * if there are no rows in the table, an array of zero length 
     * (Tag[0]) is returned.
     */
    public static Tag[] loadAllTags (Connection conn) 
	throws SQLException
    {
	String statement = "select tag_author, tag_ins_time, tag_text from " +
	    "loco_graffiti_tags order by tag_ins_time desc";
	
	Statement stmt = null;
	ResultSet rs = null;
	Vector v = new Vector();
	
	try {
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery (statement);
	    
	    while (rs.next())
	    {
		// uses private constructor
		v.addElement (new Tag (rs.getString (1), rs.getTimestamp (2),
				       rs.getString (3)));
	    }
	}
	finally
	{
	    if (rs != null)  rs.close();
	    if (stmt != null)  stmt.close();
	}
	
	Tag[] tags = new Tag[v.size()];
	v.copyInto (tags);

	return (tags);
    }
    

    /** 
     * This is called from the init() method of the graffitiWallHandler to 
     * make sure the graffitiWall has a few entries to start out
     * with.  Will insert the following line up to ten times in
     * the graffiti wall:  'this is an example tag' from 'anon'
     * @see org.locomotive.module.graffiti.GraffitiWallHandler#init
     */
    public static void prepopulate (Connection conn)
	throws SQLException
    {
	String statement = 
	    "select count(*), max(tag_id) from loco_graffiti_tags";
	Statement stmt = null;
	ResultSet rs = null;
	PreparedStatement pstmt = null;
	
	try 
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery (statement);
	    
	    if (! rs.next())
	    {
		throw new SQLException ("count(*) loco_graffiti_tags failed!");
	    }
	    
	    int row_count = rs.getInt(1);
	    int max_tag_id = rs.getInt(2);
	    
	    if (row_count < maxRows)
	    {
		statement = null;

		// set the statement
		switch (config.DB_TYPE)
		{
		case SQL.DB_TYPE_ORACLE:
		    statement = "insert into loco_graffiti_tags values " +
			" (loco_tag_id_sq.nextval, 'anon', " + 
			"'This is an example tag', ?)";
		case SQL.DB_TYPE_INFORMIX:
		case SQL.DB_TYPE_SOLID:
		    statement = "insert into loco_graffiti_tags " +
			" (tag_id, tag_author, tag_text, tag_ins_time) " +
			" values (?, 'anon', 'This is an example tag', ?)";
		    break;
		case SQL.DB_TYPE_CLOUDSCAPE:
		    statement = "insert into loco_graffiti_tags values (" +
			"(CLASS org.locomotive.loco.Cloudscape)." +
			"getNextGraffitiTagID( GETCURRENTCONNECTION()), " +
			" 'anon', 'This is an example tag', ?)";
		    break;
		case SQL.DB_TYPE_MYSQL:
		    statement = "insert into loco_graffiti_tags " +
			" (tag_id, tag_author, tag_text, tag_ins_time) " +
			" values (0, 'anon', 'This is an example tag', ?)";
		    break;
		}
		
		pstmt = conn.prepareStatement(statement);
		Timestamp t = new Timestamp(System.currentTimeMillis());
		for (int i = 0; i < (maxRows - row_count); i++)
		{
		    // a better alternative to using setTimestamp() is to use
		    // SYSDATE (Oracle) or CURRENT (Informix), for the current
		    // date and time.  Here, we're using setTimestamp() just
		    // to give you an example of how to set a value in a 
		    // PreparedStatement.
		    switch (config.DB_TYPE)
		    {
		    case SQL.DB_TYPE_ORACLE:
		    case SQL.DB_TYPE_CLOUDSCAPE:
		    case SQL.DB_TYPE_MYSQL:
			pstmt.setTimestamp(1, t);
			pstmt.executeUpdate();
			break;

		    case SQL.DB_TYPE_INFORMIX:
		    case SQL.DB_TYPE_SOLID:
			max_tag_id++;
			pstmt.setInt(1, max_tag_id);
			pstmt.setTimestamp(2, t);
			pstmt.executeUpdate();
			break;
		    }
		}
	    }
	}
	finally
	{
	    if (rs != null)  rs.close();
	    if (stmt != null)  stmt.close();
	    if (pstmt != null)  pstmt.close();
	}
    }
}
