/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.pub;

import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;
import org.locomotive.loco.perm.*;
import org.locomotive.server.Log;
import org.locomotive.util.*;
import java.sql.*;
import java.util.*;
import java.lang.*;
import java.io.*;


/**
 * This class is the main handler for the Locomotive Personal Website
 * Publisher.  This example module of the Locomotive allows users to create
 * and design their own website by filling out text fields, inserting
 * images, and arranging all of the HTML elements in a pre-arranged
 * lay-out determined by the template selected by the user.  The Locomotive
 * stores all of the necessary information (text strings, images, which
 * template is used, etc.) and automatically dynamically creates the page
 * for viewing when the handler string is parsed to determine the
 * specific user's page.  This module also displays an index of published
 * webpages, beginning with the most recent.
 *
 * NOTE: The image path (image_path) variable below must be modified
 * 
 * All templates for the Publisher Module are in the templates/pub dir.
 * The tokens and their corresponding methods are:
 * <BR><PRE>
 * Token          Method              Template
 * [no token]     showMain()          pub/main.tmpl
 * index          showDefaultIndex()  pub/index.tmpl
 * approve        showApprove()       pub/approve.tmpl
 * censor         showCensor()        pub/censor.tmpl
 * create         showCreate()        pub/create.tmpl
 * preview        showPreview()       pub/preview.tmpl 
 * final          showFinal()         pub/final.tmpl
 * </PRE>
 */

public class PublisherHandler extends GenericHandler 
{
    private PermissionSet pSet = null;


    /**
     * This should get set in this Handler's properties file- see
     * the init() method.  Get's put into the subs for every hit.
     */
    private static String document_root = "/usr/local/web/htdocs";

    /**
     * This should get set in this Handler's properties file- see
     * the init() method.  Get's put into the subs for every hit.
     */
    private static String image_path = "/images/pub";

    /**
     * The version string for the entire Publisher system - this
     * gets inserted into the client's environment every time a hit comes
     * in to this handler.
     */
    private static String version_string = "1.0";

    /**
     * Constant used to determine how many page listings to show per index
     * page.
     */
    public static final int num_listings_per_index_page = 10;

    /**
     * Permission that allows a user to allow or censor newly user-created
     * web pages.
     */
    public static final String PUB_ADMIN = "PUB_ADMIN";

    /**
     * main entry point called by RequestManager for each new request.
     */
    public void handleRequest (HandlerData hd_in)
    {
	super.handleRequest(hd_in);
	hd = hd_in;

	// call the Loco library call for standard insertions.  
	StandardInsertions.insertAll (hd);

	steam_vars.put("PUB_VERSION", version_string);
	steam_vars.put("document_root", document_root);
	steam_vars.put("image_path", image_path);
	
	// parse url tokens for proper function
	if (url_tokens.length < 2)
	{
	    showMain();
	}
	else if ( url_tokens[1].equals("index"))
	{
	    showDefaultIndex();
	}
	else if ( url_tokens[1].equals("own"))
	{
	    showOwn();
	}
	else if ( url_tokens[1].equals("approve"))
	{
	    showApprove();
	}
	else if ( url_tokens[1].equals("censor"))
	{
	    showCensor();
	}
	else if (url_tokens[1].equals ("create"))
	{
	    showCreate();
	}
	else if (url_tokens[1].equals ("preview"))
	{
	    showPreview();
	}
	else if (url_tokens[1].equals ("final"))
	{
	    showFinal();
	}
	else
	{
	    // Unidentified handler
	    displayPage ("error/unknown_handler_option.tmpl", steam_vars);
	}
    }

    /**
     * Displays the main start page for this module.  If the user has
     * the PUB_ADMIN permission, then they are given links to the approval
     * and censored index pages.
     */
    protected void showMain()
    {
	getPermissions();

	if(pSet.isSet(PUB_ADMIN))
	{
	    String links = "<P>\n" +
		"Admin Links:<BR>\n" +
		"<A HREF=\"pub?approve\">View pages awaiting approval.</A>\n" +
		"<BR>\n" +
		"<A HREF=\"pub?censor\">View censored pages.</A>\n" +
		"<BR>\n";

	    steam_vars.put("ADMIN", links);
	}

	displayPage ("pub/main.tmpl", steam_vars);
    }

    /**
     * Displays a page showing a list of published pages.  The number token
     * represents which range of pages to display.  No number token will
     * simply default to the first page of indexes.
     * @param status status of pages we want to see an index for
     * @return true if no error occurs and the index is not empty
     */
    protected boolean showIndex(int status)
    {
	int index_page = 1;

	if (url_tokens.length > 2)
	{
	    index_page = Integer.parseInt(url_tokens[2]);
	}

	int low_range = (index_page-1)*num_listings_per_index_page + 1;
	int high_range = index_page*num_listings_per_index_page;

	int maxPages = PublisherLib.getNumPages(conn, status);
	if(low_range > maxPages)
	{
	    // Out of range- return false
	    return(false);
	}

	if(high_range > maxPages)
	{
	    high_range = maxPages;
	}
	
	SortableTable index_table = PublisherLib.getIndexTable(
	    conn, low_range, high_range, status);

	if(index_table == null)
	{
	    // No pages in index - return false
	    return(false);
	}

	int max_index_pages = maxPages/num_listings_per_index_page;
	if(maxPages%num_listings_per_index_page != 0)
	{
	    max_index_pages++;
	}
	
	// The page number of this index:
	steam_vars.put("index_page", String.valueOf(index_page));
	// The maximum number of all index pages:
	steam_vars.put("max_index_pages", String.valueOf(max_index_pages));
	// The index number of the first published page in this index:
	steam_vars.put("start_index",  String.valueOf(low_range));
	// The actual table object:
	steam_vars.put("index_table", index_table);

	// For debugging
	steam_vars.put("index_table_html", index_table.toHTML());

	// Everything successful
	return(true);
    }

    /**
     * Shows the index of published pages owned by a user.
     */
    protected void showOwn()
    {
	if(user.isAnonymous())
	    showDefaultIndex();

	SortableTable table = PublisherLib.getUserPageIndex(
	    conn, user.getUserid());
	
	if(table == null)
	{
	    showDefaultIndex();
	}

	// The actual table object:
	steam_vars.put("index_table", table);

	// For debugging
	steam_vars.put("index_table_html", table.toHTML());	 

	displayPage ("pub/own_index.tmpl", steam_vars);
    }

    /**
     * Shows the index of published pages that are not-yet-approved and 
     * approved.  This is to be viewed by all.
     */
    protected void showDefaultIndex()
    {
	if (showIndex(PublisherLib.NYA_AND_APPR))
	{
	    // See if they have pages so that we can give them a link to
	    // an index of their own pages
	    if (!user.isAnonymous())
	    {
		SortableTable table = PublisherLib.getUserPageIndex(
		    conn, user.getUserid());
		
		if(table != null)
		{
		    steam_vars.put("own_index", "true");
		}
		
	    }

	    // Flush the index template
	    displayPage ("pub/index.tmpl", steam_vars);
	}
	else
	{
	    // Just return them to the main publisher page
	    displayPage ("pub/main.tmpl", steam_vars);
	}

    }

    /**
     * Displays a page showing the list of not-yet-approved published pages.
     * User must have the "PUB_ADMIN" permission to see this index.
     * Token number decides which page of the index to display.
     */
    protected void showApprove()
    {
	getPermissions();

	// See if the user has permission to be here
	if (!pSet.isSet(PUB_ADMIN))
	{
	    // Does not have permission.  Show them the unknown handler
	    // error page.
	    displayPage ("error/unknown_handler_option.tmpl", steam_vars);
	    return;
	}
	
	if (hd.form_hash != null)
	{
	    // Evaluate the user submitted form
	    Enumeration e = hd.form_hash.keys();
	    while(e.hasMoreElements())
	    {
		try
		{
		    String key = e.nextElement().toString();
		    int page_id = Integer.parseInt(key);

		    String status = (hd.form_hash.get(key)).toString();

		    int status_int = PublisherLib.NOT_YET_APPROVED;
		    if(status.equals("APPROVED"))
		    {
			status_int = PublisherLib.APPROVED;
		    }
		    else if(status.equals("CENSORED"))
		    {
			status_int = PublisherLib.CENSORED;
		    }

		    PublisherLib.changeStatus(conn, page_id, status_int);
		}
		catch (NumberFormatException nfe)
		{
		    Loco.log(Log.FATAL_ERROR, 
			     "Error in PublisherHandler: " +
			     StringUtil.exceptionStackTrace(nfe));
		    log (0, "Error in PublisherHandler: " +
			    StringUtil.exceptionStackTrace(nfe));
		}
	    }
	}

	if (showIndex(PublisherLib.NOT_YET_APPROVED))
	{
	    displayPage ("pub/approve.tmpl", steam_vars);
	}
	else
	{
	    steam_vars.put("MESSAGE", "<B>There are currently no pages to " +
			"review.</B>\n");

	    // index is empty...
	    displayPage ("pub/generic.tmpl", steam_vars);
	}
    }

    /**
     * Displays a page showing the list of censored published pages.
     * User must have the "PUB_ADMIN" permission to see this index.
     * Token number decides which page of the index to display.
     */
    protected void showCensor()
    {
	getPermissions();

	// See if the user has permission to be here
	if (!pSet.isSet(PUB_ADMIN))
	{
	    // Does not have permission.  Show them the unknown handler
	    // error page.
	    displayPage ("error/unknown_handler_option.tmpl", steam_vars);
	    return;
	}

	if (hd.form_hash != null)
	{
	    // Evaluate the user submitted form
	    Enumeration e = hd.form_hash.keys();
	    while(e.hasMoreElements())
	    {
		try
		{
		    String key = e.nextElement().toString();
		    int page_id = Integer.parseInt(key);

		    String status = (hd.form_hash.get(key)).toString();

		    if(status.equals("UNCENSOR"))
		    {
			PublisherLib.changeStatus(
			    conn, page_id, PublisherLib.NOT_YET_APPROVED);
		    }
		}
		catch (NumberFormatException nfe)
		{
		    Loco.log(Log.FATAL_ERROR, 
			     "Error in PublisherHandler: " +
			     StringUtil.exceptionStackTrace(nfe));
		}
	    }
	}

	if (showIndex(PublisherLib.CENSORED))
	{
	    displayPage ("pub/censor.tmpl", steam_vars);
	}
	else
	{
	    steam_vars.put("MESSAGE", "<B>There are currently no censored " +
			"pages.</B>\n");

	    // index is empty...
	    displayPage ("pub/generic.tmpl", steam_vars);
	}
    }

    /**
     * Displays the page used to create a web page.
     */
    protected void showCreate()
    {
	// Check to see if we have the third token, the page id.
	// No page id means that a user is creating a page for the first time
	// or that showPreview sent us here with errors from the create form.

	if (url_tokens.length > 2)
	{
	    // Get the page id
	    int page_id = Integer.parseInt(url_tokens[2]);
	    
	    steam_vars.put("id", String.valueOf(page_id));

	    // A user is re-modifying her already published page.
	    // First we must see if this is the correct user.
	    boolean theGoAhead = false;
	    
	    if(PublisherLib.isOwner(conn, page_id, user.getUserid()))
		// this person is the owner
		theGoAhead = true;
		else
		{
		    getPermissions();
		    if(pSet.isSet(PUB_ADMIN))
			// Let admins modify the page
			theGoAhead = true;
		}
	    
	    if(theGoAhead)
	    {
		// Get the page info from the database and repopulate the
		// page form.
		PublisherLib.getPageInfo(conn, page_id, steam_vars);
		Loco.log(9, "image 1: " + steam_vars.get("image_one"));

	    }
	}

	// Flush the create template
	displayPage ("pub/create.tmpl", steam_vars);
    }

    /**
     * Displays a preview of a published page.  Essentially the same as
     * showing the final version of the page, except with the addition
     * of 'Modify' and 'Publish' buttons so that the user has the option
     * to do more editing before actually publishing the page for public
     * viewing. 
     */
    protected void showPreview()
    {
	// Check to see if we have the third token, the page id.
	// No page id means that a user is previewing a page for the first
	// time.

	int id = -1;

	if (url_tokens.length > 2)
	{
	    // Get the page id
	    id = Integer.parseInt(url_tokens[2]);
	}

	if (hd.form_hash == null)
	{
	    // Nothing for us to display
	    showCreate();	    
	}
	
	String title = form_data.getString("title", "title", 1, 64);

	String body_one = form_data.getString("body_one", "body one", 1, 2000);
	String body_two = form_data.getString("body_two", "body two", 1, 2000);
	
	String caption_one =
	    form_data.getString("caption_one", "caption one", 1, 128);
	String caption_two =
	    form_data.getString("caption_two", "caption two", 1, 128);
	
	String template = form_data.getString("template", "template", 1, 128);

	// Get the original image filenames
	String image_one = form_data.getString("image_one.filename",
					       "filename of image one", 1,128);
	String image_two = form_data.getString("image_two.filename",
					       "filename of image two", 1,128);

	// Get our temporary uploaded filepaths and names
	String temp_image_one = form_data.getString("image_one.path",
						    "image_one.path", 1, 128);
	String temp_image_two = form_data.getString("image_two.path",
						    "image_two.path", 1, 128);

	// Get our posted file name if there was one
	String pub_image_one =
	    form_data.getString("pub_image_one", "pub_image one", 1, 128);
	String pub_image_two =
	    form_data.getString("pub_image_two", "pub_image two", 1, 128);


	// Do error checks
	boolean no_image_one = (image_one.length() == 0);
	boolean no_image_two = (image_two.length() == 0);

	StringBuffer error_msg = new StringBuffer();
	if (title.length() == 0)
	{
	    error_msg.append("You did not enter a title.<BR>");
	}
	if (body_one.length() == 0)
	{
	    error_msg.append("You did not enter any text for body1.<BR>");
	}
	if (body_two.length() == 0)
	{
	    error_msg.append("You did not enter any text for body2.<BR>");
	}
	if (body_one.length() == 0 && pub_image_one.length() == 0)
	{
	    error_msg.append("You did not select an image file for image1." +
			     "<BR>");
	}
	if (body_two.length() == 0 && pub_image_two.length() == 0)
	{
	    error_msg.append("You did not select an image file for image2." +
			     "<BR>");
	}
	if (caption_one.length() == 0)
	{
	    error_msg.append("You did not enter any text for caption1.<BR>");
	}
	if (caption_two.length() == 0)
	{
	    error_msg.append("You did not enter any text for caption2.<BR>");
	}
	if (template.length() == 0)
	{
	    error_msg.append("You did not select a layout format.<BR>");
	}

	if (error_msg.length() > 0)
	{
	    steam_vars.put("error_msg", error_msg);

	    // Repopulate subs
	    steam_vars.put("title", title);
	    steam_vars.put("body_one", body_one);
	    steam_vars.put("body_two", body_two);
	    steam_vars.put("caption_one", caption_one);
	    steam_vars.put("caption_two", caption_two);
	    steam_vars.put("template", template);

	    if (image_one.length() > 0)
		steam_vars.put("image_one", image_one);
	    else if (pub_image_one != null)
		steam_vars.put("image_one", pub_image_one);
	    
	    if (image_two.length() > 0)
		steam_vars.put("image_two", image_two);
	    else if (pub_image_two != null)
		steam_vars.put("image_two", pub_image_two);

	    // Send them back to the create form so they can do it right
	    showCreate();
	    return;
	}
	else
	{
	    // Good entry - let's put it in the database
	    if(id == -1)
	    {
		// This is the first time (there is no id yet)
		PublisherLib.insertPage(conn,
					user.getUserid(),
					title,
					body_one,
					body_two,
					caption_one,
					caption_two,
					template);

		// Get the id from the newly created page
		id = PublisherLib.getMostRecentId(conn,
						  user.getUserid());

		// Create the filenames we will store the images as, which
		// will be in the form "[id]_image_[one|two]"
		pub_image_one = id + "_image_one";
		pub_image_two = id + "_image_two";

		// Put the image filenames into the database
		PublisherLib.insertImagesIntoPage(conn, id, pub_image_one,
						  pub_image_two);
	    }
	    else
	    {
		if (! no_image_one)  // we got an image to upload
		{
		    pub_image_one = id + "_image_one";

		    // Create file objects for the temp files we are trying 
		    // to copy
		    File f = new File(temp_image_one);
		    if (f.exists() && f.canRead() && f.length() > 1)
		    {
			File file_image_one = 
			    new File(document_root + image_path, pub_image_one);

			f.renameTo(file_image_one);
		    }
		    else
		    {
			log (Log.FATAL_ERROR, "Could not find image 1 " +
			     "file in upload: " + temp_image_one);
			displayPage ("error/generic_error.tmpl", steam_vars);
			return;
		    }
		}

		if (! no_image_two)
		{
		    pub_image_two = id + "_image_two";

		    File f = new File(temp_image_two);
		    if (f.exists() && f.canRead() && f.length() > 1)
		    {
			File file_image_two = 
			    new File(document_root + image_path, pub_image_two);

			f.renameTo(file_image_two);
		    }
		    else
		    {
			log (Log.FATAL_ERROR, "Could not find image 2 " +
			     "file in upload: " + temp_image_two);
			displayPage ("error/generic_error.tmpl", steam_vars);
			return;
		    }
		}
		
		// There is already a row entry for this page.
		// Just do an update
		PublisherLib.reInsertPage(conn,
					  id,
					  title,
					  body_one,
					  body_two,
					  pub_image_one,
					  pub_image_two,
					  caption_one,
					  caption_two,
					  template);
	    }
	    
	    // Now repopulate subs hash from the database
	    PublisherLib.getPageInfo(conn, id, steam_vars);

	    
	    if(!user.isAnonymous())
	    {
		steam_vars.put("author", user.getUsername());
	    }

	    steam_vars.put("preview", "true");
	    
	    displayPage ("pub/" + (String) steam_vars.get("template"),
			 steam_vars);
	}
    }

    /**
     * Displays the final published version of a page.  If the page is
     * censored, only users with PUB_ADMIN permission can see the
     * page.  The token indicates the id of the page to display.
     */
    protected void showFinal()
    {
	// Check to see if we have the third token, the page id.
	if (url_tokens.length < 3)
	{
	    showDefaultIndex();
	}

	int page_id = Integer.parseInt(url_tokens[2]);
	
	int status = PublisherLib.getStatus(conn, page_id);

	boolean theGoAhead = false;

	// Check the status of the page for public viewing
	if(status >= PublisherLib.NOT_YET_APPROVED)
	{
	    theGoAhead = true;
	}
	else
	{
	    getPermissions();
	    if(pSet.isSet(PUB_ADMIN))
	    {
		// Admins are allowed to see this page
		theGoAhead = true;
	    }
	}

	if(theGoAhead)
	{
	    PublisherLib.getPageInfo(conn, page_id, steam_vars);
	    displayPage ("pub/" + (String) steam_vars.get("template"),
			 steam_vars);
	}
    }

    /**
     * Prepares the PermissionSet object for permission checking.
     */
    protected void getPermissions()
    {
	try
	{
	    pSet = UserPerm.getPermissions(user.getUserid(), conn);
	}
	catch(SQLException se)
	{
	    log (Log.FATAL_ERROR,  "An error occured accessing permissions" +
		 " for Userid " + user.getUserid() + ": " +
		 StringUtil.exceptionStackTrace(se));
	    displayPage ("error/generic_error.tmpl", steam_vars);
	    return;
	}
    }


    /**
     * The required init method, as defined in org.locomotive.loco.Handler.  Here
     * we check to see if the Properties object for this Handler
     * exists, and if it does, then we get the image_path variable out of it.
     *
     */
    public boolean init (Log log)
    {
	Properties p = Loco.getHandlerProperties((Handler) this);
	if (p != null)
	{
	    if (p.containsKey("document_root"))
		document_root = (String) p.get("document_root");
	    log.log(Log.DEBUG, "Publisher: document root is at " + 
		document_root);
	    if (p.containsKey("image_path"))
		image_path = (String) p.get("image_path");
	    log.log(Log.DEBUG, "Publisher: image path is at " + 
		    image_path);
	}

	return (true);
    }

    /**
     * The required shutdown method, as defined in org.locomotive.loco.Handler.  In
     * this example, we don't actually need to do anything.
     * @see org.locomotive.loco.Handler#shutdown
     */
    public boolean shutdown (Log log)
    {
	return (true);
    }

}

