/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.pub;

import java.util.*;
import java.sql.*;
import java.io.*;
import java.text.*;
import org.locomotive.loco.*;
import org.locomotive.util.*;

/**
 * This class provides a series of static methods which return Database 
 * information used by the Locomotive Publisher system.
 */
public class PublisherLib
{
    /**
     * Constants used to denote the different statuses that a published page
     * can be in:
     */
    public static final int CENSORED = -1;
    public static final int NOT_YET_APPROVED = 0;
    public static final int APPROVED = 1;
    public static final int NYA_AND_APPR = 2; // Not Yet Approved AND Approved

    /**
     * Creates a page entry into the database for the first time.  Note that
     * images are omitted because the images need the id of this entry
     * first to be appended to their names.
     * @param conn        database connection
     * @param userid      page creator
     * @param title       title of page
     * @param body_one    first body text field
     * @param body_two    second body text field
     * @param caption_one String caption for first image
     * @param caption_two String caption for second image
     * @param template    String name of template file
     *
     * @return false if a database error occurs
     */
    public static boolean insertPage(Connection conn,
				     int userid,
				     String title,
				     String body_one,
				     String body_two,
				     String caption_one,
				     String caption_two,
				     String template)
    {
	String statement = "insert into loco_publisher " +
	    "(id, userid, title, modified, status, template, body_one, " +
	    "body_two, caption_one, caption_two) ";

	switch (LocoConfig.DB_TYPE)
	{
	case (SQL.DB_TYPE_INFORMIX):
	case (SQL.DB_TYPE_MYSQL):
	    statement += "values (0, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
	    break;
	case (SQL.DB_TYPE_ORACLE):
	    statement += 
		"values (pub_seq.nextval, ?, ?, SYSDATE, ?, ?, ?, ?, ?, ?)";
	    break;
	}
	    
	try
	{
	    PreparedStatement pstmt = conn.prepareStatement(statement);

	    pstmt.setInt(1, userid);
	    pstmt.setString(2, title);
	    pstmt.setTimestamp(3, new Timestamp(System.currentTimeMillis()));
	    pstmt.setInt(4, NOT_YET_APPROVED);
	    pstmt.setString(5, template);
	    pstmt.setString(6, body_one);
	    pstmt.setString(7, body_two);
	    pstmt.setString(8, caption_one);
	    pstmt.setString(9, caption_two);

	    pstmt.executeUpdate();
	    pstmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + statement + ": exception: " +
		    StringUtil.exceptionStackTrace(se));	    

	    return(false);
	}

	return(true);
    }

    /**
     * Inserts images into a page entry.  Must be done afterwards because
     * the image name has to have the id name appended to it, and the id
     * is not given until after an entry for the page has been made.
     * @param conn database connection
     * @param id        page id
     * @param image_one first image filename
     * @param image_two second image filename
     *
     * @return false if a database error occurs
     */
    public static boolean insertImagesIntoPage(Connection conn, int id,
					       String image_one,
					       String image_two)
    {
	String statement = "update loco_publisher " +
	    "set image_one = ?, " +
	    "image_two = ? " +
	    "where id = " + id; 
	
	try
	{
	    PreparedStatement pstmt = conn.prepareStatement(statement);
	    
	    pstmt.setString(1, image_one);
	    pstmt.setString(2, image_two);
	    
	    pstmt.executeUpdate();
	    pstmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + statement + ": exception: " +
		    StringUtil.exceptionStackTrace(se));	    

	    return(false);
	}

	return(true);	
    }

    /**
     * Re-inserts a page entry into the database.
     * @param id          id of page
     * @param body_one    first body text field
     * @param body_two    second body text field
     * @param image_one   name of first image file
     * @param image_two   name of second image file
     * @param caption_one String caption for first image
     * @param caption_two String caption for second image
     * @param template    String name of template file
     *
     * @return false if a database error occurs
     */
    public static boolean reInsertPage(Connection conn,
				       int id,
				       String title,
				       String body_one,
				       String body_two,
				       String image_one,
				       String image_two,
				       String caption_one,
				       String caption_two,
				       String template)
    {
	String statement = "update loco_publisher " +
	    "set title = ?, " +
	    "body_one = ?, " +
	    "body_two = ?, " +
	    "image_one = ?, " +
	    "image_two = ?, " +
	    "caption_one = ?, " +
	    "caption_two = ?, " +
	    "template = ? where id = " + id; 

	try
	{
	    PreparedStatement pstmt = conn.prepareStatement(statement);

	    pstmt.setString(1, title);
	    pstmt.setString(2, body_one);
	    pstmt.setString(3, body_two);
	    pstmt.setString(4, image_one);
	    pstmt.setString(5, image_two);
	    pstmt.setString(6, caption_one);
	    pstmt.setString(7, caption_two);
	    pstmt.setString(8, template);

	    pstmt.executeUpdate();
	    pstmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + statement + ": exception: " +
		    StringUtil.exceptionStackTrace(se));	    

	    return(false);
	}

	return(true);
    }

    /**
     * Returns the most recently inserted page id of a user.
     * @param conn database connection
     * @param userid userid of user
     *
     * @return page id, -1 if none exists
     * 
     */
    public static int getMostRecentId(Connection conn, int userid)
    {
	String query = "select max(id) from loco_publisher " +
	    "where userid = " + userid;

	int id = -1;	

	try
	{
	    Statement stmt = conn.createStatement();
	    ResultSet rs = stmt.executeQuery(query);
	    
	    if(rs.next())
		id = rs.getInt(1);

	    rs.close();
	    stmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + query + ": exception: " +
		    StringUtil.exceptionStackTrace(se));	    
	}

	return(id);
    }

    /**
     * Returns all the fields of a user page (identified by its id) in a
     * SortableTable.
     * @param conn database connection
     * @param id page id
     * @param hash_in Hashtable to put these values in:
     *   id (page id)
     *   userid 
     *   title
     *   modified (date)
     *   status (ie. NOT_YET_APPROVED)
     *   template (filename)
     *   body_one (text)
     *   body_two
     *   image_one (filename)
     *   image_two (filename)
     *   caption_one (text)
     *   caption_two
     */
    public static void getPageInfo(Connection conn, int id, Hashtable hash_in)
    {
	String query = "select " +
	    "id, userid, title, modified, status, template, body_one, " +
	    "body_two, image_one, image_two, caption_one, caption_two " +
	    "from loco_publisher " +
	    "where id = " + id;

	Hashtable page_hash = hash_in;

	try
	{
	    Statement stmt = conn.createStatement();
	    ResultSet rs = stmt.executeQuery(query);
	    
	    if(rs.next())
	    {
		page_hash.put("id", new Integer(rs.getInt(1)));
		page_hash.put("userid", new Integer(rs.getInt(2)));
		page_hash.put("title", rs.getString(3));
		page_hash.put("modified", rs.getDate(4));
		page_hash.put("status", new Integer(rs.getInt(5)));
		page_hash.put("template", rs.getString(6));
		page_hash.put("body_one", rs.getString(7));
		page_hash.put("body_two", rs.getString(8));
		page_hash.put("image_one", rs.getString(9));
		page_hash.put("image_two", rs.getString(10));
		page_hash.put("caption_one", rs.getString(11));
		page_hash.put("caption_two", rs.getString(12));

	    }

	    rs.close();
	    stmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + query + ": exception: " +
		    StringUtil.exceptionStackTrace(se));	    

	}
    }

    /**
     * Verifies if a user is the owner of this page or not.
     * @param conn database connection
     * @param id page id
     * @param userid id of user
     *
     * @return true if the user is the creator of this page
     */
    public static boolean isOwner(Connection conn, int id, int userid)
    {
	String query = "select userid from loco_publisher " +
	    "where id = " + id;
	
	boolean result = false;
	
	try
	{
	    Statement stmt = conn.createStatement();
	    ResultSet rs = stmt.executeQuery(query);
	    
	    if(rs.next())
		result = (userid == rs.getInt(1));

	    rs.close();
	    stmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + query + ": exception: " +
		    StringUtil.exceptionStackTrace(se));	    
	}

	return(result);
    }

    /**
     * Returns the number of published pages in the database according to
     * status.
     * @param conn   database connection
     * @param status constant indicating the status of the pages to count
     *
     * @return -1 if a database error occurs
     */
    public static int getNumPages(Connection conn, int status)
    {
	String query = "";
	
	if(status < NYA_AND_APPR)
	{
	    query = "select count(*) from loco_publisher " +
		"where status = " + status;
	}
	else
	{
	    query = "select count(*) from loco_publisher " +
		"where status >= " + NOT_YET_APPROVED;
	}

	try
	{
	    Statement stmt = conn.createStatement();
	    ResultSet rs = stmt.executeQuery(query);

	    if(rs.next())
	    {
		int numPages = rs.getInt(1); 
		rs.close();
		stmt.close();
		
		return(numPages);
	    }
	    else
	    {
		rs.close();
		stmt.close();

		return(0);
	    }
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + query + ": exception: " +
		    StringUtil.exceptionStackTrace(se));
	    return(-1);
	}
    }

    /**
     * Returns a table containing index information of published pages between
     * the two ranges inclusive, beginning with most recent page.
     * @param conn database connection
     * @param low_range index 1 is the most recent page, the lowest range
     * @param high_range should not exceed the number of published pages in
     * the database
     * @param status constant indicating the status of the pages to fetch
     *
     * @return SortableTable with the following columns: id (page id)
     *                                                   modified (date)
     *                                                   username (author)
     *                                                   title
     */
    public static SortableTable getIndexTable(Connection conn, int low_range,
					      int high_range, int status)
    {
	// The corresponding ids of the low and high ranges from the database.
	// Remember that the lowest range, index 1, will be the highest id
	// in the database, since we are starting with the most recent page.
	// Therefore high_id will be associated with the low_range, and
	// vice-versa.
	int low_id = 0;
	int high_id = 0;

	String status_clause = "";
	
	switch(status)
	{
	case CENSORED:
	    status_clause = "status = " + CENSORED;
	    break;
	case NOT_YET_APPROVED:
	    status_clause = "status = " + NOT_YET_APPROVED;
	    break;
	case APPROVED:
	    status_clause = "status = " + APPROVED;
	    break;
	case NYA_AND_APPR:
	    status_clause = "status >= " + NOT_YET_APPROVED;
	    break;
	default:
	    Loco.log (0, "PublisherLib.getIndexTable received improper status");
	    return(null);
	}
	    

	String query = "select id from loco_publisher where " +
	    status_clause + " order by id desc";

	Statement stmt = null;
	ResultSet rs = null;

	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);


	    // Iterate through the ids until we get to the first one we want
	    for(int i = 1; i < low_range; i++)
	    {
		if (!rs.next())
		{
		    // Not enough entries to get this requested range
		    return(null);
		}
	    }

	    // Now get the corresponding ids of the low range and the high
	    // range
	    int last_id = 0;
	
	    int i = low_range;
	    while(i <= high_range)
	    {
		if (rs.next())
		{
		    last_id = rs.getInt(1);
		    
		    if (i == low_range)
			high_id = last_id;
		    else if (i == high_range)
			low_id = last_id;
		}
		else
		{
		    if(high_id == 0)
			// we never got to the first range
			return(null);
		    else
		    {
			// we got some of the range, but not all
			low_id = last_id;
			break;
		    }
		}
		
		i++;
	    }

	    rs.close();
	    stmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + query + ": exception: " +
		    StringUtil.exceptionStackTrace(se));
	    return(null);
	}

	query = "select pub.id as id, pub.modified as modified, " + 
	    "l.username as username, pub.title as title " +
	    "from loco_publisher pub, loco_users l " +
	    "where pub.userid = l.userid " +
	    " and id >= " + low_id + 
	    " and id <= " + high_id + " and " + status_clause +
	    " order by id desc";

	SortableTable table = SortableTable.loadQuery(conn, query);
	if ( table == null)
	    return(null);

	table.formatDateColumn(2, "MM/d/yyyy");
	
	return(table);
    }

    /**
     * Returns an index of pages of a single user.
     * @param conn database connection
     * @param userid id of user
     *
     * @return SortableTable with the following columns: id (page id)
     *                                                   modified (date)
     *                                                   username (author)
     *                                                   title
     *                                                   status
     */
    public static SortableTable getUserPageIndex(Connection conn, int userid)
    {
	String query = "select pub.id as id, pub.modified as modified, " + 
	    "l.username as username, pub.title as title, " +
	    "pub.status as status " +
	    "from loco_publisher pub, loco_users l " +
	    "where pub.userid = l.userid " +
	    " and pub.userid = " + userid +
	    " and status >= " + NOT_YET_APPROVED +
	    " order by id desc";

	SortableTable table = SortableTable.loadQuery(conn, query);
	if ( table == null)
	    return(null);

	table.formatDateColumn( 2, "MM/d/yyyy");
	
	return(table);
    }

    /**
     * Changes the status of a page in the database.
     * @param conn database connection
     * @param page_id id of the web page
     *
     * @return -2 if page does not exist
     */
    public static int getStatus(Connection conn, int page_id)
    {
	String query = "select status from loco_publisher " +
	    "where id = " + page_id;
 
	int status = -2;

	try
	{
	    Statement stmt = conn.createStatement();
	    ResultSet rs = stmt.executeQuery(query);

	    if(rs.next())
	    {
		status = rs.getInt(1);
	    }

	    rs.close();
	    stmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + query + ": exception: " +
		    StringUtil.exceptionStackTrace(se));
	}	

	return(status);
    }

    /**
     * Changes the status of a page in the database.
     * @param conn database connection
     * @param page_id id of the web page
     * @param status status to change the page to
     *
     * @return false if a database error occurs
     */
    public static boolean changeStatus(Connection conn, int page_id,
				       int status)
    {
	String query = "update loco_publisher set status = " + status +
	    " where id = " + page_id;
	
	try
	{
	    Statement stmt = conn.createStatement();
	    stmt.executeUpdate(query);

	    stmt.close();
	}
	catch(SQLException se)
	{
	    Loco.log (0, "SQLException occurred in PublisherLib - query: " 
		    + query + ": exception: " +
		    StringUtil.exceptionStackTrace(se));
	    
	    return(false);
	}	

	return(true);
    }
}



