/**
 * PersonalPrefServlet.java 
 *
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License. 
 *
 * The initial develop of this code under the LPL Leverage Information Systems.
 * Portions created by Leverage are Copyright (C) 1998 Leverage Information Systems.
 * All Rights reserved.
 */

package org.locomotive.module.userpref;

import java.lang.*;
import java.text.SimpleDateFormat;
import java.sql.*;
import java.util.Hashtable;
import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.locomotive.loco.*;
import org.locomotive.loco.servlet.*;
import org.locomotive.loco.perm.*;
import org.locomotive.util.*;
import org.locomotive.server.Log;

/** 
 * A simple User Preference System. Allows a user to modify his 
 * preferences/personal info.  To use this module, add the servlet
 * to your hrt.conf file. An example would be:
 * <PRE>
 * usrpref           org.locomotive.module.userpref.PersonalPrefServlet
 * </PRE>
 *
 * entry point from UserPrefServlet.java - main logic of servlet
 * This servlet accepts the following URLS: <PRE>
 * usrpref or            displays personal page. if permissions allow,
 * usrpref?personal      modifications may be made to certain fields.
 * </PRE>
 */
public class PersonalPrefServlet extends LocoServlet
{
    // max number of rows returned on a search command
    final static int SEARCH_RESULT_MAX = 500;


    /**
     * determines whether hit requires handling new data,
     * then displays the page
     */
    public void doGet(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
        // holds valid form values and user id of user to be modified
        // in the db (not necessarily the same as the user accessing the page)

        UserPref userPref = new UserPref();
        StringBuffer errors = new StringBuffer();

        // user must be logged in to view user preferences  
        if  (user.isAnonymous())
        { 
	    steam_vars.put("UPREF_MESG_WHY", "You must be logged in to access"+
			   " User Preferences");
	    displayPage("userpref/noaccess.tmpl", steam_vars);  
	    logEvent("usrpref", "personal", "noaccess");
	    return;
        }

	// loads userPref object with db info in LOCO_USER_PREFS
	try {
	    userPref = new UserPref(conn, user.getUserid());
	}
	catch (SQLException se)
	{
	    //serious error.
	    handleException(se, "UserPref(int id) ");
	    return;
	}          
                    
	// case user doesn't have db entry in LOCO_USER_PREFS 
	// loads userprefs with userid and username
	if (userPref.userName.equals(""))
	    userPref = new UserPref(user.getUserid(), user.getUsername());

	// loads permission values 
	try {
	    SortableTable user_perms = 
		UserPerm.getPermissions(user.getUserid(), conn).getPermTable();

	    // makes perm info available to template      
	    steam_vars.put("user_perms", user_perms);

	}
	catch (SQLException se)
	{
	    //serious error.
	    handleException(se, "Loading Permissions");
	    return;
	} 

	// makes user pref info available to template      
	formatUserPrefsToSteam(userPref, steam_vars);

	displayPage("userpref/personal.tmpl", steam_vars);
	logEvent( "userpref", "personal", "viewed");
	return;
    }

    
    public void doPost(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	// if the form is empty, then pretened it's a 'GET' request.
        if (form_data.isEmpty())
	{
	    doGet(req, resp);
	    return;
	}
	
        UserPref userPref = new UserPref();
        StringBuffer errors = new StringBuffer();

        if (form_data.getField("SAVE_CHANGES")!=null)
        {

           // ensures user is viewing his/her own user data
           userPref.userId = String.valueOf(user.getUserid());

           if(saveFormValuesToUserPref(form_data, userPref, errors))
           {
              try {
		  UserPref.storeUserPrefsToDB(conn, userPref, errors);
              }
	      catch (SQLException se) 
              {
                  //serious error.  
                  handleException(se, "storeUPToDB");
                  return;                
              }
              catch (LoadUserException lue)
              {
                  //serious error. 
	          handleException(lue, "storeUPToDB");
                  return;  
              }                   
	   }    

           if (errors.length() > 0)
              steam_vars.put("UPREF_EDIT_ERR", errors.toString());
           else
              steam_vars.put("EDIT_COMPLETE", "true");

           //save form vals for redisplay
           form_data.copyAllIntoSteam(steam_vars);

           displayPage("userpref/personal.tmpl", steam_vars);
           logEvent("usrpref", "personal", "edited");
           return;
	}
	else  // unrecognized token string
	{
	    displayPage ("error/unknown_servlet_option.tmpl", steam_vars);
	}
    }

  

       
    public static boolean saveFormValuesToUserPref(FormData formData, 
			        UserPref upi, StringBuffer errorBuf)
    {
       upi.userId = String.valueOf(formData.getInt(
			     "UPREF_USERID", "user id"));

       upi.userName = formData.getString("UPREF_USERNAME", "username",
					    0, 32);

       upi.fName = formData.getString("UPREF_FIRSTNAME", "first name", 
					  0,32);

       upi.mName = formData.getString("UPREF_MIDDLENAME", "middle name",
					  0,32);

       upi.lName = formData.getString("UPREF_LASTNAME", "middle name",
					  0,32);

       upi.email = formData.getEmailAddress("UPREF_EMAILADDR", "email",
					  0,64);
       formData.ignoreErrorCode(FormData.IS_EMPTY);

       upi.hAddr = formData.getString("UPREF_HOMEADDR", "home address",
					  0,32);

       upi.hCity = formData.getString("UPREF_HOMECITY", "home city",
					  0,32);

       upi.hState = formData.getString("UPREF_HOMESTATE", "home state",
					  0,2);
         
       upi.hZip = formData.getString("UPREF_HOMEZIP", "home zip code",
					  0,10);
       formData.ignoreErrorCode(FormData.IS_NULL);

       upi.hCountry = formData.getString("UPREF_HOMECOUNTRY", 
					     "home country",0,32);

       upi.wAddr = formData.getString("UPREF_WORKADDR", "work address",
					  0,32);

       upi.wCity = formData.getString("UPREF_WORKCITY", "work city",
					  0,32);

       upi.wState = formData.getString("UPREF_WORKSTATE", "work state",
					  0,2);
         
       upi.wZip = formData.getString("UPREF_WORKZIP", "work zip code",
					 0,10);

       upi.wCountry = formData.getString("UPREF_WORKCOUNTRY", 
					     "work country",0,32);

       upi.hPhone = formData.getString("UPREF_HOMEPHONE", "home phone",
					 9, 16);
       formData.ignoreErrorCode(FormData.IS_EMPTY);

       upi.wPhone = formData.getString("UPREF_WORKPHONE", "work phone",
					 9, 16);
       formData.ignoreErrorCode(FormData.IS_EMPTY);
        
       upi.passHint = formData.getString("UPREF_PASSHINT", "password hint",
					  0,32);

       upi.passWd = formData.getTrimmedString("UPREF_PASSWD1", "password",
						5, 12);
       formData.ignoreErrorCode(FormData.IS_EMPTY);
       String passWd2 =  formData.getTrimmedString("UPREF_PASSWD2", 
					       "password(confirm)", 5, 12);
       formData.ignoreErrorCode(FormData.IS_EMPTY);

       errorBuf.append(formData.allErrorMsgs());

       if (!passWd2.equals(""))
          if (!upi.passWd.equals(passWd2))
             errorBuf.append("password fields do not match");

       if (errorBuf.toString().length() == 0)
           return true;
       else
           return false;
    }

    /**
    * attempts to stuff form values back into the environment 
    * so the user will still be able to view the form data upon reload.
    * for example, in the event the user posts values that violate 
    * the constraints, this eliminates the burden of having to re-enter 
    * all the form data for the next try
    */
    private void saveFormValues()
    {
       form_data.copyAllIntoSteam(steam_vars);
    }

    public static void formatUserPrefsToSteam(UserPref info, 
				Hashtable steamVars)
    {
       steamVars.put("UPREF_USERID", info.userId);
       steamVars.put("UPREF_USERNAME", info.userName);
       steamVars.put("UPREF_FIRSTNAME", info.fName);
       steamVars.put("UPREF_MIDDLENAME", info.mName);
       steamVars.put("UPREF_LASTNAME", info.lName);
       steamVars.put("UPREF_EMAILADDR", info.email);
       steamVars.put("UPREF_HOMEADDR", info.hAddr);
       steamVars.put("UPREF_HOMECITY", info.hCity);
       steamVars.put("UPREF_HOMESTATE", info.hState);
       steamVars.put("UPREF_HOMEZIP", info.hZip);
       steamVars.put("UPREF_HOMECOUNTRY", info.hCountry);
       steamVars.put("UPREF_WORKADDR", info.wAddr);
       steamVars.put("UPREF_WORKCITY", info.wCity);
       steamVars.put("UPREF_WORKSTATE", info.wState);
       steamVars.put("UPREF_WORKZIP", info.wZip);
       steamVars.put("UPREF_WORKCOUNTRY", info.wCountry);
       steamVars.put("UPREF_HOMEPHONE", info.hPhone);
       steamVars.put("UPREF_WORKPHONE", info.wPhone);
       steamVars.put("UPREF_PASSHINT", info.passHint);
       steamVars.put("UPREF_PASSWD1", info.passWd);
       return;
    } 
}

