/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.module.userpref;

import java.lang.*;
import java.sql.*;
import java.util.*;

import org.locomotive.loco.*;
import org.locomotive.loco.handler.*;
import org.locomotive.loco.perm.*;
import org.locomotive.util.*;
import org.locomotive.server.Log;

/** 
 * A simple User Management System which allows the admin user to
 * view and modify users, user preferences, and permissions.
 * </PRE> */

public class UserMgtHandler extends GenericHandler
{
    // max number of rows returned on a search command
    final static int SEARCH_RESULT_MAX = 500;

    /**
     * entry point from UserPrefHandler.java - main logic of handler
     * This handler accepts the following URLS: <PRE>
     * userpref?manage          displays page,if user has ums admin priveleges
     * userpref?manage+help     same as above (with help instructions)
     * userpref?manage+userinfo+%id loads info on user with the id value  
     * </PRE>
     */
    public void handleRequest (HandlerData hd_in)
    {
	super.handleRequest(hd_in);

        // holds valid form values and user id of user to be modified
        // in the db (not necessarily the same as the user accessing the page)
        UserPref user_pref = new UserPref();

        StringBuffer errors = new StringBuffer(); // holds error messages
 
	// enforces privelege for viewing user management system
        try {
           if (!UserPref.hasAdminPriveleges(conn,user.getUserid()))
           {
	      //pushes value to template via steam var
              steam_vars.put("UPREF_MESG_WHY", "No view access.");

              displayPage("userpref/noaccess.tmpl", steam_vars);  
              logEvent ("userpref", "manage", "noviewaccess");      
              return;
	   }
        } catch (SQLException se)
	{
	    //serious error
            handleException(se, "UserMgtHandler-AdminCheck");
            return;
	}

        // if form_data is empty, then we had an HTTP 'GET' request
        // otherwise it was a 'POST'


        // this section handles a GET request. 
        // for this request, it's either a blank management page load,
	// a search table load for the management page, or a display
        // of information on a particular user

        if (form_data.isEmpty())
        {         
           if (url_tokens.length == 4)
	   {
  	      // handles case userpref?manage+userinfo+%id
              if (url_tokens[2].equals("userinfo"))
	      {
                 int user_id = 0;

                 // confirms the 3rd token is an integer and store the value
                 try
                 {
                    user_id = Integer.parseInt(url_tokens[3]);
		 } catch(NumberFormatException nfe)
	         {
                    displayPage("error/unknown_handler_option.tmpl", 
				steam_vars);
                    return;
	         }
	
	         // creates userPref object from database
                 try {
                    user_pref = new UserPref(conn, user_id);
   	         }catch (SQLException se)
	         {
	            //serious error.
	            handleException(se, "UserPref(int id) ");
                    return;
	         }          

                 // gets the user's permission set
	         try{

                    SortableTable user_perms =  UserPerm.getPermissions(
                       user_id, conn).getPermTable();

                    // makes perm info available to template      
                    steam_vars.put("user_perms", user_perms);

   	         }catch (SQLException se)
	         {
	            //serious error.
	            handleException(se, "Loading Permissions");
                    return;
	         } 

                 // makes user pref info available to template      
                 PersonalPrefHandler.formatUserPrefsToSteam(user_pref,
							    steam_vars);
                 
                 displayPage("userpref/managePrefs.tmpl", steam_vars);
                 logEvent("userpref", "manage", "userviewed");
                 return;
	      }
	   }

           if (url_tokens.length == 3)
	   {
 	      // handles case userpref?manage+help
              if (url_tokens[2].equals("help"))
              {
 		 // saves search form vals for redisplay
                 form_data.copyAllIntoSteam(steam_vars);

		 // sets steam var to display help text 
                 steam_vars.put("UPREF_REQUEST_HELP", "true");

                 // init permsission list for delete option
                 setPermList();

                 displayPage("userpref/manage.tmpl", steam_vars);    
                 logEvent ("userpref", "manage", "helpviewed");
                 return;
	      }   
	   }
           if (url_tokens.length == 2)
	   {
 	      // handles case userpref?manage

              // init permsission list for delete option
              setPermList();

              displayPage("userpref/manage.tmpl", steam_vars);           
              logEvent ("userpref", "manage", "mgtviewed");
              return;
           }
	}
	   

        // this section handles a 'POST' request which could be a Search,
        // Clear Fields, Save Changes, Create Permission or Delete Permission
        // command

	// handles case userpref?manage with "SEARCH" POST
        if (form_data.getField("SEARCH")!=null)
        {
           // performs search, handles errors, and adds a table of 
	   // search results into the steam_vars
           performSearch();

           // inits permsission list for delete option
           setPermList();

 	   // saves search form vals for redisplay
           form_data.copyAllIntoSteam(steam_vars);

           displayPage("userpref/manage.tmpl", steam_vars); 
           logEvent ("userpref", "manage", "search");
           return;
      
        }

        //handle case userpref?manage with "CLEAR_FIELDS" POST
        if (form_data.getField("CLEAR_FIELDS")!=null)
	{
           // inits permsission list for delete option
           setPermList();

           // reloads page without saving data.
           displayPage("userpref/manage.tmpl", steam_vars);           
           return;
	}

	// handles case userpref?manage+userinfo & with "SAVE_CHANGES" POST
        if (form_data.getField("SAVE_CHANGES")!=null)
        {
            // updates user prefs in db
            if (PersonalPrefHandler.saveFormValuesToUserPref(
		form_data, user_pref, errors))
	    {
               try {
                  UserPref.storeUserPrefsToDB(conn, user_pref, errors);
               }catch (SQLException se) 
               {
	           // serious error.  
                    handleException(se, "storeUPToDB");
                    return;                
               }
               catch (LoadUserException lue)
               {
                  // serious error. 
                  handleException(lue, "storeUPToDB");
                  return;  
               }
            }

            // updates user permissions
            PermissionSet perm_set= 
               new PermissionSet(Integer.parseInt(user_pref.userId));
            String perm_name = "";
            Enumeration e = form_data.getHashtable().keys();
             
            // roots out the perms from all of the other form values
            // and sets the permission in a permission set.  
	    for(int i = 0; e.hasMoreElements(); i++)
	    {
               perm_name = (String) e.nextElement();
               if (perm_name.startsWith("PERM_"))
                  perm_set.set(perm_name.substring(5));
	    }
            
            // stores the users new permission set to the db 
            try 
            {
               UserPerm.setPermissions(perm_set, conn);
	    }
            catch (SQLException se)
	    {
               //serious error. 
               handleException(se, "Set Permissions");
               return;  
	    }

            // sets steam error values
            if (!errors.toString().equals(""))
               steam_vars.put("UPREF_EDIT_ERR", errors.toString());
            else
               steam_vars.put("EDIT_COMPLETE", "true");

            // gets the user's permission set
	    try{

               SortableTable user_perms =  UserPerm.getPermissions(
                  Integer.parseInt(user_pref.userId), conn).getPermTable();

               // makes perm info available to template      
               steam_vars.put("user_perms", user_perms);

   	    }catch (SQLException se)
	    {
	       //serious error.
	       handleException(se, "Loading Permissions");
               return;
	    } 

            // saves form vals for redisplay
            form_data.copyAllIntoSteam(steam_vars);

            displayPage("userpref/managePrefs.tmpl", steam_vars);
            logEvent("userpref", "manage", "edited");
            return;
        }
	      
        // handles case userpref?manage with "CREATE_PERMISSION" POST
        if (form_data.getField("CREATE_PERMISSION")!=null)
	{
           String perm_name = form_data.getString("NEW_PERMISSION", 
                  "new permission name", 1, 32);

           // confirms the perm_name value conforms to constraints
           if(form_data.hasAnyErrors())
	   {
              errors.append(form_data.allErrorMsgs());
	   }
           else
	   {
	      // adds newly created permission to the db
              try {
	         Permission perm = Permission.createPermission(conn,perm_name);
	      }catch (SQLException se) 
              {
	         // checks for violation of the unique value constraint 
		 // on the perm name value.  val is specific to ORACLE vendor
                 int PERM_ALREADY_EXISTS = 1;

                 if (se.getErrorCode() == PERM_ALREADY_EXISTS)  
                    errors.append("Permission "+ perm_name+ 
		       " already exists.  You must use a different name.");
                 else
                 {
	            //serious error.  
                    handleException(se, "UMHandler-CREATEPERM");
                    return;
	         }                
              }
	   }

           // pushes error messages or successful message to template
           if (!errors.toString().equals(""))
              steam_vars.put("PERM_CREATE_ERR", errors.toString());
           else 
              steam_vars.put("PERM_CREATE_COMPLETE", "complete");


           // saves form vals for redisplay
           form_data.copyAllIntoSteam(steam_vars);
                 
           // inits permsission list for delete option
           setPermList();

           displayPage("userpref/manage.tmpl", steam_vars);

           logEvent("userpref", "perm", "create");
           return;       
	}

        // handles case userpref?manage with "DELETE_PERMISSION" POST
        if (form_data.getField("DELETE_PERMISSION")!=null)
	{
           String perm_name = form_data.getString("PERM_TO_DEL", 
                  "permission to delete", 1, 32);

           try {
	      Permission.deletePermission(conn, perm_name);
	   }catch (SQLException se) 
           {
	      //serious error.  
              handleException(se, "UMHandler-DELETEPERM");
              return;	
           }
         
           steam_vars.put("PERM_DELETE_COMPLETE", "complete");

           // saves form vals for redisplay
           form_data.copyAllIntoSteam(steam_vars);

           // inits permsission list for delete option
           setPermList();

           displayPage("userpref/manage.tmpl", steam_vars);
           logEvent("userpref", "perm", "delete");
           return;       
	}

	// unmatched token string
	displayPage ("error/unknown_handler_option.tmpl", steam_vars);
	

    }  // end processHandler


     /**
      *  if the search forms are not blank, an sql query is performed 
      *  and the manage page template variable UPREF_SEARCHLIST is loaded with
      *  the query results in an html table format.
      */
      private void performSearch()
      {
	 // holds search field values
         String username="", userid="", lastname="", emailaddr="";

         // loads the search forms
         userid = form_data.getField("UPREF_SEARCH_USERID");
         username = form_data.getField("UPREF_SEARCH_USERNAME");
         lastname = form_data.getField("UPREF_SEARCH_LASTNAME");
         emailaddr = form_data.getField("UPREF_SEARCH_EMAILADDR");

         
         StringBuffer resultTable = new StringBuffer();
         StringBuffer errors = new StringBuffer();

         boolean success;
         try{
            success = UserPref.getSearchResultTableRows(conn,
               username, lastname, emailaddr, userid, resultTable, errors,
               SEARCH_RESULT_MAX);
	 }catch (SQLException se) 
         {
	    //serious error.  
            handleException(se, "storeUPToDB");
            return;                
         }

         if (!success)
            steam_vars.put("SEARCH_NOT_SUCC", errors.toString());

	 if (resultTable.length() !=0)
            steam_vars.put("UPREF_SEARCHRESULTS", resultTable.toString());
         else 
            steam_vars.put("SEARCH_NOT_SUCC", "No matching entries found.");
	 
      }

      /*generate list of permissions for delete perm option in template */
      private void setPermList()
      {
	 // defaults steam var to zero 
         steam_vars.put("NUMBER_OF_PERMS", "0");

         // gets a vector of all permission names from the db
         Vector perm_vector;
         try{

            perm_vector = Permission.getPermissions(conn);

         }catch(SQLException se)
         {
            handleException(se, "loading perm list");
            return;
         }

         // convert the vector of names to an array so the steam construct
         // [loop number-of-times [colname array-variable]] can be used

         steam_vars.put("NUMBER_OF_PERMS", 
                        Integer.toString(perm_vector.size()));

         String[] perm_list = new String[perm_vector.size()];

         for (int i = 0; i<perm_vector.size(); i++)
         {
            perm_list[i] = 
      	     ((Permission)perm_vector.elementAt(i)).toString();
         }

         steam_vars.put("perm_list", perm_list);
      }
 
}


