/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf;

import org.locomotive.obf.event.*;
import java.io.*;


/**
 * Provides implementation of the basic functionality of a readable and
 * writable business object.  This implementation includes adding and
 * removing validity listeners and firing of validity events.
 *
 * @see    DataObject
 * @see    EditableDataObject
 *
 * @author David Warnock
 * @author Erik Erskine
 */

public abstract class AbstractRwBusinessObject
implements UidBusinessObject, Validation, ValidityListener {

  /**
   * Flag to denote the initial validity, or the last validity when the last
   * validity event was fired.
   */

  protected boolean wasValid;
  

  private transient ValidityManager validityManager = new ValidityManager();


  /**
   * Called whenever an <code>EditableDataObject</code> that this business
   * object contains and listens to changes validity.
   * The overall validity of this business object is then
   * re-examined and a new validity event fired if necessary.<br>
   * Calling this method directly is not recommended, it is better to let the
   * data fields fire their own validity events if possible.
   *
   * @param ve the event
   */

  public void validityChanged(ValidityEvent ve) {
    boolean v = isValid();
    if (v != wasValid) {
      validityManager.fireValidityEvent(this, v);
      wasValid = v;
    }
  }


  /**
   * Adds the specified validity listener to receive validity change events
   * from this object.
   * 
   * @see ValidityListener
   * @see ValidityEvent
   * @see AbstractRwBusinessObject#removeValidityListener
   * @param vl the validity listener
   */

  public void addValidityListener(ValidityListener vl) {
    validityManager.addValidityListener(vl);
  }


  /** Removes the specified validity listener so that it no longer reveives
   * validity changed events from this component.
   * 
   * @see ValidityListener
   * @see ValidityEvent
   * @see AbstractRwBusinessObject#addValidityListener
   * @param vl the validity listener
   */

  public void removeValidityListener(ValidityListener vl) {
    validityManager.removeValidityListener(vl);
  }


  /**
   * Tests whether or not this business object is valid.<br>
   * Usually a business object is valid if all it's
   * <code>EditableDataObject</code> fields are valid,
   * but this is not always the case.
   */

  public abstract boolean isValid();


  // Serialization

  private void readObject(ObjectInputStream s)
  throws ClassNotFoundException, IOException {
    s.defaultReadObject();
    validityManager = new ValidityManager();
  }


  // Cloning

  protected Object clone() throws CloneNotSupportedException {
    AbstractRwBusinessObject copy = (AbstractRwBusinessObject) super.clone();
    copy.validityManager = new ValidityManager();
    return copy;
  }

}
