/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf;


/**
 * A process to get several business objects for a given class.
 *
 * @see    GetProcess
 * @see    GetDefaultProcess
 *
 * @author Erik Erskine
 */

public class GetMultipleProcess implements BusinessProcess {

  public static final int UNLIMITED = -1;

  private Class busClass;
  private Object where, order;
  private int limit;


  /**
   * Creates a new process.
   */

  public GetMultipleProcess() {
  }


  /**
   * Creates a new process.
   *
   * @param      busClass  the class of business object to get.  This class
   */

  public GetMultipleProcess(Class busClass) {
    this(busClass, null, null, -1);
  }


  /**
   * Creates a new process.
   *
   * @param      busClass  the class of business object to get.  This class
   *                       must implement the <code>UidBusinessObject</code>
   *                       interface.
   * @param      where     an object to use to decide which objects to get.
   */

  public GetMultipleProcess(Class busClass, Object where) {
    this(busClass, where, null, -1);
  }


  /**
   * Creates a new process.
   *
   * @param      busClass  the class of business object to get.  This class
   * @param      limit     an upper limit on how many objects to get.
   */

  public GetMultipleProcess(Class busClass, int limit) {
    this(busClass, null, null, limit);
  }


  /**
   * Creates a new process.
   *
   * @param      busClass  the class of business object to get.  This class
   * @param      where     an object to use to decide which objects to get.
   * @param      limit     an upper limit on how many objects to get.
   */

  public GetMultipleProcess(Class busClass, Object where, int limit) {
    this(busClass, where, null, limit);
  }


  /**
   * Creates a new process.
   *
   * @param      busClass  the class of business object to get.  This class
   *                       must implement the <code>UidBusinessObject</code>
   *                       interface.
   * @param      where     an object to use to decide which objects to get.
   * @param      order     an object to use to decide the order in which
   */

  public GetMultipleProcess(Class busClass, Object where, Object order) {
    this(busClass, where, order, -1);
  }


  /**
   * Creates a new process.
   *
   * @param      busClass  the class of business object to get.  This class
   *                       must implement the <code>UidBusinessObject</code>
   *                       interface.
   * @param      where     an object to use to decide which objects to get.
   * @param      order     an object to use to decide the order in which
   *                       objects are returned.
   * @param      limit     an upper limit on how many objects to get.
   */

  public GetMultipleProcess(Class busClass, Object where, Object order, int limit) {
    this.busClass = busClass;
    this.where = where;
    this.order = order;
    this.limit = limit;
  }


  /**
   * Returns the class of business object this process will get.
   *
   * @see        GetProcess#setBusClass(java.lang.Class)
   */

  public Class getBusClass() {
    return busClass;
  }


  /**
   * Sets the class of business object this process will get.
   *
   * @param      busClass the class of business object to get.
   *             This class must implement must implement the
   *             <code>BusinessObject</code> interface.
   * @see        GetProcess#getBusClass()
   */

  public void setBusClass(Class c) {
    busClass = c;
  }


  /**
   * Returns the object describing which business objects to get.
   *
   * @see        GetMultipleProcess#setWhere(java.lang.Object);
   */

  public Object getWhere() {
    return where;
  }


  /**
   * Sets the object describing which business objects to get.
   *
   * @see        GetMultipleProcess#getWhere();
   */

  public void setWhere(Object o) {
    where = o;
  }


  /**
   * Returns the object describing the order in which
   * the business objects should be returned.
   *
   * @see        GetMultipleProcess#setOrder(java.lang.Object);
   */

  public Object getOrder() {
    return order;
  }


  /**
   * Sets the object describing the order in which
   * the business objects should be returned.
   *
   * @see        GetMultipleProcess#getOrder();
   */

  public void setOrder(Object o) {
    order = o;
  }


  /**
   * Returns the maximum number of business objects that will be returned.
   *
   * @return     the maximum number returned,
   *             or <code>-1</code> if there is no limit
   * @see        GetMultipleProcess#setLimit(int);
   */

  public int getLimit() {
    return limit;
  }


  /**
   * Sets the maximum number of business objects that will be returned.
   *
   * @param      l the maximum number returned,
   *             or <code>-1</code> if there is no limit
   * @see        GetMultipleProcess#getLimit();
   */

  public void setLimit(int l) {
    limit = l;
  }

}
