/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf;

import org.locomotive.obf.event.RefreshListener;
import org.locomotive.obf.event.RefreshManager;
import org.locomotive.obf.server.*;
import org.locomotive.obf.client.*;
import java.sql.Connection;


public class LocalConnection {

  private ConnectionToLocalServer connectionToServer;
  private ConnectionToLocalClient connectionToClient;



  // --- Client-side connection ---


  private static class ConnectionToLocalServer implements ConnectionToServer {

    private RefreshManager refreshManager = new RefreshManager();
    private ConnectionToLocalClient server = null;

    void setConnectionToClient(ConnectionToLocalClient con) {
      server = con;
    }


    public Object doProcess(BusinessProcess p) throws ObfException {
      return server.receiveProcess(p);
    }


    Object receiveProcess(BusinessProcess p) throws ObfException {

      if (p instanceof RefreshProcess) {
        RefreshProcess rp = (RefreshProcess) p;

        switch (rp.getType()) {
        case RefreshProcess.INSERT:
          refreshManager.insertedObject(rp.getBusinessObject());
          break;
        case RefreshProcess.UPDATE:
          refreshManager.updatedObject(rp.getBusinessObject());
          break;
        case RefreshProcess.DELETE:
          refreshManager.deletedObject(rp.getBusinessObject());
          break;
        case RefreshProcess.REFRESH:
          refreshManager.refreshObjects(rp.getBusClass());
          break;
        }

      } else {
        throw new ObfException("Client can't deal with "+p.getClass());
      }

      return null;

    }


    // --- Refresher methods ---

    public void addRefreshListener(RefreshListener listener, Class busClass) {
      System.out.println("client side local connection: addRefreshListener(listener, "+busClass+")");
      try {

        // if this connection is not already listening to the class,
        // we nned to communicate and say this connection is listening
        if (refreshManager.isMonitoringClass(busClass) == false) {
          AddClientListenerProcess process =
          new AddClientListenerProcess(server, busClass);
          doProcess(process);
        }

      } catch (ObfException oe) {
        // what now?

      }

      refreshManager.addRefreshListener(listener, busClass);
    }



    public void removeRefreshListener(RefreshListener listener, Class busClass) {
      System.out.println("client side local connection: removeRefreshListener(listener, "+busClass+")");
      refreshManager.removeRefreshListener(listener, busClass);
      try {

        // if no listeners are left that are interested in this class
        // communicate and say we should stop listening
        if (refreshManager.isMonitoringClass(busClass) == false) {
          RemoveClientListenerProcess process =
          new RemoveClientListenerProcess(server, busClass);
          doProcess(process);
        }

      } catch (ObfException oe) {
        // what now?

      }
    }

  }



  // --- Server-side connection ---



  private static class ConnectionToLocalClient implements ConnectionToClient {

    private Connection localCon = null;

    protected ConnectionToLocalServer client;

    void setConnectionToServer(ConnectionToLocalServer con) {
      client = con;
    }


    public Object doProcess(BusinessProcess process) throws ObfException {
      return client.receiveProcess(process);
    }


    Object receiveProcess(BusinessProcess p) throws ObfException {

      // get handler for the process

      ProcessHandler h = ServerSupporter.getInstance().getProcessHandler(p.getClass());

      if (localCon == null) localCon = ServerSupporter.getInstance().getConnection();
      return h.doProcess(p, localCon);
    }


    // RefreshListener

    public void insertedObject(UidBusinessObject busObject) {

      try {
        doProcess(new RefreshProcess(RefreshProcess.INSERT, busObject, null));
      } catch (ObfException e) {
        // remove this client connection from the ClientNotificationQueue

      }

    }


    public void updatedObject(UidBusinessObject busObject) {

      try {
        doProcess(new RefreshProcess(RefreshProcess.UPDATE, busObject, null));
      } catch (ObfException e) {
        // remove this client connection from the ClientNotificationQueue

      }

    }


    public void deletedObject(UidBusinessObject busObject) {

      try {
        doProcess(new RefreshProcess(RefreshProcess.DELETE, busObject, null));
      } catch (ObfException e) {
        // remove this client connection from the ClientNotificationQueue

      }

    }


    public void refreshObjects(Class busClass) {

      try {
        doProcess(new RefreshProcess(RefreshProcess.REFRESH, null, busClass));
      } catch (ObfException e) {
        // remove this client connection from the ClientNotificationQueue

      }

    }



  }



  // --- Methods

  public LocalConnection() {
    connectionToServer = new ConnectionToLocalServer();
    connectionToClient = new ConnectionToLocalClient();

    connectionToServer.setConnectionToClient(connectionToClient);
    connectionToClient.setConnectionToServer(connectionToServer);
  }

  public ConnectionToServer getConnectionToServer() {
    return connectionToServer;
  }

  public ConnectionToClient getConnectionToClient() {
    return connectionToClient;
  }


}
