/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.client;


import org.locomotive.obf.*;
import org.locomotive.obf.data.UniqueId;
import org.locomotive.obf.data.UniqueIdModel;
import org.locomotive.obf.event.*;
import org.locomotive.obf.client.event.*;


import java.io.*;
import java.util.Vector;
import java.util.Enumeration;


public abstract class AbstractDataEntryController implements DataEntryController {

  private Vector dataEntryModeChangeListeners = new Vector();
  private Vector dataChangeListeners = new Vector();
  private ValidityManager validityHandler = new ValidityManager();

  // this keeps track of the uid
  private UniqueIdModel model;
  private UniqueId lastUid;
  
  private Manager mgr;
  private boolean wasValid;
  private byte currentMode = DataEntryModeChangeEvent.BROWSE;



  public AbstractDataEntryController() {
    this(new UniqueIdModel());
  }

  public AbstractDataEntryController(UniqueIdModel m) {
    model = m;
    model.addModificationListener(this);
    wasValid = isValid();
  }

  public void setManager(Manager m) {
    mgr = m;
  }

  public Manager getManager() {
    return mgr;
  }


  public UniqueIdModel getUidModel() {
    return model;
  }



  public abstract BusinessObject getBusinessObject(Class c);

  public BusinessObject getBusinessObject(Class c, Object where) {
    return null;
  }

  
  public Object doProcess(BusinessProcess process) throws ObfException {
    return mgr.doProcess(process);
  }

  public void modified(ModificationEvent me) {
    // uid model has changed - need to update things
    lastUid = (UniqueId) model.getData();
    setBusinessObjectsToNull();
    fireDataChangeEvent();
  }




  // --- Can change the data entry mode and notify listeners


  public void addDataEntryModeChangeListener(DataEntryModeChangeListener l) {
    dataEntryModeChangeListeners.addElement(l);
  }

  public void removeDataEntryModeChangeListener(DataEntryModeChangeListener l) {
    dataEntryModeChangeListeners.removeElement(l);
  }

  protected synchronized void fireDataEntryModeChangeEvent(byte oldMode, byte newMode) {
    DataEntryModeChangeEvent d = new DataEntryModeChangeEvent(this, oldMode, newMode);
    for (Enumeration e=dataEntryModeChangeListeners.elements(); e.hasMoreElements(); ) {
      DataEntryModeChangeListener l = (DataEntryModeChangeListener) e.nextElement();
      l.dataEntryModeChange(d, oldMode, newMode);
    }
  }






  public void addDataChangeListener(DataChangeListener l) {
    dataChangeListeners.addElement(l);
  }

  public void removeDataChangeListener(DataChangeListener l) {
    dataChangeListeners.removeElement(l);
  }

  protected synchronized void fireDataChangeEvent() {
    DataChangeEvent d = new DataChangeEvent(this);
    for (Enumeration e=dataChangeListeners.elements(); e.hasMoreElements(); ) {
      DataChangeListener l = (DataChangeListener) e.nextElement();
      l.dataChange(d);
    }
  }




  protected void setMode(byte mode) {
    if (mode != currentMode) {
      fireDataEntryModeChangeEvent(currentMode, mode);
      currentMode = mode;
    }
  }


  // ---

  protected void insert() throws ObfException {
    throw new RuntimeException("insert() must be overriden in controller");
  }
  
  protected void update() throws ObfException {
    throw new RuntimeException("insert() must be overriden in controller");
  }

  protected void delete() throws ObfException {
    throw new RuntimeException("insert() must be overriden in controller");
  }


  abstract protected void setBusinessObjectsToNull();



  // check these out

  protected void beforeEdit()   throws ObfException {};
  protected void beforeNew()    throws ObfException {};
  protected void beforeDelete() throws ObfException {};
  protected void afterDelete()  throws ObfException {};
  protected void beforeInsert() throws ObfException {};
  protected void afterInsert()  throws ObfException {};
  protected void beforeUpdate() throws ObfException {};
  protected void afterUpdate()  throws ObfException {};
  protected void beforeCancel() {};
  protected void afterCancel()  {};

  protected void cancellingInsert() {};
  protected void cancellingUpdate() {};
  protected void cancellingDelete() {};


  // methods called when listening to button bar

  public void newData(DataEntryNewEvent e) {
    lastUid = (UniqueId) model.getData();
    model.setData(null);
    setBusinessObjectsToNull();
    try {
      beforeNew();
      setMode(DataEntryModeChangeEvent.NEW);

      fireDataChangeEvent();
    } catch (ObfException oe) {
      oe.showMessageDialog();
      model.setData(lastUid);
      setBusinessObjectsToNull();
      setMode(DataEntryModeChangeEvent.BROWSE);
      fireDataChangeEvent();

    }
  }


  public void editData(DataEntryEditEvent e) {
    try {
      beforeEdit();
      setMode(DataEntryModeChangeEvent.EDIT);
    } catch (ObfException oe) {
      oe.showMessageDialog();
      setBusinessObjectsToNull();
      setMode(DataEntryModeChangeEvent.BROWSE);
      fireDataChangeEvent();
    }
  }


  public void deleteData(DataEntryDeleteEvent e) {
    try {

      beforeDelete();
      delete();
      afterDelete();

    } catch (ObfException oe) {
      cancellingDelete();
      oe.showMessageDialog();
    
    } finally {
      setBusinessObjectsToNull();
      fireDataChangeEvent();
    }
  }


  public void saveData(DataEntrySaveEvent e) {
    boolean inserting = (currentMode == DataEntryModeChangeEvent.NEW);
    try {

      if (currentMode == DataEntryModeChangeEvent.NEW) {
        beforeInsert();
        insert();
        afterInsert();
      } else if (currentMode == DataEntryModeChangeEvent.EDIT) {
        beforeUpdate();
        update();
        afterUpdate();
      }

    } catch (ObfException oe) {
      if (currentMode == DataEntryModeChangeEvent.NEW) {
        cancellingInsert();
        model.setData(lastUid);
      
      } else if (currentMode == DataEntryModeChangeEvent.EDIT)
        cancellingUpdate();

      oe.showMessageDialog();

    } finally {
      setBusinessObjectsToNull();
      fireDataChangeEvent();
      setMode(DataEntryModeChangeEvent.BROWSE);
    }
  }



  public void cancelData(DataEntryCancelEvent e) {
    Debug.msg(this.getClass(), "CANCEL");

    if (currentMode == DataEntryModeChangeEvent.NEW)
      model.setData(lastUid);
    
    beforeCancel();
    setBusinessObjectsToNull();
    setMode(DataEntryModeChangeEvent.BROWSE);
    afterCancel();

    fireDataChangeEvent();
  }



  // check these out?

  public void updatedObject(UidBusinessObject busObject) {}
  public void insertedObject(UidBusinessObject busObject) {}
  public void deletedObject(UidBusinessObject busObject) {}
  public void refreshObjects(Class busClassClass) {}




  // --- Validation

  public void validityChanged(ValidityEvent ve) {
    boolean v = isValid();
    if (v != wasValid) {
      validityHandler.fireValidityEvent(this, v);
      wasValid = v;
    }
  }

  public void addValidityListener(ValidityListener vl) {
    validityHandler.addValidityListener(vl);
  }

  public void removeValidityListener(ValidityListener vl) {
    validityHandler.removeValidityListener(vl);
  }

  public abstract boolean isValid(); 



}
