/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.client;

import org.locomotive.obf.*;
import org.locomotive.obf.event.*;
import org.locomotive.obf.data.UniqueId;
import java.util.Vector;
import java.util.Enumeration;


/**
 * A manager for use as the client part of a two-tier system.
 * <p>
 * This class talks to the server using a <code>ConnectionToServer</code>
 * object, which can be either a local or remote conncetion.  This class
 * also provides implementation of the <code>Refresher</code> and
 * <code>RefreshListener</code> interfaces.
 *
 * @author David Warnock
 * @author Erik Erskine
 */

public class ObfManager implements Manager, RefreshListener {

  private static ObfManager instance = null;

  private ConnectionToServer connection;
  private RefreshManager refreshManager = new RefreshManager();


  protected ObfManager() {
  }


  /**
   * Return an instance of the <code>ObfManager</code>,
   * which is based on the singleton model.
   */

  public static ObfManager getInstance() {
    if (instance == null) {
      synchronized(ObfManager.class) {
        if (instance == null)
          instance = new ObfManager();
      }
    }
    return instance;
  }


  /**
   * Sets the connection that this manager will use to talk to a server.
   *
   * @param      con an connection to a server.
   */

  public void setConnection(ConnectionToServer con) {
    connection = con;
  }


  /**
   * Execute a business process and return the result.
   *
   * @param      process       the process to execute.
   * @exception  ObfException  if the process failed to execute successfully.
   */

  public Object doProcess(BusinessProcess process) throws ObfException {
    return connection.doProcess(process);
  }



  // Refresher

  /**
   * Adds the specified refresh listener to be notified of
   * refreshes from this object.
   *
   * @param      listener the listener to be added
   * @param      busClass the business object class to listen to
   */

  public void addRefreshListener(RefreshListener listener, Class busClass) {
    //System.out.println("manager addRefreshListener("+busClass+")");
    //try {
      if (refreshManager.isMonitoringClass(busClass) == false) {
        connection.addRefreshListener(this, busClass);
      }
    //} catch (ObfException oe) {
    //}

    refreshManager.addRefreshListener(listener, busClass);
  }


  /**
   * Removes the specified refresh listener so that it no longer receives
   * notification of refreshes from this object.
   *
   * @param      listener the listener to be removed
   * @param      busClass the business object class to listen to
   */

  public void removeRefreshListener(RefreshListener listener, Class busClass) {
    //System.out.println("manager removeRefreshListener("+busClass+")");
    refreshManager.removeRefreshListener(listener, busClass);
    //try {
      if (refreshManager.isMonitoringClass(busClass) == false) {
        connection.removeRefreshListener(this, busClass);
      }
    //} catch (ObfException oe) {
    //}

  }



  // RefreshListener

  /**
   * Invoked whenever this object receives notification of an
   * updated business object.
   *
   * @param      busObject the updated business object
   */

  public void updatedObject(UidBusinessObject busObject) {
    refreshManager.updatedObject(busObject);
  }


  /**
   * Invoked whenever this object receives notification of an
   * inserted business object.
   *
   * @param      busObject the inserted business object
   */

  public void insertedObject(UidBusinessObject busObject) {
    refreshManager.insertedObject(busObject);
  }


  /**
   * Invoked whenever this object receives notification of an
   * deleted business object.
   *
   * @param      busObject the deleted business object
   */

  public void deletedObject(UidBusinessObject busObject) {
    refreshManager.deletedObject(busObject);
  }


  /**
   * Called whenever this object receives notification of a business object
   * class that needs to be refreshed.
   *
   * @param      busClass the business object class that needs refreshing
   */

  public void refreshObjects(Class busClass) {
    refreshManager.refreshObjects(busClass);
  }






  // added by ee

  public Vector getBusinessObjects(Class busClass) throws ObfException {
    return (Vector) doProcess(new GetMultipleProcess(busClass));
  }

  public Vector getBusinessObjects(Class busClass, int iMaxRows) throws ObfException {
    return (Vector) doProcess(new GetMultipleProcess(busClass, iMaxRows));
  }

  public Vector getBusinessObjects(Class busClass, Object whereObj) throws ObfException {
    return (Vector) doProcess(new GetMultipleProcess(busClass, whereObj));
  }

  public Vector getBusinessObjects(Class busClass, Object whereObj, int iMaxRows) throws ObfException {
    return (Vector) doProcess(new GetMultipleProcess(busClass, whereObj, iMaxRows));
  }

  public Vector getBusinessObjects(Class busClass, Object whereObj, Object orderByObj) throws ObfException {
    return (Vector) doProcess(new GetMultipleProcess(busClass, whereObj, orderByObj));
  }

  public Vector getBusinessObjects(Class busClass, Object whereObj, Object orderByObj, int iMaxRows) throws ObfException {
    return (Vector) doProcess(new GetMultipleProcess(busClass, whereObj, orderByObj, iMaxRows));
  }

  public UidBusinessObject getBusinessObject(Class busClass, UniqueId whereObj) throws ObfException {
    return (UidBusinessObject) doProcess(new GetProcess(busClass, whereObj));
  }

  public BusinessObject getDefaultBusinessObject(Class busClass) throws ObfException {
    return (BusinessObject) doProcess(new GetDefaultProcess(busClass));
  }

  public UniqueId insertBusinessObject(UidBusinessObject busObj) throws ObfException {
    return (UniqueId) doProcess(new InsertProcess(busObj));
  }

  public void updateBusinessObject(UidBusinessObject busObj) throws ObfException {
    doProcess(new UpdateProcess(busObj));
  }

  public void deleteBusinessObject(UidBusinessObject busObj) throws ObfException {
    doProcess(new DeleteProcess(busObj));
  }



}


