/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.client.swing;

import org.locomotive.obf.BusinessObject;
import org.locomotive.obf.client.*;
import org.locomotive.obf.client.event.*;
import javax.swing.JPanel;


/**
 * A swing panel component that provides a view onto all or part of a
 * business object.  A panel contains a business object and subclasses
 * should add other swing components to view or edit fields in this
 * business object.
 *
 * @author Erik Erskine
 * @see    FixedPanel
 * @see    VariablePanel
 */

public abstract class AbstractPanel extends JPanel
implements DataEntryModeChangeListener {

  /**
   * The business object that this panel looks at.
   */

  protected BusinessObject busObject;


  /**
   * Creates a new panel.
   */

  protected AbstractPanel() {
  }


  /**
   * Creates a new panel that listens to the specified controller for
   * data entry mode changes.
   * <p>
   * <b>[Note: this listener never gets removed - need to fix this]</b>
   *
   * @see        AbstractPanel#browse()
   * @see        AbstractPanel#edit()
   */

  protected AbstractPanel(DataEntryController c) {
    c.addDataEntryModeChangeListener(this);
  }


  /**
   * Returns the business object that this panel is currently looking at.<br>
   * If this panel is a <code>FixedPanel</code>, this object will be constant,
   * if a <code>VariablePanel</code>, it can be changed.
   *
   * @see  VariablePanel.setBusinessObject(org.locomotive.obf.BusinessObject);
   */

  public BusinessObject getBusinessObject() {
    return busObject;
  }
  

  /**
   * Invoked when the controller (if there was one specified) changes data
   * entry mode.  Subclasses should override the <code>browse()</code> and
   * <code>edit()</code> methods instead of this.  This method should not be
   * called manually, instead the <code>browse()</code> and
   * <code>edit()</code> methods should be called.
   *
   * @see        AbstractPanel#browse()
   * @see        AbstractPanel#edit()
   */

  public final void dataEntryModeChange(DataEntryModeChangeEvent e,
                                        byte oldMode, byte newMode) {
    switch (newMode) {
    case DataEntryModeChangeEvent.BROWSE : browse(); break;
    case DataEntryModeChangeEvent.NEW    :
    case DataEntryModeChangeEvent.EDIT   : edit(); break;
    }
  }


  /**
   * Changes this panel's mode to browsing.  If a controller is specified this
   * method will be invoked whenever the panel receives notification of a
   * change in the data entry mode. Can also be called manually.
   * Subclasses should implement this method only if they are used for both
   * browsing and editing of business objects, and alter the panel
   * so it is suitable for viewing.
   *
   * @see        AbstractPanel#edit()
   */

  public void browse() {}


  /**
   * Changes this panel's mode to editing.  If a controller is specified this
   * method will be invoked whenever the panel receives notification of a
   * change in the data entry mode. Can also be called manually.
   * Subclasses should implement this method only if they are used for both
   * browsing and editing of business objects, and alter the panel
   * so it is suitable for editing only.
   *
   * @see        AbstractPanel#browse()
   */

  public void edit() {}

}

