/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.client.swing;

import org.locomotive.obf.client.*;
import org.locomotive.obf.client.event.*;
import javax.swing.JPanel;


/**
 * A swing panel component that can support browsing and editing modes.  If a
 * controller is specified this form will listen to it for changes in the data
 * entry mode, and adjust itself accordingly.  This class does not provide a
 * a view onto a business object in the way <code>AbstractPanel</code> does,
 * instead it would typically be used to contain several panels.
 *
 * @author Erik Erskine
 * @see    FixedPanel
 * @see    VariablePanel
 */

public class DataEntryForm extends JPanel implements DataEntryModeChangeListener {


  /**
   * Creates a new form.
   */

  public DataEntryForm() {
  }


  /**
   * Creates a new form that listens to the specified controller for
   * data entry mode changes.
   * <p>
   * <b>[Note: this listener never gets removed - need to fix this]</b>
   *
   * @see        DataEntryForm#browse()
   * @see        DataEntryForm#edit()
   */

  public DataEntryForm(DataEntryController c) {
    c.addDataEntryModeChangeListener(this);
  }


  /**
   * Invoked when the controller (if there was one specified) changes data
   * entry mode.  Subclasses should override the <code>browse()</code> and
   * <code>edit()</code> methods instead of this.  This method should not be
   * called manually, instead the <code>browse()</code> and
   * <code>edit()</code> methods should be called.
   *
   * @see        DataEntryForm#browse()
   * @see        DataEntryForm#edit()
   */

  public final void dataEntryModeChange(DataEntryModeChangeEvent e,
                                        byte oldMode, byte newMode) {
    switch (newMode) {
    case DataEntryModeChangeEvent.BROWSE : browse(); break;
    case DataEntryModeChangeEvent.NEW    :
    case DataEntryModeChangeEvent.EDIT   : edit(); break;
    }
  }


  /**
   * Changes this form's mode to browsing.  If a controller is specified this
   * method will be invoked whenever the form receives notification of a
   * change in the data entry mode. Can also be called manually.
   * Subclasses should implement this method only if they need to
   * differentiate between browsing and editing,
   *
   * @see        DataEntryForm#edit()
   */

  public void browse() {}


  /**
   * Changes this form's mode to editing.  If a controller is specified this
   * method will be invoked whenever the form receives notification of a
   * change in the data entry mode. Can also be called manually.
   * Subclasses should implement this method only if they need to
   * differentiate between browsing and editing,
   *
   * @see        DataEntryForm#browse()
   */

  public void edit() {}

}


