/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.client.swing;


import org.locomotive.obf.*;
import org.locomotive.obf.event.*;
import org.locomotive.obf.data.*;
import org.locomotive.obf.client.*;
import org.locomotive.obf.client.swing.*;
import org.locomotive.obf.client.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;



public class ObfTable extends JTable implements EditableField, ModificationListener, ListSelectionListener, TableModelListener, DataEntryModeChangeListener, FocusListener {

  private ObfTableModel tableModel;
  private Color normalColor;
  private UniqueIdModel uidModel;
  


  public ObfTable(ObfTableModel m) {
    super(m);
    tableModel = m;
    setEnabled(true);
    setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    setDataObject(new UniqueIdModel());
    normalColor = getSelectionBackground();
    addFocusListener(this);
  }


  public Class getDataObjectClass() {
    return UniqueIdModel.class;
  }

  // --- General


  // stops the tab key being used by the table
  public boolean isManagingFocus() {
    return false;
  }
  
  public JScrollPane createScrollingTable() {
    int c = tableModel.getColumnCount();
    JScrollPane sp = new JScrollPane(this);
    sp.setPreferredSize(new Dimension(c*100,100));
    sp.setMinimumSize(new Dimension(c*100,100));
    return sp;
  }

/*
  // put this in
  public void doCleanup() {
    if (controller != null) {
      controller.removeDataEntryModeChangeListener(this);
    }
    //lbm.removeTableModelListener(this);
    Debug.msg(getClass(), "Done Cleanup");
  }
*/



  // --- EditableField stuff ---

  public EditableDataObject getDataObject() {
    return uidModel;
  }


  public void setDataObject(EditableDataObject o) {
    if (o == null || o instanceof UniqueIdModel) {
      
      if (uidModel != null)
        uidModel.removeModificationListener(this);
      uidModel = (UniqueIdModel) o;
      if (uidModel != null)
        uidModel.addModificationListener(this);

    } else
      throw new IllegalArgumentException("setDataObject method in ObfTable was passed a "+o.getClass()+", expected an UniqueIdModel");

    // force table to act upon a modification
    if (uidModel != null) {
      modified(new ModificationEvent(uidModel));
    } else {
      clearSelection();
    }
  }


  public void modified(ModificationEvent me) {
    if (me.getSource() == uidModel) {
      UniqueId u = uidModel.getUniqueId();
      int row = (u != null) ? tableModel.findUniqueId(u) : -1;

      if (row >= 0) {
        setRowSelectionInterval(row, row);
      } else {
        clearSelection();
      }
    }
  }


  private void updateModel() {
    if (uidModel != null) {
      if (getHighlightedRow() >= 0) {
        // a row is selected, so write its uid into the document
        uidModel.setData((UniqueId) tableModel.getUniqueIdAt(getHighlightedRow()));
      } else {
        uidModel.setData(null);
      }
    }
  }




  // ---


  public void tableChanged(TableModelEvent tme) {
    int next = -1;

    //System.out.println("*** tableModel looks like this: "+getModel().getColumnCount()+", "+getModel().getRowCount());
    //((ObfTableModel) getModel()).printModel();

    Debug.msg(getClass(), "+ tableChanged, type="+tme.getType()+", highlighed row="+getHighlightedRow());

    if (tme.getType() == TableModelEvent.DELETE && getHighlightedRow() == tme.getLastRow())
      clearSelection();

    super.tableChanged(tme);

    paintImmediately(0, 0, (getSize()).width, (getSize()).height);

    Debug.msg(getClass(), "- tableChanged, highlighed row="+getHighlightedRow());

    // need to check here if the current row is changed or deleted

    //notifyDataChanged();
    //valueChanged(null);
    //t.paintImmediately(0, 0, (t.getSize()).width, (t.getSize()).height );
  }




  public void setRowSelectionInterval(int index0, int index1) {
    super.setRowSelectionInterval(index0, index1);
    ensureRowIsVisible(index0);
  }





  public void valueChanged(ListSelectionEvent lse) {
    
    if (lse.getValueIsAdjusting() == false) {
      Debug.msg(getClass(), "+ value changed (not adjusting)");
    }

    // call the superclass
    super.valueChanged(lse);

    if (lse.getValueIsAdjusting() == false) {
      updateModel();
      paintImmediately(0, 0, (getSize()).width, (getSize()).height);
      Debug.msg(getClass(), "- value changed (not adjusting)");
    }

  }




  // return the index of the row that is highlighted, or -1 if there is no row highlighted

  public int getHighlightedRow() {
    int[] r = getSelectedRows();
    return r.length > 0 ? r[0] : -1;
  }



  // select the first row in the table and return true if there is one, otherwise return false

  public boolean selectFirst() {
    Debug.msg(getClass(), "+ selectFirst");

    if (tableModel.getRowCount() > 0) {
      setRowSelectionInterval(0, 0);
      updateModel();
      Debug.msg(getClass(), "- selectFirst (yes)");
      return true;
    
    } else {
      clearSelection();
      updateModel();
      Debug.msg(getClass(), "- selectFirst (no)");
      return false;
    }
  }




  public void ensureRowIsVisible(int index) {
    Rectangle rowBounds = getCellRect(index, 0, true);
    if (rowBounds != null) {
      scrollRectToVisible(rowBounds);
    }
  }



  public void dataEntryModeChange(DataEntryModeChangeEvent e, byte oldMode, byte newMode) {
    //Debug.msg(getClass(), "source is "+e.getSource()+", controller is "+controller);
    //if (e.getSource() == controller) {
      setEnabled(newMode == DataEntryModeChangeEvent.BROWSE);
      setVisible(newMode == DataEntryModeChangeEvent.BROWSE);
    //}
  }



  public void focusGained(FocusEvent fe) {
    setSelectionBackground(normalColor.darker());
    setSelectionForeground(Color.white);
    paintImmediately(0, 0, (getSize()).width, (getSize()).height);
  }


  public void focusLost(FocusEvent fe) {
    setSelectionBackground(normalColor);
    setSelectionForeground(Color.black);
    paintImmediately(0, 0, (getSize()).width, (getSize()).height);
  }





}
