/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.client.swing;

import org.locomotive.obf.BusinessObject;
import org.locomotive.obf.client.*;
import org.locomotive.obf.client.event.*;
import javax.swing.JPanel;


/**
 * A panel that provides a view onto a business object.  The business object
 * can be changed using the <code>setBusinessObject()</code> method, so this
 * panel can look at different business objects.
 * The panel should provide a means of editing or viewing this object.<br>
 * If this panel is not required to look at different business objects then a
 * <code>FixedPanel</code> should be used for simplicity.
 *
 * @author Erik Erskine
 * @see    FixedPanel
 */

public abstract class VariablePanel extends AbstractPanel
implements DataChangeListener {

  /**
   * The class of business object this panel can look at
   */

  protected Class busClass;

  /**
   * The controller that this panel should use (optional)
   */
  protected DataEntryController controller;

  private boolean active = true;
  private boolean missedDataChange = false;
  

  /**
   * Creates a new panel that will look at business objects
   * of the specified class.
   *
   * @param      c the class of business object
   */

  protected VariablePanel(Class c) {
    busClass = c;
  }


  /**
   * Creates a new panel that will look at business objects
   * of the specified class.  This will listen to the controller for changes
   * in data entry mode, and for data change events (where the panel should
   * view a differemt business object).  If the panel is active it will ask
   * the controller for the new business object and display it.
   * <p>
   * <b>[Note: this listener never gets removed - need to fix this]</b>
   *
   * @see        AbstractPanel#browse()
   * @see        AbstractPanel#edit()
   */

  protected VariablePanel(Class c, DataEntryController ctr) {
    super(ctr);
    busClass = c;
    controller = ctr;
    controller.addDataChangeListener(this);
  }


  /**
   * Invoked by the controller (if there is one) when this panel should
   * display a different.  The panel will then ask the controller for this
   * object if it is active or when it next becomes active.
   *
   * @see        VariablePanel#isActive
   * @see        VariablePanel#setActive
   */

  public void dataChange(DataChangeEvent e) {
    BusinessObject busObj;
    if (active && controller != null) {
      busObj = controller.getBusinessObject(busClass);
      setBusinessObject(busObj);
      missedDataChange = false;
    } else {
      missedDataChange = true;
    }
  }


  /**
   * Sets the business object this panel should look at.
   *
   * @param      busObject the business object.  The class of the object must
   *             match that specified in the constructor.
   * @see        AbstractPanel#getBusinessObject()
   */

  public abstract void setBusinessObject(BusinessObject busObject);


  /**
   * Makes this panel active or inactive.  If a panel is active and it
   * receives a data change event, it will ask the controller for the
   * new business object.  If it is not active and it receives the event,
   * it will not ask the controller immediately but only when it becomes
   * active again.  This saves panels (and the controller) getting new
   * business objects when they do not need to (eg if they are not visible).
   * If an inactive panel receives several data change events it still
   * only needs to get one business object when it becomes active again.
   *
   * @param     b <code>true</code> to make the panel active,
   *            <code>false</code> to make the panel inactive
   * @see       VariablePanel#isActive()
   */

  public void setActive(boolean b) {
    if (b != active) {
      active = b;
      if (missedDataChange && controller != null) {
        BusinessObject busObj = controller.getBusinessObject(busClass);
        setBusinessObject(busObj);
        missedDataChange = false;
      }
    }
  }


  /**
   * Returns <code>true</code> if this panel is active,
   * <code>false</code> otherwise.
   * @see        VariablePanel#setActive(boolean)
   */

  public boolean isActive() {
    return active;
  }

}

