/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;

import org.locomotive.obf.event.*;
import java.io.*;


/**
 * A basic implementation of an <code>EditableDataObject</code> that
 * handles modification, validation, storage of original data, as well as
 * getting and setting the data itself.
 * <p>
 * Subclasses are responsible for implementing the <code>isValid</code> and
 * <code>isModified</code> methods, as well as <code>getDataClass</code>
 * which stores the class of the base data contained within this object.
 *
 * @author Erik Erskine
 *
 * @see    DataObject
 * @see    EditableDataObject
 * @see    BooleanModel
 * @see    EnumModel
 * @see    UniqueIdModel
 */

public abstract class AbstractEditableDataObject
implements EditableDataObject, Cloneable {

  /**
   * The base object contained within this object.
   */
  protected Object value;

  /**
   * The original base object as passed to the constructor.
   */
  protected Object originalValue;

  /**
   * Is the base object allowed to be <code>null</code>?
   */
  protected boolean allowNull;

  
  private transient ModificationManager modificationManager =
    new ModificationManager();
  private transient ValidityManager validityManager = new ValidityManager();



  /**
   * Creates a new <code>AbstractEditableDataObject</code> with current and
   * original values both <code>null</code>.
   */

  public AbstractEditableDataObject() {
    this(null);
  }


  /**
   * Creates a new <code>AbstractEditableDataObject</code> with current and
   * original values set to <code>orig</code>.
   */

  public AbstractEditableDataObject(Object orig) {
    value = orig;
    originalValue = orig;
  }


  /**
   * Returns the current value of the base object.
   */

  public Object getData() {
    return value;
  }
  

  /**
   * Sets the base object.  Also checks that <code>data</code> is of a
   * valid class, using <code>getDataClass()</code>.
   *
   * @param      data the new value
   * @exception  IllegalArgumentException if an object of an incompatible
   *             class was passed.
   */

  public void setData(Object data) {
    if (data == null || getDataClass().isAssignableFrom(data.getClass())) {
      value = data;
      modificationManager.fireModificationEvent(this);
    } else
      throw new ClassCastException("setData() method in "+getClass().getName()
                                   +" was passed "+data.getClass()
                                   +", expected "+getDataClass());
  }


  public void setDataQuietly(Object data) {
    if (data == null || getDataClass().isAssignableFrom(data.getClass())) {
      value = data;
    } else
      throw new ClassCastException("setDataQuietly() method in "
                                   +getClass().getName()
                                   +" was passed "+data.getClass()
                                   +", expected "+getDataClass());
  }


  /**
   * Return the original data.
   */
  
  public Object getOriginalData() {
    return originalValue;
  }


  /**
   * Returns a string representation of this object.
   */

  /*
  public String toString() {
    return getClass().getName()+
      "[value="+value+",originalValue="+originalValue+"]";
  }
  */

  public String toString() {
    return isModified() ? "("+value+"/"+originalValue+")" : "("+value+")";
  }



  // --- Validation ---

  /**
   * Specifies whether or not the base object is allowed to be null.
   * This is used for validation purposes.
   * @see        AbstractEditableDataObject#isNullAllowed
   */

  public void setNullAllowed(boolean b) {
    allowNull = b;
  }

  
  /**
   * Returns <code>true</code> if the base object is allowed to be null.
   * This is used for validation purposes although subclasses must make
   * use of it themselves in their <code>isValid()</code> implementation.
   * @see        AbstractEditableDataObject#setNullAllowed
   */

  public boolean isNullAllowed() {
    return allowNull;
  }
  
  
  /**
   * Adds the specified validity listener to receive validity events
   * from this object.
   *
   * @param      vl the listener to be added
   * @see        #removeValidityListener
   */

  public void addValidityListener(ValidityListener vl) {
    validityManager.addValidityListener(vl);
  }


  /**
   * Removes the specified validity listener so that it no longer receives
   * validity events from this object.
   *
   * @param      vl the listener to be added
   * @see        #addValidityListener
   */

  public void removeValidityListener(ValidityListener vl) {
    validityManager.removeValidityListener(vl);
  }



  // --- Modification ---


  /**
   * Adds the specified modification listener to receive modification events
   * from this object.
   *
   * @param      ml the listener to be added
   * @see        #removeModificationListener
   */
  
  public void addModificationListener(ModificationListener ml) {
    modificationManager.addModificationListener(ml);
  }
  
  
  /**
   * Removes the specified modification listener so that it no longer receives
   * modification events from this object.
   *
   * @param      ml the listener to be added
   * @see        #addModificationListener
   */

  public void removeModificationListener(ModificationListener ml) {
    modificationManager.removeModificationListener(ml);
  }



  // Serialization

  private void readObject(ObjectInputStream s) throws ClassNotFoundException, IOException {
    s.defaultReadObject();
    modificationManager = new ModificationManager();
    validityManager = new ValidityManager();
  }


  // Cloning

  public Object clone() throws CloneNotSupportedException {
    AbstractEditableDataObject copy = (AbstractEditableDataObject) super.clone();
    copy.modificationManager = new ModificationManager();
    copy.validityManager = new ValidityManager();
    return copy;
  }


}

