/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.math.BigDecimal;
import java.util.Locale;


/**
 * A document to handle editing of decimals.
 * <br>This uses arbitrary precision decimal numbers using the
 * <code>BigDecimal</code> class instead of floating point numbers.
 * <p>
 * [PENDING: more javadoc to be included]
 *
 * @author Erik Erskine
 * @see    java.math.BigDecimal
 * @see    IntegerDocument 
 * @see    StringDocument 
 * @see    DateDocument 
 */


public class BigDecimalDocument extends ObfDocument {

  private BigDecimal minValue = null;
  private BigDecimal maxValue = null;
  
  private DecimalFormat numberFormat;
  private DecimalFormatSymbols symbols;

  private int decimalPlaces = -1;
  private String displayPrefix = "";
  private String displaySuffix = "";



  /**
   * Creates a new document with the initial value <code>null</code>
   */

  public BigDecimalDocument() {
    this(null);
  }


  /**
   * Creates a new document with the initial value <code>d</code>,
   * converted immediately into a <code>BigDecimal</code>
   */
  
  public BigDecimalDocument(double d) {
    this(new BigDecimal(d));
  }


  /**
   * Creates a new document with the initial value <code>bd</code>
   */

  public BigDecimalDocument(BigDecimal bd) {
    super(bd);

    numberFormat =
      (DecimalFormat) DecimalFormat.getInstance(Locale.getDefault());
    symbols = numberFormat.getDecimalFormatSymbols();
    setData(bd);
  }
  

  /**
   * Always returns <code>BigDecimal.class</code>.
   */

  public Class getDataClass() {
    return BigDecimal.class;
  }


  public BigDecimal getBigDecimal() {
    return (BigDecimal) getData();
  }

  public BigDecimal getOriginalBigDecimal() {
    return (BigDecimal) getOriginalData();
  }

  public void setBigDecimal(BigDecimal bd) {
    setData(bd);
  }


/*
  // is this necessary?
  public boolean isNullAllowed() {
    return false;
  }
*/


  /**
   * Returns the decimal format associated with this document.
   */

  public DecimalFormat getNumberFormat() {
    return numberFormat;
  }


  /**
   * Sets the decimal format associated with this document.
   */

  public void setDecimalFormat(DecimalFormat df) {
    numberFormat = df;
  }


  /**
   * Updates the document text whenever the value changes.
   * This method should only be called automatically by the superclass.
   *
   * @see #documentChanged
   */

  public void updateText() {
    // alter the string, replacing the exisiting numbers
    // with the values from value

    if (value != null) {
      if (decimalPlaces >= 0)
        value = ((BigDecimal) value).setScale(decimalPlaces,
                                              BigDecimal.ROUND_HALF_UP);

      if (numberFormat != null)
        replaceDocumentText(addTrailingZeros(numberFormat.format(
          ((BigDecimal) value).doubleValue())));
    
    } else {
      replaceDocumentText("");
    }
  }



  /**
   * Updates the value whenever the document text changes.
   * This method should only be called automatically by the superclass.
   *
   * @see updateText
   */

  protected void documentChanged() {
    calculate();  
  }



  private void calculate() {
    try {
      value = new BigDecimal(numberFormat.parse(getDocumentText()).doubleValue());

    } catch (Exception e) {
      value = null;
    }
  }



  /**
   * Returns <code>true</code> if the string is valid.
   * <p>
   * [pending: more info on the validation routine - maybe not here]
   */

  public boolean isValid() {
    boolean valid = true;

    if (value == null)
      valid = isNullAllowed();
    else {
      if (minValue != null)
        valid = valid && ((BigDecimal) value).compareTo(minValue) >= 0;

      if (maxValue != null)
        valid = valid && ((BigDecimal) value).compareTo(maxValue) <= 0;
    }
    
    return valid;
  }


  // --- BigDecimal-specific stuff


  /**
   * Returns the minimum value allowed, or <code>null</code> if there is none
   */

  public BigDecimal getMinValue() {
    return minValue;
  }


  /**
   * Sets the minimum value allowed.
   *
   * @see        #getMinValue
   * @see        #getMaxValue
   * @see        #setMaxValue
   */
  
  public void setMinValue(BigDecimal bd) {
    minValue = bd;
  }


  /**
   * Returns the maximum value allowed, or <code>null</code> if there is none
   *
   * @see        #setMaxValue
   * @see        #getMinValue
   * @see        #setMinValue
   * @see        #setRange
   */

  public BigDecimal getMaxValue() {
    return maxValue;
  }


  /**
   * Returns the maximum value allowed, or <code>null</code> if there is none
   *
   * @see        #setMaxValue
   * @see        #getMinValue
   * @see        #setMinValue
   * @see        #setRange
   */

  public void setMaxValue(BigDecimal bd) {
    maxValue = bd;
  }
  

  /**
   * Returns the number of decimal places.
   *
   * @return the exact number of decimal places used, or <code>-1</code>
   *         if there is no limit
   * @see    #setDecimalPlaces
   */

  public int getDecimalPlaces() {
    return decimalPlaces;
  }


  /**
   * Sets the exact number of decimal places.
   *
   * @param      p <code>-1</code> means no limit
   * @see        #getDecimalPlaces
   */

  public void setDecimalPlaces(int p) {
    decimalPlaces = p;
  }
  


/*
  // are these ever used? / is this the correct place for them?
  public void setDisplayPrefix(String s) { displayPrefix = s; }
  public void setDisplaySuffix(String s) { displaySuffix = s; }
*/



  /**
   * Set the document to allow only unsigned values
   * (ie positive values).
   */
  
  public void setUnsigned() {
    if (minValue != null)
      minValue = minValue.max(new BigDecimal(0));
    else
      minValue = new BigDecimal(0);
  }


  /**
   * Explicitly sets the sign of the current value.
   *
   * @param      plus <code>true</code> to make the value positive
   * @param      plus <code>false</code> to make the value negative
   * @see        #inc
   * @see        #dec
   */

  public void setSign(boolean plus) {
    if (value != null) {
      if (plus)
        value = ((BigDecimal) value).abs();
      else
        value = ((BigDecimal) value).abs().negate();
      updateText();
    }
  }


  /**
   * Increment the value by <code>amount</code>.
   *
   * @see        #dec
   */

  public void inc(int amount) {
    if (value != null)
      value = ((BigDecimal) value).add(new BigDecimal(amount));
    else
      value = new BigDecimal(0);

    if (maxValue != null) value = ((BigDecimal) value).min(maxValue);
    if (minValue != null) value = ((BigDecimal) value).max(minValue);

    updateText();
  }
  

  /**
   * Decrement the value by <code>amount</code>.
   *
   * @see        #inc
   */

  public void dec(int amount) {
    inc(-amount);
  }
  




  // check this

  private String addTrailingZeros(String str) {
    if (decimalPlaces > 0) {
      
      int dppos = str.indexOf(symbols.getDecimalSeparator());
      if (dppos == -1) {
        str += symbols.getDecimalSeparator();
        dppos = str.length() - 1;
      }
      String fraction = str.substring(dppos);
      for (int x=fraction.length(); x<=decimalPlaces; x++)
        str += symbols.getZeroDigit();
    }

    return str;
  }


}
