/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;


import java.util.*;
import java.text.DateFormat;

 
/**
 * A document to handle editing of dates.
 * <p>
 * [PENDING: more javadoc to be included]
 *
 * @author Erik Erskine
 * @see    StringDocument 
 * @see    IntegerDocument 
 * @see    BigDecimalDocument 
 */


public class DateDocument extends ObfDocument {


  /**
   * The date format to use...
   */

  protected DateFormat dateFormat;

  private static final int PIVOT = 90;
  
  private Calendar cal = null;
  private int datePos=0, monthPos=1, yearPos=2;
  private boolean nullAllowed = false;
  private Calendar minCal = null, maxCal = null;



  /**
   * Creates a new document with the initial value <code>null</code>
   */

  public DateDocument() {
    this(null);
  }
  

  /**
   * Creates a new document with the initial value <code>d</code>
   */

  public DateDocument(Date d) {
    super(d);
    dateFormat = DateFormat.getDateInstance(DateFormat.SHORT, Locale.getDefault());
    doDateOrder();

    // get rid of the time!
    if (originalValue != null) {
      Calendar c = Calendar.getInstance();
      c.setTime((Date) originalValue);
      c.set(Calendar.HOUR, 0);
      c.set(Calendar.MINUTE, 0);
      c.set(Calendar.SECOND, 0);
      c.set(Calendar.MILLISECOND, 0);
      originalValue = c.getTime();
    }

    // this next line should be done as part of super()
    // maybe it isn'tworking because the date order is wrong?
    setData(d);
  }


  /**
   * Always returns <code>Date.class</code>.
   */

  public Class getDataClass() {
    return Date.class;
  }


  public Date getDate() {
    return (Date) getData();
  }

  public Date getOriginalDate() {
    return (Date) getOriginalData();
  }

  public void setDate(Date d) {
    setData(d);
  }



  /**
   * Returns the current date contained in this document.
   * @see        #setData
   */

  public Object getData() {
    return (cal == null) ? null : cal.getTime();
  }


  /**
   * Sets the current date contained in this document
   * @exception  ClassCastException if not passed a
   *             <code>Date</code> object.
   * @see        #getData
   */

  public void setData(Object data) {
    if (data == null) {
      cal = null;
      replaceDocumentText("");
    
    } else if (data instanceof Date) {
      cal = Calendar.getInstance();
      cal.setTime((Date) data);
      clearTimeOfDay();
      updateText();

    } else
      throw new ClassCastException("setData method in "+getClass().getName()+
                                   " was passed "+data.getClass().getName()+
                                   ", expected "+getDataClass().getName());
  }


  /**
   * Returns the date format associated with this document.
   */

  public DateFormat getDateFormat() {
    return dateFormat;
  }


  /**
   * Sets the date format associated with this document.
   */

  public void setDateFormat(DateFormat df) {
    dateFormat = df;
    doDateOrder();
    updateText();
  }




  
  /**
   * Updates the document text whenever the value changes.
   * This method should only be called automatically by the superclass.
   *
   * @see #documentChanged
   */

  public void updateText() {
    // tidies up the string in the document, formatting it in the short form
    // for the locale - uses the value of cal (correct after each keypress)
    // to produce the new string.
    // If cal == null (blank or invalid date), it DOES NOT clear the string -
    // this is correct
    // NB the call to replaceText() will have the side effect of calling
    // calculate() again but this is generally ok because calculate() doesn't
    // change the text so we don't loop, but it does cause a modification
    // event even if nothing has changed
    
    if (cal != null && dateFormat != null)
      replaceDocumentText(make4Digits(dateFormat.format(cal.getTime())));
  }
  


  /**
   * Updates the value whenever the document text changes.
   * This method should only be called automatically by the superclass.
   *
   * @see updateText
   */

  protected void documentChanged() {
    calculate();
  }



  private void calculate() {
    // take a string and works out what cal is, called every time the document
    // changes
    // the string must consist of 3 numbers, separated by some other character
    // the positions of day month and year as set by doDateOrder() are used
    // if a string could not be resolved, cal is set to null, otherwise it
    // contains the date
    // NB 2 digit years use the pivot and are converted to either 19xx or 20xx

    StringTokenizer st = new StringTokenizer(getDocumentText(), "/.- ");

    try {
      int[] val = new int[3];
      val[0] = Integer.parseInt(st.nextToken());
      val[1] = Integer.parseInt(st.nextToken());
      val[2] = Integer.parseInt(st.nextToken());

      if (val[yearPos] < 1000)
        val[yearPos] += (val[yearPos] < PIVOT) ? 2000 : 1900;

      cal = Calendar.getInstance();
      cal.set(Calendar.DATE, val[datePos]);
      cal.set(Calendar.MONTH, val[monthPos] - 1);
      cal.set(Calendar.YEAR, val[yearPos]);
      clearTimeOfDay();

    } catch (Exception e) {
      cal = null;
    }

  }
  

  /**
   * Returns <code>true</code> if the string is valid.
   * <p>
   * [pending: more info on the validation routine - maybe not here]
   */

  public boolean isValid() {

    if (cal == null)
      return isNullAllowed();
    else {
      if (minCal != null && minCal.after(cal)) return false;
      if (maxCal != null && maxCal.before(cal)) return false;
    }
    return true;
  }


  // --- Date stuff



  private void doDateOrder() {

    // takes the magic date 31 December 1999 and converts it into the short
    // format used by the required locale.
    // this arranges the day, month and year fields into the correct order for
    // that locale, examines this string, searching for the values 31, 12,
    // and 99 to determine the order of these values in strings typed in
    // by the user

    Calendar c = Calendar.getInstance();
    c.set(1999, Calendar.DECEMBER, 31);
    String s = dateFormat.format(c.getTime());
    
    int sd = s.indexOf("31");
    int sm = s.indexOf("12");
    int sy = s.indexOf("99");
  
    datePos = (sd>sm ? 1 : 0) + (sd>sy ? 1 : 0);
    monthPos = (sm>sd ? 1 : 0) + (sm>sy ? 1 : 0);
    yearPos = (sy>sd ? 1 : 0) + (sy>sm ? 1 : 0);
  }




  /**
   * Returns the earliest allowed date, or <code>null</code> if there is none
   *
   * @see        #setMinDate
   * @see        #getMaxDate
   * @see        #setMaxDate
   * @see        #setRange
   */

  public Date getMinDate() {
    return (minCal == null) ? null : minCal.getTime();
  }


  /**
   * Sets the earliest allowed date
   *
   * @param      d the earliest date, or <code>null</code> if there is none
   * @see        #getMinDate
   * @see        #getMaxDate
   * @see        #setMaxDate
   * @see        #setRange
   */

  public void setMinDate(Date d) {
    if (d != null) {
      minCal = Calendar.getInstance();
      minCal.setTime(d);
    } else
      minCal = null;
  }


  /**
   * Returns the latest allowed date, or <code>null</code> if there is none
   *
   * @see        #setMaxDate
   * @see        #getMinDate
   * @see        #setMinDate
   * @see        #setRange
   */

  public Date getMaxDate() {
    return (maxCal == null) ? null : maxCal.getTime();
  }


  /**
   * Sets the latest allowed date
   *
   * @param      d the latest date, or <code>null</code> if there is none
   * @see        #getMaxDate
   * @see        #getMinDate
   * @see        #setMinDate
   * @see        #setRange
   */

  public void setMaxDate(Date d) {
    if (d != null) {
      maxCal = Calendar.getInstance();
      maxCal.setTime(d);
    } else
      maxCal = null;
  }
  

  /**
   * Sets range of allowed dates
   *
   * @param      min the earliest date, or <code>null</code> if there is none
   * @param      max the latest date, or <code>null</code> if there is none
   * @see        #getMinDate
   * @see        #setMinDate
   * @see        #getMaxDate
   * @see        #setMaxDate
   */

  public void setRange(Date min, Date max) {
    (minCal = Calendar.getInstance()).setTime(min);
    (maxCal = Calendar.getInstance()).setTime(max);
  }




  /**
   * Increment the date by 1 day.
   * @see        #incDay
   * @see        #decDay
   * @see        #incMonth
   * @see        #decMonth
   * @see        #incYear
   * @see        #decYear
   */

  public void incDay() {
    alterDate(1,0,0);
  }


  /**
   * Decrement the date by 1 day.
   * @see        #decDay
   * @see        #incMonth
   * @see        #decMonth
   * @see        #incYear
   * @see        #decYear
   */

  public void decDay() {
    alterDate(-1,0,0);
  }


  /**
   * Increment the date by 1 month.
   * @see        #decMonth
   * @see        #incDay
   * @see        #decDay
   * @see        #incYear
   * @see        #decYear
   */

  public void incMonth() {
    alterDate(0,1,0);
  }


  /**
   * Decrement the date by 1 month.
   * @see        #incMonth
   * @see        #incDay
   * @see        #decDay
   * @see        #incYear
   * @see        #decYear
   */

  public void decMonth() {
    alterDate(0,-1,0);
  }


  /**
   * Increment the date by 1 year.
   * @see        #decYear
   * @see        #incDay
   * @see        #decDay
   * @see        #incMonth
   * @see        #decMonth
   */

  public void incYear() {
    alterDate(0,0,1);
  }


  /**
   * Decrement the date by 1 year.
   * @see        #incYear
   * @see        #incDay
   * @see        #decDay
   * @see        #incMonth
   * @see        #decMonth
   */

  public void decYear() {
    alterDate(0,0,-1);
  }
  

  /**
   * Alter the date by the specified days, months and years.
   * <br>Use negative numbers to decrement.
   *
   * @param      the number of days to increment by
   * @param      the number of months to increment by
   * @param      the number of years to increment by
   * @see        #setToday
   * @see        #setTomorrow
   * @see        #setYesterday
   */

  public void alterDate(int days, int months, int years) {
    if (cal != null) {
      cal.add(Calendar.DATE, days);
      cal.add(Calendar.MONTH, months);
      cal.add(Calendar.YEAR, years);
      updateText();
    }
  }


  /**
   * Set the date to today.
   *
   * @see        #setToday
   * @see        #setTomorrow
   * @see        #setYesterday
   */

  public void setToday() {
    cal = Calendar.getInstance();
    updateText();
  }


  /**
   * Set the date to tomorrow.
   * <br>This is not relative to the current value in the document,
   * use <code>incDay</code> to increment the day.
   *
   * @see        #setToday
   * @see        #setYesterday
   * @see        #incDay
   */

  public void setTommorrow() {
    cal = Calendar.getInstance();
    incDay();
  }


  /**
   * Set the date to yesterday.
   * <br>This is not relative to the current value in the document,
   * use <code>decDay</code> to decrement the day.
   *
   * @see        #setToday
   * @see        #setTomorrow
   * @see        #decDay
   */

  public void setYesterday() {
    cal = Calendar.getInstance();
    decDay();
  }





  // uses cal to examine a string representing cal
  // and includes the century if necessary

  private String make4Digits(String s) {
    String y4 = ""+cal.get(Calendar.YEAR);
    String y2 = y4.substring(2);

    int y4pos = s.indexOf(y4);
    // year is never in the middle
    int y2pos = yearPos == 0 ? s.indexOf(y2) : s.lastIndexOf(y2);

    if (y4pos == -1)
      s = s.substring(0, y2pos) + y4 + s.substring(y2pos+2);
    
    return s;
  }




  // added by ee 3 aug 99 - only want the date, not the time
  private void clearTimeOfDay() {
    if (cal != null) {
      cal.set(Calendar.HOUR, 0);
      cal.set(Calendar.MINUTE, 0);
      cal.set(Calendar.SECOND, 0);
      cal.set(Calendar.MILLISECOND, 0);
    }
  }


}
