/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;


import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.util.Locale;


/**
 * A document to handle editing of integers.
 * <p>
 * [PENDING: more javadoc to be included]
 *
 * @author Erik Erskine
 * @see    StringDocument 
 * @see    DateDocument 
 * @see    BigDecimalDocument 
 */


public class IntegerDocument extends ObfDocument {

  private Integer minVal = null, maxVal = null;
  private NumberFormat numberFormat;


  /**
   * Creates a new document with the initial value <code>null</code>
   */

  public IntegerDocument() {
    this((Integer) null);
  }


  /**
   * Creates a new document with the initial value <code>i</code>
   */
  
  public IntegerDocument(int i) {
    this(new Integer(i));
  }


  /**
   * Creates a new document with the initial value <code>i</code>
   */

  public IntegerDocument(Integer i) {
    super(i);
    numberFormat = DecimalFormat.getInstance(Locale.getDefault());
    setData(i);
  }


  /**
   * Always returns <code>Integer.class</code>.
   */

  public Class getDataClass() {
    return Integer.class;
  }


  /**
   * Returns the decimal format associated with this document.
   */

  public DecimalFormat getNumberFormat() {
    return (DecimalFormat) numberFormat;
  }


  /**
   * Sets the decimal format associated with this document.
   */

  public void setDecimalFormat(DecimalFormat df) {
    numberFormat = df;
  }


  /**
   * Updates the document text whenever the value changes.
   * This method should only be called automatically by the superclass.
   *
   * @see #documentChanged
   */

  public void updateText() {
    // tidies up the string in the document, formatting it nicely,
    // based on value
    // if value == null (blank or invalid number),
    // it DOES NOT clear the string - I think this is the desired behaviour

    // NB the call to replaceText() will have the side effect of calling
    // calculate() again - this is generally ok because calculate() doesn't
    // change the text so we don't loop :-), but it does cause a modification
    // event even if nothing has changed

    if (value != null) {
      if (numberFormat != null)
        replaceDocumentText(numberFormat.format(
          ((Integer) value).intValue()));
      else
        replaceDocumentText(value.toString());
    }
  }


  /**
   * Updates the value whenever the document text changes.
   * This method should only be called automatically by the superclass.
   *
   * @see updateText
   */

  protected void documentChanged() {
    calculate();
  }



  // take a string and works out what value is,
  // called every time the document changes

  private void calculate() {
    try {
      if (numberFormat != null)
        value = new Integer(numberFormat.parse(getDocumentText()).intValue());
      else
        value = new Integer(getDocumentText());
    } catch (Exception e) {
      value = null;
    }
  }

  

  /**
   * Returns <code>true</code> if the string is valid.
   * <p>
   * [pending: more info on the validation routine - maybe not here]
   */

  public boolean isValid() {
    if (value == null)
      return isNullAllowed();
    else {
      int i = ((Integer) value).intValue();
      if (minVal != null && (minVal.intValue() > i)) return false;
      if (maxVal != null && (maxVal.intValue() < i)) return false;
    }
    return true;
  }



  // --- Integer-specific stuff

  /**
   * Returns the minimum value allowed, or <code>null</code> if there is none
   *
   * @see        #setMinValue
   * @see        #getMaxValue
   * @see        #setMaxValue
   * @see        #setRange
   */

  public Integer getMinValue() {
    return minVal;
  }


  /**
   * Sets the minimum value allowed.
   *
   * @see        #getMinValue
   * @see        #getMaxValue
   * @see        #setMaxValue
   * @see        #setRange
   */
  
  public void setMinValue(int i) {
    minVal = new Integer(i);
  }


  /**
   * Sets the minimum value allowed.
   *
   * @param      i the minimum value, or <code>null</code> if there is none
   * @see        #getMinValue
   * @see        #getMaxValue
   * @see        #setMaxValue
   * @see        #setRange
   */

  public void setMinValue(Integer i) {
    minVal = i;
  }


  /**
   * Returns the maximum value allowed, or <code>null</code> if there is none
   *
   * @see        #setMaxValue
   * @see        #getMinValue
   * @see        #setMinValue
   * @see        #setRange
   */

  public Integer getMaxValue() {
    return maxVal;
  }


  /**
   * Sets the maximum value allowed.
   *
   * @see        #getMaxValue
   * @see        #getMinValue
   * @see        #getMaxValue
   * @see        #setRange
   */

  public void setMaxValue(int i) {
    maxVal = new Integer(i);
  }


  /**
   * Sets the maximum value allowed.
   *
   * @param      i the maximum value, or <code>null</code> if there is none
   * @see        #getMaxValue
   * @see        #getMinValue
   * @see        #setMinValue
   * @see        #setRangeValue
   */

  public void setMaxValue(Integer i) {
    maxVal = i;
  }


  /**
   * Sets the minimum and maximum values.
   *
   * @see        #getMinValue
   * @see        #setMinValue
   * @see        #getMaxValue
   * @see        #setMaxValue
   */

  public void setRange(int min, int max) {
    minVal = new Integer(min); maxVal = new Integer(max);
  }

  
  /**
   * Sets the minimum and maximum values.
   * <br><code>null</code> means no limit.
   *
   * @see        #getMinValue
   * @see        #setMinValue
   * @see        #getMaxValue
   * @see        #setMaxValue
   */

  public void setRange(Integer min, Integer max) {
    minVal = min; maxVal = max;
  }



  

  /**
   * Explicitly sets the sign of the current value.
   *
   * @param      plus <code>true</code> to make the value positive
   * @param      plus <code>false</code> to make the value negative
   * @see        #inc
   * @see        #dec
   */

  public void setSign(boolean plus) {
    if (value != null) {
      int i = ((Integer) value).intValue();
      value = plus ? new Integer(Math.abs(i)) : new Integer(0 - Math.abs(i));
      updateText();
    }
  }


  /**
   * Increment the value by <code>amount</code>.
   *
   * @see        #dec
   */

  public void inc(int amount) {
    if (value != null)
      value = new Integer(((Integer) value).intValue() + amount);
    else
      value = new Integer(0);

    int i = ((Integer) value).intValue();
    if (minVal != null && (minVal.intValue() > i)) value = minVal;
    if (maxVal != null && (maxVal.intValue() < i)) value = maxVal;
    updateText();
  }


  /**
   * Decrement the value by <code>amount</code>.
   *
   * @see        #inc
   */

  public void dec(int amount) {
    inc(-amount);
  }
  



}
