/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;

import org.locomotive.obf.event.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.io.*;


/**
 * An abstract implementation of a swing document that is an 
 * <code>EditableDataObject</code>.  This provides basic implementation
 * of handling modification, validation, storage of original data, as well as
 * getting and setting the data itself.
 * <p>
 * Subclasses are responsible for implementing the <code>isValid()</code>
 * method, as well as <code>getDataClass()</code>
 * which stores the class of the base data contained within this object.
 * <p>
 * A modification event is fired each time the document changes.
 *
 * @author Erik Erskine
 *
 * @see    StringDocument
 * @see    IntegerDocument
 * @see    DateDocument
 * @see    BigDecimalDocument
 */


public abstract class ObfDocument extends PlainDocument
implements EditableDataObject, DocumentListener, Cloneable {
  
  /**
   * The base object contained within this wrapper.
   */
  protected Object value;

  /**
   * The original base object as specified in the constructor.
   */
  protected Object originalValue;



  private boolean dirty = false, oldValidity = false;
  private transient ModificationManager modificationManager = 
    new ModificationManager();
  private transient ValidityManager validityManager = new ValidityManager();
  private boolean fireEvents = true;
  private boolean nullAllowed = true;



  /**
   * Sets both the current and original values to <code>null</code>.
   */

  public ObfDocument() {
    this(null);
  }


  /**
   * Sets both the current and original values to <code>null</code>.
   */

  public ObfDocument(Object orig) {
    originalValue = orig;
    addDocumentListener(this);
    setData(orig);
  }


  /**
   * Returns the current value of the base object.
   */

  public Object getData() {
    return value;
  }


  /**
   * Sets the base object.  Also checks that <code>data</code> is of a
   * valid class, using <code>getDataClass()</code>.
   * It then calls <code>updateText()</code> for the subclass to
   * insert the correct text into the document.
   *
   * @param      data the new value
   * @exception  ClassCastException if an object of an incompatible
   *             type (different to that specified by
   *             <code>getDataClass()</code>) was passed.
   * @see        #updateText
   */

  public void setData(Object data) {
    if (data == null || data.getClass() == getDataClass()) {
      value = data;
      updateText();
    } else
      throw new ClassCastException("setData method in "+getClass().getName()+
                                   " was passed "+data.getClass()+
                                   ", expected "+getDataClass());
  }


  /**
   * Sets the text in the swing document based on the value of the base
   * object.  Called whenever the value of the base object changes.
   * Subclasses must override this to set the correct text based on their
   * base object.
   *
   * @see        #setData
   */

  abstract public void updateText();


  /**
   * Return the original data.
   */

  public Object getOriginalData() {
    return originalValue;
  }



  // --- Validation ---

  public void setNullAllowed(boolean b) {
    nullAllowed = b;
  }

  public boolean isNullAllowed() {
    return nullAllowed;
  }


  /**
   * Adds the specified validity listener to receive validity events
   * from this object.
   *
   * @param      vl the listener to be added
   * @see        #removeValidityListener
   */

  public void addValidityListener(ValidityListener vl) {
    validityManager.addValidityListener(vl);
  }


  /**
   * Removes the specified validity listener so that it no longer receives
   * validity events from this object.
   *
   * @param      vl the listener to be added
   * @see        #addValidityListener
   */

  public void removeValidityListener(ValidityListener vl) {
    validityManager.removeValidityListener(vl);
  }



  // --- Modification ---


  /**
   * Return true if the document text has changed.
   */

  public boolean isModified() {
    Object o1 = getData();
    Object o2 = getOriginalData();

    if (o1 == null)
      return o2 != null;
    else
      return !o1.equals(o2);
  }


  /**
   * Adds the specified modification listener to receive modification events
   * from this object.
   *
   * @param      ml the listener to be added
   * @see        #removeModificationListener
   */

  public void addModificationListener(ModificationListener ml) {
    modificationManager.addModificationListener(ml);
  }


  /**
   * Removes the specified modification listener so that it no longer receives
   * modification events from this object.
   *
   * @param      ml the listener to be added
   * @see        #addModificationListener
   */

  public void removeModificationListener(ModificationListener ml) {
    modificationManager.removeModificationListener(ml);
  }
  


  // --- Internal stuff to listen to the document


  /**
   * Called when the document changes.<br>
   * This method should not normally be called manually or overridden.
   * The <code>documentChanged()</code> method enables subclasses to
   * respond to changes in the document.
   *
   * @see        #documentChanged
   */

  public void changedUpdate(DocumentEvent de) {
    if (fireEvents) processDocumentChange(de);
  }


  /**
   * Called when the document changes.<br>
   * This method should not normally be called manually or overridden.
   * The <code>documentChanged()</code> method enables subclasses to
   * respond to changes in the document.
   *
   * @see        #documentChanged
   */

  public void insertUpdate(DocumentEvent de)  {
    if (fireEvents) processDocumentChange(de);
  }


  /**
   * Called when the document changes.<br>
   * This method should not normally be called manually or overridden.
   * The <code>documentChanged()</code> method enables subclasses to
   * respond to changes in the document.
   *
   * @see        #documentChanged
   */

  public void removeUpdate(DocumentEvent de) {
    if (fireEvents) processDocumentChange(de);
  }





  private void processDocumentChange(DocumentEvent de) {
    if (de.getDocument() == this) {
      dirty = true;

      // do anything specific in a subclass before firing out
      // modification and validity events
      documentChanged();

      modificationManager.fireModificationEvent(this);

      boolean newValidity = isValid();
      if (newValidity != oldValidity) {
        validityManager.fireValidityEvent(this, newValidity);
        oldValidity = newValidity;
      }
    }
  }


  /**
   * Called whenever the document changes, but <i>before</i> any validation
   * or modification events are fired.  This should be overridden by
   * subclasses to recalculate the value of the base object (based on the
   * document's text string).
   */

  abstract protected void documentChanged();



  // --- Convenience methods ---

  /**
   * Returns the text in the document, removing the need to to call
   * <code>getText</code> and handle BadLocationExceptions that never occur.
   */

  protected synchronized String getDocumentText() {
    try {
      return getText(0, getLength());
    } catch (BadLocationException e) {
      return null;
    }
  }


  /**
   * The same as calling the <code>remove</code> and <code>insertString</code>
   * methods, but removes the need for BadLocationException checking and
   * also makes sure only a single document change event is fired
   * (for efficency reasons), so only one modification event will be fired.
   */

  protected synchronized void replaceDocumentText(String str) {

    // this normally generates two document events - for removing and
    // inserting, so the first is not fired (efficiency reasons)
    // if str is null or "", insertString will not fire it's event,
    // so we need to fire the remove event anyway!

    try {
      if (str != null && str.length() > 0)
        fireEvents = false;
      remove(0, getLength());
      
      fireEvents = true;
      insertString(0, str, null);
    } catch (BadLocationException e) {
      // never happens
    }
  }


  /**
   * Returns a string representation of this object.
   */

  public String toString() {
    return getClass().getName() + "[value="+value+",original="+originalValue+
      ",modified="+isModified()+",valid="+isValid()+"]";
  }



  // Serialization

  private void readObject(ObjectInputStream s) throws ClassNotFoundException, IOException {
    s.defaultReadObject();
    modificationManager = new ModificationManager();
    validityManager = new ValidityManager();
    addDocumentListener(this);
  }


  // Cloning

  public Object clone() throws CloneNotSupportedException {
    ObfDocument copy = (ObfDocument) super.clone();
    copy.listenerList = new EventListenerList();
    copy.modificationManager = new ModificationManager();
    copy.validityManager = new ValidityManager();
    copy.addDocumentListener(copy);
    return copy;
  }

}
