/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;

import javax.swing.text.*;


/**
 * A document supporting <code>EditableDataObject</code> to handle
 * strings.  This provides functionality to validate text, enforce and
 * check case, allow/forbid certain characters etc.
 * <p>
 * [PENDING: more javadoc to be included]
 *
 * @author Erik Erskine
 * @see    IntegerDocument
 * @see    BigDecimalDocument
 * @see    DateDocument
 */


public class StringDocument extends ObfDocument {
  

  /**
   * Constant to represent lowercase text.
   */
  public static final int LOWERCASE = -1;
  
  /**
   * Constant to represent mixed case text.
   */
  public static final int MIXEDCASE = 0;
  
  /**
   * Constant to represent uppercase text.
   */
  public static final int UPPERCASE = 1;
  
  /**
   * Constant to represent allowing specified characters.
   */
  public static final boolean ALLOW = true;
  
  /**
   * Constant to represent forbidding specified characters.
   */
  public static final boolean FORBID = false;



  private int maxLength = 0, minLength = 0;
  private String checkedCharacters = "";
  private boolean allowCheckedCharacters = FORBID;
  private int textCase = MIXEDCASE;
  private String pictureMask = "";




  /**
   * Create a new document with the initial value <code>null</code>.
   * <br><b>Note: check if is should be allowed?</b>
   */

  public StringDocument() {
    super();
  }


  /**
   * Create a new document with the initial value <code>orig</code>.
   */

  public StringDocument(String orig) {
    super(orig);
  }


  /**
   * Always returns <code>String.class</code>.
   */

  public Class getDataClass() {
    return String.class;
  }


  public String getString() {
    return (String) getData();
  }

  public String getOriginalString() {
    return (String) getOriginalData();
  }

  public void setString(String s) {
    setData(s);
  }



  /**
   * Updates the document text whenever the value changes.
   * This method should only be called automatically by the superclass.
   *
   * @see #documentChanged
   */

  public void updateText() {
    replaceDocumentText((String) value);
  }


  /**
   * Updates the value whenever the document text changes.
   * This method should only be called automatically by the superclass.
   *
   * @see updateText
   */

  protected void documentChanged() {
    value = getDocumentText();
  }




  // --- String-specific stuff


  /**
   * Returns the currently set text case, one of
   * <code>LOWERCASE</code>, <code>MIXEDCASE</code> or <code>UPPERCASE</code>.
   *
   * @see #setCase
   */
  
  public int getCase() {
    return textCase;
  }


  /**
   * Sets the currently set text case, one of
   * <code>LOWERCASE</code>, <code>MIXEDCASE</code> or <code>UPPERCASE</code>.
   * This will convert all characters entered to the desired case.
   *
   * @see #getCase
   */

  public void setCase(int c) {
    textCase = c;
  }


  /**
   * Returns the maximum length of string allowed.
   * @see        #setMaxLength
   * @see        #getMinLength
   * @see        #setMinLength
   */

  public int getMaxLength() {
    return maxLength;
  }


  /**
   * Sets the maximum length of string allowed.  If the string reaches this
   * length characters will not be inserted.
   *
   * @see        #getMaxLength
   * @see        #getMinLength
   * @see        #setMinLength
   * @see        #setLength
   */

  public void setMaxLength(int l) {
    maxLength = l;
  }


  /**
   * Returns the minimum length of string allowed.
   * @see        #setMinLength
   * @see        #getMaxLength
   * @see        #setMaxLength
   */

  public int getMinLength() {
    return minLength;
  }


  /**
   * Sets the minimum length of string allowed.
   *
   * @see        #getMinLength
   * @see        #getMaxLength
   * @see        #setMaxLength
   * @see        #setLength
   */

  public void setMinLength(int l) {
    minLength = l;
  }


  /**
   * Sets the maximum and minimum lengths.
   *
   * @see        #getMinLength
   * @see        #setMinLength
   * @see        #getMaxLength
   * @see        #setMaxLength
   */

  public void setLength(int min, int max) {
    minLength = min;
    maxLength = max;
  }


  /**
   * Returns the string of characters that are either allowed or forbidden.
   *
   * @see        #setCheckedCharacters
   * @see        #getAllowCheckedCharacters
   */

  public String getCheckedCharacters() {
    return checkedCharacters;
  }


  /**
   * Returns <code>true</code> if the checked characters are the only
   * characters to be allowed in the string, or <code>false</code> if
   * these characters are not allowed but any other character is.
   *
   * @see        #getCheckedCharacters
   * @see        #setCheckedCharacters
   */

  public boolean getAllowCheckedCharacters() {
    return allowCheckedCharacters;
  }


  /**
   * Sets the checked characters, their use.
   *
   * @param      s the string of characters
   * @param      b <code>true</code> if these should be the only characters
   *             that are allowed to be entered, <code>false</code> if
   *             these characters are not allowed, but any other character
   *             is.
   * @see        #getCheckedCharacters
   * @see        #getAllowCheckedCharacters
   */

  public void setCheckedCharacters(String s, boolean b) {
    checkedCharacters = s;
    allowCheckedCharacters = b;
  }



  // --- Validation

  private boolean containsInvalidCharacters(String s) {
    boolean found = false;

    if (checkedCharacters != null) {
      char[] checked = checkedCharacters.toCharArray();
      char[] text = s.toCharArray();

      for (int x=0; x<checked.length; x++)
	for (int y=0; y<text.length; y++)
	  if (checked[x] == text[y]) found = true;
    }
    return (found != allowCheckedCharacters);
  }

  
  private String convertCase(String s) {
    if (textCase > MIXEDCASE)
      return s.toUpperCase();
    else if (textCase < MIXEDCASE)
      return s.toLowerCase();
    else return s;
  }


  // create a temporary string with the text inserted that we can check
  private String testInsert(int offset, String str, AttributeSet attr) {
    String s = null;
    try {
      s = getText(0, offset) + str + getText(offset, (getLength()-offset));
    } catch (BadLocationException e) {
    }
    return s;
  }


  // validate text entry as the user types, not allowing any text to be
  // inserted that breaks the rules
  private boolean isValidEntry(int offset, String str, AttributeSet attr) {
    String s = testInsert(offset, str, attr);

    boolean result = (maxLength == 0 || s.length() <= maxLength) &&
      !containsInvalidCharacters(str);
  
    return result;
  }



  /**
   * Returns <code>true</code> if the string is valid.
   * <p>
   * [pending: more info on the validation routine - maybe not here]
   */

  public boolean isValid() {

    String s = null;
    try {
      s = getText(0, getLength());
    } catch (BadLocationException e) {
    }

    // test the string
    boolean result = (maxLength == 0 || s.length() <= maxLength) &&
      (s.length() >= minLength) && !containsInvalidCharacters(s);

    return result;

  }


  /**
   * Overrides <code>insertString</code> in <code>PlainDocument</code> to
   * perform some validation checks.
   *
   * @exception  BadLocationException in the same way PlainDocument does
   */


  public void insertString(int offset, String str, AttributeSet attr)
  throws BadLocationException {
    if (str != null && isValidEntry(offset, str, attr))
      super.insertString(offset, convertCase(str), attr);
  }


}
