/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;


/**
 * Implementation of <code>UniqueId</code> using an integer as an
 * identifier.<p>
 *
 * Contains an integer representing the identifier.  This integer can never be
 * <code>null</code> itself; instead the whole object is <code>null</code>.
 * This is to avoid discrepancies between the value <code>null</code> and
 * an object containing <code>null</code> as it's value.<p>
 *
 * <code>UniqueIntegerId</code> objects are constant, their values cannot be
 * changed. 
 *
 * @author Erik Erskine
 */

public final class UniqueIntegerId implements UniqueId {

  private int value;


  /**
   * Creates an <code>UniqueIntegerId</code> object with the value
   * <code>i</code>.
   *
   * @param      i the unique integer to act as the id
   */
  
  public UniqueIntegerId(int i) {
    value = i;
  }


  /**
   * Returns the integer value contained by this unique id
   */
  
  public int getInt() {
    return value;
  }


  /**
   * Tests this <code>UniqueIntegerId</code>object for equality with
   * another.<br>
   * Note: Future versions may use the <code>Comparable</code> interface
   * in Java 2.
   
   
   * @return    <code>0</code> if this object represents the same unique
   *            identifier as the argument; negative if this object is ordered
   *            before the argument; positive if this object is ordered after
   *            the argument.
   * @param     o the object to compare with this one
   * @see       UniqueIntegerId#equals(java.lang.Object);
   * @exception ClassCastException if the argument is not of type
   *            <code>UniqueIntegerId</code>
   */
  
  public int compareTo(Object o) {
    // this correctly throws ClassCastException or NullPointerException
    return value - ((UniqueIntegerId) o).value;
  }


  /**
   * Compares this UniqueIntegerId to the specified object.
   * The result is <code>true</code> if and only if the argument is not 
   * <code>null</code> and is a <code>UniqueIntegerId</code> object that
   * represents the same unique identifier as this object. 
   *
   * @param   obj  the object to compare this object with.
   * @return  <code>true</code> if they are equal;
   *          <code>false</code> otherwise.
   * @see     UniqueIntegerId#compareTo(java.lang.Object)
   * @see     UniqueIntegerId#hashCode()
   */

  public boolean equals(Object o) {
    if ((o != null) && (o instanceof UniqueIntegerId)) {
      return value == ((UniqueIntegerId) o).value;
    } else {
      return false;
    }
  }


  /**
   * Returns a hashcode for this identifier.
   *
   * @return  a hash code value for this object. 
   * @see     UniqueIntegerId#equals(java.lang.Object)
   */

  public int hashCode() {
    return value;
  }


  /**
   * Returns a string representation of this object
   */

  public String toString() {
    return "("+getClass().getName()+":"+value+")";
  }

}

