/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.data;


/**
 * Implementation of <code>UniqueId</code> using a string as an identifier.<p>
 *
 * Contains a string representing the identifier.  This string can never be
 * <code>null</code> itself; instead the whole object is <code>null</code>.
 * This is to avoid discrepancies between the value <code>null</code> and
 * an object containing <code>null</code> as it's value.<p>
 *
 * <code>UniqueStringId</code> objects are constant, their values cannot be
 * changed. 
 *
 * @author David Warnock
 * @author Erik Erskine
 */

public final class UniqueStringId implements UniqueId {

  private String value;


  /**
   * Creates an <code>UniqueStringId</code> object with the string
   * <code>str</code>.  Note that <code>str</code> must not be
   * <code>null</code>.  The static <code>create(String)</code> method
   * provides a safe way of creating the object and checking for 
   * <code>null</code>.
   *
   * @param      str the unique string to act as the id
   * @exception  NullPointerException if <code>str == null</code>
   * @see        UniqueStringId#create
   */
  
  public UniqueStringId(String str) {
    if (str == null)
      throw new NullPointerException(
        "UniqueStringId must not be constructed with null, "+
        "use the static create(String) method instead");

    value = str;
  }


  /**
   * A safe constructor method that checks for <code>null</code>.
   *
   * <p>The following example illustrates the use of this method:
   * <p> <code>UniqueStringId.create("a") </code>
   * <br><code>new UniqueStringId("a")    </code>
   * -- both return an id object with a value of <code>a</code>
   * <p> <code>new UniqueStringId(null)   </code>
   * -- fails with an NullPointerException
   * <br><code>UniqueStringId.create(null)</code>
   * --  returns <code>null</code> without trying to creating an instance
   *
   * @return  a UniqueStringId object representing the string s, or
   *          <code>null</code> if the <code>s == null</code>
   */
   
  public static UniqueStringId create(String s) {
    return(s == null) ? null : new UniqueStringId(s);
  }


  /**
   * Returns the string value contained by this unique id
   */
  
  public String getString() {
    return value;
  }


  /**
   * Tests this <code>UniqueStringId</code>object for equality with
   * another.<br>
   * Note: Future versions may use the <code>Comparable</code> interface
   * in Java 2.
   *
   * @return    <code>0</code> if this object represents the same unique
   *            identifier as the argument; negative if this object is ordered
   *            before the argument; positive if this object is ordered after
   *            the argument.  The ordering is lexicographic based on the
   *            strings.
   * @param     o the object to compare with this one
   * @see       UniqueStringId#equals(java.lang.Object);
   * @exception ClassCastException if the argument is not of type
   *            <code>UniqueStringId</code>
   */
  
  public int compareTo(Object o) {
    // this correctly throws ClassCastException or NullPointerException
    return value.compareTo(((UniqueStringId) o).value);
  }


  /**
   * Compares this UniqueStringId to the specified object.
   * The result is <code>true</code> if and only if the argument is not 
   * <code>null</code> and is a <code>UniqueStringId</code> object that
   * represents the same unique identifier as this object. 
   *
   * @param   obj  the object to compare this object with.
   * @return  <code>true</code> if they are equal;
   *          <code>false</code> otherwise.
   * @see     UniqueStringId#compareTo(java.lang.Object)
   * @see     UniqueStringId#hashCode()
   */

  public boolean equals(Object o) {
    if ((o != null) && (o instanceof UniqueStringId)) {
      return value.equals(((UniqueStringId) o).value);
    } else {
      return false;
    }
  }


  /**
   * Returns a hashcode for this identifier.
   *
   * @return  a hash code value for this object. 
   * @see     UniqueStringId#equals(java.lang.Object)
   */

  public int hashCode() {
    return value.hashCode();
  }


  /**
   * Returns a string representation of this object
   */

  public String toString() {
    return "("+getClass().getName()+":"+value+")";
  }

}

