/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.event;

import java.util.Vector;
import java.util.Enumeration;
import java.io.*;


/**
 * Provides a basic implementation of the <code>Modifiable</code> interface to
 * handle adding, removing and notification of modification listeners.
 * <p>
 * Classes implementing the <code>Modifiable</code> interface can use this
 * to avoid writing repetetive code to maintain a list of listeners.
 * <p>
 * The following example illustrates the use of this class:
 * <p>
 *
 * <pre>
 *     class Foo implements Modification ... {
 * &nbsp;
 *       private ModificationManager modificationHandler =
 *         new ModificationManager();
 * &nbsp;
 *       public boolean isModified() {
 *         ...
 *       }
 * &nbsp;
 *       public void addModificationListener(ModificationListener ml) {
 *         modificationHandler.addModificationListener(ml);
 *       }
 * &nbsp;
 *       public void removeModificationListener(ModificationListener ml) {
 *         modificationHandler.removeModificationListener(ml);
 *       }
 * &nbsp; 
 *       ...
 *      // some change occurs and you want to notify the listeners
 *      modificationHandler.fireModificationEvent(this);
 *       ...
 * &nbsp;
 *     } 
 * </pre>
 * 
 * @author Erik Erskine
 * @see org.locomotive.obf.event.Modifiable 
 * @see org.locomotive.obf.event.ModificationEvent 
 * @see org.locomotive.obf.event.ModificationListener 
 *
 */


public class ModificationManager implements Modifiable {

  /**
   * The list of listener objects.
   */

  protected Vector listeners = new Vector();


  /**
   * Meaningless, but it has to be here because this class is
   * <code>Modifiable</code>.
   */

  public boolean isModified() {
    throw new RuntimeException("Do not call isModified() in the "+
                               "ModificationManager - "+
                               "it's only meant for notifying listeners");
  }


  /**
   * Adds the specified modification listener to receive modification events
   * from this object.
   *
   * @param      ml the listener to be added
   * @see        #removeModificationListener
   */
  
  public synchronized void addModificationListener(ModificationListener ml) {
    listeners.addElement(ml);
  }


  /**
   * Removes the specified modification listener so that it no longer receives
   * modification events from this object.
   *
   * @param      ml the listener to be removed
   * @see        #addModificationListener
   */

  public synchronized void removeModificationListener(ModificationListener ml) {
    listeners.removeElement(ml);
  }
  

  /**
   * Notify all the listeners that the <code>source</code> object has changed
   *
   * @param      source the <code>Modifiable</code> object
   */

  public void fireModificationEvent(Modifiable source) {
    Enumeration e = listeners.elements();
    while (e.hasMoreElements()) {
      ModificationListener vl = (ModificationListener) e.nextElement();
      vl.modified(new ModificationEvent(source));
    }
  }


  public String toString() {
    return getClass().getName()+"["+listeners.size()+" listeners]";

  }

}



