/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.event;

import org.locomotive.obf.UidBusinessObject;
import java.util.*;


/**
 * Provides functionality to receive refreshes and in turn notify
 * listeners of these refreshes.  This class handles passing on of inserted,
 * updated and deleted business objects as well as changes to a business
 * class.
 * <p>
 * [PENDING: more documentation to follow]
 * 
 * @see org.locomotive.obf.event.Refresher 
 * @see org.locomotive.obf.event.RefreshListener 
 *
 * @author Erik Erskine
 */

public class RefreshManager implements Refresher, RefreshListener {

  private Hashtable listeners = new Hashtable();


  /**
   * Creates a new refresh manager
   */

  public RefreshManager() {
  }


  /**
   * Adds the specified refresh listener to receive refresh
   * notifications from this object.
   *
   * @param      listener the listener to be added
   * @param      busClass the business object class to listen to
   */

  public synchronized void addRefreshListener(RefreshListener listener,
                                              Class busClass) {
    Vector v = (Vector) listeners.get(busClass);

    if (v == null) {
      v = new Vector();
      listeners.put(busClass, v);
    }
    v.addElement(listener);
  }


  /**
   * Removes the specified refresh listener so that it no longer receives
   * refresh notifications from this object.
   *
   * @param      listner  the listener to be removed
   * @param      busClass the business object class to listen to
   */

  public synchronized void removeRefreshListener(RefreshListener listener,
                                                 Class busClass) {
    Vector v = (Vector) listeners.get(busClass);

    if (v != null) {
      v.removeElement(listener);
    }
  }



  // added by ee - need documentation
  // nb: it's important to clone the vector so that when we remove from it
  // we don't mess up the enumeration
  public synchronized void removeRefreshListeners(RefreshListener listener) {
    Enumeration eh = listeners.elements();
    while (eh.hasMoreElements()) {
      Vector v = (Vector) eh.nextElement();

      // for each key (busclass)
      Enumeration ev = ((Vector) v.clone()).elements();
      while (ev.hasMoreElements()) {

        // remove this listener if it's the one we're looking for
        RefreshListener rl = (RefreshListener) ev.nextElement();
        if (rl.equals(listener)) {
          v.removeElement(rl);
        }

      }
    }

  }

  public synchronized void removeRefreshListeners(Class busClass) {
    listeners.remove(busClass);
  }




  /**
   * Returns true if there are any listeners listening to the specified
   * business object class.
   *
   * @param      busClass the business object class
   */

  public synchronized boolean isMonitoringClass(Class busClass) {
    Vector v = (Vector) listeners.get(busClass);

    return (v != null && v.size() > 0);
  }



  /**
   * Invoked whenever this object receives notification of an
   * updated business object.
   *
   * @param      busObject the updated business object
   */

  public synchronized void updatedObject(UidBusinessObject busObject) {
    Vector v = (Vector) listeners.get(busObject.getClass());
    if (v != null) {
      Enumeration e = v.elements();
      while (e.hasMoreElements()) {
        RefreshListener l = (RefreshListener) e.nextElement();
        l.updatedObject(busObject);
      }
    }
  }


  /**
   * Invoked whenever this object receives notification of an
   * inserted business object.
   *
   * @param      busObject the inserted business object
   */

  public synchronized void insertedObject(UidBusinessObject busObject) {
    Vector v = (Vector) listeners.get(busObject.getClass());
    if (v != null) {
      Enumeration e = v.elements();
      while (e.hasMoreElements()) {
        RefreshListener l = (RefreshListener) e.nextElement();
        l.insertedObject(busObject);
      }
    }
  }


  /**
   * Invoked whenever this object receives notification of an
   * deleted business object.
   *
   * @param      busObject the deleted business object
   */

  public synchronized void deletedObject(UidBusinessObject busObject) {
    Vector v = (Vector) listeners.get(busObject.getClass());
    if (v != null) {
      Enumeration e = v.elements();
      while (e.hasMoreElements()) {
        RefreshListener l = (RefreshListener) e.nextElement();
        l.deletedObject(busObject);
      }
    }
  }


  /**
   * Invoked whenever this object receives notification of a business object
   * class that needs to be refreshed.
   *
   * @param      busClass the business object class that needs refreshing
   */

  public synchronized void refreshObjects(Class busClass) {
    Vector v = (Vector) listeners.get(busClass);
    if (v != null) {
      Enumeration e = v.elements();
      while (e.hasMoreElements()) {
        RefreshListener l = (RefreshListener) e.nextElement();
        l.refreshObjects(busClass);
      }
    }
  }


}
