/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.event;

import java.util.Vector;
import java.util.Enumeration;


/**
 * Provides a basic implementation of the <code>Validation</code> interface to
 * handle adding, removing and notification of validity listeners.
 * <p>
 * Classes implementing the <code>Validation</code> interface can use this
 * to avoid writing repetetive code to maintain a list of listeners.
 * <p>
 * The following example illustrates the use of this class:
 * <p>
 *
 * <pre>
 *     class Foo implements Validity ... {
 * &nbsp;
 *       private ValidityManager validityHandler = new ValidityManager();
 * &nbsp;
 *       public boolean isModified() {
 *         ...
 *       }
 * &nbsp;
 *       public void addValidityListener(ValidityListener ml) {
 *         validityHandler.addValidityListener(ml);
 *       }
 * &nbsp;
 *       public void removeValidityListener(ValidityListener ml) {
 *         validityHandler.removeValidityListener(ml);
 *       }
 * &nbsp; 
 *       ...
 *      // some change occurs and you want to notify the listeners
 *      validityHandler.fireValidityEvent(this);
 *       ...
 * &nbsp;
 *     } 
 * </pre>
 * 
 * @author Erik Erskine
 * @see org.locomotive.obf.event.Validation 
 * @see org.locomotive.obf.event.ValidityEvent 
 * @see org.locomotive.obf.event.ValidityListener 
 *
 */

public class ValidityManager implements Validation {


  /**
   * The list of listener objects.
   */

  protected Vector listeners = new Vector();


  /**
   * Meaningless, but it has to be here because this class is
   * <code>Validation</code>.
   */

  public boolean isValid() {
    throw new RuntimeException("Do not call isValid() in the ValidityManager"+
                               " - it's only meant for notifying listeners");
  }


  /**
   * Adds the specified validity listener to receive validity events
   * from this object.
   *
   * @param      vl the listener to be added
   * @see        #removeValidityListener
   */

  public synchronized void addValidityListener(ValidityListener vl) {
    listeners.addElement(vl);
  }


  /**
   * Removes the specified validity listener so that it no longer receives
   * validity events from this object.
   *
   * @param      vl the listener to be removed
   * @see        #addValidityListener
   */

  public synchronized void removeValidityListener(ValidityListener vl) {
    listeners.removeElement(vl);
  }
  

  /**
   * Notify all the listeners that the <code>source</code> object has changed
   *
   * @param      source the <code>Validation</code> object
   */
  
  public void fireValidityEvent(Validation source, boolean validity) {
    Enumeration e = listeners.elements();
    while (e.hasMoreElements()) {
      ValidityListener vl = (ValidityListener) e.nextElement();
      vl.validityChanged(new ValidityEvent(source, validity));
    }
  }
  
}


