/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.server;

import org.locomotive.obf.*;
import org.locomotive.obf.data.UniqueId;
import java.sql.*;
import java.util.Vector;


/**
 * Interface for business object handler classes.  The business object
 * handlers are responsible for retreiving, inserting, updating and
 * deleting objects from the database.  Notification of changes is not done
 * here - instead this is done in the process handlers.
 * <p>
 * [Pending: more documentation to follow]
 *
 * @author David Warnock
 * @author Erik Erskine
 * @see    
 */

public interface BusinessObjectHandler {

  /**
   * Returns a vector of business objects.
   *
   * @param      busClass  the class of business object to get.  This class
   *                       must implement the <code>UidBusinessObject</code>
   *                       interface
   * @param      where     an object to use to decide which objects to get,
   *                       or <code>null</code> to get all objects
   * @param      order     an object to use to decide the order in which
   *                       business objects are returned, or <code>null</code>
   *                       if this is not important
   * @param      limit     an upper limit on how many objects to get, or
   *                       <code>-1</code> to indicate no limit
   * @param      con       a database connection to use
   *
   * @exception  SQLException if a database-access error occurs
   * @exception  ClassCastException if the handler cannot get objects
   *             of type <code>busClass</code>, or the handler did not
   *             understand the type of <code>where</code> or
   *             <code>order</code>
   * @exception  ObfException if another error occurs
   */

  public Vector getBusinessObjects(Class busClass, Object where, Object order,
                                   int max, Connection con)
  throws ObfException, SQLException;


  /**
   * Returns a single business object, or <code>null</code> if it was not
   * found.
   *
   * @param      busClass  the class of business object to get.  This class
   *                       must implement the <code>UidBusinessObject</code>
   *                       interface
   * @param      uniqueId  the unique identifier of the business object to get
   * @param      con       a database connection to use
   *
   * @exception  SQLException if a database-access error occurs
   * @exception  ClassCastException if the handler cannot get objects
   *             of type <code>busClass</code>
   * @exception  ObfException if another error occurs
   */

  public UidBusinessObject getBusinessObject(Class busClass,
                                             UniqueId uniqueId,
                                             Connection con)
  throws ObfException, SQLException;


  /**
   * Returns the default business object for a given class.
   *
   * @param      busClass  the class of business object to get.  This class
   *                       must implement the <code>UidBusinessObject</code>
   *                       interface
   * @param      con       a database connection to use
   *
   * @exception  SQLException if a database-access error occurs
   * @exception  ClassCastException if the handler cannot get objects
   *             of type <code>busClass</code>
   * @exception  ObfException if another error occurs
   */

  public BusinessObject getDefaultBusinessObject(Class busClass,
                                                 Connection con)
  throws ObfException, SQLException;



  /**
   * Inserts a business object into the database, allocating it a
   * unique identifier
   *
   * @param      busObject the business object to insert.
   * @param      con       a database connection to use
   *
   * @exception  SQLException if a database-access error occurs
   * @exception  ClassCastException if the handler cannot insert objects
   *             of the type of busObject
   * @exception  ObfException if another error occurs
   */

  public UniqueId insertBusinessObject(UidBusinessObject busObject,
                                       Connection con)
  throws ObfException, SQLException;


  /**
   * Updates a business object in the database
   *
   * @param      busObject the business object to update.
   * @param      con       a database connection to use
   *
   * @exception  SQLException if a database-access error occurs
   * @exception  ClassCastException if the handler cannot update objects
   *             of the type of busObject
   * @exception  ObfException if another error occurs
   */

  public void updateBusinessObject(UidBusinessObject busObject,
                                   Connection con)
  throws ObfException, SQLException;


  /**
   * Deletes a business object from the database
   *
   * @param      busObject the business object to delete.
   * @param      con       a database connection to use
   *
   * @exception  SQLException if a database-access error occurs
   * @exception  ClassCastException if the handler cannot delete objects
   *             of the type of busObject
   * @exception  ObfException if another error occurs
   */

  public void deleteBusinessObject(UidBusinessObject busObject,
                                   Connection con)
  throws ObfException, SQLException;

}
