/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.server;

import org.locomotive.obf.*;
import org.locomotive.obf.event.*;
import java.util.*;


/**
 * Server-side thread that handles notification to clients of inserted,
 * updated and deleted business objects that have been reloaded, and also
 * of business object classes that need reloading.
 *.<p>
 * The thread handles adding and removing of listeners, and adding of
 * business objects that have been reloaded<code to the queue.  The thread
 * invokes the <code>RefreshListener</code>methods of all listeners.
 * It will be interrupted whenever a reloaded business object is added to
 * the queue but it is a  low priority thread and should not hold up the
 * server at peak activity.
 *
 * @author Erik Erskine
 */

public class ClientNotificationQueue extends Thread implements Refresher, RefreshListener {

  private static final int INSERT  = 0;
  private static final int UPDATE  = 1;
  private static final int DELETE  = 2;
  private static final int REFRESH = 3;


  private class QueueElement {
    int type;
    UidBusinessObject busObject;
    Class busClass;

    QueueElement(int t, UidBusinessObject o, Class c) {
      type = t;
      busObject = o;
      busClass = c;
    }
  }

  private RefreshManager refreshManager = new RefreshManager();
  private Vector changes = new Vector(); // of QueueElements

  private static ClientNotificationQueue instance;



  private ClientNotificationQueue() {
    setPriority(Thread.MIN_PRIORITY);
  }


  /**
   * Returns the singleton instance of the client notification queue.
   * If the thread is not already running it is started.
   */

  public static ClientNotificationQueue getInstance() {
    if (instance == null) {
      synchronized (ClientNotificationQueue.class) {
        if (instance == null) {
          instance = new ClientNotificationQueue();
          instance.start();
        }
      }
    }
    return instance;
  }


  /**
   * This method should not be called manually, use <code>getInstance()</code>
   * to start the thread.
   * @see        ClientNotificationQueue#getInstance()
   */

  public void run() {
    while (true) {
      examineQueue();
      suspend();
    }
  }


  /**
   * Add notification of an inserted business object to the queue.
   *
   * @param      busObject the inserted business object
   */

  public void insertedObject(UidBusinessObject busObject) {
    QueueElement qe = new QueueElement(INSERT, busObject,
                                       busObject.getClass());
    changes.addElement(qe);
    resume();
  }


  /**
   * Add notification of an updated business object to the queue.
   *
   * @param      busObject the updated business object
   */

  public void updatedObject(UidBusinessObject busObject) {
    QueueElement qe = new QueueElement(UPDATE, busObject,
                                       busObject.getClass());
    changes.addElement(qe);
    resume();
  }


  /**
   * Add notification of an deleted business object to the queue.
   *
   * @param      busObject the deleted business object
   */

  public void deletedObject(UidBusinessObject busObject) {
    QueueElement qe = new QueueElement(ReloadNotice.DELETE,
                                       busObject, busObject.getClass());
    changes.addElement(qe);
    resume();
  }


  /**
   * Add notification of a business object class
   * that needs refreshing to the queue.
   *
   * @param      busClass the business object class
   */

  public void refreshObjects(Class busClass) {
    QueueElement qe = new QueueElement(ReloadNotice.REFRESH, null, busClass);
    changes.addElement(qe);
    resume();
  }




  /**
   * Adds the specified refresh listener to be notified of
   * refreshes placed on the queue.
   *
   * @param      listener the listener to be added
   * @param      busClass the business object class to listen to
   */

  public void addRefreshListener(RefreshListener listener, Class busClass) {
    refreshManager.addRefreshListener(listener, busClass);
  }


  /**
   * Removes the specified refresh listener so that it no longer receives
   * notification of refreshes placed on the queue.
   *
   * @param      listener the listener to be removed
   * @param      busClass the business object class to listen to
   */

  public void removeRefreshListener(RefreshListener listener, Class busClass) {
    refreshManager.removeRefreshListener(listener, busClass);
  }

  public void removeRefreshListeners(RefreshListener listener) {
    refreshManager.removeRefreshListeners(listener);
  }

  public void removeRefreshListener(Class busClass) {
    refreshManager.removeRefreshListeners(busClass);
  }


  // -------------------------

  // called by the thread

  private synchronized void examineQueue() {
    Enumeration e = changes.elements();
    while (e.hasMoreElements()) {

      QueueElement qe = (QueueElement) e.nextElement();
      switch (qe.type) {
      case INSERT:
        refreshManager.insertedObject(qe.busObject);
        break;
      case UPDATE:
        refreshManager.updatedObject(qe.busObject);
        break;
      case DELETE:
        refreshManager.deletedObject(qe.busObject);
        break;
      case REFRESH:
        refreshManager.refreshObjects(qe.busClass);
        break;
      }
    }
    changes = new Vector();
  }



}


