/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.server;

import org.locomotive.obf.data.UniqueId;


/**
 * Notice of a business object that has been inserted, updated or deleted, or
 * a complete business object class that has been changed.  This class does
 * not contain the new business object, it merely implies that the objects
 * need to be reloaded.
 *
 * @author David Warnock
 * @author Erik Erskine
 * @see    ReloadListener
 * @see    ReloadQueue
 */

public class ReloadNotice {

  /**
   * A business object has been inserted (0)
   */
  public static final int INSERT = 0;

  /**
   * A business object has been updated (1)
   */
  public static final int UPDATE = 1;

  /**
   * A business object has been deleted (2)
   */
  public static final int DELETE = 2;

  /**
   * A business object class needs to be reloaded (3)
   */
  public static final int REFRESH = 3;



  private int type;
  private Class busClass;
  private UniqueId id;


  /**
   * Creates a new reload notice
   *
   * @param      type can be one of <code>INSERT</code>, <code>UPDATE</code>,
   *             <code>DELETE</code> or <code>REFRESH</code>
   * @param      busClass the class of business object to be reloaded
   * @param      id the unique id of the exact business object (in the case of
   *             <code>REFRESH</code> this is ignored)
   * @exception  InvalidArgumentException if the type was not valid.
   */

  public ReloadNotice(int type, Class busClass, UniqueId id) {

    if (type == INSERT || type == UPDATE ||
        type == DELETE || type == REFRESH) {
      this.type = type;
      this.busClass = busClass;
      this.id = id;
    } else
      throw new IllegalArgumentException("Invalid type: "+type);
  }


  /**
   * Returns the type of reload notice.
   *
   * @return     <code>INSERT</code>, <code>UPDATE</code>,
   *             <code>DELETE</code> or <code>REFRESH</code>
   */

  public int getType() {
    return type;
  }


  /**
   * Returns the class of business object to be reloaded.
   */

  public Class getBusClass() {
    return busClass;
  }


  /**
   * Returns the unique identifier of the exact business object to be
   * reloaded in the case of insert, update or delete.  This value is
   * <code>null</code> if a complete reload of a class is necessary.
   */

  public UniqueId getUniqueId() {
    return id;
  }


}
