/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.server;

import org.locomotive.obf.data.UniqueId;
import java.sql.Connection;
import java.util.*;


/**
 * Server-side thread that handles notifications of inserted, updated and
 * deleted business objects and business object classes, that need reloading.
 * <p>
 * The thread handles adding and removing of listeners, and adding of
 * <code>ReloadNotice</code> objects to the queue.  The thread invokes the
 * <code>reload()</code> method of all listeners.  It will be interrupted
 * whenever a <code>ReloadNotice</code> is added to the queue but it is a 
 * low priority thread and should not hold up the server at peak activity.
 *
 * @author David Warnock
 * @author Erik Erskine
 * @see    ReloadNotice
 * @see    ReloadListener
 */

public class ReloadQueue extends Thread {


   private static ReloadQueue instance;

   private Hashtable listeners = new Hashtable();
   // Hashtable of Vectors of Classes (reloadListeners), key is busClass

   private Vector notices = new Vector();
   // Vector of reloadNotices


   private Connection reloadCon = null;

   private ReloadQueue() {
      setPriority(Thread.MIN_PRIORITY);
   }


   /**
    * Returns the singleton instance of the reload queue.
    * If the thread is not already running it is started.
    */

   public static ReloadQueue getInstance() {
      if (instance == null) {
         synchronized (ReloadQueue.class) {
            if (instance == null) {
               instance = new ReloadQueue();
               instance.start();
            }
         }
      }
      return instance;
   }


   /**
    * This method should not be called manually, use <code>getInstance()</code>
    * to start the thread.
    * @see        ReloadQueue#getInstance()
    */

   public void run() {
      while (true) {
         doReload();
         suspend();
      }
   }


   /**
    * Adds a reload notice to the queue.
    *
    * @see        ReloadQueue#addReloadNotices(java.util.Vector)
    */

   public void addReloadNotice(ReloadNotice notice) {
      notices.addElement(notice);
      resume();
   }


   /**
    * Adds a series of reload notices to the queue in one go.  This is useful
    * when building up a series of notices during a transaction and adding them
    * to the queue when that transaction is committed.
    *
    * @param      noticeVector a vector of <code>ReloadNotice</code> objects
    * @see        ReloadQueue#addReloadNotice(ReloadNotice)
    */

   public void addReloadNotices(Vector noticeVector) {
      Enumeration e = noticeVector.elements();

      while (e.hasMoreElements())
         notices.addElement((ReloadNotice) e.nextElement());

      resume();
   }


   /**
    * Add a class that is a <code>ReloadListener</code> that should listen for
    * reload notices for a specified business object class.  On recieving these
    * notices an instance of that class should be obtained that should then
    * handle the reloading.
    *
    * @param      listenerClass the class of listener for which an instance of
    *             that class should receive notification.
    * @param      busClass the class of business objects to listen to.
    * @see ReloadQueue#removeReloadListener(java.lang.Class, java.lang.Class);
    * @see ReloadQueue#removeReloadListeners(java.lang.Class);
    */

   public void addReloadListener(Class listenerClass, Class busClass) {
      // check this class can be a ReloadListener
      Vector v = (Vector) listeners.get(busClass);
      if (v == null) {
         v = new Vector();
         listeners.put(busClass, v);
      }
      v.addElement(listenerClass);
   }


   /**
    * Removes a class that is a <code>ReloadListener</code> so that it no longer
    * listens for reload notices for a specified business object class.
    * <p>
    * <b>Not implemented yet</b>
    *
    * @param      listenerClass the class of listener.
    * @param      busClass the class of business objects to stop listening to.
    * @see ReloadQueue#addReloadListener(java.lang.Class, java.lang.Class);
    * @see ReloadQueue#removeReloadListeners(java.lang.Class);
    */

   public void removeReloadListener(Class listenerClass, Class busClass) {
      throw new RuntimeException("removeReloadListener not supported yet");
   }


   /**
    * Removes a class that is a <code>ReloadListener</code> so that it no longer
    * listens for reload notices for any business object classes.
    * <p>
    * <b>Not implemented yet</b>
    *
    * @param      listenerClass the class of listener.
    * @see ReloadQueue#addReloadListener(java.lang.Class, java.lang.Class);
    * @see ReloadQueue#removeReloadListener(java.lang.Class, java.lang.Class);
    */

   public void removeReloadListeners(Class listenerClass) {
      throw new RuntimeException("removeReloadListeners not supported yet");
   }


   // called by the thread

   private synchronized void doReload() {
      Enumeration e = notices.elements();
      while (e.hasMoreElements()) {
         ReloadNotice notice = (ReloadNotice) e.nextElement();
         fireReload(notice);
      }
      notices = new Vector();
   }


   private void fireReload(ReloadNotice notice) {
      Vector v = (Vector) listeners.get(notice.getBusClass());

      if (v != null) {
         Enumeration e = v.elements();
         while (e.hasMoreElements()) {
            Class listenerClass = (Class) e.nextElement();

            try {
               // better to get it from some sort of pool
               ReloadListener r = (ReloadListener) listenerClass.newInstance();

               if (reloadCon == null) reloadCon = ServerSupporter.getInstance().getConnection();
               r.reload(notice, reloadCon);
            } catch (InstantiationException ie) {
               // do something
               
            } catch (IllegalAccessException ie) {
               // do something
               
            }

         }
      }

   }



}


