/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.server;

import org.locomotive.obf.*;
import org.locomotive.loco.Loco;
import org.locomotive.server.*;
import java.sql.*;
import java.util.*;


/**
 * A utility class that provides functionality useful to servers
 * and handlers.  This includes:
 * <ul>
 * <li>mapping of business objects and processes to handlers
 * <li>allocation of handlers - this class will eventually use
 *     pools to keep a handlers available and avoid creating them
 *     each time they are used
 * <li>allocating of database connections
 * <li>starting, committing and rolling back transactions
 *     for a given connection.
 * </ul>
 * <p>
 * [pending: more javadoc to follow]
 *
 * @author David Warnock
 * @author Erik Erskine
 * @see    
 */

public class ServerSupporter {

   // made protected so that subclasses can be used.
   protected static ServerSupporter instance;

   private Hashtable businessObjectHandlers = new Hashtable();
   private Hashtable processHandlers = new Hashtable();

   protected boolean useLocalDatabase = false;
   protected String jdbcDriver, url, user, password;


   protected ServerSupporter() {

      // set the default process handlers
      setProcessHandler(GetProcess.class, GetProcessHandler.class);
      setProcessHandler(GetDefaultProcess.class,
                        GetDefaultProcessHandler.class);
      setProcessHandler(GetMultipleProcess.class,
                        GetMultipleProcessHandler.class);
      setProcessHandler(InsertProcess.class, InsertProcessHandler.class);
      setProcessHandler(UpdateProcess.class, UpdateProcessHandler.class);
      setProcessHandler(DeleteProcess.class, DeleteProcessHandler.class);
      setProcessHandler(AddClientListenerProcess.class,
                        AddClientListenerProcessHandler.class);
      setProcessHandler(RemoveClientListenerProcess.class,
                        RemoveClientListenerProcessHandler.class);

      // load the custom process/business object handlers
      // put this in some config file so you don't have to edit code every time you add a handler
      setBusinessObjectHandler(org.locomotive.obf.example.ExampleLbBusinessObject.class, org.locomotive.obf.example.ExampleLbHandler.class);
      setBusinessObjectHandler(org.locomotive.obf.example.ExampleRwBusinessObject.class, org.locomotive.obf.example.ExampleRwHandler.class);
   }


   /**
    * Returns a singleton instance of the server supporter.
    */

   public static ServerSupporter getInstance() {
      if (instance == null) {
         synchronized (ServerSupporter.class) {
            if (instance == null)
               instance = new ServerSupporter();
         }
      }

      return instance;
   }


   /**
    * Set the server to use a database other than the locomotive database.
    * <p>
    * [Pending: more documentation to follow]
    * <p>
    * <b>This may change</b>
    *
    * @param      jdbcDriver  the name of the jdbc driver class to use
    * @param      url         the database url
    * @param      user        the database user
    * @param      password    the user's password
    */

   public void initializeLocalDatabase(String jdbcDriver, String url,
                                       String user, String password) {
      useLocalDatabase = true; 
      this.jdbcDriver = jdbcDriver;
      this.url = url;
      this.user = user;
      this.password = password;

      try {
         Class.forName(jdbcDriver);
      } catch (ClassNotFoundException e) {
         e.printStackTrace();
      }
   }


   /**
    * Gets a database connection to use, from either the locmotive or a local
    * database.  This connection will not be used again until it is given
    * back using the <code>dropConnection()</code> method.
    *
    * @see        ServerSupporter#dropConnection(java.sql.Connection)
    */

   public Connection getConnection() {
      if (useLocalDatabase) {
         // get a local connection, change this to use a pool
        try {
           return ( DriverManager.getConnection(
                                               url, user, password) );
        } catch (SQLException se) {
          se.printStackTrace();
          return null;  // do something
        }

      } else {
         return Loco.getDBConnection();
      }
   }


   /**
    * Returns a database connection for re-use.
    * <p>
    * <b>Required:</b> A way of either returning connections for re-use or
    * disposing of them if corrupt.
    *
    * @param      con the connection to return
    * @see        ServerSupporter#getConnection()
    */

   public void dropConnection(Connection con) {
      // this is a workaround because sometimes we lock up the connections
      // fix this so that Loco.putBackBDConnection()
      // is used instead to be efficient

      if (useLocalDatabase) {
         try {
            con.close();
         } catch (SQLException e) {
            // ?
            
         }
      } else {

         //try {
         //con.close();
         //} catch (SQLException e) {
         //e.printStackTrace();
         //}

//Loco.disposeDBConnection(con);
         Loco.putBackDBConnection(con);
//System.out.println("@ disposed of connection, count="+Loco.getOpenDBConnectionCount());

      }
   }


   /**
    * Maps a business object class to a handler class, so that handlers of the
    * type <code>handlerClass</code> will be used when dealing with business
    * objects of type <code>busClass</code>.
    *
    * @param      busClass     the business object class.  This must implement
    *                          the <code>BusinessObject</code> interface
    * @param      handlerClass the handler class.  This must implement
    *                          the <code>BusinessObjectHandler</code> interface
    * @exception  ClassCastException if either classes were invalid
    */

   public void setBusinessObjectHandler(Class busClass, Class handlerClass) {
      businessObjectHandlers.put(busClass, handlerClass);

   }


   /**
    * Maps a process class to a handler class, so that handlers of the
    * type <code>handlerClass</code> will be used to execute processes of
    * type <code>processClass</code>.
    *
    * @param      processClass the business process class.  This must implement
    *                          the <code>BusinessProcess</code> interface
    * @param      handlerClass the handler class.  This must implement
    *                          the <code>ProcessHandler</code> interface
    * @exception  ClassCastException if either classes were invalid
    */

   public void setProcessHandler(Class processClass, Class handlerClass) {
      processHandlers.put(processClass, handlerClass);
   }


   /**
    * Returns an instance of a business object handler to use when dealing
    * with business objects of type <code>busClass</code>
    *
    * @param      busClass     the business object class.  This must implement
    *                          the <code>BusinessObject</code> interface
    * @exception  ClassCastException if busClass was invalid
    */

   public BusinessObjectHandler getBusinessObjectHandler(Class busClass)
   throws ObfException {

      try {
         Class handlerClass = (Class) businessObjectHandlers.get(busClass);
         if (handlerClass != null)
            return (BusinessObjectHandler) handlerClass.newInstance();
         else
            throw new ObfException("No handler class defined for "+
                                   busClass.getName());

      } catch (InstantiationException e) {
         return null; // do something
      } catch (IllegalAccessException e) {
         return null; // do something
      }
   }


   /**
    * Returns an instance of a process handler to use to execute
    * processes of type <code>processClass</code>
    *
    * @param      processClass the business process class.  This must implement
    *                          the <code>BusinessProcess</code> interface
    * @exception  ClassCastException if processClass was invalid
    */

   public ProcessHandler getProcessHandler(Class processClass)
   throws ObfException {

      try {
         Class handlerClass = (Class) processHandlers.get(processClass);
         if (handlerClass != null)
            return (ProcessHandler) handlerClass.newInstance();
         else
            throw new ObfException("No handler class defined for "+
                                   processClass.getName());

      } catch (InstantiationException e) {
         return null; // do something
      } catch (IllegalAccessException e) {
         return null; // do something
      }
   }


   /**
    * Start a transaction for the given connection.
    *
    * @param      con the database connection
    * @exception  SQLExecption if a database-access error occurs
    * @see        ServerSupporter#commitTransaction(java.sql.Connection)
    * @see        ServerSupporter#rollbackTransaction(java.sql.Connection)
    */

   public void startTransaction(Connection con) throws SQLException {
      con.setAutoCommit(false);
   }


   /**
    * Commit a transaction for the given connection.
    *
    * @param      con the database connection
    * @exception  SQLExecption if a database-access error occurs
    * @see        ServerSupporter#startTransaction(java.sql.Connection)
    * @see        ServerSupporter#rollbackTransaction(java.sql.Connection)
    */

   public void commitTransaction(Connection con) throws SQLException {
      con.commit();
      con.setAutoCommit(true);
   }


   /**
    * Roll back a transaction for the given connection.
    *
    * @param      con the database connection
    * @see        ServerSupporter#startTransaction(java.sql.Connection)
    * @see        ServerSupporter#commitTransaction(java.sql.Connection)
    */

   public void rollbackTransaction(Connection con) {
      try {
         con.rollback();
         con.setAutoCommit(true);
      } catch (SQLException e) {
         
      }
   }


}
