/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.server;



import org.locomotive.obf.Debug;
import org.locomotive.obf.data.*;
import java.util.Date;
import java.math.BigDecimal;
import java.text.DateFormat;
import java.text.SimpleDateFormat;


/**
 * Default renderer for JDBC databases.
 * <p>
 * <b>Note: these classes may change to make use of the
 * <code>org.locmotive.util.SQL</code> class</b>
 *
 * @author Erik Erskine
 * @see    org.locomotive.util.SQL
 */

public class SqlRendererJdbc implements SqlRenderer {

  protected DateFormat dateFormat = new SimpleDateFormat("dd-MMM-yyyy");


  private static SqlRendererJdbc instance = null;

  public static SqlRendererJdbc getInstance() {
    if (instance == null) {
      synchronized(SqlRendererJdbc.class) {
        if (instance == null)
          instance = new SqlRendererJdbc();
      }
    }
    return instance;
  }



  public String getSqlNull() {
    return "null";
  }

  public String getSql(String s) {
    return "'"+s+"'";
  }
  
  public String getSql(Date d) {
    return "'" + dateFormat.format(d) + "'";
  }

  public String getSql(int i) {
    return ""+i;
  }

  public String getSql(boolean b) {
    return b ? "1" : "0";
  }

  public String getSql(BigDecimal d) {
    return d.toString();
  }



  public String getSql(Object value) {
    if (value == null) 
      return getSqlNull();

    else if (value instanceof String)
      return getSql((String) value);

    else if (value instanceof Date)
      return getSql((Date) value);

    else if (value instanceof Integer)
      return getSql(((Integer) value).intValue());

    else if (value instanceof Boolean)
      return getSql(((Boolean) value).booleanValue());

    else if (value instanceof BigDecimal)
      return getSql((BigDecimal) value);

    else
      return "'" + value.toString() + "'";
  }




  public String getSql(UniqueStringId uid) {
    return (uid != null) ? "'"+uid.getString()+"'" : "null";
  }

  public String getSql(UniqueIntegerId uid) {
    return (uid != null) ? "'"+uid.getInt()+"'" : "null";
  }




  public String getSql(DataObject value) {
    return getSql(value, false);
  }



  public String getSql(DataObject value, boolean upperCase) {
    String sql = getSql(value.getData());

    if (upperCase && value.getData() != null)
      sql = sql.toUpperCase();

    return sql;
  }





  // -----------


  // Produces assignments of the form column='value'


  public String getSqlAssignment(String column, UniqueStringId u) {
    return (u != null) ? (column+"='"+u.getString()+"'") : column+"=null";
  }


  public String getSqlAssignment(String column, UniqueIntegerId u) {
    return (u != null) ? (column+"='"+u.getInt()+"'") : column+"=null";
  }


  public String getSqlAssignment(String column, DataObject value) {
    return getSqlAssignment(column, value, false);
  }


  public String getSqlAssignment(String column, DataObject value,
                                 boolean upperCase) {
    String sql = getSql(value.getData());
    
    if (upperCase && value.getData() != null)
      sql = sql.toUpperCase();

    return column+"="+sql;
  }


  public String getSqlAssignment(String column, Object value) {
    return column+"="+value;
  }






  // Produces comparisons of the form column='value'

  public String getSqlCompare(String column, UniqueStringId u) {
    return (u != null) ? (column+"='"+u.getString()+"'") : column+" is null";
  }

  
  public String getSqlCompare(String column, UniqueIntegerId u) {
    return (u != null) ? (column+"='"+u.getInt()+"'") : column+" is null";
  }

  
  public String getSqlCompare(String column, DataObject value) {
    return getSqlCompare(column, value, false);
  }


  public String getSqlCompare(String column, DataObject value,
                              boolean upperCase) {
    String sql = getSql(value.getData());
    
    if (upperCase && value.getData() != null)
      sql = sql.toUpperCase();

    return (sql != null) ? column+"="+sql : column+" is null";
  }


  public String getSqlCompare(String column, Object value) {
    return column+"="+value;
  }





  // As above but uses the original data

  public String getSqlCompareOriginal(String column,
                                      EditableDataObject value) {
    return getSqlCompareOriginal(column, value, false);
  }


  public String getSqlCompareOriginal(String column, EditableDataObject value,
                                      boolean upperCase) {
    String sql = getSql(value.getOriginalData());
    
    if (upperCase && value.getData() != null)
      sql = sql.toUpperCase();

    return (sql != null) ? column+"="+sql : column+" is null";
  }




}


