/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL are Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.obf.server;

import org.locomotive.obf.BusinessProcess;
import org.locomotive.obf.ObfException;
import org.locomotive.obf.data.UniqueId;
import java.util.Vector;
import java.util.Enumeration;
import java.sql.Connection;


/**
 * Partial handler implementation for processes that should be done inside
 * a transaction.  Subclasses should override the <code>doTransaction()</code>
 * method to execute their process within a transaction.  This class also
 * keeps track of reload notices that should be put on the reload queue only
 * if the transaction is committed.
 *
 * @author Erik Erskine
 * @author David Warnock
 */

public abstract class TransactionProcessHandler implements ProcessHandler {

  private Vector reloadNotices = new Vector();

  /**
   * Add a reload notice that should be added to the main reload queue only
   * if the transaction is committed.
   */

  protected void addReloadNotice(ReloadNotice n) {
    reloadNotices.addElement(n);
  }


  /**
   * Add a reload notice that should be added to the main reload queue only
   * if the transaction is committed.
   *
   * @param      type can be one of <code>ReloadNotice.INSERT</code>,
   *             <code>ReloadNotice.UPDATE</code>,
   *             <code>ReloadNotice.DELETE</code> or
   *             <code>ReloadNotice.REFRESH</code>
   * @param      busClass the class of business object to be reloaded
   * @param      id the unique id of the exact business object (in the case of
   *             <code>ReloadNotice.REFRESH</code> this is ignored)
   * @exception  InvalidArgumentException if the type was not valid.
   */

  protected void addReloadNotice(int type, Class busClass, UniqueId id) {
    reloadNotices.addElement(new ReloadNotice(type, busClass, id));
  }


  /**
   * Clear this local reload queue if the transaction is rolled back, so that
   * the notices do not get put on the main queue.
   */

  protected void clearReloadQueue() {
    reloadNotices = new Vector();
  }


  /**
   * Put the reload notices that have been added locally onto the main reload
   * queue if the transaction is committed.
   */

  protected void commitReloadQueue() {
    ReloadQueue.getInstance().addReloadNotices(reloadNotices);
    reloadNotices = new Vector();
  }


  /**
   * Execute a specified process inside a transaction and return the result.
   * <br>Subclasses should not override this method, but override
   * <code>doTransaction()</code> instead.
   *
   * @param      p     a process to execute
   * @param      con   a database connection to use
   * @exception  ClassCastException if the handler could not deal with
   *             processes of the type of <code>p</code>
   *             ObfException if another error occurs
   */

  public Object doProcess(BusinessProcess process, Connection con)
  throws ObfException {
    return doProcess(process, con, true);
  }


  /**
   * Execute a specified process, inside a transaction if specified,
   * and return the result.
   * <br>Subclasses should not override this method, but override
   * <code>doTransaction()</code> instead.
   * <p>
   * <b>Note: this has a bug</b> - if called from within another process
   * handler the reload notices in this process get put on the
   * reload queue even if the enclosing process fails and that transaction
   * is rolled back.
   *
   * @param      p            a process to execute
   * @param      con          a database connection to use
   * @param      transaction  <code>true</code> if this process should
   *                          be done inside its own transaction,
   *                          <code>false</code> if it will be done
   *                          as part of another process with a transaction
   * @exception  ClassCastException if the handler could not deal with
   *             processes of the type of <code>p</code>
   *             ObfException if another error occurs
   */

  public Object doProcess(BusinessProcess process, Connection con,
                          boolean transaction)
  throws ObfException {
    ServerSupporter supporter = ServerSupporter.getInstance();
    Object result = null;
    
    try {
      if (transaction)
        supporter.startTransaction(con);

      result = doTransaction(process, con);
      
      if (transaction)
        supporter.commitTransaction(con);

      commitReloadQueue();
//      supporter.dropConnection(con);
    
    } catch (Exception e) {
      supporter.rollbackTransaction(con);
      clearReloadQueue();
      e.printStackTrace();
      throw new ObfException(""+e);
    }

    return result;
  }


  /**
   * Executes the specified process.  This method is always invoked inside a
   * transaction.
   *
   * @param      p            a process to execute
   * @param      con          a database connection to use
   * @param      transaction  <code>true</code> if this process should
   *                          be done inside its own transaction,
   *                          <code>false</code> if it will be done
   *                          as part of another process with a transaction
   * @exception  ClassCastException if the handler could not deal with
   *             processes of the type of <code>p</code>
   *             ObfException if another error occurs
   */

  protected abstract Object doTransaction(BusinessProcess p, Connection con)
  throws Exception;


}
