/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
*/

package org.locomotive.router;

import java.io.*;
import java.net.*;
import java.util.*;


class HttpRequest {

  private String method, uri, path, queryString, version;
  private Hashtable headers = new Hashtable();
  private InputStream in;

  /**
   * Creates a new http request given the input stream.
   * Reads and parses the header.
   *
   * @exception  IOException if the stream could not be read
   */

  HttpRequest(InputStream in) throws IOException {
    this.in = in;
    doFirstLine();
    doHeaders();
  }



  private static String readLine(InputStream in) throws IOException {
    String result = "";

    while (true) {
      int ch = in.read();
      
      if (ch == -1) {
        // stop if we get to the end of the input
        break;

      } else if (ch == 10) {
        // stop if we get the LF character (http should use CRLF everywhere)
        break;

      } else if (ch != 13) {
        result += (char) ch;
      }

    }

    return result;
  }


  private void doFirstLine() throws IOException {
    String str = readLine(in);
    StringTokenizer st = new StringTokenizer(str);
    method  = st.hasMoreTokens() ? st.nextToken() : null;
    uri     = st.hasMoreTokens() ? st.nextToken() : null;
    version = st.hasMoreTokens() ? st.nextToken() : null;

    StringTokenizer st2 = new StringTokenizer(uri, "?");
    path        = st2.hasMoreTokens() ? st2.nextToken() : "";
    queryString = st2.hasMoreTokens() ? st2.nextToken() : "";
  }


  private void doHeaders() throws IOException {
    Vector lines = new Vector();
    boolean stop = false;
    while (!stop) {
      String line = readLine(in);

      // is this the blank line?
      if (line != null && line.length() > 0) {
        System.out.println("- "+line);
        lines.addElement(line);
      } else
        stop = true;
    }

    // now go through each line
    Enumeration e = lines.elements();
    while (e.hasMoreElements()) {
      String line = (String) e.nextElement();

      // need more cheking here
      int colon = line.indexOf(':');
      String key = line.substring(0, colon);
      String value = line.substring(colon+1).trim();

      headers.put(key, value);
    }
  }


  /**
   * Returns the method, eg GET, POST etc.
   */

  String getMethod() {
    return method;
  }



  String getUri() {
    return uri;
  }
  



  String getPath() {
    return path;
  }
  

  String getQueryString() {
    return queryString;
  }


  void translatePath(String from, String to) {
    int start = path.indexOf(from);

    if (start == 0 && path.length() > from.length())
      path = to + path.substring(from.length());

    else if (start == 0 && path.length() == from.length())
      path = to;

    if (path.length() == 0)
      path = "/";
  }


  /**
   * Returns the http version
   */

  String getVersion() {
    return version;
  }


  /**
   * Returns a header in the http request.
   */

  String getHeader(String key) {
    return(String) headers.get(key);
  }


  /**
   * Returns an enumeration of all the header keys.
   */

  Enumeration headerKeys() {
    return headers.keys();
  }


  /**
   * Returns an input stream that has read past the header
   * and is now positioned before any POST data
   */

  InputStream getInputStream() {
    return in;
  }


  /**
   * Returns the length of any POST data, or <code>null</code>
   * if there is none
   */

  Integer getContentLength() {
    try {
      return new Integer((String) headers.get("Content-length"));
    } catch (Exception e) {
      return null;
    }
  }

  String getContentType() {
    return (String) headers.get("Content-type");
  }

  String getAccept() {
    return (String) headers.get("Accept");
  }

  String getCookie() {
    return (String) headers.get("Cookie");
  }

}

