/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Sundayta Ltd.
 * Portions created by Sundayta are Copyright (C) 1999 Sundayta Ltd.
 * All Rights reserved.
 */

package org.locomotive.router;

import java.util.ResourceBundle;
import java.io.*;
import java.net.*;
import org.locomotive.server.*;
import org.locomotive.directory.*;


/**
 * The router provides a mechanism for handling HTTP requests from clients and
 * dispatching them to a locomotive, thereby eliminating the need for a web
 * server to be able to run the locomotive.
 * <p>
 * The router listens for HTTP connections from clients, parses the headers
 * and then either forwardis them to a locomotive or to a real web server,
 * depending on the URL.  The locomotive or web server are both optional, and
 * can be run on either the same machine as the router or a different one.
 * <p>
 * <b>Still to do:</b> use the bean directory for configuration
 *
 * @author Erik Erskine (erik@sundayta.co.uk)
 */

public class Router implements LocomotiveService {

  private ThreadPoolManager tpm;
  private Address webServerAddress, locomotiveAddress;
  private ServerSocket ss;
  private Socket cs;
  private int port, backlog=8, minPooledThreads=3, maxPooledThreads=8;



  /**
   * Creates a new router
   */

  private Router() {
  }


  /**
   * Start the service.
   */

  public void startService() throws ServiceException {

    if (port == 0)
      throw new ServiceException("Port not defined");

    tpm = new ThreadPoolManager(null, minPooledThreads, maxPooledThreads);

    try {
      ss = new ServerSocket (port, backlog);
    } catch (IOException e) {
      throw new ServiceException(""+e);
    }

    System.out.println("Router started on port "+port);

    int connections = 0;
    while (true) {

      try {
        cs = ss.accept();
      } catch (IOException e) {
        System.err.println("Init Failed with ioexception: " + e);
        System.exit(-1);
      }

      connections++;
      System.out.println("Got connection: " + connections);

      PooledThread pt = tpm.getPooledThread ();
      ClientConnection cc = new ClientConnection(this, cs, connections);

      pt.loadRunnable(cc);
      pt.go();
    }

  }


  /**
   * Stop the service. 
   */

  public void stopService() throws ServiceException {
    // stop!
  }


  /**
   * Return true if this service is running, false otherwise.
   */

  public boolean isRunning() {
    return false;
  }





  /**
   * Returns the port this router uses.
   * @see        Router#setPort(int)
   */

  public int getPort() {
    return port;
  }


  /**
   * Sets the port this router will use.
   * @see        Router#getPort()
   */

  public void setPort(int port) {
    this.port = port;
  }


  /**
   * Returns the maximum length of the queue of clients waiting
   * at any given time.
   * @see        Router#setBacklog(int)
   */

  public int getBacklog() {
    return backlog;
  }


  /**
   * Sets the maximum length of the queue of clients waiting
   * at any given time.
   * @see        Router#getBacklog()
   */

  public void setBacklog(int backlog) {
    this.backlog = backlog;
  }


  /**
   * Returns the minimum number of threads used by the router.
   * @see        Router#setMinPooledThreads(int)
   * @see        Router#getMaxPooledThreads()
   * @see        Router#setMaxPooledThreads(int)
   */

  public int getMinPooledThreads() {
    return minPooledThreads;
  }
  
  
  /**
   * Sets the minimum number of threads used by the router.
   * @see        Router#getMinPooledThreads()
   * @see        Router#getMaxPooledThreads()
   * @see        Router#setMaxPooledThreads(int)
   */

  public void setMinPooledThreads(int min) {
    minPooledThreads = min;
  }


  /**
   * Returns the maximum number of threads used by the router.
   * @see        Router#getMinPooledThreads()
   * @see        Router#setMinPooledThreads(int)
   * @see        Router#setMaxPooledThreads(int)
   */

  public int getMaxPooledThreads() {
    return maxPooledThreads;
  }


  /**
   * Sets the maximum number of threads used by the router.
   * @see        Router#getMinPooledThreads()
   * @see        Router#setMinPooledThreads(int)
   * @see        Router#getMaxPooledThreads()
   */

  public void setMaxPooledThreads(int max) {
    maxPooledThreads = max;
  }





  /**
   * Sets the address of the locomotive.
   *
   * @param      url the url as accessed from the router's machine
   * @param      port the port
   */

  public void setLocomotive(String url, int port) {
    locomotiveAddress = new Address(url, port);
  }


  /**
   * Sets the address of the web server.
   *
   * @param      url the url as accessed from the router's machine
   * @param      port the port
   */

  public void setWebServer(String url, int port) {
    webServerAddress = new Address(url, port);
  }


  /**
   * Returns the address of the locomotive.
   */

  Address getLocomotiveAddress() {
    return locomotiveAddress;
  }


  /**
   * Returns the address of the web server.
   */

  Address getWebServerAddress() {
    return webServerAddress;
  }



  /**
   * Utility method to display an HTML document on the given output stream,
   * containing an error message.
   *
   * @param      message the message to display between the
   *             <code>&lt;body&gt;</code> and <code>&lt;/body&gt;</code>
   * @param      os the output stream to write to
   */

  static void showErrorMessage(String message, OutputStream os) {
    System.out.println("showing error message: "+message);
    PrintWriter pw = new PrintWriter(os);
    pw.println("<html><head><title>Error</title></head>");
    pw.println("<body>"+message+"</body></html>");
    pw.flush();
  }


  /**
   * Utility method to copy an exact number of bytes
   * from one stream to another.
   *
   * @param      in the input stream
   * @param      out the output strem
   * @param      num the number of bytes to be copied
   * @see        HttpRequest#getContentLength()
   * @exception  IOException if an i/o error occurred
   */

  static void copyStream(InputStream in, OutputStream out, int num)
  throws IOException {

    for (int x=0; x<num; x++) {
      int b = in.read();
      System.out.print(" "+x+"("+(byte)b+")");
      out.write(b);
      System.out.print(",");
      out.flush();
    }
  }



  /**
   * Startup method to use the router as a standalone application.
   */

  public static void main(String args[]) {
    Router r = new Router();

    String configFile = "LocoRouter";
    if (args.length > 0)
      configFile = args[0];

    r.configure(configFile);

    try {
      r.startService();
    } catch (ServiceException e) {
      e.printStackTrace();
    }

  }






  // configuration methods
  // this lot will eventually be replaced by bean directory configuration


  private void configure(String configFile) {
    
    ResourceBundle rb = null;
    try {
      rb = ResourceBundle.getBundle(configFile);
    } catch (Exception e) {
    }

    setPort(readConfig(rb, "client.connection.port", 5420));
    setBacklog(readConfig(rb, "client.connection.backlog", 16));
    setMinPooledThreads(readConfig(rb, "pooled.threads.min", 4));
    setMaxPooledThreads(readConfig(rb, "pooled.threads.max", 16));

    // set the locomotive

    String lu = readConfig(rb, "locomotive.url", null);
    int    lp = readConfig(rb, "locomotive.port", 0);
    if (lu != null && lp != 0) {
      setLocomotive(lu, lp);
    }

    // set the web server

    String wu = readConfig(rb, "web.server.url", null);
    int    wp = readConfig(rb, "web.server.port", 0);
    if (wu != null && wp != 0) {
      setWebServer(wu, wp);
    }
  }



  private static int readConfig(ResourceBundle rb, String sKey,
                                int defaultValue) {
    int i;
    String s = null;
    try {
      s = rb.getString(sKey);
      if (s != null) i = Integer.parseInt(s);
      else i = defaultValue;
    } catch (Exception e) {
      i = defaultValue;
    }

    return i;
  }


  private static String readConfig(ResourceBundle rb, String sKey,
                                   String defaultValue) {
    String s = null;
    try {
      s = rb.getString(sKey);
    } catch (Exception e) {
      s = defaultValue;
    }
    if (s == null) s = defaultValue;
    return s;
  }

}

