/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import java.util.*;
import java.io.*;
import org.locomotive.util.*;

/**
 * The Config class provides a way to maintain configuration information.
 * It contains it's own error checking, and can return configuration
 * in either an integer or String form.  Any configs in a config file
 * must have defuaults specified in the setupDefaults() method.  So,
 * to create your own config, you would have to override this method
 * to setup your own set of configs.
 * <P> The config file should specify first the name of the config, followed
 * by at least one space, and then the value.  For example:<PRE>
 *     SYSTEM_INSTANCE_ID                      sv_one
 *</PRE>
 * All lines which begin with an '#' are read as comments, i.e. they are not
 * evaluated.
 */
public class Config
{
    private String config_file_path;
    private Hashtable config_hash;
    private Hashtable command_line_override_hash;
    private boolean config_file_path_overridden;
    private boolean dump_configs;
    private boolean dump_verbose;

    /** When the config_file_path is equal to this, it indicates that the 
     * configuration was loaded from the BeanDirectory services, not a 
     * config file.
     **/
    public static final String BEAN = "BEAN_DIRECTORY_NOT_FILE";

    /** a constraint for use in the addConfig method which will check to
     * make sure the value of the config is not empty or null, and is an 
     * integer */
    public String forint = "DEFINED TIDY NO_WHITESPACE ISINT";

    /** a constraint for use in the addConfig method which will check to
     * make sure the value of the config is neither empty nor null. */
    public String forstring = "DEFINED TIDY NOTNULL NO_WHITESPACE";

    
    /** Initializes a config object, with the specified arguments, from
     * the specified file path.
     * @param args a series of srguments may be specified at construction.
     * these include:<UL>
     * <LI>-h  help- prints usage.
     * <LI>-f  sets config file path.  will override the specified defaultpath.
     * May be used to override BeanDirectory-style configuration.
     * <LI>-d  dumps configs
     * <LI>-dv dumps configs- verbose
     * <LI>-O  enters any number of properties into Config file. 
     *         Usage: NAME=VALUE
     * </UL>
     * Only config properties which already exist in the default config tables
     * can be added.  If property is NAME="", then the value is entered as "", 
     * erasing any previous values.  
     * @param defaultpath the path of the config file to load.  If equal to 
     * the constant BEAN, no config file is used. 
     */
    public Config (String args[], String defaultpath) 
	throws IllegalArgumentException
    {
	this();
	init(args, defaultpath);
    }
    
    /** Constructor for Bean Directory.  Should be followed by a call to 
     * init(args, BEAN). 
     **/
    public Config () 
    {
	config_hash = new Hashtable(50);
	command_line_override_hash = new Hashtable();
	dump_configs = false;
	dump_verbose = false;
	config_file_path_overridden = false;

	// set up all default vars and constraints
	setupDefaults();
	// This must be called before the BeanDirectory initializer
	// tries to set variables.
    }

    public void init (String args[], String defaultpath) 
	throws IllegalArgumentException
    {
	config_file_path = defaultpath;

	// do this to get non-default config_file_path, plus any switches
	if (! parse_command_line_args (args))
	{
	    throw new IllegalArgumentException();
	}

	String errors;

	if (!config_file_path.equals(BEAN))
	{
	    // override defaults from config file
	    errors = loadConfigsFromFile (config_file_path);
	    if (! errors.equals(""))
	    {
		System.out.println("Configuration File Errors:\n");
		System.out.println(errors);
		throw new IllegalArgumentException();
	    }
	}

	// apply changes in command_line_override_hash
	doCommandLineOverride ();

	// validate the resulting config hash
	errors = validate();
	if (! errors.equals(""))
	{
	    System.out.println("Config Errors:\n");
	    System.out.println(errors);
	    throw new IllegalArgumentException();
	}

	if (dump_configs)
	{
	    doConfigDump ();
	    throw new IllegalArgumentException();
	}
    }

    public void genericSetInteger (String name, Integer val)
    {
	genericSet(name, val.toString());
    }

    // This method is exactly the same as genericSet; it exists for 
    // naming consistency.
    public void genericSetString (String name, String val)
    {
	genericSet(name, val);
    }

    // This method helps adapt the Config class to Java Beans
    public void genericSet (String name, String val)
    {
	try {
	    changeConfig (name, val);
	}
	catch (ConfigNotFoundException e)
	{
	    System.out.println("Bad config name: " + e.getMessage());
	}
	catch (ConfigInitializationException e)
	{
	    // doesn't satisfy constraints
	    System.out.println("Config constraint violation: " + 
			       e.getMessage());
	}
    }

    public Integer genericGetInteger (String name)
    {
       try {
	   return new Integer(getInt(name));
       }
       catch (ConfigNotFoundException e)
       {
	   return null;
       }
    }

    // This method helps adapt the Config class to Java Beans
    public String genericGetString (String name)
    {
       try {
	   return getString(name);
       }
       catch (ConfigNotFoundException e)
       {
	   return null;
       }
    }

    /** Override this method to set up your own defaults.  Any configs 
     * in the file which are not specified here will cause an error
     * to be thrown when the file is loaded.
     * to add a config to the defaults, use the addConfig method.
     * @see #addConfig
     */
    public void setupDefaults()
    {
	// Historical Note:  example configs are from the original MAS
	addConfig ("MAS_VERSION", "1.0b3", 
		   forstring, "MAS System Version String");

	addConfig ("SYSTEM_TAG", "EUS", 
		   forstring, "System Identity Tag");
	addConfig ("SYSTEM_SERVER_ROOT_PATH", "/home/mas", 
		   forstring, "Basedir for all file paths - absolute");

	addConfig ("LOCO_MCMS_SERVICES", "off", 
		   forstring, "MCMS services toggle switch \"on\" to enable");

	addConfig ("LOCO_PORT", "5420", 
		   forint, "TCP Port for Application Server");
	addConfig ("LOCO_DB_MIN_THREADS", "4", 
		   forint, "Minimum number of pre-opened DB RunProcessors");
	addConfig ("LOCO_DB_MAX_THREADS", "16", 
		   forint, "Grow limit for pre-opened DB RunProcessors");
	addConfig ("LOCO_LOG_LEVEL", "10", 
		   forint, "LOCO logging verbosity value");
	addConfig ("LOCO_LOG_PATH", "logs/LOCO-%.log", 
		   forstring, "Relative path for Application Server logs");

	addConfig ("LOCO_EVENTS_LOG_LEVEL", "10", 
		   forint, "Application Server Events logging verbosity");
	addConfig ("LOCO_EVENTS_LOG_PATH", "logs/events-%.log", 
		   forstring, "path to Application Server events log");

	addConfig ("LOCO_SESSION_EXPIRE_TIME", "3600", 
		   forint, "Dormant seconds allowed before session expires");
	addConfig ("LOCO_TEMPLATE_ROOT_PATH", "templates", 
		   forstring,"Application Server template root relative path");

	addConfig ("LOCO_URL_MAIN", "http://www.minds.com/cgi-bin/maslink.cgi", 
		   forstring, "Full URL to Application Server main entrance");
	addConfig ("LOCO_URL_HOMEPAGE", "command?stat+home", 
		   forstring, "relative URL for Application Server homepage");
	addConfig ("LOCO_COOKIE_DOMAIN_MAIN", ".minds.com", 
		   forstring, "Domain for Application Server cookies");

	addConfig ("LOCO_DB_URL", "jdbc:weblogic:oracle", 
		   forstring, "Application Server database connect url");
	addConfig ("LOCO_DB_SERVERNAME", "", 
		   "DEFINED TIDY NO_WHITESPACE", 
		   "Application Server Database Server Name");
	addConfig ("LOCO_DB_USERNAME", "masuser", 
		   forstring, "Application Server username for Database");
	addConfig ("LOCO_DB_PASSWORD", "masuserpw", 
		   forstring, "Application server password for Database");
	addConfig ("LOCO_DB_DATABASE", "", 
		   "DEFINED TIDY NO_WHITESPACE", 
		   "Application Server database name");

	addConfig ("MCMS_MCE_MMO_COMPILED_PATH", "/mmo/mce", 
		   forstring, 
		   "Base directory for mmo location in compiled MCEs");
	addConfig ("MCMS_MCE_MMO_PATH", "/usr/local/www/mmo/mce", 
		   forstring, "Absolute path to MCE MMO base directory");

	addConfig ("SMTP_SERVER_NAME", "mail.minds.com", 
		   forstring, "Hostname for outgoing mail SMTP server");
	addConfig ("SMTP_SERVER_PORT", "25", 
		   forint, "Port for outgoing mail SMTP server");
	addConfig ("SMTP_SERVER_TIMEOUT", "60", 
		   forint, "SMTP client delivery timeout");
	addConfig ("EMAIL_VERIFY_FROM", "", 
		   forstring + " ISEMAIL", 
		   "From address on outbound email");
	addConfig ("EMAIL_NEWREG_ALERT_TO", "", 
		   forstring + " ISEMAIL", 
		   "To address for new registration alert summary");
	addConfig ("EMAIL_NEWVCC_ALERT_TO", "", 
		   forstring + " ISEMAIL", 
		   "To address for vcc submission alert summary");

	addConfig ("ENGAGED_SYSTEM_BASE_URL", 
	    "http://www.minds.com/cgi-bin/EMPane.cgi", 
		   forstring, "Base URL for engaged system");

	addConfig ("MCMS_RECEIVER_HOSTNAME", "enterprise.minds.com", 
		   forstring, "hostname where MCMSReceiver lives");
	addConfig ("MCMS_RECEIVER_PORT", "5923", 
		   forint, "Port on which MCMS Receiver is listening");
	addConfig ("MCMS_RECEIVER_LOG_LEVEL", "10", 
		   forint, "mcms receiver log level");
	addConfig ("MCMS_RECEIVER_LOG_PATH", "logs/receiver-%.log", 
		   forstring, "path for mcms_receiver logs");
	addConfig ("MCMS_RECEIVER_MCE_MMO_PATH", "/usr/local/www/mmo/mce", 
		   forstring, "Base directory for MCE MMO storage");
	addConfig ("MCMS_RECEIVER_DB_URL", "jdbc:weblogic:oracle", 
		   forstring, "Database URL for MCMS Receiver connection");
	addConfig ("MCMS_RECEIVER_DB_USERNAME", "mas_eus_user", 
		   forstring,"Database username for MCMS Receiver connection");
	addConfig ("MCMS_RECEIVER_DB_PASSWORD", "mas_eus_userpw", 
		   forstring,"Database password for MCMS Receiver connection");
    }

    /** returns the path of the file the config is read in from.
     * If return value is equal to constant BEAN, no config file was used. */
    public String getConfigFilePath ()
    {
	return (config_file_path);
    }

    private void doCommandLineOverride ()
    {
	for (Enumeration e = command_line_override_hash.keys(); 
	     e.hasMoreElements(); )
	{
	    String name = (String) e.nextElement();
	    ConfigEntry ce = (ConfigEntry) config_hash.get (name);
	    ce.storeVal((String) command_line_override_hash.get (name));
	    ce.altered = true;
	}
    }

    private String validate ()
    {
	String name;
	ConfigEntry ce;
	StringBuffer error_sb = new StringBuffer ();
	
	for (Enumeration e = config_hash.keys(); e.hasMoreElements(); )
	{
	    name = (String) e.nextElement();
	    ce = (ConfigEntry) config_hash.get (name);
	    if (ce != null)
	    {
		ce.svalue = ce.sc.check (ce.svalue, error_sb);
	    }
	    else
	    {
		error_sb.append ("CE for " + name + " was null!");
	    }
	}

	return (error_sb.toString());
    }


    /** This method will allow you to add a config, regardless of
     * whether or not it exists in the defaults, or in the config file
     * @param name the name of the config
     * @param val the value of the config
     * @param conslist the list of constraints to apply to a config
     * value when it is loaded.  Constraint usage follows that for the
     * org.locomotive.util.StringConstraint.  Two values are provided in this
     * class - forint and forstring.
     * @param desc a short description of the config
     * @see org.locomotive.util.StringConstraint
     */
    protected void addConfig (String name, String val, 
				   String conslist, String desc)
    {
	if ((val != null) && (name != null) && (! name.equals("")))
	{
	    StringConstraint sc = new StringConstraint (name, conslist);
	    ConfigEntry ce = new ConfigEntry (name, val, sc, desc);
	    config_hash.put (name, ce);
	}
    }

    /**
     * this method changes an already created config to a specified value.
     * @param name: the name of the config
     * @param value the new value for the config.
     * @returns the value of changed config, which may have been altered 
     *          to allow it to match the config's constraints.
     * @exception ConfigNotFoundException if the config does not exist.
     * @exception ConfigInitializationException if the config value does not
     *            meet the required contraints
     */
    public String changeConfig(String name, String value) 
	throws ConfigNotFoundException, ConfigInitializationException
    {
	StringBuffer msgbuf = new StringBuffer();

        // we have two tokens which we believe to contain 
	// a name value pair
	// check if it is a valid token
	if (! config_hash.containsKey(name))
	{
	    throw new ConfigNotFoundException(" bogus config name: " + name );
	}
	else
	{
	    // check the constraints on this value
	    ConfigEntry ce = (ConfigEntry) config_hash.get (name);
	    
	    // sc.check can return modified version of token
	    value = ce.sc.check (value, msgbuf);
	    
	    // if no errors, add to the config_hash
	    if (msgbuf.length() == 0)
	    {
		ce.storeVal (value);
		ce.altered = true;
		config_hash.put (name, ce);
		return value;    // the value sc.check returned to us
	    }
	    else
	    {
		throw new ConfigInitializationException(msgbuf.toString());
	    }
	}
    }

    /** returns the value of a config as a string */
    public String getString (String name) throws ConfigNotFoundException
    {
	if ((name == null) || (name.equals("")))
	{
	    throw new ConfigNotFoundException("name cannot be null");
	}

	ConfigEntry ce = (ConfigEntry) config_hash.get (name);

	if (ce == null)
	{
	    throw new ConfigNotFoundException ("config entry " + name + 
					       " was not found");
	}

	return (ce.svalue);
    }

    /** returns the value of a config as an int, if possible */
    public int getInt (String name) throws ConfigNotFoundException
    {
	if ((name == null) || (name.equals("")))
	{
	    throw new ConfigNotFoundException("name cannot be null");
	}

	ConfigEntry ce = (ConfigEntry) config_hash.get (name);
	if (ce == null)
	{
	    throw new ConfigNotFoundException ("config entry " + name + 
					       " was not found");
	}
	return (ce.ivalue);
    }

    /** prints all the configs and their present values, along with
     * their description, to standard out.
     */
    public void doConfigDump ()
    {
	for (Enumeration e = config_hash.keys(); e.hasMoreElements(); )
	{
	    String name = (String) e.nextElement();
	    ConfigEntry ce = (ConfigEntry) config_hash.get (name);

	    String output = ce.name + "\t" + ce.svalue + "\t" + ce.ivalue + 
		"\t" + String.valueOf (ce.altered);

	    if (dump_verbose)
	    {
		output = output + "\n" + ce.description + "\n";
	    }
	    System.out.println (output);
	}
    }

    /** returns true if the first character of the string is '#' */
    protected boolean isComment (String s)
    {
	if (s.length() == 0)
	{
	    return (true);
	}
	
	if (s.charAt(0) == '#')
	{
	    return (true);
	}

	return (false);
    }


    /** actually reads in the file and loads the configs.  Call this
     * to reload a file during runtime.
     * @return an error message if there were problems reading in the configs,
     * or an empty string if all went well.
     */
    protected String loadConfigsFromFile (String filename)
    {
	FileInputStream fis;
	BufferedReader br;

	try
	{
	    fis = new FileInputStream (filename);
	}
	catch (FileNotFoundException fnfe)
	{
	    return ("Config File not found at: " + filename);
	}
	
	br = new BufferedReader (new InputStreamReader (fis));

	StringBuffer msgbuf = new StringBuffer();
	int linenum = 0;
	int token_count = 0;
	String tokens[] = new String[2];
	String line;
	int msg_size_temp = 0;
	
	try
	{
	    while ((line = br.readLine()) != null)
	    {
		linenum++;
		// tokenize the lines
		StringTokenizer st = new StringTokenizer(line);

		// skip blank lines
		if ((token_count = st.countTokens()) == 0)
		{
		    continue;
		}
		else if (token_count == 1)
		{
		    // if only one token, make sure it is a comment
		    tokens[0] = st.nextToken();
		    if (! isComment(tokens[0]))
		    {
			if (! config_hash.containsKey(tokens[0]))
			{
			    msgbuf.append("Line: " + linenum 
				  + " bogus token: " + tokens[0] + "\n");
			}
			else 
			{
			    msgbuf.append("Line: " + linenum 
				  + " valid token: " + tokens[0] 
					  + ", requires value\n");
			}
		    }
		}
		else 
		{
		    // hmm, at least two tokens
		    tokens[0] = st.nextToken();
		    tokens[1] = st.nextToken();

		    if (isComment (tokens[0]))
		    {
			continue;
		    }
		    else
		    {
			try   // try to change the config from the default
			{
			    changeConfig(tokens[0], tokens[1]);
			} 
			catch (ConfigNotFoundException e) // couldn't find it
			{
			    msgbuf.append(e.getMessage());
			}
			catch (ConfigInitializationException e) // tokens[1]
			{                                    // didn't match sc
			    msgbuf.append(e.getMessage());
			}
		    }
		}
	    }
	}
	catch (IOException ioe)
	{
	    return ("Error Reading Config File: " + ioe.toString());
	}
	
	return (msgbuf.toString());
    }

    /** prints the usage for arguments accepted by the constructor for
     * this object
     */
    public static void printsyntax (String s)
    {
	System.out.println(s + "\nSyntax:\n" + 
	   "classpath [-f config_file_path] [-h] [-O NAME=VAL] [-d] [-dv]\n" +
	   "VAL can be set to null string by using -O NAME=\"\"\n");
    }

    /** returns an enumeration of all the config names */
    public Enumeration getConfigNamesEnumeration()
    {
	return(config_hash.keys());
    }

    /**  parses the command line options. see the constructor */
    private boolean parse_command_line_args (String args[]) 
    {
	for (int i = 0; i < args.length; i++)
	{
	    if (args[i].equals("-h"))
	    {
		printsyntax("");
		return (false);
	    }
	    else if (args[i].equals ("-f"))
	    {
		i++;
		if (args.length > i)
		{
		    config_file_path = args[i];
		    config_file_path_overridden = true;
		}
		else
		{
		    printsyntax ("missing path parameter");
		    return (false);
		}
	    }
	    else if (args[i].equals ("-d"))
	    {
		dump_configs = true;
	    }
	    else if (args[i].equals ("-dv"))
	    {
		dump_configs = true;
		dump_verbose = true;
	    }
	    else if (args[i].equals ("-O"))
	    {
		i++;
		if (args.length <= i)
		{
		    printsyntax ("missing NAMEVALUE pair");
		    return (false);
		}

		StringTokenizer st = new StringTokenizer (args[i], "=");
		if (st.countTokens() != 2)
		{
		    printsyntax ("name and value must be separated by '='");
		    return (false);
		}
		
		String token1 = st.nextToken();
		String token2 = st.nextToken();
		
		if (! config_hash.containsKey (token1))
		{
		    printsyntax ("invalid config '" + token1 + "'");
		    return (false);
		}   
		
		if (token2.equals("\"\""))
		{
		    command_line_override_hash.put (token1, "");
		}
		else
		{
		    command_line_override_hash.put (token1, token2);
		}
	    }
	    else
	    {
		printsyntax ("unknown option");
		return (false);
	    }
	}
	return (true);
    }
}

/** kindly little wrapper class for config objects */
class ConfigEntry
{
    String name;
    String svalue;
    StringConstraint sc;
    String description;

    int ivalue;
    boolean altered;
    
    public ConfigEntry (String nm, String val_str, StringConstraint strc, 
			String desc)
    {
	name = nm;
	sc = strc;
	description = desc;
	altered = false;
	
	storeVal (val_str);
    }

    public void storeVal (String s)
    {
	svalue = s;

	try 
	{
	    ivalue = Integer.parseInt (s);
	}
	catch (NumberFormatException nfe)
	{
	    ivalue = 0;
	}
    }
}
