/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import java.util.Hashtable;
import java.util.Stack;
import java.sql.*;
import org.locomotive.util.ObjectPool;
import org.locomotive.directory.LocomotiveService;
import org.locomotive.directory.ServiceException;

/**
 *  Manages a pool of JDBC connections.  Will broker the connections
 * out until there are no more to give out, whereupon it will either
 * attempt to start a new connection and send that out, or 
 * block if the maximum number of connections are already open until a 
 * connection is available.
 */
public class DBConnectionPool extends ObjectPool implements LocomotiveService {
   // init
   private String driver_classname;
   private String db_url;
   private String db_username;
   private String db_password;

   /** change this to whatever you like... */
   private static int LOG_LEVEL = 3;
   // David Warnock change to support alternative DBConnectionPools
   public static int getDefaultLogLevel() { return LOG_LEVEL; }


   /** creates a new pool, and fills it with the specified minumum number
    * of database connections.
    * @param driver_classname_in the class name for the JDBC driver to
    *        use.
    * @param url_in the URL to use with the driver.  See the driver
    *        documanetation for more details.
    * @param username_in the username to connect to the database with
    * @param password_in the user's password
    * @param min_open_in the minumum number of connections to remain
    *        in the pool
    * @param max_open_in the maximum number of connections to 
    *        maintain at any given time
    * @param l the log to use to log any major events
    * @param id the unique id for this pool- pass -1 to not use an id
    */
   public DBConnectionPool (String driver_classname_in, String url_in, 
                            String username_in, String password_in, 
                            int min_open_in, int max_open_in, Log l) 
   throws InitializationException
   {
      super(min_open_in, max_open_in, l, LOG_LEVEL);

      this.driver_classname = driver_classname_in;
      this.db_url = url_in;
      this.db_username = username_in;
      this.db_password = password_in;

      try {
         Class.forName(driver_classname);
      } catch (ClassNotFoundException cnfe) {
         throw new InitializationException ("DBConnPool: driver load: " + 
                                            cnfe.toString());
      }
   }

   /**
    * Default constructor for the connection pool.
    */
   public DBConnectionPool(){   
   }

   /** creates a new database connection */
   public Object createNewObject () throws InitializationException
   {
      try {
         return ( DriverManager.getConnection(
                                             db_url, db_username, db_password) );
      } catch (SQLException se) {
         String err = "open connection failed";
         logError(err, se);
         throw new InitializationException ("ERROR: DBConnPool: " + 
                                            err + ": " + se.toString());
      }
   }


   /** calls connection.close() on the object. */
   public void destroyObject(Object obj)
   {
      try {
         if (obj instanceof Connection && ! ( (Connection) obj).isClosed()) {
            ( (Connection) obj).close();
         }
      } catch (SQLException se) {
         logError ("conn.close() failed", se);
      }
   }


   /**
    * gets a JDBC Connection object from the pool.  If the pool is currently
    * empty (meaning all connections are in use by other threads) and 
    * the current total is less than the configured maximum, a new 
    * connection is created and returned.  Otherwise this method blocks 
    * until a connection is released by another thread.  
    */ 
   public Connection getConnection()
   {
      return ( (Connection) getObject());
   }


   /**
    * returns this connection to the pool, allowing other threads to 
    * have access to it.  This should be called when the calling thread 
    * no longer needs the connection.
    */
   public void returnToPool(Object obj)
   {
      if (obj instanceof Connection) {
         try {
            if (! ((Connection) obj).isClosed()) {
               super.returnToPool(obj);
            } else {
               super.wipeFromPool(obj);
            }         
         } catch (SQLException se) {
            logError("connection cannot be returned! wiping...", se);
            super.wipeFromPool(obj);
         }
      }

   }

   /**
    * Start the DBConnectionPool service.
    * @exception ServiceException if the connection pool 
    *            not could be started.
    */
   public void startService() throws ServiceException {
      try {
         Class.forName(driver_classname);
      } catch (ClassNotFoundException cnfe) {
         throw new ServiceException("Failed loading database driver " +
                                    " class = " + driver_classname + 
                                    ". " + cnfe);
      }
      super.startService();
   }

   /**
    * Get name of the driver class for this connection pool.
    * @return Returns name of the driver class for this connection pool.
    */
   public String getDriverClassName(){
      return driver_classname;
   }

   /**
    * Set the driver class name for this connection pool.
    * @param driverClassName The name of the driver class for 
    *			      this connection pool.
    */
   public void setDriverClassName(String driverClassName){
      this.driver_classname = driverClassName;
   }

   /**
    * Get the url to the database in this connection pool.
    * @return Return the url to the database in this connection pool.
    */
   public String getUrl(){
      return db_url;
   }

   /**
    * Set the url to the database in this connection pool.
    * @param url the url to the database in this connection pool.
    */
   public void setUrl(String url){
      this.db_url = url;
   }

   /**
    * Get the database username to use for connections in this pool.
    * @return Returns the database username to use for connections
    *	        in this pool.
    */
   public String getUserName(){
      return db_username;
   }

   /**
    * Set the database username to use for connections in this pool.
    * @param userName The database username.
    */
   public void setUserName(String userName){
      this.db_username = userName;
   }

   /**
    * Get the password for the username.
    * @return Return the password.
    */
   private String getPassword(){
      return db_password;
   }

   /**
    * Set the password for the username.
    * @param password The password for the username.
    */
   public void setPassword(String password){
      this.db_password = password;
   }

}
