/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import org.locomotive.loco.*;
import org.locomotive.util.*;
import java.io.*;
import java.util.*;

/**
 * This class takes care of parsing out forms HTTP forms. The multipart forms 
 * algorithm just isn't very pretty, and probably should be rewritten...
 */
public class FormParser
{

    final static int LF   = 10;  // '\n'
    final static int CR   = 13;  // '\r'
    final static int DASH = 45;  // '-'

    final static int FILE = 1001;
    final static int TEXT = 1000;
    
    private String upload_path;

    InputStream is;       // inputstream to read from
    LocoConfig config;   // the config file to read path and max content lengths
    Hashtable form_hash;  // hashtable to return
    Hashtable cgi_envars_hash;  // holds the cgi environment
    Log server_log = null; // log to log to
    byte [] buffer;        // holds our data
    int total_read;        // the number of bytes read
    int cl;                // the content length
    int pointer;           // wehre we are in the buffer
    int num_read;          // the number of bytes read from the inputstream
    
    byte [] boundary;      // the boundary string as a byte array


    /**
     * instantiates the parser.
     * @param cgi_hash the hashtable containing the cgi environment
     * @param in the inputstream to read from
     * @param conf the config file to read such attributes as maximum
     *        form content size and the upload directory
     * @param log the log to write to
     * @exception org.locomotive.loco.ParseFormException if the content-length 
     *        is corrupted
     */
    public FormParser(InputStream in, Hashtable cgi_hash, LocoConfig conf,
		      Log log)
        throws ParseFormException
    {
        is = in;
        if (conf != null) config = conf;
        cgi_envars_hash = cgi_hash;
        server_log = log;
        cl = 0;
    
        String cls = (String) cgi_envars_hash.get("CONTENT_LENGTH");
        try 
        {
            cl = (cls == null ) ? 0 : Integer.parseInt(cls);
        }
        catch (NumberFormatException nfe)
        {
            throw new ParseFormException(
            "ERROR: Request: convert content-length (POST) failed");
        }
    }
    

    /**
     * this method returns a hashtable with all the data from the form
     * it will figure out whether the form is normal or 
     * mulitpart, and handle it accordingly.  <BR>
     * For multipart forms, all body parts containing a 'filename'
     * attribute are written to a file in the directory specified by
     * by the LOCO_UPLOAD_PATH config variable.
     * @exception org.locomotive.loco.ParseFormException if the content length of the 
     *            form being sent is longer than the _FORM_CONTENT_LENGTH 
     *            configs,
     *            or if the form is corrupted.
     */
    public Hashtable parseForm() throws ParseFormException
    {
        form_hash = new Hashtable();
	num_read = 0;
	total_read = 0;

	// if multipart, will look something like 
        // CONTENT_TYPE: multipart/form-data; boundary=-----------941115
        String formtype = (String) cgi_envars_hash.get("CONTENT_TYPE");
	if (formtype != null && formtype.indexOf("multipart") != -1)
        {
	    // Each body part is seperated by two dashes plus the boundary

	    int bindex = formtype.indexOf ("boundary=");
	    
	    if (bindex == -1 || ((bindex + 9) >= formtype.length()))
	    {
		parseNormalForm();
	    }
	    else
	    {
		String bstr = "--" + formtype.substring (bindex + 9);
		//  log(8, "boundary:" + boundary);   
		boundary = bstr.getBytes();
		parseMultipartForm();
	    }
        }
        else
        {
            parseNormalForm();
        }
        return form_hash;
    }
            


    /**
     * This method parses out form name-value pairs into a hashtable.
     * @return true if the parse is successful; false if otherwise.
     */
    private void parseNormalForm() throws ParseFormException
    {
        String namevalue = "";
        String name = "";
        String value = "";
	int maxlen = config.getInt("NORMAL_FORM_CONTENT_LENGTH");

        if ((cl < 0) || (cl > maxlen))
        {
            // content length is too big or zero 
            throw new ParseFormException("ERROR: invalid form size: " + 
                                     cl + " bytes!");
        }

	log (4, "Expecting " + cl + " bytes of form data");

        buffer = new byte [cl];

	try 
	{
	    while (total_read < cl)
	    {
		if ((num_read = is.read(buffer, total_read, cl - total_read)) 
		    == -1)
		{
		    // premature eof reading stdin 
		    throw new ParseFormException(
			"ERROR: premature eof reading stdin");
		}
		else
		{
		    total_read += num_read;
		}
	    }
	}
	catch (IOException ioe)
	{
	    throw new ParseFormException("Error reading from stdin: " + 
					 ioe.toString());
	}

        // now we have the form contents in a byte buffer
        String formdata = new String(buffer);
    
        log (8, formdata);   // for debugging!!
    
        StringTokenizer formdatast = new StringTokenizer (formdata, "&");
        StringTokenizer nvst;
        
        // name value pairs
        while ( formdatast.hasMoreTokens() )
        {
            namevalue = formdatast.nextToken();
            nvst = new StringTokenizer (namevalue, "=");
            name = nvst.nextToken();
    
            if (nvst.hasMoreTokens())
            {
                value = nvst.nextToken();
            }
            else 
            {
                value = "";
            }
    
            name = StringUtil.CGIUnescape (name);
            value = StringUtil.CGIUnescape (value);
    
            log (8, "name: " + name + ", value: " + value);
    
            if (form_hash.get(name) == null)
            {
                form_hash.put(name, value);
            }
            else
            {
                // form variable is already defined, might be a multiselect
                // list, so append the values together
                String value_previous = (String) form_hash.get (name);
                form_hash.put (name, value_previous + "," + value);
            }
        }
    }





    /**
     * this method does the parsing work for multipart forms.
     * @return the form_hash with the relevant information from the form
     * @throws ParseFormException if the form cannot be parsed correctly
     */ 
    private void parseMultipartForm() throws ParseFormException
    {
        pointer = 0;
        buffer = new byte [8192];
        boolean in_form = true;
	int maxlen = config.getInt("MULTIPART_FORM_CONTENT_LENGTH");
        
        if ((cl < 0) || (cl > maxlen) ) 
        {                               
            // too big or zero length
            throw new ParseFormException("ERROR: invalid form size: " + 
                                    cl + " bytes!");
        }

        log(8, "Expecting " + cl + " bytes of content");

	upload_path = config.getString("LOCO_UPLOAD_PATH");
	
	if (! upload_path.equals("none") && ! upload_path.startsWith("/"))
	{
	    upload_path = config.getString ("SYSTEM_SERVER_ROOT_PATH") + "/" + 
	        upload_path;
	}

        try
        {
            num_read = is.read(buffer);
            total_read+=num_read;
    
            if (num_read == -1)
            {
                throw new ParseFormException("Error reading from inputstream");
            }
    
            // skip the first boundary & line breaks
            pointer += boundary.length + 2;  
            
            while (in_form)
            {
                // read Headers
                String name =  readHeaders();
        
                // if there's a content type write it to file
                // both of these throw ParseFormExceptions
		if (form_hash.containsKey(name + ".filename"))
                {
                    writeData(name, FILE);
                }
                else       // otherwise, throw it in the form_hash
                {
                    writeData(name, TEXT);
                }
        
		// end of form reached if buffer is
		// equal to the end of the body - '-' = 45
                if ( buffer[pointer] == DASH  ) 
                {
                    in_form = false;
                }
                else if (buffer[pointer] == LF)
                {
                    pointer ++;
                }
                else
                {
                    throw new ParseFormException(
                        "Corrupted Form- bad boundary construction");
                }
            }
    
            log(4, "Multipart form parsed successfully");
            log (8, "Total read: " + total_read);
        }
        catch (IOException e)
        {
            throw new ParseFormException("Problem parsing multipart form:" + e);
        }
    }


    /** reads multipart headers from the input stream and parses them out
     * @return the name of the form object whose headers it parsed
     */
    private String readHeaders() throws IOException, ParseFormException
    {
        String name = "";
        int l_start = pointer;  // holds the place where the newest line started

        for (; pointer < num_read; pointer++)
        {
	    // we have to read more into the buffer.  
	    // Copy from the beginning of the line into the new buffer 
	    // and read as much as we can
            if (pointer == num_read - 1)
            {
		System.arraycopy(buffer, l_start, buffer, 0, 
				 pointer - l_start+1);
		pointer -= l_start;

                num_read = is.read(buffer, pointer + 1, 
				   buffer.length - pointer - 1);  
                total_read +=num_read;
                num_read += pointer + 1;
		l_start = 0;

                if (num_read == -1)
                {
                    throw new ParseFormException(
                       "Multipart form error: Unexpected EOF during data read");
                }
            }
    
            if (buffer [pointer] != LF)  // keep on reading into the buffer
            {
                continue;
            }
                
            // do the rest of this only when we encounter a line break
            if (pointer - l_start <= 2)  // end of header
            {
                pointer ++;
                return name;
            }
    
            String line = new String(buffer, l_start, pointer - l_start);
            
            l_start = pointer;
            pointer ++;
                
            StringTokenizer first = new StringTokenizer(line, ";");
            StringTokenizer s = 
                new StringTokenizer(first.nextToken().trim(), ":");
    
            String value = "";
            String str = s.nextToken().trim(); 
                
            if (! str.equalsIgnoreCase("Content-Disposition"))
            {
                value = 
		    StringUtil.CGIUnescape(removeQuotes(s.nextToken().trim()));
                str = name + "." + StringUtil.CGIUnescape(str);
                log (8, "name: " + str + " value: " +  value);
                form_hash.put(str, value);
            }
                
            while (first.hasMoreTokens())
            {
                s = new StringTokenizer(first.nextToken(), "=");
                str = s.nextToken().trim();
    
                if (str.equalsIgnoreCase("name"))
                {
                    value = s.nextToken().trim();
                    name = StringUtil.CGIUnescape(removeQuotes(value));
                }
                else
                {
		    value = StringUtil.CGIUnescape(
			removeQuotes(s.nextToken().trim()));
    	            str = name + "." + StringUtil.CGIUnescape(str);
                    log (8, "name: " + str + "\tvalue: " +  value);
                    form_hash.put(str, value);
                }
            }       
        }
        return name;
    }



    /** reads the normal body parts of a multipart form
     * @param name the name of the form input object we're parsing the data for
     * @param datatype constant TEXT if the object is text, 
     *                 or FILE if the object is a file
     * @exception IOException if there was a stream reading error
     * @exception ParseFormException if there was a problem parsing the text
     */
    private void writeData(String name, int datatype)
         throws IOException, ParseFormException
    {
        int l_start = pointer;     // holds the place in the current buffer
                            	   // where we started our read  
        boolean boundTest = false; // whether we are testing to see
	                           // if we are at the boundary 
        int bInd = 0;          // how many of the boundary bytes we have matched


        StringBuffer text = null;
        String filename = null;
        File file = null;
        FileOutputStream fos = null;

	// set up the variables
	switch (datatype)
	{
	case FILE:
	    filename = name + System.currentTimeMillis();
	    file = new File(upload_path, filename);
	    if (file.exists())
		throw new ParseFormException (
		    "Could not write file: File already exists!");

	    fos = new FileOutputStream(file);
	    break;

	case TEXT:
	    text = new StringBuffer();
	    break;
	}

	// the main loop
        for (; pointer < num_read; pointer++)
        {
	    // we read longer than the content length
            if (total_read > cl)
            {
               throw new ParseFormException("Form Content-length is corrupted");
            }
    
             // the last few bytes matched the boundary
             // we're done- break out of the loop
            if (bInd == boundary.length) 
            {
		// write the last item, if there is one
		if ((pointer - bInd- 2- l_start) > 0)
		{
		    switch (datatype)
		    {
		    case FILE:
			fos.write(buffer, l_start, 
				  pointer - bInd - 2 - l_start);
			break;
		    case TEXT:
			text.append( new String(buffer, l_start, 
						pointer - bInd- 2- l_start) );
			break;
		    }
		}
                pointer ++;
                break;
            }
            
            // if we are testing for boundary, check to see if the pointer
	    // byte equals the next byte in the boundary
            if (boundTest)
            {
                if (buffer [pointer] == boundary[bInd])
                    bInd++;
                else
                {
                    boundTest = false;
                    bInd = 0;
                }
            }
    
	    // all boundaries begin with a linefeed.  if pointer byte is one,
	    // then start looking for the boundary
            if (buffer[pointer] == LF)
            {
                boundTest = true;
            }
    
            // we're at the end of the buffer
            if (pointer == num_read - 1)
            {
		// only true if we're presently testing for a boundary
                if (bInd > 0 )
                {
                    // write everything up to the beginning of the matching 
                    // lines. Write that stuff to the beginning of the buffer.
                    //  Read the new stuff past it. Re-initialize our pointers
		    switch (datatype)
		    {
		    case FILE:
			fos.write(buffer, l_start, pointer-bInd-l_start);
			break;
		    case TEXT:
			text.append( new String(buffer, l_start, 
						pointer-bInd-l_start) );
			break;
		    }
                    System.arraycopy(buffer, pointer - bInd+1, buffer, 0, 
				     bInd+1);
                    pointer = bInd;
                    l_start = 0;

                    num_read = is.read(buffer, pointer + 1, 
				       buffer.length - pointer - 1);
                    total_read += num_read;
                    num_read +=pointer + 1;
                }
                else
                {
                    // write it all, read the whole buffer in if we can
		    switch (datatype)
		    {
		    case FILE:
			fos.write(buffer, l_start, num_read - l_start );
			break;
		    case TEXT:
			text.append(new String(buffer, l_start, 
					       num_read - l_start ));
			break;
		    }
		    
                    num_read = is.read(buffer);
                    pointer = -1;
                    l_start = 0;
                    total_read += num_read;
                }

                if (num_read == -1)
                {
                    throw new ParseFormException(
                        "Multipart form error: Unexpected EOF during data read");
                }
            }
        }
	// clean up, put our info into the hashtable and return
	String str = null;
	String value = null;

	switch (datatype)
	{
	case FILE:
	    fos.close();
	    str = name + ".path";
	    value = upload_path + "/" + filename;
	    break;
	case TEXT:
	    str = name;
	    value = text.toString();
	    break;
	}
	form_hash.put(str, value);
	log (8, "name: " + str + " value: " + value);
    }


    /** logs the message to the log specifed in the constructor, using the
      * specified log level 
      */
    public void log(int loglevel, String message)
    {
        if (server_log == null)  // basic hack to let us see the message
            System.out.println(message);
        else
        {
            server_log.log(loglevel, message);
        }
    }

    /** removes quotes from first and last positions in a string */
    private static String removeQuotes(String s)
    {
        String quote = "\"";
        if (s.startsWith(quote) && s.endsWith(quote) )
        {
            return s.substring(1, s.length() - 1);
        }
        else
        {
            return s;
        }
    }


}   // end of MultipartParser class


