/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import java.io.*;
import java.util.*;
import java.text.SimpleDateFormat;
import org.locomotive.directory.LocomotiveService;
import org.locomotive.directory.ServiceException;

/**
 * A simple logging facility.  Data can be logged to a file, along with 
 * a simple datestamp (datestamp can be turned off if desired), similiar
 * in form to the UNIX syslog facility.                              <BR><BR>
 *
 * At midnight, the previous days logfile will be closed, and a new one will
 * be opened.  A percent symbol should be placed into the log_path 
 * to allow the server to correctly open the new file.  An example 
 * log_path is the string &quot;logs/loco1-%.log&quot;.  The standard log
 * of the Loco is of this type, as is the Loco EventLog.
 * @see org.locomotive.loco.Loco
 */
public class Log implements LocomotiveService
{
    private String error_string;
    private RandomAccessFile raf;
    private boolean isopen = false;
    private int log_level = 0;
    private int day_opened = 0;
    private String original_input_path;
    private String root_path;
    private SimpleDateFormat sdf; 
    private Calendar cal;
    private StringBuffer sb;
    private boolean prepend_date = true;  // Default to datestamping

    /** If you don't want to do any logging at all, set the logging level 
       of the Log instance to this value.  Equivalent to 0. */
    public static final int NO_LOGGING = 0;

    /** Fatal error- Handler cannot continue.  Equivalent to 1. */
    public static final int FATAL_ERROR = 1;

    /** Critical Error- Handler severely disabled. Equivalent to 2.  */
    public static final int CRITICAL_ERROR = 2;

    /** Error. equivalent to 3. */
    public static final int ERROR = 3;

    /** Potential Problem.  Equivalent to 4. */
    public static final int WARNING = 4;

    /** Important information, including the URL tokens for each request.
     *  Equivalent to 5. */
    public static final int NOTICE = 5;

    /** Useful information.  Equivalent to 6. */
    public static final int INFO = 6;

    /** Debug information. Useful for development. Equivalent to 7. */
    public static final int DEBUG = 7;

    /** More Debug information, including the CGI environment for each
     * request.  Equivalent to 8. */
    public static final int DEBUG_VERBOSE = 8;
    
    /** All debug information, including internal Locomotive state logging.
     * Equivalent to 9. */
    public static final int DEBUG_REALLY_VERBOSE = 9;

    private boolean isRunning = false; 


    /**
     * Create a new Log.
     * @param lev   the logging level for this log
     * @param root  the path to the server root.  
     * @param log_path the log path, relative to the server root.  If this 
     *               string contains a '%' (percent), the current date will 
     *               be substituted into the filename at that location in the 
     *               path.
     */
    public Log (int lev, String root, String log_path) throws IOException
    {
	log_level = (lev < NO_LOGGING) ? NO_LOGGING : lev;

	root_path = root;
	original_input_path = log_path;

	init();
    }
    
    /**
     * Default constructor for the Log class.
     */
    public Log(){
    }

    private synchronized void init() throws IOException {
	sdf = new SimpleDateFormat ("yyyy-MM-dd HH:mm:ss z");
	cal = Calendar.getInstance ();
	sb = new StringBuffer ();

	if (! createNewLog ())
	{
	    throw new IOException (error_string);
	}
    }
    
    /**
     * Start the log service.
     * @exception ServiceException is thrown if the log service not
     *		  could be started.
     */
    public synchronized void startService() throws ServiceException {
	try {
	    init();
	} catch(IOException ioe){
	    throw new ServiceException(ioe.toString());
	}
	isRunning = true;
    }
    
    /**
     * Stop the service.
     * @exception ServiceException is throws if this service 
     *		  not could be stopped.
     */
    public synchronized void stopService() throws ServiceException {
	closeLog();
	isRunning = false;
    }

    /** 
     * Check if the service is running.
     * @return false if the service not is running, true otherwise.
     */
    public boolean isRunning(){
	return isRunning;
    }
    
    /** 
     * Adjusts the logging level to the specified level.  Calls to
     * log () with a level higher than that level previously set with
     * a setLevel() call or the constructor will be ignored.  Example:
     * if your current logging level in the Log object is ERROR, 
     * calls with logging level DEBUG will be ignored, 
     * but with level CRITICAL_ERROR it will be written 
     * to the logfile.  
     *                                                              <BR><BR>
     * @param newlev the new logging level.
     */
    public synchronized void setLevel(int newlev)
    {
	log_level = (newlev < NO_LOGGING) ? 
	    NO_LOGGING : newlev;
    }

    // called internally when new day
    private synchronized boolean createNewLog ()
    {
	isopen = false;
	error_string = "";

	if (original_input_path == null)
	{
	    return (true);
	}
	
	String logpath = getFullLogPath();
	try 
	{
	    raf = new RandomAccessFile(logpath, "rw");
	    raf.seek (raf.length());
	}
	catch (IOException ioe)
	{
	    error_string = ioe.toString();
	    return (false);
	}
	isopen = true;
	return (true);
    }

    private String getFullLogPath ()
    {

	String path = original_input_path;

	if (path.indexOf ("/") != 0)
	{
	    path = root_path + "/" + path;
	}

	Date dt = new java.util.Date ();
	SimpleDateFormat localsdf = new SimpleDateFormat ("yyyyMMdd");

	cal.setTime (dt);
	day_opened = cal.get (Calendar.DAY_OF_MONTH);
	
	StringBuffer sb = new StringBuffer ();
	char c;
	
	for (int i = 0; i < path.length(); i++)
	{
	    c = path.charAt (i);
	    if (c != '%')
	    {
		sb.append (c);
	    }
	    else
	    {
		sb.append (localsdf.format (dt));
	    }
	}
	return (sb.toString());
    }
    
    /**
     * logs an line to the logfile, containing the string s.  If the 
     * lev parameter is greater than the current log level, the call 
     * will simply return, without writing.  
     * @param lev  the level of this log request.  If higher than the 
     *             current logging level, this request will be ignored
     * @param s    the string to be written to the log
     */
    public synchronized void log(int lev, String s)
    {
	if (lev <= log_level)
	{
	    Date dt = new Date();
	    cal.setTime (dt);
	    if (day_opened != cal.get (Calendar.DAY_OF_MONTH))
	    {
		closeLog();
		createNewLog();
	    }
	    
	    if (isopen)
	    {
		// Prepend the date only if we should..
		if (prepend_date)
		{
		    sb.append (sdf.format (dt));
		    sb.append (" ");
		}
		sb.append (s);
		sb.append ("\n");

		try
		{
		    raf.writeBytes (sb.toString());
		}
		catch (IOException ioe)
		{
		    closeLog();
		    System.out.println("write to log failed: " + 
				       ioe.toString());
		}
		
		sb.setLength (0);
	    }
	}
    }

    /**
     * Closes this Log.  Any subsequent writes to this Log are simply 
     * ignored.  
     */
    public void closeLog ()
    {
	isopen = false;

	try
	{
	    raf.close();
	}
	catch (IOException ioe) {}
    }
    
    /**
     * Set the root path of that the log file will be written 
     * relative to. Normally this is the server install directory.
     *
     * @param rootPath Root path for the log.
     */  
    public void setRootPath(String rootPath)
    {
	this.root_path = rootPath;
    }

    /**
     * Get the root path.
     * @return Returns the root path.
     */
    public String getRootPath()
    {
	return root_path;
    }

    /**
     * Relative path name to the log file.
     * @param logPath Relative pathname of the log.If this 
     *               string contains a '%' (percent), the current date will 
     *               be substituted into the filename at that location in the 
     *               path.
     */
    public void setLogPath(String logPath)
    {
	this.original_input_path = logPath;
    }

    /**
     * Get relative path to the log file.
     * @return Returns the relative path to the log file.
     */
    public String getLogPath()
    {
	return original_input_path;
    }

    /**
     * Call this method with a prepend value of false to turn off the
     * automatic prepending of the date on each log output line.
     *                                                              <BR><BR>
     * @param prepend make this true if you want the Log object to
     *        prepend the date to the beginning of your log output
     *        lines, or false to turn off the datestamping.
     */
    public synchronized void setPrependDate(boolean prepend)
    {
	prepend_date = prepend;
    }
}
