/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import java.util.*;
import java.io.*;
import org.locomotive.util.*;

/**
 * A handy config file that allows the specification of multiple values per
 * name and the reading of data either from the method call data passed in
 * or from a file path specified. After parsing the config data you can then
 * get them with the traditional getString and getInt methods found in
 * org.locomotive.server.Config or you can ask for specific indexes. You should
 * extend this class and override the setupDefaults function to make your own
 * configs. This version does not use string constraints. Nor does it have
 * any command line stuff. Nor does it print to standard out. As such, errors
 * are reported (as they kind of are in the old Config) with exceptions and
 * the like, so watch those ConfigInitializationException's and
 * ConfigNotFoundException's. It does some constraining based on type. The
 * strings that represent integers have to be Integer.parseInt'able, the
 * strings have to be non-null and non-emptystring. Right now the email
 * constraints are the same as strings, though we may put in a x@x constraint
 * for it... It also allows you to turn off the name-checking... that is, you
 * can have "unknown" config variables be read in and set. Override
 * enforceConfigTypes to do this (see it for more info.) Override
 * getDefaultConfigTypes to tell this parser what to expect on this unfamiliar
 * lines (it won't try to guess for you...)
 */
public abstract class MultiConfig
{
    /**
     * These three members STRING, INTEGER, and EMAIL are type values you
     * can use when specifying an array of types for the addConfig method.
     */
    public final static int STRING = 0;
    public final static int INTEGER = 1;
    public final static int EMAIL = 2;

    /**
     * These three members are the type arrays, <B>provided for convenience
     * only.</B> I.e. instead of &quot;new int[] {INTEGER}&quot; you can just
     * use one of these (in the case of that example, just use "INTEGER").
     */
    public final static int[] ONE_STRING_TYPE = {STRING};
    public final static int[] ONE_INT_TYPE = {INTEGER};
    public final static int[] ONE_EMAIL_TYPE = {EMAIL};

    /**
     * Same as above, only containing two-in-a-row, just in case it's handy.
     * I.e. &quot;TWO_STRING_TYPE&quot; is equivalent to
     * &quot;new int[] {STRING, STRING}&quot;.
     */
    public final static int[] TWO_STRING_TYPE = {STRING, STRING};
    public final static int[] TWO_INT_TYPE = {INTEGER, INTEGER};
    public final static int[] TWO_EMAIL_TYPE = {EMAIL, EMAIL};

    private String local_string_data;
    private Hashtable config_hash;
    private boolean local_isFile = false;

    /**
     * Make the config object with the file path or the string data as
     * specified. If isFile is true, it will interperet the string as a
     * file path and try to load it. If isFile is false it will use the
     * string passed in as the config file itself. I.e. this string should
     * then contain all the newlines and formatting and everything that the
     * actual config file would were it an external file. Throws an
     * ConfigInitializationException if it sees errors loading a file or
     * checking the supplied strings against their type constraints.
     * Use within a program with no file access might look like this:
     * <XMP>
     *  StringBuffer sb = new StringBuffer(123);
     * 	sb.append("OTHER_EXAMPLE stringone stringtwo\n");
     * 	sb.append("YET_ANOTHER_EXAMPLE stringone stringtwo 323 stringthree\n");
     * 	sb.append("AND_ANOTHER_EXAMPLE stringone 457\n");
     *
     * 	MultiConfigExtension mc = new MultiConfigExtension(sb.toString(),
     *                                                     false);
     * </XMP>
     *
     * Config files and the strings of the same content are pretty felxible.
     * White space is ignored. Newlines signify a new data item, the first item
     * on the line is the name and following items (seperated by whitespace)
     * are its associated values.
     * @param string_data  a filepath or a buffer of data in the config format
     * @param isFile      true if string_data is a file
     * @exception ConfigInitializationException if the string data is bogus or 
     *            if there was trouble loading from the file
     */
    public MultiConfig(String string_data, boolean isFile)
	throws ConfigInitializationException
    {
	config_hash = new Hashtable(50);
	local_string_data = string_data;
	local_isFile = isFile;

	// set up all default vars and constraints
	if (!setupDefaults())
	{
	    throw new ConfigInitializationException("Internal Configuration " +
						    "Object Errors");
	}

	String errors = null;

	// override defaults from config info
	errors = loadConfigs(local_string_data, isFile);

	if (! errors.equals(""))
	{
	    throw new ConfigInitializationException("Configuration Data " + 
						    "Errors:\n" + errors);
	}

	// validate the resulting config hash
	errors = validate();

	if (! errors.equals(""))
	{
	    throw new ConfigInitializationException("Configuration Data " +
						    "Errors:\n" + errors);
	}
    }

    /**
     * Override this method as shown below to set up your own configs.
     * 
     * Historical Note: some example configs are from the original MAS.
     *
     * <XMP>
     * boolean setupDefaults()
     * {
     *  boolean b = true;
     * 
     * 	b &= addConfig ("MAS_EXAMPLE", new String[] {"1.0b3"},
     *   new int[] {STRING}, "MAS System Version String");
     *
     * 	b &= addConfig ("LOCO_EXAMPLE", new String[] {"5420"},
     *       new int[] {INTEGER}, "TCP Port for Application Server");
     *
     * 	b &= addConfig ("AND_ANOTHER_EXAMPLE",
     * 					 new String[] {"stringone", "457"},
     * 					 new int[] {STRING, INTEGER},
     * 					 "another example");
     *
     * 	b &= addConfig ("MAS_VERSION", oneString("1.0b3"), ONE_STRING_TYPE,
     * 	     "MAS System Version String");
     *
     * 	b &= addConfig ("LOCO_PORT", oneString("5420"), ONE_INT_TYPE,
     * 	     "TCP Port for Application Server");
     *
     * 	b &= addConfig ("OTHER_EXAMPLE", twoString("stringone", "stringtwo"),
     * 		   TWO_STRING_TYPE, "another example");
     *
     * 	b &= addConfig ("YET_ANOTHER_EXAMPLE",
     * 		   new String[] {"stringone", "stringtwo"},
     * 		   TWO_STRING_TYPE, "another example");
     *
     *  b &= addConfig ("AND_THE_LAST_ONE",
     * 		   twoString("stringone", "hi"),
     * 		   new int[] {STRING, STRING}, "descriptive text");
     *
     *  return(b);
     * }
     * </XMP>
     */
    public abstract boolean setupDefaults();

    /**
     * Override this function to simply return true or false. If it returns
     * true then the config object will expect that it knows the form of
     * any line it encounters. If it returns false then it will look to
     * getDefaultConfigTypes() for the form and try to make the line fit
     * that. This implementation (i.e. the default) returns true.
     */
    public boolean enforceConfigTypes()
    {
	return(true);
    }

    /**
     * If the enforceConfigNames method you implement returns false,
     * then when the parser encoutners a strange line it will use the
     * types specified in this method as the default form of the line.
     * Lines that don't match this form will be rejected with errors.
     * If the enforceConfigNames method you implement returns true,
     * this will never be called. You can override this method to
     * specify what form the parser should assume when it encounters
     * unfmailiar lines and the enforceConfigTypes function returns
     * false. This implementation (i.e. the default) returns ONE_STRING_TYPE.
     */
    public int[] getDefaultConfigTypes()
    {
	return(ONE_STRING_TYPE);
    }

    /**
     * Tells us if this config file came from a file path that we loaded.
     */
    public boolean isFromFile()
    {
	return(local_isFile);
    }

    /**
     * Tells us if this config file came from string information given to it
     * in the constructor method call.
     */
    public boolean isFromString()
    {
	return(!local_isFile);
    }

    /**
     * Gives us the file path this config file was constructed with (null if
     * it was actually constructed with string config data passed in...)
     */
    public String getConfigFilePath()
    {
	if (local_isFile)
	    return(local_string_data);
	else
	    return(null);
    }

    /**
     * Gives us the string config data given to the config object when it was
     * created (null if it was actually created with a file path.)
     */
    public String getConfigStringData()
    {
	if (!local_isFile)
	    return(local_string_data);
	else
	    return(null);
    }

    // Checks each MultiConfigEntry for ok-ness with its type constraints
    private String validate()
    {
	String name;
	MultiConfigEntry ce;
	StringBuffer error_sb = new StringBuffer ();

	for (Enumeration e = config_hash.keys(); e.hasMoreElements(); )
	{
	    name = (String) e.nextElement();
	    ce = (MultiConfigEntry) config_hash.get (name);
	    if (ce != null)
	    {
		error_sb.append(ce.check());
	    }
	    else
	    {
		error_sb.append ("CE for " + name + " was null!");
	    }
	}

	return (error_sb.toString());
    }

    /**
     * Create and add a config entry to the cache we're maintaining.
     * @param name the name, or key, of the entry you wish to create
     * @param vals the array of strings that represent the values you
     * wish to add
     * @param types the array of types that describe the values you are
     * adding (see INTEGER, STRING, EMAIL, et. al. above for info about types).
     * Each type value in this array should correspond to the string value
     * in the same position of the array.
     * @param desc the string description of the config entry, perhaps
     * containing info about each value in the array, perhaps not.
     */
    protected boolean addConfig(String name, String[] vals, int[] types,
				String desc)
    {
	if ((vals != null) && (types != null) && (name != null) &&
	    (! name.equals("")) && (vals.length == types.length))
	{
	    MultiConfigEntry ce = new MultiConfigEntry (name, vals, types, desc);
	    config_hash.put (name, ce);
	    return(true);
	}
	else return(false);
    }

    private boolean addDefaultTypedConfig(String name, String[] vals,
					  String desc)
    {
	if ((vals != null) && (name != null) &&
	    (! name.equals("")) &&
	    (vals.length == getDefaultConfigTypes().length))
	{
	    MultiConfigEntry ce = new MultiConfigEntry(name, vals,
						       getDefaultConfigTypes(),
						       desc);
	    config_hash.put (name, ce);
	    return(true);
	}
	else return(false);	
    }

    /**
     * Returns the array of string objects for the config entry matching the
     * supplied name (i.e. key) - will throw an exception if &quot;name&quot;
     * is null or &quot;&quot; or the config entry found is the null object.
     * @param name the name of the config 
     * @exception if a config of the mentioned name is not available
     */
    public String[] getStringArray(String name) throws ConfigNotFoundException
    {
	if ((name == null) || (name.equals("")))
	{
	    throw new ConfigNotFoundException("name can not be null or \"\"");
	}

	MultiConfigEntry ce = (MultiConfigEntry) config_hash.get (name);

	if (ce == null)
	{
	    throw new ConfigNotFoundException ("config entry " + name +
					       " was null");
	}

	return (ce.svalues);
    }

    /**
     * Returns the array of integer values for the config entry matching the
     * supplied name (i.e. key) - will throw an exception if &quot;name&quot;
     * is null or &quot;&quot; or the config entry found is the null object.
     * @param name the name of the config 
     * @exception if a config of the mentioned name is not available
     */
    public int[] getIntArray(String name)
	throws ConfigNotFoundException
    {
	if ((name == null) || (name.equals("")))
	{
	    throw new ConfigNotFoundException("name can not be null or \"\"");
	}

	MultiConfigEntry ce = (MultiConfigEntry) config_hash.get (name);

	if (ce == null)
	{
	    throw new ConfigNotFoundException("config entry " + name + 
					      " was null");
	}

	return (ce.ivalues);
    }

    /**
     * Gets the array of strings matching the key passed in in &quot;name&quot;
     * and returns the string found at the index specified - will throw an
     * exception if the index specified is invalid for this config entry or
     * &quot;name&quot; is null or &quot;&quot; or the config entry found is
     * the null object.
     * @param name the name of the config 
     * @param index - the position of the requested data in the string array
     * @exception if a config of the mentioned name is not available
     */
    public String getStringByIndex(String name, int index)
	throws ConfigNotFoundException
    {
	String[] s = getStringArray(name);

	if ((s.length > index) && (index >= 0)) return(s[index]);
	else throw new ConfigNotFoundException("config entry " + name +
					       " has no value at index " +
					       index);
    }

    /**
     * Gets the array of integers matching the key passed in in
     * &quot;name&quot; and returns the value found at the index specified -
     * will throw an exception if the index specified is invalid for this
     * config entry or &quot;name&quot; is null or &quot;&quot; or if the
     * config entry found is the null object.
     * @param name the name of the config 
     * @param index - the position of the requested data in the int array
     * @exception if a config of the mentioned name is not available
     */
    public int getIntByIndex(String name, int index)
	throws ConfigNotFoundException
    {
	int[] i = getIntArray(name);

	if ((i.length > index) && (index >= 0)) return(i[index]);
	else throw new ConfigNotFoundException("config entry " + name +
					       " has no value at index " +
					       index);
    }

    /**
     * Gets the first element of the array of strings that come from the
     * config entry that matches the key (i.e. &quot;name&quot;) specified.
     * @see getStringByIndex 
     * @param name the name of the configuration value
     * @exception if a config of the mentioned name is not available
     */
    public String getString(String name)
	throws ConfigNotFoundException
    {
	return(getStringByIndex(name, 0));
    }

    /**
     * Gets the first element of the array of integers that come from the
     * config entry that matches the key (i.e. &quot;name&quot;) specified.
     * @see getIntByIndex 
     * @param name the name of the configuration value
     * @exception if a config of the mentioned name is not available
     */
    public int getInt(String name)
	throws ConfigNotFoundException
    {
	return(getIntByIndex(name, 0));
    }

    /**
     * Standard toString with a twist - the boolean determines if the
     * descriptive text comes out as well.
     */
    public String toString(boolean verbose)
    {
	StringBuffer sb = new StringBuffer(75);

	for (Enumeration e = config_hash.keys(); e.hasMoreElements(); )
	{
	    String name = (String) e.nextElement();
	    MultiConfigEntry ce = (MultiConfigEntry) config_hash.get (name);

	    sb.append(ce.toString(verbose));
	}

	return(sb.toString());
    }

    /**
     * Calls the other toString with false.
     */
    public String toString()
    {
	return(toString(false));
    }

    private boolean isComment(String s)
    {
	if (s == null) return(false);

	if (s.length() == 0)
	{
	    return (true);
 	}

	if (s.charAt(0) == '#')
	{
	    return (true);
	}

	return (false);
    }

    private String loadConfigs(String string_data, boolean isFile)
    {
	FileInputStream fis = null;
	BufferedReader br = null;

	if (string_data == null)
	{
	    if (isFile)
		return("\"" + string_data + "\" passed as file name.");
	    else
		return("\"" + string_data + "\" passed as config " +
		       "description.");
	}
	else if (string_data.equals("")) return("");
	// reeturn emtpy stringbuffer tostring form, so "" not 'null'.

	if (isFile)
	{
	    try
	    {
		fis = new FileInputStream(string_data);
	    }
	    catch (FileNotFoundException fnfe)
	    {
		return ("Config File not found at: " + string_data);
	    }

	    br = new BufferedReader(new InputStreamReader(fis));
	}
	else
	{
	    br = new BufferedReader(new CharArrayReader(
		string_data.toCharArray()));
	}

	StringBuffer msgbuf = new StringBuffer();
	int linenum = 0;
	int token_count = 0;
	String name = null;
	Vector tok_vec = new Vector(2);
	String line;

	try
	{
	    while ((line = br.readLine()) != null)
	    {
		linenum++;
		// tokenize the lines
		StringTokenizer st = new StringTokenizer(line);

		// skip blank lines
		if ((token_count = st.countTokens()) == 0)
		{
		    continue;
		}
		else if (token_count == 1)
		{
		    // if only one token, make sure it is a comment
		    name = st.nextToken();

		    if (! isComment(name))
		    {
			if (! config_hash.containsKey(name))
			{
			    msgbuf.append("Line: " + linenum +
					  " bogus config name, also has no " +
					  "value: " + name + "\n");
			}
			else
			{
			    msgbuf.append("Line: " + linenum +
					  " valid config name: " + name +
					  ", requires value\n");
			}
		    }
		}
		else
		{
		    // hmm, at least two tokens

		    name = st.nextToken();

		    if (isComment (name))
		    {
			continue;
		    }
		    else
		    {
			tok_vec.removeAllElements();

			while (st.hasMoreTokens())
			{
			    tok_vec.addElement(st.nextToken());
			}

			// we have two tokens which we believe to contain 
			// a name/value list pair

			// check if it is a valid token
			if (( !config_hash.containsKey(name) ) &&
			    enforceConfigTypes())
			{
			    msgbuf.append("Line: " + linenum +
					  " bogus config name: " + name +
					  "\n");
			}
			else
			{
			    if ((!config_hash.containsKey(name)) &&
				!enforceConfigTypes())
			    {
				// then we want to dynamically create an
				// entry. We do it this way (create it and
				// then write it again below) so that it
				// will be marked as "modified".
				if (tok_vec.size() !=
				    getDefaultConfigTypes().length)
				{
				    msgbuf.append("Line: " + linenum +
						  " (dynamic) had " +
						  "inconsistant number of " +
						  "values: " + name + "\n");
				    continue;
				}
				else
				{
				    String svals[] = new String[
					tok_vec.size()];

				    tok_vec.copyInto(svals);

				    if (!addDefaultTypedConfig(name, svals,
						       "Dynamically created " +
						       "config entry. No " +
						       "description " +
						       "available."))
				    {
					msgbuf.append("Line: " + linenum +
						      " could not be " +
						      "dynamically " +
						      "created.");
					continue;
				    }
				}
			    }

			    MultiConfigEntry ce = (MultiConfigEntry)
				config_hash.get(name);

			    String svals[] = new String[tok_vec.size()];

			    tok_vec.copyInto(svals);

			    if (svals.length != ce.types.length)
			    {
				msgbuf.append("Line: " + linenum +
					      " had inconsistant number of " +
					      "values: " + name + "\n");
			    }
			    else
			    {
				ce.storeVal(svals);
				ce.altered = true;

				msgbuf.append(ce.check());

				// If there were any previous errors as well we
				// won't add anything new to the hash but we do
				// check above to make sure it's OK...
				if (msgbuf.length() == 0)
				    config_hash.put(name, ce);
			    } // end of if svals and types are the same length
			} // end if the config hash has the key
		    } // end if the line is not a comment
		} // end if at least 2 tokens
	    } // end while loop that reads file

	    if (isFile) fis.close();

	    br.close();

	} // end try
	catch (IOException ioe)
	{
	    return ("Error Reading Config File: " + ioe.toString());
	}

	return (msgbuf.toString());
    }

    /**
     * Returns the enumeration of all the config entry names (i.e. an
     * enumeration of all the keys).
     */
    public Enumeration getConfigNamesEnumeration()
    {
	return(config_hash.keys());
    }

    /**
     * Give a string array of one element as provided. For use with addConfig.
     */
    protected String[] oneString(String instring)
    {
	return(new String[] {instring});
    }

    /**
     * Give a string array of two elements as provided. For use with addConfig.
     */
    protected String[] twoString(String instringone, String instringtwo)
    {
	return(new String[] {instringone, instringtwo});
    }
}

class MultiConfigEntry
{
    String name;
    String description;

    String[] svalues;
    int[] ivalues;
    int[] types;

    boolean altered;

    protected MultiConfigEntry(String name, String[] values, int[] types,
			       String description)
    {
	if ((values == null) || (types == null) || (name == null)) return;
	else if (values.length != types.length) return;

	this.name = name;
	this.description = description;
	this.types = types;

	storeVal (values);

	altered = false;
    }

    protected void storeVal (String[] s)
    {
	if (s.length != types.length) return;

	svalues = s;

	ivalues = new int[svalues.length];

	for (int i = 0; i < s.length; i++)
	{
	    try 
	    {
		ivalues[i] = Integer.parseInt (s[i]);
	    }
	    catch (NumberFormatException nfe)
	    {
		ivalues[i] = 0;
	    }
	}
    }

    protected String check()
    {
	StringBuffer sb = new StringBuffer(10);

	for (int i = 0; i < svalues.length; i++)
	{
	    if (types[i] == MultiConfig.INTEGER)
	    {
		if (svalues[i] == null)
		{
		    sb.append("CE " + name + " value #" + (i + 1) +
			      " is an integer and its string is null.");
		}
		else if (svalues[i] == "")
		{
		    sb.append("CE " + name + " value #" + (i + 1) +
			      " is an integer and its string is \"\"");
		}
		else
		{
		    try
		    {
			Integer.parseInt(svalues[i]);
		    }
		    catch (NumberFormatException nfe)
		    {
			sb.append("CE " + name + " value #" + (i + 1) +
				  " is an integer type but does not parse.");
		    }
		}
	    }
	    else if (types[i] == MultiConfig.STRING)
	    {
		if (svalues[i] == null)
		{
		    sb.append("CE " + name + " value #" + (i + 1) +
			      " is a string and is null.");
		}
		else if (svalues[i] == "")
		{
		    sb.append("CE " + name + " value #" + (i + 1) +
			      " is a string and is \"\"");
		}
	    }
	    else if (types[i] == MultiConfig.EMAIL)
	    {
		if (svalues[i] == null)
		{
		    sb.append("CE " + name + " value #" + (i + 1) +
			      " is an email address and its string is null.");
		}
		else if (svalues[i] == "")
		{
		    sb.append("CE " + name + " value #" + (i + 1) +
			      " is an email address and its string is \"\"");
		}
/*		else
		{
		    StringTokenizer st = new StringTokenizer(svalues[i], "@");

		    if (st.countTokens() != 2)
		    {
			sb.append("CE " + name + " value #" + (i + 1) +
				  " is an email address and is not " +
				  "syntactically correct");
		    }
		    else
		    {

		    }
		}
		*/
	    }
	}

	return(sb.toString());
    }

    protected String toString(boolean verbose)
    {
	return(name + "\t" + stringFromArray(svalues) +
	       "\t" + stringFromArray(ivalues) + "\t" +
	       stringFromArray(types) + "\t" +
	       (altered ? "nondefault" : "default") +
	       (verbose ? " \"" + description + "\"" : "") + "\n");
    }

    private static String stringFromArray(int[] i_ra)
    {
	StringBuffer sb = new StringBuffer(10);

	sb.append("[");

	for (int i = 0; i < i_ra.length; i++)
	{
	    if (i != i_ra.length - 1)
		sb.append(i_ra[i]).append(";");
	    else
		sb.append(i_ra[i]);
	}

	return(sb.append("]").toString());
    }

    private static String stringFromArray(String[] s_ra)
    {
	StringBuffer sb = new StringBuffer(10);

	sb.append("[");

	for (int i = 0; i < s_ra.length; i++)
	{
	    if (i != s_ra.length - 1)
		sb.append(s_ra[i]).append(";");
	    else
		sb.append(s_ra[i]);
	}

	return(sb.append("]").toString());
    }
}
