/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import java.util.*;
import java.net.*;
import java.io.*;

/**
 * This class can be used to have the VM deliver an email to an external
 * SMTP server - typically UNIX sendmail.
 */
public class SMTPClient implements Runnable 
{
    private String server_name;
    private int server_port;
    private String address_to;
    private String address_from;
    private String subject_line;
    private String body;
    private Log log;
    private int log_id = 0;
    private String helo_namestring = "use_lookup";
    
    private Socket sock;
    private InputStream is;
    private OutputStream os;
    private BufferedReader br;
    private DataOutputStream dos;
    
/* attachment stuff added by ck - ugly! */
    private String attach_body;
    private String filename;
    private String body_header = 
    "Content-Type: text/plain; charset=\"us-ascii\"\r\n";
    private String boundary = "--=====================_881123928==_";
    private String attach_header = 
    "Content-Type: text/plain; charset=\"iso-8859-1\"\r\nContent-Transfer-Encoding: quoted-printable\r\nContent-Disposition: attachment; filename=\"";
    private int timeout = 20;  // timeout in seconds


    /** Creates a new SMTPClient.
     * @param svrname the name of the SMTP server
     * @param svrport the port the server is listening on for requests
     * @param to the email address to send the message to
     * @param from the email address the the message is from
     * @param subject the subject of the message
     * @param body_text the main text of the message
     */
    public SMTPClient (String svrname, int svrport, String to, String from,
		       String subject, String body_text)
    {
	server_name = svrname;
	server_port = svrport;
	address_to = to;
	address_from = from;
	subject_line = subject;
	body = body_text;
	helo_namestring = "use_lookup";
    }



    /** Creates a new SMTPClient.
     * @param svrname the name of the SMTP server
     * @param svrport the port the server is listening on for requests
     * @param to the email address to send the message to
     * @param from the email address the the message is from
     * @param subject the subject of the message
     * @param body_text the main text of the message
     * @param id a number to identify this client by for logging purposes
     */
    public SMTPClient (String svrname, int svrport, String to, String from,
		       String subject, String body_text, int id)
    {
	server_name = svrname;
	server_port = svrport;
	address_to = to;
	address_from = from;
	subject_line = subject;
	body = body_text;
	log_id = id;
	helo_namestring = "use_lookup";
    }

    /** Creates a new SMTPClient.
     * @param svrname the name of the SMTP server
     * @param svrport the port the server is listening on for requests
     * @param to the email address to send the message to
     * @param from the email address the the message is from
     * @param subject the subject of the message
     * @param body_text the main text of the message
     * @param id a number to identify this client by for logging purposes
     * @param helo the name of the client host to be used for the SMTP 
     *        HELO handshake.  If it has the value &quot;use_lookup&quot;
     *        the SMTPClient will attempt a name lookup on the localhost. 
     *        This is the default behavior for the other constructors. 
     * @deprecated  -  This was added to get around bugs that some sites 
     *         have been having, thought to be a problem with the JDK or 
     *         Solaris.  You probably don't need it.   
     */
    public SMTPClient (String svrname, int svrport, String to, String from,
		       String subject, String body_text, int id, String helo)
    {
	server_name = svrname;
	server_port = svrport;
	address_to = to;
	address_from = from;
	subject_line = subject;
	body = body_text;
	log_id = id;
	helo_namestring = (helo == null) ? "use_lookup" : helo;
    }

    /** Creates a new SMTPClient, with an attachment.  Presently, you
     * have to read in the file yourself, then specify the file name
     * and the contents of the file as a string.  This will be eventually
     * changed to specifying just a File object.  This method of
     * attachment does not work for binary files.
     * @param svrname the name of the SMTP server
     * @param svrport the port the server is listening on for requests
     * @param to the email address to send the message to
     * @param from the email address the the message is from
     * @param subject the subject of the message
     * @param body_text the main text of the message
     * @param id a number to identify this client by
     * @param to_attach the text of the attachment
     * @param filename the name of the file the attachment text was read
     *        in from
     */
    public SMTPClient(String svrname, int svrport, String to, String from,
		      String subject, String body_text, int id, 
		      String to_attach, String filename)
    {
	
	this(svrname, svrport, to, from, subject, body_text, id);
	this.attach_body = to_attach;
	if (filename == null)
	    this.filename = "";
	else
	    this.filename = filename;
    }
    
    /** set the number of seconds before the Mail client will stop
     * trying the send the message
     */
    public void setTimeout(int seconds)
    {
	if (seconds >= 0)
	    timeout = seconds;
    }


    /**
     * Starts the delivery process.  To avoid hanging the calling
     * thread, use true for use_own_thread to allow asynchronous delivery
     * "log" may be null if no logging is desired.
     */
    public void go (Log log, boolean use_own_thread)
    {
	this.log = log;

	if (use_own_thread)
	{
	    Thread t = new Thread (this);
	    t.start();
	}
	else
	{
	    run ();
	}
    }


    /** required to implement Runnable, to allow sending mail messages
     * on their own thread.  
     * @see #go
     */
    public void run ()
    {
	// get the servername
	InetAddress server_address[];

	try
	{
	    server_address = InetAddress.getAllByName (server_name);
	}
	catch (UnknownHostException uhe)
	{
	    log (0, "ERROR: " + uhe.toString());
	    return;
	}

	// open the socket
	try
	{
	    sock = new Socket (server_address[0], server_port);
	    sock.setSoTimeout(timeout*1000);
	    is = sock.getInputStream();
	    os = sock.getOutputStream();
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: " + ioe.toString());
	    return;
	}

	br = new BufferedReader (new InputStreamReader (is));
	dos = new DataOutputStream(os);

	String header = getHeader();
	String output;
	String resp_line;

	try
	{
	    // read greeting
	    resp_line = br.readLine();
	    log (5, "greeting: " + resp_line);
	    log(1, "Attempting mail send to " + address_to);

	    if (! resp_line.startsWith ("220"))
	    {
		closeAllWithError("failed at greeting");
		return;
	    }

	    // attempt lookup if 
	    if (helo_namestring.equalsIgnoreCase ("use_lookup"))
	    {
		helo_namestring = InetAddress.getLocalHost().getHostName();
	    }

	    // send helo
	    output = "HELO " + helo_namestring + "\r\n";
	    dos.writeBytes(output);
	    dos.flush();

	    // check for 250
	    resp_line = br.readLine();
	    if (! resp_line.startsWith ("250"))
	    {
		closeAllWithError("failed after HELO");
		return;
	    }

	    // send "MAIL"
	    output = "MAIL FROM: " + address_from + "\r\n";
	    dos.writeBytes (output);
	    dos.flush();

	    // check for 250
	    resp_line = br.readLine();
	    if (! resp_line.startsWith ("250"))
	    {
		closeAllWithError("failed after MAIL");
		return;
	    }

	    // send "RCPT"
	    output = "RCPT TO: " + address_to + "\r\n";
	    dos.writeBytes (output);
	    dos.flush();

	    // check for 250
	    resp_line = br.readLine();
	    if (! resp_line.startsWith ("250"))
	    {
		closeAllWithError("failed after RCPT");
		return;
	    }

	    // send "DATA"
	    output = "DATA\r\n";
	    dos.writeBytes (output);
	    dos.flush();

	    // check for 354
	    resp_line = br.readLine();
	    if (! resp_line.startsWith ("354"))
	    {
		closeAllWithError("failed after DATA");
		return;
	    }

	    // send Headers
	    // send quoted data
	    StringBuffer outputBuf = new StringBuffer();

	    outputBuf.append(header);
	    outputBuf.append(body);

	    if (attach_body != null)
	    {
		// send attachment
		outputBuf.append("\r\n\r\n");
		outputBuf.append(boundary);
		outputBuf.append("\r\n");
		outputBuf.append(attach_header);
		outputBuf.append(filename);
		outputBuf.append("\"\r\n\r\n");
		outputBuf.append(attach_body);
		outputBuf.append(boundary);
		outputBuf.append("--\r\n");
	    }

	    outputBuf.append("\r\n.\r\n");
	    dos.writeBytes (outputBuf.toString());
	    dos.flush();

	    // check for 250
	    resp_line = br.readLine();
	    if (! resp_line.startsWith ("250"))
	    {
		closeAllWithError("failed after body write");
		return;
	    }

	    // send "quit"
	    output = "quit\r\n";
	    dos.writeBytes (output);
	    dos.flush();

	    // check for 221
	    resp_line = br.readLine();
	    if (! resp_line.startsWith ("221"))
	    {
		closeAllWithError("failed after quit");
		return;
	    }
	    closeAll();

	    log (5, "Successful Delivery");
	}
	catch (IOException ioe)
	{
	    closeAllWithError(ioe.toString());
	}
    }

    private String getHeader()
    {
	StringBuffer header = new StringBuffer();
	header.append("To: ");
	header.append(address_to);
	header.append("\r\nFrom: ");
	header.append(address_from);
	header.append("\r\nSubject: ");
	header.append(subject_line);
	header.append("\r\nMime-Version: 1.0\r\n");
	if (attach_body != null) 
	{
	    header.append("Content-Type: multipart/mixed; ");
	    header.append("boundary=\"=====================_881123928==_\"\r\n");
	    header.append(boundary + "\r\n");
	    header.append("Content-Type: text/plain; charset=\"us-ascii\"\r\n");
	}
	header.append("\r\n");
	return header.toString();
    }

    private void closeAllWithError(String s)
    {
	log (0, "ERROR: " + s);
	closeAll();
    }

    private void closeAll ()
    {
	try
	{
	    br.close();
	    dos.close();
	    os.close();
	    is.close();
	    sock.close();
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: close failed");
	}
    }

    private void log (int level, String s)
    {
	if (log != null)
	{
	    log.log(level, "SMTPClient(" + log_id + "): " + s);
	}
	else
	{
	    System.err.println("SMTPClient: " + s);
	}
    }
}
