/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import org.locomotive.loco.*;
import org.locomotive.server.Log;

import java.util.*;
import java.io.*;
import java.net.*;
import java.sql.*;

/**
 * An abstract way to route services to pattern Strings.  This presently
 * loads a file which maps a token or an extension to a class, as well as a
 * path to a properties file.  It will take care of loading and parsing the
 * file, and loading the properties.  Moreover, it'll even reload the
 * file, or an individual service, for you.
 * Here's the file format: <PRE>
 *  &lt;url_pattern&gt;  &lt;classname&gt;  [&lt;path to properties file&gt;]
 *  // comments after the double slash
 * </PRE>
 * To get this to work for you, you'll have to override the following: <UL>
 * <LI>routeRequest() - this actually sends a request to particular service
 * <LI>putInDefaults()- the default services for any given tokens.  
 * <LI>createNewServiceEntry() - this class uses another class, ServiceEntry,
 *     to help it do the work.  You'll have extend that class and override
 *     a couple of its methods.  This method should return your subclass.
 * <LI>ServiceEntry.initialize() - this should initialize a particular service.
 * <LI>ServiceEntry.shutdown() - this should shutdown a particular service.
 * </UL>
 */
public abstract class ServiceRoutingTable
{
    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>AVAILABLE</I></CODE>, that means that this ServiceEntry's
     * service is available to serve requests.
     */
    public static final int AVAILABLE = 0;

    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>UNAVAILABLE_AND_WAITING</I></CODE>, that means that
     * this ServiceEntry's service is currently unavailable, but may
     * become available again after <CODE><I>unavailable_seconds
     * </I></CODE> have elapsed.
     */
    public static final int UNAVAILABLE_AND_WAITING = 1;

    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>PERMANENTLY_UNAVAILABLE</I></CODE>, that means that
     * this ServiceEntry's service is unavailable, and probably won't
     * be available again until the system administrator(s) restart
     * the Locomotive.
     */
    public static final int PERMANENTLY_UNAVAILABLE = 2;

    /**
     * If the instance member <CODE><I>availability</I></CODE> is set to
     * <CODE><I>SHUTDOWN</I></CODE>, that means that the Object factory
     * for this service has been shut down.
     */
    public static final int SHUTDOWN = 255;

    protected Log server_log;

    /** The Hashtable that stores token-to-service mappings */
    protected Hashtable service_tokens_hash;

    /** The Hashtable that stores extension-to-service mappings */
    protected Hashtable service_extensions_hash;

    /** The Hashtable that stores service-to-propsfile mappings */
    protected Hashtable service_configs;

    private String table_file_path;
    private boolean reload = true;
    private String classname;
    private org.apache.java.lang.AdaptiveClassLoader loader;
    private Vector classpath;

    /** this is for whether or not we should reload every time we sense
     * a new class file */
    private boolean auto_reload = true;  

    /**
     * creates a new (but uninitialized) HanderRouting Table
     * @param path the path to the table file
     * @param log  a server log
     * @param classpath a Vector of paths to search for service classes in
     * @throw IllegalArgumentException if the classpath contains things the
     *        class loader doesn't like
     */
    public ServiceRoutingTable (String path, Log log, Vector classpath, 
				boolean auto_reload)
    {
	table_file_path = path;
	server_log = log;
	this.classpath = classpath;
	this.auto_reload = auto_reload;

	classname = getClass().getName();
	int index = classname.lastIndexOf('.');
	if (index > 0)
	{
	    classname = classname.substring(index + 1);
	}
    }
 
    /**
     * initializes or re-initializes this object's routing structures
     * based on the defaults and the contents of the file at the location 
     * specified in 'path' in the Constructor.  Returns true if successful.
     * @exception InitializationException if the file could not be loaded
     *            correctly, or if any of the Services could not be loaded
     *            successfully.
     */
    public void initialize() throws InitializationException
    {
	try
	{
	    loader = new org.apache.java.lang.AdaptiveClassLoader(classpath);
	}
	catch (IllegalArgumentException iae)
	{
	    throw new InitializationException (
		"Could not create class loader: " + iae.toString());
	}

        synchronized (this)
        {
            service_tokens_hash = new Hashtable();
            service_extensions_hash = new Hashtable();
            service_configs = new Hashtable();

            putInDefaults();
        }

	parseTableFile();

	reload = false;
    }



    /**
     * Attempts to shut down each of the services
     */
    public boolean shutdown()
    {
	boolean success = true;
        Enumeration e = null;

	// First, we'll shut down the services mapped to tokens
        if (null != service_tokens_hash)
        {
            e = service_tokens_hash.elements();
            while (e.hasMoreElements())
            {
                ServiceEntry service_entry = (ServiceEntry) e.nextElement();

                // synchronize here so that nothing else can access this
                // service while we're shutting it down
                synchronized (service_entry)
	        {
                    log(9, "Shutting down " + service_entry.theclass);

                    // this'll log the error
                    if (! service_entry.shutdown(server_log))
                    {
                        success = false;
                    }
		
                    service_tokens_hash.remove(service_entry.pattern);
                    service_configs.remove(service_entry.theclass);
                }
            }
        }

	// Next, we'll shut down the services mapped to extensions
        if (null != service_extensions_hash)
        {
            e = service_extensions_hash.elements();
            while (e.hasMoreElements())
            {
                ServiceEntry service_entry = (ServiceEntry) e.nextElement();

                // synchronize here so that nothing else can access this
                // service while we're shutting it down
                synchronized (service_entry)
                {
                    log(9, "Shutting down " + service_entry.theclass);

                    // this'll log the error
                    if (! service_entry.shutdown(server_log))
                    {
                        success = false;
                    }
		
                    service_extensions_hash.remove(service_entry.pattern);
                    service_configs.remove(service_entry.theclass);
                }
            }
        }

	return (success);
    }


    
    /** Override this to return the ServiceEntry class you
     * need.  It should just call the constructor for a 
     * subclass of serviceEntry, then return the newly contructed object.
     * This should never be called in your code- call addService instead,
     * which will call this, then register the newly created service.
     */
    protected abstract ServiceEntry createServiceEntry(String pattern,
						       String classname,
						       String config_path);


    /**
     * Creates and adds a new service to the ServiceRoutingTable.  
     * Note that this will call the createService() method to instantiate
     * a new serviceEntry.  Be sure to override it to create the service
     * entry you need.
     * @param pattern the pattern that the service is mapped to.
     * @param classname the service class's name.
     * @param config_path the path to the objects' properties file.  Leave
     *        this null if the service doesn't have a properties file.
     * @exception InitializationException if the ServiceEntry could not
     *        be loaded successfully.
     */
    public void addService(String pattern, String classname,
			      String config_path) 
	throws InitializationException
    {
	ServiceEntry service_entry = 
	    createServiceEntry(pattern, classname, config_path);
	service_entry.load(server_log);

//	log(9, "Debug: entry loaded. Props: " + service_entry.props);
	// we have the props- now stick 'em in
	if (service_entry.props != null)
	{
	    /* FIXME: The below code makes the (bad) assumption that this
	     *        Locomotive can have exactly one set of configs per
	     *        servlet class!  Servlet container users/admins may
	     *        want to use the same servlet configured multiple
	     *        different ways in the same servlet container instance!
	     *        Here's what the Servlet 2.2 spec says about this, in
	     *        section 3.3.1, "Loading and Instantiation":
	     *
	     *            It is important to note that there can be more
	     *        than one instance of a given Servlet class in the
	     *        servlet container. For example, this can occur where
	     *        there was more than one servlet definition that
             *        utilized a specific servlet class with different
	     *        initialization parameters.
	     */
	    service_configs.put(service_entry.theclass, service_entry.props);
	}

//	log(9, "Debug: Config hash: " + service_configs.toString());


	try { service_entry.initialize(); }
	catch (InitializationException ie) // remove em if we had problems
	{ 
	    service_configs.remove(service_entry.theclass);
	    throw ie;
	}

	// Right here we'll check to see if it's a token mapping
	// (otherwise it's an extension mapping)
	int index = pattern.indexOf('*');
	if (index == -1)
	    service_tokens_hash.put(pattern, service_entry);
	else
	    service_extensions_hash.put(pattern, service_entry);
    }


    /** 
     * Returns true if there is a service associated with this pattern
     */
    public boolean containsService(String pattern)
    {
	return (service_tokens_hash.containsKey(pattern) ||
		service_extensions_hash.containsKey(pattern));
    }


    /** 
     * This is the method which actually routes a request to a particular
     * service.  This must be overridden to each different ServiceRoutingTable.
     * 
     * @param request_id the unique id for this request
     * @param sock the socket this request connected on
     * @param is the inputstream from that socket
     * @param os the outputstream from that socket
     * @param reply_protocol the protocol type we received the request from
     * @param cgi_envars the CGI variables passed in with the request
     * @param form_hash the form data, if there was any
     * @param url_tokens a tokenized form of the url
     * @param request_uri the unparsed url tokens
     * @param bstamp_id the browser stamp id that came in from the
     *        bstamp cookie for this request
     * @param conn the Database connection for this request
     * @param session the Session object for this request
     * @param user the user for this request
     */
    public abstract void routeRequest(int request_id,
				      Socket sock, 
				      InputStream is, 
				      OutputStream os,
				      Hashtable cgi_envars,
				      Hashtable form_hash,
				      Hashtable steam_vars,
				      String [] url_tokens,
				      String request_uri, 
				      long bstamp_id,
				      Response resp,
				      Connection conn, 
				      Session session,
                                      Hashtable attributes);


    /**
     * returns the service class for the specified pattern
     * NOTE: I think this is only used for Handlers!  --JB
     */
    public Class routeToClass(LocoConfig config, String pattern)
    {
	ServiceEntry service_entry = getService(config, pattern);
	if (service_entry == null)
	{
	    return (null);
	}
	else
	{
	    return (service_entry.theclass);
	}
    }


    /**
     * Finds and returns the service Object associated with the specified
     * request uri.  This method first tries to match the first token to
     * a token mapping (like "base"), and then tries to match the request uri
     * to an extension mapping (like "*.jsp").
     */
    public Object routeToService(LocoConfig config, String request_uri,
				 String first_token)
    {
	// Get the ServiceEntry by calling routeToServiceEntry()..
	ServiceEntry service_entry = routeToServiceEntry(config, request_uri,
							 first_token);

	if (service_entry != null) return (service_entry.service);

	// We couldn't find a service for this request.
	return null;
    }

    /**
     * Returns the ServiceEntry Object appropriate for the given config,
     * URI, and first token.  Returns null if a suitable service entry
     * cannot be found.  
     *                                                              <BR><BR>
     * @param config the LocoConfig Object.
     * @param request_uri the URI of the request.
     * @param first_token the first token of the request.
     */
    public ServiceEntry routeToServiceEntry(LocoConfig config,
                                            String request_uri,
                                            String first_token)
    {
	// Make sure that there's no query string on the end of the uri
	int index = request_uri.indexOf('?');
	if (index != -1) request_uri = request_uri.substring(0, index);

	ServiceEntry service_entry = null;

	// We'll try to match the first token to a token mapping
	if (first_token != null && first_token.length() > 0)
	    service_entry = getService(config, first_token);
	if (service_entry != null) return (service_entry);

	// No token mapping match, so see if there's an extension mapping match

	// See if there's an extension mapping that matches this uri
	String test_pattern = "";
	index = request_uri.lastIndexOf('.');
	if (index != -1)
	{
	    test_pattern = "*" + request_uri.substring(index);
	}
	service_entry = getService(config, test_pattern);

	return service_entry;
    }

    /**
     * returns the Service Entry object appropriate for the given config and 
     * pattern string.  Returns null if a suitable service cannot be found.  
     * Note that if a reload has been issued and certain classes could not
     * be found the method may also return null (indicative of error)
     *
     * <P>That it is synchronized should mean that we don't have to worry
     * about synchonizing the methods inside, like initialize.  We may want
     * to rethink this...
     */
    public synchronized ServiceEntry getService (LocoConfig config, 
						    String pattern)
    {
	if ( pattern == null || pattern.length() < 1)
	{
	    pattern = "default";
	}


	/* auto reload code- pretty bad- if it doesn't work, we've got
	 * major problems
	 */
	if (auto_reload && loader.shouldReload())
	{
	    loader.reinstantiate();
	    log(0, "Reloading Service Routing Table classes...");
	    try { initialize(); }
	    catch (InitializationException ie)
	    {
		log(0, "ERROR: getService: could not reload service: " + 
		    ie.toString());

		// now what?  we're screwed.  
		return (null);
	    }

	}

	// Right here we'll check to see if it's a token mapping (otherwise
	// it's an extension mapping)
	boolean isTokenPattern = false;
	int index = pattern.indexOf('*');
	if (index == -1) isTokenPattern = true;

	ServiceEntry service_entry = null;
	if (isTokenPattern)
            service_entry =(ServiceEntry) service_tokens_hash.get(pattern);
	else
	    service_entry =(ServiceEntry) service_extensions_hash.get(pattern);

	// if null (pattern match failure) pass to unknown 
	// if no unknown service, pass back up a null
	if (service_entry == null)
	{
	    if (isTokenPattern) return null;

	    service_entry = (ServiceEntry) service_tokens_hash.get ("unknown");
	    if (null == service_entry)
	    {
		return (null);
	    }
	}

	// we assume at this point we've got a valid service
	if (service_entry.cached)
	{
	    return (service_entry);
	}
	else
	{
	    try
	    {
		reloadService(pattern);
	    }
	    catch (InitializationException ie)
	    {
		log(0, "ERROR: getService: could not reload service: " +
		    ie.toString());
	    }

	    if (isTokenPattern)
		service_entry =(ServiceEntry) service_tokens_hash.get(pattern);
	    else
		service_entry =
		    (ServiceEntry) service_extensions_hash.get(pattern);

	    if (null == service_entry)
	    {
		return (null);
	    }

	    return (service_entry);
	}
    }


    /**
     * Marks a service for reloading, which means it will be reloaded on the
     * next request for it, as opposed to right now, which is what the
     * reloadService() method does.
     */
    public void markServiceForReloading(String pattern)
    {
	ServiceEntry service_entry = null;

	// Right here we'll check to see if it's a token mapping (otherwise
	// it's an extension mapping)
	int index = pattern.indexOf('*');
	if (index == -1) 
	    service_entry =(ServiceEntry) service_tokens_hash.get(pattern);
	else
	    service_entry =(ServiceEntry) service_extensions_hash.get(pattern);
	
	if (service_entry != null)
	{
	    service_entry.cached = false;
	}
    }



    /** 
     * reloads a single service - that is, creates a new Service routing
     * table entry, attempts to reload the properties, and if all
     * goes well, replaces the old service with the new one.
     */
    public void reloadService(String pattern)
	throws InitializationException
    {
	// Right here we'll check to see if it's a token mapping (otherwise
	// it's an extension mapping)
	boolean isTokenPattern = false;
	int index = pattern.indexOf('*');
	if (index == -1) isTokenPattern = true;

	ServiceEntry service_entry = null;
	if (isTokenPattern)
            service_entry =(ServiceEntry) service_tokens_hash.get(pattern);
	else
	    service_entry =(ServiceEntry) service_extensions_hash.get(pattern);

	if (service_entry == null)
	{
	    return;  // no service to reload
	}

	ServiceEntry service_entry_new = createServiceEntry(
	    service_entry.pattern, service_entry.classname,
	    service_entry.config_path);
	service_entry_new.load(server_log);  // this may throw an exception

	// save the old ones in case we have problems
	Properties old_props = 
	    (Properties) service_configs.get(service_entry.theclass);

	synchronized (service_configs)
	{
	    try
	    {
		// put in the new ones
		if (service_entry_new.props != null)
		{
		    service_configs.put(service_entry_new.theclass, 
					service_entry_new.props);
		}
		
		service_entry_new.initialize(); 
	    }
	    catch (InitializationException ie) // make everything better
	    { 
		service_configs.remove(service_entry_new.theclass);
		service_configs.put(service_entry.theclass, old_props);
		throw ie;
	    }

	    synchronized (service_entry)
	    {
		// if we made it here, the new service loaded okay, and we can
		// replace the old one with it.
		
		// at least try to shut the old one down
		service_entry.shutdown(server_log);  
		// get rid of old props
		service_configs.remove(service_entry.theclass); 
		service_entry = service_entry_new;

		if (isTokenPattern)
		    service_tokens_hash.put(pattern, service_entry);
		else
		    service_extensions_hash.put(pattern, service_entry);
	    }
	}
    }



    /**
     * return the properties object for a given Class.  This properties 
     * file will always have the path to the properties file as an attribute,
     * called PROPERTIES_PATH.
     */
    public Properties getProperties(Class theclass)
    {
	return ( (Properties) service_configs.get(theclass));
    }

    /**
     * Override this to put in service defaults.  This class assumes
     * there should be a service installed for the 'unknown' token,
     * and one installed for the 'default' token.
     */
    protected abstract void putInDefaults() throws InitializationException;


    /**
     * loads hooks and class file names from a text file external to the 
     * Loco
     */
    private void parseTableFile() throws InitializationException
    {
	// if our input filepath was none, don't look for external modules
	// in a file
	if (table_file_path.equals ("none"))
	{
	    return;
	}
	
	log (1, "Now loading ServiceRoutingTable file: " + table_file_path);

	// table_file_path should contain full path to config file
	// try and open it.

	FileReader fr = null;
	try
	{
	    fr = new FileReader (table_file_path);
	}
	catch (FileNotFoundException fnfe)
	{
	    throwException("load table: " + fnfe.toString());
	}

	StreamTokenizer st = new StreamTokenizer (fr);
	st.eolIsSignificant (true);
	st.slashSlashComments (true);
	st.wordChars ('_', '_');
	st.wordChars ('*', '*');
	
	boolean got_pattern = false;
	boolean got_classname = false;
	boolean got_config_path = false;
	boolean successful = true;

	Hashtable token_maps = new Hashtable();
	Hashtable extension_maps = new Hashtable();
	Hashtable configs = new Hashtable();

	String pattern = ""; 
	String classname = "";
	String config_path = null;

	int ttype;
	
	try
	{
	    // here's our main loop.  We have a couple of state pointers:
	    // 1. got_pattern for url_token (1st token on line)
	    // 2. got_classname for class (2nd token on line)
	    // 3. got_config_path for path to properties (3rd token on line)
	    // 3 is optional; the rest are required
	    while ((ttype = st.nextToken()) != StreamTokenizer.TT_EOF)
	    {
//              log(9, "line " + st.lineno() + " ttype: " + st.sval);
		if (ttype == StreamTokenizer.TT_EOL)
		{
		    if (got_pattern && !got_classname)
		    {
			got_pattern = false;
			got_config_path = false;
			throwException( "load table: missing classname at " +
					"line: " + st.lineno());
		    }
		    else
		    {
			// get ready for next line
			got_pattern = false;
			got_classname = false;
			got_config_path = false;
		    }
		}
		else if (ttype == StreamTokenizer.TT_WORD || ttype == '\'' || 
			 ttype == '"')
		{
		    if (!got_pattern)
		    {
			pattern = st.sval;
			got_pattern = true;
		    }
		    else if (!got_classname)
		    {
			classname = st.sval;
			got_classname = true;

                        // Right here we'll check to see if it's a token
                        // mapping (otherwise it's an extension mapping)
			int index = pattern.indexOf('*');
                        if (index == -1)
			    token_maps.put(pattern, classname);
			else
			    extension_maps.put(pattern, classname);
		    }
		    else if (!got_config_path)
		    {
			config_path = st.sval;
			got_config_path = true;
			configs.put(pattern, config_path);
		    }
		    else
		    {
			log (2, "ERROR: load table: extra junk ignored, " +
					"line " + st.lineno());
		    }
		}
		else  // StreamTokenizer.TT_NUMBER
		{
		    throwException("load table: unexpected number, line " + 
				   st.lineno());
		}
	    }
	}
	catch (IOException ioe)
	{
	    throwException("load table: " +  ioe.toString());
	}
	finally
	{
	    try { if (fr != null) fr.close(); }
	    catch (IOException e) {}
	}

	//
	// We loaded the file -- now let's load the classes and properties.
	//

	//First, the token mappings...
	Enumeration e = token_maps.keys();
	
	while (e.hasMoreElements())
	{
	    pattern = (String) e.nextElement();
	    classname = (String) token_maps.get (pattern); 
	    config_path = (String) configs.get(pattern);

	    // if we have an error, let's log it, but let's keep going
	    try
	    {
		addService(pattern, classname, config_path);  
	    }
	    catch (InitializationException ie)
	    {
		log(0, "ERROR: could not load service for " + pattern +
		    ": " + ie);
	    }
	}

	//Then, the extension mappings...
	e = extension_maps.keys();
	
	while (e.hasMoreElements())
	{
	    pattern = (String) e.nextElement();
	    classname = (String) extension_maps.get (pattern); 
	    config_path = (String) configs.get(pattern);

	    // if we have an error, let's log it, but let's keep going
	    try
	    {
		addService(pattern, classname, config_path);  
	    }
	    catch (InitializationException ie)
	    {
		log(0, "ERROR: could not load service for " + pattern +
		    ": " + ie);
	    }
	}
    }


    /** Log with the correct class name */
    protected void log(int log_level, String message)
    {
	server_log.log(log_level, classname + ": " + message);
    }

    /** throw an initializationException with the correct classname */
    protected void throwException(String message) 
	throws InitializationException
    {
	throw new InitializationException(
	    classname + ": " + message);
    }


    /** 
     * This little class is the companion to the ServiceRoutingTable.  The
     * ServiceRoutingTable creates these, and keeps a bunch of these around.
     * To use this class, you'll want to override two methods: 
     * <CODE>initialize()</CODE>, and  <CODE>shutdown()</CODE>, which will
     * do whatever's appropriate to do at the beginning of and at the end of
     * your service's life.  If your service requires a persistent object,
     * then create it in the intialize() method and store it in the service
     * field.
     */
    public abstract class ServiceEntry
    {
	protected boolean cached = false;
	public String pattern = null; 
	protected String classname = null;
	protected Class theclass = null;
	protected String config_path = null;
	protected Properties props = null;
	protected Object service = null;

	/**
	 * This ServiceEntry's service may not be available to serve
	 * requests at certain times.  It starts off being available,
	 * and it may later be marked as either temporarily or permanently
	 * unavailable.
	 * @see org.locomotive.loco.servlet.ServletFactory
	 */
	protected int availability = AVAILABLE;

	/**
	 * We'll keep track of any waiting periods so that we can tell
	 * users what to expect.  This unavailability timeout is only
	 * valid when availability is set to UNAVAILABLE_AND_WAITING.
	 */
	protected int unavailable_seconds = 0;

	/**
	 * If the service is either temporarily or permanently unavailable,
	 * we should have a reason for the unavailability that we can
	 * show to the client.
	 */
	protected String unavailability_reason = null;

	/**
	 * The date/timestamp of when the last availability reporting event
	 * occurred.  It is stored as the number of milliseconds elapsed
	 * between Jan 1, 1970 00:00:00 GMT and the actual time this event
	 * occurred.
	 */
	protected long last_availability_event_timestamp = 0L;

	/** 
	 * Creates a new ServiceEntry.  
	 * @param pat the pattern to map the service to
	 * @param cname the service's class name
	 * @param c_path the path to the service's properties file
	 */
	protected ServiceEntry(String pat, String cname, String c_path)
	{
	    pattern = pat;
	    classname = cname;
	    config_path = c_path;
	}

	/** 
	 * This will load the config path and create the class object. 
	 * The initialize() method below will still need to be called to 
	 * acually initialize the object.
	 * @exception InitializationException if the class couldn't be found
	 *            or loaded, or if the properties couldn't be loaded.
	 */
	protected synchronized void load(Log server_log)
            throws InitializationException
	{
	    FileInputStream fis = null;
	    String errmsg = "classload failed for " + pattern + ": ";

	    try
	    {
		// Alright- here we use our classloader to load our
		// lovely class.  Let's see how it works...
		theclass = loader.loadClass(classname);
	    
		//log(9, "Debug: Attempting to load class: " + theclass);

		if (config_path != null)
		{
		    props = new Properties();
		    fis = new FileInputStream(config_path);
		    props.load(fis);
		    log(9, "Loading properties for " +
			classname + " at " + config_path);

		    // now, as a bonus, let's add the path to the prop
		    // so our friendly developers can change the prop
		    // and write the changes to file...
		    props.put("PROPERTIES_PATH", config_path);
		}
	    }
	    catch (IOException e)
	    {
		props = null;
		throwException(errmsg + "could not load properties " +
			       "at " + config_path + ": " + e.toString());
	    }
	    catch (ClassNotFoundException cnfe)
	    {
		theclass = null;
		throwException(cnfe.toString());
	    }
	    catch (LinkageError le)  
	    {
		// will catch subclasess ExceptionInInitializerError and
		// NoClassDefFoundError too
		theclass = null;
		throwException (le.toString());
	    }
	    finally
	    {
		try
		{
		    if (fis != null) fis.close();
		}
		catch (IOException e)
                {
                }
	    }
	}

	/** This needs to be overridden to instantiate and create the
	 * service.  You should catch all exceptions and rethrow them
	 * as InitializationExceptions.  You Should never need to call
	 * this- load(), above, will call it, after the properties file
	 * has been loaded.
	 */
	protected abstract void initialize() throws InitializationException;

	/**
	 * This should also be overridden to call whatever cleanup/shutdown
	 * methods you need.  Again, exceptions should be caught and logged.
	 */
	public abstract boolean shutdown(Log server_log);

	/**
	 * Gets the number of seconds that this ServiceEntry's service is
	 * unavailable for.
	 */
	public int getUnavailableSeconds()
	{
	    return unavailable_seconds;
	}

	/**
	 * Sets the number of seconds that this ServiceEntry's service is
	 * unavailable for.
	 *                                                           <BR><BR>
	 * @param unavailable_seconds the amount of time during which
	 *        this ServiceEntry's servlets are unavailable.
	 */
	public void setUnavailableSeconds(int unavailable_seconds)
	{
	    this.unavailable_seconds = unavailable_seconds;
	}

	/**
	 * Returns the availability status of this ServiceEntry's service.
	 */
	public int getAvailability()
	{
	    return availability;
	}

	/**
	 * Sets the availability code for this ServiceEntry's service.
	 *                                                           <BR><BR>
	 * @param availability The new availability state, should be one
	 *        of the following: <CODE><I>ServiceEntry.AVAILABLE,
	 *        ServiceEntry.UNAVAILABLE_AND_WAITING,
	 *        ServiceEntry.PERMANENTLY_UNAVAILABLE.           </I></CODE>
	 */
	public synchronized void setAvailability(int availability)
	{
	    // Set the new availability value
	    this.availability = availability;
	}

	/**
	 * Returns the reason for the unavailability.
	 */
	public String getUnavailabilityReason()
	{
	    return unavailability_reason;
	}

	/**
	 * Sets the reason for the unavailability.
	 *                                                           <BR><BR>
	 * @param reason the text reason to use.
	 */
	public void setUnavailabilityReason(String reason)
	{
	    unavailability_reason = reason;
	}

	/**
	 * Gets the timestamp of the last availability event.  It is stored
	 * as the number of milliseconds elapsed between Jan 1, 1970
	 * 00:00:00 GMT and the actual time this event occurred.
	 */
	public long getLastAvailabilityEventTimestamp()
	{
	    return last_availability_event_timestamp;
	}

	/**
	 * Sets the timestamp of the last availability event.  It is stored
	 * as the number of milliseconds elapsed between Jan 1, 1970
	 * 00:00:00 GMT and the actual time this event occurred.
	 *                                                           <BR><BR>
	 * @param last_availability_event_timestamp the date/timestamp
	 *        when the ServiceRoutingTable last received a report
	 *        about service availability.
	 */
	public void setLastAvailabilityEventTimestamp(
            long last_availability_event_timestamp)
	{
	    this.last_availability_event_timestamp =
		last_availability_event_timestamp;
	}

	/**
	 * Regardless of what the service Object is, use this method to
	 * get a reference to it.
	 */
	public Object getService()
	{
	    return service;
	}
    }
}

