/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.server;

import java.util.*;
import org.locomotive.directory.LocomotiveService;
import org.locomotive.directory.ServiceException;

public class ThreadPoolManager implements LocomotiveService
{
    private Log log;
    private int min_threads = 0;
    private int max_threads = 0;
    private int current_thread_count = 0;
    private int in_use_count = 0; 
    private Stack pooled_thread_stack;
    private boolean isRunning = false;

    public ThreadPoolManager (Log lg, int min, int max)
    {
	log = lg;
	min_threads = min;
	max_threads = max;
	init();
    }
    

    /**
     * Default constructor for the thread pool manager.
     */
    public ThreadPoolManager(){
    }
  
    private void init(){
	pooled_thread_stack = new Stack ();
	
	// push the initial threads onto the stack
	for (int i = 0; i < min_threads; i++)
	{
	    PooledThread pt = new PooledThread (this, log, 
						current_thread_count);
	    Thread t = new Thread (pt);
	    t.start ();
	    pooled_thread_stack.push (pt);
	    current_thread_count++;
	}
    }

    /**
     * Start the thread pool manager.
     * @exception ServiceException is thrown if the manager could not start.
     */
    public void startService() throws ServiceException {
	init();
	isRunning = true;
    }
    /**
     * Stop the thread pool manager.
     * @exception ServiceException is thrown if the manager could not stop.
     */
    public void stopService() throws ServiceException {
	isRunning = false;
    }
    
    /**
     * Check if the service running.
     * @return Returns true if the service is running, 
     *	       false otherwise.
     */
    public boolean isRunning(){
	return isRunning;
    }

    public synchronized PooledThread getPooledThread ()
    {
	boolean sent_warning = false;

	while (pooled_thread_stack.empty ())
	{
	    // if we can add a thread do so, else sleep
	    if (current_thread_count < max_threads)
	    {
		PooledThread pt = new PooledThread (this, log, 
						    current_thread_count);
		Thread t = new Thread (pt);
		t.start ();
		current_thread_count++;
		in_use_count++;
		return (pt);
	    }
	    else
	    {

		// Log the fact that we're stuck here (good output for admins
		// so they know that the pool should probably be made larger)!
		if (null != log)
		{
		    log.log(Log.WARNING,
		        "ThreadPoolManager: all Threads in use!  Waiting..");
		    sent_warning = true;
		}

		try
		{
		    wait();
		}
		catch (InterruptedException ie) {}
	    }
	}

	if (sent_warning && null != log)
	{
	    // Log the fact that we're done waiting (this could have been a
	    // while)
	    log.log(Log.WARNING, "ThreadPoolManager: Done waiting.");
	}

	in_use_count++;
	return ((PooledThread) pooled_thread_stack.pop ());
    }

    public synchronized void returnToPool (PooledThread pt)
    {
	pooled_thread_stack.push (pt);
	in_use_count--;
	notify();
    }

    public int getActiveCount ()
    {
	return (in_use_count);
    }
    
    public int getPoolSize ()
    {
	return (current_thread_count);
    }

    public int getMaxCount()
    {
	return(max_threads);
    }

    public int getMinCount()
    {
	return(min_threads);
    }

    /**
     * Set the maximum number of threads this pool can have.
     * @param maxThreads Maximum number of threads for this pool.
     */
    public void setMaxThreads(int maxThreads){
	this.max_threads = maxThreads;
    }

    /**
     * Get the maximum number of threads for this pool.
     * @return Returns the maximum number of threads for this pool.
     */
    public int getMaxThreads(){
	return max_threads;
    }

    /**
     * Set the minimum number of threads this pool can have.
     * @param minThreads Minimum number of threads for this pool.
     */
    public void setMinThreads(int minThreads){
	this.min_threads = minThreads;
    }

    /**
     * Get the minimum number of threads for this pool.
     * @return Returns the minimum number of threads for this pool.
     */
    public int getMinThreads(){
	return min_threads;
    }
    
    /**
     * Set the server log for this pool.
     * @param log the log
     */
    public void setLog(Log log){
	this.log = log;
    }

}
