/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 2000 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.servlets;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;

import javax.servlet.*;
import javax.servlet.http.*;

/**
 * The MultiThreadedHttpServlet class is a Servlet class you can use as
 * a superclass for multithreaded servlet-container-independant servlets
 * you're writing.  All of the functionality it provides should work on
 * any Servlet 2.1-compliant container.  This class and your own subclass
 * should be all that you need.
 *                                                                   <BR><BR>
 * Notice that this class does not implement SingleThreadModel, which means
 * that subclasses of this class will be treated as multithreaded servlets.
 * Only one instance of each kind of multithreaded servlet will be
 * instantiated in the servlet container, and that one instance will serve
 * any/all requests simultaneously.
 *
 * @author Jason Brittain <brittainj@diamtech.com>
 */
public class MultiThreadedHttpServlet extends HttpServlet
{
    /**
     * The service method is the main entry point for Servlets.  All this
     * method does is call the setup() method to get everything ready for
     * the request, and then call super.service() to send the request
     * off to the appropriate do() method (i.e. doGet(), doPost(), etc).
     * <P>If you choose to override this method in a sub class, don't forget
     * to call the setup() method at the beginning of your version.
     * @see #setup
     * @see javax.servlet.http.HttpServlet.html#service
     */ 
    public void service(HttpServletRequest request,
			HttpServletResponse response)
	throws IOException, ServletException
    {
	response.setContentType("text/html");

	super.service(request, response);
    }

    /**
     * This is a convenience method you can use when you catch
     * serious Exceptions for which you want a stack trace and 
     * you need to send the user an error code, which can help you debug.
     *                                                               <BR><BR>
     * @param exception the exception to print to the server log.
     * @param short_code the error message to send to the user
     * @param request the HttpServletRequest for this Thread
     * @param response the HttpServletResponse for this Thread
     */
    public void handleException(Exception exception, String short_code,
				HttpServletRequest request,
				HttpServletResponse response)
    {
	log(getClass().getName() + ": ERROR: " + short_code + "." +
	    exceptionStackTrace(exception));

	try
	{
	    response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
			       short_code);
	}
	catch (IOException io_exception)
	{
	    // We caught an IOException, so we better log it.
	    log(getClass().getName() + ": ERROR: caught Exception while " +
		"sending response." + exceptionStackTrace(io_exception));    
	}
    }

    /**
     * If you want to send back a single string as the response to the user,
     * you can use this method.  (You don't need to put your message within
     * &lt;HTML&gt;&lt;BODY&gt;...&lt;/BODY&gt;&lt;/HTML&gt;)
     *                                                               <BR><BR>
     * @param message the message you'd like to display
     * @param request the HttpServletRequest for this Thread
     * @param response the HttpServletResponse for this Thread
     */
    public void displayString(String message, HttpServletRequest request,
                              HttpServletResponse response) throws IOException
    {
	if (message == null || message.equals(""))
	{
	    try
	    {
		response.sendError(
		    HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
		    "no message!");
	    }
	    catch (IOException exception)
	    {
		// We caught an IOException, so we better log it.
		log(getClass().getName() + ": ERROR: caught Exception while " +
		    "sending response." + exceptionStackTrace(exception));    
	    }
	    return;
	}
	
	PrintWriter print_writer = response.getWriter();
	if (message.indexOf("<HTML>") != -1)
	{
	    print_writer.print(message);
	    print_writer.flush();
	}
	else
	{
	    print_writer.print("<HTML><BODY>" + message + "</BODY></HTML>");
	    print_writer.flush();
	}
    }

    /**
     * Simply tells the user that they typed a bad URL, by sending the message:
     * "I'm sorry, I don't understand that URL."
     *                                                              <BR><BR>
     * @param response the HttpServletResponse for this Thread
     */
    public void sayBadURL(HttpServletResponse response) throws IOException
    {
	try
	{
	    response.sendError(response.SC_NOT_FOUND, 
			       "I'm sorry, I don't understand that URL.\n");
	}
	catch (IOException exception)
	{
	    // We caught an IOException, so we better log it.
	    log(getClass().getName() + ": ERROR: caught Exception while " +
		"sending response." + exceptionStackTrace(exception));    
	}
    }

    /**
     * A utility function - given the exception, give us a string that
     * contains the stack trace for it.  This method was copied verbatim
     * from org.locomotive.util.StringUtil.java for inclusion in this
     * self-contained container-portable class.
     */
    public static String exceptionStackTrace(Throwable e)
    {
	StringWriter sw = new StringWriter();
	e.printStackTrace(new PrintWriter(sw));
	return("\n" + sw.toString());
    }
}
