/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.steam;

import java.util.*;
import java.lang.reflect.Array;

import org.locomotive.util.SortableTable;
import org.locomotive.util.GrowableChar;


/** 
 * This class handles arrays. 
 * It will accept either array names or array elements,  which are specifed
 * as the array name plus an index in parenthesis Example- somename(5).
 * If the array expression is contained within a table expression then 
 * the array is evaluated as a column in the table.
 * If only the array name is entered in the constructor, the when the 
 * ArrayExpr is evaluated the first element of the array is returned.
 * Note that indexes start at zero, just like in normal arrays, so the 
 * first element has an index of zero. Note also that this will evaluate 
 * non-array objects exactly the same as VariableExpr does.  It'll just 
 * take longer.
 */
class ArrayExpr extends MixedExpr
{
    protected int index = 0;            // index of the array to return
    protected String var_name = null;

    // holds whether this object refers to the array as a whole, or 
    //only an element  in the array
    protected boolean isElement = false;

    protected SortableTable table = null;        // stays null if array
    protected TemplateCacheEntry parent = null;


    void printTree (String s)
    {
	System.out.println(s + "ARRAY: " + var_name + "\tIndex: " + index);
	for (int i=0; i < expr.length; i++)
	{
	    expr[i].printTree ("\t" + s + "(" + i +  "): ");
	}
    }

    /** presently deals with sortable tables and arrays.  Should implement
     * vectors too, and JDK 1.2 containers too. */
    public void evaluate (Hashtable ht, GrowableChar growchar)
    {
	// if we have no array name, evaluate like a normal mixed expr
	if (!evaluateArgs(ht))
	{
	    super.evaluate(ht, growchar);
	}

//	    Object temp = ht.get("loop_index");
//	    if (temp != null)
//		System.out.println("loop_index: " + temp.toString());
//	    System.out.println(var_name + " table: " + table);
//	    printTree( var_name + " ");

	if (table != null) // if it got set in evaluateArray
	{
	    evaluateTableColumn(ht, growchar);
	}
	else
	{
	    evaluateArray(ht, growchar);
	}

	table = null;
	isElement = false;
	index = 0;
	var_name = null;
    }

    public boolean notEmpty(Hashtable ht)
    {
	boolean b = false;

	if (! evaluateArgs(ht))
	    b = false;
	else
	    b = doNotEmpty(ht);

	table = null;
	isElement = false;
	index = 0;
	var_name = null;
	return b;
    }


    private boolean doNotEmpty (Hashtable ht)
    {
//	    org.locomotive.loco.Loco.log(0, "notEmpty-ARRAY: " + var_name + 
//				 "\tIndex: " + index);

	// if table is set, then we're a column in a table
	if (table != null)  
	{
	    Object o = table.get(index, var_name);
	    return (o != null && (o.toString().length() > 0) );
	}

	Object array = ht.get(var_name);

	if (array == null) 
	    return false;

	if (array.getClass().isArray())
	{
	    if (isElement || index > 0)  // we're  an element
	    {
		return ((Array.get(array, index) != null) &&
			!(Array.get(array, index).toString().equals("")));
	    }
	    else
	    {
		// loop through the array
		for (int i = 0; i < Array.getLength(array); i++)
		{
		    if ((Array.get(array, i) != null) &&
			!(Array.get(array, i).toString().equals("")))
			return true;
		}
		return false;
	    }
	}
	else  // we have a normal variable
	    return(array.toString().length() > 0);
    }


    /** looks for a column in a SortableTable, and reads the row number
     * of that column given by the index.  Index is set by TableExpr. */
    private void evaluateTableColumn(Hashtable ht, GrowableChar growchar)
    {
	if (index < table.getNumRows() )
	{
	    Object to_return = table.get(index, var_name);
	    if (to_return != null)
	    {
		growchar.add(to_return.toString());
	    }
	}
//	org.locomotive.loco.Loco.log(0, "evaluate-ARRAY-table: " + var_name + 
//			     "\tIndex: " + index + "\tisElement" + isElement);
    }


    /** looks in the hashtable for an array, and reads the indexth element.
     * Index is set by LoopExpr, or by simply evaluating the template. */
    private void evaluateArray(Hashtable ht, GrowableChar growchar)
    {
	// System.out.println("ARRAY: " + var_name + "\tIndex: " + index);
	Object array = ht.get(var_name);

	if (array == null || array.toString().length() <= 0)
	    return;

	if (array.getClass().isArray())
	{
	    array = ht.get(var_name);

	    if (index < Array.getLength(array) )
	    {
		Object to_return = Array.get(array, index);
		if (to_return != null)
		    growchar.add(to_return.toString());
	    }
	}
	else  // this variable is a normal variable.  Just enter it
	{
	    growchar.add (array.toString());
	}
//	org.locomotive.loco.Loco.log(0, "evaluate-ARRAY: " + var_name + 
//			     "\tIndex: " + index + "\tisElement " + isElement);
    }
    

    /** Get the name out of the 1st expression and set it.
     * Get the index out of the 2nd expression and set it.
     * Get the right TableExpr out of the parent table_stack and
     * set tableExpr to it.  Set the index to that table's index
     * @return false if the name cannot be evaluated. true if otherwise
     */
    private boolean evaluateArgs(Hashtable ht)
    {
	// get the name
	if (expr != null && expr.length > 0 && expr[0] != null 
	    && expr[0].notEmpty(ht) ) 
	{
	    GrowableChar g = new GrowableChar(40);
	    expr[0].evaluate(ht, g);
	    var_name = g.toString();
//		System.out.println("array name: " + var_name);
	}
	else  // no array name- we're done
	{
	    var_name = null;
	    isElement = false;
	    return false;
	}

	// if our 2nd expression is null or empty, we have no index
	if ( expr.length > 1 && expr[1] != null &&  expr[1].notEmpty(ht) )
	{
	    GrowableChar index_str = new GrowableChar(32);
	    expr[1].evaluate(ht, index_str);
	    
	    try 
	    {
		index = Integer.parseInt(index_str.toString());
		isElement = true;
	    }
	    catch (NumberFormatException e)
	    {
//		org.locomotive.loco.Loco.log(0, "NumberFormatException: " + index_str);
		isElement = false;
	    }  // just use the default index
	}
	else
	{
	    isElement = false;
	}

	// finally, check to see if the var_name is the column of any
	// of the table in our parent's table_stack
	org.locomotive.util.data.Stack s = new org.locomotive.util.data.Stack();
	while (! parent.table_stack.isEmpty())
	{
	    TableExpr temp = (TableExpr) parent.table_stack.pop();
	    s.push(temp);
//		System.out.println("table_stack\t" + parent.table_stack);

	    if (temp.table != null && 
		temp.table.getColumnIndex(var_name) != -1)
	    {  // we have a table for this column
		table = temp.table;
		if (! isElement)
		    index = temp.index;
		break;
	    }
	}

	while (! s.isEmpty())
	{
	    parent.table_stack.push(s.pop());
	}

	// the table is still null- set the index to loop_index
	if (table == null && (! isElement) && ht.containsKey("loop_index"))
	{
	    index = ((Number) ht.get("loop_index")).intValue();
	}

	return true;
    }

}

