/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.steam;

import java.io.*;
import java.util.*;

import org.locomotive.util.GrowableChar;
import org.locomotive.server.Log;

/**
 * The CachedPageLoader evalutes templates from the filesystem or 
 * passed in as a string according to the rules of the LOCO
 * Template Language.  Templates are compiled the first time they are 
 * evaluated, then the compiled version is cached and used for subsequent 
 * re-evaluation.  See the documentation for the template language for 
 * the details of its usage.
 */
public class CachedPageLoader
{
    static int max_template_size;
    static int check_template_freq;
    static String template_root;
    
    private static Log server_log;
    private static int logging_id = 0;
    
    private static Hashtable Template_Cache = new Hashtable();

    /*
     * initializes the CachedPageLoader class.
     * @param templ_root   The template root.  Any load statements will load 
     *                 and include new templates relative to this path
     * @param check_freq  how many times to SKIP checking whether
     *              templates have changed on disk.  (For speed.)
     * @param initial_bufsize  The initial size of the template output buffer.
     *                   If too small, the buffer will have to be resized
     *                   (causing extra memory allocation).  
     */
    public static void initialize (String templ_root, int check_freq,
				   int initial_bufsize)
    {
	template_root = templ_root;
	check_template_freq = (check_freq > 1) ? check_freq : 1;
	max_template_size = (initial_bufsize > 0) ? initial_bufsize : 1;
    }
    
    
    /**
     * Contructs a CachedPageLoader with the specified debugging 
     * log/identifier for internal logging.  
     * @param slog  a destination log. May be null, in which case the 
     *               CPL will not log 
     * @param id    a unique id to be used in the output log.  ignored if slog
     *              is null;
     */
    public CachedPageLoader (Log slog, int id)
    {
	server_log = slog;
	logging_id = id;
    }


    /**
     * Evaluate the given filename, relative to the given template root 
     * with the given substitution environment.  The enviroment is read 
     * write, and can contain name/values set during the evaluation process
     * @param t_path   the path to the template, relative to the template_root
     *                 passed in via initialize()
     * @param subs     The environment (in/out)
     */
    public String evalTemplate (String t_path, Hashtable subs)
    {
	// in case people want to use "true" or "yes" as constants:
	if (!subs.containsKey("true")) subs.put("true", "true");
	if (!subs.containsKey("TRUE")) subs.put("TRUE", "TRUE");
	if (!subs.containsKey("True")) subs.put("True", "True");
	if (!subs.containsKey("yes")) subs.put("yes", "yes");
	if (!subs.containsKey("YES")) subs.put("YES", "YES");
	if (!subs.containsKey("Yes")) subs.put("Yes", "Yes");

	MixedExpr top_expr = null;

	if (t_path.startsWith("/"))
	{
	    top_expr = cachedTemplateExpr (template_root + t_path);
	}
	else
	{
	    top_expr = cachedTemplateExpr (template_root + "/" + t_path);
	}

	GrowableChar growchar = new GrowableChar (max_template_size);
// This would help debug:
//	top_expr.printTree("");
	top_expr.evaluate (subs, growchar);
	
	return (growchar.toString());
    }


    /**
     * Evaluate the following string.  Does NOT insert it into the cache.
     * @param string_to_eval  The string to evaluate.
     * @param subs            The environment (in/out)
     * @param allow_load_cmds Whether or not to allow LOAD commands to load
     *                        in templates from files (using the template_root
     *                        passed in via initialize())
     */
    public String evalString (String string_to_eval, Hashtable subs,
			      boolean allow_load_cmds)
    {
	// in case people want to use "true" or "yes" as constants:
	if (!subs.containsKey("true")) subs.put("true", "true");
	if (!subs.containsKey("TRUE")) subs.put("TRUE", "TRUE");
	if (!subs.containsKey("True")) subs.put("True", "True");
	if (!subs.containsKey("yes")) subs.put("yes", "yes");
	if (!subs.containsKey("YES")) subs.put("YES", "YES");
	if (!subs.containsKey("Yes")) subs.put("Yes", "Yes");

	TemplateCacheEntry entry = new TemplateCacheEntry(string_to_eval, 
							  allow_load_cmds);
	GrowableChar growchar = new GrowableChar (max_template_size);
//	entry.top_expr.printTree("");
	entry.top_expr.evaluate (subs, growchar);
	
	return (growchar.toString());
    }
    
    
    /**
     * This function is necessary in order to do [LOAD ...] operations.
     * LoadExpr's don't know what the template root is, so they have to get
     * it externally from somewhere.
     */
    static String templateRoot ()
    {
	return template_root;
    }
    
    /**
     * This function is also necessary in order to do [LOAD ...] operations.
     */
    static int checkTemplateFreq ()
    {
	return check_template_freq;
    }

    
    /**
     * XXXXXXXX TAKE THIS OUT LATER.  ONLY FOR TESTING. 
     * Note: made not public by Thede for 3.3. Simply change to public 
     * and recompile to test from command line...
     *
    public static void main(String argv[])
    {
	initialize ("/home/ck/src/templates", 1, 20000);
	Hashtable ht = new Hashtable();

	CachedPageLoader cpl = new CachedPageLoader(null, 1);
	if (argv != null && argv.length > 1 && argv[0] != null)
	{
	    if (argv[0].equals("file"))
		System.out.println (cpl.evalTemplate (argv[1], ht));
	    
	    else if (argv[0].equals("string"))
		System.out.println (cpl.evalString (argv[1], ht, true));
	}

    }
    */    
    
    /*
      private static void print (String s)
      {
      System.out.println(s);
      }
      */
    
    /*----------------------------------------------------------------------
     *
     * The following are functions having to do with the cache for templates:
     *
     *----------------------------------------------------------------------
     */
    
    /**
     * For testing only.  This returns a string saying whether a file is
     * in the cache, and, if so, when it was read in.
     */
    static String wasInCache (String filename)
    {
	if (Template_Cache.containsKey (filename))
	{
	    TemplateCacheEntry entry = 
		(TemplateCacheEntry) Template_Cache.get (filename);
	    return (filename + " " + entry.count + " " + entry.modified_date);
	}
	else
	    return "   No mod date.";
    }
    

    /** 
     * This method will get a template expression tree either from
     * the cache or from a file.  If it is not in the cache, it will parse
     * the file and store the new template expression tree into the cache.
     * If it is in the cache, then it will nevertheless check whether the
     * template file has changed.  However, for efficiency purposes, 
     * it will not check every time;  it will do so only every N times,
     * where N = "check_template_freq".
     * <p>If the file does not exist, or there were other problems, then it
     * will return a tree with a single empty node which evaluates to "".
     */
    protected static synchronized MixedExpr cachedTemplateExpr (String filename)
    {
	return (cachedTemplateExpr(filename, true));
    }

    /** 
     * This method will get a template expression tree either from
     * the cache or from a file.  If it is not in the cache, it will parse
     * the file and store the new template expression tree into the cache.
     * If it is in the cache, then it will nevertheless check whether the
     * template file has changed.  However, for efficiency purposes, 
     * it will not check every time;  it will do so only every N times,
     * where N = "check_template_freq".
     * <p>If the file does not exist, or there were other problems, then it
     * will return a tree with a single empty node which evaluates to "".
     * @param evaluate whether or not to evaluate the file and parse it into
     * a template tree 
     */
    protected static synchronized MixedExpr cachedTemplateExpr (String filename,
							      boolean evaluate)
    {
	TemplateCacheEntry entry =
	    (TemplateCacheEntry) Template_Cache.get (filename);
	
	File template_file = new File (filename);

	// if the file does not exist, then return an empty expr.
	if (! template_file.exists())
	{   
	    return new MixedExpr(); 
	}

	long modified_date = template_file.lastModified();

	if (entry == null)
	{
	    log (10, filename + " not found in cache."); // debug
	    entry = new TemplateCacheEntry (filename, modified_date, evaluate);
	    Template_Cache.put (filename, entry);
	}
	else if (entry.evaluated != evaluate) 
	{
	    // if our cached version is evaluated and our new version shouldn't
	    // be, or vice versa, we have to reload
	    log (9, filename + " must be reloaded"); // debug
	    entry = new TemplateCacheEntry (filename, modified_date, evaluate);
	    Template_Cache.remove (filename);
	    Template_Cache.put (filename, entry);
	}
	else if (++entry.count >= check_template_freq) // increment entry count
	{
	    entry.count = 0;
	    if (entry.modified_date < modified_date)
	    {
		// debug
		log (10, filename + " is outdated -- reloading...");
		entry = 
		    new TemplateCacheEntry (filename, modified_date, evaluate);
		Template_Cache.remove (filename);
		Template_Cache.put (filename, entry);
	    }
	}
	
	return (entry.top_expr);
    }
    

    /**
     * This static method can be called to flush the template cache.
     * All currently cached compiled templates will be discarded. 
     * When a subsequent request for template evaluation is made, the 
     * corresponding template is reloaded, compiled, and evaluated, then the 
     * compiled version is again stored in the template cache.
     */
    public static synchronized void clearTemplateCache ()
    {
	Template_Cache.clear();
    }

    /**
     * returns a char buffer with file contents given a relative path.
     * returns empty char array on error.
     */
    static char[] loadfile (String full_path)
    {
	byte buf[];
	FileInputStream template_fis;
	int rawsize = 0;

	log (5, "loading template: " + full_path);

	// open the file
	try
	{
	    template_fis = new FileInputStream(full_path);
	}
	catch (FileNotFoundException fnfe)
	{
//	    print("file: " + full_path + " not found.");
	    log (0, "ERROR: loadfile: open:" + fnfe.toString());
	    return (new char[0] );
	}

	// discover file size
	try 
	{
	    rawsize = template_fis.available();
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: loadfile(): getsize: " + ioe.toString());
	    try 
	    {
		template_fis.close();
	    }
	    catch (IOException otherioe)
	    {
		log (0, "ERROR: exception closing!");
	    }
	    return (new char[0]);
	}

	log (8, "file size is: " + rawsize);

	// attempt to fully load the file into buffer
	buf = new byte[rawsize];
	int totalread = 0;
	try 
	{
	    int numread = 0;
	    while ((numread = template_fis.read (buf, totalread, 
						 rawsize - totalread)) > 1)
	    {
//		log (8, "read " + numread + " from file");
		totalread += numread;
	    }
	}
	catch (IOException ioe)
	{
	    log (0, "ERROR: loadfile: slurpfile: " + ioe.toString());

	    try
	    {
		template_fis.close();
	    }
	    catch (IOException otherioe)
	    {
		log (0, "ERROR: exception closing!");
	    }
	    return (new char[0]);
	}

	// early exit if empty
	if (totalread == 0)
	{
	    // empty file 
	    try 
	    {
		template_fis.close();
	    }
	    catch (IOException otherioe)
	    {
		log (0, "ERROR: exception closing!");
	    }
	    return (new char[0]);
	}
	
	// copy the contents
	char template_chars[] = new char[totalread];
	for (int i = 0; i < totalread; i++)
	{
	    template_chars[i] = (char) (buf[i] & 0xFF);
	}

	// close the template file
	try 
	{
	    template_fis.close();
	}
	catch (IOException otherioe)
	{
	    log (0, "ERROR: exception closing!");
	}

	return (template_chars);
    }

    
    /**
     * logs specific to the CPL
     */
    private static void log (int level, String s)
    {
	if (server_log != null)
	{
	    server_log.log (level, "CPL(" + logging_id + ") " + s);
	}
    }

}

