/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developer of this code under the LPL is Leverage Information 
 * Systems. Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */ 

package org.locomotive.steam;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.locomotive.server.InitializationException;


/**
 * SteamServlet can be used to interpret and display steam templates.
 * The first time templates are read, they are automatically compiled
 * can cached, for high performance.  They are reloaded whenever the
 * template file modification date changes.  This class uses any
 * extra path information it finds tacked on the the URL to figure
 * out which template to display.  If it doesn't find a file type suffix
 * in the URL, it will automatically add &quot;.tmpl&quot; to the
 * end of the URL.  For instance: <BR>
 * <PRE> 
 * disp/base/main.tmpl     maps to $template.root/base/main.tmpl
 * disp/base/main          maps to $template.root/base/main
 * disp/base/general/page1 maps to $template.root/base/general/page1.tmpl
 * </PRE> 
 * The $template.root variable is one of the properties that is loaded
 * during servlet initialization.  To set it, put the following line
 * In your Servlet config file: 
 * <PRE><CODE>
 * template.root = /your/template/root/path
 * </PRE></CODE>
 * SteamServlet will not start up without this line in your configuration
 * file.  Consult your Servlet engine documentation on how to add 
 * configurations for Servlets.
 */
public class SteamServlet extends HttpServlet
{

    CachedPageLoader cpl;
    String template_root;


    /**
     * forwards the request to toGet()
     * @see #doGet
     */
    public void doPost(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	doGet(req, resp);
    }
	

    /**
     * Send the requested page to the client, using the rules specified in the
     * class documentation above.
     */
    public void doGet(HttpServletRequest req, HttpServletResponse resp)
	throws ServletException, IOException
    {
	Hashtable steam_vars = 
	    (Hashtable) req.getAttribute("locomotive.steam_vars");

	if (steam_vars == null) steam_vars = new Hashtable();
	
	String template_path = shellEscape(req.getPathInfo());

        // check that the url_tokens are not null or empty, and have
        // only shell safe characters
        if (template_path != null && template_path.length() > 0)
	{
	    if (template_path.lastIndexOf('.') < 0)
	    {
		template_path += ".tmpl";
	    }
		
	    String result = cpl.evalTemplate(template_path, steam_vars);
	    // if there is no file in that file path, then
	    // displayPage() will return an empty string, "", in which
	    // case we should tell the user that the URL was faulty.
	    if (result.equals(""))
	    {
		resp.sendError(resp.SC_NOT_FOUND, "ERROR: " +
			       template_path + " Not found on this server.");
		log("template not available: " + template_root + template_path);
	    }
	    else
	    {
		PrintWriter out = resp.getWriter();
		out.print(result);
		out.flush();
	    }
		
	}
	else
	{
	    resp.sendError(resp.SC_NOT_FOUND, "ERROR: " +
			   template_path + " Not found on this server.");
	}
    }


    /**
     * This Initializes the Cached Page Loader.  This takes a single input
     * parameter: <CODE>template.root</CODE> which signifies the document
     * root to load pages from. If this servlet is run inside
     * the Locomotive, make sure you set the template root to the
     * same place as the loco template root, or else all the other
     * handlers will not operate correctly.  
     */
    public void init(ServletConfig config) throws ServletException
    {
	super.init(config);
	template_root = config.getInitParameter("template.root");
	if (template_root == null)
	    throw new ServletException("You must specify template.root in " +
				       "the DisplayPageServlet config " +
				       "properties file");

	CachedPageLoader.initialize(template_root, 1,  16384);
	cpl = new CachedPageLoader(null, 1);

    }
	
    /**
     * Shell escapes a given string.  Currently remove the following: 
     * <BLOCKQUOTE><B>|&nbsp;&nbsp;<&nbsp;&nbsp;>&nbsp;&nbsp;!
     * &nbsp;&nbsp;$&nbsp;&nbsp;,&nbsp;&nbsp;\</B></BLOCKQUOTE>
     */
    public static String shellEscape(String s)
    {
	if (s == null) return ("");

	String badchars = "|<>!$,\\";
	StringBuffer sb = new StringBuffer();
	char [] input = s.toCharArray();
	for (int i = 0; i < input.length; i++)
	{
	    if (badchars.indexOf(input[i]) < 1)
	    {
		sb.append(input[i]);
	    }
	}
	
	return (sb.toString());
    }

}
