/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.steam;

import java.util.*;
import java.io.*;

import org.locomotive.util.GrowableChar;
import org.locomotive.util.data.Stack;


/**
 * TemplateCacheEntry represents one entry in the template cache table.
 * It knows how to parse a file into a MixedExpr tree with subcomponents.
 */
class TemplateCacheEntry
{
    public String filename;
    public long modified_date;
    public MixedExpr top_expr;
    public int count = 0;
    public boolean allow_loads = true;

    protected char input[];
    protected int input_cursor = 0;
    protected boolean exp_ended = false;
    protected boolean evaluated = true;

    // supports nested table exprs inside the template
    protected org.locomotive.util.data.Stack table_stack = null;  
    

    public TemplateCacheEntry (String the_filename)
    {
	File file = new File (the_filename);
	modified_date = file.lastModified();
	count = 0;
	table_stack = new org.locomotive.util.data.Stack();

	input = CachedPageLoader.loadfile (the_filename);
	top_expr = new MixedExpr();
	parseMixedExpr (top_expr, true, false);
    }

    
    public TemplateCacheEntry (String the_filename, long the_modified_date)
    {
	this(the_filename, the_modified_date, true);
    }

    
    /** This version allows you to choose whether you want the file
     * evaluated.  If evaluate is false, the the top_expr is a TextExpr
     * with the entire file as its contents 
     */
    public TemplateCacheEntry (String the_filename, long the_modified_date,
			       boolean evaluate)
    {
	evaluated = evaluate;
	filename = the_filename;
	modified_date = the_modified_date;
	count = 0;
	table_stack = new org.locomotive.util.data.Stack();
	    
	input = CachedPageLoader.loadfile (filename);

	if (evaluate)
	{
	    top_expr = new MixedExpr();
	    parseMixedExpr (top_expr, true, false);
	}
	else
	{
	    top_expr = new TextExpr(new String(input));
	}
    }
		


    /**
     * This version will parse a string, instead of a file.
     * It ignores [LOAD ...] commands, because you might want only the
     * string itself to be evaluated.  If you also want to allow LOADs,
     * then use the other version with strings.
     *
     public TemplateCacheEntry (String string_to_parse)
     {
     allow_loads = false;
     input = string_to_parse.toCharArray();
     top_expr = new MixedExpr();
     parseMixedExpr (top_expr, true, false);
     }
     * Not currently used
     */
    
    /**
     * This version will parse a string, instead of a file.
     * It will optionally allow [LOAD ...] commands, using the template root
     * passed into CachedPageLoader.initialize().
     * @param string_to_parse  The string to parse into an expression tree.
     * @param allow_load_cmds  Whether to allow load commands.
     */
    public TemplateCacheEntry (String string_to_parse, 
			       boolean allow_load_cmds)
    {
	table_stack = new org.locomotive.util.data.Stack();
	allow_loads = allow_load_cmds;
	input = string_to_parse.toCharArray();
	top_expr = new MixedExpr();
	parseMixedExpr (top_expr, true, false);
    }
    
    

    private void parseMixedExpr (MixedExpr mixed_expr,
				 boolean is_top_expr,
				 boolean is_quote_expr)
    {
	boolean done = false;
	
	if (input.length == 0)
	{
	    return;
	}
	
	GrowableChar growchar = new GrowableChar (1024);
	
	char c;

	while (!done && (input_cursor < input.length))
	{
	    c = input[input_cursor];
	    if (c == '\\')
	    {
		// quote next char if there is one
		if (input.length > (input_cursor + 1))
		{
		    if (growchar.cursorpos == growchar.buf.length)
		    {
			growchar.resize();
		    }
		    growchar.buf[growchar.cursorpos++] = 
			input[input_cursor + 1];
		    input_cursor = input_cursor + 2;
		}
		else
		{
		    input_cursor++;
		}
	    }

	    // if we see one of these tokens, then we know we've seen the
	    // end of the previous expression.  We should do something
	    // about it.
	    // NOTE: if we see a lone unbalanced ']' and we're at the
	    // top level, we should just ignore it, else we'll think
	    // that we're finished.
	    
	    else if ((c == ']' && (!is_top_expr)) ||
		     (c == '[') ||
		     (!is_top_expr && !is_quote_expr && 
		      Character.isWhitespace(c)))
	    {
		input_cursor++;
		if (growchar.length() > 0)
		{
		    String s = growchar.toString();

		    // At the top level or inside a quote, 
		    // everything is text.
		    if (is_top_expr || is_quote_expr)
		    {
			mixed_expr.insert (new TextExpr (s));
		    }
		    else
		    {
			// but if we're inside a [...] tag, then everything
			// that is not inside [quote ...] is treated like a
			// variable or a numeric constant.
			try {
			    Integer.parseInt(s);
			    mixed_expr.insert (new TextExpr (s));
			}
			catch (NumberFormatException ex) {
			    mixed_expr.insert (new VariableExpr(s));
			}
		    }
		}
		growchar.clear();
		
		if (c == ']' && (!is_top_expr))
		{
		    done = true;
		}
		else if (c == '[')
		{
		    mixed_expr.insert (parseSpecialExpr ());
		}
		else if (!is_top_expr && !is_quote_expr &&
			 Character.isWhitespace(c))
		{
		    skipWhitespace();
		}
		
	    }
	    else
	    {
		// add the character to the output
		if (growchar.cursorpos == growchar.buf.length)
		{
		    growchar.resize();
		}
		growchar.buf[growchar.cursorpos++] = c;
		input_cursor++;
	    }
	}
	    
	// Finally, append the last remnants, if there were any:
	if (growchar.length() > 0)
	    mixed_expr.insert (new TextExpr (growchar.toString()));
	    
	return;
    }

    private void skipWhitespace()
    {
	while ((input_cursor < input.length)
	       && (Character.isWhitespace(input[input_cursor])))
	{
	    input_cursor++;
	}
    }

    private String getToken ()
    {
	boolean done = false;
	StringBuffer sb = new StringBuffer();
	char c;

	skipWhitespace();
	while (!done && (input_cursor < input.length))
	{
	    c = input[input_cursor];

	    if (c == '\\')
	    {
		// quoted character
		if (input_cursor < (input.length - 1))
		{
		    input_cursor++;
		    sb.append (input[input_cursor]);
		}
		else
		{
		    input_cursor++;
		    done = true;
		}
	    }
	    else if (c == ']')
	    {
//		input_cursor++;
//		exp_ended = true;
		done = true;
	    }
	    else if (Character.isWhitespace(c))
	    {
		input_cursor++;
		done = true;
	    }
	    else if (c == '[')
	    {
		done = true;
	    }
	    else 
	    {
		input_cursor++;
		sb.append (c);
	    }
	}
//	log (8, "getToken returning: " + sb.toString());

	return (sb.toString());
    }

    private MixedExpr parseSpecialExpr ()
    {
	MixedExpr expr;
	    
	// ignore whitespace

	String token = getToken ();
	if (token.length() == 0)
	{
	    expr = new TextExpr("");
	    ignoreUntilClosed();
	    return (expr);
	}
	
	String lowercase_token = token.toLowerCase();
	
	if (lowercase_token.equals("if"))
	{
	    expr = new IfExpr();
	}
	else if (lowercase_token.equals("load") && allow_loads)
	{
	    expr = new LoadExpr ();

	    // we MUST create a TextExpr next, because if we don't, then
	    // the filename will be interpreted as a VariableExpr instead.
	    expr.insert(new TextExpr (getToken()));

	    // we'll try to get the 'noeval' too, if it exists...
	    expr.insert(new TextExpr (getToken())); 
	    ignoreUntilClosed();
	    return (expr);
	}
	else if (lowercase_token.equals("not"))
	{
	    expr = new NotExpr ();
	}
	else if (lowercase_token.equals("or"))
	{
	    expr = new OrExpr ();
	}
	else if (lowercase_token.equals("and"))
	{
	    expr = new AndExpr ();
	}
	else if (lowercase_token.equals("eq") || 
		 lowercase_token.equals("="))
	{
	    expr = new EqualExpr();
	}
	else if (lowercase_token.equals(">"))
	{
	    expr = new GtrExpr();
	}
	else if (lowercase_token.equals(">=") || 
		 lowercase_token.equals("=>"))
	{
	    expr = new GtrEqExpr();
	}
	else if (lowercase_token.equals("<"))
	{
	    expr = new LessExpr();
	}
	else if (lowercase_token.equals("<=") || 
		 lowercase_token.equals("=<"))
	{
	    expr = new LessEqExpr();
	}
	else if (lowercase_token.equals("!=") || 
		 lowercase_token.equals("neq") ||
		 lowercase_token.equals("ne"))
	{
	    expr = new NotEqExpr();
	}
	else if (lowercase_token.equals("+"))
	{
	    expr = new AddNumbersExpr();
	}
	else if (lowercase_token.equals("-"))
	{
	    expr = new SubtractNumbersExpr();
	}
	else if (lowercase_token.equals("*"))
	{
	    expr = new MultiplyNumbersExpr();
	}
	else if (lowercase_token.equals("/"))
	{
	    expr = new DivideNumbersExpr();
	}
	else if (lowercase_token.equals("set"))
	{
	    // !!! known bug/feature: maybe [set d] should unset d,
	    // just like [set d []], but instead, it slurps in the
	    // entire rest of the template!  This is because getToken()
	    // eats the closing ']'.
	    
	    expr = new SetExpr();
	    
	    // the reason we MUST create a TextExpr next is because we're
	    // setting the NAME of the variable.  [set A B] does NOT mean
	    // "set (what A evaluates to) to B", but rather,
	    // "set A itself to B".  Otherwise, A would be parsed as a
	    // VariableExpr instead.  SetExpr requires A to be a TextExpr
	    // in order to be able to set it.
	    expr.insert(new TextExpr (getToken()));
	    
	    skipWhitespace();
		
	    // then, treat the rest of the line as if it were 
	    // [quote ...]'ed.
	    parseMixedExpr (expr, false, true);
	    return (expr);
	}
//	else if (lowercase_token.equals("var"))
//	{
//	    token = getToken();
//	    expr = new VariableExpr (token);
//	    ignoreUntilClosed();
//	    return (expr);
//	}
	else if (lowercase_token.equals("colname") || 
		 lowercase_token.equals("array"))
	{
	    expr = new ArrayExpr();
	    ((ArrayExpr) expr).parent = this;
	    expr.insert(new TextExpr (getToken()));
	    parseMixedExpr (expr, false, true);
	    return (expr);
	}
	else if (lowercase_token.equals("quote"))
	{
	    expr = new MixedExpr();
	    parseMixedExpr (expr, false, true);
	    return (expr);
	}
	else if (lowercase_token.equals("loop"))
	{
	    expr = new LoopExpr();
	    parseMixedExpr (expr, false, true);
	    return (expr);
	}
	else if (lowercase_token.equals("table"))
	{
	    expr = new TableExpr();
	    ((TableExpr) expr).parent = this;
	    expr.insert(new TextExpr (getToken()));  // get the name
	    parseMixedExpr (expr, false, true);
	    return (expr);
	}
	else  // treat the token as if it equaled "var"
	{
	    expr = new VariableExpr (token);
	    ignoreUntilClosed();
	    return (expr);
	}

//	log (8, "End parseSpecialExpr: " + expr.toString());
	parseMixedExpr (expr, false, false);
	
	return (expr);
    }

    
    private void ignoreUntilClosed()
    {
	boolean done = false;
	char c;
	int dp = 0;

	if (exp_ended)
	{
	    exp_ended = false;
	    return;
	}

	while ((!done) && (input_cursor < input.length))
	{
	    c = input[input_cursor];
	    
	    if (c == ']')
	    {
		dp--;
	    }
	    else if (c == '[')
	    {
		dp++;
	    }
	    else if (c == '\\')
	    {
		input_cursor++;
	    }

	    if (dp == -1)
	    {
		done = true;
	    }
	    input_cursor++;
	}
    }   
}



