/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.*;

/**
 * Utility routine for manipulating blocks of Base64 data
 */
public class Base64
{
    protected static byte base64lookup[];
    private static boolean base64lookup_initialized = false;
    protected static byte base64reverse_lookup[];
    private static boolean base64reverse_lookup_initialized = false;

    /**
     * like it says
     */
    private static void initialize_base64_lookup()
    {
	base64lookup = new byte[64];

	for (int i = 0; i < 26; i++)
	{
	    base64lookup[i] = (byte) (65 + i);
	}
	for (int i = 26; i < 52; i++)
	{
	    base64lookup[i] = (byte) (97 + (i - 26));
	}
	for (int i = 52; i < 62; i++)
	{
	    base64lookup[i] = (byte) (48 + (i - 52));
	}

	// plus '+'
	base64lookup[62] = 43;

	// forward-slash '/'
	base64lookup[63] = 47;

	base64lookup_initialized = true;
    }

    /**
     * like it says
     */
    private static void initialize_base64reverse_lookup ()
    {
	base64reverse_lookup = new byte[128];

	// set everything to -1
	for (int i = 0; i < 128; i++)
	{
	    base64reverse_lookup[i] = (byte) -1;
	}

	// stuff values
	for (int i = 0; i < 26; i++)
	{
	    base64reverse_lookup[65 + i] = (byte) i;
	}
	for (int i = 26; i < 52; i++)
	{
	    base64reverse_lookup[97 + (i - 26)] = (byte) i;
	}
	for (int i = 52; i < 62; i++)
	{
	    base64reverse_lookup[48 + (i - 52)] = (byte) i;
	}
	
	// do '+'
	base64reverse_lookup[43] = 62;
	
	// do '/'
	base64reverse_lookup[47] = 63;
	base64reverse_lookup_initialized = true;
    }

    /**
     * convert a byte array to a base64 byte array
     * output is 4/3 size of input, plus extra 4 bytes 66% of the time
     * @param input  an array of binary byte data
     */ 
    public static byte[] base64Encode (byte input[])
    {
	if (! base64lookup_initialized)
	    initialize_base64_lookup();

	int input_cursor = 0;
	int output_cursor = 0;

	// get correctly sized output
	int outputsize = (input.length / 3) * 4 
	    + (((input.length % 3) != 0) ? 4 : 0);

	byte output[] = new byte[outputsize];

	int block;

	while ((input.length - input_cursor) > 2)
	{
	    // get front six of first byte
	    block = 0x00000000 | (byteToInt (input[input_cursor]) >>> 2);
	    output[output_cursor++] = base64lookup[block & 0x3F];
	    
	    // get tail two plus front four
	    block = 0x00000000 | (byteToInt (input[input_cursor++]) << 4) | 
		(byteToInt (input[input_cursor]) >>> 4);
	    output[output_cursor++] = base64lookup[block & 0x3F];
	    
	    // get back four plus front two
	    block = 0x00000000 | (byteToInt (input[input_cursor++]) << 2) | 
		(byteToInt (input[input_cursor]) >>> 6);
	    output[output_cursor++] = base64lookup[block & 0x3F];
	    
	    // get tail six
	    block = 0x00000000 | byteToInt (input[input_cursor++]);
	    output[output_cursor++] = base64lookup[block & 0x3F];
	}
	
	// do tail processing 
	if ((input.length - input_cursor) == 0)
	{
	    return (output);
	}
	else if ((input.length - input_cursor) == 1)
	{
	    // get front six of first byte
	    block = 0x00000000 | (byteToInt (input[input_cursor]) >>> 2);
	    output[output_cursor++] = base64lookup[block & 0x3F];

	    // get tail two 
	    block = 0x00000000 | (byteToInt (input[input_cursor++]) << 4);
	    output[output_cursor++] = base64lookup[block & 0x3F];

	    output[output_cursor++] = 61;
	    output[output_cursor++] = 61;
	    return (output);
	}
	else // must be 2 remaining
	{
	    // get front six of first byte
	    block = 0x00000000 | (byteToInt (input[input_cursor]) >>> 2);
	    output[output_cursor++] = base64lookup[block & 0x3F];
	    
	    // get tail two plus front four
	    block = 0x00000000 | (byteToInt (input[input_cursor++]) << 4) | 
		(byteToInt (input[input_cursor]) >>> 4);
	    output[output_cursor++] = base64lookup[block & 0x3F];
	    
	    // get back four
	    block = 0x00000000 | (byteToInt (input[input_cursor++]) << 2);
	    output[output_cursor++] = base64lookup[block & 0x3F];

	    output[output_cursor++] = 61;   
	    return (output);
	}
    }

    /**
     * decodes input base64, returns binary byte array.  May throw 
     * IllegalArgumentException is the input is not valid Base64-text
     * @param input  a byte array containing Base64-text
     */
    public static byte[] base64Decode (byte input[])
    {
	if ((input.length == 0) || (input == null))
	{
	    return new byte[0];
	}
	
	// mimecode must be divisible by 4
	if ((input.length % 4) != 0)
	{
	    throw new IllegalArgumentException();
	}

	// allocate output byte array
	byte output[];
	if (input[input.length - 1] != '=')
	{
	    output = new byte[input.length / 4 * 3];
	}
	else if (input[input.length - 2] == '=')
	{
	    output = new byte[input.length / 4 * 3 - 2];
	}
	else
	{
	    output = new byte[input.length / 4 * 3 - 1];
	}

	// process 4 input bytes = 3 output bytes at a time
	int output_cursor = 0;
	int input_cursor = 0;
	for (input_cursor = 0; (input.length - input_cursor) > 4; 
	     input_cursor = input_cursor + 4)
	{
	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor]) << 2) | 
		(mimeByteToVal(input[input_cursor + 1]) >> 4));
		
	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor + 1]) << 4) |
		(mimeByteToVal(input[input_cursor + 2]) >> 2));

	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor + 2]) << 6) |
		mimeByteToVal(input[input_cursor + 3]));
	}

	// handle the padded '=' endings
	if ((output.length % 3) == 0)
	{
	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor]) << 2) |
		(mimeByteToVal(input[input_cursor + 1]) >> 4));
		
	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor + 1]) << 4) |
		(mimeByteToVal(input[input_cursor + 2]) >> 2));

	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor + 2]) << 6) |
		mimeByteToVal(input[input_cursor + 3]));
	}
	else if ((output.length % 3) == 1)
	{
	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor]) << 2) |
		(mimeByteToVal(input[input_cursor + 1]) >> 4));
	}
	else 
	{
	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor]) << 2) |
		(mimeByteToVal(input[input_cursor + 1]) >> 4));
		
	    output[output_cursor++] = (byte)(
		(mimeByteToVal(input[input_cursor + 1]) << 4) |
		(mimeByteToVal(input[input_cursor + 2]) >> 2));
	}
	
	return (output);
    }

    private static byte mimeByteToVal (byte b)
    {
	if (b < 0)
	    throw new IllegalArgumentException();

	if (! base64reverse_lookup_initialized)
	    initialize_base64reverse_lookup();
	
	if (base64reverse_lookup[b] == -1)
	{
	    throw new IllegalArgumentException();
	}

	return (base64reverse_lookup[b]);
    }

    /**
     * converts a byte to an it, as if the byte were "unsigned"
     */
    static int byteToInt (byte b)
    {
	if (b >= 0)
	{
	    return ( (int) b);
	}
	return ((int) b + 256);
    }
}

