/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.StringTokenizer;

/**
 * This class contains some useful utilities, mostly for dealing with 
 * credits cards, and some for dealing with dollar currency formats.
 */
public class Commerce
{
    /**
     * Decide if a given credit card number is in good form or not using Luhn
     * Checksum algorithm.
     *
     * Number the digits from right to left, starting at 0. At odd locations
     * perform i = (2*i)%9. Sum all the digits. sum%10 must be 0.
     *
     */
    public static boolean isValidCreditCard(long number)
    {
	if (number <= 0) return(false);

	String snum = (new Long(number)).toString();

	int len = snum.length();

	char schars[] = new char[len];

	snum.getChars(0, len, schars, 0);

	int sum = 0;

	// Start at the end of the array
	int index = len - 1;

	while (index >= 0)
	{
	    // len-index-1 is the numbering, right to left,
	    // starting at zero for the rightmost, of the position
	    // "index" in the zero-indexed array of length len. If
	    // that numbering value & 1 == 1 then we are at an odd location.
	    // Otherwise it's an even location.
	    if (((len - index - 1) & 1) == 1)
	    {
		// If that value was odd, distort the schars value at that
		// location according to the Luhn algorithm.
		schars[index] = Character.forDigit(
		    ((Character.digit(schars[index], 10) * 2) % 9),
		    10);
	    }

	    // Whether it was an even (undistorted) or odd (distorted in the
	    // above if clause) sum it up into our running tally.
	    sum += Character.digit(schars[index], 10);

	    index--;
	}

	// Now that the sum has been computed, tell us if it mod 10 is 0.
	return((sum % 10) == 0);
    }

    /**
     * Checks basic syntax stuff for a creit card string. Takes out -'s,
     * spaces, and any other non-digit stuff. Does not check max or min
     * size of the number value (though negative values are not possible since
     * -'s are removed...). Returns -1 if the number specified is
     * invalid.
     */
    public static long parseCCNum(String cc_str)
    {
	// I know, I know. It's there because of some database-related
	// reason. Probably should return Long and null if it failed...
	long failure = -1;

	if (cc_str == null) return(failure);
	if (cc_str.equals("")) return(failure);

	StringBuffer cc_strb = new StringBuffer();
	int index = 0;
	int max = cc_str.length();
	boolean keepgoin;

	while (index < max)
	{
	    keepgoin = true;

	    // Skip over garbage, take out only digits.
	    while (keepgoin)
	    {
		if (index < max)
		{
		    if (!Character.isDigit(cc_str.charAt(index)))
		    {
			index++;
		    }
		    else keepgoin = false;
		}
		else keepgoin = false;
	    }

	    // Get the digit, put it in our buffer (cc_strb)
	    if (index < max)
		cc_strb.append(cc_str.charAt(index));

	    index++;
	}

	long result;

	// Translate the number now in cc_strb
	try { result = Long.parseLong(cc_strb.toString()); }
	catch (NumberFormatException nfe) { return (failure); }

	// Catch negative values (shouldn't be possible since we took
	// out dashes, but what the hey...
	if (result < 0) return(failure);

	return(result);
    }

    /**
     * Given the expiration string of a credit card number, try to
     * figure out the month. I made a small improvement to this and
     * have tested it some. It should work. (Casey)
     */
    public static String parseCCExpMonth(String cc_exp_str)
    {
	if (cc_exp_str == null) return(null);

	StringTokenizer tok = new StringTokenizer(cc_exp_str, "/");
	String result = null;

	// result is now null

	if (tok.hasMoreTokens())
	{
	    result = tok.nextToken();
	}

	// If no more tokens (i.e. bad syntax) result is still null

	try
	{
	    // Parsing a null string throws NumberFormatException
	    int value = Integer.parseInt(result);

	    // Make sure it's a valid month
	    if ((value < 1) || (value > 12)) result = null;
	    else result = ("" + value);

	    if (result != null) // If we did get something
	    {
		// Insert 0 place holder for small values
		if (value < 10)
		    result = "0" + result;
	    }
	}
	catch (NumberFormatException nfe) { result = null; }

	return(result);
    }

    /**
     * Given the expiration string of a credit card number, try to
     * figure out the year. I made a small improvement to this and
     * have tested it some. It should work. (Casey)
     */
    public static String parseCCExpYear(String cc_exp_str)
    {
	if (cc_exp_str == null) return(null);

	StringTokenizer tok = new StringTokenizer(cc_exp_str, "/");
	String result = null;

	if (tok.hasMoreTokens())
	{
	    tok.nextToken();

	    if (tok.hasMoreTokens())
	    {
		result = tok.nextToken();
	    }
	}

	try
	{
	    int value = Integer.parseInt(result);

	    if ((value < 0) || (value > 99)) result = null;
	    else result = ("" + value);

	    if (result != null)
	    {
		if (value < 10)
		    result = "0" + result;
	    }
	}
	catch (NumberFormatException nfe) { result = null; }

	return(result);
    }

    /**
     * Given a price string, make it presentable by adding .00 if it's a
     * straight integer or adding 0 if it's a 1.6 form. Does NOT add a $.
     * Probably should in the future. It's not there because purple moon
     * is using insertion points and adding their own $ to strings that
     * this is returning. Ah well. I could make some adapter code but what
     * the hell. You'll live.
     */
    public static String cleanUpPrice(String price)
    {
	if (price == null) return("");
	if (price.equals("")) return("");

	StringTokenizer tok = new StringTokenizer(price, ".");

	if (tok.countTokens() != 2)
	    return(price + ".00");
	else
	{
	    tok.nextToken();
	    if (tok.nextToken().length() == 1)
		return(price + "0");
	}

	return(price);
    }
    
    
    /**
     * Pass in an int representing the number of cents (== the number
     * of dollars * 100), and you'll get back a String representing it in
     * money format.  ex: if num_cents is 534, then returns "5.34".
     * if num_cents is negative, will take its absolute value.
     */
    public static String printPriceFromCentValue (int num_cents)
    {
	String sign = "";
	if (num_cents < 0)
	{
	    sign = "-";
	    num_cents = -num_cents;
	}
	return sign + (num_cents/100) + "." +
	    ((num_cents % 100 < 10) ? "0" : "") + (num_cents % 100);
    }
    
    public static int getValueInCentsFromPriceStr (String price)
	throws NumberFormatException
    {
	if (price.indexOf('.') != price.length() - 3)
	    throw new NumberFormatException();
	
	return (Integer.parseInt(
	    price.substring(0, price.length() - 3) + 
	    price.substring(price.length() - 2)));
    }
}
