/**
 * DoubleHashtable.java - a Hashtable of Hashtables, where you
 * use the first key to find a Hashtable, and then use the second key
 * to search within that Hashtable to find your item.
 *
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.*;

/**
 * DoubleHashtable is a hashtable of hashtables, 
 * where you use the first key to find a hashtable, and then 
 * use the second key to search within that hashtable to find your item.
 *
 * You can clear entire Hashtables at a time.  Neat.
 * It even does periodic cleanup so that unused rows or columns will not
 * stick around and take up space as empty Hashtables.
 */
public class DoubleHashtable
{
    Hashtable master_hashtable = null;
    Hashtable empty_hashtable = null;
    
    // every 500 times we remove an item, do a periodic check to remove
    // rows or columns that are not being used, to be more efficient.
    public static final int CLEANUP_INTERVAL = 500;
    int periodic_cleanup = 0;
    
    // let's just say that the average number of columns for any row is about
    // 5, and the average number of rows for any columns is about the same.
    public static final int AVERAGE_ENTRIES_PER_ROW = 5;

    public boolean conserve_memory = false;
    
    /**
     * If you want to conserve memory usage, and you know that your
     * child Hashtables are temporary, then you should set
     * conserve_memory to true in order to make this class throw away
     * empty child Hashtables once in a while.
     */
    public DoubleHashtable (boolean conserve_memory)
    {
        master_hashtable = new Hashtable ();
        empty_hashtable = new Hashtable ();
	this.conserve_memory = conserve_memory;
    }
    
    /**
     * Default constructor sets up the DoubleHashtable with the 
     * conserve memory property set to false.
     */
    public DoubleHashtable(){
	this(false);
    }

    
    
    public synchronized void put (Object child_tables_key, Object items_key,
				  Object data_item)
    {
        get_or_create_child_table(child_tables_key).put(items_key, data_item);
    }
    
    public Object get (Object child_tables_key, Object items_key)
    {
        return (get_table_at (child_tables_key).get (items_key));
    }

    
    public Hashtable getChildTable (Object child_tables_key)
    {
	return (get_or_create_child_table(child_tables_key));
    }
    
    
    public Enumeration getAllChildTableKeys ()
    {
        return (master_hashtable.keys());
    }
    
    /**
     * This is a VERY expensive operation -- it traverses the entire table.
     */
    public synchronized Enumeration getAllElements ()
    {
        // estimate about 5 entries per key, on average?
        Vector v = new Vector (master_hashtable.size() * 5);
        
        Enumeration e = master_hashtable.elements ();
        while (e.hasMoreElements())
        {
            Enumeration e2 = ((Hashtable) e.nextElement()).elements ();
            while (e2.hasMoreElements())
            {
                v.addElement (e2.nextElement());
            }
        }
        
	return v.elements();
    }
    
    
    public Enumeration getKeysOfChildTable (Object child_tables_key)
    {
        return (get_table_at (child_tables_key).keys ());
    }

    public Enumeration getElementsOfChildTable (Object child_tables_key)
    {
        return (get_table_at (child_tables_key).elements ());
    }
    
    /**
     * traverses the ENTIRE Hashtable of Hashtables -- a very expensive
     * operation!  Don't do this if you can possibly avoid it!
     */
    public synchronized boolean contains (Object data_item)
    {
        Enumeration e = master_hashtable.elements ();
        
        while (e.hasMoreElements())
        {
            if (((Hashtable) e.nextElement()).contains (data_item))
                return (true);
        }
        
        return (false);
    }
    
    
    public boolean contains (Object child_tables_key, Object items_key)
    {
        return (get_table_at (child_tables_key).containsKey (items_key));
    }
    
    public boolean containsChildTableKey (Object child_tables_key)
    {
        return (! (get_table_at (child_tables_key).isEmpty()));
    }
    
    
    public Object remove (Object child_tables_key, Object items_key)
    {
        if (conserve_memory && periodic_cleanup++ > CLEANUP_INTERVAL)
            removeEmptyChildTables();

        return (get_table_at (child_tables_key).remove (items_key));
    }
    
    public synchronized Hashtable removeChildTable (Object child_tables_key)
    {
	return ((Hashtable) master_hashtable.remove(child_tables_key));
    }

    public synchronized void clear ()
    {
        master_hashtable.clear();
    }

    
    /**
     * Removes child Hashtables which are empty.  Of course, this
     * doesn't mean that they won't be used again, but if they are,
     * then they can simply be created again by
     * get_or_create_child_table(), so there's no danger of losing
     * tables.  Anybody can call this method at any time, if you need
     * to conserve memory. 
     */
    synchronized void removeEmptyChildTables()
    {
        Enumeration e = master_hashtable.keys();
        
        while (e.hasMoreElements())
        {
	    Object child_tables_key = e.nextElement();
            Hashtable child_table =
		(Hashtable) master_hashtable.get(child_tables_key);
            if (child_table.isEmpty())
            {
                master_hashtable.remove(child_tables_key);
            }
        }
    }
    
    
    /**
     * Returns the Hashtable at "key", either by row or column.  If
     * you never created a Hashtable at "key", you should get an empty
     * Hashtable back as a result.
     * @param child_tables_key - the key which locates that child Hashtable.
     */
    Hashtable get_table_at (Object child_tables_key)
    {
	Hashtable child_table =
	    (Hashtable) master_hashtable.get(child_tables_key);
        if (child_table == null)
        {
	    return (empty_hashtable);
	}
	else
	{
            return (child_table);
        }
    }


    /**
     * Creates a child Hashtable at the location "child_tables_key", in the
     * master_hashtable.
     * If there is already a Hashtable at the location "child_tables_key", then
     * will simply return that Hashtable.
     *
     * We need to synchronize parts of this method to be thread-safe, but 
     * we're not going to synchronize this entire method, because that would
     * penalize the lookup time every time.  Instead, we're going to 
     * synchronize ONLY when the location "child_tables_key" is new.
     */
    Hashtable get_or_create_child_table (Object child_tables_key)
    {
        Hashtable child_table =
	    (Hashtable) master_hashtable.get (child_tables_key);
	
	if (child_table == null)
        {
            synchronized (master_hashtable) {
                
                // here, we're repeating
		// "if (master_hashtable.containsKey (child_tables_key))"
                // because if two threads are here at the same time, one
                // of them will create the table first, and then the other
                // one executes afterwards.
                // it's worth the penalty of doing this if-statement twice
                // because, after a while, all the tables will have been
                // created, and this will no longer be a penalty.
                
		child_table =
		    (Hashtable) master_hashtable.get (child_tables_key);
		
		if (child_table == null)
		{
		    // create a Hashtable about twice as big as the average
		    // (expected) number of entries per row or column.
		    // Hashtable sizes are about 50 + 4*N.
		    child_table = new Hashtable (AVERAGE_ENTRIES_PER_ROW * 2);
		    master_hashtable.put (child_tables_key, child_table);
                }
            }
        }
	
	return (child_table);
    }

    public static void main (String argv[])
    {
        DoubleHashtable tab = new DoubleHashtable(false);
        tab.put ("Fred", "1", "Fred1");
        System.out.println ("Row: " + tab.master_hashtable.toString());
        
        tab.put ("Fred", "1", "Fred1");
        System.out.println ("Row: " + tab.master_hashtable.toString());
        
        tab.put ("Fred", "2", "Fred2");
        System.out.println ("(Fred, 2) = " + tab.get("Fred", "2"));
        System.out.println ("Row: " + tab.master_hashtable.toString());
        
        tab.put ("Wilma", "11", "Wilma11");
        System.out.println ("Row: " + tab.master_hashtable.toString());
        
        tab.put ("Barney", "21", "Barney21");
        System.out.println ("Row: " + tab.master_hashtable.toString());
        
        tab.remove ("Barney", "2");
        System.out.println ("Row: " + tab.master_hashtable.toString());
        
        tab.remove ("Barney", "21");
        System.out.println ("Row: " + tab.master_hashtable.toString());
        
        tab.removeChildTable ("Fred");
        System.out.println ("Row: " + tab.master_hashtable.toString());

        tab.removeEmptyChildTables();
        System.out.println ("Row: " + tab.master_hashtable.toString());

        tab.removeChildTable ("Fred");
        System.out.println ("Row: " + tab.master_hashtable.toString());

        tab.put ("Fred", "2", "Fred2");
        System.out.println ("Row: " + tab.master_hashtable.toString());
    }


}
