/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.*;

/**
 * A class for matching string A against string B, where one of the strings
 * may have one or more wildcard characters (which are dots, '.') which can
 * match any single character.
 * 
 * For example, both 'banana' and 'mo.key' will match
 * "the monkey ate the banana".  You would say:
 * <pre>
 *     keywords = { "banana", "monkey" }; 
 *     ExpressionSearch expr = new ExpressionSearch(keywords);
 *     boolean matched = expr.matches("the monkey ate the banana");
 * </pre>
 * Perhaps later we'll add a larger subset of regular expressions.
 *
 * This class is implemented using a Trie, to optimize access speed.  For
 * example, if keywords = { "monkey", "moniker" }, then the Trie would look like:
 * <pre>
 *             __ K  E  Y
 *  M  O  N __/
 *            \__ I  K  E  R
 * </pre>
 */
public class ExpressionSearch
{
    private SearchNode top_node;
    private Vector next_nodes1;
    private Vector next_nodes2;
    private boolean wildcard_mode = false;
    
    /**
     * Create a new expression search object that will search for the 
     * input String 
     */
    public ExpressionSearch (String s)
    {
	addWord (s);
    }

    /**
     * Creates a new expression search object.  This method will 
     * cause an internal 'compiled' search tree to be generated from 
     * the list of input expressions.
     */
    public ExpressionSearch (String expression_list[])
    {
	for (int i = 0; i < expression_list.length; i++)
	{
	    addWord (expression_list[i]);
	}
    }
    
    public void setWildcardMode (boolean wm)
    {
	wildcard_mode = wm;
    }

    /**
     * returns true if the input string contains the expression
     */
    public boolean matches (String s)
    {
	if (s == null || s.equals ("") || top_node.childern_v == null)
	{
	    return (false);
	}
	
	SearchNode current_node;
	next_nodes1 = new Vector();
	next_nodes2 = new Vector();
	Vector source_nodes = next_nodes2;
	Vector dest_nodes = next_nodes1;
	Vector tmp_nodes;

	char sbuf[] = s.toCharArray();
	
	int pos = 0;
	Enumeration e;
	Enumeration e2;
	for (pos = 0; pos < sbuf.length; pos++)
	{
	    // check children of root node, place on next tree
	    e = top_node.childern_v.elements();
	    while (e.hasMoreElements())
	    {
		current_node = (SearchNode) e.nextElement();

		if (current_node.value == sbuf[pos])
		{
		    if (current_node.childern_v == null)
		    {
			return (true);
		    }
		    else
		    {
			dest_nodes.addElement(current_node);
		    }
		}
		else if (wildcard_mode && (sbuf[pos] == '.'))
		{
		    if (current_node.childern_v == null)
		    {
			return (true);
		    }
		    else
		    {
			dest_nodes.addElement(current_node);
		    }		    
		}
	    }
	    
	    // process source to dest (effectively mini-roots) 
	    e = source_nodes.elements();
	    while (e.hasMoreElements())
	    {
		current_node = (SearchNode) e.nextElement();

		e2 = current_node.childern_v.elements();
		while (e2.hasMoreElements())
		{
		    current_node = (SearchNode) e2.nextElement();
		    if (current_node.value == sbuf[pos])
		    {
			// if no more child nodes or we are at the last
			if (current_node.childern_v == null)
			{
			    return (true);
			}
			else
			{
			    dest_nodes.addElement(current_node);
			}
		    }
		    else if (wildcard_mode && (sbuf[pos] == '.'))
		    {
			if (current_node.childern_v == null)
			{
			    return (true);
			}
			else
			{
			    dest_nodes.addElement(current_node);
			}		    
		    }
		}
	    }

	    // clear source
	    source_nodes.setSize(0);

	    // switch dest and source
	    tmp_nodes = source_nodes;
	    source_nodes = dest_nodes;
	    dest_nodes = tmp_nodes;
	}

	return (false);
    }

    private void addWord (String s)
    {
	SearchNode current_node;
	SearchNode child_node;
	
	if (top_node == null)
	{
	    top_node = new SearchNode('.');
	}

	// place each character in the tree
	int len = s.length();
	current_node = top_node;
	
	for (int pos = 0; pos < len; pos++)
	{
	    if (current_node.childern_v != null)
	    {
		// if we are adding a word that is even shorter, chop
		// off the bottom of the tree
		boolean placed = false;
		Enumeration e = current_node.childern_v.elements();

		// cruise through current childern, see if they have same value
		while (!placed && e.hasMoreElements())
		{
		    child_node = (SearchNode) e.nextElement();
		    
		    if (child_node.value == s.charAt(pos))
		    {
			placed = true;

			// trucate tree if we have a shorter input expression
			// than one currently in the tree
			if (pos == len -1)
			{
			    child_node.childern_v = null;
			}
			else if (child_node.childern_v == null)
			{
			    // don't go any further if we aleady have 
			    // the base in place 
			    // aka don't add 'asswipe' if we already have 'ass'
			}
			else
			{
			    current_node = child_node;
			}
		    }
		}
		
		// the current character is a new one, create new node
		if (!placed)
		{
		    child_node = new SearchNode (s.charAt(pos));
		    current_node.addChild (child_node);
		    current_node = child_node;
		}
	    }
	    else 
	    {
		child_node = new SearchNode (s.charAt(pos));
		current_node.addChild (child_node);
		current_node = child_node;
	    }
	}
    }
}

class SearchNode 
{
    char value;
    Vector childern_v; 

    /** 
     * new search node with char value
     */
    SearchNode (char c)
    {
	value = c;
	childern_v = null;
    }
    
    /**
     * append a new SearcNode to this node's list of childern
     */
    void addChild (SearchNode sn)
    {
	if (childern_v == null)
	{
	    childern_v = new Vector();
	}
	
	childern_v.addElement (sn);
    }
}
