/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.*;

/**
 * A FIFO is essentially a variable length list.  Trying to stuff another
 * object onto a full FIFO will cause the calling thread to block until 
 * one of two things occurs: another thread removes an object, freeing space
 * for another, or the FIFO is closed (also via another thread).  If the 
 * FIFO is closed, any blocked callers will be thrown a run time exception -
 * IndexOutOfBoundsException
 */
public class FIFO
{
    private boolean closed = false;
    private int max_size = 0;
    private int size = 0;
    private FIFOStorage head;
    private FIFOStorage tail;

    /**
     * create a new empty FIFO with max size of 8.
     */
    public FIFO ()
    {
	max_size = 8;
	tail = head = null;
    }
    
    /**
     * create a new FIFO, with the specified max size.
     * @param msz   the maximum number of objects in the FIFO at once
     */
    public FIFO (int msz)
    {
	max_size = (msz < 1) ? 1 : msz;
	tail = head = null;
    }

    /**
     * close the fifo.  This will cause other thread who are trying to 
     * stuff new objects onto the FIFO to receive IndexOutOfBoundsExceptions
     * Threads taking objects out of the FIFO can continue to do so until
     * the FIFO is emptied, at which point any attempts to pull new objects
     * will result in an IndexOutOfBoundsException being thrown
     */
    public synchronized void close()
    {
	closed = true;
	notify ();
    }

    /**
     * Stuffs an object into the FIFO.  Will block if the FIFO is currently
     * full and will complete when there is enough room.  If the FIFO is 
     * closed or is closed by another thread while the caller is blocked, 
     * stuff will return with false.
     * @param o  the object to be stuffed into the FIFO
     */
    public synchronized boolean stuff (Object o)
    {
	if (closed)
	{
	    return (false);
	}

	while ((size >= max_size) && (! closed))
	{
	    try
	    {
		wait ();
	    }
	    catch (InterruptedException ie) {}
	}
	
	if (closed) 
	{
	    return (false);
	}

	if (head == null)
	{
	    head = new FIFOStorage (o);
	    tail = head;
	    size = 1;
	}
	else
	{
	    FIFOStorage oldtail = tail;
	    tail = new FIFOStorage (o);
	    tail.next = oldtail;
	    oldtail.previous = tail;
	    size++;
	}

	notify ();
	return (true);
    }

    /**
     * Grabs the next Object off the FIFO queue.  Will block until 
     * an object is available. An exception will be thrown if the 
     * FIFO is closed at the time of the call or if closed while the caller
     * is blocked
     */
    public synchronized Object yank () throws IndexOutOfBoundsException
    {
	while (head == null)
	{
	    if (closed)
	    {
		throw new IndexOutOfBoundsException();
	    }
	    try
	    {
		wait ();
	    }
	    catch (InterruptedException ie) {}
	}

	// get object out
	Object outbound = head.o;

	FIFOStorage oldhead = head;
	head = head.previous;
	if (head != null)
	{
	    head.next = null;
	}
	oldhead.previous = null;

	size--;
	notify ();
	return (outbound);
    }

    /**
     * true if there are currently objects in the queue.  Note that 
     * this state may change while the caller isn't looking if multiple 
     * threads are accesing the same FIFO.
     */
    public boolean hasContents ()
    {
	return (size > 0);
    }

    /**
     * current size (number of objects in the FIFO).  Like hasContents()
     * this state may change if the FIFO is being accessed by multiple 
     * concurrent threads
     */
    public synchronized int size ()
    {
	return (size);
    }
}

class FIFOStorage 
{
    FIFOStorage previous = null;
    FIFOStorage next = null;
    Object o = null;

    FIFOStorage (Object obj)
    {
	o = obj;
    }
}
