/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

/**
 * manages a pool of numbers.  Marks numbers that are used and will not
 * give them out again until they are marked as free.  Getting a new
 * number will always give you the lowest available in the set.
 */
public class LowNumberKeeper 
{
    static final int NUMBER_IN_USE = 1;
    static final int NUMBER_NOT_IN_USE = 0;
    private int values[];
    private int maximum; 
    private int first; 
    
    /**
     * creates a new keeper.  max - first + 1 gives the size of the 
     * pool
     * @param first - the first number in the series that is available
     * @parm  max   - the last number that will be available
     */
    public LowNumberKeeper (int fst, int max) throws IllegalArgumentException
    {
	first = fst;
	maximum = max;
	
	if (maximum < first)
	{
	    throw new IllegalArgumentException();
	}
	
	values = new int [maximum - first + 1];
	for (int i = 0; i < values.length; i++)
	{
	    values[i] = NUMBER_NOT_IN_USE;
	}
    }
    
    /**
     * gets the next available number in the sequence.  
     * @throws ArrayIndexOutOfBoundsException if no numbers are available
     */
    public int getNextNumber () throws ArrayIndexOutOfBoundsException
    {
	for (int i = 0; i < values.length; i++)
	{
	    if (values[i] == NUMBER_NOT_IN_USE)
	    {
		values[i] = NUMBER_IN_USE;
		return (first + i);
	    }
	}
	
	throw new ArrayIndexOutOfBoundsException ();
    }

    /**
     * Returns the number back to the pool so that others may use it.
     * @param val the value to be made available in the pool
     */
    public void freeNumber (int val)
    {
	// ignore values that aren't managed here
	if (val < first || val > maximum)
	{
	    return;
	}

	val = val - first;
	
	values[val] = NUMBER_NOT_IN_USE;
    }
}
