/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.*;
import org.locomotive.server.*;
import org.locomotive.directory.LocomotiveService;
import org.locomotive.directory.ServiceException;

/**
 * Manages a pool of objects.  Will broker the objects
 * out until there are no more to give out, whereupon it will either
 * attempt to create a new object and send that out, or 
 * block if the maximum number of objects are already open until an
 * object is available.
 * <P>To subclass this, you'll probably want to override the return and wipe
 * methods to do an instanceof check for your particular object type, and
 * you may want to check the object's state in returnToPool.
 */
public abstract class ObjectPool implements LocomotiveService
{
    /** the stack to use to hold the objects */
    protected Stack pe_stack;
    /** This holds objects in use.  Key: object; value: pooled entry */
    protected Hashtable pe_in_use_hash;

    protected int min_open = 0;
    protected int max_open = 0;
    protected int current_open = 0;
    protected int poolentry_id_sequence = 0;

    protected Log log;
    protected int log_level;

    protected boolean isRunning = false;
    

    /** 
     * This gets prepended to each logging message.  It will automatically
     * get set to the class name minus any packages in the constructor.
     * subclasses can set it anything they like after calling super().
     */
    protected String classname;  
    

    /** creates a new pool, and fills it with the specified minumum number
     * of objects
     * @param min_open_in the minumum number of objects to remain
     *        in the pool
     * @param max_open_in the maximum number of objects to 
     *        maintain at any given time
     */
    public ObjectPool (int min_open_in, int max_open_in)
	throws InitializationException
    {
	this(min_open_in, max_open_in, null, 0);
    }
	
    /** creates a new pool, and fills it with the specified minumum number
     * of objects
     * @param min_open_in the minumum number of objects to remain
     *        in the pool
     * @param max_open_in the maximum number of objects to 
     *        maintain at any given time
     * @param l the log to use to log any major events.  Leave as null for no 
     *        logging
     * @param log_lvl_in the log level to use when logging
     */
    public ObjectPool (int min_open_in, int max_open_in, Log l, int log_lvl_in)
	throws InitializationException
    {
	min_open = min_open_in;
	max_open = max_open_in;
	log = l;
	log_level = log_lvl_in;
	
	preInit();
    }
    

    /**
     * Default constructor for the object pool
     */
    public ObjectPool(){
    }

    /**
     * Only called the first time the object pool is initialized.
     */
    private void preInit() throws InitializationException {
	if ((min_open < 1) || (max_open < 1))
	{
	    throw new InitializationException ( 
		"Pool requires at least one object");
	}

	// initialize storage
	pe_stack = new Stack ();
	pe_in_use_hash = new Hashtable (max_open);

	// get the class name
	classname = getClass().getName();
	int index = classname.lastIndexOf('.');
	if (index > 0)
	{
	    classname = classname.substring(index + 1);
	}
    }
    

    /** This needs to be called after the pool is created */
    public synchronized void initialize() throws InitializationException
    {
	// create the minimum number of objects
	for (int i = 0; i < min_open; i++)
	{
	    pe_stack.push (new PoolEntry(createNewObject()));

	    log ("init: opened (" + current_open + ")");

	    current_open++;
	}
    }	

    /**
     * Start the object pool service. Subclasses should make
     * sure to call this method if they override, so the object pool
     * is initialized correctly.
     * @exception ServiceException is thrown if an error occured to startup.
     */
    public synchronized void startService() throws ServiceException {
	try {
	    preInit();
	    initialize();
	} catch(InitializationException iae){
	    throw new ServiceException(iae.toString());
	}
	isRunning = true;
    }
    /**
     * Stop the object pool service. 
     * @exception ServiceException if the object pool 
     *		  not could be stopped.
     */
    public synchronized void stopService() throws ServiceException {
	closeCurrentPoolMembers();
	isRunning = false;
    }

    /**
     * Check if the object pool is started.
     * @return Return true if the object pool is started, 
     *	       false otherwise.
     */
    public boolean isRunning(){
	return isRunning;
    }



    /** 
     * This is called internally whenever a new object is created and
     * added to the pool. Override this for each subclass.  
     */
    protected abstract Object createNewObject() throws InitializationException;
	
    
    /**
     * This should be overridden in sub classes to clean up any of an object's
     * resources before the object get's removed from the pool.
     */
    protected abstract void destroyObject(Object obj);


    /**
     * gets an object from the pool.  If the pool is currently
     * empty (meaning all objects are in use by other threads) and 
     * the current total is less than the configured maximum, a new 
     * object is created and returned.  Otherwise this method blocks 
     * until an object is released by another thread.  
     */ 
    public synchronized Object getObject() 
    {
	PoolEntry pe;
	
	// if a pe is available on the stack
	if (! pe_stack.empty ())
	{
	    pe = (PoolEntry) pe_stack.pop ();

	    log("pe " + pe.id + " : " + pe.usage_count);

	    pe_in_use_hash.put (pe.object, pe);

	    pe.usage_count++;
	    return (pe.object);
	}

	// if the Stack is empty, see if we can't create another object
	try
	{
	    if (current_open < max_open)
	    {
		pe = new PoolEntry(createNewObject());
		pe_in_use_hash.put (pe.object, pe);
		pe.usage_count++;
		current_open++;
		return (pe.object);
	    }
	}
	catch (InitializationException ie) 
	{
	    // what should we do?  The subclass can log the exception in
	    // createNewObject, if they like
	}

	//
	// we're gonna have to wait
	//

	// Log the fact that we're stuck here (good output for admins
	// so they know that the pool should probably be made larger)!
	int previous_log_level = getLogLevel();
	setLogLevel(Log.WARNING);
	log(classname + ": All Objects in use!  Waiting..");

	// Wait until the PoolEntry Stack is no longer empty...
	while (pe_stack.empty())
	{
	    try
	    {
		wait();
	    }
	    catch (InterruptedException ie) {}
	}

	// Log the fact that we're done waiting (this could have been a while)
	log(classname + ": Done waiting.");
	setLogLevel(previous_log_level);

	pe = (PoolEntry) pe_stack.pop ();
	pe_in_use_hash.put (pe.object, pe);

	pe.usage_count++;
	return (pe.object);
    }


    /**
     * returns the current number of open db objects in the pool
     */ 
    public int getPoolSize ()
    {
	return (current_open);
    }

    /**
     * returns this object to the pool, allowing other threads to 
     * have access to it.  This should be called when the calling thread 
     * no longer needs the object.  Subclasses may want to override to
     * check the type of the Object being returned.
     */
    public synchronized void returnToPool (Object obj)
    {

	// pull up pe from cold storage
	PoolEntry pe = (PoolEntry) pe_in_use_hash.remove(obj);

	if (pe != null)
	{
	    if (pe.usable)
	    {
		// add it back into the Stack for others to use
		pe_stack.push (pe);
		notify(); // notify any blocked getObjects()s
	    }
	    else
	    {
		// wipe the object
		current_open--;
		
		log ("removed pe " + pe.id + " after " + pe.usage_count + 
		     " uses");   
		
		destroyObject(obj);
	    }
	}
    }


    /**
     * removes the specified object from the pool. Usually used if an
     * object has been put into some bad state, like a closed connection,
     * for instance.
     */
    public synchronized void wipeFromPool (Object obj)
    {
	PoolEntry pe = (PoolEntry) pe_in_use_hash.get (obj);
	pe_in_use_hash.remove (obj);

	if (pe != null)
	{
	    current_open--;

	    log ("removed pe " + pe.id + " after " + pe.usage_count + 
		 " uses");   
	}

	destroyObject(obj);
    }

    /**
     * Closes all objects in the pool at the time of the call.
     * If other objects are returned to the pool after this call, 
     * they will not be closed, and this call would need to be made again.
     */
    public synchronized void closeCurrentPoolMembers ()
    {
	PoolEntry pe;
	
	// mark all the connection in use so they'll we wiped when they return
	Enumeration e = pe_in_use_hash.elements();
	while (e.hasMoreElements())
	{
	    ((PoolEntry) e.nextElement()).usable = false;
	}

	// wipe all the connections in the pool
	while (! pe_stack.empty())
	{
	    pe = (PoolEntry) pe_stack.pop ();
	    current_open--;
	    log ("removed pe " + pe.id + " after " + pe.usage_count + 
		 " uses");   
	    destroyObject(pe.object);
	}
    }

    /**
     * Each PoolEntry has a numeric ID, this method increments it and
     * gives it out.
     */
    protected synchronized int getNextId()
    {
	return (poolentry_id_sequence++);
    }


    /** 
     * Log a statement.  Will provide the class name for the object
     * pool at the the beginning of the statement.  Will use the log
     * level and log specified in the constructor; if they're null,
     * then it won't log 
     */
    protected void log(String message)
    {
	if (log != null)
	{
	    log.log(log_level, classname + ": " + message);
	}
    }

    /** 
     * Log an error.  Will provide the class name for the object
     * pool at the the beginning of the statement.  Will use the log
     * level and log specified in the constructor; if they're null,
     * then it won't log 
     */
    protected void logError(String message, Throwable t)
    {
	if (log != null)
	{
	    log.log(log_level, "ERROR: " + classname + ": " + message + ": " + 
		StringUtil.exceptionStackTrace(t));
	}
    }

    /**
     * The internal class to use as our Object entry.
     */
    public class PoolEntry 
    {
	public Object object;
	public int id = 0;
	public int usage_count = 0;
	public boolean usable = false;
    
	public PoolEntry (Object o)
	{
	    object = o;
	    id = getNextId();
	    usage_count = 0;
	    usable = true;
	}

    }

    /**
     * Set the minimum size of the object pool.
     * @param minSize The minimum size of the object pool.
     */
    public void setMinSize(int minSize){
	this.min_open = minSize;
    }

    /**
     * Get the minimum size of the object pool.
     * @return The minimum size of the object pool.
     */
    public int getMinSize(){
	return min_open;
    }

    /**
     * Set the maximum size of the object pool.
     * @param maxSize The maximum size of the object pool.
     */
    public void setMaxSize(int maxSize){
	this.max_open = maxSize;
    }

    /**
     * Get the maximum size of the object pool.
     * @return The maximum size of the object pool.
     */
    public int getMaxSize(){
	return max_open;
    }

    /**
     * Set the log for this object pool.
     * @param log The log object for this connetion pool.
     */
    public void setLog(Log log){
	this.log = log;
    }

    /**
     * Set the log level for this object pool.
     * @param logLevel The log level for this object pool.
     */
    public void setLogLevel(int logLevel){
	this.log_level = logLevel;
    }
    
    /**
     * Get the log level for this object pool.
     * @return The log level for this object pool.
     */
    public int getLogLevel(){
	return log_level;
    }
}


