/*
 * The contents of this file are subject to the Locomotive Public License
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You
 * may not use this file except in compliance with the License; You may
 * obtain a copy of the License at http://www.locomotive.org/license/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The initial developers of this code under the LPL is Leverage Information
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.*;
import java.sql.*;  // for timestamp
import java.text.SimpleDateFormat;
import java.text.ParsePosition;

/**
 * This class contains some useful utilities for dealing with SQL statements
 * and results of queries.
 */
public class SQL
{

    // to help us out with the database- and to keep the permissions
    // seperate from the Locomotive
    public final static int DB_TYPE_ORACLE =        0;
    public final static int DB_TYPE_INFORMIX =      1;
    public final static int DB_TYPE_SOLID =         2;
    public final static int DB_TYPE_CLOUDSCAPE =    3;
    public final static int DB_TYPE_MYSQL =         4;
    public final static int DB_TYPE_POSTGRESQL =    5;
    public final static int DB_TYPE_INTERBASE  =    6;


    /**
     * This will return one of this class's final static int DB_TYPES for
     * the database connection supplied.
     */
    public static int getDBType(Connection conn) throws SQLException
    {
	DatabaseMetaData dbmd = conn.getMetaData();
	String db_type = dbmd.getDatabaseProductName().toLowerCase();
	
	if (db_type.equalsIgnoreCase("oracle"))
	{
	    return ( SQL.DB_TYPE_ORACLE );
	}
	else if (db_type.equalsIgnoreCase("mysql"))
	{
	    return ( SQL.DB_TYPE_MYSQL );
	}
	else if (db_type.equalsIgnoreCase("informix"))
	{
	    return ( SQL.DB_TYPE_INFORMIX );
	}
	else if (db_type.equalsIgnoreCase("solid"))
	{
	    return ( SQL.DB_TYPE_SOLID );
	}
	else if (db_type.equalsIgnoreCase("cloudscape"))
	{
	    return ( SQL.DB_TYPE_CLOUDSCAPE );
	}
	else if (db_type.equalsIgnoreCase("postgresql"))
	{
	    return ( SQL.DB_TYPE_POSTGRESQL );
	}
	else if (db_type.equalsIgnoreCase("interbase"))
	{
	    return ( SQL.DB_TYPE_INTERBASE );
	}
	else
	{
	    throw new SQLException("Database " + db_type + " not supported.");
	}
    }	


    /**
     * A handy pre-configured format parser for date text fields. <BR>
     *
     * Takes an input string in the form <BR>
     * &quot;YYYY-MM-DD HH:MM:SS&quot;<BR> or
     * &quot;YYYY-MM-DD&quot;<BR> or
     * &quot;YY-MM-DD&quot;<BR> or
     * &quot;YY-MM-DD HH:MM:SS&quot;<BR>
     *
     * @exception IllegalArgumentException  if the input format does not match
     */
    public static Timestamp stringToTimestamp (String s) throws
        IllegalArgumentException
    {
	java.util.Date d;
	Timestamp ts = new Timestamp(System.currentTimeMillis());

	// try first format
	SimpleDateFormat sdf = new SimpleDateFormat ("yyyy-MM-dd kk:mm:ss");
	sdf.setLenient (false);
	
	d = sdf.parse (s, new ParsePosition (0));
	if (d != null)
	{
	    ts.setTime (d.getTime());
	    return (ts);
	}
	
	// try next format
	sdf = new SimpleDateFormat ("yyyy-MM-DD");
	sdf.setLenient (false);
	
	d = sdf.parse (s, new ParsePosition (0));
	if (d != null)
	{
	    ts.setTime (d.getTime());
	    return (ts);
	}

	// try next format

	sdf = new SimpleDateFormat ("yy-MM-DD kk:mm:ss");
	sdf.setLenient (false);
	
	d = sdf.parse (s, new ParsePosition (0));
	if (d != null)
	{
	    ts.setTime (d.getTime());
	    return (ts);
	}

	// try final format
	sdf = new SimpleDateFormat ("yy-MM-DD");
	sdf.setLenient (false);
	
	d = sdf.parse (s, new ParsePosition (0));
	if (d != null)
	{
	    ts.setTime (d.getTime());
	    return (ts);
	}

	throw new IllegalArgumentException ("unrecognizable format");
    }

    /**
     * builds an SQL &quot;IN clause&quot; using the input column name
     * and number array
     * @param colname - the name of the column in the db
     * @param numarray - used as the data for the clause contents
     */
    public static String buildINClause (String colname, int numarray[])
    {
	if (numarray.length == 0)
	{
	    return ("1 = 1");
	}
	
	if (numarray.length == 1)
	{
	    return (colname + " = " + numarray[0]);
	}
	
	StringBuffer sb = new StringBuffer();

	sb.append (colname);
	sb.append (" IN (");

	int stop = numarray.length - 1;
	for (int i = 0; i < stop; i++)
	{
	    sb.append (String.valueOf (numarray[i]));
	    sb.append (", ");
	}
	
	sb.append (String.valueOf (numarray[stop]));
	sb.append (")");
	return (sb.toString());
    }

    /**
     * builds an SQL &quot;NOT IN clause&quot; using the input column name
     * and number array
     * @param colname - the name of the column in the db
     * @param numarray - used as the data for the clause contents
     */
    public static String buildNOTINClause (String colname, int numarray[])
    {
	if (numarray.length == 0)
	{
	    return ("1 = 1");
	}
	
	if (numarray.length == 1)
	{
	    return (colname + " != " + numarray[0]);
	}
	
	StringBuffer sb = new StringBuffer();

	sb.append (colname);
	sb.append (" NOT IN (");

	int stop = numarray.length - 1;
	for (int i = 0; i < stop; i++)
	{
	    sb.append (String.valueOf (numarray[i]));
	    sb.append (", ");
	}
	
	sb.append (String.valueOf (numarray[stop]));
	sb.append (")");
	return (sb.toString());
    }

    /**
     * quotes a String so it can be used in a statement, putting it within
     * single quotes, '...'.
     * Escapes all single quotes ' within the String.
     */
    public static String quoteString (String s)
    {
	if (s == null)
	{
	    return ("''");
	}

	int total = 0;
	char input[] = s.toCharArray();
	
	for (int i = 0; i < input.length; i++)
	{
	    if (input[i] == '\'')
	    {
		total++;
	    }
	}

	char output[] = new char[input.length + total + 2];
	
	int pos = 0;
	output[0] = '\'';
	pos++;

	for (int i = 0; i < input.length; i++)
	{
	    if (input[i] == '\'')
	    {
		output[pos] = '\'';
		pos++;
		output[pos] = '\'';
		pos++;
	    }
	    else
	    {
		output[pos] = input[i];
		pos++;
	    }
	}
	
	output[pos] = '\'';
	return (new String (output));
    }

    /**
     * creates a quoted String, enclosing the String argument within '%..%',
     * to be used in a LIKE '%...%' clause.
     */
    public static String quoteStringMidLike (String s)
    {
	if (s == null)
	{
	    return ("'%'");
	}

	int total = 0;
	char input[] = s.toCharArray();
	
	for (int i = 0; i < input.length; i++)
	{
	    if (input[i] == '\'')
	    {
		total++;
	    }
	}

	char output[] = new char[input.length + total + 4];
	
	int pos = 0;
	output[pos++] = '\'';
	output[pos++] = '%';

	for (int i = 0; i < input.length; i++)
	{
	    if (input[i] == '\'')
	    {
		output[pos++] = '\'';
		output[pos++] = '\'';
	    }
	    else
	    {
		output[pos++] = input[i];
	    }
	}
	
	output[pos++] = '%';
	output[pos] = '\'';
	return (new String (output));
    }

    /**
     * creates a quoted String, enclosing the String argument within '..%',
     * to be used in a LIKE '...%' clause for searching for prefixes.
     */
    public static String quoteStringEndLike (String s)
    {
	if (s == null)
	{
	    return ("'%'");
	}
	int total = 0;
	char input[] = s.toCharArray();
	
	for (int i = 0; i < input.length; i++)
	{
	    if (input[i] == '\'')
	    {
		total++;
	    }
	}

	char output[] = new char[input.length + total + 3];
	
	int pos = 0;
	output[pos++] = '\'';

	for (int i = 0; i < input.length; i++)
	{
	    if (input[i] == '\'')
	    {
		output[pos++] = '\'';
		output[pos++] = '\'';
	    }
	    else
	    {
		output[pos++] = input[i];
	    }
	}
	
	output[pos++] = '%';
	output[pos] = '\'';
	return (new String (output));
    }

/*
    public static BitSet parseBitSet(String bss)
    {
	if (bss == null) return (new BitSet());

	BitSet bs = new BitSet();

	int i = 0;
	int max = bss.length();

	while (i < max)
	{
	    if (bss.charAt(i) == '1')
		bs.set(i);
	    else
		bs.clear(i);

	    i++;
	}

	return(bs);
    }

    public static String bitSetToString(BitSet bs)
    {
	if (bs == null) return("");

	int max = bs.size();

	if (max <= 0) return("");

	StringBuffer out = new StringBuffer();

	int i = 0;

	while (i < max)
	{
	    out.append(bs.get(i) ? "1" : "0");

	    i++;
	}

	return(out.toString());
    }
    */


    /**
     * Everybody knows that SQL cannot do bit-wise logical
     * operations, such as bit-wise AND or ORs.  But, you can
     * simulate the same kind of thing using the following
     * principle: if you need to test whether the 17th bit is on
     * in your number, you can use the following test in SQL:
     * <pre>
     *    if (number &gt;= 2^17) and (number/2^18 - floor (number/2^18) &gt;= 0.5)...
     * </pre>
     * This method will generate the SQL equivalent of the above statement.
     * The first part, number >= 2^17, is to add speed to the search.
     * Example of using this method:
     * <pre>
     *    String criterion = statementToTestBitInNumber ("loco_users.flags", 1&lt;&lt;17);
     * </pre>
     * will yield the string:
     * <pre>
     *    (loco_users.flags &gt;= 131072) and
     *       ((loco_users.flags/262144) - floor(loco_users.flags/262144) &gt;= 0.5)
     * </pre>
     *
     * @param column_name -- the name of the column in your database table.
     * @param num_with_nth_bit_on -- a number with only one bit turned on,
     * the bit you're interested in.  You can pass in 1 &lt;&lt; <i>nth-bit</i>.
     */
    public static String statementToTestBitInNumber (String column_name,
						     int num_with_nth_bit_on)
    {
	StringBuffer statement = new StringBuffer(120);
	statement.append ("((");
	statement.append (column_name);
	statement.append ("/");
	statement.append (num_with_nth_bit_on * 2);
	statement.append (") - floor(");
	statement.append (column_name);
	statement.append ("/");
	statement.append (num_with_nth_bit_on * 2);
	statement.append (") >= 0.5) and (");
	statement.append (column_name);
	statement.append (" >= ");
	statement.append (num_with_nth_bit_on);
	statement.append (")");
	
	return statement.toString();
    }


    /**
     * executes a query (which should request a single number)
     * and returns an int.
     * @throws SQLException if the query fails to return a number.
     */
    public static int getInt(Connection conn, String query)
	throws SQLException
    {
	if (query == null || query.length() == 0)
	    return -1;
	
	Statement stmt = null;
	ResultSet rs = null;
	int toReturn = -1;
	
	try {
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);
	    if (rs.next())
	    {
		toReturn = rs.getInt(1);
	    }
	    else
	    {
		throw new SQLException("Warning: org.locomotive.util.SQL.getInt(): "+
				       " Your query, '" + query +
				       "', returned no results!");
	    }
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}

	return toReturn;
    }

    /**
     * executes a query (which should request a single number)
     * and returns a long.
     * @throws SQLException if the query fails to return a number.
     */
    public static long getLong(Connection conn, String query)
	throws SQLException
    {
	if (query == null || query.length() == 0)
	    return -1;
	
	Statement stmt = null;
	ResultSet rs = null;
	long toReturn = -1;
	
	try
	{
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(query);
	    if (rs.next())
	    {
		toReturn = rs.getLong(1);
	    }
	    else
	    {
		throw new SQLException(
		    "Warning: org.locomotive.util.SQL.getLong(): "+
		    " Your query, '" + query + "', returned no results!");
	    }
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}

	return toReturn;
    }

    /**
     * executes a single SQL update (which may be either an insert or an
     * update).
     * @returns the number of rows updated.
     * @throws SQLException if the update fails.
     */
    public static int update(Connection conn, String statement)
	throws SQLException
    {
	if (statement == null || statement.length() == 0)
	    return 0;
	
	Statement stmt = null;
	try
	{
	    stmt = conn.createStatement();
	    return (stmt.executeUpdate(statement));
	}
	finally
	{
	    if (stmt != null) stmt.close();
	}
    }
}
