/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import org.locomotive.util.data.*;
import java.util.*;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.sql.*;
import java.io.*;


/** 
 * SortableTable is a data structure that allows for easy and persistent storage
 * of java.sql.ResultSets.  SortableTables, unlike ResultSets, are not tied to
 * any database connections, and so can be maintained without tying up database
 * connections.  SortableTables also have the added functionality of sorting by 
 * column.  Both columns and rows may be added after the object is constructed.
 * However, adding columns is very expensive, and rows which are added can 
 * only be added to the end of the structure.
 */
public class SortableTable 
{
    /** used to denote an ascending sort */
    public final static int ASCENDING  = -1024;

    /** used to denote a descending sort */
    public final static int DESCENDING = -1025;

    /** used to denote a number comparison for sorting */
    public final static int NUMBER = 1024;

    /** used to denote a String comparison for sorting */
    public final static int STRING = 1025;

    /** used to denote a Date comparison for sorting */
    public final static int DATE = 1026;


    private int numRows;             // the number of rows in the table
    private int numColumns;          // the number of columns in the table

    private int rowCapacity;         // the present capacity for rows
    private int columnCapacity;      // the present capacity for columns

    private int [] SQLTypes;         // the SQLType of the corresponding
                                     // column in the database 
    private String [] names;         // the names of each of the columns
    private Hashtable namesHash;     // keys are column names, 
                                     // values are the column indexes

    private Object [] [] table;      // The array holding the objects

    private int rowCapacityIncrement;     // number of rows to add in resize
    private int columnCapacityIncrement;  // number of columns to add in resize





    /**
     * Creates a SortableTable.  <PRE>
     * Defaults: initial row capacity = 10;   initial column capacity = 15
     *           row capacity increment = 10;   column capacity increment = 15
     *</PRE>
     */
    public SortableTable ()
    {
	this(10, 15, 10, 15);
    }

    
    /**
     * Creates a SortableTable whose initial size is given by the specified 
     * parameters.  <PRE>
     * Defaults: row resize increment = 10;   column resize increment = 15
     * </PRE>
     */
    public SortableTable (int rowCapacity, int columnCapacity)
    {
	this(rowCapacity, columnCapacity, 10 , 15);
    }
    

    /**
     * creates a sortable table with the specified number of rows and columns.
     */
    public SortableTable (int rowCapacity, int columnCapacity, 
			  int rowCapacityIncrement, int columnCapacityIncrement)
    {
	this.rowCapacity = rowCapacity;
	this.columnCapacity = columnCapacity;
	this.rowCapacityIncrement = rowCapacityIncrement;
	this.columnCapacityIncrement = columnCapacityIncrement;
	numColumns = 0;
	numRows = 0;

	table = new Object [rowCapacity] [columnCapacity];	
	names = new String [columnCapacity];
	namesHash = new Hashtable();
	SQLTypes = new int [columnCapacity];
    }


    /** 
     * Creates a new table out of a subset of rows of another table
     * All the columns of the source table are copied to the destination
     * table. Column indexes are preserved- i.e. column 4 in the source
     * table will also be column 4 in the destination table.
     * The column names of the source table will be preserved also.
     * @param source the table to load the data from
     * @param startRow the row of the source table to begin loading from
     * @param endRow the final row to load into the new table
     */
    public SortableTable (SortableTable source, int startRow, 
		      int endRow) 
    {
	this(endRow - startRow + 1, source.getNumColumns(), 10 , 15);
	this.numColumns = source.getNumColumns();
	this.numRows = endRow - startRow + 1;

	for (int i = 0; i < numColumns; i++)
	{
	    String s = source.getColumnName(i);
	    if (s != null)
		setColumnName(i, s);
	    SQLTypes[i] = source.getSQLType(i);
	}
    
	for (int i = 0; i < numRows; i++)
	{
	    for (int j = 0; j < numColumns; j++)
	    {
		table [i] [j] = source.get(startRow + i, j);
	    }
	}
    }


    /**
     * returns the number of rows in the table.
     */
    public int getNumRows()
    {
	return numRows;
    }


    /**
     * returns the number of columns in the table.
     */
    public int getNumColumns()
    {
	return numColumns;
    }


    /**
     * sets the number of rows to the specified size.  Note that if the column
     * size is set smaller than the present number of rows, rows will be
     * deleted!
     */
    public void setRowCapacity(int rowCapacity)
    {
	if (rowCapacity < numRows)
	{
	    for (int i = rowCapacity; i < numRows; i++)
	    {
		table [i] = null;
	    }		

	    numRows = rowCapacity;
	}
	else if (rowCapacity > this.rowCapacity)
	{
	    Object [] [] temp = new Object [rowCapacity] [columnCapacity];
	    System.arraycopy(table, 0, temp, 0, numRows);
	    table = temp;
	}
	else  // don't change the capacity
	    return;

	this.rowCapacity = rowCapacity;
    }


    /**
     * sets the number of columns to the specified size.  Note that if the
     * column size is set smaller than the present number of rows, rows will be
     * deleted!  Will not actually resize the data structure, however. 
     * This is a very expensive method, and should be avoided if possible 
     */
    public void setColumnCapacity(int columnCapacity)
    {
	if (columnCapacity < numColumns)
	{
	    // ditch the column names 
	    for (int j = columnCapacity; j < numColumns; j++)
	    {
		String toRemove = names [j];
		if (toRemove != null)
		    namesHash.remove(toRemove);
		names [j] = null;
		SQLTypes[j] = Types.NULL;
	    }

	    // ditch the data
	    for (int i = 0; i < numRows; i++)
	    {
		for (int j = columnCapacity - 1; j < numColumns; j++)
		    table [i] [j] = null;
	    }

	    numColumns = columnCapacity;
	}
	else if (columnCapacity > this.columnCapacity)
	{
	    // resize the names array
	    String [] temp = new String [columnCapacity];
	    System.arraycopy(names, 0, temp, 0, names.length);
	    names = temp;

	    // resize the SQLTypes
	    int [] temp3 = new int [columnCapacity];
	    System.arraycopy(SQLTypes, 0, temp3, 0, SQLTypes.length);
	    SQLTypes = temp3;

	    // resize each row
	    for (int i = 0; i < rowCapacity; i++)
	    {
		Object [] temp2 = new Object [columnCapacity];
		System.arraycopy(table [i], 0, temp2, 0, numColumns);
		table [i] = temp2;
	    }
	} else  // don't change the capacity
	    return;
	
	this.columnCapacity = columnCapacity;
    }


    /**
     * loads data into the table from a ResultSet, beginning at the row whose
     * index is specified, into the columns beginning with and following the
     * columns whose index is specified.  
     * @param rs the ResultSet to load the information from
     * @param startRow the index of the row to begin loading into
     * @param startColumn the index of the column to begin loading into
     * @param setNames whether to set the column names using the 
     *                 ResultSetMetaData.  Will also set the SQLType.
     * @exception SQLException if the resultSetMetaData cannot be accessed
     */
    public void load (ResultSet rs, int startRow, 
		      int startColumn, boolean setNames)
	throws SQLException
    {
	ResultSetMetaData rsmd = rs.getMetaData();
	int rsNumCols = rsmd.getColumnCount() + startColumn;
	
	// resize if we have to
	if (rsNumCols > columnCapacity)
	{
	    setColumnCapacity(rsNumCols + columnCapacityIncrement);
	}

	// increase the number of columns if we need to
	if (rsNumCols > numColumns)
	    numColumns = rsNumCols;
	
	// set the names if requested
	if (setNames)
	{
	    for (int i = startColumn; i < rsNumCols; i++)
	    {
		setColumnName(i, rsmd.getColumnLabel(i-startColumn+1));
		setSQLType(i, rsmd.getColumnType(i-startColumn+1));
	    }
	}

//	org.locomotive.loco.Loco.log(0, "NumCols: " + numColumns + 
//			   "\tcolCapacity: " + columnCapacity);

	// load the data
	for (int rowIndex = startRow; rs.next(); rowIndex ++)
	{
	    // resize if we have to
	    if (rowIndex == rowCapacity - 1)
		setRowCapacity(rowCapacity + rowCapacityIncrement);

	    // increase the number of rows if we need to
	    if (rowIndex == numRows)
		numRows ++;

//	    org.locomotive.loco.Loco.log(0, "NumRows: " + numRows + 
//				 "\trowCapacity: " + rowCapacity);

	    for (int colIndex = startColumn; colIndex < rsNumCols;  colIndex++ )
	    {
		table [rowIndex] [colIndex] = 
		    rs.getObject(colIndex-startColumn+1);
	    }
	}
    }


    /**
     * loads data into the table from a ResultSet, starting at the first row
     * and first column in the table
     * @param rs the ResultSet to load the information from
     * @param setNames whether to set the column names using the 
     *        ResultSetMetaData
     * @exception SQLException if the resultSetMetaData cannot be accessed
     */
    public void load (ResultSet rs, boolean setNames) throws SQLException
    {
	load(rs, 0, 0, setNames);
    }



    /** 
     * Creates a new table out of a subset of rows of another table
     * All the columns of the source table are copied to the destination
     * table. Column indexes are preserved- i.e. column 4 in the source
     * table will also be column 4 in the destination table.
     * The column names of the source table will be preserved also.
     * @param source the table to load the data from
     * @param startRow the row of the source table to begin loading from
     * @param endRow the final row to load into the new table
     *
    public void load (SortableTable table, int startRow, int endRow)
    {
	this.setColumnCapacity(table.getNumColumns());
	this.setRowCapacity(table.getNumRows());

	this.numRows = endRow - startRow + 1;
	this.numColumns = source.getNumColumns();
	
	for (int i = 0; i < numColumns; i++)
	{
	    String s = source.getColumnName(i);
	    if (s != null)
		setColumnName(i, s);
	    SQLTypes[i] = source.getSQLType(i);
	}
	
	for (int i = 0; i <  numRows; i++)
	{
	    for (int j = 0; j < numColumns; j++)
	    {
		table [i] [j] = source.get(startRow + i, j);
	    }
	}
    }

    */


    /**
     * inserts an object at the given indexes. The table is resized if the 
     * index is greater than the present table size.
     */
    public void insert (Object obj, int rowIndex, int columnIndex) 
    {

	// resize rows
	if (rowIndex > rowCapacity - 1)
	{
	    int toSet = rowCapacityIncrement + rowCapacity;
	    if (rowIndex >= toSet)
		toSet = rowIndex + rowCapacityIncrement;
	    setRowCapacity(toSet);
	}
	
	// resize columns
	if (columnIndex > columnCapacity - 1)
	{
	    int toSet = columnCapacityIncrement + columnCapacity;
	    if (columnIndex >= toSet)
		toSet = columnIndex + columnCapacityIncrement;
	    setColumnCapacity(toSet);
	}

	// increase the number of rows if we need to
	if (rowIndex >= numRows) {
	    numRows = rowIndex+1;
//	    org.locomotive.loco.Loco.log(0, "numRows resize: " + rowIndex + 
//				 " " + numRows);
	}

	// increase the number of columns if we need to
	if (columnIndex >= numColumns)
	    numColumns = columnIndex+1;

	table [rowIndex] [columnIndex] = obj;
    }

    
    /** 
     * inserts an object at the given row and column name.  The table is
     * resized if the row index is greater than the present table size.
     * @exception IllegalArgumentException if the column name is not
     * valid
     */
    public void insert(Object obj, int rowIndex, String columnName)
	throws IllegalArgumentException
    {
	int columnIndex = getColumnIndex(columnName);
	if (columnIndex == -1)
	{
	    throw new IllegalArgumentException ("Column " + columnName +
						" not in table");
	}
	insert(obj, rowIndex, columnIndex);
    }



    /**
     * removes an object stored at the specified indexes from the table
     * @return the object being removed. Note that removed objects will not 
     * resize the number of rows or columns.  the only real way to remove 
     * entire rows or columns is to resize the table using setColumnCapacity 
     * and setRowCapacity...  Sorry!
     */
    public Object remove (int rowIndex, int columnIndex)
    {
	Object toReturn = table [rowIndex] [columnIndex];
	table [rowIndex] [columnIndex] = null;
	return toReturn;
    }

    
    /**
     * returns the object stored at the given indexes.  Returns null if no 
     * is stored at those indexes. Returns null if the indexes are out of 
     * bounds     
     */
    public Object get (int rowIndex, int columnIndex)
    {
	if (numRows > rowIndex && numColumns > columnIndex)
	    return table [rowIndex] [columnIndex];
	else
	    return null;
    }


    /**
     * returns the object stored at the given indexes.  Returns null if no
     * object is stored at those indexes. Or if the indexes are out of bounds,
     * or if the column name does not exist
     */
    public Object get (int rowIndex, String columnName)
    {
	int columnIndex = getColumnIndex(columnName);
	if (columnIndex != -1 && numRows > rowIndex)
	    return table [rowIndex] [columnIndex];
	else
	    return null;
    }


    /**
     * sets the name of the column at the specified column index.
     */
    public void setColumnName (int columnIndex, String name)
    {
	// resize columns
	if (columnIndex > columnCapacity - 1)
	{
	    int toSet = columnCapacityIncrement + columnCapacity;
	    if (columnIndex >= toSet)
		toSet = columnIndex + columnCapacityIncrement;
	    setColumnCapacity(toSet);
	}

	// increase the number of columns if we need to
	if (columnIndex >= numColumns)
	    numColumns = columnIndex+1;


	String toRemove = names [columnIndex];
	if (toRemove != null && namesHash.containsKey(toRemove))
	    namesHash.remove(toRemove);

	names [columnIndex] = name;
	namesHash.put(name, new Integer(columnIndex));
    }


    /** 
     * returns the name of the columns at the specified column index.
     * Returns null if the name is not set, or if the index is out of the
     * array bounds
     */
    public String getColumnName (int columnIndex)
    {
	if (columnIndex < names.length)
	    return names [columnIndex];
	else
	    return null;
    }


    /**
     * returns the column index denoted by the specified column name.
     * Returns -1 if the specified name is not a column name
     */
    public int getColumnIndex (String name)
    {
	if (namesHash.containsKey(name))
	    return ((Integer) namesHash.get(name)).intValue();

	String temp = name.toUpperCase();
	if (namesHash.containsKey(temp))
	    return ((Integer) namesHash.get(temp)).intValue();
	
	temp = name.toLowerCase();
	if  (namesHash.containsKey(temp))
	    return ((Integer) namesHash.get(temp)).intValue();

	// we never found it - return -1
	return -1;
    }



    /**
     * Sets the column SQL Type.  Useful in case the an element of the
     * column is null but must be reinserted into the database.  
     * The actual int value for the SQL type is specified by the 
     * XOPEN SQL standard.
     * @see java.sql.Types
     * @param columnIndex the column whose SQL Type is to be set
     * @param SQLType one of the types specified in java.sql.Types
     */
    public void setSQLType(int columnIndex, int SQLType)
    {
	// resize columns
	if (columnIndex > columnCapacity - 1)
	{
	    int toSet = columnCapacityIncrement + columnCapacity;
	    if (columnIndex >= toSet)
		toSet = columnIndex + columnCapacityIncrement;
	    setColumnCapacity(toSet);
	}

	// increase the number of columns if we need to
	if (columnIndex >= numColumns)
	    numColumns = columnIndex+1;

	SQLTypes[columnIndex] = SQLType;
    }

    
    /** returns the column's SQL type.  Usefule for inserting nulls
     * into the database using prepared statements.  
     * The actual int value for the SQL type is specified by the 
     * XOPEN SQL standard.
     * @see java.sql.Types
     * @param columnIndex the column whose type to return.
     * @return the SQLType, or Types.NULL if it is not set
     */
    public int getSQLType(int columnIndex)
    {
	if (columnIndex >= SQLTypes.length)
	    return Types.NULL;
	else
	    return SQLTypes[columnIndex];
    }


    /**
     * sorts the table by the specified column name.  Uses a heap sort.
     *  Note that in most cases it will be a lot faster to sort a result set 
     * using an 'order by' sql clause in the query that the table is loaded 
     * from.
     * @param columnName the column name to sort by
     * @param type the kind of information in the column to sort (Given by the
     *             constants NUMBER, DATE, or STRING)
     * @param order the order (given by the constants ASCENDING or DESCENDING)
     *              to sort by
     */
    public void sort (String columnName, int type, int order)
    {
	sort(getColumnIndex(columnName), type, order);
    }

    
    /**
     * sorts the table by the specified column index.  Uses a heap sort.
     * This is slightly faster than the other sorting method.  Note that
     * in most cases it will be a lot faster to sort a result set using 
     * an 'order by' sql clause in the query that the table is loaded from.
     * @param columnIndex the column index to sort by
     * @param type the kind of information in the column to sort (Given by the
     *             constants NUMBER, DATE, or STRING)
     * @param order the order (given by the constants ASCENDING or DESCENDING)
     *              to sort by
     */
    public void sort (int columnIndex, int type, int order)
    {
	if (columnIndex > numColumns -1)
	    throw new NoSuchElementException("Column does not exist: " +
					     "Col: " + columnIndex + 
					     "Total: " + numColumns);

	Heap heap = new Heap();
	ComparisonKey key = null;

	switch (type)
	{
	case NUMBER:
	    for (int i = 0; i < numRows; i++)
	    {
		Number num = (Number) table[i][columnIndex];
		if (num == null)
		    key = new NumberKey (0);
		else
		    key = new NumberKey(num);

		heap.insert(key, table [i]);
	    }
	    break;
	case STRING:
	    for (int i = 0; i < numRows; i++)
	    {
		String s = (String) table[i][columnIndex] ;
		if (s == null)
		   key = new StringKey("");
		else
		    key = new StringKey(s);

		heap.insert(key, table [i]);
	    }
	    break;
	case DATE:
	    for (int i = 0; i < numRows; i++)
	    {
		Date date = (Date) table[i][columnIndex];
		if (date == null)
		    key = new DateKey(new Date(0));
		else
		    key = new DateKey(date);

		heap.insert(key, table [i]);
	    }
	    break;
	default:
	    throw new IllegalArgumentException(
		"SortableTable: Sort: Type does not exist");
	}

	switch (order)
	{
	case ASCENDING:
	    for (int i = numRows - 1; i >=0; i--)
	    {
		table[i] = (Object []) heap.remove();
	    }
	    break;
	case DESCENDING:
	    for (int i = 0; i < numRows; i++)
	    {
		table[i] = (Object []) heap.remove();
	    }
	    break;
	default:
	    throw new IllegalArgumentException (
		"SortableTable: Sort: Sort order does not exist");
	}
    }


    /**
     * returns an Enumeration of the rows in the table.   The objects in the
     * enumeration are simply arrays of objects.  Note that nextElement will
     * return null if the next row in the enumeration is null...
     */
    public Enumeration rows()
    {
	return new RowEnumerator();
    }


    /**
     * Returns an enumeration of all the objects in the table.  It enumerates
     * row by row- i.e. all the columns of the first row are returned, all the
     * columns of the second row are returned, etc. Note that nextElement will
     * return null if the next element in the table is null....
     */
    public Enumeration elements()
    {
	return new TableEnumerator();
    }


    /** 
     * Returns a string representation of the table. All it does is write
     * the row and column indexes of each object and call the objects toString()
     * method, so be careful!
     */
    public String toString()
    {
	StringBuffer sb = new StringBuffer();

	// print the header
	sb.append("row\t");
	for (int i = 0; i < numColumns; i++)
	    sb.append(names [i] + "\t");
	sb.append("\n\n");

	for (int i = 0; i < numRows; i++)
	{
	    sb.append(i + "\t");
	    for (int j = 0; j < numColumns; j++)
		sb.append(table [i] [j] + "\t");
	    sb.append("\n");
	}	

	return sb.toString();
    }


    /**
     * returns an HTML representation of the table, for easy data checking of
     * handlers.
     */
    public String toHTML()
    {
	StringBuffer sb = new StringBuffer();
	String obj = "";
	
	sb.append("<TABLE BORDER=1>\n\t<TR>\n");
	
        // print the column names
	sb.append("\t\t<TH ALIGN=CENTER>Row</TH>\n");
	for (int i = 0; i < numColumns; i++)
	{
	    obj = "";
	    if (names [i] != null)
		obj = names [i];

	    sb.append("\t\t<TH ALIGN=CENTER>" + obj + "</TH>\n");
	}
	sb.append("\t</TR>\n");

        // print the column types
	sb.append("\t<TR>\n");
	sb.append("\t\t<TH ALIGN=CENTER>SQL Types</TH>\n");
	for (int i = 0; i < numColumns; i++)
	{
	    sb.append("\t\t<TH ALIGN=CENTER>");
	    sb.append(printSQLType(SQLTypes[i]));
	    sb.append("</TH>\n");
	}
	sb.append("\t</TR>\n");

	// print the table body
	for (int i = 0; i < numRows; i++)
	{
	    sb.append("\t<TR>\n");
	    sb.append("\t\t<TD ALIGN=CENTER>" + i + "</TD>\n");
	    for (int j = 0; j < numColumns; j++)
	    {
		obj = "";
		if (table [i] [j] != null)
		    obj = table [i] [j].toString();

		sb.append("\t\t<TD ALIGN=CENTER>" + obj + "</TD>\n");
	    }
	    sb.append("\t</TR>\n");
	}
	sb.append("</TABLE>");

	return sb.toString();
    }


    /** change an entire column of timestamps to a column of 
     * strings containing formatted dates.  The format is the default
     * given to us by java.text.SimpleDateFormat
     * @param table the table to update
     * @param column the column to update
     * @param format the format string to use
     * @see java.text.SimpleDateFormat
     */
    public void formatDateColumn(int column, String format) 
    {
	try
	{ 
	    // change the dates...
	    SimpleDateFormat sdf = new SimpleDateFormat(format);
	    for (int i = 0; i < getNumRows(); i++)
	    {
		Object obj = get(i, column);
		if (obj != null)
		{
	insert(sdf.format((java.util.Date)obj), i, column);
		}
	    }
	} catch (Exception e) 
	{
//	    Loco.log (0, "ERROR: DiscussionLib: couldn't format date: " + e);
	}
    }

    /** change nulls in a column to an empty string (&quot;&quot;)
     * @param table the table to update
     * @param column the column to update
     * @param alt the string to substitute
     */
    public void changeNullsToString(int column,	String alt) 
    {
	try
	{ 
	    for (int i = 0; i < getNumRows(); i++)
	    {
		Object obj = get(i, column);
		if (obj == null)
		{
		    insert(alt, i, column);
		}
	    }
	} catch (Exception e) {
//    Loco.log (0, "ERROR: DiscussionLib: couldn't substitute null: " + e);
	}
    }



    /**
     * loads a query specified by the string into the SortableTable
     * that's returned.  Will return null if there is a SQLException.
     * @see loadFromQuery(Connection conn, String query), which throws
     * an SQLException. 
     */
    public static SortableTable loadQuery(Connection conn, String query)
    {
	PreparedStatement stmt = null;
	ResultSet rs = null;
	SortableTable table = null;

	try
	{
	    stmt = conn.prepareStatement(query);
	    rs = stmt.executeQuery();
	    table = new SortableTable();
	    table.load(rs, true);
	}
	catch (SQLException e)
	{

	    printMsg(0, "ERROR: SortableTable: loadQuery: " +
		     e + ": Error #: " + e.getErrorCode() + ": SQLState: " + 
		     e.getSQLState() + " query: " + query);
	}
	finally
	{
	    try
	    {
		if (rs != null) rs.close();
		if (stmt != null) stmt.close();
	    } 
	    catch (SQLException e) {}
	}

        if (table == null || table.getNumRows() < 1)
            return null;
        else
            return table;
    }	


    /** loads a query specified by the string into the SortableTable that's
     * returned.
     * @throws SQLException, passed up from the load (ResultSet, boolean)
     * method.
     */
    public static SortableTable loadFromQuery(Connection conn, String query)
	throws SQLException
    {
	PreparedStatement stmt = null;
	ResultSet rs = null;
	SortableTable table = null;

	try
	{
	    stmt = conn.prepareStatement(query);
	    rs = stmt.executeQuery();
	    table = new SortableTable();
	    table.load(rs, true);
	    return table;
	}
	finally
	{
	    if (rs != null) rs.close();
	    if (stmt != null) stmt.close();
	}
    }	



    /** 
     * Updates a table in the database based on the information in the
     * SortableTable. Constructs a SQL statement of the form: <PRE>
     * UPDATE &lt;tableName&gt; 
     *        SET &lt;column n name&gt; = &lt;column n value&gt; 
     *        WHERE &lt;constraintcol 0&lt; = &lt;column n value&gt;</PRE>
     * and runs it for each row in the sortableTable.
     *
     * @param tableName the name of the table to update
     * @param constraintCols an array containing the names of the columns
     *        whose values should be used to constrain each update
     * @return the number of rows updated, or -1 if there was an error
     */
    public int doUpdate(Connection conn, String tableName, 
			String [] constraintCols)
    {
	if (conn == null) return -1;

	PreparedStatement stmt = null;
	int rows = 0;
	int presentRow = 0;
	int insert = 0;
	// this holds whether or ot each column should be part of the
	// update clause.  Columns are not part of the update clause
	// if they are in the constraint clause, or if they are null
	boolean [] toUpdate = new boolean [numColumns];

	for (int i = 0; i < numColumns; i++)
	{
	    if (names [i] != null && names[i].length() > 0)
	    {
		toUpdate [i] = true;
		for (int j = 0; j < constraintCols.length; j++)
		{
		    if (names[i].equals(constraintCols[j]))
		    {
			toUpdate[i] = false;
			break;
		    }
		}
	    }
	    else
	    {
		toUpdate [i] = false;
	    }
	}

	// construct the query
	StringBuffer query = new StringBuffer("UPDATE ");
	query.append(tableName);
	query.append(" SET ");
	for (int i = 0; i < numColumns; i++)
	{
	    if (toUpdate[i])
	    {
		query.append(names[i]);
		query.append(" = ?, ");
	    }
	}
	query.setLength(query.length() - 2);  // last comma
 	query.append(" WHERE ");
	for (int i = 0; i < constraintCols.length; i++)
	{
	    if (constraintCols[i] != null && constraintCols[i].length() > 0)
	    {
		query.append(constraintCols[i]);
		query.append(" = ? AND ");
	    }
	    query.setLength(query.length() - 5); // last and
	}

	// now run it
	try
	{
	    stmt = conn.prepareStatement(query.toString());
	    for (int i = 0; i < numRows; i++)
	    {
		insert = 1;
		presentRow = i;
		// set the values to update
		for (int j = 0; j < numColumns; j++)
		{
		    if (toUpdate[j])
		    {
//			org.locomotive.loco.Loco.log(8, "Row: " + i + " type for " + 
//					     insert + " = " + SQLTypes[j] + 
//					     " : " + table [i] [j]);
			if (table[i] [j] == null)
			{
			    stmt.setNull(insert++, SQLTypes[j]);
			}
			else if (SQLTypes[j] == Types.LONGVARCHAR)
			{
			    // ach- write the inputstream for the BLOB
			    String blob = table[i] [j].toString();
			    ByteArrayInputStream pbis = 
				new ByteArrayInputStream( blob.getBytes());
			    stmt.setAsciiStream(insert++, pbis, 
						blob.length());
			}
			else
			{
			    stmt.setObject(insert++, table[i][j]);
			}
		    }
		}
		// set the constraint values
		for (int j = 0; j < constraintCols.length; j++)
		{
		    if (constraintCols[j] != null && 
			constraintCols[j].length() > 0)
		    {
			Object obj = get(i, constraintCols[j]);
//			org.locomotive.loco.Loco.log(8, "Row: " + i + " cons type " + 
//					     insert + " = " + SQLTypes[j] + 
//					     " : " + table [i] [j]);

			if (obj == null)
			{
			    int index = getColumnIndex(constraintCols[j]);
			    if (index < 0) throw new IllegalArgumentException(
				"Constraint column is not in the table");

			    stmt.setNull(insert++, SQLTypes[index]);
			}
			else
			{
			    stmt.setObject(insert++, obj);
			}
		    }
		}
		rows += stmt.executeUpdate();
	    }
	}
	catch (SQLException e)
	{
	    rows = -1;
	    StringBuffer errmsg  = new StringBuffer();
	    errmsg.append("ERROR: SortableTable: doUpdate: ");
	    errmsg.append(e); errmsg.append(": Row ");
	    errmsg.append(presentRow); errmsg.append(": (");
	    for (int i = 0; i < numColumns; i++) 
	    {
		if (toUpdate[i]) 
		{
		    errmsg.append(table[presentRow][i]);
		    errmsg.append(", ");
		}
	    }
	    errmsg.setLength(errmsg.length() -2);
	    errmsg.append(") "); errmsg.append(query.toString());

	    printMsg(0, errmsg.toString());
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}

	return rows;
    }


    

    /** 
     * Inserts each row of the SortableTable into a table in the database.
     * Constructs a SQL statement of the form: <PRE>
     * INSERT INTO &lt;tableName&gt;
     *        (&lt;column 1 name&gt, &lt;column 2 name&gt, &lt;column n name&gt)
     *        VALUES 
     *        (&lt;column 1 value&gt, &lt;column 2 value&gt, &lt;column n value&gt)</PRE>
     * and runs it for each row in the sortableTable.
     */
    public int doInsert(Connection conn, String tableName)
    {
	if (conn == null) return -1;

	PreparedStatement stmt = null;
	int rows = 0;
	int presentRow = 0;
	int insert = 0;
	// this holds whether or ot each column should be part of the
	// update clause.  Columns are not part of the update clause
	// if they are in the constraint clause, or if they are null
	boolean [] toInsert = new boolean [numColumns];

	for (int i = 0; i < numColumns; i++)
	{
	    if (names [i] != null && names[i].length() > 0)
	    {
		toInsert [i] = true;
	    }
	    else
	    {
		toInsert [i] = false;
	    }
	}
	
	
	// first construct the query
	StringBuffer query = new StringBuffer("INSERT INTO ");
	query.append(tableName); query.append("( ");
	for (int i = 0; i < numColumns; i++)
	{
	    if (toInsert[i])
	    {
		query.append(names[i]);
		query.append(", ");
	    }
	}
	query.setLength(query.length() - 2);  // comma
	query.append(") VALUES ( ");
	for (int i = 0; i < numColumns; i++)
	{
	    if (toInsert[i])
	    {
		query.append("?, ");
	    }
	}
	query.setLength(query.length() - 2);  // comma
	query.append(")");
	
	try
	{
	    stmt = conn.prepareStatement(query.toString());
	    for (int i = 0; i < numRows; i++)
	    {
		insert = 1;
		presentRow = i;
		for (int j = 0; j < numColumns; j++)
		{
		    if (toInsert[j])
		    {
			if (table[i] [j] == null)
			{
			    stmt.setNull(insert++, SQLTypes[j]);
			}
			else if (SQLTypes[j] ==  Types.LONGVARCHAR)
			{ 
			    // ach- write the inputstream for the BLOB
			    String blob = table[i][j].toString();
			    ByteArrayInputStream pbis = 
				new ByteArrayInputStream( blob.getBytes());
			    stmt.setAsciiStream(insert++, pbis, 
						blob.length());
			}
			else
			{
			    stmt.setObject(insert++, table[i][j]);
			}
		    }
		}
		rows += stmt.executeUpdate();
	    }
	}
	catch (SQLException e)
	{
	    rows = -1;
	    StringBuffer errmsg  = new StringBuffer();
	    errmsg.append("ERROR: SortableTable: doInsert: ");
	    errmsg.append(e); errmsg.append(": Row ");
	    errmsg.append(presentRow); errmsg.append(": (");
	    for (int i = 0; i < numColumns; i++) 
	    {
		if (toInsert[i]) 
		{
		    errmsg.append(table[presentRow][i]); errmsg.append(", ");
		}
	    }
	    errmsg.setLength(errmsg.length() -2);
	    errmsg.append(") "); errmsg.append(query.toString());

	    printMsg(0, errmsg.toString());
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}

	return rows;
    }




    /** Deletes all the Rows in the SortableTable.  Uses every row value
     * as a constraints for the delete SQL statement, which is very
     * inefficient.  To set the columns to use as constraints explicitly
     * use the other doDelete method.
     * 
     * @param tableName the name of the table to update
     * @return the number of rows deleted, or -1 if there was an error
     */
    public int doDelete(Connection conn, String tableName)
    {
	return doDelete(conn, tableName, names);
    }


    /** deletes all the rows in the sortableTable from the named table.
     * Constructs a SQL statement of the form: <PRE>
     * DELETE FROM  &lt;tableName&gt;
     *        WHERE &lt;constraintCols 0&gt = &lt;column n value&gt;
     *          AND &lt;constraintCols 1&gt = &lt;column n value&gt;
     * </PRE>
     * where the constraintCol names specify the column values to use
     * as constraints.  The statement is run for every row in the table.
     * <P> Note also that if a row contains any null elements, then the
     * row will not be deleted, because nulls cannot be recognized as
     * valid constraints using the '=' operator in SQL.
     *
     * @param tableName the name of the table to update
     * @param constraintCols an array containing the names of the columns
     *        whose values should be used to constrain each update
     * @return the number of rows deleted, or -1 if there was an error
     */
    public int doDelete(Connection conn, String tableName, 
			String [] constraintCols)
    {
	if (conn == null) return -1;

	PreparedStatement stmt = null;
	int rows = 0;
	int presentRow = 0;
	int insert = 0;
	
	// construct the query
	StringBuffer query = new StringBuffer("DELETE FROM ");
	query.append(tableName); query.append(" WHERE ");
	for (int i = 0; i < constraintCols.length; i++)
	{
	    if (constraintCols[i] != null && constraintCols[i].length() > 0)
	    {
		query.append(constraintCols[i]);
		query.append(" = ? AND ");
	    }
	}
	query.setLength(query.length() - 5);  // last AND

	try
	{
	    stmt = conn.prepareStatement(query.toString());
	    for (int i = 0; i < numRows; i++)
	    {
		insert = 1;
		presentRow = i;
		// set the constraint values
		for (int j = 0; j < constraintCols.length; j++)
		{
		    if (constraintCols[j] != null && 
			constraintCols[j].length() > 0)
		    {
			Object obj = get(i, constraintCols[j]);
			if (obj == null)
			{
			    int index = getColumnIndex(constraintCols[j]);
			    if (index < 0) throw new IllegalArgumentException(
				"Constraint column is not in the table");

			    stmt.setNull(insert++, SQLTypes[index]);
			}
			else
			{
			    stmt.setObject(insert++, obj);
			}
		    }
		}
		rows += stmt.executeUpdate();
	    }
	}
	catch (SQLException e)
	{
	    rows = -1;
	    StringBuffer errmsg  = new StringBuffer();
	    errmsg.append("ERROR: SortableTable: doDelete: ");
	    errmsg.append(e); errmsg.append(": Row ");
	    errmsg.append(presentRow); errmsg.append(": (");
	    for (int i = 0; i < numColumns; i++) 
	    {
		if (constraintCols[i] != null && 
			constraintCols[i].length() > 0)
		{
		    errmsg.append(table[presentRow][i]); errmsg.append(", ");
		}
	    }
	    errmsg.setLength(errmsg.length() -2); errmsg.append(") ");
	    errmsg.append(query.toString());

	    printMsg(0, errmsg.toString());
	}
	finally
	{
	    try { if (stmt != null) stmt.close(); } 
	    catch (SQLException e) {}
	}

	return rows;
    }
	


    /** takes an int from java.sql.Types and returns it's name */
    public static String printSQLType(int SQLType)
    {
	String name = null;
	switch (SQLType)
	{
	case Types.BIGINT:        name = "BIGINT"; break;
	case Types.BINARY:        name = "BINARY"; break;
	case Types.BIT:           name = "BIT"; break;
	case Types.CHAR:          name = "CHAR"; break;
	case Types.DATE:          name = "DATE"; break;
	case Types.DOUBLE:        name = "DOUBLE"; break;
	case Types.FLOAT:         name = "FLOAT"; break;
	case Types.INTEGER:       name = "INTEGER"; break;
	case Types.LONGVARBINARY: name = "LONGVARBINARY"; break;
	case Types.LONGVARCHAR:   name = "LONGVARCHAR"; break;
	case Types.NULL:          name = "NULL"; break;
	case Types.NUMERIC:       name = "NUMERIC"; break;
	case Types.REAL:          name = "REAL"; break;
	case Types.SMALLINT:      name = "SMALLINT"; break;
	case Types.TIME:          name = "TIME"; break;
	case Types.TIMESTAMP:     name = "TIMESTAMP"; break;
	case Types.TINYINT:       name = "TINYINT"; break;
	case Types.VARBINARY:     name = "VARBINARY"; break;
	case Types.VARCHAR:       name = "VARCHAR"; break;
	default:                  name = "-UNKNOWN-"; break;
	}
	
	return (name + " (# " + SQLType + ")");
    }


    /** print a message to the log, or, if the log is null, to standard out
     */
    private static void printMsg(int logLevel, String msg)
    {
/*  Loco should not be ref'ed in this file.  will fix later.
	try 
	{
	    org.locomotive.loco.Loco.log(logLevel, msg);
	    
	} catch (NullPointerException ne)
	{
	    System.err.println(msg);
	}

*/
    }
	

/*=============== INNER CLASSES ====================*/



    /**
     * This class enumerates through all the rows in the table
     */
    private class RowEnumerator implements Enumeration
    {
	int rowIndex;  // one less than the first row
	
	/** constructor */
	RowEnumerator()
	{
	    rowIndex = -1;
	}
	
	/** Returns true if there are more rows, returns false otherwise */
	public boolean hasMoreElements()
	{
	    return (rowIndex < numRows - 1);
	}
	
	/** Returns the next row of the SortableTable */
	public Object nextElement()
	{
	    if (rowIndex++ == numRows)
		throw new NoSuchElementException("SortableTable: no more rows");
	    return table [rowIndex];
	}
	
    }  // end of RowEnumerator
    
    
    
    /**
     * This class enumerates through all the objects in the table
     */
    private class TableEnumerator implements Enumeration
    {
	int rowIndex;
	int columnIndex;

	/** constructor */
	TableEnumerator()
	{
	    rowIndex    = 0;
	    columnIndex = -1;
	}
	
	/** Returns true if there are more objects, returns false otherwise */
	public boolean hasMoreElements()
	{
	    if (rowIndex < numRows - 1)
		return true;
	    else 
		return (columnIndex < numColumns - 1);
	}
	
	/** Returns the next object in the SortableTable, going through each
	 * column in each row.  See SortableTable.elements()
	 */
	public Object nextElement()
	{
	    if (rowIndex + 1 == numRows)
		throw new NoSuchElementException(
		    "SortableTable: no more elements");
	    if (columnIndex < numColumns - 1)
		columnIndex ++;
	    else
	    {
		columnIndex = 0;
		rowIndex ++;
	    }
	    return table [rowIndex] [columnIndex];
	}

    }  // end of TableEnumerator





} // end of SortableTable class

