/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util;

import java.util.*;
import java.text.*;
import java.io.StringWriter;
import java.io.PrintWriter;

/**
 * Utilities for manipulating strings. Should be relatively robust.
 *
 * <p>
 * A simple example:
 *
 * <xmp>
 * String s = "hello \"there\" you\\me"; // produces: hello "there" you\me
 *
 * System.out.println(encodeString(s));
 * // produces: ["hello \"there\" you\\me"] (less the [ and ] brackets)
 *
 * Vector v = new Vector();
 * v.addElement(s);
 * v.addElement(s);
 *
 * System.out.println(decodeStrings(encodeStrings(v)));
 * // produces: [hello "there" you\me, hello "there" you\me]
 *
 * </xmp>
 *
 * Also, there are many methods of the form, "get...()", for parsing
 * values in a specific format or of a specific type out of a string.
 * These methods are analogous to the methods of the same name in the
 * FormData class.  Refer to the documentation in that class for an
 * understanding of how to use these get...() methods.
 * Also, @see ParseStringException.
 */
public class StringUtil
{
    /*---------------------- general String utilities ---------------------*/
    
    /**
     * Takes a string that represents a list of strings and parses them out.
     * Supports using the "\" operator to include "'s in a string, like java.
     * Any non-terminated strings should be kept up to the end of the input
     * string (i.e., using ['s and ]'s as my meta string characters,
     * ["stringone", "Stringtwo", "stringthree] will give a vector of
     * [stringone], [Stringtwo], and [stringthree]. Commas aren't needed in
     * between strings. It just looks for the next ["]. So
     * ["stringone""stringtwo""stringthree"] works just as well. Remember, the
     * ['s I just put in there to be clear about what the strings are... those
     * are treated as normal characters.
     * The "\" character allows anything after itself (\'s at the end of an
     * input string are kept as characters.) So
     * ["stringone" "\"stringtwo"garbage235478"\"string\"three\""] comes out as
     * [stringone] ["stringtwo] and ["string"three"] ... [\\] gives [\], etc.
     * <p> Why? It allows some handiness. Like if you want to save a bunch
     * of strings in one field in a database and want to allow quotes and
     * slashes and what not.
     */
    public static Vector decodeStrings(String s)
    {
	if (s == null) return(null);
	if (s.equals("")) return(null);

	int ssize = s.length();
	Vector v = new Vector();
	StringBuffer sb = new StringBuffer();

	int i = 0;

	while (i < ssize)
	{
	    while ((i < ssize) && !(s.charAt(i) == '\"'))
	    {
		i++;
	    }

	    i++;
	    sb.setLength(0);

	    while ((i < ssize) && !(s.charAt(i) == '\"'))
	    {
		if (s.charAt(i) == '\\')
		{
		    if (i + 1 < ssize)
		    {
			i++;
			sb.append(s.charAt(i));
			i++;
		    }
		    else
		    {
			sb.append(s.charAt(i));
			i++;
		    }
		}
		else
		{
		    sb.append(s.charAt(i));
		    i++;
		}
	    }

	    i++;

	    v.addElement(sb.toString());
	}

	return(v);
    }

    /**
     * encode strings with no seperator (saves some space.) Same thing
     * as encodeStrings(s, "") (call that for even more efficiency :-) )
     * This will decode just fine. It just doesn't sacrifice anything for
     * readability.
     */
    public static String encodeStringsEfficiently(Enumeration e)
    {
	return(encodeStrings(e, ""));
    }

    /**
     * encode strings with no seperator (saves some space.) Same thing
     * as encodeStrings(s, "") (call that for even more efficiency :-) )
     * This will decode just fine. It just doesn't sacrifice anything for
     * readability.
     */
    public static String encodeStringsEfficiently(Vector v)
    {
	return(encodeStrings(v.elements(), ""));
    }

    /**
     * encode strings with the default seperator ", "... i.e. gives you
     * "stringone", "string two", "etc..."
     */
    public static String encodeStrings(Enumeration e)
    {
	return(encodeStrings(e, ", "));
    }

    /**
     * encode strings with the default seperator ", "... i.e. gives you
     * "stringone", "string two", "etc..."
     */
    public static String encodeStrings(Vector v)
    {
	return(encodeStrings(v.elements(), ", "));
    }

    /**
     * Take a vector  of strings and make a big string of encoded strings.
     * This can then be decoded with decodeStrings.
     * Seperator is what comes between each string. Nothing is needed for
     * parsing to work correctly, but adding something like ", " between each
     * string may improve readability. This is NOT the special seperator used
     * by the decoder - this can not be changed (as of now) and is ". This is
     * what comes between the string units. So "hi", "there" has a seperator
     * of [, ] (less the brackets.) The "'s around those words have to be
     * there for the decoder to decode them correctly.
     * An empty vector gives you the empty string back. A vector with
     * one string element which is the empty string gives you the string [""]
     * back (less the [ and ]). A null vector or other errors (like
     * non-strings in the vector) gives you null.
     */
    public static String encodeStrings(Vector v, String seperator)
    {
	return(encodeStrings(v.elements(), seperator));
    }

    /**
     * Take an enumeration of strings and make a big string of encoded strings.
     * This can then be decoded with decodeStrings.
     * Seperator is what comes between each string. Nothing is needed for
     * parsing to work correctly, but adding something like ", " between each
     * string may improve readability. This is NOT the special seperator used
     * by the decoder - this can not be changed (as of now) and is ". This is
     * what comes between the string units. So "hi", "there" has a seperator
     * of [, ] (less the brackets.) The "'s around those words have to be
     * there for the decoder to decode them correctly.
     * An empty enumeration gives you the empty string back. A enumeration with
     * one string element which is the empty string gives you the string [""]
     * back (less the [ and ]). A null enumeration or other errors (like
     * non-strings in the enumeration) gives you null.
     */
    public static String encodeStrings(Enumeration e, String seperator)
    {
	if (e == null) return(null);

	StringBuffer b = new StringBuffer();
	String s = null;

	if (e.hasMoreElements())
	{
	    try
	    {
		s = (String) e.nextElement();
	    }
	    catch (ClassCastException cce)
	    {
		return(null);
	    }
	}

	if (s != null) b.append(encodeString(s));

	while(e.hasMoreElements())
	{
	    try
	    {
		s = (String) e.nextElement();
	    }
	    catch (ClassCastException cce)
	    {
		return(null);
	    }

	    if (s != null)
		b.append(seperator + encodeString(s));
	}

	if (b.length() != 0)
	{
	    return(b.toString());
	}
	else return("");
    }

    /**
     * Take a string and replace all the "'s with \"'s and the \'s with
     * \\'s and surrounds the string with "'s which is necessary for
     * a concatenated list of strings to be sucessfully decoded.
     */
    public static String encodeString(String s)
    {
	if (s == null) return(null);

	StringBuffer b = new StringBuffer(s.length());

	char[] c_ra = s.toCharArray();
	int max = c_ra.length;
	char curchar;

	for (int i = 0; i < max; i++)
	{
	    curchar = c_ra[i];

	    if (curchar == '\\')
	    {
		b.append("\\\\");
	    }
	    else if (curchar == '\"')
	    {
		b.append("\\\"");
	    }
	    else b.append(curchar);
	}

	return("\"" + b.toString() + "\"");
    }

    /*-------------------------- HTML utilities -------------------------*/
    
    private static char hexToChar (char c1, char c2)
    {
	int first = 0;
	int second = 0;
	
	first = Character.digit (c1, 16);
	second = Character.digit (c2, 16);
	
	if ((first == -1) || (second == -1))
	{
	    return ((char) -1);
	}
	return ((char) (16 * first + second));
    }

    /**
     * takes CGIEncoded strings (normally from the contents of the 
     * STDIN in a POST) and unquotes them
     */
    public static String CGIUnescape (String s)
    {
	char ca[] = new char[s.length()];
	s.getChars(0, s.length(), ca, 0);

	StringBuffer outputsb = new StringBuffer();
	
	for (int i = 0; i < ca.length; i++)
	{
	    if (ca[i] == '+')
	    {
		outputsb.append(" ");
	    }
	    else if ((ca[i] == '%') && (ca.length > (i + 2) ))
	    {
		outputsb.append(hexToChar(ca[i+1], ca[i+2]));
		i = i + 2;
	    }
	    else
	    {
		outputsb.append(ca[i]);
	    }
	}
	    
	return (outputsb.toString());
    }

    /**
     * Decodes a string encoded with URLEncode. 
     * @param s - the URL encoded string.
     * @returns - a decoded version of the input
     */
    public static String URLDecode (String s)
    {
	char input[] = s.toCharArray();
	
	StringBuffer output = new StringBuffer();

	for (int i = 0; i < input.length; i++)
	{
	    if ((input[i] == '%') && (input.length > (i+2)))
	    {
		output.append ( hexToChar (input[i+1], input[i+2]));
		i = i + 2;
	    }
	    else
	    {
		output.append (input[i]);
	    }
	}
	return (output.toString());
    }

    /**
     * takes an arbitrary string, returns a URLEncoded String 
     * everything but letters and numbers are escaped via the "%" sign.  
     * The result is suitable for returning to a browser.
     * @param s  -  the input string.
     * @returns  - a URL Encoded string.
     */
    public static String URLEncode (String s)
    {
	char input[] = s.toCharArray();
	StringBuffer output = new StringBuffer();
	char c; 
	
	for (int i = 0; i < input.length; i++)
	{
	    c = input [i];
	    if ( Character.isLetter (c) || Character.isDigit(c))
	    {
		output.append(c);
	    }
	    else
	    {
		output.append("%" + Integer.toString ( (int) c, 16));
	    }
	}
	return (output.toString());
    }

    /**
     * HTML quotes all values in the input hash.
     */
    public static void htmlQuoteHash (Hashtable ht)
    {
	Enumeration e = ht.keys();
	
	String keyname;
	String value;
	while (e.hasMoreElements())
	{
	    keyname = (String) e.nextElement();
	    value = (String) ht.get(keyname);
	    if (value != null) 
	    {
		ht.put (keyname, htmlQuoteString (value));
	    }
	}
    }

    /**
     * quotes any string for interpreted display of an HTML page by a browser.
     * For instance an &amp; will be quoted to &amp;amp;
     */
    public static String htmlQuoteString (String s)
    {
	if (s == null)
	    return ("");

	char input[] = s.toCharArray();
	
	StringBuffer output = new StringBuffer();

	for (int i = 0; i < input.length; i++)
	{
	    if (input[i] == '\"')
	    {
		output.append("&quot;");
	    }
	    else if (input[i] == '<')
	    {
		output.append("&lt;");
	    }
	    else if (input[i] == '>')
	    {
		output.append("&gt;");
	    }
	    else if (input[i] == '&')
	    {
		output.append("&amp;");
	    }
	    else
	    {
		output.append(input[i]);
	    }
	}
	    
	return (output.toString());
    }

    
    /**
     * Shell escapes a given string.  Currently remove the following: 
     * <BLOCKQUOTE><B>|&nbsp;&nbsp;&lt;&nbsp;&nbsp;&gt;&nbsp;&nbsp;!
     * &nbsp;&nbsp;$&nbsp;&nbsp;,&nbsp;&nbsp;\</B></BLOCKQUOTE>
     */
    public static String shellEscape(String s)
    {
	if (s == null) return ("");

	String badchars = "|<>!$,\\";
	StringBuffer sb = new StringBuffer();
	char [] input = s.toCharArray();
	for (int i = 0; i < input.length; i++)
	{
	    if (badchars.indexOf(input[i]) < 1)
	    {
		sb.append(input[i]);
	    }
	}
	
	return (sb.toString());
    }
		


    /**
     * A utility function - given the exception, give us a string that
     * contains the stack trace for it.
     */
    public static String exceptionStackTrace(Throwable e)
    {
	StringWriter sw = new StringWriter();
	e.printStackTrace(new PrintWriter(sw));
	return("\n" + sw.toString());
    }

    /**
     * A utility function to trim a string to a given length.
     * null strings, strings that are less than the length, etc are
     * returned untouched. max_length is the maximum length, NOT the
     * maximum index.
     */
    public static String clip(String s, int max_length)
    {
	if (s == null) return(s);

	// remember, substring's end-index is EXclusive!
	if (s.length() > max_length)
	    return(s.substring(0, max_length));
	else return(s);
    }

    /*----------- utilities for parsing data with special formats -----------*/
        
    /**
     * tells whether this string contains an int.  If you want the int value,
     * don't use this method, just use Integer.parseInt().
     */
    public static boolean isInt (String string_value)
    {
	try {
	    Integer.parseInt (string_value);
	    return true;
	}
	catch (NumberFormatException ex)
	{
	    return false;
	}
    }
    
    
    /**
     *  trims string_containing_digits and returns a string containing only
     *  digits. 
     * @exception org.locomotive.util.ParseStringException
     * If there are non-digits (other than leading and
     *  trailing whitespace), then throws ParseStringException 
     *  with an error code of DOESNT_FIT_TYPE.
     */
    public static String getDigits (String string_containing_digits)
	throws ParseStringException
    {
	return getSpecialChars ("digits", string_containing_digits,
				0, string_containing_digits.length(), true);
    }
    
    
    /**
     * tells whether string_containing_digits contains only digits.
     * if it's null or empty, returns true.
     * Does not throw any ParseStringException.
     */
    public static boolean isDigits (String string_containing_digits)
    {
	if (string_containing_digits == null || 
	    string_containing_digits.equals(""))
	    return true;

	char[] ch = string_containing_digits.toCharArray();
	for (int i=0; i < ch.length; i++)
	    if (!Character.isDigit(ch[i]))
		return false;
	return true;
    }
    
    
    /**
     *  if this field contains non-digits, then the non-digits will be removed.
     * @exception org.locomotive.util.ParseStringException
     *  Also, if it does contain non-digits, and warn_non_digits is true, then
     *  throws ParseStringException with an error code of DOESNT_FIT_TYPE.
     */
    public static String getDigits (String string_containing_digits,
				    int minimum_length, int maximum_length,
				    boolean warn_non_digits)
	throws ParseStringException
    {
	return getSpecialChars ("digits", string_containing_digits,
				minimum_length, maximum_length,
				warn_non_digits);
    }

    
    /**
     * tells whether string_containing_alphanum contains only alphanumerics.
     * If it's null or empty, returns true.
     * Does not throw any ParseStringException.
     */
    public static boolean isAlphanum (String string_containing_alphanum)
    {
	if (string_containing_alphanum == null ||
	    string_containing_alphanum.equals(""))
	    return true;
	
	char[] ch = string_containing_alphanum.toCharArray();
	for (int i=0; i < ch.length; i++)
	    if (!Character.isLetterOrDigit(ch[i]))
		return false;
	return true;
    }
    
    
    /**
     *  trims string_containing_alphanum and returns a string containing only
     *  alphanumerics. 
     * @exception org.locomotive.util.ParseStringException
     * If there are non-alphanumerics, then 
     *  throws ParseStringException with an error code of DOESNT_FIT_TYPE.
     */
    public static String getAlphanum (String string_containing_alphanum)
	throws ParseStringException
    {
	return getSpecialChars ("letters or digits", 
				string_containing_alphanum,
				0, string_containing_alphanum.length(), true);
    }
    
    
    /**
     * if this field contains non-alphanumerics, then they will be removed.
     * @exception org.locomotive.util.ParseStringException
     * If it does contain non-alphanumerics, and warn_non_alphanum is true,
     * then throws ParseStringException with an error code of DOESNT_FIT_TYPE.
     */
    public static String getAlphanum (String string_containing_alphanum,
				      int minimum_length, int maximum_length,
				      boolean warn_non_alphanum)
	throws ParseStringException
    {
	return getSpecialChars ("letters or digits",
				string_containing_alphanum, 
				minimum_length, maximum_length,
				warn_non_alphanum);
    }
    
    
    /**
     * This method trims the leading and trailing whitespace in
     * a string, and returns its value. 
     * @exception org.locomotive.util.ParseStringException
     * If the field is shorter than the minimum length, or
     * if it is longer than the maximum length, throws a ParseStringException
     * with an error code of TOO_SHORT or TOO_LONG, respectively.
     * However, it will NOT throw an exception if there is whitespace in the 
     * field.
     */
    public static String getTrimmedString (String string_value,
					   int minimum_length,
					   int maximum_length) 
	throws ParseStringException
    {
	return getSpecialChars ("trimmed_string", string_value,
				minimum_length, maximum_length,
				false); 
    }
    
    
    
    static String getSpecialChars (String char_type, String string_value,
				   int minimum_length, int maximum_length,
				   boolean warn_bad_chars)
	throws ParseStringException
    {
	if (string_value == null)
	{
	    throw new ParseStringException (ParseStringException.IS_NULL);
	}
	    
	char[] chars = string_value.toCharArray();
	char[] new_value = new char[chars.length];
	int counter = 0;
	    
	boolean found_bad_char = false;
	    
	if (char_type.equals("digits"))
	{
	    for (int i=0; i < chars.length; i++)
		if (Character.isDigit (chars[i]))
		    new_value[counter++] = chars[i];
		else if (warn_bad_chars)
		    throw new ParseStringException (
			ParseStringException.DOESNT_FIT_TYPE);
	}
	else if (char_type.equals("letters or digits"))
	{
	    for (int i=0; i < chars.length; i++)
		if (Character.isLetterOrDigit (chars[i]))
		    new_value[counter++] = chars[i];
		else if (warn_bad_chars)
		    throw new ParseStringException (
			ParseStringException.DOESNT_FIT_TYPE);
	}
	else if (char_type.equals("trimmed_string"))
	{
	    new_value = chars;
	    counter = chars.length;
	}
	    
	    
	if (counter < minimum_length)
	{
	    throw new ParseStringException (ParseStringException.TOO_SHORT);
	}
	else if (counter > maximum_length)
	{
	    throw new ParseStringException (ParseStringException.TOO_LONG);
	}
	    
	return new String(new_value, 0, counter);
    }
    
    
    /**
     * This method removes all whitespace in string_value.
     * It will NOT throw an error if there is whitespace in the field.
     */
    public static String removeWhitespaces (String string_value)
    {
	if (string_value == null || string_value.equals(""))
	{
	    return string_value;
	}
	else
	{
	    char[] chars = string_value.toCharArray();
	    char[] new_value = new char[chars.length];
	    int counter = 0;
		
	    for (int i=0; i < chars.length; i++)
		if (!Character.isSpaceChar (chars[i]))
		    new_value[counter++] = chars[i];
	    
	    return new String(new_value, 0, counter);
	}
    }
    

    /**
     * tries to parse a field as a Date in one of the following formats:
     * "MMM d, yyyy", "MM-d-yy", "MM-d-yyyy", "MM/d/yy",
     * "MM/d/yyyy", "d.MM.yy", "d.MM.yyyy"
     * @return If it could not parse the field as a Date, will return null.
     */  
    public static java.util.Date getDate (String string_containing_date)
    {
	if (string_containing_date == null ||
	    string_containing_date.equals("")) 
	{
	    return null;
	}
	    
	string_containing_date = string_containing_date.trim();
	    
	String possible_formats[] = {"MMM d, yyyy", "MM-d-yy", "MM-d-yyyy",
				   "MM/d/yy", "MM/d/yyyy", "d.MM.yy",
				   "d.MM.yyyy"}; 
	    
	for (int i=0; i < possible_formats.length; i++)
	{
	    try {
		SimpleDateFormat format = 
		    new SimpleDateFormat (possible_formats [i]);
		format.setLenient(true);
		
		java.util.Date result_date = 
		    format.parse (string_containing_date);
		
		// One final check: since SimpleDateFormat.parse() will even
		// parse days > 31, we should check to make sure that the
		// resulting date really is the right date, by checking that
		// the day-of-month of the resulting date is contained within
		// string_containing_date.
		
		if (result_date != null &&
		    (string_containing_date.indexOf
		     (new SimpleDateFormat("d").format (result_date)) != -1))
		{
		    return result_date;
		}
	    }
	    catch (ParseException exception) { }
	}
	
	return null;
    }
    
    
    /**
     * tries to parse the three fields, month_field, day_field, and
     * year_field, and create a Date out of them. 
     * The month_field argument can be either a number from 1-12 or at least
     * the first three letters of the name of a month, upper or lower case.
     * @exception org.locomotive.util.ParseStringException
     * If one of the three fields is null, will throw a ParseStringException.
     */
    public static java.util.Date getDate (String month_field, String day_field,
					  String year_field) 
	throws ParseStringException
    {
	if (month_field == null || day_field == null || year_field == null)
	{
	    throw new ParseStringException (ParseStringException.IS_NULL);
	}
	
	month_field = month_field.trim();
	day_field = day_field.trim();
	year_field = year_field.trim();
	
	if (Character.isLetter (month_field.charAt (0)) &&
	    month_field.length() > 2)
	{
	    return getDate (month_field.substring(0,3) + " " + 
			    day_field + ", " + year_field);
	}
	else if (Character.isDigit (month_field.charAt (0)))
	{
	    return getDate (month_field + "/" + day_field + "/" + year_field);
	}
	else
	{
	    throw new ParseStringException (
		ParseStringException.DOESNT_FIT_TYPE);
	}
    }
    
    
    /**
     * Will tell you whether the email address is a valid SMTP email address.
     * The email address should have at least one '@', at least one '.' after
     * then '@', and should have only alphanumeric characters or characters
     * in: [+@.'%-_!] (not including the brackets [ ]).
     * (This method was copied largely from StringConstraint.checkForIsEmail())
     */
    public static boolean isEmailAddress (String email_address)
    {
	if (email_address == null || email_address.length() < 6)
	{
	    return false;
	}
	
	// start real basic here with "username@domain.tld"
	// this may need to get squirrelly
	    
	// check characters first
	char c;
	char[] email_address_chars = email_address.trim().toCharArray();
	for (int i = 0; i < email_address_chars.length; i++)
	{
	    c = email_address_chars[i];
		
	    if (! ( Character.isLetterOrDigit(c) ||
		    (c == '+') || (c == '@') || (c == '.') || (c == '\'') ||
		    (c == '%') || (c == '-') || (c == '_') || (c == '!') ))
	    {
		return false;
	    }
	}
	
	
	// make sure we have userpart@hostpart, there can be only one "@"
	StringTokenizer st = new StringTokenizer (email_address, "@");

	if (st.countTokens() != 2)
	    return false;
	
	// domainname should have at least one period, NOT at the beginning
	// or end.
	st.nextToken();
	String domainname = st.nextToken();

	// apparently, Java's StringTokenizer class will IGNORE ".." if
	// the token separator is '.'.  In other words, it won't even think
	// that there are tokens there.  However, DNS doesn't allow domain
	// names with two dots next to each other.

	if (domainname.startsWith(".") || domainname.endsWith(".") ||
	    domainname.indexOf("..") > -1)
	    return false;

	StringTokenizer ost = new StringTokenizer (domainname, ".");

	if (ost.countTokens() < 2)
	{
	    return false;
	}
	
	while (ost.hasMoreTokens())
	{
	    String next_token = ost.nextToken();
	    if (next_token.length() == 0)
		return false;
	}
	
	return true;
    }

    /**
     * Given an array of strings, convert it to one string with the
     * separator as given.
     * Converts separator to "" if it is null.
     * @returns ("" + null) if a null string_array is fed in,
     * "" if a zero-length string array is fed in.
     */
    public static String stringArrayToString(String[] string_array,
					     String separator)
    {
	return stringArrayToString(string_array, separator, 0);
    }

    /**
     * Given an array of strings, convert it to one string with the
     * separator as given.
     * Converts separator to "" if it is null. start_index is set to 0 if
     * less than 0.
     * @returns ("" + null) if a null string_array is fed in,
     * "" if a zero-length string array is fed in or if start_index is
     * greater than string_array.length - 1
     */
    public static String stringArrayToString(String[] string_array,
					     String separator,
					     int start_index)
    {
	if (string_array == null) return("" + null);
	if (string_array.length == 0) return("");
	
	if (separator == null) separator = "";

	int length = string_array.length;
	if (start_index < 0) start_index = 0;

	StringBuffer s = new StringBuffer();

	if (start_index < length)
	    s.append(string_array[start_index]);
	
	for (int i = start_index + 1; i < length; i++)
	{
	    s.append(separator);
	    if (string_array[i] != null)
		s.append(string_array[i]);
	}

	return(s.toString());
    }
    
    
    /**
     * Given an array of ints, convert it to one string with the
     * separator as given.
     * @returns ("" + null) if a null array is fed in,
     * "" if a zero-length array is fed in,
     * converts separator to "" if it is null.
     */
    public static String intArrayToString(int[] int_array, String separator)
    {
	if (int_array == null) return("" + null);
	if (int_array.length == 0) return("");
	
	if (separator == null) separator = "";

	int length = int_array.length;

	StringBuffer s = new StringBuffer();

	if (length > 0)
	    s.append(int_array[0]);
	
	for (int i = 1; i < length; i++)
	{
	    s.append(separator);
	    s.append(int_array[i]);
	}

	return(s.toString());
    }
    
    
    /**
     * This is the opposite of stringArrayToString().
     * Given a string containing a separator, break it into an array of 
     * strings and return that.
     */
    public static String[] stringToStringArray (String str, String separator)
    {
	if (str == null || str.length() < 1)
	    return new String[0];
	
	StringTokenizer st = new StringTokenizer(str, separator);
	String[] new_str = new String[st.countTokens()];
	
	int i = 0;
	while (st.hasMoreTokens())
	{
	    new_str[i++] = (String) st.nextToken();
	}
	
	return new_str;
    }		
    
    /**
     * This is the opposite of intArrayToString().
     * Given a string containing a separator, break it into an array of 
     * strings and return that.
     */
    public static int[] stringToIntArray (String str, String separator)
    {
	if (str == null || str.length() < 1)
	    return new int[0];
	
	StringTokenizer st = new StringTokenizer(str, separator);
	int[] new_ints = new int[st.countTokens()];
	
	int i = 0;
	while (st.hasMoreTokens())
	{
	    try {
		new_ints[i] = Integer.parseInt((String)st.nextToken());
		i++;
	    }
	    catch (NumberFormatException ex)
	    {
		new_ints[i++] = -1;
	    }
	}
	
	return new_ints;
    }		
    
}
