/**
 * The AlarmThread is a class which can manage a large
 * number of "alarm" events, processing each event as its alarm
 * goes off.  The alarm events must trigger within a period of time
 * which is a global parameter of the AlarmThread.
 * <P>
 * The AlarmThread is implemented as an array of Vectors.
 * Specifically, let's say, for example, that we created
 * an AlarmThread which could store alarms which triggered within
 * one hour.  In this case, the array would contain 3600 Vectors.
 * Each Vector would contain all the alarm events that were due
 * to trigger in one particular second.
 * <P>
 * Let's take an example.  Let's say you create a new alarm event,
 * which is supposed to go off in half an hour.  Then,
 * you register this new event with the AlarmThread.
 * The AlarmThread will calculate which of the 3600 Vectors to put
 * your new event by using the following formula:
 * 
 * <PRE>(System.currentTimeMillis()/1000 + 1800) % 3600</PRE>
 * 
 * In other words, the array of Vectors simply takes
 * (the alarm time of your event, expressed as the number
 * of seconds since January 1, 1970) modulo 3600, to decide
 * which "bucket" to put your event.
 * <P>
 * The AlarmThread processes alarms by simply walking down
 * the array.  In other words, the AlarmThread may wake up every
 * 10 seconds, for example, and process
 * all the alarm events which have triggered in the last 10 seconds.
 * In order to find all of these alarm events,
 * the AlarmThread simply retrieves the Vector for each second,
 * again by using the formula T % 3600, where T stands for that
 * particular second.
 * <P>
 * Notice that as long as all the alarm events are triggered within
 * 3600 seconds, they will be processed at exactly the time that
 * they are supposed to be triggered.  Therefore, one of the
 * strict requirements of this implementation is that the alarms
 * must all trigger within a specific period of time, which is the
 * SECONDS_IN_PERIOD variable.
 * <P>
 * <B>Discussion of alternative implementations:</B>
 * <P>
 * Several other people have suggested some good alternative
 * ways to implement the AlarmThread.  One such alternative is to
 * implement the list of alarm events as a linked list instead of
 * an array of Vectors.  To find a particular alarm
 * event, we could use either a binary tree that is sorted by
 * absolute time, or we can use a Hashtable which finds alarm
 * events based on their trigger time.  Then, inserting and
 * removing alarm events would be an inexpensive operation.  
 * <P>
 * These alternatives have much to recommend them.  Their major advantage
 * over my implementation is that they may save some amount of memory.
 * However, one has to remember that linked lists, Hashtables,
 * or binary trees also take up additional memory, as well as
 * more memory management overhead.  In general, allocating many small
 * objects, such as linked list nodes or binary tree nodes, use up more
 * memory management time than large things like Vectors.
 * In addition, deleting items always requires some execution time,
 * particularly when searching for the right item.  In the
 * AlarmThread class itself, I don't delete any alarm events until
 * they are actually triggered; and then, I delete entire blocks (Vectors)
 * of alarm events at a time.  In the subclasses of
 * AlarmThread, such as ExpirationThread and AutoSaveThread, I simply
 * mark each alarm event as "deleted" if it is supposed to be
 * deleted, and, later, when I encounter these already-deleted
 * entries, I simply ignore them.
 * <P>
 * Another disadvantage of the alternative implementations is that
 * there is more potential for memory leakage, with more complicated
 * algorithms, especially if I'm using data structures that I write myself,
 * instead of using proven data structures such as Java 2.0 Collections or
 * Rogue Wave or Objectspace.  Another way to say this is that
 * I will need to spend more time reinventing my own data structures
 * and testing them just to make the AlarmThread work, which I feel
 * is not a good way to leverage my own time.  In fact, I think that it is
 * reasonable to revisit the implementation of the AlarmThread after Java 2.0
 * is officially released for Linux and FreeBSD, and other popular platforms,
 * so that we can leverage the Collections Framework.
 * <P>
 * I believe that the AlarmThread can be of general use in the
 * future.  However, currently it is specific to the GlobalDBCache.
 * It is used to implement the automatic expiration and auto-saving
 * features in the GlobalDBCache.  Since it is used in this specific
 * function, the AlarmThread has some special properties beyond
 * simple management of alarm events.
 * <UL>
 * <LI> First of all, the AlarmThread allows you to limit the
 * maximum number of items stored in itself.  The reason why it
 * implements this feature is because this automatically makes it
 * a LRU (Least Recently Used) cache, which is a common type
 * of cache which is useful to have.
 * <LI> Second, since the AlarmThread is integrated with
 * the GlobalDBCache, then, when items are deleted from
 * the GlobalDBCache, the AlarmThread needs to delete these items
 * as well.  Now, I made an implementation decision here.
 * Instead of actually deleting items, I simply set a flag
 * in each item saying whether it was deleted or not.  This flag
 * is the InstanceCacheEntry variable called is_in_cache.
 * Whenever this flag is false, the AlarmThread should simply
 * ignore this item.  The reason I did this was because
 * I felt that deleting these items was an extra overhead
 * which had no benefits other than freeing up a little memory
 * ahead of time.  If each item does not take up much space,
 * the execution overhead to delete it may be unnecessary.
 * </UL>
 * <P>
 * Other notes:
 * <UL>
 * <LI> If you add an item to always_ringing, then it will NEVER be
 * removed, unless you yourself explicitly remove it in your own
 * processAlarm() method, by setting your is_removed flag to true
 * so that calling hasBeenRemoved() on your object returns true.
 * So, if you set the maximum_number_of_entries argument in
 * the constructor, and always_ringing contains more than
 * maximum_number_of_entries objects, it doesn't matter;
 * the AlarmThread will not delete any objects.  It will ONLY
 * delete an object from always_ringing if hasBeenRemoved()
 * returns true.  Remember this if you are concerned about memory
 * usage!
 * </UL>
 */
package org.locomotive.util.cache;

import java.util.*;
import java.sql.*;

public abstract class AlarmThread implements Runnable
{
    Vector alarms_this_period[];
    
    // total_number_of_entries is the total number of objects that
    // have registered their alarms here.
    int total_number_of_entries;

    Thread alarm_thread;

    public int SECONDS_IN_PERIOD, WAKE_UP_INTERVAL, MAXIMUM_NUMBER_OF_ENTRIES;

    int SIZE_OF_BUFFER;

    long last_time_checked, current_time;

    Vector always_ringing;
    
    boolean keep_running = true;


    /**
     * @param number_of_seconds_in_period your alarm events must
     * trigger within a certain maximum duration of time.  This
     * parameter expresses that time as a number of seconds.
     * @param wake_up_interval how often this AlarmThread should wake
     * up and check for alarm events which have triggered.
     * @param maximum_number_of_entries if you want to restrict the
     * capacity for the total number of alarm events, set this
     * parameter to that number.  if you do not want to restrict the
     * capacity, set this parameter to 0.
     */
    public AlarmThread(int number_of_seconds_in_period, int wake_up_interval,
		       int maximum_number_of_entries)
    {
	if (number_of_seconds_in_period < 10)
	    number_of_seconds_in_period = 3600;
	
	if (wake_up_interval <= 0 ||
	    wake_up_interval >= number_of_seconds_in_period)
	    wake_up_interval = 10;
	
	if (maximum_number_of_entries < 0)
	    maximum_number_of_entries = 0;
	
	this.SECONDS_IN_PERIOD = number_of_seconds_in_period;
	this.WAKE_UP_INTERVAL = wake_up_interval;
	this.MAXIMUM_NUMBER_OF_ENTRIES = maximum_number_of_entries;

	this.SIZE_OF_BUFFER =
	    number_of_seconds_in_period + wake_up_interval + 5;

//	print ("SIZE_OF_BUFFER = " + SIZE_OF_BUFFER);
//	print ("MAXIMUM_NUMBER_OF_ENTRIES = " +
//	          MAXIMUM_NUMBER_OF_ENTRIES);
//	print ("WAKE_UP_INTERVAL = " + WAKE_UP_INTERVAL);

	// the reason we make SIZE_OF_BUFFER slightly bigger is to fix
	// the following problem.  Suppose we woke up at time T and we
	// wanted to process all the objects which have expired in the
	// last ten minutes.  Now, let's say that someone else, at
	// exactly time T, registered a new object with an expiration
	// time of 3599, which is a legitimate expiration time.  Now,
	// let's say that SECONDS_IN_PERIOD = 3600 in this example.
	// Then, if SIZE_OF_BUFFER were exactly the same size as
	// SECONDS_IN_PERIOD, then this new object would be registered
	// at exactly (3599 modulo 3600), which is T - 1.  Thus, this
	// new object would be detected immediately by our expiration
	// process, because it occurred within the last 10 minutes,
	// between T - 10 and T.  This is not what we want.  However,
	// if we added a little buffer to SIZE_OF_BUFFER, then the new
	// object would be registered at T - 15 instead.  Voila.

	always_ringing = new Vector();

	alarms_this_period = new Vector [SIZE_OF_BUFFER];

	for (int i=0; i < SIZE_OF_BUFFER; i++)
	    alarms_this_period [i] = new Vector();

	total_number_of_entries = 0;

	last_time_checked = System.currentTimeMillis()/1000;

	alarm_thread = new Thread (this);
	alarm_thread.start();
    }


    /**
     * The run() method only takes care of sleeping and waking up.
     */
    public void run()
    {
	// in the future, we can change keep_running to false to stop
	// this
	// thread gracefully.
	keep_running = true;

        while (keep_running)
        {
	    // sleep....
            try {
                Thread.sleep (WAKE_UP_INTERVAL * 1000);
            }
            catch (InterruptedException e) {
                keep_running = false;
            }

            findRingingObjects();
        }
    }
    
    
    /**
     * Allows you to stop this thread gracefully, without forcing it
     * to die a violent death.
     */
    public void shutdown()
    {
	keep_running = false;
    }


    /**
     * This method finds all the alarm events which have triggered while
     * this AlarmThread was sleeping.  It also checks all the alarm events
     * which always trigger, which are stored in the Vector always_ringing.
     */
    void findRingingObjects()
    {
	// now, let's first process all the "always check" items:
	// (I felt that Vector methods are synchronized, and probably
	// slower in general than array methods, so that's why I'm
	// copying the always_ringing Vector into an array before
	// iterating over each element.)

	if (always_ringing.size() > 0)
	{
	    Object always_ringing_snapshot [];

	    synchronized (always_ringing)
	    {
		always_ringing_snapshot = new Object[always_ringing.size()];
		always_ringing.copyInto(always_ringing_snapshot);
	    }

	    int size = always_ringing_snapshot.length;
	    for (int i=0; i < size; i++)
	    {
		processAlarm(always_ringing_snapshot[i]);
	    }

	    // periodically, we should clean out the obsolete entries
	    // in always_ringing.  Deciding whether an entry is
	    // obsolete is up to each particular subclass to decide,
	    // by implementing hasBeenRemoved().
	    if (Math.random() < 0.2)
		clearOutDeadItems (always_ringing);
	}


	// next, let's handle all the objects in the list of regular
	// timed alarms.

	long current_time = System.currentTimeMillis()/1000;

	long must_check_until = last_time_checked;

	if (MAXIMUM_NUMBER_OF_ENTRIES != 0 &&
	    total_number_of_entries > MAXIMUM_NUMBER_OF_ENTRIES)
	{
	    int totalObjects_left = total_number_of_entries;
	    while (totalObjects_left > MAXIMUM_NUMBER_OF_ENTRIES)
	    {
		Vector alarms = (Vector) alarms_this_period [
		    (int) (must_check_until % SIZE_OF_BUFFER)];
		totalObjects_left -= alarms.size();
		must_check_until++;
	    }

/**********	    
	    print ("total # items = " + total_number_of_entries +
		   " > " + MAXIMUM_NUMBER_OF_ENTRIES +
		   ", so we need to get rid of " +
		   (must_check_until - last_time_checked) +
		   " entries, which removes: " +
		   (total_number_of_entries - totalObjects_left) +
		   " items.");
*********/
	}
	
	
	if (must_check_until < current_time)
	    must_check_until = current_time;
	
	for (long which_second = last_time_checked;
	     which_second < must_check_until;
	     which_second++)
	{
	    // we can't use a long within a Java array, so we have to
	    // convert to an int first.
	    Vector alarms_this_second_vector = (Vector) alarms_this_period [
		(int) (which_second % SIZE_OF_BUFFER)];
	    Object alarms_this_second [];

//	    print ("processing alarms [" + (which_second % SIZE_OF_BUFFER)
//		      + "]...");
	    
	    // Because Vector methods are synchronized and therefore
	    // much slower, I'm going to copy the contents of
	    // alarms_this_second_vector into an array instead.

	    synchronized (alarms_this_second_vector)
	    {
		alarms_this_second =
		    new Object [alarms_this_second_vector.size()];
		alarms_this_second_vector.copyInto(alarms_this_second);
		
		// removing all elements does NOT free up the space
		// already allocated for this Vector, which is GOOD,
		// because we don't want so much garbage collection
		// going on, nor do we want to re-allocate new space
		// every time for this Vector; we just want to get rid
		// of the items themselves.
		alarms_this_second_vector.removeAllElements();
	    }

	    // remember to reduce total_number_of_entries appropriately:
	    int size = alarms_this_second.length;
	    total_number_of_entries -= size;

	    // iterate through these alarms and process each one.

	    for (int j=0; j < size; j++)
	    {
		Object alarm = alarms_this_second[j];

		if (getAlarmTime (alarm) > must_check_until)
		{
		    // this object's alarm has not yet started
		    // ringing.  We better save its future alarm time.
		    registerAlarm (alarm);
		}
		else
		{
		    // this object is ringing.  Let's process its alarm.
		    processAlarm (alarm);
		}
	    }
	}

	last_time_checked = current_time;
    }


    /**
     * This method is called to periodically clean up the
     * always_ringing list, or, in other words, the
     * "check these items every time you wake up" list.
     */
    void clearOutDeadItems (Vector list)
    {
	Object new_list [];
	synchronized (list)
	{
	    new_list = new Object [list.size()];
	    list.copyInto (new_list);
	    list.removeAllElements();
	}

	int size = new_list.length;
	total_number_of_entries -= size;

	for (int i=0; i < size; i++)
	{
	    if (! hasBeenRemoved (new_list [i]))
	    {
		list.addElement(new_list [i]);
		total_number_of_entries++;
	    }
	    else
	    {
//		print ("removed: " + new_list[i] + " from always_ringing.");
	    }
	}
    }

    // for debugging:
    public void print (String string)
    {
	//	Loco.log (9, ((System.currentTimeMillis()/1000) % SIZE_OF_BUFFER) + 
	//		  " ____> " + string);
    }

    
    /**
     * Your alarm-handling system should implement this method,
     * which tells whether this object has already been removed or not.
     */
    abstract boolean hasBeenRemoved (Object object);


    /**
     * Your alarm-handling system should implement this method.
     * @return the time when this object's alarm will trigger
     */
    abstract long getAlarmTime (Object object);

    
    /**
     * Your alarm-handling system should implement this method.
     * It should process an alarm going off.
     */
    abstract void processAlarm (Object object);


    /**
     * Your alarm-handling system should implement this method.
     * It should add an alarm into the system (either into
     * alarms_this_period or always_ringing).
     */
    abstract void registerAlarm (Object object);

}
