/**
 * DBCache is a cache which allows you to read or write RelationalObjects
 * or any other kind of Object to and from the database (or some other source
 * which you define).
 * <P>
 * This cache can automatically expire or auto-save your Objects, as well.
 * Specifically, expiration means that if an Object in the cache
 * has not been referenced for some period of time (which you can define),
 * it will be automatically removed from the cache.
 * Auto-saving an Object means that a thread will periodically check
 * whether a RelationalObject has any dirty data which need to be saved to the
 * database.  if it does have dirty data, then it will be saved.
 * <P>
 * <H4> Example </H4>
 * <P>
 * Let's run through an example of how this API would really work in
 * practice.  First, we have to create or get hold of an existing DBCache 
 * object. Here we will use the default DBCache for Loco :
 * <PRE>
 * DBCache dbCache = Loco.getGlobalDBCache();
 * </PRE>
 * Second, I will implement a RelationalObject called AddressBook.
 * AddressBook knows how to read and write itself to the database, as a
 * relational database table with several columns: first_name, last_name, and
 * address.  Next, I register AddressBook with the DBCache:
 * <PRE>
 *     dbCache.register ("AddressBook_table", new AddressBook().getClass(), 0, 3600, false);
 * </PRE>
 * Now, the DBCache knows that whenever I'm talking about
 * "AddressBook_table", I'm talking about the AddressBook class.  Also,
 * DBCache will automatically save dirty data in my AddressBook objects
 * to the database, at an interval set by default in DBCache (for
 * example, this might be every 60 seconds).  I also want DBCache to
 * expire AddressBook objects which have not been accessed for more than an
 * hour.  I'm defining access as either get() or contains() or put().
 * <P>
 * Now, let's say that AddressBook_table in the database does not contain
 * anything right now.  Then, I might say:
 * <PRE>
 *     AddressBook my_book = new AddressBook ("Ben", "Lai", "no address");
 *     dbCache.put ("AddressBook_table", "Ben", my_book);
 * </PRE>
 * Then, after 60 seconds or less, dbCache will detect that there is a
 * new AddressBook in its cache, and it will write that AddressBook into the
 * database, using the insert () command of AddressBook.  dbCache knows
 * that this AddressBook is a new one because it keeps track of that
 * information.
 * <P>
 * Now, let's say that, a few minutes later, I need my AddressBook again.
 * Note that my current Servlet or Handler is a completely different one from
 * the one which created the AddressBook in the database, which probably
 * ended already.
 * <PRE>
 *    AddressBook my_book =
 *        (AddressBook) dbCache.get("AddressBook_table", "Ben");
 * my_book.setAddress ("1 Embarcadero Centre");
 * </PRE>
 * Now, if I'm a good programmer, then inside my setAddress() method, I set
 * my dirty flag to true, so that, no more than 60 seconds from now,
 * dbCache will write my new changes to the database.
 * <P>
 * Now, let's say that I log off, and I don't return until the next day.  In
 * that case, then my AddressBook should expire from dbCache, which
 * checks every hour for expired AddressBooks.
 * <P>
 * Now, what if I want to make sure that the database always has the latest
 * version of my AddressBook?  There are two ways to implement this.  First,
 * I can set the auto_save_frequency to 1 when I register the AddressBook
 * class.  Then, dbCache will check once per second for dirty
 * AddressBook data.  However, this method may require a lot of CPU time,
 * especially if there are a lot of AddressBook objects.  The second
 * technique is for me to set the auto_save_frequency to -1, so that the
 * dbCache NEVER checks.  Then, it is up to me as the programmer to
 * make sure that, whenever I change any data in AddressBook, I immediately
 * save my AddressBook to the database.  Actually, since I know that I'm not
 * perfect, I would rather set auto_save_frequency to 300, or 5 minutes, just
 * in case there is someplace where I modify data in my AddressBook object
 * and I forgot to save it to the database.  This guarantees that all
 * AddressBook data in the database will be at least as recent as 5 minutes
 * ago, without using a lot of CPU.
 * <P>
 * <H4>Implementation </H4>
 * <P>
 * dbCache is primarily implemented using a Hashtable which uses
 * the TwoKey class as the key (the TwoKey object simply combines the
 * hashCode() results of two objects into one, and does equals() on
 * both objects).  This seems to be the cleanest way to implement it,
 * and seems to offer the best performance.  I had considered using a
 * HashtableOfHashtables, but that's two hash lookups instead of one,
 * so I changed to using one Hashtable.  If we need the functionality
 * of the HashtableOfHashtables class later, it's easy to convert back.
 * <P>
 *DBCache uses the ExpirationThread and the AutoSaveThread to
 * check when objects have expired from the cache, or when they need
 * to be auto-saved.  This includes removing or re-setting the alarms
 * set by the get(), put(), and remove() methods.
 */
package org.locomotive.util.cache;

import java.util.*;
import java.sql.*;
import java.io.*;
import org.locomotive.util.*;
import org.locomotive.server.DBConnectionPool;
import org.locomotive.server.Log;
import org.locomotive.server.Config;
import org.locomotive.loco.Loco;
import org.locomotive.loco.LocoConfig;


public class DBCache
{
    DBConnectionPool connectionPool;

    int maxEntries = -1;

    
    // the default frequency for how long an object can remain in the
    // cache (used primarily by the ExpirationThread)
    // expressed in number of seconds
    int defaultExpirationTime = -1;


    // the maximum time that any object can remain in this cache,
    // expressed in number of seconds
    int maxExpirationTime = -1;

    // how long the ExpirationThread sleeps before waking up to check
    // whether objects have expired, expressed in number of seconds
    int expirationSleepInterval = 60;

    // the default frequency for an object's auto-save interval, which is
    // how often to check whether a particular object has dirty data,
    // expressed in number of seconds
    int defaultAutoSaveInterval = 60;


    // the maximum auto-save interval for any object in this cache,
    // expressed in number of seconds.  This should be the same as
    // MAX_EXPIRATION_TIME, because the AlarmThread has only a certain
    // capacity.
    int maxAutoSaveInterval = maxExpirationTime;

    // how long the AutoSaveThread sleeps before waking up to check
    // dirty data in objects, expressed in number of seconds
    int autoSaveSleepInterval = 60;

    Hashtable class_metadata = new Hashtable();
  
    DoubleHashtable instance_cache = new DoubleHashtable();

  
    /**
     * The ExpirationThread handles all expirations of Objects from the cache.
     * The following constructor creates an ExpirationThread
     * which can remember Objects for 3600 seconds;
     * which will wake up every 10 seconds to find expired Objects;
     * and which will store a maximum of 10,000 Objects before forcing Objects
     * to expire (even though they have not yet reached their requested
     * expiration time).
     */
    ExpirationThread expiration_thread;

    /**
     * The AutoSaveThread periodically saves the dirty data
     * of every Object in the cache.
     * The following constructor creates an AutoSaveThread
     * which can remember Objects for 3600 seconds, and
     * which will wake up every 20 seconds to check for dirty data.
     * (We don't need to set a maximum for the number of Objects to store
     * in the AutoSaveThread, because the ExpirationThread will manage that
     * for us.  That's why the last argument is 0, meaning:
     * "no maximum, go have a party!".)
     */
    AutoSaveThread auto_save_thread;

    Log server_log;

    /**
     * Startup the service.
     * This should be called after the service has been 
     * configured properly.
     */
    public void startService(){
	expiration_thread = new ExpirationThread (this, maxExpirationTime,
						  expirationSleepInterval,
						  maxEntries);

	auto_save_thread = new AutoSaveThread (this, maxAutoSaveInterval,
					       autoSaveSleepInterval,
					       0);
    }

    /**
     * Default constructor for the database cache. The Connection 
     * pool and other parameters should be set using bean interface methods, 
     * and the startService() method should be used to start up the DBCache
     * service. 
     */
    public DBCache(){
    }

    /**
     * Construct a database cache reading default values from a Config 
     * and with a specific connection pool.
     * @param config The Config to get default values from.
     * @param connectionPool The connection pool to use for this cache.
     */
    public DBCache(Config config, DBConnectionPool connectionPool,
		   Log server_log)
    {
	this(config, connectionPool);

	this.server_log = server_log;
    }
    
    /**
     * Construct a database cache reading default values from a LocoConfig 
     * and with a specific connection pool.
     * @param config The LocoConfig to get default values from.
     * @param connectionPool The connection pool to use for this cache.
     */
    public DBCache(Config config, DBConnectionPool connectionPool){
	this.connectionPool = connectionPool;

	maxEntries = -1;
    
	defaultExpirationTime = config.getInt("CACHE_DEFAULT_EXPIRATION_TIME");

	maxExpirationTime = config.getInt("CACHE_MAX_EXPIRATION_TIME");
    
	expirationSleepInterval = config.getInt("CACHE_EXPIRATION_SLEEP_INTERVAL");
    
	defaultAutoSaveInterval = 
	    config.getInt("CACHE_DEFAULT_AUTO_SAVE_INTERVAL");

	maxAutoSaveInterval = maxExpirationTime;

	autoSaveSleepInterval = config.getInt("CACHE_AUTO_SAVE_SLEEP_INTERVAL");
    }

    public void setMaxEntries(int maxEntries){
	this.maxEntries = maxEntries;
    }

    public int getMaxEntries(){
	return maxEntries;
    }

    /**
     * the default frequency for how long an object can remain in the
     * cache (used primarily by the ExpirationThread)
     * expressed in number of seconds.
     */
    public void setDefaultExpirationTime(int defaultExpirationTime){
	this.defaultExpirationTime = defaultExpirationTime;
    }

    public int getDefaultExpirationTime(){
	return defaultExpirationTime;
    }

    /** 
     * The maximum time that any object can remain in this cache,
     * expressed in number of seconds.
     */
    public void setMaxExpirationTime(int maxExpirationTime){
	this.maxAutoSaveInterval = this.maxExpirationTime = maxExpirationTime;
    }

    public int getMaxExpirationTime(){
	return maxExpirationTime;
    }

    /**
     * how long the ExpirationThread sleeps before waking up to check
     * whether objects have expired, expressed in number of seconds
     */
    public void setExpirationSleepInterval(int expirationSleepInterval){
	this.expirationSleepInterval = expirationSleepInterval;
    }

    public int getExpirationSleepInterval(){
	return expirationSleepInterval;
    }
    /**
     * how long the AutoSaveThread sleeps before waking up to check
     * dirty data in objects, expressed in number of seconds
     */     
    public void setDefaultAutoSaveInterval(int defaultAutoSaveInterval){
	this.defaultAutoSaveInterval = defaultAutoSaveInterval;
    }

    public int getDefaultAutoSaveInterval(){
	return defaultAutoSaveInterval;
    }

    /**
     * how long the AutoSaveThread sleeps before waking up to check
     * dirty data in objects, expressed in number of seconds
     */
    public void setAutoSaveSleepInterval(int autoSaveSleepInterval){
	this.autoSaveSleepInterval = autoSaveSleepInterval;
    }

    public int getAutoSaveSleepInterval(){
	return autoSaveSleepInterval;
    }

    /**
     * Set the connection pool this cache.
     * @param connectionPool The connection pool to use for this cache.
     */
    public void setConnectionPool(DBConnectionPool connectionPool){
	this.connectionPool = connectionPool;
    }

    /**
     * Get the connection pool for this cache.
     * @return Returns the connectionPool used by this cache;
     */
    public DBConnectionPool getConnectionPool(){
	return connectionPool;
    }

  

    /**
     * The register() method allows you to register your Objects that
     * you want to store in the database, with the DBCache.
     * There are two benefits in registering:
     * <UL>
     * <LI> if there is a naming conflict, a CacheNamingException
     * will be thrown, which conveniently alerts you to that problem.
     * <LI> once you have registered a short, mnemonic
     * short_class_name for your RelationalObject class, you can use
     * that name wherever you need to refer to your class, which is
     * much more convenient and readable.
     * </UL>
     * @param short_class_name should have only alphanumeric
     * characters and '_'.
     * @param auto_save_frequency_in_seconds how
     * often the DBCache should check for dirty data for each
     * RelationalObject, and write it to the database.  Expressed as
     * the number of seconds between each periodic check.
     * <ul>
     * <li>If this number is -1, then it means "never save to the database".
     * <li>If this number is 0, then it means use DBCache's
     * default auto-save frequency.
     * <li>If this number is 1, then it means that the DBCache
     * will check "as frequently as possible" -- that is, each time
     * the DBCache wakes up, at the interval you specify.
     * <li>If this number is > 3600, it will be treated as if it were
     * 3600.  In other words, we think that it makes sense to flush
     * dirty information at least once per hour.  if you want to
     * flush dirty information at an interval longer than an hour,
     * use -1 ("never save") and simply rely on your own scheduling
     * algorithm.
     * </ul>
     * @param expiration_in_seconds how long, in seconds, an Object
     * can remain in the cache, without being accessed by
     * DBCache's get() or put() methods, before it
     * is removed.  We recommend that you make this number reasonably
     * large, for example, at least 60 seconds, but more typically,
     * 1800 (1/2 hour) or 3600 (1 hour).  Otherwise, why bother using
     * the DBCache at all?
     * <ul>
     * <li>If this number is -1, then it means "never expires".
     * <li>If this number is 0, then it means use DBCache's
     * default expiration period.
     * <li>If this number is > 3600, it will be treated as if it were
     * 3600.  In other words, the DBCache only has room to
     * store up to an hour's worth of data.  Or, you can write your
     * own expiration algorithm by combining -1 ("never expire") with
     * the remove() method.
     * </ul>
     * @param affects_other_sessions if this is true, then it means
     * that we will need to have a consistent view for this
     * RelationalObject across multiple Locomotive servers.  This is
     * very complicated, and will be implemented later.  for now,
     * this argument is unused; it's just a placeholder.
     */
    public void register (String short_class_name,
			  Class class_to_register,
			  long expiration_frequency_in_seconds,
			  long auto_save_frequency_in_seconds,
			  boolean affects_other_sessions)
	throws CacheNamingException
    {
	// check that short_class_name has only alphanumeric characters or '_'
	if (short_class_name == null || short_class_name.length() == 0)
	    throw new CacheNamingException ("empty short_class_name!");

	StringTokenizer tokens = new StringTokenizer (short_class_name, "_");
	while (tokens.hasMoreTokens())
	    {
		if (!StringUtil.isAlphanum (tokens.nextToken()))
		    {
			throw new CacheNamingException (
							"illegal character in short_class_name: " +
							short_class_name + "!");
		    }
	    }

	// Check that short_class_name is not already registered
	if (class_metadata.contains (short_class_name))
	    {
		throw new CacheNamingException (
						short_class_name +
						" already exists as a registered class name!");
	    }

	// Create some meta-information associated with this class,
	// containing auto_save_frequency_in_seconds and expiration_in_seconds,
	// as well as whether this class implements RelationalObject,
	// and then put the class and the meta-information into a
	// Hashtable.
	
	boolean is_relational_object = false;
	Object new_instance;

	try
	    {
		new_instance = class_to_register.newInstance();
		// find out whether this class implements RelationalObject.
		is_relational_object = (new_instance instanceof RelationalObject);
		if (! is_relational_object &&
		    ! (new_instance instanceof Serializable))
		    {
			throw new CacheNamingException (
							class_to_register + " must implement either the " +
							"RelationalObject interface or the Serializable " +
							"interface!");
		    }
	    }
	catch (Throwable e)
	    {
		log(0, "ERROR: DBCache.register(): " +
		    class_to_register.toString() + ": " +
		    StringUtil.exceptionStackTrace(e));
	    }

	// make sure that the expiration frequency is set to a
	// legitimate value.
	long expiration_frequency = expiration_frequency_in_seconds;

	if (expiration_frequency != -1)
	    {
		if (expiration_frequency == 0)
		    expiration_frequency = defaultExpirationTime;
		else if (expiration_frequency > maxExpirationTime)
		    expiration_frequency = maxExpirationTime - 1;
		else if (expiration_frequency < expirationSleepInterval)
		    expiration_frequency = expirationSleepInterval - 1;
	    }


	// make sure that the auto-save frequency is set to a legitimate value.
	long auto_save_frequency = auto_save_frequency_in_seconds;

	if (auto_save_frequency != -1)
	    {
		if (auto_save_frequency == 0)
		    auto_save_frequency = defaultAutoSaveInterval;
		else if (auto_save_frequency > maxAutoSaveInterval)
		    auto_save_frequency = maxAutoSaveInterval - 1;
		else if (auto_save_frequency < autoSaveSleepInterval)
		    auto_save_frequency = autoSaveSleepInterval - 1;
	    }

	class_metadata.put (short_class_name,
			    new MetaDataEntry (short_class_name,
					       class_to_register,
					       is_relational_object,
					       auto_save_frequency,
					       expiration_frequency));
	
	//	print ("registering " + class_to_register + " as '" +
	//	          short_class_name + "', expires = " + expiration_frequency +
	//	          ", auto-save = " + auto_save_frequency + ".");
    }



    /**
     * the default register method is:
     */
    public void register (String short_class_name,
			  Class class_to_register)
	throws CacheNamingException
    {
	register (short_class_name, class_to_register, 0, 0, false);
    }
    

    /**
     * This method retrieves an object from the cache.  This method
     * will "refresh" this Object's expiration time; in other words,
     * its expiration time in the cache will be pushed out. 
     */
    public Object get (String short_class_name, Object key)
	throws SQLException, CacheNamingException
    {
	if (short_class_name == null || short_class_name.length() == 0)
	    throw new CacheNamingException ("empty short_class_name!");

	if (key == null)
	    throw new CacheNamingException ("empty key!");

	// if there is an Object in the cache which matches the key
	// and short_class_name, then return that Object.
	InstanceCacheEntry instance_cache_entry =
	    (InstanceCacheEntry) instance_cache.get (short_class_name, key);

	if (instance_cache_entry != null && instance_cache_entry.is_in_cache)
	    {
		// Remember to refresh expiration-time here.
		instance_cache_entry.expiration_time =
		    System.currentTimeMillis()/1000 +
		    instance_cache_entry.class_data.expiration_frequency;
	    
		//	print ("reset expiration time for " + instance_cache_entry + " to " +
		//	       instance_cache_entry.expiration_time % SIZE_OF_BUFFER + ".");
	    
		return (instance_cache_entry.object);
	    }
	else
	    {
		MetaDataEntry class_data =
		    (MetaDataEntry) class_metadata.get (short_class_name);
	    
		if (class_data == null)
		    throw new CacheNamingException (
						    short_class_name + " is not a registered class!");

		Object new_object = loadObject (class_data, key);

		if (new_object == null)
		    return (null);

		instance_cache_entry =
		    new InstanceCacheEntry (class_data, key, new_object);

		instance_cache.put (short_class_name, key,
				    instance_cache_entry);
		expiration_thread.registerAlarm (instance_cache_entry);
		auto_save_thread.registerAlarm (instance_cache_entry);

		return (new_object);
	    }
    }


    protected  Object loadObject (MetaDataEntry class_data, Object key)
	throws SQLException
    {
	// if the class is identified as a RelationalObject,
	// then use the RelationalObject method, load(), to load
	// that Object from the database.
	// Else, load that Object from the database as a Serializable.
	if (class_data.is_relational_object)
	    {
		RelationalObject new_relational_object;
	    
		try
		    {
			new_relational_object =
			    (RelationalObject) (class_data.my_class.newInstance());
		    }
		catch (Throwable e)
		    {
			log(0, "ERROR: DBCache.loadObject(): " +
			    class_data.my_class.toString() + ": " +
			    StringUtil.exceptionStackTrace(e));
			return null;
		    }
	    
		Connection conn = null;

		try {
		    if(connectionPool != null)
			conn = connectionPool.getConnection();
		    new_relational_object.load (conn, key);
		} finally {
		    if (conn != null) 
			connectionPool.returnToPool(conn);
		}

		if (!new_relational_object.existsInDB())
		    return (null);
		else
		    return (new_relational_object);
	    }
	else
	    {
		return (loadSerializedObjectFromDB (class_data, key));
	    }
    }


    protected Object loadSerializedObjectFromDB (
						 MetaDataEntry class_data, Object key)
	throws SQLException
    {
	if (!((key instanceof Integer) || (key instanceof Long) ||
	      (key instanceof String)))
	    {
		throw new SQLException (
					"The key for the class: " + class_data.name +
					" must be an Integer, a Long, or a String, instead of a " +
					key.getClass().getName());
	    }
	
	Object new_object = null;
	Statement stmt = null;
	ResultSet rs = null;

	String search_key = null;
	if (key instanceof String)
	    search_key = SQL.quoteString(key.toString());
	else
	    search_key = key.toString();

	String statement = "select object from loco_cache_objects " +
	    " where class_name = " + SQL.quoteString(class_data.name) +
	    " and key = " + search_key;

	//	print ("loading SERIALIZED object using: '" + statement + "'...");

	Connection conn = null;

	try {
	    if(connectionPool != null)
		conn = connectionPool.getConnection();
	    stmt = conn.createStatement();
	    rs = stmt.executeQuery(statement);
	    if (rs.next())
		{
		    ObjectInputStream obis = new ObjectInputStream(
								   new BufferedInputStream(rs.getBinaryStream(1)));
		    new_object = obis.readObject();
		    obis.close();
		}
	    else
		{
		    // if the Object could not be found in the database,
		    // then simply return null, just like a Hashtable
		    // returns null when it does not contain an item.
		    return (null);
		}
	} catch (Throwable ex) {
	    throw new SQLException("DBCache.get(" +
				   class_data.name +
				   ", " + key.toString() + "): " +
				   StringUtil.exceptionStackTrace(ex));
	} finally {
	    try {
		if (rs != null) rs.close();
		if (stmt != null) stmt.close();
	    } catch (SQLException se) {}
	    
	    if (conn != null) 
		connectionPool.returnToPool(conn);
	}
	
	return (new_object);
    }



    /**
     * "convenience" get() method using an int key:
     */
    public Object get (String short_class_name, int int_key)
	throws SQLException, CacheNamingException
    {
	return (get (short_class_name, new Integer (int_key)));
    }

    /**
     * "convenience" get() method using a long key:
     */
    public Object get (String short_class_name, long long_key)
	throws SQLException, CacheNamingException
    {
	return (get (short_class_name, new Long (long_key)));
    }


    /**
     * This method puts an object in the cache.
     * NOTE: if the item is dirty, this method will write
     * it into the database immediately.
     */
    public void put (String short_class_name, Object key,
		     Object new_object)
	throws SQLException, CacheNamingException
    {
	if (short_class_name == null || short_class_name.length() == 0)
	    throw new CacheNamingException ("empty short_class_name!");
	
	if (key == null)
	    throw new CacheNamingException ("empty key!");
	
	if (new_object == null)
	    throw new CacheNamingException (
					    "cannot put a null Object into the cache!");
	
	// if there is an Object in the cache which matches the key
	// and short_class_name, then simply overwrite that Object.
	// NOTE: this segment of code may also be useful in the
	// future, because we need to remove or reset the alarm for
	// this Object.
	InstanceCacheEntry instance_cache_entry =
	    (InstanceCacheEntry) instance_cache.get(short_class_name, key);

	if (instance_cache_entry != null && instance_cache_entry.is_in_cache)
	    {
		instance_cache_entry.object = new_object;

		// Remember to refresh expiration-time here.
		instance_cache_entry.expiration_time =
		    System.currentTimeMillis()/1000 +
		    instance_cache_entry.class_data.expiration_frequency;
		//	print ("reset expiration time for " + instance_cache_entry + " to " +
		//	       instance_cache_entry.expiration_time % SIZE_OF_BUFFER + ".");
	    }
	else
	    {
		MetaDataEntry class_data =
		    (MetaDataEntry) class_metadata.get (short_class_name);
		if (class_data == null)
		    throw new CacheNamingException (
						    short_class_name + " is not a registered class!");

		instance_cache_entry =
		    new InstanceCacheEntry (class_data, key, new_object);

		instance_cache.put (short_class_name, key,
				    instance_cache_entry);
		expiration_thread.registerAlarm (instance_cache_entry);
		auto_save_thread.registerAlarm (instance_cache_entry);
	    }

	// if the new Object is a RelationalObject, and it is dirty,
	// then write it into the database.
	if ((new_object instanceof RelationalObject) &&
	    ((RelationalObject) new_object).isDirty())
	    {
		Connection conn = null;

		try {
		    if(connectionPool != null)
			conn = connectionPool.getConnection();
		
		    RelationalObject relational_object =
			(RelationalObject) new_object;
		    if (relational_object.existsInDB())
			relational_object.update (conn);
		    else
			relational_object.insert (conn);
		
		} finally {
		    if (conn != null) 
			connectionPool.returnToPool(conn);
		}
	    }
    }


    /**
     * This method will write an Object into the database,
     * without saving it into the cache.  This method neither uses nor sets
     * the dirty flag of a RelationalObject explicitly.
     * NOTE: short_class_name must be registered with the DBCache.
     */
    public boolean writeToDB (String short_class_name, Object key,
			      Object new_object)
	throws SQLException, CacheNamingException
    {
	if (short_class_name == null || short_class_name.length() == 0)
	    throw new CacheNamingException ("empty short_class_name!");

	if (key == null)
	    throw new CacheNamingException ("empty key!");

	if (new_object == null)
	    throw new CacheNamingException (
					    "cannot write a null Object into the database!");

	if (new_object instanceof RelationalObject)
	    {
		Connection conn = null;

		try {
		    if(connectionPool != null)
			conn = connectionPool.getConnection();
		    boolean result = false;
		
		    RelationalObject relational_object =
			(RelationalObject) new_object;
		    if (relational_object.existsInDB())
			result = (relational_object.update (conn));
		    else
			result = (relational_object.insert (conn));
		
		    return (result);
		
		} finally {
		    if (conn != null) 
			connectionPool.returnToPool(conn);
		}

	    }
	else
	    {
		return (writeSerializedObjectToDB (short_class_name, key,
						   new_object));
	    }
    }


    public boolean writeSerializedObjectToDB (String short_class_name,
					      Object key,
					      Object new_object)
	throws SQLException, CacheNamingException
    {
	if (!((key instanceof Integer) || (key instanceof Long) ||
	      (key instanceof String)))
	    {
		throw new CacheNamingException (
						"The key for the class: " + short_class_name +
						" must be an Integer, a Long, or a String, instead of a " +
						key.getClass().getName());
	    }
	
	Connection conn = null;
	PreparedStatement pstmt = null;
	    
	try {
	    if(connectionPool != null)
		conn = connectionPool.getConnection();
	    conn.setAutoCommit(false);
	    
	    String search_key = null;
	    if (key instanceof String)
		search_key = SQL.quoteString(key.toString());
	    else
		search_key = key.toString();

	    String update_statement =
		"update loco_cache_objects set object = ? " +
		" where class_name = " +
		SQL.quoteString(short_class_name) +
		" and key = " + search_key;

	    String insert_statement = "insert into loco_cache_objects " +
		" (class_name, key, Object) values (" +
		SQL.quoteString(short_class_name) + ", " +
		search_key + ", ?)";
	    
	    pstmt = conn.prepareStatement (update_statement);
	    
	    // is 1 k enough?
	    ByteArrayOutputStream bos = new ByteArrayOutputStream(1024);
	    ObjectOutputStream oos = new ObjectOutputStream(bos);
	    oos.writeObject(new_object);
	    oos.flush();
	    ByteArrayInputStream bis =
		new ByteArrayInputStream(bos.toByteArray());
	    pstmt.setBinaryStream(2, bis, bis.available());
	    
	    int updated_successfully = pstmt.executeUpdate();
	    pstmt.close();
	    pstmt = null;
	    
	    if (updated_successfully > 0)
		{
		    if (! conn.getAutoCommit()) {
			conn.commit();
			conn.setAutoCommit(true);
		    }
		    return (true);
		}
	    else
		{
		    pstmt = conn.prepareStatement (insert_statement);
		    pstmt.setBinaryStream(2, bis, bis.available());
		
		    pstmt.executeUpdate();
		    pstmt.close();
		    pstmt = null;
		
		    if (! conn.getAutoCommit()) {
			conn.commit();
			conn.setAutoCommit(true);
		    }

		    return (true);
		}
	    
	} catch (SQLException se) {
	    if (! conn.getAutoCommit()) {
		conn.rollback();
		conn.setAutoCommit(true);
	    }
	    throw se;
	} catch (IOException ioe) {
	    throw new SQLException (
				    "DBCache couldn't write serialized Object: " +
				    ioe.toString());
	} finally {
	    if (pstmt != null)
		pstmt.close();

	    conn.setAutoCommit(true);
	    if (conn != null) 
		connectionPool.returnToPool(conn);
	}
    }


    /**
     * This method removes an object from the cache.  It does NOT delete
     * the Object from the database.  (To delete it from the database,
     * call the deleteFromDB() method)
     * NOTE: this Object will now no longer be auto-saved to the database,
     * even if it has dirty data.  That becomes your own responsibility.
     * @see deleteFromDB()
     */
    public Object remove (String short_class_name, Object key)
	throws SQLException, CacheNamingException
    {
	InstanceCacheEntry entry =
	    (InstanceCacheEntry) instance_cache.remove(short_class_name, key);

	if (entry == null)
	    {
		return null;
	    }
	else
	    {
		// now, we set the flag for telling the
		// ExpirationThread and the AutoSaveThread that
		// this item is no longer in the cache, and they
		// should not act on it.
		entry.is_in_cache = false;
		Object target_object = entry.object;
		// also, empty up some memory
		entry.object = null;
		return (target_object);
	    }
    }


    /**
     * This method deletes an object from the database, but NOT from
     * the cache.  This method is in DBCache because
     * DBCache seems like the only place with a general library
     * for dealing with Object serialization I/O with the database.
     * if you want to delete a RelationalObject instead of a
     * serialized generic Object, then just call the delete() method
     * for that RelationalObject.
     * @return whether the Object was successfully deleted from the database. 
     */
    public boolean deleteSerializedObjectFromDB (
						 String short_class_name, Object key)
	throws SQLException, CacheNamingException
    {
	if (short_class_name == null || short_class_name.length() == 0)
	    throw new CacheNamingException ("empty short_class_name!");

	if (key == null)
	    throw new CacheNamingException ("empty key!");

	if (!((key instanceof Integer) || (key instanceof Long) ||
	      (key instanceof String)))
	    {
		throw new CacheNamingException (
						"The key for the class: " + short_class_name +
						" must be an Integer, a Long, or a String, instead of a " +
						key.getClass().getName());
	    }
	
	Connection conn = null;

	try {
	    if(connectionPool != null)
		conn = connectionPool.getConnection();
	    
	    boolean result = (0 < SQL.update (
					      conn, "delete loco_cache_objects where class_name = " +
					      SQL.quoteString(short_class_name) + " and key = " +
					      ((key instanceof String) ? SQL.quoteString((String) key) :
					       key.toString())
					      ));
	    
	    return (result);
	    
	} finally {
	    if (conn != null) 
		connectionPool.returnToPool(conn);
	}
    }

    
    /**
     * Tells whether an Object is in the DBCache.
     * This method DOES NOT "refresh" this Object; in other words,
     * its expiration time in the cache IS NOT updated.
     */
    public boolean contains (String short_class_name, Object key)
	throws CacheNamingException
    {
	if (short_class_name == null || short_class_name.length() == 0)
	    throw new CacheNamingException ("empty short_class_name!");
	
	if (key == null)
	    throw new CacheNamingException ("empty key!");
	
	InstanceCacheEntry instance_cache_entry =
	    (InstanceCacheEntry) instance_cache.get (short_class_name, key);

	return (instance_cache_entry != null &&
		instance_cache_entry.is_in_cache);
    }
    
    
    /**  
     * This method will remove all the items for this class from the cache.
     * @param save_dirty_data if this is true, then each item will be
     * auto-saved to the database if it has dirty data.
     */
    public void clearCacheForClass (String short_class_name,
				    boolean save_dirty_data)
	throws CacheNamingException
    {
	if (short_class_name == null || short_class_name.length() == 0)
	    throw new CacheNamingException ("empty short_class_name!");
	
	if (! save_dirty_data)
	    {
		Enumeration items =
		    instance_cache.getElementsOfChildTable (short_class_name);
		while (items.hasMoreElements())
		    {
			InstanceCacheEntry entry =
			    (InstanceCacheEntry) items.nextElement();
			entry.is_in_cache = false;
			
			// also, empty up some memory
			entry.object = null;
		    }
	    }
	
	instance_cache.removeChildTable (short_class_name);
    }
    
    /**
     * This is just a convenience method for ExpirationThread and
     * AutoSaveThread. 
     */
    public void saveRelationalObjectIfDirty (InstanceCacheEntry entry)
	throws SQLException
    {
	if (entry.is_in_cache && (entry.class_data.auto_save_frequency != -1) &&
	    (entry.object instanceof RelationalObject) &&
	    ((RelationalObject) entry.object).isDirty())
	    {
		Connection conn = null;

		try {
		    if(connectionPool != null)
			conn = connectionPool.getConnection();
		
		    RelationalObject relational_object =
			(RelationalObject) entry.object;
		    if (relational_object.existsInDB())
			relational_object.update (conn);
		    else
			relational_object.insert (conn);
		
		} finally {
		    if (conn != null) 
			connectionPool.returnToPool(conn);
		}
	    }
    }
    
    
    public static void main (String argv[])
    {
	DBCache dbCache = new DBCache(Loco.getConfig(), null);
	
	String states [] =
	{ "Kansas", "California", "Michigan" // "Ohio", "Florida", "Massachusetts"
	};
	
	String cities [] =
	{ "Seattle", "Chicago", "Miami" // "Los Angeles", "San Francisco", "Vancouver"
	};
	
	String people [] =
	{ "Ben", "Dave", "Thede", "Piyush", "Chris", "Chip"
	};

	try {
	    dbCache.register ("states", new RelationalObject1().getClass(), 4, -1, false);
	    dbCache.register ("cities", new String().getClass(),  5, 5, false);
	    dbCache.register ("people", new String().getClass(), 20, 5, false);
	
	    for (int i=0; i < states.length; i++)
		{
		    dbCache.put ("states", states [i], new RelationalObject1());
		}
	
	    try { Thread.sleep (2000); } catch (InterruptedException e) {}
	
	    for (int i=0; i < cities.length; i++)
		{
		    //	        dbCache.put ("cities", cities [i], cities [i]);
		}
	
	    /*********
		      for (int i=0; i < people.length; i++)
		      {
		      dbCache.put ("people", people [i], people [i]);
		      }
	    *******/
	

	    // now, periodically test that these items are now in the cache:
	    for (int j=0; j < 4; j++)
		{
	    
		    for (int i=0; i < states.length; i++)
			{
			    //		    dbCache.print(states [i] + " -> " + dbCache.get ("states", states [i]));

			    RelationalObject1 rel =
				(RelationalObject1) dbCache.get ("states", states [i]);
			    if (rel != null)
				rel.setDirty(true);
		
			    dbCache.print(states [i] + (dbCache.contains ("states", states [i]) ?
						" exists!" : " NOT exist"));
			}
	    
		    for (int i=0; i < cities.length; i++)
			{
			    dbCache.print(cities [i] + " -> " + dbCache.get ("cities", cities [i]));
		
			    dbCache.print(cities [i] + (dbCache.contains ("cities", cities [i]) ?  " exists!" : " NOT exist"));
		
			    /*****
				  dbCache.print (people [i] + " -> " + get ("people", people [i]));
		
				  dbCache.print (people [i] + (dbCache.contains ("people", people [i]) ?  " exists!" : " NOT exist"));
			    ****/
			}
	    
		    System.out.println("");
		    dbCache.print ("sleeping...");

		    try { Thread.sleep (2000); } catch (InterruptedException e) {}

		    dbCache.print ("waking up!");
		    System.out.println("");
		}
	
	    // now, let's try removing some of these states:
	
	    dbCache.put ("states", "Massachusetts", "Massachusetts" );
	    dbCache.print("Massachusetts -> " + dbCache.get ("states", "Massachusetts"));
	    dbCache.print("Massachusetts" + (dbCache.contains ("states", "Massachusetts") ?  " exists!" : " NOT exist"));
	
	    dbCache.remove ("states", "Massachusetts");
	    dbCache.print("Massachusetts -> " + dbCache.get ("states", "Massachusetts"));
	    dbCache.print("Massachusetts" + (dbCache.contains ("states", "Massachusetts") ?  " exists!" : " NOT exist"));
	
	    //	try { Thread.sleep (4000); } catch (InterruptedException e) {}
	
	    dbCache.put ("states", "Massachusetts", "Nebraska" );
	    dbCache.print("Massachusetts -> " + dbCache.get ("states", "Massachusetts"));
	    dbCache.print("Massachusetts" + (dbCache.contains ("states", "Massachusetts") ?  " exists!" : " NOT exist"));
	
	    try { Thread.sleep (4000); } catch (InterruptedException e) {}
	
	    dbCache.print("Massachusetts -> " + dbCache.get ("states", "Massachusetts"));
	    dbCache.print("Massachusetts" + (dbCache.contains ("states", "Massachusetts") ?  " exists!" : " NOT exist"));
	
	    dbCache.put ("states", "Kansas", "Kansas" );
	    dbCache.print("Kansas -> " + dbCache.get ("states", "Kansas"));
	    dbCache.print("Kansas" + (dbCache.contains ("states", "Kansas") ?  " exists!" : " NOT exist"));
	
	    dbCache.remove ("states", "Kansas");
	    dbCache.print("Kansas -> " + dbCache.get ("states", "Kansas"));
	    dbCache.print("Kansas" + (dbCache.contains ("states", "Kansas") ?  " exists!" : " NOT exist"));
	
	    dbCache.remove ("states", "Kansas");
	    dbCache.print("Kansas -> " + dbCache.get ("states", "Kansas"));
	    dbCache.print("Kansas" + (dbCache.contains ("states", "Kansas") ?  " exists!" : " NOT exist"));
	
	    dbCache.put ("states", "Kansas", "Montana" );
	    dbCache.print("Kansas -> " + dbCache.get ("states", "Kansas"));
	    dbCache.print("Kansas" + (dbCache.contains ("states", "Kansas") ?  " exists!" : " NOT exist"));
	
	    dbCache.put ("states", "Kansas", "Hawaii" );
	    dbCache.print("Kansas -> " + dbCache.get ("states", "Kansas"));
	    dbCache.print("Kansas" + (dbCache.contains ("states", "Kansas") ?  " exists!" : " NOT exist"));
	
	}
	catch (SQLException se)
	    {
		dbCache.print (StringUtil.exceptionStackTrace(se));
	    }
	catch (CacheNamingException cne)
	    {
		dbCache.print (StringUtil.exceptionStackTrace(cne));
	    }
	
	dbCache.expiration_thread.shutdown();
	dbCache.auto_save_thread.shutdown();
    }


    /**
     * log to the DBCache server log
     * @param lev the log level
     * @param s a string to be logged
     */
    public void log (int lev, String s)
    {
	if (server_log != null)
	    server_log.log (lev, s);
	else
	    System.out.println (s);
    }


    // for debugging...
    void print (String string)
    {
	log (9, ((System.currentTimeMillis()/1000) %
		 (maxExpirationTime + expirationSleepInterval + 5)) +
	     " ____> " + string);
    }
    
}

class RelationalObject1 implements RelationalObject
{
    boolean is_dirty = false;
    
    public boolean load (Connection conn, Object key) throws SQLException
    {
	System.out.println("RelationalObject1.load()!");
	return true;
    }

    public boolean update (Connection conn) throws SQLException
    {
	System.out.println("RelationalObject1.update()!");
	return true;
    }

    public boolean insert (Connection conn) throws SQLException
    {
	System.out.println("RelationalObject1.insert()!");
	return true;
    }

    public boolean delete (Connection conn) throws SQLException
    {
	System.out.println("RelationalObject1.delete()!");
	return true;
    }

    public boolean existsInDB()
    { return true; }

    public boolean isDirty ()
    {
	return is_dirty;
    }
    
    public void setDirty (boolean is_dirty)
    {
	this.is_dirty = is_dirty;
    }
}

/*******

In theory, the previously mentioned methods are pretty much all we
need in the DBCache class.  However, the following might be
useful methods too.  What do you all think?  Implementing them is
relatively easy using a HashtableOfHashtables.  But it might be
good to keep the API for the cache relatively simple, that's all.

    public boolean DBTableIsRegistered (String short_class_name)

    public Enumeration allRegisteredDBTableNames()

    public int numElementsInDBTableCache (String short_class_name)

    public Enumeration elementsOfDBTableCache (String short_class_name)

    public void setAutoSaveFrequency (String short_class_name, long number_of_seconds)

    public long getAutoSaveFrequency (String short_class_name)

    public void setExpirationTime (String short_class_name, long number_of_seconds)

    public long getExpirationTime (String short_class_name)

******/

/**
 * Not in use anymore, since it faster to use the DoubleHashtable than
 * using the TwoKey solution.
class TwoKey
{
    String class_name;
    Object key;
    
    public TwoKey (String class_name, Object key)
    {
	this.class_name = class_name;
	this.key = key;
    }
    
    public int hashCode()
    {
	return (class_name.hashCode() * 31 + key.hashCode());
    }
    
    public boolean equals (Object another_twokey)
    {
	TwoKey two_key = (TwoKey) another_twokey;
	return (class_name.equals(two_key.class_name) &&
		key.equals(two_key.key));
    }
}
*/

class MetaDataEntry
{
    String name;
    boolean is_relational_object;
    long auto_save_frequency;
    long expiration_frequency;
    Class my_class;

    public MetaDataEntry (String name, 
			  Class my_class,
			  boolean is_relational_object,
			  long auto_save_frequency,
			  long expiration_frequency)
    {
	this.name = name;
	this.my_class = my_class;
	this.is_relational_object = is_relational_object;
	this.auto_save_frequency = auto_save_frequency;
	this.expiration_frequency = expiration_frequency;
    }
}

class InstanceCacheEntry
{
    MetaDataEntry class_data;
    Object key;
    Object object;
    
    long auto_save_time;
    long expiration_time;
    
    // the following flag will allow the DBCache, the
    // ExpirationThread, and the AutoSaveThread to communicate to each
    // other that this InstanceCacheEntry is no longer in the
    // DBCache, so we should not auto-save it.
    boolean is_in_cache;

    public InstanceCacheEntry (MetaDataEntry class_data,
			       Object key,
			       Object Object)
    {
	this.class_data = class_data;
	this.key = key;
	this.object = Object;
	long current_time = System.currentTimeMillis()/1000;
	this.auto_save_time = current_time + class_data.auto_save_frequency;
	this.expiration_time = current_time + class_data.expiration_frequency;
	this.is_in_cache = true;
    }

    
    public String toString()
    {
	return (class_data.name + "[" + key + "] = " + object);
    }
    
}


/**
 * The ExpirationThread is an implementation of the AlarmThread which
 * works within the DBCache framework to automatically expire
 * objects from the DBCache.  Each object in the DBCache
 * has an expiration time which is registered with the
 * ExpirationThread.  When this expiration time triggers, that object
 * is expired in the ExpirationThread, and the ExpirationThread will
 * call DBCache.remove() on that object as well.
 * <P>
 * In the ExpirationThread or the DBCache, I don't actually
 * delete any alarm events or objects individually; instead, I simply
 * mark each alarm event as "deleted" if it is supposed to be deleted,
 * and, later, when I encounter these already-deleted entries, I
 * simply ignore them.  I delete entire multiple-second ranges of
 * alarm events at a time.  In implementation terms, I throw away lots
 * of Vectors.
 */
class ExpirationThread extends AlarmThread
{
    DBCache dbCache;

    public ExpirationThread(DBCache dbCache, 
			    int number_of_seconds_in_period,
			    int wake_up_interval,
			    int maximum_number_of_entries)
    {
	super (number_of_seconds_in_period, wake_up_interval,
	       maximum_number_of_entries);
	this.dbCache = dbCache;
    }
    
    
    public boolean hasBeenRemoved (Object object)
    {
	return (! ((InstanceCacheEntry) object).is_in_cache);
    }
    
    public long getAlarmTime (Object object)
    {
	return (((InstanceCacheEntry) object).expiration_time);
    }
    
    
    public void processAlarm (Object object)
    {
	InstanceCacheEntry entry = (InstanceCacheEntry) object;
	
	if (entry.is_in_cache)
	    {
		try {
		    dbCache.saveRelationalObjectIfDirty (entry);
		}
		catch (SQLException se) {
		    log(0, "ERROR: " + StringUtil.exceptionStackTrace(se));
		}
	    
		// now, let's remove this object from the DBCache.
		// (a nice side effect of this is that the AutoSaveThread
		// will also remove this object from the auto-save list if
		// entry.is_in_cache == false)
		try {
		    dbCache.remove (entry.class_data.name, entry.key);

		    //		print ("" + entry +
		    //		       " has been expired and removed from the cache.");
		}
		catch (SQLException se) {
		    log(0, "ERROR: " + StringUtil.exceptionStackTrace(se));
		}
		catch (CacheNamingException cne)
		    {
			log(0, "ERROR: " + StringUtil.exceptionStackTrace(cne));
		    }
	    
		// just to be extra sure:
		entry.is_in_cache = false;
	    }
	else
	    {
		//	    print (entry + " ALREADY removed. is_in_cache = false.");
	    }
    }
    
    
    /**
     * This method will add an object to the expirations list, to be
     * expired later. 
     */
    public void registerAlarm (Object object)
    {
	InstanceCacheEntry entry = (InstanceCacheEntry) object;
	
	// if this object still exists
	// in the DBCache, then we will expire it in the future.
	if (! entry.is_in_cache || entry.class_data.expiration_frequency == -1)
	    return;

	Vector future_alarm;
	
	synchronized (alarms_this_period)
	    {
		future_alarm = (Vector) alarms_this_period [
							    (int) (entry.expiration_time % SIZE_OF_BUFFER)];
	    }
	
	future_alarm.addElement (entry);
	//	print ("" + entry + " is added to alarms [" +
	//	          entry.expiration_time % SIZE_OF_BUFFER + "].");
	total_number_of_entries++;
    }
    
    /**
     * log to the DBCache server log
     * @param lev the log level
     * @param s a string to be logged
     */
    public void log (int lev, String s)
    {
	dbCache.log (lev, s);
    }
}


/**
 * The AutoSaveThread is an implementation of the AlarmThread which
 * works within the DBCache framework to automatically
 * "auto-save" objects.  In other words, it periodically checks
 * objects to see if they contain dirty data, and if they do, it
 * writes them out to the database.
 * <P>
 * Each object in the DBCache has an auto-save interval which is
 * registered with the ExpirationThread.  Whenever this auto-save time
 * triggers, that object is checked for the dirty data, and the
 * AutoSaveThread will call
 * DBCache.saveRelationalobjectIfDirty() on that object if there
 * are dirty data.
 * <P>
 * In the AutoSaveThread or the DBCache, I don't actually
 * delete any alarm events or objects individually; instead, I simply
 * mark each alarm event as "deleted" if it is supposed to be deleted,
 * and, later, when I encounter these already-deleted entries, I
 * simply ignore them.  I delete entire multiple-second ranges of
 * alarm events at a time.  In implementation terms, I throw away
 * big Vectors instead of throwing away many many little objects.
 */
class AutoSaveThread extends AlarmThread
{
    DBCache dbCache;
    public void print (String string)
    {
	// override the AlarmThread print() for now, so it's easier to
	// debug

	log (9, ((System.currentTimeMillis()/1000) % SIZE_OF_BUFFER) + 
	     " ====> " + string);
    }
    
    public AutoSaveThread (DBCache dbCache, 
			   int number_of_seconds_in_period,
			   int wake_up_interval,
			   int maximum_number_of_entries)
    {
	super (number_of_seconds_in_period, wake_up_interval,
	       maximum_number_of_entries);
	this.dbCache = dbCache;
    }
    
    
    public boolean hasBeenRemoved (Object object)
    {
	return (! ((InstanceCacheEntry) object).is_in_cache);
    }

    public long getAlarmTime (Object object)
    {
	return (((InstanceCacheEntry) object).auto_save_time);
    }


    public void processAlarm (Object object)
    {
	InstanceCacheEntry entry = (InstanceCacheEntry) object;

	if (entry.is_in_cache)
	    {
		try {
		    dbCache.saveRelationalObjectIfDirty (entry);
		}
		catch (SQLException se) {
		    log(0, "ERROR: " + StringUtil.exceptionStackTrace(se));
		}

		// If an entry either ALWAYS or NEVER needs to be auto-saved
		// (auto-save frequency == -1 or <= WAKE_UP_INTERVAL)
		// then we shouldn't register it again.
	    
		if (entry.class_data.auto_save_frequency > WAKE_UP_INTERVAL)
		    {
			entry.auto_save_time = System.currentTimeMillis()/1000 +
			    entry.class_data.auto_save_frequency;
			//		print ("reset auto-save time for " + entry + " to: " +
			//		          entry.auto_save_time % SIZE_OF_BUFFER + ".");
			registerAlarm (entry);
		    }
	    }
	else
	    {
		//	    print (entry + " has already been removed. is_in_cache = false.");
	    }
    }


    /**
     * This method will add an object to the auto-save list, to be
     * auto-saved later. 
     */
    public void registerAlarm (Object object)
    {
	InstanceCacheEntry entry = (InstanceCacheEntry) object;

	// if this object no longer exists in the DBCache, or if
	// it should never be auto-saved, then we should not register
	// it to be saved in the future.
	
	if (! entry.is_in_cache || entry.class_data.auto_save_frequency == -1)
	    return;

	if (entry.class_data.auto_save_frequency <= WAKE_UP_INTERVAL)
	    {
		always_ringing.addElement (entry);
		total_number_of_entries++;
		//	    print ("" + entry + " is added to always_ringing.");
		return;
	    }

	Vector future_alarm;

	synchronized (alarms_this_period)
	    {
		future_alarm = (Vector) alarms_this_period [
							    (int) (entry.auto_save_time % SIZE_OF_BUFFER)];
	    }

	future_alarm.addElement (entry);
	//	print ("" + entry + " is added to alarms [" +
	//	          entry.auto_save_time % SIZE_OF_BUFFER + "].");
	total_number_of_entries++;
    }


    /**
     * This method will reset the auto_save time to NOW + auto_save_frequency.
     * NOTE: it does NOT register an alarm event for this object!
     * That will be automatically done by the AutoSaveThread
     * whenever it encounters this object.
     */
    public void resetAutoSaveTime (Object object)
    {
	InstanceCacheEntry entry = (InstanceCacheEntry) object;

	// if this object still exists
	// in the DBCache, then we will auto-save it in the future.
	// However, if it should never be auto-saved, or it should ALWAYS
	// be autoSaved (in other words, if we need to check
	// it every time this thread wakes up), then it doesn't make sense to
	// reset the auto-save time.
	if (! entry.is_in_cache ||
	    entry.class_data.auto_save_frequency <= WAKE_UP_INTERVAL)
	    return;

	entry.auto_save_time = System.currentTimeMillis()/1000 +
	    entry.class_data.auto_save_frequency;
	
	//	print ("reset auto-save time for " + entry +
	//	          " to: " + entry.auto_save_time % SIZE_OF_BUFFER + ".");
    }

    /**
     * log to the DBCache server log
     * @param lev the log level
     * @param s a string to be logged
     */
    public void log (int lev, String s)
    {
	dbCache.log (lev, s);
    }
}
