/**
 * The RelationalObject interface is an interface which works with
 * the GlobalDBCache to define some standard methods for reading and
 * writing to and from the database.  If you are interested in using
 * the GlobalDBCache to cache objects which are stored persistently in
 * the database, your object must implement this interface in order
 * for the GlobalDBCache to be able to read and write it to the
 * database.
 */
package org.locomotive.util.cache;

import java.sql.*;

public interface RelationalObject
{
    /**
     * The GlobalDBCache calls this method to load your object
     * from the database.
     * NOTE: remember to set your dirty flag to false inside this method.
     * ALSO: remember to set your exists_in_DB flag inside this method
     * to true or false, depending out whether this object was found.
     * NOTE 2: even though Java does not allow us to specify whether a
     * method can be synchronized in an interface, we strongly recommend
     * that you implement this method as a synchronized method.
     * @return whether this RelationalObject was successfully loaded
     * (for example, it might not exist in the database).
     */
    public boolean load (Connection conn, Object key) throws SQLException;
    
    /**
     * The GlobalDBCache calls this method to write your object
     * to the database, if your object already exists in the database.
     * NOTE: remember to reset your dirty flag to false inside this method!
     * ALSO: remember to set your exists_in_DB flag to true inside this method.
     * NOTE 2: we recommend that you implement this method as a
     * synchronized method.
     * @return whether this object was written to the database.  (for
     * example, this object might not exist in the database)
     */
    public boolean update (Connection conn) throws SQLException;

    /**
     * The GlobalDBCache calls this method to create your new object
     * in the database, if your object DOES NOT already exist in the database.
     * NOTE: remember to reset your dirty flag to false inside this method!
     * ALSO: remember to set your exists_in_DB flag to true inside this method.
     * NOTE 2: we recommend that you implement this method as a
     * synchronized method.
     * @return whether this object was created in the database.
     */
    public boolean insert (Connection conn) throws SQLException;

    /**
     * The GlobalDBCache calls this method to delete your object
     * from the database.
     * @return whether this object was deleted from the database.
     * ALSO: remember to set your exists_in_DB flag to false inside this
     * method. (for example, this object might not exist in the database)
     */
    public boolean delete (Connection conn) throws SQLException;

    /**
     * returns the value of the flag which tells whether this object
     * already exists in the DB or not.
     */
    public boolean existsInDB();

    /**
     * @return the value of the dirty flag.
     * The "dirty" flag is used by the GlobalDBCache to determine
     * whether to write any changes to the database,
     * at the interval you schedule.
     */
    public boolean isDirty ();
}
