/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util.data;

import java.util.Enumeration;
import java.util.NoSuchElementException;



/**
 * this class is an abstract implementation of a data tree, and can be 
 * extended to for heaps or binary search trees.
 * it proivides implementations for two kinds of data structures:
 *     1. where there is a key and a data type
 *     2. where the key is the data type
 * All keys must implement the comparisonKey class
 * @see org.locomotive.util.data.ComparisonKey
 */
public class DataTree {
    
    protected TreeNode root;		// the root node of the tree
    protected int count;		// the number of items in the tree
    
    
    /**
     * construct an empty tree
     * sets number of items in the tree to zero
     */
    public DataTree() {
	count = 0;
    }
    
    /**
     * create a tree with the node as it's root
     * @param node the TreeNode to set the root as
     */
    public DataTree(TreeNode node) {
    	root = node;
	count = 1;
    }
    
    /**
     * constructs a new tree with a key and it's data as the root 
     * @param newKey the ComparisonKey to place at the root
     * @param newData the data to place in the root node
     */
    public DataTree(ComparisonKey newKey, Object newData) {
	this();
	root = new TreeNode(newKey, newData);
	count =1;		
    }
    
    /**
     * constructs a tree with the comparisonKey as the key and the data
     * for the root node.
     * @param newObject the ComparisonKey to place at the root
     */
    public DataTree(ComparisonKey newObject) {
	this();
	root = new TreeNode(newObject);
	count =1;		
    }
    
    /**
     * inserts a new object into tree
     * @param newKey the key to use for the object
     * @param newData the ComparisonKey to enter
     */
    public void insert (ComparisonKey newKey, Object newData) {
	count ++;
    }
    
    /**
     * inserts a node whose comparisonKey is its data
     */
    public void insert (ComparisonKey newKey) {
	count ++;
    }
    
    /**
     * returns true if the tree is empty, false if otherwise
     */
    public boolean isEmpty() {
	return (count == 0);
    }
    
    
    /** finds the object node containing the object with the 
     * specified key and returns it.  Returns null if key is
     * not in the tree
     */
    public TreeNode find(ComparisonKey key) {
	return null;
    }
    
    /**
     * returns the root node of the tree.  Returns null if tree is
     * empty
     */
    public TreeNode root() {
	return root;
    }
    
    /**
     * this method returns the number of elements in the tree
     */ 
    public int size() {
	return count;
    }
    


    /** a node for a binary search tree */
    protected class TreeNode {
    
	private ComparisonKey key;     // the key to reference node with
	private Object        data;    // the data in the node
	private TreeNode   leftChild;
	private TreeNode   rightChild;
	private TreeNode   parent;

	/** Constructor takes a key and it's corresponding data */
	public TreeNode( ComparisonKey key, Object data ) {
	    this.key = key;
	    this.data = data;
	}

	/** Constructor for Nodes with keys as data */
	public TreeNode(ComparisonKey data) {
	    this(data, null);
	}

	/** Convenience method for comparing the keys of two tree nodes */
	public int compare( TreeNode node ) 
	    throws ComparisonKeyException
	{
	    return( key.compare( node.key ) );
	}


	/** Access key */
	public ComparisonKey getKey() {
	    return key;
	}

	/** access data */
	public Object getData() {
	    if (data != null)
		return data;
	    else
		return key;
	}

	/** Access parent */
	public TreeNode getParent() {
	    return parent;
	}

	public TreeNode getLeftChild() {
	    return leftChild;
	}

	public TreeNode getRightChild() {
	    return rightChild;
	}

	/** returns true if node has no children */
	public boolean hasNoChildren() {
	    if ( (this.leftChild == null) && (this.rightChild == null) )
		return true;
	    else
		return false;
	}

	/** returns true if node is a leaf node (if node has < 2 children) */
	public boolean leafNode() {
	    if ( (this.leftChild == null) || (this.rightChild == null) )
		return true;
	    else
		return false;
	}

	/** returns the data contained in the node:  either node.data or
	 * node.key if node.data is null */
	public Object returnData() {
	    if (this.data != null)
		return this.data;
	    else return this.key;
	}
    
	/** Change key */
	public void setKey( ComparisonKey newKey ) {
	    this.key = newKey;
	}

	/** Change data */
	public void setData( Object newData ) {
	    this.data = newData;
	}

	/** Swap data with another node */
	public void swapData(TreeNode node) {
	    TreeNode tempNode = new TreeNode(node.getKey(), node.getData());
	    node.setKey(this.getKey());
	    node.setData(this.getData());
	    this.setKey(tempNode.getKey());
	    this.setData(tempNode.getData());
	}

	/** Set left child of this node and parent pointer of
	 * new left child.  */
	public void setLeftChild( TreeNode newChild ) {
	    if (newChild != null)
		newChild.parent = this;
	    leftChild = newChild;
	}

	/** Set right child of this node and parent pointer of
	* new right child.  */
	public void setRightChild( TreeNode newChild ) {
	    if (newChild != null)
		newChild.parent = this;
	    rightChild = newChild;
	}
    
	public String toString() {
	    return "key: " + key + "\tdata: " + data;
	}
    
    }  // end of TreeNode class




    
}  // end of DataTree class



