/*
 * The contents of this file are subject to the Locomotive Public License 
 * (LPL), a derivative of the Mozilla Public License, version 1.0. You 
 * may not use this file except in compliance with the License; You may 
 * obtain a copy of the License at http://www.locomotive.org/license/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The initial developers of this code under the LPL is Leverage Information 
 * Systems.  Portions created by Leverage are Copyright (C) 1998 Leverage 
 * Information Systems. All Rights reserved.
 */

package org.locomotive.util.data;


import java.util.Enumeration;
import java.util.NoSuchElementException;



/*============================== HEAP CLASS =============================*/
/**
 * this class implements a heap ADT.  it is useful for implementing
 * priority queues among other things.  It requires TreeNode class
 * and some implementation of the ComparisonKey class.  
 * an enumeration may be created from it by using the elements() method
 *
 * @see TreeNode
 * @see IntegerKey
 * @see DataTree
 */
public class Heap extends DataTree
{
 

    /**
     * this method inserts a ComparisonKey object into the heap in an
     * appropriate place
     *
     * @param newObject the key of the object to be inserted
     * @param newData the data held in the new node
     * PostCondition: the heap is reheapified
     */
    public void insert(ComparisonKey newKey, Object newData ) 
    {
	super.insert(newKey, newData);	 // count ++
	TreeNode newNode = new TreeNode(newKey, newData);
	TreeNode parent;
	if (count == 1) 
	    root = newNode;
	else 
	{
	    parent = getNode(count/2);
	    if (count%2 == 1)
		parent.setRightChild(newNode);
	    else
		parent.setLeftChild(newNode);
	}
				
	TreeNode compareNode = newNode;
	while (compareNode != root) 
	{
	    parent = compareNode.getParent();
	    if (compareNode.compare(parent) > 0) 
	    {
		compareNode.swapData(parent);
	    }
	    compareNode = parent;
	}								 
    }


    /**
     * this method inserts a ComparisonKey object into the heap in an
     * appropriate place.	Used when the data held is the key
     *
     * @param newObject the key of the object to be inserted
     * @param newData the data held in the new node
     * PostCondition: the heap is reheapified
     */
    public void insert(ComparisonKey newKey ) 
    {	
	insert(newKey, null);
    }	


    /**
     * this method removes the highest-valued element from the heap
     * 
     * @returns the highest valued element in the heap
     * 
     * @PreConsition: the heap is a well-formed
     * @PostCondition: the heap is reheapified
     */
    public Object remove() throws ComparisonKeyException 
    {
	Object returnData;
	if (count == 0)
	    return null;
	else 
	{
	    returnData = root.getData();
	    if(returnData == null)
		returnData = root.getKey();
	}
				
	if (count == 1) 
	{
	    root = null;
	    count --;
	    return returnData;
	}
	
	root.setData( (getNode(count)).getData() );
	root.setKey( (getNode(count)).getKey() );	
	
	if (count%2 == 1)
	    getNode(count/2).setRightChild(null);
	else
	    getNode(count/2).setLeftChild(null);
					
	TreeNode compareNode = root;
	// System.out.println("leafNode? " + compareNode.leafNode());
	while (!compareNode.hasNoChildren()) 
	{
	    TreeNode tempNode;
	    if (compareNode.getRightChild() != null) 
	    {
		if ((compareNode.getLeftChild()).compare(
		    compareNode.getRightChild() ) > 0 )
		    tempNode = compareNode.getLeftChild();
		else
		    tempNode = compareNode.getRightChild();
	    } else 
		tempNode = compareNode.getLeftChild();
	
	    if (compareNode.compare(tempNode) < 0) 
	    {
		compareNode.swapData(tempNode);
	    }
	    compareNode = tempNode;
	}
	if (count > 0)
	    count --;
	return returnData;		
    }



    /** Look at maximum element without removing it.   */
    public TreeNode peek() 
    {
	return root;
    }

		

    /** returns true if the heap is empty   */
    public boolean isEmpty() 
    {
	if (root == null)
	    return true;
	else
	    return false;
    }


    /**
     * Search for an object (obj) in the heap such that
     * obj.equals(key) is true.	Return a pointer to the object and 
     * make the node where it is found the current node (e.g., for 
     * a subsequent remove() call).
     * returns null if the object is not found
     * If more than one object would match, return the   *first  * such 
     * object in the heap.
     */
    public TreeNode find( ComparisonKey key ) 
    {
	TreeNode myNode;
	for (int i = 1; i<=count; i++) 
	{
	    myNode = getNode(i);
	    if ( (myNode.getKey()).compare(key) == 0) 
	    {
		return myNode;
	    }
	}
	return null;
    }


    /** returns the node at the specified index   */
    TreeNode getNode(int index) 
    {
	TreeNode parent = root;
	for (int i = lg(index) -1; i>=0; i--) 
	{
	    if ( ( (index/(int)Math.pow(2, i)) %2 ) == 1) 
	    {
		parent = parent.getRightChild();
	    } else 
	    {
		parent = parent.getLeftChild();
	    }
	}
	//System.out.println("getnode() " + 
	//		 ((IntegerKey)parent.getData()).toString() );
	return parent;
    }				 
		


    /** creates an enumeration of the heap   */
    public Enumeration elements() 
    {
	return new HeapIterator();
    }



    /** returns the floor of the log (base 2) of a number)   */
    static int lg(int value) 
    {
	int j = 0;
	for (int i = value; i>1; i = i/2, j++);
	return j;
    }



    /*===================== HEAP ITERATOR CLASS =====================  */
    /**
     * HeapIterator class creates a pointer to a node in a heap, thus
     * allowing one to access any item in a heap, not just the first
     */
    public class HeapIterator implements Enumeration 
    {
	TreeNode current;
	int index;
	
	/** Construct a new iterator on the heap and make the first node
	 * to be the "current" node	  */
	public HeapIterator() 
	{
	    current = root;
	    index = 1;
	}
	
	/**
	 * Return the object stored at the current node in the heap
	 * Should throw a NoSuchElementException if no current node.
	 */
	public Object peek() 
	{
	    return current.getData();
	}
	
	/** Move iterator to point to next node in heap and return the item
	 * Should throw a NoSuchElementException if already at end of heap
	 */
	public Object next() throws NoSuchElementException 
	{
	    if (getNode(count) == current )
		throw new NoSuchElementException (
		    "There are no more objects in this heap");
	    index ++;
	    current = getNode(index);
	    return current.getData();
	}
	
	/** Move iterator to point to previous node in heap and return the 
	 * item. Should throw a NoSuchElementException if already at 
	 * beginning of heap
	 */
	public Object pred() throws NoSuchElementException 
	{
	    if (current == root)
		throw new NoSuchElementException (
		    "You are already at the beginning of the heap");
	    index --;
	    current = getNode(index);
	    return current.getData();
	}
	
	
	/** The following two methods implement the enumeration interface
	 * Returns true if we have not yet reached the end of the heap
	 * false otherwise.
	 */
	public boolean hasMoreElements() 
	{ 
	    if (current == getNode(count))
		return false;
	    else 
		return true;
	}
	
	/** Return the next item in the heap	  */
	public Object nextElement() 
	{
	    index ++;
	    current = getNode(index);
	    return current;
	}
	
    } // end of HeapIterator class

}	// end of heap class

